Package: modsem
Type: Package
Title: Latent Interaction (and Moderation) Analysis in Structural
        Equation Models (SEM)
Version: 1.0.22
Authors@R: 
    c(person(given = "Kjell", family = "Solem Slupphaug", 
             email = "slupphaugkjell@gmail.com", role = c("aut", "cre"),
             comment = c(ORCID = "0009-0005-8324-2834")),
      person(given = "Mehmet", family="Mehmetoglu",
             email = "mehmetm@ntnu.no", role=c("ctb"),
             comment = c(ORCID="0000-0002-6092-8551")),
      person(given = "Matthias", family="Mittner",
             email = "matthias.mittner@uit.no", role=c("ctb"),
             comment = c(ORCID="0000-0003-0205-7353")))
Maintainer: Kjell Solem Slupphaug <slupphaugkjell@gmail.com>
Description: 
    Estimation of interaction (i.e., moderation) effects between latent variables
    in structural equation models (SEM). 
    The supported methods are:
      The constrained approach (Algina & Moulder, 2001).
      The unconstrained approach (Marsh et al., 2004).
      The residual centering approach (Little et al., 2006).
      The double centering approach (Lin et al., 2010).
      The latent moderated structural equations (LMS) approach (Klein & Moosbrugger, 2000).
      The quasi-maximum likelihood (QML) approach (Klein & Muthén, 2007)
    The constrained- unconstrained, residual- and double centering- approaches
    are estimated via 'lavaan' (Rosseel, 2012), whilst the LMS- and QML- approaches
    are estimated via 'modsem' it self. Alternatively model can be
    estimated via 'Mplus' (Muthén & Muthén, 1998-2017).
    References:
    Algina, J., & Moulder, B. C. (2001). 
      <doi:10.1207/S15328007SEM0801_3>.
      "A note on estimating the Jöreskog-Yang model for latent variable interaction using 'LISREL' 8.3."
    Klein, A., & Moosbrugger, H. (2000). 
      <doi:10.1007/BF02296338>.
      "Maximum likelihood estimation of latent interaction effects with the LMS method."
    Klein, A. G., & Muthén, B. O. (2007). 
      <doi:10.1080/00273170701710205>.
      "Quasi-maximum likelihood estimation of structural equation models with multiple interaction and quadratic effects."
    Lin, G. C., Wen, Z., Marsh, H. W., & Lin, H. S. (2010). 
      <doi:10.1080/10705511.2010.488999>.
      "Structural equation models of latent interactions: Clarification of orthogonalizing and double-mean-centering strategies."
    Little, T. D., Bovaird, J. A., & Widaman, K. F. (2006). 
      <doi:10.1207/s15328007sem1304_1>.
      "On the merits of orthogonalizing powered and product terms: Implications for modeling interactions among latent variables."
    Marsh, H. W., Wen, Z., & Hau, K. T. (2004). 
      <doi:10.1037/1082-989X.9.3.275>.
      "Structural equation models of latent interactions: evaluation of alternative estimation strategies and indicator construction."
    Muthén, L.K. and Muthén, B.O. (1998-2017).  
      "'Mplus' User’s Guide. Eighth Edition."
      <https://www.statmodel.com/>.
    Rosseel Y (2012). 
      <doi:10.18637/jss.v048.i02>.
      "'lavaan': An R Package for Structural Equation Modeling." 
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
LinkingTo: Rcpp, RcppArmadillo
Imports: Rcpp, purrr, stringr, lavaan, rlang, nlme, dplyr, mvnfast,
        stats, fastGHQuad, mvtnorm, ggplot2, parallel, plotly, Deriv,
        MASS, Amelia, grDevices, cli, RhpcBLASctl, memoise
Depends: R (>= 4.1.0)
URL: https://modsem.org, https://github.com/kss2k/modsem
Suggests: knitr, rmarkdown, ggpubr, RColorBrewer, MplusAutomation
VignetteBuilder: knitr
Config/roxygen2/version: 8.1.0
NeedsCompilation: yes
Packaged: 2026-08-20 12:57:39 UTC; kss
Author: Kjell Solem Slupphaug [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-8324-2834>),
  Mehmet Mehmetoglu [ctb] (ORCID:
    <https://orcid.org/0000-0002-6092-8551>),
  Matthias Mittner [ctb] (ORCID: <https://orcid.org/0000-0003-0205-7353>)
Repository: CRAN
Date/Publication: 2026-08-21 07:00:09 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.character,LavToken)
S3method(as.character,matrix)
S3method(as.data.frame,simple_slopes)
S3method(as.logical,matrix)
S3method(assignSubClass,LavClosure)
S3method(assignSubClass,LavMathExpr)
S3method(assignSubClass,LavName)
S3method(assignSubClass,LavNumeric)
S3method(assignSubClass,LavOperator)
S3method(assignSubClass,LavToken)
S3method(bootstrap_modsem,"function")
S3method(bootstrap_modsem,modsem_da)
S3method(bootstrap_modsem,modsem_pi)
S3method(centered_estimates,lavaan)
S3method(centered_estimates,modsem_da)
S3method(coef,modsem_da)
S3method(coef,modsem_mplus)
S3method(coef,modsem_pi)
S3method(coefficients,modsem_da)
S3method(coefficients,modsem_pi)
S3method(compare_fit,modsem_da)
S3method(compare_fit,modsem_pi)
S3method(estimate_h0,modsem_da)
S3method(estimate_h0,modsem_pi)
S3method(evalToken,LavAdd)
S3method(evalToken,LavBlank)
S3method(evalToken,LavComment)
S3method(evalToken,LavFunction)
S3method(evalToken,LavInteraction)
S3method(evalToken,LavModify)
S3method(evalToken,LavName)
S3method(evalToken,LavNumeric)
S3method(evalToken,LavOperator)
S3method(evalToken,LavSeperator)
S3method(evalToken,LavSubtract)
S3method(evalToken,LavToken)
S3method(evalToken,LeftBracket)
S3method(evalToken,RightBracket)
S3method(fitsToken,LavBlank)
S3method(fitsToken,LavClosure)
S3method(fitsToken,LavComment)
S3method(fitsToken,LavName)
S3method(fitsToken,LavNumeric)
S3method(fitsToken,LavOperator)
S3method(fitsToken,LavString)
S3method(is_interaction_model,modsem_da)
S3method(is_interaction_model,modsem_pi)
S3method(modsem_inspect,lavaan)
S3method(modsem_inspect,modsem_da)
S3method(modsem_inspect,modsem_pi)
S3method(modsem_predict,modsem_da)
S3method(modsem_predict,modsem_pi)
S3method(nobs,modsem_da)
S3method(nobs,modsem_mplus)
S3method(nobs,modsem_pi)
S3method(parameter_estimates,lavaan)
S3method(parameter_estimates,modsem_da)
S3method(parameter_estimates,modsem_mplus)
S3method(parameter_estimates,modsem_pi)
S3method(print,ModsemMatrix)
S3method(print,ModsemParTable)
S3method(print,ModsemRelcorr)
S3method(print,ModsemSymmetricMatrix)
S3method(print,ModsemVector)
S3method(print,modsem_da)
S3method(print,modsem_mplus)
S3method(print,modsem_partable_summary)
S3method(print,modsem_pi)
S3method(print,simple_slopes)
S3method(print,summary_da)
S3method(print,summary_modsem_pi)
S3method(print,summary_mplus)
S3method(standardized_estimates,data.frame)
S3method(standardized_estimates,lavaan)
S3method(standardized_estimates,modsem_da)
S3method(standardized_estimates,modsem_mplus)
S3method(standardized_estimates,modsem_pi)
S3method(summary,modsem_da)
S3method(summary,modsem_mplus)
S3method(summary,modsem_pi)
S3method(var_interactions,data.frame)
S3method(var_interactions,modsem_da)
S3method(var_interactions,modsem_mplus)
S3method(vcov,modsem_da)
S3method(vcov,modsem_mplus)
S3method(vcov,modsem_pi)
export(bootstrap_modsem)
export(centered_estimates)
export(colorize_output)
export(compare_fit)
export(default_settings_da)
export(default_settings_pi)
export(estimate_h0)
export(extract_lavaan)
export(fit_modsem_da)
export(get_pi_data)
export(get_pi_syntax)
export(is_interaction_model)
export(modsem)
export(modsem_coef)
export(modsem_da)
export(modsem_inspect)
export(modsem_mimpute)
export(modsem_mplus)
export(modsem_nobs)
export(modsem_pi)
export(modsem_predict)
export(modsem_vcov)
export(modsemify)
export(parameter_estimates)
export(plot_interaction)
export(plot_jn)
export(plot_surface)
export(relcorr_single_item)
export(set_modsem_colors)
export(simple_slopes)
export(standardize_model)
export(standardized_estimates)
export(summarize_partable)
export(trace_path)
export(twostep)
export(var_interactions)
importFrom(Rcpp,sourceCpp)
importFrom(memoise,memoise)
importFrom(stats,
  coef,
  coefficients,
  nobs,
  vcov
)
useDynLib(modsem, .registration = TRUE)

# `modsem` <img src="man/figures/modsem.png" alt="Logo" align = "right" height="139" class="logo">
<!-- badges: start -->
[![R-CMD-check](https://github.com/kss2k/modsem/actions/workflows/checks.yml/badge.svg)](https://github.com/kss2k/modsem/actions/workflows/checks.yml)
[![Tests](https://github.com/kss2k/modsem/actions/workflows/tests.yml/badge.svg)](https://github.com/kss2k/modsem/actions/workflows/tests.yml)
[![CRAN](https://www.r-pkg.org/badges/version/modsem)](https://cran.r-project.org/package=modsem)
[![PKGDOWN-Build](https://github.com/kss2k/modsem/actions/workflows/pkgdown.yml/badge.svg)](https://github.com/kss2k/modsem/actions/workflows/pkgdown.yml)
[![](https://cranlogs.r-pkg.org/badges/grand-total/modsem)](https://cran.r-project.org/package=modsem)
[![GitHub Clones](https://img.shields.io/badge/dynamic/json?color=success&label=Clone&query=count&url=https://gist.githubusercontent.com/Kss2k/131dd94e938508da4f039ec2a4ecb256/raw/clone.json&logo=github)](https://github.com/MShawon/github-clone-count-badge)

<!-- badges: end -->
`modsem` is an `R`-package for estimating interaction (i.e., moderation) effects between latent variables
in structural equation models (SEMs). See https://www.modsem.org for a tutorial.

# Installation
`modsem` is available on `CRAN` and `GitHub`, and can be installed as follows:

```R
# From CRAN 
install.packages("modsem")

# Latest version from GitHub
install.packages("remotes")
remotes::install_github("kss2k/modsem", build_vignettes = TRUE)
```
**Note**: The package needs to be compiled from source on `macOS` (if installing via `GitHub`) and `Linux`.
If you have issues installing the package on `macOS`, you might need to install the `gfortran` compiler.
A `C++` compiler is also required, but should be installed by default on most systems.
See the [R for macOs](https://cran.r-project.org/bin/macosx/tools/) page for more information.

If you're using `Windows`, consider installing [`OpenBLAS in R for Windows`](https://github.com/david-cortes/R-openblas-in-windows) 
for better perfmance. If you're using a `Linux` distribution, consider installing
the [`ropenblas` package](https://CRAN.R-project.org/package=ropenblas)

# Methods/Approaches

There are a number of approaches for estimating interaction effects in SEM. 
In `modsem()`, the `method = "method"` argument allows you to choose which to use.
Different approaches can be categorized into two groups: 
Product Indicator (PI) and Distribution Analytic (DA) approaches.

## Product Indicator (PI) Approaches:
- `"ca"` = constrained approach (Algina & Moulder, 2001)
    - Note that constraints can become quite complicated for complex models, 
      particularly when there is an interaction including enodgenous variables.
      The method can therefore be quite slow. 
- `"uca"` = unconstrained approach (Marsh, 2004)
- `"rca"` = residual centering approach (Little et al., 2006)
- `"dblcent"` = double centering approach (Marsh., 2013)
  - default 
- `"pind"` = basic product indicator approach (not recommended)

## Distribution Analytic (DA) Approaches
- `"lms"` = The Latent Moderated Structural equations (LMS) approach, see the [vignette](https://modsem.org/articles/lms_qml.html)
- `"qml"` = The Quasi Maximum Likelihood (QML) approach, see the [vignette](https://modsem.org/articles/lms_qml.html)
- `"mplus"` = `Mplus`
  - estimates model through `Mplus`, if it is installed

# Examples 

## Elementary Interaction Model (Kenny & Judd, 1984; Jaccard & Wan, 1995)
```R
library(modsem)

m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3
  
  # Inner model
  Y ~ X + Z + X:Z 
'

# Double centering approach
est1_dca <- modsem(m1, oneInt)
summary(est1_dca)

# Constrained approach
est1_ca <- modsem(m1, oneInt, method = "ca")
summary(est1_ca)

# QML approach 
est1_qml <- modsem(m1, oneInt, method = "qml")
summary(est1_qml, standardized = TRUE) 

# LMS approach 
est1_lms <- modsem(m1, oneInt, method = "lms") 
summary(est1_lms)
```

## Theory Of Planned Behavior
```R
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

# double centering approach
est_tpb_dca <- modsem(tpb, data = TPB, method = "dblcent")
summary(est_tpb_dca)

# Constrained approach using Wrigths path tracing rules for generating
# the appropriate constraints
est_tpb_ca <- modsem(tpb, data = TPB, method = "ca") 
summary(est_tpb_ca)

# LMS approach 
est_tpb_lms <- modsem(tpb, data = TPB, method = "lms")
summary(est_tpb_lms, standardized = TRUE) 

# QML approach 
est_tpb_qml <- modsem(tpb, data = TPB, method = "qml") 
summary(est_tpb_qml, standardized = TRUE)
```
## Interactions between two observed variables
```R
est2 <- modsem('y1 ~ x1 + z1 + x1:z1', data = oneInt, method = "dblcent")
summary(est2)
```

## Interaction between an obsereved and a latent variable 
```R
m3 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  
  # Inner model
  Y ~ X + z1 + X:z1
'

est3 <- modsem(m3, oneInt, method = "dblcent", 
               res.cov.method = "none") # res.cov.method = "simple" will lead
                                        # to an unidentifiable model. Instead we
                                        # constrain them to zero
summary(est3)
```

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)

## ----setup--------------------------------------------------------------------
# library(modsem)

## ----eval = FALSE-------------------------------------------------------------
# 
# m1 <- '
# # Outer Model
# X =~ x1 + x2 + x3
# Y =~ y1 + y2 + y3
# Z =~ z1 + z2 + z3
# 
# # Inner model
# Y ~ X + Z + X:Z
# '
# 
# # Product Indicator Approaches
# modsem(m1, data = oneInt, method = "ca")
# modsem(m1, data = oneInt, method = "uca")
# modsem(m1, data = oneInt, method = "rca")
# modsem(m1, data = oneInt, method = "dblcent")
# 
# # Distribution Analytic Approaches
# modsem(m1, data = oneInt, method = "mplus")
# modsem(m1, data = oneInt, method = "lms")
# modsem(m1, data = oneInt, method = "qml")

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)

## ----setup--------------------------------------------------------------------
library(modsem)

## -----------------------------------------------------------------------------
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)
plot_interaction(x = "X", z = "Z", y = "Y",
                 vals_z = -3:3, model = est1)

## -----------------------------------------------------------------------------
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "lms")
plot_interaction(x = "INT", z = "PBC", y = "BEH",
                 vals_z = c(-0.5, 0.5), model = est2)

## -----------------------------------------------------------------------------
m1 <-  '
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9

  visual ~ speed + textual + speed:textual
'

est1 <- modsem(m1, data = lavaan::HolzingerSwineford1939, method = "ca")
plot_jn(x = "speed", z = "textual", y = "visual", model = est1, max_z = 6)

## -----------------------------------------------------------------------------
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "qml")
plot_jn(x = "INT", z = "PBC", y = "BEH", model = est2,
        min_z = -1.5, max_z = -0.5)

## ----eval = EVAL_DEFAULT------------------------------------------------------
# tpb <- "
# # Outer Model (Based on Hagger et al., 2007)
#   ATT =~ att1 + att2 + att3 + att4 + att5
#   SN =~ sn1 + sn2
#   PBC =~ pbc1 + pbc2 + pbc3
#   INT =~ int1 + int2 + int3
#   BEH =~ b1 + b2
# 
# # Inner Model (Based on Steinmetz et al., 2011)
#   INT ~ ATT + SN + PBC
#   BEH ~ INT + PBC
#   BEH ~ PBC:INT
# "
# 
# est2 <- modsem(tpb, TPB, method = "qml")
# plot_surface(x = "INT", z = "PBC", y = "BEH", model = est2)

## -----------------------------------------------------------------------------
set.seed(2934269)

var_X      <- 1
var_Z      <- 1
cov_X_Z    <- 0.2

zeta_Y     <- 0.4

gamma_Y_X  <-  0
gamma_Y_Z  <-  0
gamma_Y_XZ <-  1
gamma_Y_ZZ <-  3 # exclude for now
gamma_Y_XX <- -1 # exclude for now


lambda_1   <- 1
lambda_2   <- .7
lambda_3   <- .8


epsilon    <- 0.2
beta_1     <- 1.2
beta_2     <- 0.8
beta_3     <- 1.5
N          <- 1500

residual <- \(epsilon) rnorm(N, sd = sqrt(epsilon))
create_ind <- \(lv, beta, lambda, epsilon) beta + lambda * lv + residual(epsilon)


Phi <- matrix(c(var_X, cov_X_Z,
                cov_X_Z, var_Z), nrow = 2)
XI <- mvtnorm::rmvnorm(N, sigma = Phi)

X <- XI[, 1]
Z <- XI[, 2]

Y <-
  gamma_Y_X * X +
  gamma_Y_Z * Z +
  gamma_Y_XZ * X * Z +
  gamma_Y_XX * X * X +
  gamma_Y_ZZ * Z * Z +
  residual(zeta_Y)

x1 <- create_ind(X, beta_1, lambda_1, epsilon)
x2 <- create_ind(X, beta_2, lambda_2, epsilon)
x3 <- create_ind(X, beta_3, lambda_2, epsilon)

z1 <- create_ind(Z, beta_1, lambda_1, epsilon)
z2 <- create_ind(Z, beta_2, lambda_2, epsilon)
z3 <- create_ind(Z, beta_3, lambda_2, epsilon)

y1 <- create_ind(Y, beta_1, lambda_1, epsilon)
y2 <- create_ind(Y, beta_2, lambda_2, epsilon)
y3 <- create_ind(Y, beta_2, lambda_2, epsilon)


data.sim <- data.frame(
   x1, x2, x3,
   z1, z2, z3,
   y1, y2, y3
)

## ----eval = EVAL_DEFAULT------------------------------------------------------
# model <- '
#   X =~ x1 + x2 + x3
#   Z =~ z1 + z2 + z3
#   Y =~ y1 + y2 + y3
# 
#   Y ~ X + Z + X:X + Z:Z + X:Z
# '
# 
# est3 <- modsem(model, data = data.sim, method = "qml")

## ----eval = EVAL_DEFAULT------------------------------------------------------
# plot_interaction(x = "X", z = "Z", y = "Y", model = est3, vals_z = c(-1, 0, 1))

## ----eval = EVAL_DEFAULT------------------------------------------------------
# plot_surface(x = "X", z = "Z", y = "Y", model = est3)

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- TRUE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)

## ----setup--------------------------------------------------------------------
library(modsem)

## -----------------------------------------------------------------------------
library(modsem)

fit <- modsem('y1 ~ x1 + z1 + x1:z1', data = oneInt, method = "lms")
summary(fit, standardized = TRUE)

## -----------------------------------------------------------------------------
# Interaction effect between a latent and an observed variable
m1 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 + z2 # Z is latent
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ Z:X
'

lms1 <- modsem(m1, oneInt, method = "lms")

# Interaction effect between two observed variables
m2 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed
  x1 ~~ 0.1 * x1 # Specify a variance for the measurement error

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

lms2 <- modsem(m2, oneInt, method = "lms")
summary(lms2)

## ----error = TRUE-------------------------------------------------------------
try({
m2 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

lms3 <- modsem(m2, oneInt, method = "lms")
})

## -----------------------------------------------------------------------------
m3 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

qml3 <- modsem(m3, oneInt, method = "qml")
summary(qml3)

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)

## ----setup--------------------------------------------------------------------
# library(modsem)

## -----------------------------------------------------------------------------
# tpb_2so <- '
#   # First order latent variables
#   ATT =~ att1 + att2 + att3
#   SN  =~ sn1 + sn2 + sn3
#   PB =~ pb1 + pb2 + pb3
#   PC =~ pc1 + pc2 + pc3
#   BEH =~ b1 + b2
# 
#   # Higher order latent variables
#   INT =~ ATT + SN
#   PBC =~ PC + PB
# 
#   # Structural model
#   BEH ~ PBC + INT + INT:PBC
# '
# 
# est_lms_2so <- modsem(tpb_2so, data = TPB_2SO, method = "lms")
# summary(est_lms_2so)

## -----------------------------------------------------------------------------
# tpb_1so <- '
#   # First order latent variables
#   ATT =~ att1 + att2 + att3
#   SN  =~ sn1 + sn2 + sn3
#   PBC =~ pbc1 + pbc2 + pbc3
#   BEH =~ b1 + b2
# 
#   # Higher order latent variables
#   INT =~ ATT + PBC + SN
# 
#   # Structural model
#   BEH ~ PBC + INT + INT:PBC
# '
# 
# est_lms_1so <- modsem(tpb_1so, data = TPB_1SO, method = "lms", nodes = 32)
# summary(est_lms_1so)

## -----------------------------------------------------------------------------
# tpb_2so <- '
#   # First order latent variables
#   ATT =~ att1 + att2 + att3
#   SN  =~ sn1 + sn2 + sn3
#   PB =~ pb1 + pb2 + pb3
#   PC =~ pc1 + pc2 + pc3
#   BEH =~ b1 + b2
# 
#   # Higher order latent variables
#   INT =~ ATT + SN
#   PBC =~ PC + PB
# 
#   # Higher order interaction
#   INTxPBC =~ ATT:PC + ATT:PB + SN:PC + SN:PB
# 
#   # Structural model
#   BEH ~ PBC + INT + INTxPBC
# '
# 
# est_ca <- modsem(tpb_2so, data = TPB_2SO, method = "ca")
# summary(est_ca)
# 
# est_dblcent <- modsem(tpb_2so, data = TPB_2SO, method = "dblcent")
# summary(est_dblcent)

## -----------------------------------------------------------------------------
# tpb_1so <- '
#   # First order latent variables
#   ATT =~ att1 + att2 + att3
#   SN  =~ sn1 + sn2 + sn3
#   PBC =~ pbc1 + pbc2 + pbc3
#   BEH =~ b1 + b2
# 
#   # Higher order latent variables
#   INT =~ ATT + PBC + SN
# 
#   # Higher order interaction
#   INTxPBC =~ ATT:PBC + SN:PBC + PBC:PBC
# 
#   # Structural model
#   BEH ~ PBC + INT + INTxPBC
# '
# 
# est_ca <- modsem(tpb_1so, data = TPB_1SO, method = "ca")
# summary(est_ca)
# 
# est_dblcent  <- modsem(tpb_1so, data = TPB_1SO, method = "dblcent")
# summary(est_dblcent)

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)

## ----setup--------------------------------------------------------------------
# library(modsem)

## -----------------------------------------------------------------------------
# library(modsem)
# m1 <- '
# # Outer Model
#   X =~ x1 + x2 + x3
#   Z =~ z1 + z2 + z3
#   Y =~ y1 + y2 + y3
# 
# # Inner Model
#   Y ~ X + Z + X:Z
# '
# 
# lms1 <- modsem(m1, oneInt, method = "lms")
# # Standardized estimates
# summary(lms1, standardized = TRUE)

## -----------------------------------------------------------------------------
# qml1 <- modsem(m1, oneInt, method = "qml", robust.se = TRUE) # get robust standard errors
# summary(qml1, fit = TRUE, adjusted = TRUE) # get additional fit indices

## -----------------------------------------------------------------------------
# # ATT = Attitude
# # PBC = Perceived Behavioral Control
# # INT = Intention
# # SN = Subjective Norms
# # BEH = Behavior
# tpb <- '
# # Outer Model (Based on Hagger et al., 2007)
#   ATT =~ att1 + att2 + att3 + att4 + att5
#   SN =~ sn1 + sn2
#   PBC =~ pbc1 + pbc2 + pbc3
#   INT =~ int1 + int2 + int3
#   BEH =~ b1 + b2
# 
# # Inner Model (Based on Steinmetz et al., 2011)
#   INT ~ ATT + SN + PBC
#   BEH ~ INT + PBC
#   BEH ~ INT:PBC
# '
# 
# lms2 <- modsem(tpb, TPB, method = "lms", nodes = 32)
# summary(lms2)
# 
# qml2 <- modsem(tpb, TPB, method = "qml")
# summary(qml2, standardized = TRUE) # Standardized estimates

## ----include = FALSE----------------------------------------------------------
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)

## ----setup--------------------------------------------------------------------
# library(modsem)

## -----------------------------------------------------------------------------
# tpb_uk <- "
# # Outer Model (Based on Hagger et al., 2007)
#   ATT =~ att3 + att2 + att1 + att4
#   SN =~ sn4 + sn2 + sn3 + sn1
#   PBC =~ pbc2 + pbc1 + pbc3 + pbc4
#   INT =~ int2 + int1 + int3 + int4
#   BEH =~ beh3 + beh2 + beh1 + beh4
# 
# # Inner Model (Based on Steinmetz et al., 2011)
#   INT ~ ATT + SN + PBC
#   BEH ~ INT + PBC
#   BEH ~ INT:PBC
# "
# 
# fit <- modsem(tpb_uk,
#               data = TPB_UK,
#               method = "lms",
#               nodes = 32, # Number of nodes for numerical integration
#               adaptive.quad = TRUE, # Use quasi-adaptive quadrature
#               adaptive.frequency = 3, # Update the quasi-adaptive quadrature every third EM-iteration
#               adaptive.quad.tol = 1e-12, # Relative error tolerance for quasi-adaptive quad
#               algorithm ="EMA", # Use accelerated EM algorithm (Default)
#               convergence.abs = 1e-4, # Relative convergence criterion
#               convergence.rel = 1e-10, # Relative convergence criterion
#               max.iter = 500, # Maximum number of iterations
#               max.step = 1) # Maximum number of steps in the maximization step
# summary(fit)

# Functitions for specifying constraints in the constrained approach


labelFactorLoadings <- function(parTable) {
  # Firstly Label factor loadings in Partable
  loadingLabels <- apply(parTable[parTable$op == "=~", c("rhs", "lhs")],
                         MARGIN = 1,
                         FUN = function(x) createLabelLambda(x[1], x[2]))
  parTable[parTable$op == "=~", "mod"] <- loadingLabels
  parTable
}


specifyFactorLoadingsSingle <- function(parTable, relDf) {
  latentProdName <- stringr::str_c(rownames(relDf), collapse = "")

  for (indProd in colnames(relDf)) {
    indProdLabel     <- createLabelLambda(indProd, latentProdName)
    indsInProdLabels <- createLabelLambda(relDf[[indProd]], rownames(relDf))
    vecLhsRhs        <- c(indProdLabel, stringr::str_c(indsInProdLabels,
                                                       collapse = " * "))

    newRow   <- createParTableRow(vecLhsRhs, op = "==")
    parTable <- rbind(parTable, newRow)
  }

  parTable
}


specifyFactorLoadings <- function(parTable, relDfs) {
  parTable <- specifyFactorLoadingsSingle(parTable, relDfs[[1]])
  if (length(relDfs) <= 1) return(parTable)
  specifyFactorLoadings(parTable, relDfs[-1])
}


addVariances <- function(pt, exclude = character(0)) {
  # Add variance-labels if missing for LVs
  LVs <- setdiff(getLVs(pt), exclude)
  if (length(LVs) == 0) return(pt)

  specifiedLVs <- pt[pt$lhs %in% LVs &
                     pt$op == "~~" &
                     pt$rhs %in% LVs &
                     pt$lhs == pt$rhs, "lhs"] |> unique()
  toBeSpecifiedLvs <- LVs[!LVs %in% specifiedLVs]

  newRows <- lapply(toBeSpecifiedLvs, FUN = function(x)
                    createParTableRow(c(x, x), op = "~~")) |>
    purrr::list_rbind()
  pt <- rbind(pt, newRows)

  # Add variance for observed variables if missing
  observed <- getIndicators(pt, observed = TRUE)

  specifiedOvs <- pt[pt$lhs %in% observed &
                     pt$op == "~~" &
                     pt$rhs %in% observed &
                     pt$lhs == pt$rhs, "lhs"] |> unique()
  toBeSpecifiedOvs <- observed[!observed %in% specifiedOvs]

  newRows <- lapply(toBeSpecifiedOvs,
                    FUN = function(x) createParTableRow(c(x, x), op = "~~")) |>
    purrr::list_rbind()
  rbind(pt, newRows)
}


addCovariances <- function(pt) {
  # Add covariances for exogenous variables if missing
  pt  <- stripColonsParTable(pt)
  xis <- getXis(pt, isLV = FALSE, checkAny = FALSE)

  if (!length(xis))
    return(pt)

  indicators   <- unique(pt[pt$op == "=~", "rhs"])
  isLowerOrder <- xis %in% indicators

  if (all(isLowerOrder))
    return(pt)

  xis <- xis[!isLowerOrder] # we don't want to add any residual covariances
                            # only full covariances

  combos <- getUniqueCombos(xis)

  isSpecified <- \(x, y) any(pt$op == "~~" & ((pt$rhs == x & pt$lhs == y) |
                                              (pt$rhs == y & pt$lhs == x)))

  for (i in seq_len(NROW(combos))) {
    xi1 <- combos[i, "V1"]
    xi2 <- combos[i, "V2"]

    if (isSpecified(x = xi1, y = xi2))
      next

    newRow <- createParTableRow(c(xi1, xi2), op = "~~")
    pt <- rbind(pt, newRow)
  }

  pt
}


labelParameters <- function(pt) {
  latents <- getLVs(pt)
  endogenous <- latents[latents %in% pt[pt$op == "~", "lhs"]]
  exogenous <- latents[!latents %in% endogenous]
  observed <- getOVs(pt)

  # Gamma
  pt[pt$op == "~", "mod"] <-
    apply(pt[pt$op == "~", c("rhs", "lhs")],
          MARGIN = 1, FUN = function(x)
            createLabelGamma(x[[1]], x[[2]]))

  # Variances of exogenous
  pt[pt$op == "~~" & pt$lhs == pt$rhs & pt$lhs %in% exogenous, "mod"] <-
    vapply(pt[pt$op == "~~" & pt$lhs == pt$rhs &
              pt$lhs %in% exogenous, "lhs"],
           FUN.VALUE = vector("character", length = 1L),
           FUN = createLabelVar)

  # Variances of endogenous
  pt[pt$op == "~~" & pt$lhs == pt$rhs & pt$lhs %in% endogenous, "mod"] <-
    vapply(pt[pt$op == "~~" & pt$lhs == pt$rhs &
              pt$lhs %in% endogenous, "lhs"],
           FUN.VALUE = vector("character", length = 1L),
           FUN = createLabelZeta)

  # Variance of Observed Variables
  pt[pt$op == "~~" & pt$rhs %in% observed & pt$lhs == pt$rhs, "mod"] <-
    vapply(pt[pt$op == "~~" & pt$rhs %in% observed &
           pt$lhs == pt$rhs, "rhs"],
           FUN.VALUE = vector("character", length = 1L),
           FUN = createLabelVar)

  # Covariances
  pt[pt$op == "~~" & pt$lhs != pt$rhs, "mod"] <-
    apply(pt[pt$op == "~~" & pt$lhs != pt$rhs, c("lhs", "rhs")],
          MARGIN = 1, FUN = function(x)
            createLabelCov(x[[1]], x[[2]]))

  pt
}


specifyVarCovSingle <- function(parTable, relDf) {
  # This function specifies variances for latents, indicators,
  # and indicator products. It will also specifies covariances for latent
  # products, and elements int those products.
  mod_stopif(nrow(relDf) > 2, paste0("Constraints for products with more than two ",
         " elements are not supported for this method"))

  # General info
  elemsInProdTerm <- rownames(relDf)
  latentProd <- stringr::str_c(rownames(relDf), collapse = "")

  # Variance of latent product
  labelLatentProd <- createLabelVar(latentProd)
  labelsElemsInProd <- vapply(elemsInProdTerm,
                              FUN.VALUE = vector("character", length = 1L),
                              FUN = function(x) trace_path(parTable, x, x))

  labelCovElems <- trace_path(parTable, elemsInProdTerm[[1]],
                              elemsInProdTerm[[2]]) |> paste0(" ^ 2")

  lhs <- labelLatentProd
  rhs <- paste(stringr::str_c(labelsElemsInProd, collapse = " * "),
               labelCovElems, sep = " + ")
  varLatentProd <- createParTableRow(c(lhs, rhs), op = "==")

  # covariances between elems and latents
  labelsCovElemProd <-
    vapply(elemsInProdTerm, FUN.VALUE = vector("character", length = 1L),
           FUN = function(elem) createLabelCov(elem, latentProd)) # wrap in anonymous fun
                                                             # scope latentProd

  labelsCovElemProd <- labelsCovElemProd[labelsCovElemProd %in% parTable$mod]

  if (length(labelsCovElemProd)) { # should not be added in higher order models
    covsElemsProd <- lapply(labelsCovElemProd, FUN = function(x)
                              createParTableRow(c(x, "0"), op = "==")) |>
      purrr::list_rbind()
  } else covsElemsProd <- NULL

  # Variances of product indicators
  constrained.varProdInds <- vector("list", length = ncol(relDf))

  for (indProd in colnames(relDf)) {
    labelVarIndProd <- createLabelVar(indProd)

    labelsFactorLoadings <- vector("character", length = nrow(relDf))
    labelsVarLatents     <- vector("character", length = nrow(relDf))
    labelsVarInds        <- vector("character", length = nrow(relDf))

    for (latent in seq_len(nrow(relDf))) {
      labelsFactorLoadings[[latent]] <-
        createLabelLambdaSquared(relDf[latent, indProd],
                                 rownames(relDf)[[latent]])
      labelsVarLatents[[latent]] <-
        trace_path(parTable, rownames(relDf)[[latent]],
                   rownames(relDf)[[latent]])
      labelsVarInds[[latent]] <- createLabelVar(relDf[latent, indProd])
    }

    lhs  <- labelVarIndProd
    rhs1 <- paste(labelsFactorLoadings[[1]], labelsVarLatents[[1]],
                  labelsVarInds[[2]], sep = " * ")
    rhs2 <- paste(labelsFactorLoadings[[2]], labelsVarLatents[[2]],
                  labelsVarInds[[1]], sep = " * ")
    rhs3 <- paste(labelsVarInds[[1]], labelsVarInds[[2]], sep = " * ")
    rhs  <- paste(rhs1, rhs2, rhs3, sep = " + ")

    constrained.varProdInds[[indProd]] <- createParTableRow(c(lhs, rhs), op = "==")
  }

  constrained.varProdInds <- purrr::list_rbind(constrained.varProdInds)
  rbindParTable(parTable, rbind(varLatentProd,
                                covsElemsProd,
                                constrained.varProdInds))
}


specifyVarCov <- function(parTable, relDfs) {
  parTable <- specifyVarCovSingle(parTable, relDfs[[1]])
  if (length(relDfs) <= 1) return(parTable)
  specifyVarCov(parTable, relDfs[-1])
}

combineLavLabels <- function(lavLabelsCov, lavLabelsMain, currentLabels, g = 1L) {
  lavLabels <- c(lavLabelsCov, lavLabelsMain)

  if (g > 1L)
    lavLabels <- stats::setNames(sprintf("%s.g%d", lavLabels, g),
                                 nm = names(lavLabels))

  finalLabels <- currentLabels
  finalLabels[finalLabels %in% names(lavLabels)] <-
    lavLabels[names(lavLabels) %in% finalLabels]
  finalLabels
}


createLavLabels <- function(matrices, subset, etas, parTable.in = NULL) {
  lambdaX      <- createLabelsMatrix(matrices$lambdaX, op = "=~")
  lambdaY      <- createLabelsMatrix(matrices$lambdaY, op = "=~")
  thetaDelta   <- createLabelsMatrix(matrices$thetaDelta, op = "~~")
  thetaEpsilon <- createLabelsMatrix(matrices$thetaEpsilon, op = "~~")
  W            <- createLabelsMatrix(matrices$W, op = "<~")
  T            <- createLabelsMatrix(matrices$T, op = "~~")
  phi          <- createLabelsMatrix(matrices$phi, op = "~~")
  A            <- createLabelsMatrix(matrices$A, op = "~~")
  covZetaXi    <- createLabelsMatrix(matrices$covZetaXi, op = "~~", first = "cols")
  psi          <- createLabelsMatrix(matrices$psi, op = "~~")
  tauX         <- createLabelsMatrix(matrices$tauX, op = "~", first = "rows")
  tauY         <- createLabelsMatrix(matrices$tauY, op = "~", first = "rows")
  alpha        <- createLabelsMatrix(matrices$alpha, op = "~", first = "rows")
  beta0        <- createLabelsMatrix(matrices$beta0, op = "~", first = "rows")
  gammaXi      <- createLabelsMatrix(matrices$gammaXi, op = "~", first = "rows")
  gammaEta     <- createLabelsMatrix(matrices$gammaEta, op = "~", first = "rows")
  omegaXiXi    <- createLabelsOmega(matrices$omegaXiXi, parTable.in = parTable.in)
  omegaEtaXi   <- createLabelsOmega(matrices$omegaEtaXi, parTable.in = parTable.in)

  labels <- c(
    lambdaX = lambdaX,
    lambdaY = lambdaY,
    tauX = tauX,
    tauY = tauY,
    thetaDelta = thetaDelta,
    thetaEpsilon = thetaEpsilon,
    W = W,
    T = T,
    phi = phi,
    A = A,
    covZetaXi = covZetaXi,
    psi = psi,
    alpha = alpha,
    beta0 = beta0,
    gammaXi = gammaXi,
    gammaEta = gammaEta,
    omegaXiXi = omegaXiXi,
    omegaEtaXi = omegaEtaXi
  )

  labels[subset]
}


createLavLabelsCov <- function(matrices, subset) {
  if (is.null(matrices)) return(NULL)

  psi   <- createLabelsMatrix(matrices$psi, op = "~~")
  gamma <- createLabelsMatrix(matrices$gamma, op = "~", first = "rows")

  labels <- c(psi = psi, gamma = gamma)
  labels[subset]
}


createLabelsMatrix <- function(X, op = "~", first = "cols") {
  labels <- character(0L)
  rows   <- rownames(X)
  cols   <- colnames(X)

  # this is ugly, but... we have to read by cols first
  getLabel <- switch(first, cols = function(col, row) paste0(col, op, row),
                     rows = function(col, row) paste0(row, op, col))

  for (i in seq_len(NCOL(X))) for (j in seq_len(NROW(X))) {
    labels <- c(labels, getLabel(col = cols[[i]], row = rows[[j]]))
  }

  labels
}


createLabelsOmega <- function(X, parTable.in = NULL) {
  C <- \(x, y) sprintf("%s:%s", x, y)
  getIntTerm <- function(lhs, rhs) {
    xz <- C(lhs, rhs)
    zx <- C(rhs, lhs)

    if (is.null(parTable.in))         xz
    else if (xz %in% parTable.in$rhs) xz
    else                              zx
  }

  rows   <- rownames(X)
  cols   <- colnames(X)
  labels <- character(0L)

  for (i in seq_len(ncol(X))) for (j in seq_len(nrow(X))) {
    eta_i <- stringr::str_split_i(rows[[j]], pattern = "~", i = 1) # y~x -> y
    lhs_i <- stringr::str_split_i(rows[[j]], pattern = "~", i = 2) # y~x -> x
    rhs_i <- cols[[i]]
    xz    <- getIntTerm(lhs_i, rhs_i)

    labels <- c(labels, sprintf("%s~%s", eta_i, xz))
  }

  labels
}


getLavCoefs <- function(model, theta, method) {
  fullTheta <- getTransformationsTheta(model, theta, method)
  fullNames  <- names(fullTheta)
  thetaNames <- names(theta)

  if (!is.null(fullNames) && !is.null(thetaNames)) {
    isFree <- fullNames %in% thetaNames
  } else if (!is.null(fullNames) && is.null(thetaNames)) {
    isFree <- logical(length(fullTheta))
  } else {
    isFree <- seq_along(fullTheta) %in% seq_along(theta)
  }

  lavLabels <- model$params$lavLabels

  if (!is.null(lavLabels) && length(lavLabels) == length(fullTheta)) {
    names(fullTheta) <- lavLabels
  } else if (!is.null(fullNames)) {
    names(fullTheta) <- fullNames
  }

  list(all = fullTheta, free = fullTheta[isFree])
}

devtools::load_all()

tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'


testthat::expect_no_condition({
  TPB2 <- TPB
  TPB2[23:150, "att1"] <- NA

  fit_fiml <- modsem(tpb, TPB2, method = "lms", nodes = 32, missing = "fiml")

  ebm     <- modsem_predict(fit_fiml, method = "EBM", se = TRUE)
  ml      <- modsem_predict(fit_fiml, method = "ML", type = "lv")
  ml_all  <- modsem_predict(fit_fiml, method = "ML", type = "all")

  # newdata: use first 20 rows (may contain missing)
  ebm_new <- modsem_predict(fit_fiml, newdata = TPB2[1:20, ], method = "EBM")
  ml_new  <- modsem_predict(fit_fiml, newdata = TPB2[1:20, ], method = "ML")
})


testthat::expect_no_condition({
  fit_qml <- modsem(tpb, TPB, method = "qml")

  ebm     <- modsem_predict(fit_qml, method = "EBM", se = TRUE)
  ml      <- modsem_predict(fit_qml, method = "ML")
  ml_all  <- modsem_predict(fit_qml, method = "ML", type = "all")

  # newdata: use last 20 rows
  ebm_new <- modsem_predict(fit_qml, newdata = TPB[181:200, ], method = "EBM", type = "lv")
  ml_new  <- modsem_predict(fit_qml, newdata = TPB[181:200, ], method = "ML")
})


tpb.comp <- '
  ATT <~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH <~ b1 + b2

  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
'

testthat::expect_no_condition({
  fit_comp <- modsem(tpb.comp, TPB, method = "lms")

  ebm     <- modsem_predict(fit_comp, method = "EBM")
  ml      <- modsem_predict(fit_comp, method = "ML", type = "lv")

  ebm_new <- modsem_predict(fit_comp, newdata = TPB[1:20, ], method = "EBM", type = "all")
  ml_new  <- modsem_predict(fit_comp, newdata = TPB[1:20, ], method = "ML")
})


testthat::test_that("modsem_predict.modsem_da formats SE attributes like lavaan for one pattern", {
  fit_qml <- modsem(tpb, TPB[1:80, ], method = "qml", calc.se = FALSE)

  lv  <- modsem_predict(fit_qml, method = "EBM", type = "lv",  se = TRUE)
  ov  <- modsem_predict(fit_qml, method = "EBM", type = "ov",  se = TRUE)
  all <- modsem_predict(fit_qml, method = "EBM", type = "all", se = TRUE)

  testthat::expect_equal(dim(attr(lv, "se")),  c(1L, ncol(lv)))
  testthat::expect_equal(dim(attr(ov, "se")),  c(1L, ncol(ov)))
  testthat::expect_equal(dim(attr(all, "se")), c(1L, ncol(all)))

  testthat::expect_true(is.matrix(attr(lv, "acov")))
  testthat::expect_true(is.matrix(attr(ov, "acov")))
  testthat::expect_true(is.matrix(attr(all, "acov")))

  testthat::expect_null(attr(lv, "vcov"))
  testthat::expect_null(attr(lv, "pattern"))
  testthat::expect_null(attr(attr(lv, "acov"), "patterns"))
})


testthat::test_that("modsem_predict.modsem_da formats SE attributes by missingness pattern", {
  TPB2 <- TPB[1:80, ]
  TPB2[1:10, "att1"] <- NA

  fit_fiml <- modsem(
    tpb, TPB2, method = "lms", nodes = 32, missing = "fiml",
    calc.se = FALSE
  )

  lv <- modsem_predict(fit_fiml, method = "EBM", type = "lv", se = TRUE)

  testthat::expect_equal(dim(attr(lv, "se")), dim(lv))
  testthat::expect_null(attr(lv, "vcov"))
  testthat::expect_null(attr(lv, "pattern"))

  acov <- attr(lv, "acov")
  testthat::expect_true(is.list(acov))
  testthat::expect_equal(names(acov), c("pattern1", "pattern2"))
  testthat::expect_true(all(vapply(acov, is.matrix, logical(1))))

  patterns <- attr(acov, "patterns")
  testthat::expect_equal(dim(patterns)[1], 2L)
  testthat::expect_equal(rownames(patterns), c("pattern1", "pattern2"))
  testthat::expect_true("att1" %in% colnames(patterns))
  testthat::expect_true(any(!patterns[, "att1"]))

  lv_grouped <- modsem_predict(
    fit_fiml, method = "EBM", type = "lv", se = TRUE,
    drop.list.single.group = FALSE
  )

  acov_grouped <- attr(lv_grouped, "acov")
  testthat::expect_equal(names(acov_grouped), "group1")
  testthat::expect_equal(names(acov_grouped$group1), c("pattern1", "pattern2"))
  testthat::expect_equal(
    attr(acov_grouped$group1, "patterns"),
    patterns
  )
})

correctStdSolutionPI <- function(object, parTable.std) {
  # Function for correcting the standardized solution for interaction terms
  # lavaan standardizes the interaction term (X:Z) to have a variance of 1,
  # which is not correct if cov(X, Z) != 0.
  parTable <- parameter_estimates(object)
  intTerms <- object$elementsInProdNames
  cols <- c("est", "se", "ci.lower", "ci.upper")
  cols <- intersect(cols, names(parTable.std))

  vcov <- modsem_inspect(object, what = "cov.all")
  vcor <- modsem_inspect(object, what = "cor.all")

  is.multigroup <- "group" %in% colnames(parTable)
  parTable      <- addMissingGroups(parTable)
  parTable.std  <- addMissingGroups(parTable.std)
  groups        <- getGroupsParTable(parTable)

  parTable.std.new <- NULL

  for (g in groups) {
    if (is.multigroup) {
      vcov.g <- vcov[[g]]
      vcor.g <- vcor[[g]]
    } else {
      vcov.g <- vcov
      vcor.g <- vcor
    }

    parTable.g     <- parTable[parTable$group == g, , drop = FALSE]
    parTable.g.std <- parTable.std[parTable$group == g, , drop = FALSE]

    mod_stopif(!"est" %in% cols,
           "The parTable must contain the 'est' column for the standardized solution.")

    for (xz in names(intTerms)) {
      elems <- intTerms[[xz]]
      vars  <- diag(vcov.g[elems, elems])
      sds   <- sqrt(vars)

      rowsXZ <- parTable.g[parTable.g$rhs == xz & parTable.g$op == "~", , drop = FALSE]
      y <- rowsXZ$lhs[[1]]

      if (!length(y)) {
        mod_msg_warn(paste0("No endogenous variable found for interaction term '", xz, "'."))
        next
      }

      # Find the relevant exogenous variables
      struct <- parTable.g[parTable.g$lhs == y & parTable.g$op == "~", , drop = FALSE]
      xis    <- struct$rhs

      # Unstandardized terms
      vary  <- vcov.g[y, y]
      sdy   <- sqrt(vary)
      B3    <- getCoefs(y = y, x = xz, parTable = parTable.g)

      # Correlations
      combosXis <- getUniqueCombos(xis)
      combosXis <- combosXis[combosXis[[1]] == xz |
                             combosXis[[2]] == xz, , drop = FALSE]

      lxis <- combosXis[[1]]
      rxis <- combosXis[[2]]

      corrs <- vapply(seq_along(lxis), FUN.VALUE = numeric(1L),
                      FUN = \(i) vcor.g[lxis[i], rxis[i]])

      # Incorrectly standardized terms
      lcoefsIncorrect <- getCoefs(x = lxis, y = y, parTable = parTable.g.std)
      rcoefsIncorrect <- getCoefs(x = rxis, y = y, parTable = parTable.g.std)

      corrtermsIncorrect <- sum(2 * lcoefsIncorrect * rcoefsIncorrect * corrs)

      b3Incorrect <- getCoefs(y = y, x = xz, parTable = parTable.g.std)
      projVarY_XZ <- b3Incorrect ^ 2 + corrtermsIncorrect # this should be the same
      # for both the correctly, and
      # incorrectly standardized terms
      # and is the identity which makes it
      # possible to standardize the terms correctly

      # Correctly standardized terms
      b3Correct <- B3 * abs(prod(sds) / sdy) # in case some variances are negative
      # we want to make sure we don't flip
      # the sign...

      lcoefsCorrect <- lcoefsIncorrect
      rcoefsCorrect <- rcoefsIncorrect
      lcoefsCorrect[lxis == xz] <- b3Correct
      rcoefsCorrect[rxis == xz] <- b3Correct
      corrtermsCorrect <- sum(2 * lcoefsCorrect * rcoefsCorrect * corrs)

      # Calculate the correct standard deviation of the interaction term
      # Using the identity:
      #   projVarY_XZ = b3Correct ^ 2 * sd(xz) ^ 2 + sd(xz) * corrterms
      # Solve for sd(xz) using the quadratic formula:
      #   sd(xz) = (- corrterms +/- sqrt(corrterms^2 + 4 * b3Correct ^ 2 * projVarY_XZ)) / 2 * b3Correct ^ 2
      numerator <- -corrtermsCorrect + sign(projVarY_XZ) *
        sqrt(corrtermsCorrect^2 + 4 * (b3Correct ^ 2) * projVarY_XZ)
      denominator <- 2 * (b3Correct ^ 2)
      sdXZ <- numerator / denominator # correctly standardized sd(xz)

      # Correct parTable.g
      # We apply scaling factors to the parameters and the standard errors.
      # In theory we should use the delta method to calculate the new standard errors
      # but in practice it shouldn't really matter...
      lequalY  <- parTable.g.std$lhs == y
      requalXZ <- parTable.g.std$rhs == xz
      lequalXZ <- parTable.g.std$lhs == xz
      isCov    <- parTable.g.std$op == "~~"
      isCoef   <- parTable.g.std$op == "~"

      isCovTerm  <- xor(lequalXZ, requalXZ) & isCov
      isCoefTerm <- lequalY & requalXZ & isCoef
      isVarTerm  <- lequalXZ & requalXZ & isCov

      # Scaling factors
      scalefVar  <- (sdXZ^2) / 1
      scalefCoef <- b3Correct / b3Incorrect
      scalefCov  <- sdXZ

      # Apply
      parTable.g.std[isCovTerm, cols]  <- parTable.g.std[isCovTerm, cols]  * scalefCov
      parTable.g.std[isCoefTerm, cols] <- parTable.g.std[isCoefTerm, cols] * scalefCoef
      parTable.g.std[isVarTerm, cols]  <- parTable.g.std[isVarTerm, cols]  * scalefVar
    }

    etas <- unique(parTable.g.std[parTable.std.new$op == "~", "lhs"])
    varEtas <- calcVarParTable(etas, parTable.g.std)

    mod_warnif(
      any(abs(varEtas - 1) > 1e-10),
      paste0("Some variances are not equal to 1! ",
      "This indicates that the solution was not standardized correctly!")
    )

    parTable.std.new <- rbind(parTable.std.new, parTable.g.std)
  }

  # Recalculate custom parameters
  constrExprs <- sortConstrExprsFinalPt(parTable.std) # not .new

  errorWarning <- function(e) {
    mod_msg_warn(paste0("Calculation of custom parameter failed: ", e))
    NA
  }

  for (i in seq_len(NROW(constrExprs))) {
    row <- constrExprs[i, , drop=FALSE]

    expr      <- parse(text=constrExprs[i, "rhs"])
    labelList <- parTableLabelsToList(parTable.std.new) # must be updated for each iteration

    oldVal <- row$est
    newVal <- tryCatch(eval(expr, envir = labelList), error = errorWarning)

    ratio  <- newVal / oldVal
    values <- row[ , cols]

    row[, cols] <- values * ratio

    parTable.std.new <- rbind(parTable.std.new, row)
  }

  if (!is.multigroup)
    parTable.std.new <- parTable.std.new[colnames(parTable.std.new) != "group"]

  attr(parTable.std.new, "var.etas") <- varEtas
  parTable.std.new
}

PARSER_SETTINGS <- rlang::env(
  parentheses.as.string = FALSE
)


evalToken <- function(token, lhs, rhs) {
  UseMethod("evalToken")
}


#' @export
evalToken.LavOperator <- function(token, lhs, rhs) {
  if (is.LavToken(rhs)) rhs <- list(rhs)
  if (is.LavToken(lhs)) lhs <- list(lhs)

  mod_stopif(!is.atomic(lhs) && is.LavOperator(lhs$op),
    paste0("Unexpected operator ", highlightErrorToken(lhs$op)),
    only.format.header = TRUE
  )

  mod_stopif(!is.atomic(rhs) && is.LavOperator(rhs$op),
    paste0("Unexpected operator ", highlightErrorToken(rhs$op)),
    only.format.header = TRUE
  )

  list(lhs = lhs, op = token, rhs = rhs)
}


#' @export
evalToken.LavToken <- function(token, lhs, rhs) {
  token
}


#' @export
evalToken.LavName <- function(token, lhs, rhs) {
  mod_stopif(is.LavToken(rhs),
    paste0("Unexpected token ", highlightErrorToken(rhs)),
    only.format.header = TRUE
  )
  mod_stopif(is.LavToken(lhs),
    paste0("Unexpected token ", highlightErrorToken(lhs)),
    only.format.header = TRUE
  )
  token
}


#' @export
evalToken.LavNumeric <- function(token, lhs, rhs) {
  mod_stopif(is.LavToken(rhs),
    "Unexpected token:\n",
    highlightErrorToken(rhs),
    only.format.header = TRUE
  )

  mod_stopif(is.LavToken(lhs),
    "Unexpected token:\n",
    highlightErrorToken(lhs),
    only.format.header = TRUE
  )

  token
}


#' @export
evalToken.LavAdd <- function(token, lhs, rhs) {
  if (is.null(rhs)) {
    mod_msg_stop(
      paste0("Expected token after `+` ", highlightErrorToken(token)),
      only.format.header = TRUE
    )

  } else if (is.null(lhs)) {
    mod_msg_stop(
      paste0("Expected token before `+` ", highlightErrorToken(token)),
      only.format.header = TRUE
    )
  }

  if (is.LavToken(rhs)) {
    rhs <- list(rhs)
  }
  if (is.LavToken(lhs)) {
    lhs <- list(lhs)
  }
  c(lhs, rhs)
}


#' @export
evalToken.LavModify <- function(token, lhs, rhs) {
  structure(rhs, modifier = lhs)
}


#' @export
evalToken.LavSubtract <- function(token, lhs, rhs) {
  out <- paste0(token, rhs)
  attributes(out) <- attributes(rhs)

  out
}


#' @export
evalToken.LavBlank <- function(token, lhs, rhs) {
  NULL
}


#' @export
evalToken.LavInteraction <- function(token, lhs, rhs) {
  if (!"LavName" %in% class(lhs) || !"LavName" %in% class(rhs)) {
    mod_msg_stop(
      paste0("Interactions are reserved for objects ", highlightErrorToken(token)),
      only.format.header = TRUE
    )
  }
  out <- paste0(lhs, token, rhs)
  attributes(out) <- attributes(lhs)
  out
}


#' @export
evalToken.LavComment <- function(token, lhs, rhs) {
  NULL
}


#' @export
evalToken.LavFunction <- function(token, lhs, rhs) {
  functionCall <- paste0(token, stringr::str_c(unlist(rhs), collapse = ","), ")")
  out <- eval(rlang::parse_expr(functionCall), envir = modEnv)
  attributes(out) <- attributes(token)
  out
}


#' @export
evalToken.LeftBracket <- function(token, lhs, rhs) {
  if (PARSER_SETTINGS$parentheses.as.string)
    return(paste0("(", stringifyTokens(rhs), ")"))

  rhs
}


#' @export
evalToken.RightBracket <- function(token, lhs, rhs) {
  lhs
}


stringifyTokens <- function(tokens) {
  if (all(c("lhs", "op", "rhs") %in% names(tokens))) {
    lhs <- paste0(unlist(tokens$lhs), collapse = "+")
    op  <- unlist(tokens$op)[[1L]]
    rhs <- paste0(unlist(tokens$rhs), collapse = "+")
    out <- paste0(lhs, op, rhs)

  } else {
    out <- paste0(unlist(tokens), collapse="")
  }

  # This code is ugly as f*
  class(out) <- c("LavObject", "LavName", "LavToken")
  attr(out, "pos") <- attr(unlist(tokens)[[1L]], "pos")
  attr(out, "priority") <- 13
  attr(out, "lineNumber") <- attr(unlist(tokens)[[1L]], "lineNumber")

  out
}


#' @export
evalToken.LavSeperator <- function(token, lhs, rhs) {
  if ("LavToken" %in% class(rhs)) {
    rhs <- list(rhs)
  }
  if ("LavToken" %in% class(lhs)) {
    lhs <- list(lhs)
  }
  c(lhs, rhs)
}


evalTokens <- function(syntaxTree) {
  if (is.null(syntaxTree) || length(syntaxTree) == 0) {
    return(NULL)
  }

  lhs <- evalTokens(syntaxTree$lhs)
  rhs <- evalTokens(syntaxTree$rhs)

  return(evalToken(syntaxTree$node,
                   lhs = lhs,
                   rhs = rhs))
}


parseSyntaxTrees <- function(syntaxTrees) {
  lapply(syntaxTrees, evalTokens)
}


createParTableBranch <- function(syntaxTree) {
  if (is.null(syntaxTree)) {
    return(NULL)
  }
  rhs <- vector("character", length(syntaxTree[["rhs"]]))
  mod <- rhs

  for (i in seq_along(syntaxTree[["rhs"]])) {
    rhs[[i]] <- getTokenString(syntaxTree[["rhs"]][[i]])
    modifier <- getTokenString(attr(syntaxTree[["rhs"]][[i]], "modifier"))
    if (!is.null(modifier)) {
      mod[[i]] <- modifier
    }
  }
  lhs <- vapply(syntaxTree[["lhs"]], FUN.VALUE = character(1L),
                FUN = getTokenString)
  lhs <- vapply(lhs, FUN.VALUE = character(length(rhs)),
                FUN = function(x, len)
                  rep(x, len),
                len = length(rhs)) |> as.vector()
  op <- rep(getTokenString(syntaxTree$op), length(rhs))
  parTable <- data.frame(lhs = lhs, op = op, rhs = rhs, mod = mod)

  # post-processing
  parTable[parTable$op == ":=", "mod"] <- parTable[parTable$op == ":=", "lhs"]
  intercepts <- parTable$op == "~" & parTable$rhs == "1"
  parTable[intercepts, "op"]  <- "~1"
  parTable[intercepts, "rhs"] <- ""

  parTable
}


#' Generate parameter table for \code{lavaan} syntax
#'
#' @param syntax model syntax
#' @param parentheses.as.string Should parentheses be read parsed as literal strings?
#'
#' @return \code{data.frame} with columns \code{lhs, op, rhs, mod}
#' @export modsemify
#'
#' @examples
#' library(modsem)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#''
#' modsemify(m1)
modsemify <- function(syntax, parentheses.as.string = FALSE) {
  if (is.null(syntax)) return(NULL)

  mod_stopif(!is.character(syntax) && length(syntax) > 1,
    "Syntax is not a string og length 1"
  )

  on.exit({
    PARSER_SETTINGS$parentheses.as.string <- FALSE
  })

  if (parentheses.as.string)
    PARSER_SETTINGS$parentheses.as.string <- TRUE

  syntaxTrees <- createSyntaxTreesSyntax(syntax)
  parsedTrees <- parseSyntaxTrees(syntaxTrees)
  purrr::list_rbind(lapply(parsedTrees,
                           FUN = createParTableBranch))
}


ungroupParTable <- function(parTable) {
  if (!"group" %in% colnames(parTable) || length(getGroupsParTable(parTable)) <= 1L)
    return(parTable)

  parUnique <- unique(parTable[c("lhs", "op", "rhs")])
  parTableUngrouped <- NULL

  for (i in seq_len(NROW(parUnique))) {
    lhs <- parUnique[i, "lhs"]
    op  <- parUnique[i, "op"]
    rhs <- parUnique[i, "rhs"]

    modifiers <- parTable[parTable$lhs == lhs &
                          parTable$op  == op  &
                          parTable$rhs == rhs, "mod"]

    modifiers[modifiers == ""] <- NA

    if (all(is.na(modifiers)))
      mod <- ""
    else
      mod <- sprintf("c(%s)", paste0(modifiers, collapse = ", "))

    parTableUngrouped <- rbind(
      parTableUngrouped,
      data.frame(lhs = lhs, op = op, rhs = rhs, mod = mod)
    )
  }

  parTableUngrouped
}


parTableToSyntax <- function(parTable, removeColon = FALSE) {
  if (is.null(parTable)) return(NULL)

  parTable <- ungroupParTable(parTable)

  intercepts <- parTable$op == "~1"
  parTable[intercepts, "rhs"] <- "1"
  parTable[intercepts, "op"]  <- "~"

  out <- ""
  if (removeColon) {
    parTable$lhs <- stringr::str_remove_all(parTable$lhs, ":")
    parTable$rhs <- stringr::str_remove_all(parTable$rhs, ":")
    parTable$mod <- stringr::str_remove_all(parTable$mod, ":")
  }

  for (i in seq_len(NROW(parTable))) {
    if (parTable[["mod"]][[i]] != "" && parTable[["op"]][[i]] != ":=")
      modifier <- paste0(parTable[["mod"]][[i]], "*")
    else
      modifier <- ""

    line <- paste0(parTable[["lhs"]][[i]], " ",
                   parTable[["op"]][[i]], " ",
                   modifier,
                   parTable[["rhs"]][[i]], "\n")
    out <- paste0(out, line)
  }

  out
}


mergeTokens <- function(x, y) {
  mod_stopif(!"LavName" %in% class(x) || !"LavName" %in% class(x),
    paste0("Interactions are reserved for objects ", highlightErrorToken(x)),
    only.format.header = TRUE
  )

  out <- paste0(x, y)
  attributes(out) <- attributes(x)
  out

}

devtools::load_all()

tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + PBC + SN
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

splitm <- function(s) splitParTable(modsemify(s))
s1 <- splitm(tpb)
testthat::expect_true(unique(s1$parTableCov$lhs) == "INT")

testthat::expect_warning(
  est_qml <- modsem(tpb, TPB, method = "qml", auto.split.syntax = FALSE)
)
summary(est_qml)


tpb2 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  HAPPY =~ h1 + h2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + PBC + SN
  BEH ~ INT + PBC
  BEH ~ INT:PBC
  HAPPY ~ BEH:INT
'

s2 <- splitm(tpb2)
testthat::expect_true(unique(s2$parTableCov$lhs) == "INT")

tpb3 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  HAPPY =~ h1 + h2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + PBC + SN
  BEH ~ INT + PBC
  INT ~ SN:PBC
  HAPPY ~ BEH:INT
'

s3 <- splitm(tpb3)
testthat::expect_true(is.null(s3$parTableCov))

devtools::load_all()

m1 <- modsemify(
                '
y1 ~ x1
x1 ~ x2
x2 ~ x3
x3 ~ x1
                '
)
testthat::expect_warning(trace_path(m1, "x1", "x1"),
                         "non-recursive model \\(infinite loop\\)")
m2 <- modsemify(
                '
y1 ~ x1
x1 ~ x2
x2 ~ x3
x3 ~ x4
x1 ~~ x1
x2 ~~ x2
x3 ~~ x3
x4 ~~ x4
                '
)

testthat::expect_equal(
  unname(trace_path(m2, "x3", "x3")),
  "(x3~~x3 + x3~x4 ^ 2 * x4~~x4)"
)

testthat::expect_equal(
  unname(trace_path(m2, "x4", "x4")),
  "(x4~~x4)"
)

m3 <- modsemify('visual  =~ x1 + x2 + x3
                textual =~ x4 + x5 + x6
                speed   =~ x7 + x8 + x9
                visual  ~ speed + textual + speed:textual')

testthat::expect_true(is.na(trace_path(m3, "textual", "textual")))


m4 <- modsemify('
   # Outer Model
   X =~ x1 + x2 +x3
   Y =~ y1 + y2 + y3
   Z =~ z1 + z2 + z3

   # Inner model
   Y ~ X + Z + X:Z
')

testthat::expect_equal(
  unname(trace_path(m4, "X", "X", missing.cov = TRUE)),
  "(X~~X)"
)

devtools::load_all()

testthat::expect_error({
  modsemify("y ~ x + z + (x + z | cluster)", parentheses.as.string = FALSE)
})

testthat::expect_no_condition({
  modsemify("y ~ x + z + (x + z | cluster)", parentheses.as.string = TRUE)
})

testthat::expect_true(!PARSER_SETTINGS$parentheses.as.string)

devtools::load_all()
models <- list(m1 = '
               # latent variables
               ind60 =~ x1 + x2 + x3
               dem60 =~ y1 + y2 + y3 + y4
               dem65 =~ y5 + y6 + y7 + y8
               # regressions
               dem60 ~ ind60
               dem65 ~ ind60 + dem60
               # residual covariances
               y1 ~~ y5
               y2 ~~ y4 + y6
               y3 ~~ y7
               y4 ~~ y8
               y6 ~~ y8
               ',
               m3 = ' visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9 ',
               m4 = 'visual  =~ x1 + start(0.8)*x2 + start(1.2)*x3
               textual =~ x4 + start(0.5)*x5 + start(1.0)*x6
               speed   =~ x7 + start(0.7)*x8 + start(1.8)*x9',
               m5 = '# three-factor model
               visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9
               # intercepts with fixed values
               x1 + x2 + x3 + x4 ~ 0.5*1',
               m6 = '# three-factor model
               visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9
               # intercepts
               x1 ~ 1
               x2 ~ 1
               x3 ~ 1
               x4 ~ 1
               x5 ~ 1
               x6 ~ 1
               x7 ~ 1
               x8 ~ 1
               x9 ~ 1',
               m7 = ' # direct effect
               Y ~ c*X
               # mediator
               M ~ a*X
               Y ~ b*M
               # indirect effect (a*b)
               ab := a*b
               # total effect
               total := c + (a*b)
               '
)


set.seed(1234)
X <- rnorm(100)
M <- 0.5*X + rnorm(100)
Y <- 0.7*M + rnorm(100)
d7 <- data.frame(X = X, Y = Y, M = M)

data <- list(d1 = lavaan::PoliticalDemocracy,
             d3 = lavaan::HolzingerSwineford1939,
             d4 = lavaan::HolzingerSwineford1939,
             d5 = lavaan::HolzingerSwineford1939,
             d6 = lavaan::HolzingerSwineford1939,
             d7 = d7)


estimates <- vector("list", length(models))
for (i in seq_along(estimates)) {
  estimates[[i]]$lav <- tryCatch({
    est <- lavaan::sem(models[[i]], data = data[[i]])
    est
  },
  warning = function(e) {
    warning("Warning in lav model ", i)
    warning(capturePrint(e))
  },
  error = function(e) {
    est <- NA
    warning("Error in lav model ", i)
  },
  finally = {
    est
  })
  estimates[[i]]$modsem <- tryCatch({
    est <- modsem(models[[i]], data = data[[i]], estimator = "ML")
    est
  },
  warning = function(e) {
    warning("Warning in modsem model ", i)
    est
  },
  error = function(e) {
    est <- NA
    warning("Error in modsem model ", i)
  },
  finally = {
    est
  })
}

for (est in estimates) {
  lavaanEst <- lavaan::parameterEstimates(est$lav)
  lavaanEst[is.na(lavaanEst)] <- -999
  modsemEst <- lavaan::parameterEstimates(est$modsem$lavaan)
  modsemEst[is.na(modsemEst)] <- -999
  testthat::expect_equal(lavaanEst, modsemEst)
}


testthat::expect_error(modsemify('X =~ x1 x2 + x3'), regexp = "Unexpected token.*")
testthat::expect_error(modsemify('X =~ 1.0 1.0'),
                       regexp = "Unexpected token.*")
testthat::expect_error(modsemify('X =~ x1 + x2 + +x3'),
                       regexp = "Expected token before .*")
testthat::expect_error(modsemify('X =~ x1 + x2 +'),
                       regexp = "Expected token after .*")
testthat::expect_error(modsemify('X =~ x1 + x2 =~ x3'),
                       regexp = "Unexpected operator .*")

m <- '
X =~ x1 + x2 +
+x3
'
rebuiltLine <- paste0(getLines(m)[[2]], collapse="") # check that `+\n+` -> `+ +`, not `++`
testthat::expect_true(!any(grepl("\\+\\+", rebuiltLine)))

devtools::load_all()

tpb <- "
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

  INT ~ ATT + SN + PBC + ATT:PBC
  BEH ~ INT + PBC + ATT + INT:PBC + ATT:PBC
"

est_lms <- modsem(tpb, TPB, method = "lms", nodes = 32)
plot_jn(x = "ATT", z = "PBC", y = "BEH", model = est_lms, type = "indirect")
plot_jn(x = "ATT", z = "PBC", y = "BEH", model = est_lms, type = "total")

est_dblcent <- modsem(tpb, TPB, method = "dblcent")
plot_jn(x = "ATT", z = "PBC", y = "BEH", model = est_dblcent, type = "indirect")

devtools::load_all()

m1 <- '
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

  Y ~ X + Z + X:Z
'

fit_pi <- modsem(m1, oneInt)

bootstrap_modsem(fit_pi, FUN = coef, R = 10)

rsqr_diff <- function(est_h1) {
  est_h0 <- estimate_h0(est_h1, reduced = FALSE)
  r2.h1 <- modsem_inspect(est_h1, what = "r2")
  r2.h0 <- modsem_inspect(est_h0, what = "r2")

  r2.h1[names(r2.h0)] - r2.h0
}

bootstrap_modsem(fit_pi, FUN = rsqr_diff, R = 10)


fit_da <- modsem(m1, oneInt, method = "lms")
summary(fit_da)

bootstrap_modsem(fit_da, FUN = coef, R = 10L, type = "parametric")
bootstrap_modsem(fit_da, FUN = coef, R = 10L, type = "nonparametric")

tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

boot <- bootstrap_modsem(model = modsem,
                         model.syntax = tpb, data = TPB,
                         method = "dblcent", rcs = TRUE,
                         rcs.scale.corrected = TRUE,
                         rcs.mc.reps = 10000,
                         R = 10L,
                         FUN = "coef")
coef <- apply(boot, MARGIN = 2, FUN = mean, na.rm = TRUE)
se   <- apply(boot, MARGIN = 2, FUN = sd, na.rm = TRUE)

cat("Parameter Estimates:\n")
print(coef)

cat("Standard Errors: \n")
print(se)


for (method in c("lms", "dblcent")) {
  fit_rcs <- modsem(m1, oneInt, rcs = TRUE, method = method)

  testthat::expect_warning(
    bootstrap_modsem(fit_rcs, R = 10L),
    regexp = "bootstrapping.*rcs=TRUE"
  )
}

#' Estimate a \code{modsem} model using multiple imputation
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param data A dataframe with observed variables used in the model.
#'
#' @param method Method to use:
#' \describe{
#'   \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
#'   \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
#' }
#'
#' @param m Number of imputations to perform. More imputations will yield better estimates
#'   but can also be (a lot) slower.
#'
#' @param verbose Should progress be printed to the console?
#'
#' @param se How should corrected standard errors be computed? Alternatives are:
#' \describe{
#'   \item{\code{"simple"}}{Uncorrected standard errors are only calculated once, in the first imputation.
#'                          The standard errors are thereafter corrected using the distribution of the
#'                          estimated coefficients from the different imputations.}
#'   \item{\code{"full"}}{Uncorrected standard errors are calculated and aggregated for each imputation.
#'                        This can give more accurate results, but can be (a lot) slower.
#'                        The standard errors are thereafter corrected using the distribution of the
#'                        estimated coefficients from the different imputations.}
#' }
#'
#' @param ... Arguments passed to \code{\link{modsem}}.
#'
#' @details
#'  \code{modsem_impute} is currently only available for the DA approaches
#'  (LMS and QML). It performs multiple imputation using \code{Amelia::amelia}
#'  and returns aggregated coefficients from the multiple imputations, along with
#'  corrected standard errors.
#'
#' @examples
#'
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#'
#' oneInt2 <- oneInt
#'
#' set.seed(123)
#' k <- 200
#' I <- sample(nrow(oneInt2), k, replace = TRUE)
#' J <- sample(ncol(oneInt2), k, replace = TRUE)
#' for (k_i in seq_along(I)) oneInt2[I[k_i], J[k_i]] <- NA
#'
#' \dontrun{
#' est <- modsem_mimpute(m1, oneInt2, m = 25)
#' summary(est)
#' }
#' @export
modsem_mimpute <- function(model.syntax,
                           data,
                           method = "lms",
                           m = 25,
                           verbose = interactive(),
                           se = c("simple", "full"),
                           ...) {
  mod_stopif(!method %in% c("lms", "qml"),
         "Multiple imputation is only available for the LMS and QML approaches!")

  modsem_mimput_modsem_da(
    model.syntax = model.syntax,
    data = data,
    method = method,
    m = m,
    verbose = verbose,
    se = se,
    ...
  )
}


modsem_mimput_modsem_da <- function(model.syntax,
                                    data,
                                    method = "lms",
                                    m = 25,
                                    cov.syntax = NULL,
                                    verbose = TRUE,
                                    se = c("simple", "full"),
                                    calc.se = NULL,
                                    ...) {
  se <- tolower(se)
  se <- match.arg(se)

  parTable <- rbind(modsemify(model.syntax), modsemify(cov.syntax))
  ovs      <- getOVs(parTable)

  missing.cols <- !ovs %in% colnames(data)
  mod_stopif(any(missing.cols),
         paste0("Variables missing from data:\n",
         paste(ovs[missing.cols], sep = "\n")))

  data <- data[, ovs, drop = FALSE]

  if (verbose) cat("Imputing data...\n")
  imputed  <- Amelia::amelia(data, m = m, p2s = 0)

  COEF.ALL  <- vector("list", length = m)
  COEF.FREE <- vector("list", length = m)
  VCOV.ALL  <- vector("list", length = m)
  VCOV.FREE <- vector("list", length = m)
  fits <- vector("list", length = m)

  THETA <- NULL

  for (i in seq_len(m)) {
    printedLines <- utils::capture.output(split = TRUE, {
      if (verbose) printf("Fitting imputation %d/%d...\n", i, m)

      if (is.null(calc.se))
        calc.se_i <- ifelse(se == "simple" && i > 1, yes = FALSE, no = TRUE)
      else
        calc.se_i <- calc.se

      if (is.null(THETA)) {
        optimize <- TRUE
        start    <- NULL
      } else {
        optimize <- FALSE
        start    <- apply(THETA, MARGIN = 2, FUN = mean, na.rm = TRUE) # na.rm should't be necessary, ever...
      }

      data_i <- as.data.frame(imputed$imputations[[i]])

      fit_i <- tryCatch(
        modsem_da(
          model.syntax = model.syntax,
          data         = data_i,
          method       = method,
          cov.syntax   = cov.syntax,
          start        = start,
          verbose      = verbose,
          optimize     = optimize,
          calc.se      = calc.se_i,
          ...
        ), error = \(e) {print(e); NULL}
      )

      if (is.null(fit_i)) next


      fits[[i]]      <- fit_i
      COEF.ALL[[i]]  <- coef(fit_i, type = "all")
      COEF.FREE[[i]] <- coef(fit_i, type = "free")

      if (i > 1 && se == "simple") {
        VCOV.ALL[[i]]  <- VCOV.ALL[[1]]
        VCOV.FREE[[i]] <- VCOV.FREE[[1]]

      } else if (calc.se_i) {
        VCOV.ALL[[i]]  <- vcov(fit_i, type = "all")
        VCOV.FREE[[i]] <- vcov(fit_i, type = "free")

      } else {
        k.all  <- length(COEF.ALL[[i]])
        k.free <- length(COEF.FREE[[i]])
        d.all  <- names(COEF.ALL[[i]])
        d.free <- names(COEF.FREE[[i]])

        VCOV.ALL[[i]]  <- matrix(0, nrow = k.all, ncol = k.all,
                                 dimnames = list(d.all, d.all))
        VCOV.FREE[[i]] <- matrix(0, nrow = k.free, ncol = k.free,
                                 dimnames = list(d.free, d.free))
      }

      THETA <- rbind(
        THETA,
        matrix(fit_i$theta, nrow = 1, dimnames = list(NULL, names(fit_i$theta)))
      )
    })

    nprinted <- length(printedLines)
    if (i < m) eraseConsoleLines(nprinted)
  }

  failed <- vapply(fits, FUN.VALUE = logical(1L), FUN = is.null)

  if (any(failed)) {
    mod_msg_warn(sprintf("Model estimation failed in %d out of %d impuations!",
                     sum(failed), m))

    fits      <- fits[!failed]
    COEF.ALL  <- COEF.ALL[!failed]
    COEF.FREE <- COEF.FREE[!failed]
    VCOV.ALL  <- VCOV.ALL[!failed]
    VCOV.FREE <- VCOV.FREE[!failed]
    m         <- sum(!failed)
  }

  pool.all  <- rubinPool(COEF.ALL,  VCOV.ALL)
  pool.free <- rubinPool(COEF.FREE, VCOV.FREE)

  coef.all  <- pool.all$theta.bar
  vcov.all  <- pool.all$Tvcov

  coef.free <- pool.free$theta.bar
  vcov.free <- pool.free$Tvcov

  # Re-do parameter estimates
  parTable1   <- parameter_estimates(fits[[1]])
  orig.labels <- parTable1$label
  parTable1   <- getMissingLabels(parTable1)
  parTableT   <- data.frame(label = names(coef.all),
                            est.t = coef.all,
                            std.error.t = sqrt(diag(vcov.all)))

  parTable    <- leftJoin(left = parTable1, right = parTableT, by = "label")
  match       <- !is.na(parTable$est.t)

  parTable$est[match]       <- parTable$est.t[match]
  parTable$std.error[match] <- parTable$std.error.t[match]
  parTable$est.t       <- NULL
  parTable$std.error.t <- NULL
  parTable$label[!parTable$label %in% orig.labels] <- ""
  parTable <- parTable[c("lhs", "op", "rhs", "group", "label", "est", "std.error")] # remove z-statistics
  parTable <- addZStatsParTable(parTable)
  rownames(parTable) <- NULL # reset

  matrices    <- aggregateMatrices(fits, type = "main")
  covMatrices <- aggregateMatrices(fits, type = "cov")
  expected.matrices <- aggregateMatrices(fits, type = "expected")

  getScalarFit <- function(fit, field, dtype = numeric)
    vapply(fits, FUN.VALUE = dtype(1L), \(fit) fit[[field]])

  fit.out <- fits[[1]]
  fit.out$coefs.all       <- coef.all
  fit.out$coefs.free      <- coef.free
  fit.out$vcov.all        <- vcov.all
  fit.out$vcov.free       <- vcov.free
  fit.out$parTable        <- parTable
  fit.out$information     <- sprintf("Rubin-corrected (m=%d)", m)
  fit.out$FIM             <- solve(vcov.free)
  fit.out$theta           <- apply(THETA, MARGIN = 2, FUN = mean, na.rm = TRUE)
  fit.out$iterations      <- sum(getScalarFit(fits, field = "iterations"))
  fit.out$logLik          <- mean(getScalarFit(fits, field = "logLik"))
  fit.out$convergence     <- all(getScalarFit(fits, field = "convergence",
                                              dtype = logical))
  fit.out$convergence.msg <- getConvergenceMessage(fit.out$convergence,
                                                   fit.out$iterations)
  fit.out$model$matrices          <- matrices
  fit.out$model$covModel$matrices <- covMatrices
  fit.out$expected.matrices       <- expected.matrices

  fit.out$imputations <- list(fitted = fits, data = imputed)

  fit.out
}


matrixMeans <- function(matrices, na.rm = TRUE) {
  if (!length(matrices))
    return(matrix(nrow=0, ncol=0))

  # Take the mean of a list of n*p matrices
  # With the option of ignoring missing values

  flattened <- as.data.frame(lapply(matrices, FUN = c))
  means <- rowMeans(flattened, na.rm = na.rm)

  dims <- dimnames(matrices[[1]])
  n    <- NROW(matrices[[1]])
  p    <- NCOL(matrices[[1]])

  matrix(means, nrow = n, ncol = p, dimnames = dims)
}


aggregateMatrices <- function(fits, type) {
  switch(type,
    main = {
      matrices  <- fits[[1]]$model$matrices
      names.num <- NAMES_PAR_MATRICES
    },
    cov = {
      matrices  <- fits[[1]]$model$covModel$matrices
      names.num <- NAMES_PAR_MATRICES_COV
    },
    expected = {
      matrices  <- fits[[1]]$expected.matrices
      names.num <- names(matrices)
    }
  )

  if (is.null(matrices) || !length(matrices))
    return(matrices)

  m <- length(fits)

  for (i in seq_len(m)[-1]) {
    fit_i      <- fits[[i]]
    matrices_i <- switch(type,
      main     = fit_i$model$matrices,
      cov      = fit_i$model$covModel$matrices,
      expected = fit_i$expected.matrices
    )
    matrices_i <- matrices_i[names(matrices_i) %in% names.num]

    for (mat in names(matrices_i)) {
      if (is.null(matrices_i[[mat]]) || !length(matrices_i[[mat]]))
        next

      matrices[[mat]] <- matrices[[mat]] + matrices_i[[mat]]
    }
  }

  for (mat in names.num)
    matrices[[mat]] <- matrices[[mat]] / m

  matrices
}


rubinPool <- function(coef.list, vcov.list) {
  stopifnot(length(coef.list) == length(vcov.list))

  m     <- length(coef.list)
  THETA <- do.call(cbind, coef.list)
  p     <- nrow(THETA)

  # Pooled point estimate
  theta.bar <- rowMeans(THETA, na.rm = TRUE) # na.rm shouldn't be necessary
  # but just in case...

  # Within-imputation covariance (W)
  W <- matrixMeans(vcov.list, na.rm = TRUE)

  # Between-imputation covariance (B)
  THETAc <- sweep(THETA, 1, theta.bar, "-")
  B      <- (THETAc %*% t(THETAc)) / (m - 1)

  # Total Rubin-corrected covariance (T)
  Tvcov  <- W + (1 + 1/m) * B

  # Small-sample df
  df <- (m - 1) * (1 + diag(W) / ((1 + 1/m) * diag(B)))^(-2)

  list(theta.bar = theta.bar,
       W         = W,
       B         = B,
       Tvcov     = Tvcov,
       df        = df)
}

devtools::load_all()

m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'

oneInt2 <- oneInt

set.seed(123)
k <- 200
I <- sample(nrow(oneInt2), k, replace = TRUE)
J <- sample(ncol(oneInt2), k, replace = TRUE)
for (k_i in seq_along(I)) oneInt2[I[k_i], J[k_i]] <- NA

# double centering approach
est <- modsem(m1, oneInt2)
testthat::expect_true(any(is.na(est$data)))

est <- modsem(m1, oneInt2, na.rm=TRUE)
testthat::expect_true(!any(is.na(est$data)))

est <- modsem(m1, oneInt2, na.rm=FALSE)
testthat::expect_true(any(is.na(est$data)))

# Residual Centering Approach
est <- modsem(m1, oneInt2, method = "rca")
testthat::expect_true(any(is.na(est$data)))

est <- modsem(m1, oneInt2, method = "rca", na.rm=TRUE)
testthat::expect_true(!any(is.na(est$data)))

est <- modsem(m1, oneInt2, method = "rca", na.rm=FALSE)
testthat::expect_true(any(is.na(est$data)))

# lms
testthat::expect_warning(
  modsem(m1, oneInt2, method = "lms", missing = "complete", convergence.abs = 1,
         calc.se = FALSE),
  regexp = "Removing missing values list-wise"
)


testthat::expect_message(
  modsem(m1, oneInt2, method = "lms", missing = "impute", convergence.abs = 1,
         calc.se = FALSE),
  regexp = "Imputing.*"
)

# qml
testthat::expect_message(
  modsem(m1, oneInt2, method = "qml", missing = "impute", convergence.rel =1e-1,
         calc.se = FALSE),
  regexp = "Imputing.*"
)


# test multiple imputation
mimp_lms <- modsem_mimpute(m1, oneInt2, method = "lms", m = 5, cov.syntax = "")
mimp_qml <- modsem_mimpute(m1, oneInt2, method = "qml", m = 5)

print(summary(mimp_lms))
summary(mimp_qml)

testthat::expect_false(
  mimp_lms$imputations$fitted[[1]]$logLik ==
    mimp_lms$logLik
)

sd.uncorrected <- sqrt(diag(vcov(mimp_lms$imputations$fitted[[1]])))
sd.corrected   <- sqrt(diag(vcov(mimp_lms)))
testthat::expect_true(all(sd.corrected > sd.uncorrected))

# test fiml
fiml_lms <- modsem(m1, oneInt2, method = "lms", missing = "fiml")
summary(fiml_lms)

fiml_h0  <- estimate_h0(fiml_lms, calc.se = FALSE)
testthat::expect_true(fiml_h0$iterations == 2L)

testthat::expect_error(modsem(m1, oneInt2, method = "qml", missing = "fiml") ,
                       regexp = "Using FIML with QML is not available.*yet.*")

oneInt3 <- oneInt2
oneInt3$x1 <- NA
oneInt3$y2 <- NA

missing.options <- c("fiml", "ml", "direct", "casewise", "listwise",
                     "impute", "complete")
for (missing in missing.options) {
  testthat::expect_error(modsem(m1, oneInt3, method = "lms", missing = "fiml"),
                         regexp = "Please remove .*x1.*y2")
}

# test fiml with multigroup
set.seed(123)
oneIntMG <- oneInt2
oneIntMG$group <- sample(c("G1", "G2"), nrow(oneInt2), replace = TRUE)
fiml_lms_g <- modsem(m1, oneIntMG, method = "lms", missing = "fiml", group = "group")
summary(fiml_lms_g)


# test fiml with sampling weights
# Sampling weights used to be silently dropped under
# FIML (matrix subsetting in `handleMissingData()` dropped the attribute),
# and the weights must be aligned with the pattern-concatenated row
# order used by the E-step, so results must be invariant to row shuffling.

set.seed(42283)
oneIntW <- oneInt[seq_len(400), ]
oneIntW[sample(400, 60), "x2"] <- NA
oneIntW$wt <- runif(400, 0.2, 3)
oneIntWSorted <- oneIntW[order(is.na(oneIntW$x2)), ]

set.seed(7282)
oneIntWShuffled <- oneIntWSorted[sample(400), ]

fiml.w  <- modsem(m1, oneIntWSorted, method = "lms", missing = "fiml",
                  sampling.weights = "wt", calc.se = FALSE)
fiml.w2 <- modsem(m1, oneIntWShuffled, method = "lms", missing = "fiml",
                  sampling.weights = "wt", calc.se = FALSE)
fiml.nw <- modsem(m1, oneIntWSorted, method = "lms", missing = "fiml",
                  calc.se = FALSE)

# weights must actually affect the fit
testthat::expect_gt(
  max(abs(coef(fiml.w) - coef(fiml.nw)[names(coef(fiml.w))])),
  expected = 0.01
)

# and must be invariant to the row ordering of the data
testthat::expect_equal(
  unclass(coef(fiml.w)),
  unclass(coef(fiml.w2)[names(coef(fiml.w))]),
  tolerance = 1e-8
)

getJN_GridMonteCarlo <- function(x, z, y, parTable, model, min_z, max_z, sig.level, alpha,
                                    detail, sd.line, standardized, xz, greyscale,
                                    plot.jn.points = TRUE, group = NULL, group.label = NULL,
                                    type = "indirect", mc.reps = 10000, ...) {
  mod_msg_note("Using Monte-Carlo quantiles...")

  type <- match.arg(type, c("indirect", "total", "direct"))

  mean_z <- getMean(z, parTable = parTable)
  sd_z   <- sqrt(calcCovParTable(x = z, y = z, parTable = parTable))
  mod_stopif(is.na(sd_z), sprintf("Variance for %s not found in model", z))

  z_min <- min_z + mean_z
  z_max <- max_z + mean_z

  labels <- getParTableLabels(parTable, labelCol = "label")
  labels <- stringr::str_replace_all(labels, OP_REPLACEMENTS)
  parTable$label <- labels

  intTerms <- getIntTerms(parTable)

  mod_stopif(!length(intTerms), "No interaction terms were found in the model.")

  for (intTerm in intTerms) {
    elems <- stringr::str_split_1(intTerm, pattern = ":")

    if (!z %in% elems)
      next

    mod_stopif(length(elems) > 2L,
           "Johnson-Neyman plot is not available for three-way (or higher) interactions!")

    m <- elems[elems != z]

    simpleEffects <- parTable[parTable$rhs == intTerm & parTable$op == "~", , drop = FALSE]

    for (i in seq_len(NROW(simpleEffects))) {
      row   <- simpleEffects[i, ]
      cond1 <- parTable$lhs == row$lhs & parTable$rhs == m
      cond2 <- parTable$lhs == row$lhs & parTable$rhs == intTerm
      beta1 <- parTable[cond1, "label"]
      beta2 <- parTable[cond2, "label"]

      if (length(beta1) && length(beta2)) {
        parTable[cond1, "label"] <- paste0(
          "(", beta1[[1L]], "+", beta2[[1L]], "*", ".NUMERIC_VALUE__Z)"
        )
      }
    }
  }

  parTable.simple <- parTable[!grepl(":", parTable$lhs) &
                              !grepl(":", parTable$rhs), , drop = FALSE]

  paths <- getJN_PathsParTable(x = x, y = y, parTable = parTable.simple)
  mod_stopif(is.null(paths), sprintf("No paths between %s and %s were found!", x, y))

  is.direct <- attr(paths, "is.direct")

  expr_paths <- list(
    total    = paths,
    direct   = paths[is.direct],
    indirect = paths[!is.direct]
  )

  keep <- expr_paths[[type]]
  mod_stopif(!length(keep), sprintf("No %s effect between %s and %s was found!", type, x, y))
  expr <- parse(text = paste0(keep, collapse = "+"))

  V <- tryCatch(getVcovSimpleSlopes(model, standardized = standardized), error = \(e) NULL)
  coefs <- structure(parTable$est, names = labels)

  if (is.null(V)) {
    k    <- length(coefs)
    pars <- names(coefs)
    V    <- matrix(0, nrow = k, ncol = k, dimnames = list(pars, pars))
  } else {
    dimnames(V) <- list(
      stringr::str_replace_all(rownames(V), OP_REPLACEMENTS),
      stringr::str_replace_all(colnames(V), OP_REPLACEMENTS)
    )
  }

  labels.u <- unique(labels)
  V        <- expandVCOV(V, labels = labels.u)
  coefs    <- coefs[labels.u]

  draws <- mvtnorm::rmvnorm(mc.reps, mean = coefs, sigma = V)
  if (is.null(dim(draws)))
    draws <- matrix(draws, nrow = 1L)

  COEFS    <- as.data.frame(draws)

  getrow <- function(.NUMERIC_VALUE__Z) {
    slopex <- with(COEFS, eval(expr))
    ci <- stats::quantile(slopex, probs = c(sig.level / 2, 1 - sig.level / 2), na.rm = TRUE)
    point <- mean(slopex, na.rm = TRUE)
    c(point, ci, .NUMERIC_VALUE__Z)
  }

  z_range   <- seq(z_min, z_max, length.out = detail)
  jn_points <- do.call(rbind, lapply(z_range, getrow))
  colnames(jn_points) <- c("slope", "lower", "upper", "z")
  jn_points <- as.data.frame(jn_points)
  jn_points$significant <- (jn_points$lower > 0) | (jn_points$upper < 0)

  significant_everywhere <- all(jn_points$significant) || !any(jn_points$significant)

  list(
    grid = jn_points,
    jn_points = approxJN_PointsFromGrid(jn_points),
    significant_everywhere = significant_everywhere,
    mean_z = mean_z,
    sd_z = sd_z,
    min_z = z_min,
    max_z = z_max
  )
}


#' Plot Interaction Effect Using the Johnson-Neyman Technique
#'
#' This function plots the simple slopes of an interaction effect across different values of a moderator variable using the Johnson-Neyman technique. It identifies regions where the effect of the predictor on the outcome is statistically significant.
#'
#' @param x The name of the predictor variable (as a character string).
#' @param z The name of the moderator variable (as a character string).
#' @param y The name of the outcome variable (as a character string).
#' @param model A fitted model object of class \code{modsem_da}, \code{modsem_mplus}, \code{modsem_pi}, or \code{lavaan}.
#' @param min_z The minimum value of the moderator variable \code{z} to be used in the plot (default is -3). It is relative to the mean of z.
#' @param max_z The maximum value of the moderator variable \code{z} to be used in the plot (default is 3). It is relative to the mean of z.
#' @param sig.level The alpha-criterion for the confidence intervals (default is 0.05).
#' @param alpha alpha setting used in \code{ggplot} (i.e., the opposite of opacity)
#' @param detail The number of generated data points to use for the plot (default is 1000). You can increase this value for smoother plots.
#' @param sd.line A thick black line showing \code{+/- sd.line * sd(z)}. NOTE: This line will be truncated by \code{min_z} and \code{max_z} if
#' the sd.line falls outside of \code{[min_z, max_z]}.
#' @param standardized Should coefficients be standardized beforehand?
#' @param xz The name of the interaction term. If not specified, it will be created using \code{x} and \code{z}.
#' @param greyscale Logical. If \code{TRUE} the plot is plotted in greyscale.
#' @param plot.jn.points Logical. If \code{TRUE}, omit the numeric annotations for the JN-points from the plot.
#' @param type Which effect to display. One of \code{"direct"}, \code{"indirect"}, or \code{"total"}.
#' @param mc.quantiles Should JN quantiles be calculated using Monte-Carlo estimates?
#' @param mc.reps Number of Monte Carlo replicates used to approximate the confidence
#'   bands when \code{mc.quantile} is \code{TRUE} and/or \code{type} is \code{"indirect"} or \code{"total"}.
#' @param ... Additional arguments (currently not used).
#'
#' @return A \code{ggplot} object showing the interaction plot with regions of significance.
#' @details
#' The function calculates the simple slopes of the predictor variable \code{x} on the outcome variable \code{y} at different levels of the moderator variable \code{z}. It uses the Johnson-Neyman technique to identify the regions of \code{z} where the effect of \code{x} on \code{y} is statistically significant.
#'
#' When plotting indirect or total effects, the function relies on Monte Carlo
#' draws from the estimated sampling distribution of the parameters to approximate
#' the conditional effect and its confidence interval across the moderator range.
#'
#' It extracts the necessary coefficients and variance-covariance information from the fitted model object. The function then computes the critical t-value and solves the quadratic equation derived from the t-statistic equation to find the Johnson-Neyman points.
#'
#' The plot displays:
#' \itemize{
#'   \item The estimated simple slopes across the range of \code{z}.
#'   \item Confidence intervals around the slopes.
#'   \item Regions where the effect is significant (shaded areas).
#'   \item Vertical dashed lines indicating the Johnson-Neyman points.
#'   \item Text annotations providing the exact values of the Johnson-Neyman points.
#' }
#'
#' @examples
#' \dontrun{
#' library(modsem)
#'
#' m1 <-  '
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#'
#'   visual ~ speed + textual + speed:textual
#' '
#'
#' est <- modsem(m1, data = lavaan::HolzingerSwineford1939, method = "ca")
#' plot_jn(x = "speed", z = "textual", y = "visual", model = est, max_z = 6)
#' }
#' @export
plot_jn <- function(x, z, y, model, min_z = -3, max_z = 3,
                    sig.level = 0.05, alpha = 0.2, detail = 1000,
                    sd.line = 2, standardized = FALSE, xz = NULL,
                    greyscale = FALSE, plot.jn.points = TRUE,
                    type = c("direct", "indirect", "total"),
                    mc.quantiles = FALSE, mc.reps = 10000, ...) {

  mod_stopif(!inherits(model, c("modsem_da", "modsem_mplus", "modsem_pi", "lavaan")),
         "model must be of class 'modsem_pi', 'modsem_da', 'modsem_mplus', or 'lavaan'")

  type <- match.arg(type)

  if (standardized) {
    parTable <- standardized_estimates(model, correction = TRUE, colon.pi = TRUE)
  } else {
    parTable <- parameter_estimates(
      object = model,
      label.renamed.prod = TRUE,
      colon.pi = TRUE
    )
  }

  group.label <- modsem_inspect(model, what = "group.label")
  parTable <- addMissingGroups(getMissingLabels(parTable))

  plots <- list()
  for (g in getGroupsParTable(parTable)) {
    label.g    <- if (length(group.label)) group.label[[g]] else NULL
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]

    plots[[g]] <- plotJN_Group(
      x = x, z = z, y = y, parTable = parTable.g, model = model,
      min_z = min_z, max_z = max_z, sig.level = sig.level,
      alpha = alpha, detail = detail, sd.line = sd.line,
      standardized = standardized, xz = xz, greyscale = greyscale,
      plot.jn.points = plot.jn.points, group = g, group.label = label.g,
      type = type, mc.reps = mc.reps, mc.quantiles = mc.quantiles, ...
    )
  }

  if (length(plots) <= 1L)
    return(plots[[1L]])

  if (!requireNamespace("ggpubr", quietly = TRUE)) {
    printf("The `ggpubr` package is needed to arrange Johnson-Neyman plots in multigroup models!\n")
    printf("Do you want to install it? (y/n) ")
    choice <- tolower(substr(readLines(n = 1L), 1L, 1L))

    stopifnot(choice == "y")
    utils::install.packages("ggpubr")
  }

  if (requireNamespace("ggpubr", quietly = TRUE)) { # Make R CMD check happy
    ggpubr::ggarrange(plotlist = plots, labels = group.label)
  } else mod_msg_stop("The `ggpubr` package is needed to arrange Johnson-Neyman plots in multigroup models!\n")
}


getJN_GridDelta <- function(x, z, y, parTable, model, min_z, max_z, sig.level, alpha,
                            detail, sd.line, standardized, xz, greyscale,
                            plot.jn.points = TRUE, group = NULL, group.label = NULL, ...) {
  if (is.null(xz))
    xz <- paste(x, z, sep = ":")

  checkInputsSimpleSlopes(x = x, z = z, y = y, xz = xz, parTable = parTable)

  xz <- c(xz, reverseIntTerm(xz))

  # z mean/sd and plotting window
  mean_z <- getMean(z, parTable = parTable)
  sd_z   <- sqrt(calcCovParTable(x = z, y = z, parTable = parTable))
  mod_stopif(is.na(sd_z), sprintf("Variance for %s not found in model", z))
  z_min  <- min_z + mean_z
  z_max  <- max_z + mean_z

  # coefficients
  beta_x  <- parTable[parTable$lhs == y & parTable$rhs == x   & parTable$op == "~", "est"]
  beta_xz <- parTable[parTable$lhs == y & parTable$rhs %in% xz & parTable$op == "~", "est"]
  mod_stopif(length(beta_x)  == 0, "Coefficient for x not found in model")
  mod_stopif(length(beta_xz) == 0, "Coefficient for interaction term not found in model")

  VCOV <- getVcovSimpleSlopes(model, standardized = standardized)
  label_beta_x  <- parTable[parTable$lhs == y & parTable$rhs == x   & parTable$op == "~", "label"]
  label_beta_xz <- parTable[parTable$lhs == y & parTable$rhs %in% xz & parTable$op == "~", "label"]

  var_beta_x         <- VCOV[label_beta_x,  label_beta_x]
  var_beta_xz        <- VCOV[label_beta_xz, label_beta_xz]
  cov_beta_x_beta_xz <- VCOV[label_beta_x,  label_beta_xz]

  if (inherits(model, "lavaan")) {
    vcov <- lavaan::vcov
    coef <- lavaan::coef
    nobs <- lavaan::nobs
  }

  nobs <- nobs(model)
  npar <- length(coef(model))
  df_resid <- nobs - npar
  if (df_resid < 1) {
    mod_msg_warn(paste0("Degrees of freedom for residuals must be greater than 0. ",
             "Using sample size instead of degrees of freedom"))
    df_resid <- nobs
  }

  t_crit <- stats::qt(1 - sig.level / 2, df_resid)

  # Johnson–Neyman roots
  A <- beta_xz^2 - t_crit^2 * var_beta_xz
  B <- 2 * beta_x * beta_xz - 2 * t_crit^2 * cov_beta_x_beta_xz
  C <- beta_x^2 - t_crit^2 * var_beta_x
  disc <- B^2 - 4 * A * C

  significant_everywhere <- FALSE
  jn_points <- numeric(0)
  if (A == 0) {
    if (B != 0) {
      z_jn <- -C / B; z_lower <- z_jn; z_upper <- z_jn
      jn_points <- z_jn
    } else {
      significant_everywhere <- TRUE
    }

  } else if (disc < 0) {
    significant_everywhere <- TRUE

  } else if (disc == 0) {
    z_jn <- -B / (2 * A); z_lower <- z_jn; z_upper <- z_jn
    jn_points <- z_jn

  } else {
    z1 <- (-B + sqrt(disc)) / (2 * A)
    z2 <- (-B - sqrt(disc)) / (2 * A)
    z_lower <- min(z1, z2); z_upper <- max(z1, z2)
    jn_points <- c(z_lower, z_upper)
  }

  jn_points <- jn_points[is.finite(jn_points)]

  # grid and simple slopes
  z_range  <- seq(z_min, z_max, length.out = detail)
  slope    <- beta_x + beta_xz * z_range
  SE_slope <- sqrt(var_beta_x + z_range^2 * var_beta_xz + 2 * z_range * cov_beta_x_beta_xz)
  t_value  <- slope / SE_slope
  p_value  <- 2 * (1 - stats::pt(abs(t_value), df_resid))
  significant <- p_value < sig.level

  lower_all <- slope - t_crit * SE_slope
  upper_all <- slope + t_crit * SE_slope

  grid <- data.frame(
    z = z_range,
    slope = as.numeric(slope),
    lower = as.numeric(lower_all),
    upper = as.numeric(upper_all),
    significant = significant
  )

  if (!length(jn_points) && !significant_everywhere) {
    jn_points <- approxJN_PointsFromGrid(grid)
  }

  list(
    grid = grid,
    jn_points = jn_points,
    significant_everywhere = significant_everywhere,
    mean_z = mean_z,
    sd_z = sd_z,
    min_z = z_min,
    max_z = z_max
  )
}


plotJN_Group <- function(x, z, y, parTable, model, min_z, max_z, sig.level, alpha,
                         detail, sd.line, standardized, xz, greyscale,
                         plot.jn.points = TRUE, group = NULL, group.label = NULL,
                         type = c("direct", "indirect", "total"),
                         mc.reps = 10000, mc.quantiles = FALSE, ...) {

  type <- match.arg(type)

  if (type != "direct")
    mc.quantiles <- TRUE

  if (mc.quantiles) {
    result <- getJN_GridMonteCarlo( # Monte-Carlo Quantiles
      x = x, z = z, y = y, parTable = parTable, model = model, min_z = min_z,
      max_z = max_z, sig.level = sig.level, alpha = alpha, detail = detail,
      sd.line = sd.line, standardized = standardized, xz = xz, greyscale = greyscale,
      plot.jn.points = plot.jn.points, group = group, group.label = group.label,
      type = type, mc.reps = mc.reps
    )
  } else {
    result <- getJN_GridDelta( # Delta-Method (Symmetric) Quantiles
      x = x, z = z, y = y, parTable = parTable, model = model, min_z = min_z,
      max_z = max_z, sig.level = sig.level, alpha = alpha, detail = detail,
      sd.line = sd.line, standardized = standardized, xz = xz, greyscale = greyscale,
      plot.jn.points = plot.jn.points, group = group, group.label = group.label
    )
  }

  df_plot   <- result$grid
  mean_z    <- result$mean_z
  sd_z      <- result$sd_z
  z_min     <- result$min_z
  z_max     <- result$max_z
  jn_points <- result$jn_points
  significant_everywhere <- isTRUE(result$significant_everywhere)

  significant <- df_plot$significant
  sig_all <- all(significant)
  sig_any <- any(significant)

  if (sig_all || !sig_any) {
    mod_msg_note("No regions where the effect transitions between significant and non-significant.")
  }

  if (sig_any && !sig_all) {
    format_num <- function(val) formatC(val, format = "f", digits = 2)
    format_sig <- function(val) sub("^0\\.", ".", formatC(val, format = "f", digits = 2))

    transition_points <- jn_points
    transition_points <- transition_points[is.finite(transition_points)]
    transition_points <- sort(unique(transition_points))
    transition_points <- transition_points[transition_points >= z_min & transition_points <= z_max]

    run_info <- rle(significant)
    n_runs <- length(run_info$values)

    if (!length(transition_points)) {
      transition_points <- approxJN_PointsFromGrid(df_plot)
      transition_points <- transition_points[is.finite(transition_points)]
      transition_points <- transition_points[transition_points >= z_min & transition_points <= z_max]
    }

    boundaries <- c(z_min, transition_points, z_max)
    if (length(boundaries) != n_runs + 1L) {
      idx <- which(diff(as.integer(significant)) != 0)
      approx_points <- (df_plot$z[idx] + df_plot$z[idx + 1L]) / 2
      boundaries <- c(z_min, approx_points, z_max)
    }

    if (length(boundaries) == n_runs + 1L && n_runs > 1L) {
      describe_segment <- function(lower, upper, is_sig, pos, total) {
        lower_txt <- format_num(lower)
        upper_txt <- format_num(upper)
        range_txt <- if (pos == 1L) {
          sprintf("%s <= %s", z, upper_txt)
        } else if (pos == total) {
          sprintf("%s >= %s", z, lower_txt)
        } else {
          sprintf("%s is between %s and %s", z, lower_txt, upper_txt)
        }
        effect_txt <- if (is_sig) {
          sprintf("the %s effect of %s is p < %s", type, x, format_sig(sig.level))
        } else {
          sprintf("the %s effect of %s is not significant (p >= %s)", type, x, format_sig(sig.level))
        }
        sprintf("When %s, %s.", range_txt, effect_txt)
      }

      body_lines <- vapply(
        seq_len(n_runs),
        function(i) describe_segment(boundaries[[i]], boundaries[[i + 1L]],
                                     run_info$values[[i]], i, n_runs),
        character(1)
      )

      if (!is.null(group.label)) {
        header <- sprintf("Johnson-Neyman regions (group %s):", group.label)
      } else {
        header <- "Johnson-Neyman regions:"
      }
      message(sprintf("%s\n  %s", header, paste(body_lines, collapse = "\n  ")))

    } else {
      nsig <- df_plot$z[!significant]
      interval <- sprintf("[%s, %s]", format_num(min(nsig)),
                          format_num(max(nsig)))

      if (!is.null(group.label)) {
        header <- sprintf("Johnson-Neyman Interval (group %s):", group.label)
      } else {
        header <- "Johnson-Neyman Interval:"
      }

      body <- sprintf("When %s is outside the interval %s, the %s effect of %s is p < %s.",
                      z, interval, type, x, format_sig(sig.level))
      message(sprintf("%s\n  %s", header, body))
    }
  }


  # split into contiguous runs to avoid polygon bleed
  run_id <- cumsum(c(0, diff(as.integer(significant)) != 0))
  siglabel <- sprintf("p < %s", sig.level)
  significance_chr <- ifelse(significant, siglabel, "n.s.")
  Significance <- factor(significance_chr, levels = c(siglabel, "n.s."))

  x_start <- mean_z - sd.line * sd_z
  x_end   <- mean_z + sd.line * sd_z

  if (x_start < z_min && x_end > z_max) {
    mod_msg_warn("Truncating SD-range on the right and left!")
  } else if (x_start < z_min) {
    mod_msg_warn("Truncating SD-range on the left!")
  } else if (x_end > z_max) {
    mod_msg_warn("Truncating SD-range on the right!")
  }
  x_start <- max(x_start, z_min)
  x_end   <- min(x_end, z_max)
  y_start <- y_end <- 0

  hline_label <- sprintf("+/- %s SDs of %s", sd.line, z)
  data_hline <- data.frame(x_start, x_end, y_start, y_end, hline_label)

  # unified legend for colour/fill
  breaks <- c(siglabel, "n.s.", hline_label)
  values <- structure(c("cyan3", "red", "black"), names = breaks)

  y_range <- range(c(df_plot$lower, df_plot$upper, 0), na.rm = TRUE)
  if (!all(is.finite(y_range))) y_range <- c(-1, 1)

  df_plot$run_id <- run_id
  df_plot$Significance <- Significance

  # Define variables to stop R CMD check from complaining
  slope <- NULL
  lower <- NULL
  upper <- NULL

  p <- ggplot2::ggplot(df_plot, ggplot2::aes(x = z, y = slope)) +
    # single ribbon, split into runs; fill follows Significance
    ggplot2::geom_ribbon(
      ggplot2::aes(ymin = lower, ymax = upper,
                   fill = Significance, group = run_id),
      alpha = alpha, na.rm = TRUE
    ) +
    # line coloured by Significance, also split into runs for color change
    ggplot2::geom_line(
      ggplot2::aes(color = Significance, group = run_id),
      linewidth = 1, na.rm = TRUE
    ) +
    ggplot2::geom_hline(yintercept = 0, color = "black", linewidth = 0.5) +
    suppressWarnings(
      ggplot2::geom_segment(
        mapping = ggplot2::aes(x = x_start, xend = x_end, y = y_start, yend = y_end,
                               color = hline_label, fill = hline_label),
        data = data_hline, linewidth = 1.5
      )
    ) +
    ggplot2::ggtitle("Johnson-Neyman Plot") +
    ggplot2::scale_discrete_manual(
      aesthetics = c("colour", "fill"),
      name = "",
      values = values,
      breaks = breaks,
      drop = FALSE
    ) +
    ggplot2::scale_y_continuous(limits = y_range) +
    ggplot2::labs(x = z, y = paste("Slope of", x, "on", y)) +
    ggplot2::theme_minimal()

  # only show JN lines if there is a transition in the plotted window
  has_transition_in_window <- any(diff(as.integer(df_plot$significant)) != 0, na.rm = TRUE)
  if (!significant_everywhere && has_transition_in_window) {
    jn_points_in_window <- jn_points[jn_points >= z_min & jn_points <= z_max]

    if (!length(jn_points_in_window)) {
      jn_points_in_window <- approxJN_PointsFromGrid(df_plot)
      jn_points_in_window <- jn_points_in_window[jn_points_in_window >= z_min &
                                                 jn_points_in_window <= z_max]
    }

    if (length(jn_points_in_window)) {
      top_y <- suppressWarnings(max(df_plot$slope[is.finite(df_plot$slope)], na.rm = TRUE))
      if (!is.finite(top_y)) top_y <- y_range[2]

      hline_colour <- if (greyscale) "black" else "red"

      for (point in jn_points_in_window) {
        p <- p + ggplot2::geom_vline(xintercept = point, linetype = "dashed", color = hline_colour)

        if (plot.jn.points) {
          p <- p + ggplot2::annotate("text", x = point, y = top_y,
                                     label = paste("JN point:", round(point, 2)),
                                     hjust = -0.1, vjust = 1, color = "black")
        }
      }
    }
  }

  if (greyscale)
    p <- suppressMessages(p + ggplot2::scale_colour_grey() + ggplot2::scale_fill_grey())

  p
}


approxJN_PointsFromGrid <- function(grid) {
  if (is.null(grid) || !NROW(grid)) return(numeric(0))

  sig_vals <- grid$significant
  if (all(sig_vals) || !any(sig_vals)) return(numeric(0))

  transitions <- which(diff(as.integer(sig_vals)) != 0)
  if (!length(transitions)) return(numeric(0))

  approx_zero <- function(z1, z2, v1, v2) {
    if (v1 == v2) return(mean(c(z1, z2)))
    z1 - v1 * (z2 - z1) / (v2 - v1)
  }

  out <- numeric(length(transitions))

  for (i in seq_along(transitions)) {
    idx <- transitions[[i]]
    z1 <- grid$z[idx]
    z2 <- grid$z[idx + 1L]
    lower1 <- grid$lower[idx]
    lower2 <- grid$lower[idx + 1L]
    upper1 <- grid$upper[idx]
    upper2 <- grid$upper[idx + 1L]

    if ((lower1 > 0 && lower2 <= 0) || (lower1 <= 0 && lower2 > 0)) {
      out[[i]] <- approx_zero(z1, z2, lower1, lower2)
    } else if ((upper1 < 0 && upper2 >= 0) || (upper1 >= 0 && upper2 < 0)) {
      out[[i]] <- approx_zero(z1, z2, upper1, upper2)
    } else {
      out[[i]] <- mean(c(z1, z2))
    }
  }

  out[is.finite(out)]
}


getJN_PathsParTable <- function(x, y, parTable) {
  if (x == y) # exit if it's a non-recursive model
    return(NULL)

  cols <- c("lhs", "op", "rhs", "label")
  gamma <- unique(parTable[parTable$lhs == y & parTable$op == "~", cols, drop = FALSE])
  preds <- gamma[, "rhs"]

  if (!length(preds))
    return(NULL)

  paths     <- NULL
  is.direct <- NULL

  for (i in seq_along(preds)) {
    pred <- preds[[i]]

    if (pred == x) {
      paths     <- c(paths, gamma[i, "label"])
      is.direct <- c(is.direct, TRUE)

    } else {
      downstream <- getJN_PathsParTable(x = x, y = pred, parTable = parTable)

      if (!is.null(downstream)) {
        paths <- c(paths, paste0(gamma[i, "label"], "*", downstream))
        is.direct <- c(is.direct, rep(FALSE, length(downstream)))
      }
    }
  }

  attr(paths, "is.direct") <- is.direct
  paths
}

devtools::load_all()
set.seed(123)
models <- list(m1 = '
               # latent variables
               ind60 =~ x1 + x2 + x3
               dem60 =~ y1 + y2 + y3 + y4
               dem65 =~ y5 + y6 + y7 + y8
               # regressions
               dem60 ~ ind60
               dem65 ~ ind60 + dem60
               dem65 ~ ind60:dem60
               # residual covariances
               y1 ~~ y5
               y2 ~~ y4 + y6
               y3 ~~ y7
               y4 ~~ y8
               y6 ~~ y8
               ',
               m3 = ' visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9
               visual ~ speed + textual + speed:textual',
               m4 = 'visual  =~ x1 + start(0.8)*x2 + start(1.2)*x3
               textual =~ x4 + start(0.5)*x5 + start(1.0)*x6
               speed   =~ x7 + start(0.7)*x8 + start(1.8)*x9
               visual ~ speed + textual + speed:textual',
               m5 = '# three-factor model
               visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9
               visual ~ speed + textual + 0.1*speed:textual
               # intercepts with fixed values
               x1 + x2 + x3 + x4 ~ 0.5*1',
               m6 = '# three-factor model
               visual  =~ x1 + x2 + x3
               textual =~ x4 + x5 + x6
               speed   =~ x7 + x8 + x9
               visual ~ speed + textual + speed:textual
               # intercepts
               x1 ~ 1
               x2 ~ 1
               x3 ~ 1
               x4 ~ 1
               x5 ~ 1
               x6 ~ 1
               x7 ~ 1
               x8 ~ 1
               x9 ~ 1'

)



data <- list(d1 = lavaan::PoliticalDemocracy,
             d3 = lavaan::HolzingerSwineford1939,
             d4 = lavaan::HolzingerSwineford1939,
             d5 = lavaan::HolzingerSwineford1939,
             d6 = lavaan::HolzingerSwineford1939)

nativeMethods <- allNativeMethods[allNativeMethods != "pind"]
methods <- list(m1 = nativeMethods[nativeMethods != "ca"],
                m3 = nativeMethods,
                m4 = nativeMethods,
                m5 = nativeMethods,
                m6 = nativeMethods[nativeMethods != c("uca", "ca")])


estimates <- vector("list", length(models))
testthat::expect_warning({
  for (i in seq_along(estimates)) {
    model  <- models[[i]]
    data_i <- data[[i]]
    ests   <- vector("list", length(methods))
    names(ests) <- methods

    for (method in methods[[i]]) {
      ests[[method]] <- modsem(model, data_i, method=method, estimator = "ML")
    }
    estimates[[i]] <- ests
  }
}, regexp = ".*Replacing `start.*`.* in.*") # make this more informative later

# testing plot function
plot_interaction(x = "ind60", z = "dem60", y = "dem65", xz = "ind60:dem60",
                 vals_z = c(-0.5, 0.5), model = estimates[[1]][["rca"]])
plot_interaction(x = "ind60", z = "dem60", y = "dem65", xz = "ind60:dem60",
                 vals_z = c(-0.5, 0.5), model = estimates[[1]][["rca"]], standardized=TRUE)
plot_interaction(x = "speed", z = "textual", y = "visual",
                 vals_z = c(-0.5, 0.5), model = estimates[[2]][["ca"]])

testthat::expect_warning(
  plot_jn(x = "ind60", z = "dem60", y = "dem65", model = estimates[[1]][["rca"]],
          max_z = 10, min_z = -10), regex = "Degrees of freedom .*"
)
plot_jn(x = "speed", z = "textual", y = "visual", model = estimates[[2]][["ca"]],
        max_z = 6, standardized=TRUE)
plot_jn(x = "speed", z = "textual", y = "visual", model = estimates[[2]][["ca"]],
        max_z = 6)

print(summary(estimates[[1]][["rca"]], H0=FALSE))
print(estimates[[1]][["rca"]])

testthat::expect_true(summary(estimates[[1]][["rca"]], H0=FALSE)$info$version != "??")


m1 <- '
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Y ~ X + z1 + X:z1
'

testthat::expect_warning(
  modsem(m1, oneInt), regexp = "All residual covariances .*"
)

testthat::expect_no_warning(
 modsem(m1, oneInt, res.cov.method = "none")
)

testthat::expect_no_warning(
 modsem(m1, oneInt, res.cov.method = "equality")
)

m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ NA*y1 + y*y1 + y*y2 + y*y3
  Z =~ z1 + z2 + z3
  Y ~~ 1*Y
  # Inner model
  Y ~ X + Z + X:Z
'

fit <- modsem(m1, oneInt, method = "dblcent")
summary(fit)

getMethodSettingsDA <- function(method, args = NULL) {

    settings <- list(
        lms = list(verbose = interactive(),
                   optimize = TRUE,
                   nodes = 24,
                   convergence.abs = 1e-4,
                   convergence.rel = 1e-10,
                   optimizer = "nlminb",
                   center.data = FALSE,
                   standardize.data = FALSE,
                   standardize.out = FALSE,
                   standardize = FALSE,
                   mean.observed = TRUE,
                   double = FALSE,
                   calc.se = TRUE,
                   FIM = "observed",
                   OFIM.hessian = TRUE,
                   EFIM.S = 100,
                   EFIM.parametric = TRUE,
                   robust.se = FALSE,
                   R.max = 1e5,
                   max.iter = 500,
                   max.step = 1,
                   epsilon = 1e-6,
                   quad.range = Inf,
                   adaptive.quad = TRUE,
                   adaptive.quad.tol = 1e-12,
                   adaptive.frequency = 3,
                   n.threads = NULL,
                   algorithm = "EMA",
                   em.control = list(),
                   missing = "listwise",
                   orthogonal.x = FALSE,
                   orthogonal.y = FALSE,
                   auto.fix.first = TRUE,
                   auto.fix.single = TRUE,
                   auto.split.syntax = FALSE,
                   cr1s = FALSE,
                   group = NULL,
                   sampling.weights = NULL,
                   sampling.weights.normalization = "total",
                   fix.composite.var = TRUE
        ),
        qml = list(verbose = interactive(),
                   optimize = TRUE,
                   nodes = 0,
                   convergence.rel = 1e-6,
                   convergence.abs = NULL, # not relevant
                   optimizer = "nlminb",
                   center.data = FALSE,
                   standardize = FALSE,
                   standardize.data = FALSE,
                   standardize.out = FALSE,
                   mean.observed = TRUE,
                   double = FALSE,
                   calc.se = TRUE,
                   FIM = "observed",
                   OFIM.hessian = TRUE,
                   EFIM.S = 100,
                   EFIM.parametric = TRUE,
                   robust.se = FALSE,
                   R.max = 1e5,
                   max.iter = 500,
                   max.step = NULL,
                   epsilon = 1e-8,
                   quad.range = Inf,
                   adaptive.quad = FALSE,
                   adaptive.quad.tol = NULL,
                   n.threads = NULL,
                   adaptive.quad = FALSE,
                   adaptive.frequency = NULL,
                   em.control = NULL,
                   algorithm = NULL,
                   missing = "listwise",
                   orthogonal.x = FALSE,
                   orthogonal.y = FALSE,
                   auto.fix.first = TRUE,
                   auto.fix.single = TRUE,
                   auto.split.syntax = TRUE,
                   cr1s = FALSE,
                   group = NULL,
                   sampling.weights = NULL,
                   sampling.weights.normalization = "total",
                   fix.composite.var = FALSE
        )
    )

    if (is.null(args)) return(settings[method])

    settingNames <- unique(unlist(lapply(settings, FUN = names)))
    args <- args[settingNames]
    isMissing <- vapply(args, FUN.VALUE = logical(1L), FUN = is.null)
    missingArgs <- settingNames[isMissing]

    mod_stopif(!method %in% names(settings), "Unrecognized method")

    args.out <- c(settings[[method]][missingArgs], args[!isMissing])

    args.out$standardize.data <-
      args.out$standardize || args.out$standardize.data
    args.out$standardize.out <-
      args.out$standardize || args.out$standardize.out
    args.out$mean.observed <-
      !args.out$standardize && args.out$mean.observed
    args.out$OFIM.hessian <-
      args.out$OFIM.hessian && !args.out$robust.se
    args.out$center.data <- !args.out$standardize.data &&
      !args.out$mean.observed

    if (is.null(args.out$group) &&
        tolower(args.out$sampling.weights.normalization) == "group")
      args.out$sampling.weights.normalization <- "total"

    args.out
}



#' default arguments fro LMS and QML approach
#'
#' @param method which method to get the settings for
#' @return list
#' @export
#' @description
#' This function returns the default settings for the LMS and QML approach.
#' @examples
#' library(modsem)
#' default_settings_da()
default_settings_da <- function(method = c("lms", "qml")) {
  getMethodSettingsDA(method = method, args = NULL)
}

prepDataModsemDA <- function(data, allIndsXis, allIndsEtas, missing = "listwise",
                             cluster = NULL, sampling.weights = NULL) {

  if (is.null(data) || !NROW(data))
    return(list(data.full = NULL, n = 0, k = 0, p = 0, cluster = NULL, idx = NULL))

  if (!is.null(cluster)) {
    mod_stopif(length(cluster) > 1L, "`cluster` must be a single variable!")

    CLUSTER <- as.factor(data[, cluster])

  } else CLUSTER <- NULL

  if (!is.null(sampling.weights)) {
    mod_stopif(length(sampling.weights) > 1L, "`sampling.weights` must be a single variable!")
    WEIGHTS <- data[, sampling.weights]

  } else WEIGHTS <- NULL

  sortData(data, allIndsXis, allIndsEtas) |>
    castDataNumericMatrix() |>
    handleMissingData(missing = missing, CLUSTER = CLUSTER, WEIGHTS = WEIGHTS) |>
    patternizeMissingDataFIML()
}


indexData <- function(data) {
  attr(data, "idx") 
}

sortData <- function(data, allIndsXis, allIndsEtas) {
  inds    <- unique(c(allIndsXis, allIndsEtas))
  ovs     <- colnames(data)
  missing <- inds[!inds %in% ovs]

  mod_stopif(!all(inds %in% ovs), paste0("Missing observed variables in data:\n  ",
         paste(missing, collapse = ", ")))

  data[unique(c(allIndsXis, allIndsEtas))]
}


anyAllNA <- function(data) {
  any(vapply(data, FUN.VALUE = logical(1L), function(x) all(is.na(x))))
}


castDataNumericMatrix <- function(data) {
  force(data) # evaluate to check for errors
  data <- tryCatch({
    numericData <- lapplyDf(data, FUN = as.numeric)
  },
  warning = function(w) {
    mod_msg_warn(paste0("Warning in converting data to numeric: \n", w))
    numericData <- suppressWarnings(lapplyDf(data, FUN = as.numeric))
    mod_stopif(anyAllNA(numericData), "Unable to convert data to type numeric")
    numeric
  },
  error = function(e) {
    mod_msg_stop("Unable to convert data to type numeric")
  })
  as.matrix(data)
}


patternizeMissingDataFIML <- function(data) {
  # if we are not using fiml, the missing data should already have been removed...
  CLUSTER <- attr(data, "cluster")
  WEIGHTS <- attr(data, "weights")
  INDEX   <- attr(data, "index")

  Y   <- as.matrix(data)
  obs <- !is.na(Y)

  rowMissingAll <- apply(obs, MARGIN = 1, FUN = \(x) !any(x))
  colMissingAll <- apply(obs, MARGIN = 2, FUN = \(x) !any(x))
  mod_stopif(any(colMissingAll),
         paste0("Please remove variables with only missing values:\n  ",
         paste0(colnames(obs)[colMissingAll], collapse = ", ")))

  patterns <- unique(obs, MARING = 2)

  if (any(rowMissingAll)) { # remove patterns where all are missing
    # This shouldn't really happen, as it should be handled already in
    # `handleMissingData()`. Regardless, we should handle it if it ever happens
    return(patternizeMissingDataFIML(data[!rowMissingAll, , drop = FALSE]))
  }

  ids <- seq_len(NROW(patterns))
  n   <- NROW(Y)
  k   <- NCOL(Y)

  rowidx <- vector("list", length = NROW(ids))
  colidx <- vector("list", length = NROW(ids))
  data.split  <- vector("list", length = NROW(ids))
  n.pattern  <- numeric(NROW(ids))
  d.pattern  <- numeric(NROW(ids))

  for (id in ids) {
    mask  <- matrix(patterns[id, ], nrow = n, ncol = k, byrow = TRUE)
    match <- apply(obs == mask, MARGIN = 1, FUN = all)
    ridx  <- which(match)
    cidx  <- which(patterns[id, ])

    rowidx[[id]] <- ridx
    colidx[[id]] <- cidx
    data.split[[id]] <- Y[ridx, cidx, drop = FALSE]
    n.pattern[[id]] <- sum(match)
    d.pattern[[id]] <- length(cidx)
  }

  # Everything downstream (posterior matrices, densities, score contributions)
  # is in pattern-concatenated order (i.e., the stacked rows of `data.split`),
  # so the row-level vectors must be permuted accordingly. With a single
  # pattern, `perm` is the identity.
  permutation <- unlist(rowidx)

  list(
    ids        = ids,
    rowidx     = rowidx,
    colidx     = colidx,
    colidx0    = lapply(colidx, FUN = \(idx) idx - 1),
    patterns   = patterns,
    data.split = data.split,
    n.pattern  = n.pattern,
    d.pattern  = d.pattern,
    n          = NROW(data),
    k          = NCOL(data),
    p          = length(ids),
    data.full  = data,
    is.fiml    = length(ids) > 1L,
    cluster    = CLUSTER[permutation],
    weights    = WEIGHTS[permutation],
    index      = INDEX[permutation]
  )
}


handleMissingData <- function(data, missing = "listwise", CLUSTER = NULL, WEIGHTS = NULL) {
  missing       <- tolower(missing)
  completeCases <- stats::complete.cases(data)
  anyMissing    <- any(!completeCases)
  allMissing    <- all(!completeCases)
  INDEX         <- seq_len(NROW(data))

  if (!anyMissing){
    attr(data, "cluster") <- CLUSTER
    attr(data, "weights") <- WEIGHTS
    attr(data, "index")   <- INDEX
    return(data)

  } else if (allMissing) {
    missingAllCol <- apply(data, MARGIN = 2, FUN = \(x) all(is.na(x)))
    colsMissing   <- colnames(data)[missingAllCol]

    mod_msg_stop(paste0("Please remove variables with only missing values:\n  ",
          paste0(colsMissing, collapse = ", ")))
  }

  if (missing %in% c("listwise", "casewise", "complete")) {
    mod_msg_warn(paste0("Removing missing values list-wise!\n",
             "Consider using `missing=\"fiml\"`, `missing=\"impute\"`, ",
             "or the `modsem_mimpute()` function!\n"))

    out <- data[completeCases, ]
    attr(out, "cluster") <- CLUSTER[completeCases]
    attr(out, "weights") <- WEIGHTS[completeCases]
    attr(out, "index")   <- INDEX[completeCases]

    return(out)

  } else if (missing == "impute") {
    mod_msg_note("Imputing missing values. Consider using the `modsem_mimpute()` function!")

    imp  <- Amelia::amelia(data, m = 1, p2s = 0)

    imp1 <- as.matrix(as.data.frame(imp$imputations[[1]]))
    attr(imp1, "cluster") <- CLUSTER
    attr(imp1, "weights") <- WEIGHTS
    attr(imp1, "index")   <- INDEX

    return(imp1)

  } else if (missing %in% c("fiml", "ml", "direct")) {
    rowMissingAll <- apply(data, MARGIN = 1, FUN = \(x) all(is.na(x)))
    keep          <- !rowMissingAll

    data <- data[keep, , drop = FALSE] # we already know that !all(rowMissingAll)

    # NOTE: attributes must be set after subsetting, as subsetting a
    # matrix drops any custom attributes!
    attr(data, "cluster") <- CLUSTER[keep]
    attr(data, "weights") <- WEIGHTS[keep]
    attr(data, "index")   <- INDEX[keep]

    return(data)

  } else {
    mod_msg_stop(sprintf("Unrecognized value for `missing`: `%s`", missing))
  }
}

simulateDataParTable <- function(parTable, N, colsOVs = NULL, colsLVs = NULL, seed = NULL) {
  if (!is.null(seed) && exists(".Random.seed")) .Random.seed.orig <- .Random.seed
  else                                          .Random.seed.orig <- NULL

  on.exit({
    if (!is.null(.Random.seed.orig)) .Random.seed <<- .Random.seed.orig
  })

  if (!is.null(seed))
    set.seed(seed)

  parTable <- addMissingGroups(parTable)
  groups   <- getGroupsParTable(parTable)

  LV <- vector("list", length = length(groups))
  OV <- vector("list", length = length(groups))

  N.g <- floor(N / length(groups))

  for (g in groups) {
    simg <- simulateDataParTableGroup(
      parTable = parTable[parTable$group == g, , drop = FALSE],
      N        = N.g,
      colsOVs  = colsOVs,
      colsLVs  = colsLVs
    )

    LV[[g]] <- simg$lV
    OV[[g]] <- simg$oV
  }

  list(LV = LV, OV = OV)
}


simulatedGroupsToDf <- function(sim, type = "OV") {
  simt <- sim[[toupper(type)]]

  do.call(rbind, lapply(seq_along(simt), FUN = \(g)
                        cbind(as.data.frame(simt), data.frame(group = g))))
}


simulateDataParTableGroup <- function(parTable, N, colsOVs = NULL, colsLVs = NULL) {
  # Output columns
  if (is.null(colsOVs)) colsOVs <- getOVs(parTable)
  if (is.null(colsLVs)) colsLVs <- getLVs(parTable)

  # Here we simplify the model by converting all =~ to ~
  lhs <- parTable$lhs
  op  <- parTable$op
  rhs <- parTable$rhs

  # Thus we can treat indicators the same as endogenous variables
  idx <- which(op == "=~")
  parTable[idx, "lhs"] <- rhs[idx]
  parTable[idx, "op"]  <- "~"
  parTable[idx, "rhs"] <- lhs[idx]

  # endogenous variables (etas)
  etas <- getSortedEtas(parTable, isLV = FALSE, checkAny = TRUE)
  xis  <- getXis(parTable, checkAny = TRUE, isLV = FALSE)
  vars <- union(xis, etas)

  # sample disturbances
  Psi <- matrix(
    0, nrow = length(vars), ncol = length(vars),
    dimnames = list(vars, vars)
  )

  psi.idx <- which(
    parTable$lhs %in% vars & parTable$op == "~~" & parTable$rhs %in% vars
  )

  for (i in psi.idx) {
    lhs <- parTable[i, "lhs"]
    rhs <- parTable[i, "rhs"]
    est <- parTable[i, "est"]
    Psi[lhs, rhs] <- Psi[rhs, lhs] <- est
  }

  alpha <- getIntercepts(vars, parTable = parTable)

  # Keep a copy of Zeta
  Zeta <- as.matrix(mvtnorm::rmvnorm(n = N, mean = alpha, sigma = Psi))
  colnames(Zeta) <- vars
  Eta <- Zeta

  for (eta in etas) {
    idx <- which(parTable$lhs == eta & parTable$op == "~")

    y <- Eta[,eta, drop = TRUE]
    for (i in idx) {
      row <- parTable[i, , drop = FALSE]
      gamma <- row$est
      x <- getValuesFromRhs(rhs = row$rhs, mat = Eta)
      y <-  y + gamma * x
    }

    Eta[,eta] <- y
  }

  dataOVs <- Eta[,colsOVs, drop = FALSE]
  dataLVs <- Eta[,colsLVs, drop = FALSE]

  list(oV = dataOVs, lV = dataLVs)
}


getValuesFromRhs <- function(rhs, mat) {
  mod_stopif(!length(rhs), "rhs is of length zero!")

  if (!grepl(":", rhs))
    return(mat[,rhs, drop = TRUE])

  elems <- stringr::str_split_1(rhs, pattern = ":")

  if (length(elems) == 1L) # should have been caught earlier, but just in case...
    return(mat[,elems, drop = TRUE])

  mod_stopif(!length(elems),
    "zero elements in product:", rhs
  )

  mod_stopif(!all(elems %in% colnames(mat)),
    "Missing values for elements in product!",
    paste0(setdiff(elems, colnames(mat)), collapse = ", ")
  )

  xz <- rep(1, NROW(mat))

  for (x in elems)
    xz <- xz * mat[,x, drop = TRUE]

  xz
}

devtools::load_all()

m1 <- '
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + a * X:Z
  c := 10 * a
'

lms1 <- modsem(m1, oneInt, method = "lms", adaptive.quad=TRUE)

tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

lms2 <- modsem(tpb, TPB, method = "lms", nodes = 32, adaptive.quad=TRUE)
summary(lms2)


tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att3 + att2 + att1 + att4
  SN =~ sn4 + sn2 + sn3 + sn1
  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
  INT =~ int2 + int1 + int3 + int4
  BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
"

lms3 <- modsem(tpb_uk, data = TPB_UK, "lms",
               nodes=32, FIM="observed", robust.se = TRUE,
               adaptive.quad=TRUE, algorithm ="EMA")
print(summary(lms3))
#> Regressions:
#>                   Estimate  Std.Error  z.value  P(>|z|)
#>    INT ~
#>      PBC             1.037      0.036    28.46    0.000
#>      ATT            -0.060      0.030    -2.04    0.041
#>      SN              0.051      0.033     1.55    0.120
#>    BEH ~
#>      PBC             0.398      0.052     7.63    0.000
#>      INT             0.594      0.049    12.24    0.000
#>      PBC:INT         0.141      0.008    17.66    0.000

# Compared with Mplus
#> Regressions:
#>                    Estimate  Std.Error  z.value  Pr(>|z|)
#>   INT ~
#>     ATT              -0.053      0.031    -1.71     0.089
#>     SN               -0.065      0.024    -2.71     0.008
#>     PBC               1.090      0.036    30.28     0.000
#>   BEH ~
#>     PBC               0.405      0.052     7.79     0.000
#>     INT               0.588      0.048    12.25     0.000
#>     INT:PBC           0.141      0.008    17.62     0.000


nlsem <- '
ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
CAREER =~ career1 + career2 + career3 + career4
SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
'

# For such a small number of nodes it doesn't really matter if you use an
# adaptive quadrature, as all the nodes bring some value
# no warning for low number of nodes here, when using adaptive quadrature
lms4 <- modsem(nlsem, data = jordan, method = "lms",
               adaptive.quad=TRUE, OFIM.hessian = FALSE,
               nodes = 15, mean.observed = FALSE)
summary(lms4)

getMethodSettingsPI <- function(method, args) {
  defaultResCov <- "simple"
  settings <- list(
    rca = list(
        center.before = FALSE,
        center.after = FALSE,
        residuals.prods = TRUE,
        residual.cov.syntax = TRUE,
        constrained.prod.mean = FALSE,
        constrained.loadings = FALSE,
        constrained.var = FALSE,
        res.cov.method = defaultResCov,
        res.cov.across = TRUE,
        match = FALSE,
        match.recycle = FALSE),
    uca  = list(
        center.before = TRUE,
        center.after = FALSE,
        residuals.prods = FALSE,
        residual.cov.syntax = TRUE,
        constrained.prod.mean = TRUE,
        constrained.loadings = FALSE,
        constrained.var = FALSE,
        res.cov.method = defaultResCov,
        res.cov.across = TRUE,
        match = FALSE,
        match.recycle = FALSE),
    pind  = list(
        center.before = FALSE,
        center.after = FALSE,
        residuals.prods = FALSE,
        residual.cov.syntax = FALSE,
        constrained.prod.mean = FALSE,
        constrained.loadings = FALSE,
        constrained.var = FALSE,
        res.cov.method = defaultResCov,
        res.cov.across = TRUE,
        match = FALSE,
        match.recycle = FALSE),
    dblcent  = list(
        center.before = TRUE,
        center.after = TRUE,
        residuals.prods = FALSE,
        residual.cov.syntax = TRUE,
        constrained.prod.mean = FALSE,
        constrained.loadings = FALSE,
        constrained.var = FALSE,
        res.cov.method = defaultResCov,
        res.cov.across = TRUE,
        match = FALSE,
        match.recycle = FALSE),
    ca = list(
        center.before = TRUE,
        center.after = FALSE,
        residuals.prods = FALSE,
        residual.cov.syntax = TRUE,
        constrained.prod.mean = TRUE,
        constrained.loadings = TRUE,
        constrained.var = TRUE,
        res.cov.method = "ca",
        res.cov.across = FALSE,
        match = TRUE,
        match.recycle = TRUE)
    )

  settingNames <- unique(unlist(lapply(settings, FUN = names)))
  args <- args[settingNames]

  if (is.null(args)) return(settings[method])

  isMissing <- vapply(args, FUN.VALUE = logical(1L), FUN = is.null)
  missingArgs <- settingNames[isMissing]
  mod_stopif(!method %in% names(settings), "Unrecognized method")

  c(settings[[method]][missingArgs], args[!isMissing])
}


#' default arguments for product indicator approaches
#'
#' @param method which method to get the settings for
#' @return list
#' @export
#' @description
#' This function returns the default settings for the product indicator approaches
#' @examples
#' library(modsem)
#' default_settings_pi()
default_settings_pi <- function(method = c("rca", "uca", "pind", "dblcent", "ca")) {
  getMethodSettingsPI(method = method, args = NULL)
}

# functions for computing the constrained covariance matrix,
# based on causal relationships. This can make the lms method more flexible,
# as you can split the model into a non-linear, and linear part. allowing
# you to use (normally distributed) endogenous variables as non-normal
# as of now the mean-structure is excluded
covModel <- function(syntax = NULL,
                     method = "lms",
                     parTable = NULL,
                     xis.main = NULL,
                     parTable.main = NULL,
                     orthogonal.x = FALSE,
                     orthogonal.y = FALSE) {
  if (is.null(parTable) && !is.null(syntax)) parTable <- modsemify(syntax)
  if (is.null(parTable)) {
    return(list(matrices = NULL, freeParams = 0, info = list(etas = NULL, xis = NULL),
                theta = NULL, syntax = NULL, parTable = NULL))
  }

  parTable.full <- rbind(parTable.main, parTable)

  if (NROW(parTable) && any(parTable$op == "~")) {
    etas     <- getSortedEtas(parTable, isLV = FALSE, checkAny = TRUE)
    numEtas  <- length(etas)
    xis      <- getXis(parTable, checkAny = TRUE, isLV = FALSE)
    xis      <- unique(c(xis, xis.main[!xis.main %in% etas]))

    numXis   <- length(xis)
    isSimple <- FALSE
  } else {
    etas     <- character(0L)
    numEtas  <- 0L
    xis      <- xis.main
    numXis   <- length(xis)
    isSimple <- TRUE

    parTable <- parTable.main[
      (parTable.main$lhs %in% xis.main &
       parTable.main$rhs %in% xis.main &
       parTable.main$op == "~~") |
       parTable.main$op %in% c(":=", "=="), , drop = FALSE
    ]
  }

  # Gamma, i.e., [[gammaEta, gammaXi], [0, 0]] over c(etas, xis).
  # The xi rows are all zero, as xis are exogenous by construction
  listGamma <- constructGamma(c(etas, xis), c(etas, xis), parTable = parTable)
  gamma <- listGamma$numeric
  labelGamma <- listGamma$label

  # covariance matrices
  # We need the full partable to look for all relevant covariances
  listPsi <- constructPsi(
    c(etas, xis),
    parTable = parTable.full,
    orthogonal.y = orthogonal.y,
    orthogonal.x = orthogonal.x,
    has.exo = TRUE
  )

  psi <- listPsi$numeric
  labelPsi <- listPsi$label

  matrices <- list(gamma = gamma, psi = psi)

  labelMatrices <- list(gamma = labelGamma, psi = labelPsi)

  model <- list(
    info = list(
      etas = etas,
      numEtas = numEtas,
      xis = xis,
      numXis = numXis,
      is.simple = isSimple
    ),
    matrices = matrices,
    labelMatrices = labelMatrices,
    syntax = syntax,
    parTable = parTable
  )

  model
}


countFreeCovModel <- function(matrices) {
  vapply(matrices, FUN.VALUE = integer(1L),
         FUN = function(x) sum(is.na(x))) |> sum()
}


expectedCovModel <- function(model, method = "lms", sortedXis) {
  gamma <- model$matrices$gamma
  psi <- model$matrices$psi

  # `gamma` is all-zero when `is.simple`, in which case this is just the identity
  Binv <- solve(diag(nrow(gamma)) - gamma)

  sigma <- Binv %*% psi %*% t(Binv)
  sigma <- sigma[sortedXis, sortedXis]

  if (method == "lms") {
    sigma <- tryCatch(t(chol(sigma)),
                      error = function(e) {
                        sigma[TRUE] <- NaN
                        sigma
                      })
  }
  sigma
}


covModelToParTable <- function(model, method = "lms") {
  matricesEst   <- model$covModel$matrices
  matricesSE    <- model$covModelSE$matrices
  matricesNA    <- model$covModelNA$matrices
  matricesLabel <- model$covModel$labelMatrices

  if (is.null(matricesEst) || is.null(matricesNA)) return(NULL)
  if (is.null(matricesSE)) matricesSE <- matricesNA

  etas <- model$info$etas
  numXis <- model$info$numXis
  parTable <- NULL

  if (!model$covModel$info$is.simple) {
    # coefficients Structural Model
    newRows <- matrixToParTable(matricesNA$gamma,
                                matricesEst$gamma,
                                matricesSE$gamma,
                                matricesLabel$gamma,
                                op = "~",
                                rowsLhs = TRUE)
    parTable <- rbind(parTable, newRows)
  }

  newRows <- matrixToParTable(matricesNA$psi,
                              matricesEst$psi,
                              matricesSE$psi,
                              matricesLabel$psi,
                              op = "~~",
                              rowsLhs = FALSE,
                              symmetric = TRUE)
  parTable <- rbind(parTable, newRows)


  parTable <- lapplyDf(parTable, FUN = function(x) replace(x, x == -999, NA))
  # return
  parTable
}

PKG_INFO <- rlang::env(version = NULL)


UNICODE_MSG_STRINGS <- list(
  updateStatusLog0 = "\rIter=%d Mode=%s LogLik=%.2f \u0394LL=%.2g rel\u0394LL=%.2g",
  incrementIterations0 = "\rEval=%d LogLik=%.2f \u0394LL=%.2g rel\u0394LL=%.2g"
)


ASCII_MSG_STRINGS <- list(
  updateStatusLog0 = "\rIter=%d Mode=%s LogLik=%.2f dLL=%.2g reldLL=%.2g",
  incrementIterations0 = "\rEval=%d LogLik=%.2f dLL=%.2g reldLL=%.2g"
)


MSG_STRINGS <- rlang::env(
  unicode = FALSE,
  strings = ASCII_MSG_STRINGS
)


# Box-drawing characters for table output (see simple_slopes.R)
UNICODE_BOX_CHARS <- list(
  V_LINE    = "\u2502", # vertical line
  H_LINE    = "\u2500", # horizontal line
  D_CROSS   = "\u252c", # down cross
  LU_CORNER = "\u250c", # left upper corner
  LL_CORNER = "\u2514", # left lower corner
  RU_CORNER = "\u2510", # right upper corner
  RL_CORNER = "\u2518", # right lower corner
  H_DLINE   = "\u2550", # horizontal double line
  F_DCROSS  = "\u256a", # full double cross
  R_DCROSS  = "\u2561", # right double cross
  L_DCROSS  = "\u255e", # left double cross
  U_CROSS   = "\u2534"  # up cross
)


ASCII_BOX_CHARS <- list(
  V_LINE    = "|",
  H_LINE    = "-",
  D_CROSS   = "+",
  LU_CORNER = "+",
  LL_CORNER = "+",
  RU_CORNER = "+",
  RL_CORNER = "+",
  H_DLINE   = "=",
  F_DCROSS  = "+",
  R_DCROSS  = "+",
  L_DCROSS  = "+",
  U_CROSS   = "+"
)


BOX_CHARS <- rlang::env(chars = ASCII_BOX_CHARS)


getPackageVersion <- function(pkgname) {
  tryCatch({
    c(read.dcf(
      file = system.file("DESCRIPTION", package = pkgname),
      fields = "Version"
    ))
  }, error = function(e) {
    mod_msg_warn("Failed to get package version")
    "??" # replace this with a hard-coded value?
  })
}


.onLoad <- function(libname, pkgname) {
  PKG_INFO$version <- getPackageVersion(pkgname)

  # Are unicode characters supported?
  utf8 <- tryCatch(isTRUE(l10n_info()[["UTF-8"]]), error = \(e) FALSE)
  MSG_STRINGS$unicode <- utf8

  if (utf8) MSG_STRINGS$strings <- UNICODE_MSG_STRINGS
  else      MSG_STRINGS$strings <- ASCII_MSG_STRINGS

  if (utf8) BOX_CHARS$chars <- UNICODE_BOX_CHARS
  else      BOX_CHARS$chars <- ASCII_BOX_CHARS

  resetModsemColors()
}


.onAttach <- function(libname, pkgname) {
  version <- getPackageVersion(pkgname)
  message <- sprintf("This is %s (%s). Please report any bugs!", pkgname, version)

  openMP_Enabled <- checkOpenMP_Cpp()

  if (!openMP_Enabled) {
    message <- paste0(message, "\n",
                     "OpenMP is not available! Multi-threading will not work properly!")
  }

  packageStartupMessage(message)
}

#' @export
#' @param high.order.as.measr Should higher order measurement model be
#'   denoted with the \code{=~} operator? If \code{FALSE} the \code{~}
#'   operator is used.
#' @param rm.tmp.ov Should temporary (hidden) variables be removed?
#' @param label.renamed.prod Should renamed product terms keep their old (implicit) labels?
#' @param is.public Should public version of parameter table be returned?
#'   If \code{FALSE}, the internal version of the parameter table is returned.
#' @describeIn parameter_estimates Get parameter estimates of a \code{\link{modsem_da}} object
parameter_estimates.modsem_da <- function(object,
                                          high.order.as.measr = TRUE,
                                          is.public = TRUE,
                                          rm.tmp.ov = is.public,
                                          label.renamed.prod = NULL, # capture
                                          ...) {
  parTable <- object$parTable

  if (high.order.as.measr) {
    indsHigherOrderLVs <- object$model$info$indsHigherOrderLVs
    parTable <- higherOrderStruct2Measr(parTable = parTable,
                                        indsHigherOrderLVs = indsHigherOrderLVs)
    parTable <- sortParTableDA(parTable, model = object$model)
  }

  if (rm.tmp.ov)
    parTable <- removeTempOV_RowsParTable(parTable)

  modsemParTable(parTable, is.public = is.public)
}


#' summary for modsem objects
#'
#' @param object modsem object to summarized.
#' @param H0 Should a null model be estimated (used for comparison).
#' @param verbose Print progress for the estimation of null model.
#' @param r.squared Calculate R-squared.
#' @param fit Print additional fit measures.
#' @param adjusted.stat Should sample size corrected/adjustes AIC and BIC be reported?
#' @param digits Number of digits to print.
#' @param scientific Print p-values in scientific notation.
#' @param ci Print confidence intervals.
#' @param standardized Print standardized estimates.
#' @param centered Print mean centered estimates.
#' @param monte.carlo Should Monte Carlo bootstrapped standard errors be used? Only
#'   relevant if \code{standardized = TRUE}. If \code{FALSE} delta method is used instead.
#' @param mc.reps Number of Monte Carlo repetitions. Only relevant if \code{monte.carlo = TRUE},
#'   and \code{standardized = TRUE}.
#' @param loadings Print loadings.
#' @param regressions Print regressions.
#' @param covariances Print covariances.
#' @param intercepts Should intercepts be included in the output?
#' If \code{standardized = TRUE} intercepts will by default be excluded.
#' @param variances Print variances.
#' @param thresholds Print thresholds.
#' @param var.interaction If FALSE variances for interaction terms will be removed
#' from the output.
#' @param ... Additional arguments.
#' @rdname summary
#' @export
#' @examples
#' \dontrun{
#' m1 <- "
#'  # Outer Model
#'  X =~ x1 + x2 + x3
#'  Y =~ y1 + y2 + y3
#'  Z =~ z1 + z2 + z3
#'
#'  # Inner model
#'  Y ~ X + Z + X:Z
#' "
#'
#' est1 <- modsem(m1, oneInt, "qml")
#' summary(est1, ci = TRUE, scientific = TRUE)
#' }
summary.modsem_da <- function(object,
                              H0 = is_interaction_model(object),
                              verbose = interactive(),
                              r.squared = TRUE,
                              fit = FALSE,
                              adjusted.stat = FALSE,
                              digits = 3,
                              scientific = FALSE,
                              ci = FALSE,
                              standardized = FALSE,
                              centered = FALSE,
                              monte.carlo = FALSE,
                              mc.reps = 10000,
                              loadings = TRUE,
                              regressions = TRUE,
                              covariances = TRUE,
                              intercepts = TRUE,
                              variances = TRUE,
                              thresholds = TRUE,
                              var.interaction = FALSE,
                              ...) {
  method   <- object$method
  parTable <- parameter_estimates(object)
  finalModel <- object$model
  groupModels <- finalModel$models

  if (is.null(groupModels) || !length(groupModels))
    groupModels <- list(list(data = object$data))

  n.groups <- length(groupModels)
  group.labels <- finalModel$info$group.levels
  if (is.null(group.labels) || length(group.labels) != n.groups) {
    model.names <- names(groupModels)
    if (!is.null(model.names) && length(model.names) == n.groups &&
        all(nzchar(model.names))) {
      group.labels <- model.names
    } else {
      group.labels <- paste0("Group ", seq_len(n.groups))
    }
  }

  normalize_r2 <- function(x) {
    if (is.null(x)) return(NULL)
    vec <- unclass(x)
    nm  <- names(vec)
    vec <- as.numeric(vec)
    names(vec) <- nm
    vec
  }

  aggregate_r2 <- function(r2, weights) {
    if (is.null(r2)) return(NULL)
    if (!is.list(r2)) return(normalize_r2(r2))
    if (!length(r2)) return(NULL)
    r2_names <- names(r2)
    if (!is.null(names(weights)) &&
        !is.null(r2_names) &&
        length(r2_names) == length(r2) &&
        all(r2_names %in% names(weights))) {
      weights <- weights[r2_names]
    } else {
      weights <- weights[seq_along(r2)]
    }
    vars <- unique(unlist(lapply(r2, names)))
    vars <- vars[!is.na(vars) & nzchar(vars)]
    if (!length(vars)) return(NULL)
    agg <- stats::setNames(numeric(length(vars)), vars)
    total_w <- sum(weights, na.rm = TRUE)
    if (!is.finite(total_w) || total_w <= 0) {
      total_w <- length(r2)
      weights[] <- 1
    }
    for (i in seq_along(r2)) {
      vec <- normalize_r2(r2[[i]])
      if (is.null(vec) || !length(vec)) next
      w <- weights[[i]]
      if (!is.finite(w) || w < 0) w <- 0
      agg[names(vec)] <- agg[names(vec)] + vec * w
    }
    agg / total_w
  }

  extract_n <- function(dat) {
    if (is.null(dat)) return(0)
    if (!is.null(dat$n)) return(dat$n)
    if (!is.null(dat$data.full)) return(NROW(dat$data.full))
    if (!is.null(dat$data)) return(NROW(dat$data))
    0
  }

  extract_patterns <- function(dat) {
    if (is.null(dat)) return(0L)
    if (isTRUE(dat$is.fiml)) {
      if (!is.null(dat$p)) return(dat$p)
      if (!is.null(dat$ids)) return(length(dat$ids))
      return(1L)
    }
    0L
  }

  group.Ns <- vapply(groupModels, function(sub) extract_n(sub$data), numeric(1L))
  fiml.counts <- vapply(groupModels, function(sub) extract_patterns(sub$data), numeric(1L))
  fiml.flags  <- fiml.counts > 0L

  extra.cols <- NULL
  if (standardized) {
    std.col     <- "Std.all"
    extra.cols <- c(extra.cols, std.col)

    parTable <- addTransformedEstimatesPT(
      parTable    = parTable,
      values.from = "est",
      values.to   = std.col,
      FUN         = standardized_estimates,
      object      = object,
      monte.carlo = monte.carlo,
      mc.reps     = mc.reps
    )
  }

  if (centered) {
    cnt.col     <- "Cnt.all"
    extra.cols <- c(extra.cols, cnt.col)

    parTable <- addTransformedEstimatesPT(
      parTable    = parTable,
      values.from = "est",
      values.to   = cnt.col,
      FUN         = centered_estimates,
      object      = object,
      monte.carlo = monte.carlo,
      mc.reps     = mc.reps
    )
  }

  if (!var.interaction)
    parTable.out <- removeInteractionVariances(parTable)
  else
    parTable.out <- parTable

  args <- object$args
  out <- list(
    parTable        = parTable.out,
    data            = NULL,
    iterations      = object$iterations,
    logLik          = object$logLik,
    fit             = fitModsemDA_Internal(object, chisq = FALSE),
    D               = NULL,
    N               = structure(group.Ns, names = group.labels),
    N.total         = sum(group.Ns),
    method          = method,
    optimizer       = object$optimizer,
    quad            = object$info.quad,
    type.se         = object$type.se,
    type.estimates  = ifelse(standardized, "standardized", object$type.estimates),
    information     = object$information,
    is.fiml         = any(fiml.flags),
    n.fiml.patterns = structure(fiml.counts, names = group.labels),
    group.labels    = group.labels,
    npar            = length(coef(object, type = "free")),
    convergence.msg = object$convergence.msg,
    naive.fit.measures = !is.null(object$ordered.mc),
    ordered.mc = object$ordered.mc
  )

  if (H0) {
    if (any(grepl(":", parTable$rhs)) && verbose)
      cat("Estimating baseline model (H0)\n")

    est_h0 <- estimateH0Cached(object, calc.se = FALSE, warn_no_interaction = FALSE)

    out$nullModel <- est_h0
    if (is.null(est_h0)) {
      mod_msg_warn("Comparative fit to H0 will not be calculated.")
      H0        <- FALSE
      out$D     <- NULL
      out$fitH0 <- NULL

    } else {
      out$D     <- compare_fit(est_h1 = object, est_h0 = est_h0)
      out$fitH0 <- fitModsemDA_Internal(est_h0, lav.fit = TRUE)
      out$naive.fit.measures.h0 <- !is.null(est_h0$ordered.mc)
    }
  } else {
    out$D <- NULL
    out$naive.fit.measures.h0 <- FALSE
  }

  if (r.squared) {
    r2.values <- modsem_inspect(object, "r2.lv")
    out$r.squared <- aggregate_r2(r2.values, group.Ns)

    if (H0) {
      r2.h0.values <- modsem_inspect(est_h0, "r2.lv")
      out$r.squared.h0 <- aggregate_r2(r2.h0.values, group.Ns)

    } else {
      out$r.squared.h0 <- NULL
    }

  } else {
    out$r.squared <- NULL
    out$r.squared.h0 <- NULL
  }

  out$format <- list(
    digits        = digits,
    scientific    = scientific,
    adjusted.stat = adjusted.stat,
    ci            = ci,
    loadings      = loadings,
    regressions   = regressions,
    covariances   = covariances,
    intercepts    = intercepts,
    variances     = variances,
    thresholds    = thresholds,
    extra.cols    = extra.cols,
    extra.fit     = fit,
    scaled.stat   = object$args$robust.se
  )

  class(out) <- "summary_da"
  out
}


#' @export
print.summary_da <- function(x, digits = 3, ...) {
  format_value <- function(val, digits = 3, scientific = x$format$scientific) {
    if (length(val) == 0 || all(is.na(val))) {
      return("NA")
    }
    formatNumeric(val, digits = digits, scientific = scientific)
  }

  formatCount <- function(val) {
    if (length(val) == 0) return("NA")
    num <- suppressWarnings(as.numeric(val))
    out <- ifelse(is.na(num), "NA",
                  formatC(round(num), format = "f", digits = 0, big.mark = ""))
    out
  }

  # We want the width without ci and extra cols
  width.out <- getWidthPrintedParTable(
    parTable    = x$parTable,
    scientific  = x$format$scientific,
    ci          = FALSE,
    digits      = x$format$digits,
    extra.cols  = NULL
  )

  printf(x$convergence.msg)

  # Convergence and Model Info -------------------------------------------------
  is.mc.ordered <- !is.null(x$ordered.mc)

  if (is.mc.ordered) {
    header.names <- c(
      "Estimator",
      "Optimization method",
      "Naive optimization method",
      "Number of model parameters"
    )

    header.values <- c(
      paste0("MC-", stringr::str_to_upper(x$method)),
      "ROBBINS-MONRO",
      stringr::str_to_upper(x$optimizer),
      x$npar
    )
  } else {
    header.names <- c(
      "Estimator",
      "Optimization method",
      "Number of model parameters"
    )

    header.values <- c(
      stringr::str_to_upper(c(x$method, x$optimizer)),
      x$npar
    )
  }

  cat(allignLhsRhs(lhs = header.names, rhs = header.values, pad = "  ",
                   width.out = width.out), "\n", sep = "")

  nObsVec <- x$N
  if (is.null(names(nObsVec)) || !any(nzchar(names(nObsVec))))
    names(nObsVec) <- paste0("Group ", seq_along(nObsVec))

  total.n <- sum(nObsVec, na.rm = TRUE)
  sampleLines <- allignLhsRhs(
    lhs = "Number of observations",
    rhs = formatCount(total.n),
    pad = "  ",
    width.out = width.out
  )

  if (length(nObsVec) > 1L) {
    sampleLines <- paste0(
      sampleLines,
      allignLhsRhs(
        lhs = paste0("  ", names(nObsVec)),
        rhs = formatCount(nObsVec),
        pad = "  ",
        width.out = width.out
      )
    )
  }

  if (x$is.fiml) {
    fimlVec <- x$n.fiml.patterns

    if (is.null(names(fimlVec)) || !any(nzchar(names(fimlVec))))
      names(fimlVec) <- paste0("Group ", seq_along(fimlVec))

    totalPatterns <- sum(fimlVec, na.rm = TRUE)
    sampleLines <- paste0(
      sampleLines,
      allignLhsRhs(
        lhs = "Number of missing patterns",
        rhs = formatCount(totalPatterns),
        pad = "  ",
        width.out = width.out
      )
    )

    nonZero <- fimlVec > 0
    if (length(fimlVec) > 1L && any(nonZero)) {
      sampleLines <- paste0(
        sampleLines,
        allignLhsRhs(
          lhs = paste0("  ", names(fimlVec)[nonZero]),
          rhs = formatCount(fimlVec[nonZero]),
          pad = "  ",
          width.out = width.out
        )
      )
    }

  }

  cat(sampleLines, "\n", sep = "")

  # Criterion/LogLik -----------------------------------------------------------
  names <- c(
    if (isTRUE(x$naive.fit.measures)) "Naive Loglikelihood" else "Loglikelihood",
    if (isTRUE(x$naive.fit.measures)) "Naive Akaike (AIC)" else "Akaike (AIC)",
    if (isTRUE(x$naive.fit.measures)) "Naive Bayesian (BIC)" else "Bayesian (BIC)"
  )

  values <- c(
    format_value(x$logLik,  digits = 2),
    format_value(x$fit$AIC, digits = 2),
    format_value(x$fit$BIC, digits = 2)
  )

  if (x$format$adjusted.stat) {
    names  <- c(
      names,
      if (isTRUE(x$naive.fit.measures)) "Naive Corrected Akaike (AICc)" else "Corrected Akaike (AICc)",
      if (isTRUE(x$naive.fit.measures)) "Naive Adjusted Bayesian (aBIC)" else "Adjusted Bayesian (aBIC)"
    )
    values <- c(values, format_value(x$fit$AICc, digits = 2),
                format_value(x$fit$aBIC, digits = 2))
  }

  ll.title <- if (isTRUE(x$naive.fit.measures)) {
    "Naive Loglikelihood and Information Criteria:"
  } else {
    "Loglikelihood and Information Criteria:"
  }
  cat(ll.title, "\n", sep = "")
  cat(allignLhsRhs(lhs = names, rhs = values, pad = "  ",
                   width.out = width.out), "\n")

  # Intergration ---------------------------------------------------------------
  if (!is.null(x$quad)) {
    nodesDim   <- x$quad$nodes.dim
    dimVal     <- x$quad$dim
    nodesTotal <- x$quad$nodes.total

    if (length(nodesDim) || length(dimVal) || length(nodesTotal)) {
      cat("Numerical Integration:\n")
      names <- c("Points of integration (per dim)", "Dimensions",
                 "Total points of integration")
      values <- c(formatCount(nodesDim),
                  formatCount(dimVal),
                  formatCount(nodesTotal))
      cat(allignLhsRhs(lhs = names, rhs = values, pad = "  ",
                       width.out = width.out), "\n", sep = "")
    }

  }

  # Comparative fit ------------------------------------------------------------
  if (!is.null(x$D)) {
    lav.fit.h0 <- x$fitH0$lav.fit
    fnull      <- \(x) if (is.null(x)) NA else x

    cat("Fit Measures for Baseline Model (H0):\n")

    names <- c("", "Chi-square", "Degrees of Freedom (Chi-square)",
               "P-value (Chi-square)")
    values <- c("Standard",
                format_value(x$fitH0$chisq.value, digits = 2),
                x$fitH0$chisq.df,
                formatPval(x$fitH0$chisq.pvalue, scientific = x$format$scientific))

    if (x$format$scaled.stat) {
      chisq.s <- fnull(lav.fit.h0[["chisq.scaled"]])
      df.s    <- fnull(lav.fit.h0[["df.scaled"]])
      pval.s  <- fnull(lav.fit.h0[["pvalue.scaled"]])
      scale.f <- fnull(lav.fit.h0[["chisq.scaling.factor"]])
      values.scaled <- c("Scaled",
                         format_value(chisq.s, digits = 2),
                         round(df.s),
                         formatPval(pval.s, scientific = x$format$scientific),
                         format_value(scale.f, digits = 3),
                         rep("", 2))
      values <- c(values, rep("", 3))
      names <- c(names, "Scaling correction factor",
                 "  Yuan-Bentler correction (Mplus variant)", "")

    } else values.scaled <- NULL


    names <- c(names, "RMSEA")
    values <- c(values, format_value(x$fitH0$RMSEA, digits = 3))

    if (!is.null(values.scaled)) {
      rmsea.s <- fnull(lav.fit.h0[["rmsea.scaled"]])
      values.scaled <- c(values.scaled, format_value(rmsea.s, digits = 3))
    }

    if (x$format$extra.fit) {
      names  <- c(names, "CFI", "TLI", "SRMR")
      srmr <- x$fitH0$SRMR
      cfi  <- fnull(lav.fit.h0[["cfi"]])
      tli  <- fnull(lav.fit.h0[["tli"]])
      values <- c(values,
                  format_value(cfi, digits = 3),
                  format_value(tli, digits = 3),
                  format_value(srmr, digits = 3))

      if (!is.null(values.scaled)) {
        cfi.s <- fnull(lav.fit.h0[["cfi.scaled"]])
        tli.s <- fnull(lav.fit.h0[["tli.scaled"]])
        values.scaled <- c(values.scaled,
                           format_value(cfi.s, digits = 3),
                           format_value(tli.s, digits = 3),
                           "")
      }
    }

    ll.h0 <- if (isTRUE(x$naive.fit.measures.h0)) "Naive Loglikelihood" else "Loglikelihood"
    aic.h0 <- if (isTRUE(x$naive.fit.measures.h0)) "Naive Akaike (AIC)" else "Akaike (AIC)"
    bic.h0 <- if (isTRUE(x$naive.fit.measures.h0)) "Naive Bayesian (BIC)" else "Bayesian (BIC)"
    names <- c(names, "", ll.h0, aic.h0, bic.h0)
    values <- c(values, "",
                format_value(x$nullModel$logLik, digits = 2),
                format_value(x$fitH0$AIC, digits = 2),
                format_value(x$fitH0$BIC, digits = 2))

    if (!is.null(values.scaled))
      values.scaled <- c(values.scaled, rep("", 4))

    if (x$format$adjusted.stat) {
      aicc.h0 <- if (isTRUE(x$naive.fit.measures.h0)) {
        "Naive Corrected Akaike (AICc)"
      } else {
        "Corrected Akaike (AICc)"
      }
      abic.h0 <- if (isTRUE(x$naive.fit.measures.h0)) {
        "Naive Adjusted Bayesian (aBIC)"
      } else {
        "Adjusted Bayesian (aBIC)"
      }
      names <- c(names, aicc.h0, abic.h0)
      values <- c(values,
                  format_value(x$fitH0$AICc, digits = 2),
                  format_value(x$fitH0$aBIC, digits = 2))

      if (!is.null(values.scaled))
        values.scaled <- c(values.scaled, rep("", 2))
    }

    cat(allignLhsRhs(lhs = names, rhs = values, pad = "  ",
                     width.out = width.out, rhs.scaled = values.scaled), "\n")

    cat("Comparative Fit to H0 (LRT test):\n")
    names <- c("Loglikelihood change",
               "Difference test (D)",
               "Degrees of freedom (D)", "P-value (D)")
    values <- c(format_value(x$D$diff.loglik, digits = 2),
                format_value(x$D$D, digits = 2),
                x$D$df,
                formatPval(x$D$p, scientific = x$format$scientific))
    cat(allignLhsRhs(lhs = names, rhs = values, pad = "  ",
                     width.out = width.out), "\n")

  }

  # R2 -------------------------------------------------------------------------
  if (!is.null(x$r.squared)) {
    r.squared <- format_value(x$r.squared, digits = 3)
    names     <- names(r.squared)

    label_h1 <- "R-Squared Interaction Model (H1)"
    if (length(x$N) > 1) label_h1 <- paste0(label_h1, " (weighted)")

    cat(label_h1, ":\n", sep = "")
    cat(allignLhsRhs(lhs = names, rhs = r.squared,
										 pad = "  ", width.out = width.out))

    if (!is.null(x$r.squared.h0)) {
      r.squared.h0 <- format_value(x$r.squared.h0, digits = 3)
			names.h0     <- names(r.squared.h0)

      label_h0 <- "R-Squared Baseline Model (H0)"
      if (length(x$N) > 1) label_h0 <- paste0(label_h0, " (weighted)")

			cat(label_h0, ":\n", sep = "")
      cat(allignLhsRhs(lhs = names.h0, rhs = r.squared.h0, pad = "  ",
                       width.out = width.out))

      # Calculate Change (using unformatted Rsquared)
      if (!is.null(x$r.squared) && !is.null(x$r.squared.h0)) {
        r.squared.diff.raw <- x$r.squared - x$r.squared.h0
        r.squared.diff <- format_value(r.squared.diff.raw, digits = 3)
        names.diff     <- names(r.squared.diff)
        cat("R-Squared Change (H1 - H0):\n")
        cat(allignLhsRhs(lhs = names.diff, rhs = r.squared.diff,
                         pad = "  ", width.out = width.out))
      }
    }
  }

  # Parameters -----------------------------------------------------------------
  cat("\nParameter Estimates:\n")
  names <- c("Coefficients", "Information", "Standard errors")
  values <- c(x$type.estimates, x$information, x$type.se)
  cat(allignLhsRhs(lhs = names, rhs = values, pad = "  ",
                   width.out = width.out), "\n")

  groups  <- getGroupsParTable(x$parTable)
  ngroups <- length(groups)

  for (g in groups) {
    if (ngroups > 1L) {
      label <- tryCatch(x$group.labels[[g]], error = \(e) NA)
      printf("Group %d [%s]:\n\n", g, label)
    }

    select <- x$parTable$group == g
    if (g == max(groups)) select <- select | x$parTable$group == 0L

    printParTable(x$parTable[select, , drop = FALSE],
                  scientific  = x$format$scientific,
                  ci          = x$format$ci,
                  digits      = x$format$digits,
                  loadings    = x$format$loadings,
                  regressions = x$format$regressions,
                  covariances = x$format$covariances,
                  intercepts  = x$format$intercepts,
                  variances   = x$format$variances,
                  thresholds  = x$format$thresholds,
                  extra.cols  = x$format$extra.cols)
  }
}


#' @export
print.modsem_da <- function(x, digits = 3, ...) {
  cat(x$convergence.msg)
  print(parameter_estimates(x))
}


#' @export
var_interactions.modsem_da <- function(object, ...) {
  var_interactions.data.frame(parameter_estimates(object), ...)
}


#' Inspect components of a \code{modsem_da} fit
#'
#' \code{modsem_inspect.modsem_da} lets you pull lavaan-style matrices,
#' optimiser diagnostics, expected moments, or fit measures from a
#' \code{\link{modsem_da}} object.
#'
#' @param object A fitted object of class \code{"modsem_da"}.
#' @param what   Character scalar selecting what to return (see \emph{Details}).
#'               If \code{NULL} the value \code{"default"} is used.
#' @param ...    Passed straight to \code{modsem_inspect_da()}.
#'
#' @details
#' Below is a list of possible values for the \code{what} argument,
#' organised in several sections.  Keywords are \emph{case-sensitive}.
#'
#' \strong{Presets}
#'
#' \describe{
#'   \item{\code{"default"}}{Everything in \emph{Sample information}, \emph{Optimiser diagnostics}
#'   \emph{Parameter tables}, \emph{Model matrices}, and \emph{Expected-moment matrices} except
#'   the raw \code{data} slot}
#'   \item{\code{"coef"}}{Coefficients and variance-covariance matrix of both free and constrained parameters (same as \code{"coef.all"}).}
#'   \item{\code{"coef.all"}}{Coefficients and variance-covariance matrix of both free and constrained parameters (same as \code{"coef"}).}
#'   \item{\code{"coef.free"}}{Coefficients and variance-covariance matrix of the free parameters.}
#'   \item{\code{"all"}}{All items listed below, including \code{data}.}
#'   \item{\code{"matrices"}}{The lavaan-style model matrices.}
#'   \item{\code{"optim"}}{Only the items under \emph{Optimiser diagnostics}}.
#'   \item{\code{"fit"}}{A list with \code{fit.h0}, \code{fit.h1}, comparative.fit}
#' }
#'
#' \strong{Sample information:}
#'
#' \describe{
#' \item{\code{"N"}}{Number of analysed rows (integer).}
#' \item{\code{"ngroups"}}{Number of groups in model (integer).}
#' \item{\code{"group"}}{Group variable in model (character).}
#' \item{\code{"group.label"}}{Group labels (character).}
#' \item{\code{"ovs"}}{Observed variables used in model (character).}
#' }
#'
#' \strong{Parameter estimates and standard errors:}
#'
#' \describe{
#'   \item{\code{"coefficients.free"}}{Free parameter values.}
#'   \item{\code{"coefficients.all"}}{Both free and constrained parameter values.}
#'   \item{\code{"vcov.free"}}{Variance–covariance of free coefficients only.}
#'   \item{\code{"vcov.all"}}{Variance–covariance of both free and constrained coefficients.}
#' }
#'
#' \strong{Optimiser diagnostics:}
#'
#' \describe{
#'   \item{\code{"coefficients.free"}}{Free parameter values.}
#'   \item{\code{"vcov.free"}}{Variance–covariance of free coefficients only.}
#'   \item{\code{"information"}}{Fisher information matrix.}
#'   \item{\code{"loglik"}}{Log-likelihood.}
#'   \item{\code{"iterations"}}{Optimiser iteration count.}
#'   \item{\code{"convergence"}}{\code{TRUE}/\code{FALSE} indicating whether the model converged.}
#'   \item{\code{"iteration.history"}}{LMS iteration history with mode, log-likelihood, changes, elapsed time, and forecast columns, if available.}
#' }
#'
#' \strong{Parameter tables:}
#'
#' \describe{
#'   \item{\code{"partable"}}{Parameter table with estimated parameters.}
#'   \item{\code{"partable.input"}}{Parsed model syntax.}
#' }
#'
#' \strong{Model matrices:}
#'
#' \describe{
#'   \item{\code{"lambda"}}{Factor loadings.}
#'   \item{\code{"theta"}}{Residual covariance matrix for indicators.}
#'   \item{\code{"wmat"}}{Composite loading matrix for observed indicators, if present.}
#'   \item{\code{"tmat"}}{Composite residual matrix for indicators, if present.}
#'   \item{\code{"psi"}}{Residual covariance matrix for latent variables.}
#'   \item{\code{"beta"}}{Structural regression matrix among latent variables.}
#'   \item{\code{"nu"}}{Intercepts for observed variables.}
#'   \item{\code{"alpha"}}{Intercepts for latent variables.}
#' }
#' \strong{Model-implied matrices:}
#'
#' \describe{
#'   \item{\code{"cov.ov"}}{Model-implied covariance of observed variables.}
#'   \item{\code{"cov.lv"}}{Model-implied covariance of latent variables.}
#'   \item{\code{"cov.all"}}{Joint covariance of observed + latent variables.}
#'   \item{\code{"cor.ov"}}{Correlation counterpart of \code{"cov.ov"}.}
#'   \item{\code{"cor.lv"}}{Correlation counterpart of \code{"cov.lv"}.}
#'   \item{\code{"cor.all"}}{Correlation counterpart of \code{"cov.all"}.}
#'   \item{\code{"mean.ov"}}{Expected means of observed variables.}
#'   \item{\code{"mean.lv"}}{Expected means of latent variables.}
#'   \item{\code{"mean.all"}}{Joint mean vector.}
#' }
#'
#' \strong{R-squared and standardized residual variances:}
#'
#' \describe{
#'   \item{\code{"r2.all"}}{R-squared values for both observed (i.e., indicators) and latent endogenous variables.}
#'   \item{\code{"r2.lv"}}{R-squared values for latent endogenous variables.}
#'   \item{\code{"r2.ov"}}{R-squared values for observed (i.e., indicators) variables.}
#'   \item{\code{"res.all"}}{Standardized residuals (i.e., \code{1 - R^2}) for both observed (i.e., indicators) and latent endogenous variables.}
#'   \item{\code{"res.lv"}}{Standardized residuals (i.e., \code{1 - R^2}) for latent endogenous variables.}
#'   \item{\code{"res.ov"}}{Standardized residuals (i.e., \code{1 - R^2}) for observed variables (i.e., indicators).}
#' }
#'
#' \strong{Interaction-specific caveats:}
#'
#' \itemize{
#'   \item If the model contains an \emph{uncentred} latent interaction term it is centred
#'     internally before any \code{cov.*}, \code{cor.*}, or \code{mean.*} matrices are
#'     calculated.
#'   \item These matrices should not be used to compute fit-statistics (e.g.,
#'     chi-square and RMSEA) if there is an interaction term in the model.
#' }
#' @return A named list with the extracted information. If a single piece of information is returned,
#'  it is returned as is; not as a named element in a list.
#'
#' @examples
#' \dontrun{
#' m1 <- "
#'  # Outer Model
#'  X =~ x1 + x2 + x3
#'  Y =~ y1 + y2 + y3
#'  Z =~ z1 + z2 + z3
#'
#'  # Inner model
#'  Y ~ X + Z + X:Z
#' "
#'
#' est <- modsem(m1, oneInt, "lms")
#'
#' modsem_inspect(est) # everything except "data"
#' modsem_inspect(est, what = "optim")
#' modsem_inspect(est, what = "matrices")
#' }
#'
#' @export
#' @describeIn modsem_inspect Inspect a \code{\link{modsem_da}} object
modsem_inspect.modsem_da <- function(object, what = NULL, ...) {
  if (is.null(what)) what <- "default"
  modsem_inspect_da(object, what = what, ...)
}


#' @export
#' @importFrom stats vcov
vcov.modsem_da <- function(object, type = c("all", "free"), ...) {
  type <- tolower(type)
  type <- match.arg(type)

  what <- ifelse(type == "all", yes = "vcov.all", no = "vcov.free")
  modsem_inspect_da(object, what = what, ...)
}


#' @export
#' @importFrom stats coefficients
coefficients.modsem_da <- function(object, type = c("all", "free"), ...) {
  type <- tolower(type)
  type <- match.arg(type)

  what <- ifelse(type == "all", yes = "coefficients.all",
                 no = "coefficients.free")
  modsem_inspect_da(object, what = what, ...)
}


#' @export
#' @importFrom stats coef
coef.modsem_da <- function(object, type = "all", ...) {
  coefficients.modsem_da(object, type = type, ...)
}


#' @export
#' @importFrom stats nobs
nobs.modsem_da <- function(object, ...) {
  modsem_inspect_da(object, what = "N", ...)[[1]]
}


#' @describeIn standardized_estimates Method for \code{modsem_da} objects
#'
#' @param monte.carlo Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
#' standard errors; if \code{FALSE}, use the delta method (default).
#' @param mc.reps Number of Monte Carlo repetitions. Default is 10000.
#' @param tolerance.zero Threshold below which standard errors are set to \code{NA}.
#' @param rm.tmp.ov Should temporary (hidden) variables be removed?
#'
#' @export
standardized_estimates.modsem_da <- function(object,
                                             monte.carlo = FALSE,
                                             mc.reps = 10000,
                                             tolerance.zero = 1e-10,
                                             rm.tmp.ov = TRUE,
                                             ...) {
  parTable.std <- standardizedSolutionCOEFS(
    object,
    monte.carlo = monte.carlo,
    mc.reps = mc.reps,
    tolerance.zero = tolerance.zero,
    ...
  )$parTable

  if (rm.tmp.ov)
    parTable.std <- removeTempOV_RowsParTable(parTable.std)

  parTable.std
}


#' @describeIn centered_estimates Method for \code{modsem_da} objects
#'
#' @param monte.carlo Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
#' standard errors; if \code{FALSE}, use the delta method (default).
#' @param mc.reps Number of Monte Carlo repetitions. Default is 10000.
#' @param tolerance.zero Threshold below which standard errors are set to \code{NA}.
#'
#' @export
centered_estimates.modsem_da <- function(object,
                                         monte.carlo = FALSE,
                                         mc.reps = 10000,
                                         tolerance.zero = 1e-10, ...) {

  stdSolution <- centeredSolutionCOEFS(
    object,
    monte.carlo = monte.carlo,
    mc.reps = mc.reps,
    tolerance.zero = tolerance.zero, ...
  )

  stdSolution$parTable
}


#' @export
is_interaction_model.modsem_da <- function(object) {
  isTRUE(object$model$info$has.interaction)
}

modsemParseEnv <- rlang::env(
  syntaxLines = NULL
)


resetModsemParseEnv <- function() {
  modsemParseEnv$syntaxLines <- NULL
}


getCharsLine <- function(line, i = 1) {
  if (is.null(line) || nchar(line) == 0) {
    return(" ")
  } else if (i > nchar(line)) {
    return(NULL)
  }
  rest <- getCharsLine(line, i + 1)
  c(substr(line, i, i),rest)
}


getLines <- function(syntax) {
  operators <- c("=~", "<~", "<=", ">=", "==", ":=", "~~", "~",
                 "+", "*", "<-", "->", "<", ">", "-", "|", ",")
  for (op in operators) {
    pattern <- paste0("\\", op, "\\s*[\n|;]")
    replace <- paste0(op, " ")
    syntax <- stringr::str_replace_all(syntax, pattern, replace)
  }

  lines <- strsplit(syntax, "\n|;") |>
    unlist() |>
    as.list() |>
    lapply(getCharsLine)
  lines <- purrr::imap(lines,
                  function(x, lNum)
                    structure(x, lineNumber = lNum))
  lines
}


createTokensLine <- function(line, i = 1,
                             token = NULL, listTokens = list()) {
  if (i > length(line)) {
    return(appendToList(listTokens, token))
  }

  if (length(listTokens) > 0 && is.MathOperator(last(listTokens))) {
    lastToken <- last(listTokens)
    offset <- attr(lastToken, "pos") + nchar(lastToken)

    token <- buildMathExprToken(line[offset:length(line)], pos = offset)
    return(appendToList(listTokens, token))
  }


  if (is.null(token)) {
    token <- initializeToken(line[[i]], pos = i, line)
  } else {
    if (fitsToken(token, nextChar = line[[i]])) {
      token <- addCharToken(token, nextChar = line[[i]])

    } else {
      listTokens <- appendToList(listTokens, token)
      token <- initializeToken(line[[i]], pos = i, line)
    }
  }
  if ("LavComment" %in% class(token)) {
      return(listTokens)
    }
  return(createTokensLine(line, i + 1, token, listTokens = listTokens))
}


initializeToken <- function(char, pos, line) {
  nextChar <- if (pos + 1 <= length(line)) line[[pos + 1]] else ""

  nextCharIsAlpha <- grepl("[[:alpha:]\\.]", nextChar)
  nextCharIsNum   <- grepl("[[:digit:]]", nextChar)
  opChars <- c("=", "~", "*", "+", "<", ">", ":", "^", "/", "-", "|", ",")

  # optimaly this should be a switch
  if (grepl("#", char)) {
    type  <- "LavComment"
    priority <- 999
  } else if (grepl("\\s+", char)) {
    type <- "LavBlank"
    priority <- 999
  } else if (grepl("[[:alpha:]_]", char) || (char == "." & nextCharIsAlpha)) {
    type <- "LavName"
    priority <- 10
  } else if (grepl("[\\(\\)]", char)) {
    type <- "LavClosure"
    priority <- 2
  } else if (char %in% opChars) {
    type <- "LavOperator"
    priority <- 0
  } else if (grepl("[[:alnum:]]", char) || (char == "." & nextCharIsNum)) {
    type <- "LavNumeric"
    priority <- 10
  } else if (grepl('\\"' , char)) {
    type <- "LavString"
    priority <- 10
  } else {
    mod_msg_stop(
      paste0("Unrecognized class of token in line ", attr(line, "lineNumber"), " pos ", pos, "\n"),
      highlightError(line, pos = pos),
      only.format.header = TRUE
    )
  }

  structure(
    char,
    pos = pos,
    lineNumber = attr(line, "lineNumber"),
    priority = priority,
    class = c(type, "LavToken")
  )
}


buildMathExprToken <- function(restLine, pos) {
  token <- stringr::str_c(restLine, collapse = "")
  structure(token,
            pos = pos,
            lineNumber = attr(restLine, "lineNumber"),
            priority = 10,
            class = c("LavMathExpr", "LavToken"))
}


fitsToken <- function(token, nextChar) {
  UseMethod("fitsToken")
}


#' @export
fitsToken.LavName <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar", nextChar))
  # if object name ends with ( it is a function,
  # and next char belongs to a new object
  if (grepl("\\($", token)) {
    return(FALSE)
  }
  grepl("[[:alpha:][:digit:]_.\\(]", nextChar)[[1]]
}


#' @export
fitsToken.LavString <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar", nextChar))
  # if object name ends with ( it is a function,
  # and next char belongs to a new object
  if (grepl('\\"$', token)) {
    return(FALSE)
  }
  grepl("[[:graph:][:space:]]", nextChar)[[1]]
}


#' @export
fitsToken.LavOperator <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar", nextChar))

  completeToken <- paste0(token, nextChar)
  switch(completeToken,
         "=~" = TRUE,
         "~~" = TRUE,
         "<-" = TRUE,
         "->" = TRUE,
         "==" = TRUE,
         "!=" = TRUE,
         ":=" = TRUE,
         ">=" = TRUE,
         "<=" = TRUE,
         "<~" = TRUE,
         FALSE)
}


#' @export
fitsToken.LavBlank <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar", nextChar))
  grepl("\\s+", nextChar)
}


#' @export
fitsToken.LavClosure <- function(token, nextChar) {
  FALSE
}


#' @export
fitsToken.LavNumeric <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar", nextChar))
  grepl("[[:digit:].]", nextChar)
}


#' @export
fitsToken.LavComment <- function(token, nextChar) {
  TRUE
}


assignSubClass <- function(token) {
  UseMethod("assignSubClass")
}


#' @export
assignSubClass.LavOperator <- function(token) {
  switch (getTokenString(token),
          "=~" = {subClass <- "LavMeasure";     priority <- 0},
          "~"  = {subClass <- "LavPredict";     priority <- 0},
          "~~" = {subClass <- "LavCovar";       priority <- 0},
          "+"  = {subClass <- "LavAdd";         priority <- 1},
          "*"  = {subClass <- "LavModify";      priority <- 2},
          "<"  = {subClass <- "LavLessLeft";    priority <- 0},
          ">"  = {subClass <- "LavLessRight";   priority <- 0},
          "==" = {subClass <- "LavEquals";      priority <- 0},
          ":"  = {subClass <- "LavInteraction"; priority <- 2},
          ":=" = {subClass <- "LavCustom";      priority <- 0},
          ","  = {subClass <- "LavSeperator";   priority <- 0},
          "-"  = {subClass <- "LavSubtract";    priority <- 3},
          "|"  = {subClass <- "LavThreshold";   priority <- 0},
          "<~" = {subClass <- "LavMeasure";     priority <- 0},
          mod_msg_stop(
            paste0("Unrecognized operator: ", highlightErrorToken(token)),
            only.format.header = TRUE
          )
  )
  structure(token,
            class = c(subClass, class(token)),
            priority = priority)
}


#' @export
assignSubClass.LavClosure <- function(token) {
  switch(getTokenString(token),
          "("  = {subClass <- "LeftBracket";    priority <- 3},
          ")"  = {subClass <- "RightBracket";   priority <- 3},
          mod_msg_stop(paste0("Unrecognized operator: ", token))
  )
  structure(token,
            class = c(subClass, class(token)),
            priority = priority)
}


#' @export
assignSubClass.LavName <- function(token) {
  if (grepl("\\($", getTokenString(token))) {
    subClass <- "LavFunction"
    priority <- 3
  } else {
    subClass <- "LavObject"
    priority <- 3
  }
  structure(token,
            class = c(subClass, class(token)),
            priority = priority)
}


#' @export
assignSubClass.LavNumeric <- function(token) {
  token
}


#' @export
assignSubClass.LavToken <- function(token) {
  token
}


#' @export
assignSubClass.LavMathExpr <- function(token) {
  token
}


appendToList <- function(list, elem) {
  list[[length(list) + 1]] <- elem
  list
}


prioritizeTokens <- function(listTokens, i = 1, brackets = list(),
                             nLeftBrackets = 0) {
  if (is.null(listTokens) || i > length(listTokens)) {
    mod_stopif(nLeftBrackets != 0,
      paste0("Unmatched left bracket", highlightErrorToken(brackets[[1]])),
      only.format.header = TRUE
    )

    return(listTokens)

  } else if (is.RightClosure(listTokens[[i]])) {
    brackets <- brackets[-length(brackets)]
    nLeftBrackets <- nLeftBrackets - 1
    mod_stopif(nLeftBrackets < 0,
      paste0("Unmatched right bracket ", highlightErrorToken(listTokens[[i]])),
      only.format.header = TRUE
    )
  }
  getTokenPriority(listTokens[[i]]) <-
    getTokenPriority(listTokens[[i]]) + nLeftBrackets*10

  if (is.LeftClosure(listTokens[[i]])) {
    brackets <- appendToList(brackets, listTokens[[i]])
    nLeftBrackets <- nLeftBrackets + 1
  }
  prioritizeTokens(listTokens, i + 1, brackets = brackets,
                   nLeftBrackets = nLeftBrackets)
}


removeLavBlankLine <- function(line, removeComments = TRUE) {
  if (is.null(line) || length(line) == 0) {
    return(line)
  }
  isBlankOrComment <- vapply(line,
                             FUN.VALUE = logical(1L),
                             FUN = function(token) is.LavBlankOrComment(token))
  line[!isBlankOrComment]
}


tokenizeSyntax <- function(syntax, optimize = TRUE) {
  resetModsemParseEnv()
  mod_stopif(!is.character(syntax), "Syntax must be a string")

  lines <- getLines(syntax)
  modsemParseEnv$syntaxLines <- lines

  tokenizedLines <- lines |>
    lapply(createTokensLine) |>
    lapply(removeLavBlankLine) |>
    lapply(FUN = function(tokens) lapply(tokens, assignSubClass))

  tokenizedLines <- tokenizedLines |>
    lapply(prioritizeTokens)

  isEmpty <- vapply(tokenizedLines, FUN.VALUE = logical(1L),
                    FUN = function(line) is.null(line) || length(line) == 0)

  tokenizedLines[!isEmpty]
}


mergeTokensToString <- function(listTokens) {
  vapply(listTokens, FUN.VALUE = character(1L),
         FUN = getTokenString) |>
    stringr::str_c(collapse = "")
}


addCharToken <- function(token, nextChar) {
  mod_stopif(length(nextChar) != 1, paste0("Wrong length of nextChar ", nextChar))

  out <- paste0(token, nextChar)
  attributes(out) <- attributes(token)
  out
}


highlightError <- function(line, pos) {
  line <- stringr::str_c(line, collapse = "")
  indent <- "      "
  message <- paste0(indent, line, "\n",
                    indent, stringr::str_c(rep(" ", pos - 1),
                                           collapse = ""), "^")
  message
}


highlightErrorToken <- function(token) {
  lineNumber <- attr(token, "lineNumber")
  line <- modsemParseEnv$syntaxLines[[lineNumber]]
  if (is.list(line)) {
    line <- vapply(line,
                   FUN.VALUE = character(1L),
                   FUN = getTokenString)
  }
  line <- stringr::str_c(line, collapse = "")
  pos <- getTokenPosition(token)
  indent <- "      "
  message <- paste0("\n", indent, line, "\n",
                    indent, stringr::str_c(rep(" ", pos - 1),
                                           collapse = ""), "^")
  message
}


getTokenString <- function(token) {
  token[[1]]
}



getTokenPriority <- function(token) {
  attr(token, "priority")
}


`getTokenPriority<-` <- function(token, value) {
  attr(token, "priority") <- value
  token
}


getTokenPosition <- function(token) {
  attr(token, "pos")
}


is.LavToken <- function(token) {
  inherits(token, "LavToken")
}


is.LavName <- function(token, ignore.function=TRUE) {
  isName <- inherits(token, "LavName")

  if (ignore.function) {
    isName <- isName && !grepl("\\($", getTokenString(token))
  }

  isName
}


is.LavOperator <- function(token) {
  inherits(token, "LavOperator")
}


is.MathOperator <- function(token) {
  switch(getTokenString(token),
         "==" = TRUE,
         "<" = TRUE,
         ">" = TRUE,
         ":=" = TRUE,
         FALSE)
}


is.LavBlankOrComment <- function(token) {
  inherits(token, "LavBlank") || inherits(token, "LavComment")
}


is.LeftClosure <- function(token) {
  inherits(token, "LeftBracket") || inherits(token, "LavFunction")
}


is.RightClosure <- function(token) {
  inherits(token, "RightBracket")
}


is.LavClosure <- function(token) {
  inherits(token, "LavClosure")
}


is.LavOperator <- function(token) {
  inherits(token, "LavOperator")
}


is.firstClassOperator <- function(token) {
  switch(getTokenString(token),
         "=~" = TRUE,
         "~"  = TRUE,
         "~~" = TRUE,
         "<"  = TRUE,
         ">"  = TRUE,
         "==" = TRUE,
         FALSE)
}


#' @export
as.character.LavToken <- function(x, ...) {
  attributes(x) <- NULL
  x
}

#' Interaction between latent variables using product indicators
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param data dataframe
#'
#' @param method method to use:
#' \describe{
#'   \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
#'   \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
#'   \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
#'   \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
#'   \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
#' }
#'
#' @param match should the product indicators be created by using the match-strategy
#'
#' @param match.recycle should the indicators be recycled when using the match-strategy? I.e.,
#'   if one of the latent variables have fewer indicators than the other, some indicators
#'   are recycled to match the latent variable with the most indicators.
#'
#' @param standardize.data should data be scaled before fitting model
#'
#' @param first.loading.fixed Should the first factor loading in the latent product be fixed to one? Defaults to \code{FALSE}, as
#'   this already happens in \code{lavaan} by default. If \code{TRUE}, the first factor loading in the latent product is fixed to one.
#'   Manually in the generated syntax (e.g., \code{XZ =~ 1*x1z1}).'
#'
#' @param center.before should indicators in products be centered before computing products.
#'
#' @param center.after should indicator products be centered after they have been computed?
#'
#' @param residuals.prods should indicator products be centered using residuals.
#'
#' @param residual.cov.syntax should syntax for residual covariances be produced.
#'
#' @param constrained.prod.mean should syntax for product mean be produced.
#'
#' @param center.data should data be centered before fitting model
#'
#' @param constrained.loadings should syntax for constrained loadings be produced.
#'
#' @param constrained.var should syntax for constrained variances be produced.
#'
#' @param res.cov.method method for constraining residual covariances. Options are
#' \describe{
#'   \item{"simple"}{Residuals of product indicators with variables in common are allowed to covary freely. Defualt for most approches.}
#'   \item{"ca"}{Residual covariances of product indicators are constrained according to the constrained approach.}
#'   \item{"equality"}{Residuals of product indicators with variables in common are constrained to have equal covariances".
#'                     Can be useful for models where the model is unidentifiable using \code{res.cov.method == "simple"},
#'                     (e.g., when there is an interaction between an observed and a latent variable).}
#'   \item{"none"}{Residual covariances between product indicators are not specificed (i.e., constrained to zero).
#'                 Produces the same results as \code{constrained.cov.syntax = FALSE}.
#'                 Can be useful for models where the model is unidentifiable using \code{res.cov.method == "simple"},
#'                 (e.g., when there is an interaction between an observed and a latent variable).}
#' }
#'
#' @param res.cov.across Should residual covariances be specified/freed across different interaction terms.
#'   For example if you have two interaction terms \code{X:Z} and \code{X:W} the residuals of the
#'   generated product indicators \code{x1:z1} and \code{x1:w1} may be correlated. If \code{TRUE}
#'   residual covariances are allowed across different latent interaction terms. If \code{FALSE}
#'   residual covariances are only allowed between product indicators which belong to the same
#'   latent interaction term.
#'
#' @param composite.int.res.cov Should residual covariance syntax be generated for interaction
#'   terms that are fully composite (i.e., all component variables defined with \code{<~})?
#'   Defaults to \code{FALSE} because for fully composite interaction terms the residual
#'   covariances of the product indicators are not needed. Set to \code{TRUE} to force
#'   residual covariance syntax for composite interaction terms (useful for testing).
#'
#' @param auto.scale methods which should be scaled automatically (usually not useful)
#'
#' @param auto.center methods which should be centered automatically (usually not useful)
#'
#' @param estimator estimator to use in \code{lavaan}
#'
#' @param group group variable for multigroup analysis
#'
#' @param cluster cluster variable for multilevel models
#'
#' @param run should the model be run via \code{lavaan}, if \code{FALSE} only modified syntax and data is returned
#'
#' @param na.rm should missing values be removed (case-wise)? Defaults to FALSE. If \code{TRUE}, missing values are removed case-wise.
#' If \code{FALSE} they are not removed.
#'
#' @param suppress.warnings.lavaan should warnings from \code{lavaan} be suppressed?
#' @param suppress.warnings.match should warnings from \code{match} be suppressed?
#'
#' @param rcs Should latent variable indicators be replaced with reliability-corrected
#'   single item indicators instead? See \code{\link{relcorr_single_item}}.
#'
#' @param rcs.choose Which latent variables should get their indicators replaced with
#'   reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
#'   as the \code{choose} argument.
#'
#' @param rcs.res.cov.xz Should the residual (co-)variances of the product indicators
#'   created from the reliability-corrected single items (created if \code{rcs = TRUE})
#'   be specified and constrained before estimating the model? If \code{TRUE} the estimates
#'   for the constraints are approximated using a monte carlo simulation (see the \code{rcs.mc.reps} argument).
#'   If \code{FALSE} the residual variances are not specified, which usually mean that all
#'   are constrained to zero.
#'
#' @param rcs.mc.reps Sample size used in monte-carlo simulation, when approximating the
#'   the estimates of the residual (co-)variances between the product indicators formed
#'   by reliabiliyt-corrected single items (see the \code{rcs.res.cov.xz} argument).
#'
#' @param rcs.scale.corrected Should reliability corrected items be scale-corrected? If \code{TRUE}
#'   reliability-corrected single items are corrected for differences in factor loadings between
#'   the items. Default is \code{TRUE}.
#'
#' @param LAVFUN Function used to estimate the model. Defaults to \code{lavaan::sem}.
#'
#' @param ... arguments passed to \code{LAVFUN}
#'
#' @return \code{modsem} object
#' @export
#' @description
#' \code{modsem_pi()} is a function for estimating interaction effects between latent variables,
#' in structural equation models (SEMs), using product indicators.
#' Methods for estimating interaction effects in SEMs can basically be split into
#' two frameworks:
#' 1. Product Indicator based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"},
#' \code{"ca"}, \code{"pind"}), and
#' 2. Distributionally based approaches (\code{"lms"}, \code{"qml"}).
#' \code{modsem_pi()} is essentially a fancy wrapper for \code{lavaan::sem()} which generates the
#' necessary syntax and variables for the estimation of models with latent product indicators.
#' Use \code{default_settings_pi()} to get the default settings for the different methods.
#'
#' @examples
#' library(modsem)
#' # For more examples, check README and/or GitHub.
#' # One interaction
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#'
#' # Double centering approach
#' est <- modsem_pi(m1, oneInt)
#' summary(est)
#'
#' \dontrun{
#' # The Constrained Approach
#' est_ca <- modsem_pi(m1, oneInt, method = "ca")
#' summary(est_ca)
#' }
#'
#' # Theory Of Planned Behavior
#' tpb <- '
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   # Covariances
#'   ATT ~~ SN + PBC
#'   PBC ~~ SN
#'   # Causal Relationships
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC
#'   BEH ~ INT:PBC
#' '
#'
#' # Double centering approach
#' est_tpb <- modsem_pi(tpb, data = TPB)
#' summary(est_tpb)
#'
#' \dontrun{
#' # The Constrained Approach
#' est_tpb_ca <- modsem_pi(tpb, data = TPB, method = "ca")
#' summary(est_tpb_ca)
#' }
modsem_pi <- function(model.syntax = NULL,
                      data = NULL,
                      method = "dblcent",
                      match = NULL,
                      match.recycle = NULL,
                      standardize.data = FALSE,
                      center.data = FALSE,
                      first.loading.fixed = FALSE,
                      center.before = NULL,
                      center.after = NULL,
                      residuals.prods = NULL,
                      residual.cov.syntax = NULL,
                      constrained.prod.mean = NULL,
                      constrained.loadings = NULL,
                      constrained.var = NULL,
                      res.cov.method = NULL,
                      res.cov.across = NULL,
                      composite.int.res.cov = FALSE,
                      auto.scale = "none",
                      auto.center = "none",
                      estimator = "ML",
                      group = NULL,
                      cluster = NULL,
                      run = TRUE,
                      na.rm = FALSE,
                      suppress.warnings.lavaan = FALSE,
                      suppress.warnings.match = FALSE,
                      rcs = FALSE,
                      rcs.choose = NULL,
                      rcs.res.cov.xz = rcs,
                      rcs.mc.reps = 1e5,
                      rcs.scale.corrected = TRUE,
                      LAVFUN = lavaan::sem,
                      ...) {
  method <- tolower(method)

  mod_stopif(is.null(model.syntax), "No model syntax provided in modsem")
  mod_stopif(is.null(data), "No data provided in modsem")
  mod_stopif(!is.data.frame(data) && !is.matrix(data), "data must be a data.frame or matrix!")

  if (!is.null(cluster)) {
    est <- modsemPICluster(
      model.syntax = model.syntax,
      method = method,
      data = data,
      match = match,
      match.recycle = match.recycle,
      standardize.data = standardize.data,
      center.data = center.data,
      first.loading.fixed = first.loading.fixed,
      center.before = center.before,
      center.after = center.after,
      residuals.prods = residuals.prods,
      residual.cov.syntax = residual.cov.syntax,
      constrained.prod.mean = constrained.prod.mean,
      constrained.loadings = constrained.loadings,
      constrained.var = constrained.var,
      res.cov.method = res.cov.method,
      res.cov.across = res.cov.across,
      composite.int.res.cov = composite.int.res.cov,
      auto.scale = auto.scale,
      auto.center = auto.center,
      run = run,
      estimator = estimator,
      group = group,
      cluster = cluster,
      na.rm = na.rm,
      suppress.warnings.match = suppress.warnings.match,
      suppress.warnings.lavaan = suppress.warnings.lavaan,
      rcs = rcs,
      rcs.choose = rcs.choose,
      rcs.mc.reps = rcs.mc.reps,
      rcs.scale.corrected = rcs.scale.corrected,
      LAVFUN = lavaan::sem,
      ...
    )

    return(est)
  }

  data <- as.data.frame(data)

  if (rcs) { # use reliability-correct single items?
    if (!is.null(rcs.choose))
      rcs.choose <- rcs.choose[!grepl(":", rcs.choose)]

    corrected <- relcorr_single_item(
      syntax          = model.syntax,
      data            = data,
      group           = group,
      choose          = rcs.choose,
      scale.corrected = rcs.scale.corrected,
      warn.lav        = FALSE
    )

    model.syntax <- corrected$syntax
    data         <- corrected$data

    if (rcs.res.cov.xz)
      res.cov.across <- FALSE
  }

  methodSettings <-
    getMethodSettingsPI(method, args =
                        list(center.before = center.before,
                             center.after = center.after,
                             residuals.prods = residuals.prods,
                             residual.cov.syntax = residual.cov.syntax,
                             constrained.prod.mean = constrained.prod.mean,
                             constrained.loadings = constrained.loadings,
                             constrained.var = constrained.var,
                             res.cov.method = res.cov.method,
                             res.cov.across = res.cov.across,
                             first.loading.fixed = first.loading.fixed,
                             match = match,
                             match.recycle = match.recycle))

  # Get the specifications of the model
  modelSpec <- parseLavaan(model.syntax, colnames(data),
                           match = methodSettings$match,
                           suppress.warnings.match = suppress.warnings.match,
                           match.recycle = methodSettings$match.recycle)

  # Composite variables in interaction terms are not supported with method="ca"
  if (method == "ca" && length(modelSpec$composites) > 0L) {
    hasCompositeElement <- vapply(
      modelSpec$elementsInProdNames,
      FUN.VALUE = logical(1L),
      FUN = function(elems) any(elems %in% modelSpec$composites)
    )
    if (any(hasCompositeElement)) {
      mod_msg_stop(paste0('method="ca" does not support composite variables in interaction terms. ',
            'Use method="dblcent", "rca", "uca", or "pind" instead.'))
    }
  }

  # Save these for later
  input <- list(syntax = model.syntax, data = data,
                parTable = modelSpec$parTable)

  # Data Processing
  oVs        <- c(modelSpec$oVs, group)
  missingOVs <- setdiff(oVs, colnames(data))
  mod_stopif(length(missingOVs), paste0("Missing variables in data:\n", missingOVs))

  completeCases <- stats::complete.cases(data[oVs])

  if (any(!completeCases) && (is.null(na.rm) || na.rm)) {
    mod_warnif(is.null(na.rm), "Removing missing values list-wise.")
    data <- data[completeCases, ]
  }

  cont.cols <- setdiff(colnames(data), c(cluster, group))

  if (center.data)
    data[cont.cols] <- lapply(data[cont.cols], FUN = centerIfNumeric, scaleFactor = FALSE)

  if (standardize.data)
    data[cont.cols] <- lapply(data[cont.cols], FUN = scaleIfNumeric, scaleFactor = FALSE)

  prodInds <-
    createProdInds(modelSpec,
                   data = data,
                   center.before = methodSettings$center.before,
                   center.after = methodSettings$center.after,
                   residuals.prods = methodSettings$residuals.prods)
  mergedProdInds <- combineListDf(prodInds)

  # using list_cbind so that mergedProdInds can be NULL
  newData <- purrr::list_cbind(list(data, mergedProdInds))

  # Genereating a new syntax with constraints and measurmentmodel
  parTable <- addSpecsParTable(modelSpec,
                               residual.cov.syntax = methodSettings$residual.cov.syntax,
                               res.cov.method = methodSettings$res.cov.method,
                               res.cov.across = methodSettings$res.cov.across,
                               constrained.prod.mean = methodSettings$constrained.prod.mean,
                               constrained.loadings = methodSettings$constrained.loadings,
                               constrained.var = methodSettings$constrained.var,
                               firstFixed = first.loading.fixed,
                               composite.int.res.cov = composite.int.res.cov)

  newSyntax <- parTableToSyntax(parTable, removeColon = TRUE)

  if (rcs && rcs.res.cov.xz && method != "ca") { # Constrained Approach Should handle this it on its own...

    elemsxz   <- modelSpec$elementsInProdNames
    crossResCov <- simulateCrossResCovRCS(
      corrected = corrected,
      elemsInIntTerms = elemsxz,
      mc.reps = rcs.mc.reps,
      parTable = parTable,
      include.normal.inds = modelSpec$nways > 2L
    )

    newSyntax <- paste(newSyntax, crossResCov$syntax, sep = "\n")
  }

  # Interaction model
  modelSpec$prodInds <- prodInds
  modelSpec$syntax   <- newSyntax
  modelSpec$data     <- newData
  modelSpec$parTable <- parTable
  modelSpec$method   <- method

  # Extra info saved for estimating baseline model
  input$modsemArgs <- methodSettings
  input$lavArgs    <- list(estimator = estimator, cluster = cluster, group = group,
                           LAVFUN = LAVFUN,
                           rcs.res.cov.xz = rcs.res.cov.xz,
                           rcs.mc.reps = rcs.mc.reps,
                           rcs.choose = rcs.choose, ...)
  modelSpec$input  <- input

  if (run) {
    lavWrapper <- getWarningWrapper(silent = suppress.warnings.lavaan)
    lavEst <- tryCatch(LAVFUN(newSyntax, newData, estimator = estimator,
                              group = group, ...) |> lavWrapper(),
                       error = function(cnd) {
                         mod_msg_warn(conditionMessage(cnd))
                         NULL
                       })
    coefParTable <- tryCatch(lavaan::parameterEstimates(lavEst),
                             error = function(cnd) NULL)
    modelSpec$lavaan       <- lavEst
    modelSpec$coefParTable <- coefParTable
  }

  structure(
    modelSpec,
    class = c("modsem_pi", "modsem"),
    method = method,
    isRCS_Model = rcs
  )
}


createProdInds <- function(modelSpec,
                           data,
                           center.before = FALSE,
                           center.after = FALSE,
                           residuals.prods = FALSE) {
  indProds <- purrr::map2(.x = modelSpec$relDfs,
                          .y = modelSpec$indsInLatentProds,
                          .f = createIndProds,
                          data = data,
                          centered = center.before)
  if (residuals.prods) {
    indProds <- purrr::map2(.x = indProds, .y = modelSpec$indsInLatentProds,
                            .f = calculateResidualsDf, data = data)

  } else if (!is.logical(residuals.prods)) {
    mod_msg_stop("residualProds was neither FALSE nor TRUE in createProdInds")
  }

  if (center.after) {
    indProds <- lapply(indProds, FUN = function(df)
                       lapplyDf(df, FUN = function(x) x - mean(x, na.rm = TRUE)))

  }

  indProds
}


createIndProds <- function(relDf, indNames, data, centered = FALSE) {
  varnames <- unname(colnames(relDf))
  inds      <- data[indNames]
  isNumeric <- sapply(inds, is.numeric)

  mod_stopif(any(!isNumeric), "Expected inds to be numeric when creating prods")

  if (centered) {
    inds <- lapplyDf(inds, FUN = function(x) x - mean(x, na.rm = TRUE))
  }

  prods <- lapplyNamed(
    X = varnames,
    FUN = \(varname, data, relDf) multiplyIndicatorsCpp(data[relDf[[varname]]]),
    data = inds,
    relDf = relDf,
    names = varnames
  )

  # return as data.frame()
  structure(prods, row.names = seq_len(nrow(data)),
            class = "data.frame")
}


calculateResidualsDf <- function(dependentDf, independentNames, data) {
  isComplete <- stats::complete.cases(data)

  # Using purrr::list_cbind() is more efficient than cbind()
  combinedData <- purrr::list_cbind(list(dependentDf, data))
  combinedData <- combinedData[isComplete, ]

  # Getting the names of the dependent variables
  dependentNames <- colnames(dependentDf)
  # Getting formula
  formula <- getResidualsFormula(dependentNames, independentNames)

  resNoNA <- as.data.frame(stats::residuals(stats::lm(formula = formula,
                                                      combinedData)))
  colnames(resNoNA) <- dependentNames

  resNA <- dependentDf
  resNA[isComplete, ] <- resNoNA

  resNA
}


getResidualsFormula <- function(dependendtNames, indepNames) {
  formulaDep   <- paste0("cbind(", stringr::str_c(dependendtNames,
                                collapse = ", "), ")")
  formulaIndep <- stringr::str_c(indepNames, collapse = " + ")
  paste0(formulaDep, " ~ ", formulaIndep)
}


addSpecsParTable <- function(modelSpec,
                             residual.cov.syntax = FALSE,
                             res.cov.method = "equality",
                             res.cov.across = TRUE,
                             constrained.prod.mean = FALSE,
                             constrained.loadings = FALSE,
                             constrained.var = FALSE,
                             firstFixed = TRUE,
                             composite.int.res.cov = FALSE) {
  relDfs          <- modelSpec$relDfs
  latentProds     <- modelSpec$latentProds
  indProdNames    <- modelSpec$indProdNames
  parTable        <- modelSpec$parTable
  isCompositeProd <- modelSpec$isCompositeProd

  if (is.null(relDfs) || length(relDfs) < 1) return(parTable)

  # Look up composite status for each product term (FALSE when unknown)
  isCompVec <- isCompositeProd[latentProds]
  isCompVec[is.na(isCompVec)] <- FALSE

  # Composite product terms use <~ (formative); mixed/latent use =~
  measureParTable <- purrr::map2(
    .x = latentProds,
    .y = indProdNames,
    .f = function(prodName, indNames) {
      isComp <- isTRUE(isCompVec[[prodName]])
      op     <- if (isComp) "<~" else "=~"
      ff     <- if (isComp) FALSE else firstFixed
      getParTableMeasure(prodName, indNames, operator = op, firstFixed = ff)
    }
  ) |> purrr::list_rbind()
  parTable <- rbindParTable(parTable, measureParTable)

  compositeProds    <- latentProds[isCompVec]
  nonCompRelDfs     <- relDfs[!isCompVec[names(relDfs)]]
  hasNonCompProds   <- any(!isCompVec)

  # Enter label/constraint block only when non-composite products require it
  if (constrained.var || constrained.loadings || (constrained.prod.mean && hasNonCompProds)) {
    parTable <- addVariances(parTable, exclude = compositeProds) |>
      addCovariances() |>
      labelParameters() |>
      labelFactorLoadings()
  }

  if (!is.logical(residual.cov.syntax)) {
    mod_msg_stop("residual.cov.syntax is not FALSE or TRUE in generateSyntax")

  } else if (residual.cov.syntax && res.cov.method != "none") {
    # Skip residual covariances for fully-composite product terms unless
    # composite.int.res.cov = TRUE
    if (!composite.int.res.cov) {
      resCovRelDfs <- relDfs[!isCompVec[names(relDfs)]]
    } else {
      resCovRelDfs <- relDfs
    }

    if (length(resCovRelDfs) > 0) {
      # Even if `res.cov.across == TRUE` we still want to run `getParTableResCov`
      # for each interaction term, due to some important checks. If checks fail
      # we get `attr(relDf, "OK") == FALSE`
      residualCovariancesList <- purrr::map(.x = resCovRelDfs, .f = getParTableResCov,
                                            method = res.cov.method,
                                            pt = parTable,
                                            include.single.inds = FALSE)
      residualCovariances <- purrr::list_rbind(residualCovariancesList)

      if (res.cov.across) {
        # Get residual covariances across interaction terms
        # E.g.,
        # X:Z =~ x1:z1
        # X:M =~ x1:m1
        # x1:z1 ~~ x1:m1
        isOK <- vapply(residualCovariancesList, FUN.VALUE = logical(1L),
                       FUN = \(rows) attr(rows, "OK"))
        RelList <- Reduce(lapply(unname(resCovRelDfs[isOK]), FUN = as.list), f = c)
        residualCovariances <- getParTableResCov(relDf = RelList,
                                                 method = res.cov.method,
                                                 pt = parTable,
                                                 include.single.inds = TRUE)
      }

      parTable <- rbindParTable(parTable, residualCovariances)
    }
  }

  if (constrained.var)      parTable <- specifyVarCov(parTable, nonCompRelDfs)
  if (constrained.loadings) parTable <- specifyFactorLoadings(parTable, nonCompRelDfs)

  if (constrained.prod.mean) {
    # Skip mean constraint for fully-composite product terms
    nonCompProdNames    <- modelSpec$prodNames[!isCompVec[modelSpec$prodNames]]
    nonCompElemsInProds <- modelSpec$elementsInProdNames[nonCompProdNames]

    if (length(nonCompProdNames) > 0) {
      restrictedMeans <- purrr::map2(.x = nonCompProdNames,
                                     .y = nonCompElemsInProds,
                                     .f = getParTableRestrictedMean,
                                     createLabels = !constrained.var,
                                     pt = parTable) |>
        purrr::list_rbind()
      parTable <- rbindParTable(parTable, restrictedMeans)
    }
  }

  # redefine labels (using 'old' := 'new'), if any were overwritten when adding constraints
  parTable <- defineUndefinedLabels(parTable.x = modelSpec$parTable,
                                    parTable.y = parTable)

  parTable
}


# this function assumes a prod of only two latent variables no more
getParTableRestrictedMean <- function(prodName, elementsInProdName,
                                      createLabels = TRUE, pt) {
  mod_stopif(length(elementsInProdName) > 2,
         paste0("The mean of a latent prod should not be constrained when there",
         " are more than two variables in the prod term. Please use a",
         " different method \n"))

  meanLabel     <- createLabelMean(prodName)
  meanStructure <- createParTableRow(vecLhsRhs = c(prodName, ""),
                                     op = "~1", mod = meanLabel)
  covEquation   <- trace_path(pt, elementsInProdName[[1]], elementsInProdName[[2]])
  meanFormula   <- createParTableRow(vecLhsRhs = c(meanLabel, covEquation),
                                     op = "==")
  rbind(meanStructure, meanFormula)
}


multiplyIndicators <- function(df) {
  if (is.null(df)) return(NULL)
  if (ncol(df) <= 1) return(df[[1]])

  y <- cbind.data.frame(df[[1]] * df[[2]],
                        df[,-(1:2),drop = FALSE])

  multiplyIndicators(y)
}


getParTableMeasure <- function(dependentName,
                               predictorNames,
                               operator = "=~",
                               firstFixed = FALSE) {
  mod_stopif(length(dependentName) > 1, paste0("Expected dependentName ",
         "to be a single string in getParTableMeasure"))

  if (length(predictorNames) == 1 && dependentName == predictorNames) {
    # In this case it should be seen as an observed variable,
    # and should not have a measurement model
    return(NULL)
  }

  nRows    <- length(predictorNames)
  parTable <- data.frame(lhs = rep(dependentName, nRows),
                         op = rep(operator, nRows),
                         rhs = predictorNames,
                         mod = vector("character", nRows))

  if (firstFixed) parTable[["mod"]][[1]] <- "1"
  parTable
}


createParTableRow <- function(vecLhsRhs, op, mod = "") {
  data.frame(lhs = vecLhsRhs[[1]], op = op, rhs = vecLhsRhs[[2]], mod = mod)
}


#' Get \code{lavaan} syntax for product indicator approaches
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param method method to use:
#' \code{"rca"} = residual centering approach,
#' \code{"uca"} = unconstrained approach,
#' \code{"dblcent"} = double centering approach,
#' \code{"pind"} = prod ind approach, with no constraints or centering,
#' \code{"custom"} = use parameters specified in the function call
#'
#' @param data Optional. Dataset to use, usually not relevant.
#'
#' @param match should the product indicators be created by using the match-strategy
#'
#' @param ... arguments passed to other functions (e.g., \link{modsem_pi})
#'
#' @return \code{character} vector
#' @export
#' @description
#' \code{get_pi_syntax()} is a function for creating the \code{lavaan} syntax used for estimating
#' latent interaction models using one of the product indicator approaches.
#'
#' @examples
#' library(modsem)
#' library(lavaan)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#' syntax <- get_pi_syntax(m1)
#' data <- get_pi_data(m1, oneInt)
#' est <- sem(syntax, data)
#' summary(est)
get_pi_syntax <- function(model.syntax,
                          method = "dblcent",
                          match = FALSE,
                          data = NULL,
                          ...) {
  oVs       <- getOVs(model.syntax = model.syntax)

  if (is.null(data)) {
    data <- as.data.frame(matrix(0, nrow = 1, ncol = length(oVs),
                                 dimnames = list(NULL, oVs)))
  }

  modsem_pi(model.syntax, method = method, match = match,
            data = data, run = FALSE, ...)$syntax
}


#' Get data with product indicators for different approaches
#'
#' @param model.syntax \code{lavaan} syntax
#' @param data data to create product indicators from
#' @param method method to use:
#' \code{"rca"} = residual centering approach,
#' \code{"uca"} = unconstrained approach,
#' \code{"dblcent"} = double centering approach,
#' \code{"pind"} = prod ind approach, with no constraints or centering,
#' \code{"custom"} = use parameters specified in the function call
#' @param match should the product indicators be created by using the match-strategy
#'
#' @param ... arguments passed to other functions (e.g., \link{modsem_pi})
#'
#' @return \code{data.frame}
#' @export
#' @description
#' \code{get_pi_data()} is a function for creating a dataset with product indiactors used for estimating
#' latent interaction models using one of the product indicator approaches.
#'
#' @examples
#' library(modsem)
#' library(lavaan)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#' syntax <- get_pi_syntax(m1)
#' data <- get_pi_data(m1, oneInt)
#' est <- sem(syntax, data)
#' summary(est)
get_pi_data <- function(model.syntax, data, method = "dblcent",
                        match = FALSE, ...) {
  modsem_pi(model.syntax, data = data, method = method, match = match,
            run = FALSE, ...)$data
}



#' extract lavaan object from modsem object estimated using product indicators
#'
#' @param object modsem object
#'
#' @return lavaan object
#' @export
#' @examples
#' library(modsem)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#' est <- modsem_pi(m1, oneInt)
#' lav_est <- extract_lavaan(est)
extract_lavaan <- function(object) {
  if (!inherits(object, "modsem_pi")) {
    mod_msg_stop("object is not of class modsem_pi")
  }
  object$lavaan
}


modsemPICluster <- function(model.syntax = NULL,
                            data = NULL,
                            method = "dblcent",
                            match = NULL,
                            match.recycle = NULL,
                            standardize.data = FALSE,
                            center.data = FALSE,
                            first.loading.fixed = FALSE,
                            center.before = NULL,
                            center.after = NULL,
                            residuals.prods = NULL,
                            residual.cov.syntax = NULL,
                            constrained.prod.mean = NULL,
                            constrained.loadings = NULL,
                            constrained.var = NULL,
                            res.cov.method = NULL,
                            res.cov.across = NULL,
                            composite.int.res.cov = FALSE,
                            auto.scale = "none",
                            auto.center = "none",
                            estimator = "ML",
                            group = NULL,
                            cluster = NULL,
                            run = TRUE,
                            na.rm = FALSE,
                            suppress.warnings.lavaan = FALSE,
                            suppress.warnings.match = FALSE,
                            rcs = FALSE,
                            rcs.choose = NULL,
                            rcs.res.cov.xz = rcs,
                            rcs.mc.reps = 1e5,
                            rcs.scale.corrected = FALSE,
                            LAVFUN = lavaan::sem,
                            ...) {
  mod_stopif(na.rm, "`na.rm=TRUE` can currently not be paired with the `cluster` argument!")

  levelPattern <- "level([:blank:]*):([:blank:]*)([A-z]|[0-9]+)"
  levelHeaders <- unlist(stringr::str_extract_all(model.syntax, pattern=levelPattern))
  syntaxBlocks <- unlist(stringr::str_split(model.syntax, pattern=levelPattern))

  if (!length(levelHeaders)) levelHeaders <- ""
  else                       syntaxBlocks <- syntaxBlocks[-1]

  mod_stopif(length(syntaxBlocks) != length(levelHeaders), "Different number of blocks than level headers!")

  data$ROW_IDENTIFIER_ <- seq_len(nrow(data))
  has.interaction      <- FALSE
  newSyntax <- ""
  newData <- NULL

  for (i in seq_along(syntaxBlocks)) {
    syntaxBlock <- syntaxBlocks[[i]]
    levelHeader <- levelHeaders[[i]]

    parsedBlock <- modsemify(syntaxBlock)
    if (!NROW(parsedBlock)) next

    if (any(grepl(":", parsedBlock$rhs)))
        has.interaction <- TRUE

    newBlockSyntax <- get_pi_syntax(
      model.syntax = syntaxBlock,
      data = data,
      method = method,
      match = match,
      match.recycle = match.recycle,
      standardize.data = standardize.data,
      center.data = center.data,
      first.loading.fixed = first.loading.fixed,
      center.before = center.before,
      center.after = center.after,
      residuals.prods = residuals.prods,
      residual.cov.syntax = residual.cov.syntax,
      constrained.prod.mean = constrained.prod.mean,
      constrained.loadings = constrained.loadings,
      constrained.var = constrained.var,
      res.cov.method = res.cov.method,
      res.cov.across = res.cov.across,
      composite.int.res.cov = composite.int.res.cov,
      auto.scale = auto.scale,
      auto.center = auto.center,
      suppress.warnings.match = suppress.warnings.match,
      rcs = rcs,
      rcs.choose = rcs.choose,
      rcs.res.cov.xz = rcs.res.cov.xz,
      rcs.mc.reps = rcs.mc.reps,
      rcs.scale.corrected = rcs.scale.corrected
    ) |> stringr::str_replace_all(pattern = "\n", replacement = "\n\t")

    newBlockData <- get_pi_data(
      model.syntax = syntaxBlock,
      data = data,
      method = method,
      match = match,
      match.recycle = match.recycle,
      standardize.data = standardize.data,
      center.data = center.data,
      first.loading.fixed = first.loading.fixed,
      center.before = center.before,
      center.after = center.after,
      residuals.prods = residuals.prods,
      residual.cov.syntax = residual.cov.syntax,
      constrained.prod.mean = constrained.prod.mean,
      constrained.loadings = constrained.loadings,
      constrained.var = constrained.var,
      res.cov.method = res.cov.method,
      res.cov.across = res.cov.across,
      composite.int.res.cov = composite.int.res.cov,
      auto.scale = auto.scale,
      auto.center = auto.center,
      suppress.warnings.match = suppress.warnings.match,
      na.rm = FALSE,
      rcs = rcs,
      rcs.choose = rcs.choose,
      rcs.res.cov.xz = rcs.res.cov.xz,
      rcs.mc.reps = rcs.mc.reps,
      rcs.scale.corrected = rcs.scale.corrected
    )

    if (is.null(newData)) {
      newData <- newBlockData
    } else {
      newCols <- setdiff(colnames(newBlockData), colnames(newData))
      newBlockData <- newBlockData[c("ROW_IDENTIFIER_", newCols)]
      newData <- dplyr::left_join(newData, newBlockData, by = "ROW_IDENTIFIER_")
    }

    newSyntax <- paste(
      newSyntax,
      levelHeader, "\t", # indent first row
      newBlockSyntax,
      sep = "\n"
    )
  }

  modelSpec <- list(syntax = newSyntax, data = newData,
                    has.interaction = has.interaction)

  if (run) {
    lavWrapper <- getWarningWrapper(silent = suppress.warnings.lavaan)
    lavEst <- tryCatch({
        lavWrapper(LAVFUN(
          newSyntax, newData, estimator = estimator,
          group = group, cluster = cluster, ...
        ))
      },
      error = function(cnd) {
        mod_msg_warn(conditionMessage(cnd))
        NULL
      }
    )

    coefParTable <- tryCatch(lavaan::parameterEstimates(lavEst),
                             error = function(cnd) NULL)
    modelSpec$lavaan       <- lavEst
    modelSpec$coefParTable <- coefParTable
  }

  structure(modelSpec, class = c("modsem_pi", "modsem"), method = method)
}

devtools::load_all()

# --- Helpers ------------------------------------------------------------------

m_fully_comp <- '
  X <~ x1 + x2
  Z <~ z1 + z2
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

m_mixed <- '
  X =~ x1 + x2 + x3
  Z <~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

m_linear_comp <- '
  X <~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

m_latent_only <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

# --- parseLavaan: composite detection -----------------------------------------

spec_full  <- modsem:::parseLavaan(m_fully_comp, colnames(oneInt))
spec_mixed <- modsem:::parseLavaan(m_mixed,      colnames(oneInt))
spec_lin   <- modsem:::parseLavaan(m_linear_comp, colnames(oneInt))
spec_lv    <- modsem:::parseLavaan(m_latent_only, colnames(oneInt))

testthat::test_that("parseLavaan detects composites correctly", {
  testthat::expect_equal(spec_full$composites,  c("X", "Z"))
  testthat::expect_equal(spec_mixed$composites, "Z")
  testthat::expect_equal(spec_lin$composites,   "X")
  testthat::expect_length(spec_lv$composites,   0L)
})

testthat::test_that("isCompositeProd is TRUE only when ALL elements are composites", {
  testthat::expect_true(spec_full$isCompositeProd[["XZ"]])
  testthat::expect_false(spec_mixed$isCompositeProd[["XZ"]])
  testthat::expect_false(spec_lin$isCompositeProd[["XZ"]])
  testthat::expect_false(spec_lv$isCompositeProd[["XZ"]])
})

# --- Generated syntax: operator selection ------------------------------------

syntax_full  <- modsem_pi(m_fully_comp,  oneInt, run = FALSE)$syntax
syntax_mixed <- modsem_pi(m_mixed,       oneInt, run = FALSE)$syntax
syntax_lin   <- modsem_pi(m_linear_comp, oneInt, run = FALSE)$syntax

testthat::test_that("fully-composite product uses <~ in generated syntax", {
  testthat::expect_match(syntax_full, "XZ <~", fixed = TRUE)
  testthat::expect_no_match(syntax_full, "XZ =~", fixed = TRUE)
})

testthat::test_that("mixed (latent x composite) product uses =~ in generated syntax", {
  testthat::expect_match(syntax_mixed, "XZ =~", fixed = TRUE)
  testthat::expect_no_match(syntax_mixed, "XZ <~", fixed = TRUE)
})

testthat::test_that("linear composite (not in product) is passed through with <~", {
  testthat::expect_match(syntax_lin, "X <~", fixed = TRUE)
})

# --- Residual covariance toggle -----------------------------------------------

testthat::test_that("no residual covariance syntax for composite product by default", {
  testthat::expect_no_match(syntax_full, "~~", fixed = TRUE)
})

syntax_full_rcov <- modsem_pi(m_fully_comp, oneInt,
                              composite.int.res.cov = TRUE, run = FALSE)$syntax

testthat::test_that("composite.int.res.cov=TRUE adds residual covariances", {
  testthat::expect_match(syntax_full_rcov, "~~", fixed = TRUE)
})

testthat::test_that("mixed product still gets residual covariances by default", {
  testthat::expect_match(syntax_mixed, "~~", fixed = TRUE)
})

# --- method="ca" error --------------------------------------------------------

m_ca_comp_prod <- '
  X <~ x1 + x2
  Z <~ z1 + z2
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

m_ca_mixed_prod <- '
  X =~ x1 + x2 + x3
  Z <~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

# Composite is only used linearly; the interaction is purely between latent variables
m_ca_linear_only <- '
  ATT <~ att1 + att2 + att3 + att4 + att5
  SN  =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
'

testthat::test_that('method="ca" errors on fully-composite product', {
  testthat::expect_error(
    modsem_pi(m_ca_comp_prod, oneInt, method = "ca", run = FALSE),
    regexp = 'method="ca".*composite'
  )
})

testthat::test_that('method="ca" errors on mixed (latent+composite) product', {
  testthat::expect_error(
    modsem_pi(m_ca_mixed_prod, oneInt, method = "ca", run = FALSE),
    regexp = 'method="ca".*composite'
  )
})

testthat::test_that('method="ca" allows linear composites with latent products', {
  testthat::expect_no_error(
    modsem_pi(m_ca_linear_only, TPB, method = "ca", run = FALSE)
  )
})

# --- Estimation: fully-composite interaction ----------------------------------

testthat::test_that("fully-composite interaction converges with dblcent", {
  fit <- suppressWarnings(modsem_pi(m_fully_comp, oneInt, method = "dblcent",
                                    optim.gradient = "numerical"))
  testthat::expect_true(lavaan::lavInspect(fit$lavaan, "converged"))
})

testthat::test_that("fully-composite interaction converges with rca", {
  fit <- suppressWarnings(modsem_pi(m_fully_comp, oneInt, method = "rca",
                                    optim.gradient = "numerical"))
  testthat::expect_true(lavaan::lavInspect(fit$lavaan, "converged"))
})

testthat::test_that("fully-composite interaction converges with uca", {
  fit <- suppressWarnings(modsem_pi(m_fully_comp, oneInt, method = "uca",
                                    optim.gradient = "numerical"))
  testthat::expect_true(lavaan::lavInspect(fit$lavaan, "converged"))
})

# --- Estimation: mixed interaction --------------------------------------------

testthat::test_that("mixed (latent x composite) interaction converges with dblcent", {
  fit <- suppressWarnings(modsem_pi(m_mixed, oneInt, method = "dblcent",
                                    optim.gradient = "numerical"))
  testthat::expect_true(lavaan::lavInspect(fit$lavaan, "converged"))
})

# --- Regression: pure latent models still work --------------------------------

testthat::test_that("pure latent models are unaffected by composite changes", {
  for (m in c("dblcent", "rca", "uca", "ca")) {
    fit <- suppressWarnings(modsem_pi(m_latent_only, oneInt, method = m))
    testthat::expect_true(
      lavaan::lavInspect(fit$lavaan, "converged"),
      info = paste("method =", m)
    )
  }
})

devtools::load_all()
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + c * Z
"

est_m1_lms <- modsem(m1, oneInt, method = "lms", calc.se=FALSE)
testthat::expect_true(est_m1_lms$iterations == 2L) # lowest possible number of iterations for lms
testthat::expect_no_condition(summary(est_m1_lms))

est_m1_qml <- modsem(m1, oneInt, method = "qml", calc.se=FALSE)
testthat::expect_true(est_m1_qml$iterations == 1L) # lowest possible number of iterations for qml
testthat::expect_no_condition(summary(est_m1_qml))


tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + a2 * att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEHAVIOUR_VARIABLE_LONG =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ a * ATT + a * SN + a * PBC
  BEHAVIOUR_VARIABLE_LONG ~ INT + b * PBC

  INT ~~ HELLO_THERE * INT
  BEHAVIOUR_VARIABLE_LONG ~~ HELLO_THERE2 * BEHAVIOUR_VARIABLE_LONG
"

est_tpb_lms <- modsem(tpb, TPB, method = "lms", calc.se=TRUE)
testthat::expect_true(est_tpb_lms$iterations == 2L)
testthat::expect_no_condition(summary(est_tpb_lms))
print(summary(est_tpb_lms, H0 = FALSE, standardized = TRUE))

est_tpb_qml <- modsem(tpb, TPB, method = "qml", calc.se=FALSE)
testthat::expect_true(est_tpb_qml$iterations <= 1L)
testthat::expect_no_condition(summary(est_tpb_qml))


tpb_main <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ a1 * att1 + a2 * att2 + att3 + att4 + att5
  SN =~ s1 * sn1 + sn2
  PBC =~ p1 * pbc1 + pbc2 + pbc3
  INT =~ i1 * int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  BEH ~ INT + a * PBC
"

tpb_cov <- "
  INT ~ ATT + b * SN + c * PBC
"

est_split_lms <- modsem(tpb_main, data = TPB, method = "lms",
                           calc.se=FALSE, cov.syntax = tpb_cov)
testthat::expect_true(est_split_lms$iterations == 2)
testthat::expect_no_condition(summary(est_split_lms))

est_split_qml <- modsem(tpb_main, data = TPB, method = "qml",
                           calc.se=FALSE, cov.syntax = tpb_cov)
testthat::expect_true(est_split_qml$iterations == 1)
testthat::expect_no_condition(summary(est_split_qml))


dupsn1 <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ a1 * att1 + a2 * att2 + att3 + att4 + att5
  SN =~ s1 * sn1 + sn2
  PBC =~ p1 * pbc1 + pbc2 + pbc3 + sn1
  INT =~ i1 * int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + b * SN + c * PBC
  BEH ~ INT + a * PBC
"

est_dupsn1_lms <- modsem(dupsn1, TPB, method = "lms", calc.se=FALSE)
testthat::expect_true(est_dupsn1_lms$iterations == 2)
testthat::expect_no_condition(summary(est_tpb_lms))

est_dupsn1_qml <- modsem(dupsn1, TPB, method = "qml", calc.se=FALSE)
testthat::expect_true(est_dupsn1_qml$iterations == 1)
testthat::expect_no_condition(summary(est_tpb_qml))


tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  SN ~  ATT + PBC
  INT ~ ATT + PBC
  BEH ~ PBC
'
tpb_lms_2 <- modsem(tpb, TPB, method = "lms")
testthat::expect_true(tpb_lms_2$iterations == 2)
summary(tpb_lms_2, H0 = FALSE)


tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

  att4 ~~ att5
  att4 ~~ int1

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + PBC + SN
  BEH ~ PBC
'

tpb_lms_3 <- modsem(tpb, TPB, method = "lms")
testthat::expect_true(tpb_lms_3$iterations == 2)
summary(tpb_lms_3, H0 = FALSE, standardized = TRUE)


testthat::expect_error(
  modsem(tpb, TPB, method = "qml"),
  regexp = "Residual covariances between .* endogenous .* exogenous .*qml.*"
)


tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + a2 * att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

syntax_pi <- get_pi_syntax(tpb, method = "dblcent")
data_pi   <- get_pi_data(tpb, data = TPB, method = "dblcent")
est_lms_pi <- modsem(syntax_pi, data_pi, method = "lms", calc.se = FALSE)
est_qml_pi <- modsem(syntax_pi, data_pi, method = "qml", calc.se = FALSE)
testthat::expect_true(est_lms_pi$iterations == 2)
testthat::expect_true(est_qml_pi$iterations == 1)


m_NA <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ NA*y1 + y2 + y3
  Z =~ z1 + z2 + z3
  Y ~~ 1*Y

  # Inner model
  Y ~ X + Z
'

est_lms_NA <- modsem(m_NA, oneInt, method = "lms")
testthat::expect_true(est_lms_NA$iterations == 2L)

#' @describeIn modsem_predict
#' Computes factor scores or model-implied observed-variable scores for a
#' \code{\link{modsem_da}} model via MAP optimisation.
#' @export
modsem_predict.modsem_da <- function(object,
                                     newdata = NULL,
                                     method = c("EBM", "ML", "Bartlett", "Regression"),
                                     type = c("lv", "ov", "all"),
                                     standardized = FALSE,
                                     se = FALSE,
                                     drop.list.single.group = TRUE,
                                     ...) {
  method <- match.arg(
    arg        = toupper(method)[[1L]],
    choices    = c("EBM", "ML", "BARTLETT", "REGRESSION"),
    several.ok = FALSE
  )

  if      (method == "BARTLETT")   method <- "ML"
  else if (method == "REGRESSION") method <- "EBM"

  predicted <- modsemPredictDA(
    object  = object,
    newdata = newdata,
    method  = method,
    se      = se,
    drop.list.single.group = drop.list.single.group
  )

  Eta <- predicted$Eta
  Y   <- predicted$Y

  type <- match.arg(
    arg        = tolower(type)[[1L]],
    choices    = c("lv", "ov", "all"),
    several.ok = FALSE
  )

  out <- switch(type,
    lv  = Eta,
    ov  = Y,
    all = predicted$All
  )

  if (standardized) {
    predAttrs <- attributes(out)[c("se", "acov")]

    scaled <- scale(out)
    sc     <- attr(scaled, "scaled:scale")
    out    <- modsemMatrix(scaled)

    if (!is.null(predAttrs$se))
      attr(out, "se") <- sweep(predAttrs$se, MARGIN = 2L,
                               STATS = sc[colnames(predAttrs$se)], FUN = "/")
    if (!is.null(predAttrs$acov))
      attr(out, "acov") <- predictDA_RescaleAcov(predAttrs$acov, sc)
  }

  out
}


modsemPredictDA <- function(object, newdata = NULL, method = c("EBM", "ML"),
                            se = FALSE, drop.list.single.group = TRUE) {
  model    <- object$model
  submodels <- model$models

  if (is.null(newdata)) {
    DATA <- object$data
  } else {
    DATA <- prepNewdataDA(object, newdata)
  }

  ETA <- vector("list", length = length(submodels))
  YY  <- vector("list", length = length(submodels))
  ALL <- vector("list", length = length(submodels))

  for (g in seq_along(submodels)) {
    if (is.null(DATA[[g]])) next

    ETA[[g]] <- modsemPredictEtaDA_Group(
      submodel = submodels[[g]],
      data     = DATA[[g]],
      method   = method,
      se       = se
    )

    YY[[g]] <- YFromEta(
      Eta      = ETA[[g]],
      matrices = submodels[[g]]$matrices
    )

    ALL[[g]] <- cbind(
      ETA[[g]],
      YY[[g]][, setdiff(colnames(YY[[g]]), colnames(ETA[[g]])), drop = FALSE]
    )

    if (se) {
      matrices.g <- submodels[[g]]$matrices
      Lambda.g   <- predictDALambda(matrices.g)
      A.eta.g    <- attr(ETA[[g]], "acov")

      ySe <- predictDAEmptySE(
        n = nrow(YY[[g]]),
        vars = colnames(YY[[g]]),
        npatterns = length(A.eta.g)
      )
      yAcov <- vector("list", length = length(A.eta.g))

      allSe <- predictDAEmptySE(
        n = nrow(ALL[[g]]),
        vars = colnames(ALL[[g]]),
        npatterns = length(A.eta.g)
      )
      allAcov <- vector("list", length = length(A.eta.g))

      yKeep <- match(colnames(ALL[[g]])[-seq_len(ncol(ETA[[g]]))], colnames(YY[[g]]))
      Aall  <- rbind(diag(ncol(ETA[[g]])), Lambda.g[yKeep, , drop = FALSE])
      rownames(Aall) <- colnames(ALL[[g]])

      pattern.g <- attr(ETA[[g]], "pattern")

      for (p in seq_along(A.eta.g)) {
        if (is.null(A.eta.g[[p]])) next

        rows.p <- pattern.g == p
        Vy <- Lambda.g %*% A.eta.g[[p]] %*% t(Lambda.g)
        dimnames(Vy) <- list(colnames(YY[[g]]), colnames(YY[[g]]))

        Va <- Aall %*% A.eta.g[[p]] %*% t(Aall)
        dimnames(Va) <- list(colnames(ALL[[g]]), colnames(ALL[[g]]))

        yAcov[[p]]   <- Vy
        ySe[predictDASeRows(rows.p, length(A.eta.g)), ] <- matrix(
          predictDASqrtDiag(Vy),
          nrow = if (length(A.eta.g) == 1L) 1L else sum(rows.p),
          ncol = ncol(ySe),
          byrow = TRUE
        )

        allAcov[[p]] <- Va
        allSe[predictDASeRows(rows.p, length(A.eta.g)), ] <- matrix(
          predictDASqrtDiag(Va),
          nrow = if (length(A.eta.g) == 1L) 1L else sum(rows.p),
          ncol = ncol(allSe),
          byrow = TRUE
        )
      }

      names(yAcov) <- names(A.eta.g)
      names(allAcov) <- names(A.eta.g)

      attr(YY[[g]], "se")   <- ySe
      attr(YY[[g]], "acov") <- yAcov
      attr(attr(YY[[g]], "acov"), "patterns") <- attr(A.eta.g, "patterns")
      attr(YY[[g]], "pattern") <- pattern.g

      attr(ALL[[g]], "se")   <- allSe
      attr(ALL[[g]], "acov") <- allAcov
      attr(attr(ALL[[g]], "acov"), "patterns") <- attr(A.eta.g, "patterns")
      attr(ALL[[g]], "pattern") <- pattern.g
    }
  }

  # collect
  Eta <- do.call(rbind, ETA)
  Y   <- do.call(rbind, YY)
  All <- do.call(rbind, ALL)

  if (se) {
    SE.eta <- predictDACollectSEAttr(ETA, "se")
    SE.y   <- predictDACollectSEAttr(YY,  "se")
    SE.all <- predictDACollectSEAttr(ALL, "se")

    ACOV.eta <- predictDACollectGroupAttr(ETA, "acov")
    ACOV.y   <- predictDACollectGroupAttr(YY,  "acov")
    ACOV.all <- predictDACollectGroupAttr(ALL, "acov")

    group.names <- model$info$group.levels %||% paste0("group", seq_along(ACOV.eta))
    group.names <- group.names[seq_along(ACOV.eta)]

    names(ACOV.eta) <- names(ACOV.y) <- names(ACOV.all) <- group.names

  }

  # Restore original row order when newdata was split by group.
  # Each non-NULL DATA[[g]] carries an "orig.idx" attribute recording its rows'
  # positions in newdata; unlist() collects them in group-processing order so
  # order() gives the permutation back to input order.
  orig.idx <- unlist(lapply(DATA, attr, "orig.idx"))

  if (length(orig.idx)) {
    ord <- order(orig.idx)
    Eta <- Eta[ord, , drop = FALSE]
    Y   <- Y[ord, , drop = FALSE]
    All <- All[ord, , drop = FALSE]

    if (se) {
      if (nrow(SE.eta) == nrow(Eta)) SE.eta <- SE.eta[ord, , drop = FALSE]
      if (nrow(SE.y) == nrow(Y))     SE.y   <- SE.y[ord, , drop = FALSE]
      if (nrow(SE.all) == nrow(All)) SE.all <- SE.all[ord, , drop = FALSE]
    }
  }

  if (se) {
    Eta <- modsemMatrix(Eta, is.public = TRUE)
    Y   <- modsemMatrix(Y, is.public = TRUE)
    All <- modsemMatrix(All, is.public = TRUE)

    attr(Eta, "se")           <- modsemMatrix(SE.eta, is.public = TRUE)
    attr(Eta, "acov")         <- predictDAFormatGroupVcovList(
      ACOV.eta, drop.list.single.group = drop.list.single.group
    )

    attr(Y, "se")   <- modsemMatrix(SE.y, is.public = TRUE)
    attr(Y, "acov") <- predictDAFormatGroupVcovList(
      ACOV.y, drop.list.single.group = drop.list.single.group
    )

    attr(All, "se")   <- modsemMatrix(SE.all, is.public = TRUE)
    attr(All, "acov") <- predictDAFormatGroupVcovList(
      ACOV.all, drop.list.single.group = drop.list.single.group
    )

  } else {
    Eta <- modsemMatrix(Eta, is.public = TRUE)
    Y   <- modsemMatrix(Y, is.public = TRUE)
    All <- modsemMatrix(All, is.public = TRUE)

  }

  list(Eta = Eta, Y = Y, All = All)
}


modsemPredictEtaDA_Group <- function(submodel, data, method = c("EBM", "ML"), se = FALSE) {
  method <- match.arg(toupper(method), c("EBM", "ML"))

  # Convert composite LVs to observed scores upfront: replaces their multi-
  # indicator blocks with a single score column and rebuilds the patternized
  # data to match. This is analogous to how TEMP_OV wrapping handles structural
  # OVs -- the composite score becomes a perfect single-indicator of the LV.
  processed  <- convertCompositesToObs(submodel$matrices, data)
  data       <- processed$data

  # data
  data.split <- data$data.split
  patterns   <- data$patterns
  rowidx     <- data$rowidx
  n.total    <- data$n

  # matrices: fill any zero/NA theta diagonals (TEMP_OV, single-indicator composites)
  matrices <- modsemPredictPreProcessMatrices(processed$matrices, data = data)
  xptr     <- modelMatrixCacheCpp(matrices)

  .f <- switch(method,
    ML  = logLikFromZetaMLCpp,
    EBM = logLikFromZetaEBMCpp,
    mod_msg_stop(paste0("Unrecognized method: ", method, "!"))
  )

  .g <- switch(method,
    ML  = gradLogLikFromZetaMLCpp,
    EBM = gradLogLikFromZetaEBMCpp,
    mod_msg_stop(paste0("Unrecognized method: ", method, "!"))
  )

  .h <- switch(method,
    ML  = hessLogLikFromZetaMLCpp,
    EBM = hessLogLikFromZetaEBMCpp,
    mod_msg_stop(paste0("Unrecognized method: ", method, "!"))
  )

  k   <- NCOL(matrices$phi) + NCOL(matrices$psi)
  dim <- c(colnames(matrices$phi), colnames(matrices$psi))

  Eta <- matrix(NA_real_, nrow = n.total, ncol = k, dimnames = list(NULL, dim))

  if (se) {
    SE <- predictDAEmptySE(
      n = n.total,
      vars = colnames(Eta),
      npatterns = length(data.split)
    )
    ACOV <- vector("list", length(data.split))
    PATTERN <- integer(n.total)
  }

  for (p in seq_along(data.split)) {
    data.p    <- data.split[[p]]
    pattern.p <- patterns[p, , drop = TRUE]
    idx.p     <- which(pattern.p) - 1L

    n     <- NROW(data.p)
    start <- rep(0, k)
    Eta.p <- matrix(NA_real_, nrow = n, ncol = k)

    if (se) {
      VCOV.p  <- vector("list", n)
      SE.p <- predictDAEmptySE(
        n = n,
        vars = colnames(Eta),
        npatterns = length(data.split)
      )
    }

    for (i in seq_len(n)) {
      y <- data.p[i, , drop = TRUE]

      tryCatch({
        opt_i <- stats::nlminb(
          start     = start,
          objective = \(zeta) -.f(zeta = zeta, y = y, xptr = xptr, idx = idx.p),
          gradient  = \(zeta) -.g(zeta = zeta, y = y, xptr = xptr, idx = idx.p)
        )

        Eta.p[i, ] <- impliedEtaFromZetaCpp(zeta = opt_i$par, xptr = xptr)

        if (se) {
          H <- .h(
            zeta = opt_i$par,
            y    = y,
            xptr = xptr,
            idx  = idx.p
          )

          info.zeta <- -H$Hessian
          dimnames(info.zeta) <- list(dim, dim)

          vcov.zeta <- solveFIM(info.zeta)
          dimnames(vcov.zeta) <- list(dim, dim)

          J.eta <- jacobianEtaFromZeta(
            zeta = opt_i$par,
            xptr = xptr
          )
          dimnames(J.eta) <- list(dim, dim)

          vcov.eta <- J.eta %*% vcov.zeta %*% t(J.eta)
          dimnames(vcov.eta) <- list(dim, dim)

          VCOV.p[[i]]  <- vcov.eta
        }

      }, error = function(e) {
        mod_msg_warn(sprintf("nlminb() failed for pattern %i, row %i!", p, i))
        NULL
      })
    }

    Eta[rowidx[[p]], ] <- Eta.p

    if (se) {
      ACOV[[p]]  <- predictDAMeanMatrix(VCOV.p)

      SE.p[TRUE]  <- rep(predictDASqrtDiag(ACOV[[p]]), each = nrow(SE.p))

      SE[predictDASeRows(rowidx[[p]], length(data.split)), ] <- SE.p
      PATTERN[rowidx[[p]]]   <- p
    }
  }

  if (se) {
    pattern.names <- paste0("pattern", seq_along(ACOV))
    names(ACOV) <- pattern.names
    attr(ACOV, "patterns") <- predictDAPatternTable(patterns, pattern.names)

    attr(Eta, "se")               <- SE
    attr(Eta, "acov")             <- ACOV
    attr(Eta, "pattern")          <- PATTERN
  }

  Eta
}


jacobianEtaFromZeta <- function(zeta, xptr, relStep = 1e-4, minAbsPar = 0.0) {
  p <- length(zeta)
  J <- matrix(0, nrow = p, ncol = p)

  incr <- pmax(abs(zeta), minAbsPar) * relStep
  incr[incr == 0] <- relStep

  for (j in seq_len(p)) {
    zp <- zeta
    zm <- zeta
    zp[j] <- zp[j] + incr[j]
    zm[j] <- zm[j] - incr[j]

    J[, j] <- (
      impliedEtaFromZetaCpp(zeta = zp, xptr = xptr) -
      impliedEtaFromZetaCpp(zeta = zm, xptr = xptr)
    ) / (2 * incr[j])
  }

  J
}


predictDALambda <- function(matrices) {
  diagPartitionedMat(matrices$lambdaX, matrices$lambdaY)
}


predictDACollectMatrixAttr <- function(x, which) {
  mats <- lapply(x, attr, which = which)
  mats <- mats[lengths(mats) > 0L]
  do.call(rbind, mats)
}


predictDACollectSEAttr <- function(x, which) {
  mats <- lapply(x, attr, which = which)
  mats <- mats[lengths(mats) > 0L]

  if (all(vapply(mats, nrow, integer(1)) == 1L)) {
    out <- do.call(rbind, mats)
    rownames(out) <- names(mats)
    return(out)
  }

  do.call(rbind, mats)
}


predictDACollectGroupAttr <- function(x, which) {
  out <- lapply(x, attr, which = which)
  out[lengths(out) > 0L]
}


predictDAFormatGroupVcovList <- function(x, drop.list.single.group = TRUE) {
  out <- lapply(x, \(g) {
    if (is.null(g)) return(NULL)

    if (is.matrix(g))
      return(modsemMatrix(g, symmetric = TRUE, is.public = TRUE))

    pattern.table <- attr(g, "patterns")

    if (length(g) == 1L)
      return(modsemMatrix(g[[1L]], symmetric = TRUE, is.public = TRUE))

    out.g <- lapply(g, \(v) {
      if (is.null(v)) return(NULL)
      modsemMatrix(v, symmetric = TRUE, is.public = TRUE)
    })

    attr(out.g, "patterns") <- pattern.table
    out.g
  })

  if (drop.list.single.group && length(out) == 1L) out[[1L]] else out
}


predictDAMeanMatrix <- function(x) {
  x <- Filter(Negate(is.null), x)
  if (!length(x)) return(NULL)

  out <- Reduce(`+`, x) / length(x)
  dimnames(out) <- dimnames(x[[1L]])
  out
}


predictDAEmptySE <- function(n, vars, npatterns) {
  matrix(
    NA_real_,
    nrow = if (npatterns == 1L) 1L else n,
    ncol = length(vars),
    dimnames = list(NULL, vars)
  )
}


predictDASeRows <- function(rows, npatterns) {
  if (npatterns == 1L) TRUE else rows
}


predictDAPatternTable <- function(patterns, pattern.names) {
  patterns <- as.matrix(patterns)
  rownames(patterns) <- pattern.names

  colnames(patterns) <- stringr::str_replace_all(
    string = stringr::str_remove_all(colnames(patterns), pattern = TEMP_OV_PREFIX),
    pattern = OP_REPLACEMENTS_INV
  )

  patterns
}


predictDASqrtDiag <- function(x) {
  if (is.null(x)) return(rep(NA_real_, 0L))

  v <- diag(x)
  v[v < 0] <- NA_real_
  sqrt(v)
}


predictDA_RescaleAcov <- function(acov, sc) {
  if (is.matrix(acov)) {
    d <- 1 / sc[rownames(acov)]
    return(sweep(sweep(acov, 1L, d, "*"), 2L, d, "*"))
  }

  patterns.attr <- attr(acov, "patterns")
  out <- lapply(acov, \(v) if (is.null(v)) NULL else predictDA_RescaleAcov(v, sc))
  attr(out, "patterns") <- patterns.attr
  out
}


YFromEta <- function(Eta, matrices) {
  LambdaX <- matrices$lambdaX
  LambdaY <- matrices$lambdaY
  tauX    <- matrices$tauX
  tauY    <- matrices$tauY

  Lambda <- diagPartitionedMat(LambdaX, LambdaY)
  tau    <- c(tauX, tauY)

  sweep(Eta %*% t(Lambda), MARGIN = 2, STATS = tau, FUN = "+")
}


prepNewdataDA <- function(object, newdata) {
  newdata    <- as.data.frame(newdata)
  newdata    <- addStructOVColumnsDA(newdata, object)  # add TEMP_OV cols
  submodels  <- object$model$models
  group.var  <- object$args$group
  n.groups   <- length(submodels)

  if (!is.null(group.var)) {
    mod_stopif(!group.var %in% colnames(newdata),
           sprintf("Grouping variable '%s' not found in `newdata`.", group.var))

    group.levels <- object$model$info$group.levels
    mod_stopif(is.null(group.levels),
           "Model has a grouping variable but group levels could not be determined.")

    group.values <- as.character(newdata[[group.var]])

    DATA <- lapply(seq_len(n.groups), function(g) {
      cols <- colnames(submodels[[g]]$data$data.full)

      missing.cols <- setdiff(cols, colnames(newdata))
      mod_stopif(length(missing.cols),
             paste0("Missing columns in `newdata`:\n  ", paste(missing.cols, collapse = ", ")))

      rows.g <- group.values == group.levels[[g]]
      if (!any(rows.g)) return(NULL)

      mat.g  <- as.matrix(newdata[rows.g, cols, drop = FALSE])
      storage.mode(mat.g) <- "double"

      data.g <- patternizeMissingDataFIML(mat.g)
      attr(data.g, "orig.idx") <- which(rows.g)
      data.g
    })

  } else {
    cols <- colnames(submodels[[1L]]$data$data.full)

    missing.cols <- setdiff(cols, colnames(newdata))
    mod_stopif(length(missing.cols),
           paste0("Missing columns in `newdata`:\n  ", paste(missing.cols, collapse = ", ")))

    mat <- as.matrix(newdata[, cols, drop = FALSE])
    storage.mode(mat) <- "double"

    DATA <- list(patternizeMissingDataFIML(mat))
  }

  DATA
}


addStructOVColumnsDA <- function(newdata, object) {
  group.info <- object$model$info$group.info
  structovs  <- group.info$structovs
  ovIntTerms <- group.info$ovIntTerms

  if (!length(structovs)) return(newdata)

  # OV interaction terms: compute product columns (e.g. x1:x2 -> x1___INTERACTION___x2)
  for (ovInt in ovIntTerms) {
    vars     <- stringr::str_split_1(ovInt, ":")
    ovIntNew <- stringr::str_replace_all(ovInt, ":", OP_OV_INT)

    missing <- setdiff(vars, colnames(newdata))
    mod_stopif(length(missing),
           paste0("Missing columns in `newdata` required for interaction term '", ovInt, "':\n  ",
           paste(missing, collapse = ", ")))

    newdata[[ovIntNew]] <- apply(newdata[, vars, drop = FALSE], MARGIN = 1L, FUN = prod)
  }

  # Plain structural OVs: check they are present, then add TEMP_OV copy
  ovIntNewNms    <- stringr::str_replace_all(ovIntTerms, ":", OP_OV_INT)
  plainStructOVs <- setdiff(structovs, ovIntNewNms)

  missing <- setdiff(plainStructOVs, colnames(newdata))
  mod_stopif(length(missing),
    paste0("Missing columns in `newdata`:\n  ", paste(missing, collapse = ", "))
  )

  for (ov in structovs) {
    newdata[[paste0(TEMP_OV_PREFIX, ov)]] <- newdata[[ov]]
  }

  newdata
}


modsemPredictPreProcessMatrices <- function(matrices, data, pct.fill = 0.01) {
  # Fill zero/NA diagonal entries in Theta with pct.fill * empirical variance.
  # Handles TEMP_OV indicators (theta = 0 by construction) and composite score
  # columns added by convertCompositesToObs (also initialised to 0).
  fillTheta <- function(Theta) {
    if (!NROW(Theta) || !NCOL(Theta))
      return(Theta)

    nm      <- colnames(Theta)
    bad.idx <- which(diag(Theta) <= 0 | is.na(diag(Theta)))

    for (idx in bad.idx) {
      col <- nm[idx]
      Theta[idx, idx] <- pct.fill * stats::var(data$data.full[, col], na.rm = TRUE)
    }

    Theta
  }

  matrices$thetaDelta   <- fillTheta(matrices$thetaDelta)
  matrices$thetaEpsilon <- fillTheta(matrices$thetaEpsilon)

  matrices
}


convertCompositesToObs <- function(matrices, data) {
  # For each composite LV, replace its multi-indicator block in the data with a
  # single composite-score column (score = W' * indicators). The modified lambda
  # gets a loading of 1 from the LV to its score column; theta is set to 0
  # (fillTheta adds a small regularising value afterward). This is the same
  # pattern as TEMP_OV wrapping for structural OVs.
  W.mat <- matrices$W

  has.composites <- NCOL(W.mat) > 0L &&
                    any(apply(W.mat, MARGIN = 2L, FUN = \(x) any(x != 0)))

  if (!has.composites)
    return(list(matrices = matrices, data = data))

  compLVs  <- colnames(W.mat)[apply(W.mat, 2L, \(x) any(x != 0))]
  compInds <- rownames(W.mat)[apply(W.mat, 1L, \(x) any(x != 0))]

  xiInds  <- rownames(matrices$lambdaX) %||% character(0)
  etaInds <- rownames(matrices$lambdaY) %||% character(0)
  xis     <- colnames(matrices$lambdaX) %||% character(0)
  etas    <- colnames(matrices$lambdaY) %||% character(0)

  xiCompLVs  <- intersect(xis,     compLVs)
  etaCompLVs <- intersect(etas,    compLVs)
  xiRegInds  <- setdiff(xiInds,    compInds)
  etaRegInds <- setdiff(etaInds,   compInds)

  xiScoreNms  <- if (length(xiCompLVs))  paste0(".PRED_COMP__", xiCompLVs)  else character(0)
  etaScoreNms <- if (length(etaCompLVs)) paste0(".PRED_COMP__", etaCompLVs) else character(0)

  # Compute composite scores. W.mat rows and dataFull columns are both indexed
  # by indicator names, so named indexing gives correct alignment.
  # NA propagates naturally when any component indicator is missing.
  dataFull  <- data$data.full
  W.comp    <- W.mat[, compLVs, drop = FALSE]
  shared    <- intersect(rownames(W.comp), colnames(dataFull))
  compScore <- dataFull[, shared, drop = FALSE] %*% W.comp[shared, , drop = FALSE]
  colnames(compScore) <- c(xiScoreNms, etaScoreNms)

  # New data matrix: xi regular inds, xi scores, eta regular inds, eta scores.
  # This ordering matches the new lambdaX/lambdaY row ordering below.
  newDataFull <- cbind(
    dataFull[, xiRegInds,    drop = FALSE],
    compScore[, xiScoreNms,  drop = FALSE],
    dataFull[, etaRegInds,   drop = FALSE],
    compScore[, etaScoreNms, drop = FALSE]
  )

  # lambdaX: keep regular-indicator rows; add a row per xi composite with loading 1
  lX.reg  <- matrices$lambdaX[xiRegInds, , drop = FALSE]
  lX.comp <- matrix(0, nrow = length(xiCompLVs), ncol = length(xis),
                    dimnames = list(xiScoreNms, xis))
  for (i in seq_along(xiCompLVs))
    lX.comp[xiScoreNms[i], xiCompLVs[i]] <- 1

  # lambdaY: same for eta composites
  lY.reg  <- matrices$lambdaY[etaRegInds, , drop = FALSE]
  lY.comp <- matrix(0, nrow = length(etaCompLVs), ncol = length(etas),
                    dimnames = list(etaScoreNms, etas))
  for (i in seq_along(etaCompLVs))
    lY.comp[etaScoreNms[i], etaCompLVs[i]] <- 1

  # thetaDelta / thetaEpsilon: drop composite-indicator rows/cols; add score
  # rows/cols initialised to 0 (fillTheta regularises them)
  newThetaDelta   <- expandTheta(
    matrices$thetaDelta, keep = xiRegInds,  add = xiScoreNms
  )
  newThetaEpsilon <- expandTheta(
    matrices$thetaEpsilon, keep = etaRegInds, add = etaScoreNms
  )

  # tauX / tauY: tauX is stored as a 1-column matrix; extract as named vectors.
  # tau for a composite score = W' * tau_indicators (weighted sum of intercepts).
  tauX.full <- stats::setNames(
    as.vector(matrices$tauX), rownames(matrices$tauX) %||% character(0)
  )
  tauY.full <- stats::setNames(
    as.vector(matrices$tauY), rownames(matrices$tauY) %||% character(0)
  )

  tau.xiComp  <- as.vector(
    t(W.mat[xiInds,  xiCompLVs,  drop = FALSE]) %*% tauX.full[xiInds]
  )
  tau.etaComp <- as.vector(
    t(W.mat[etaInds, etaCompLVs, drop = FALSE]) %*% tauY.full[etaInds]
  )

  newTauX <- c(tauX.full[xiRegInds],  stats::setNames(tau.xiComp,  xiScoreNms))
  newTauY <- c(tauY.full[etaRegInds], stats::setNames(tau.etaComp, etaScoreNms))

  matrices$lambdaX      <- rbind(lX.reg, lX.comp)
  matrices$lambdaY      <- rbind(lY.reg, lY.comp)
  matrices$thetaDelta   <- newThetaDelta
  matrices$thetaEpsilon <- newThetaEpsilon
  matrices$tauX         <- newTauX
  matrices$tauY         <- newTauY

  list(matrices = matrices, data = patternizeMissingDataFIML(newDataFull))
}


expandTheta <- function(Theta, keep, add) {
  thetaKept <- Theta[keep, keep, drop = FALSE]
  n.add     <- length(add)
  if (!n.add) return(thetaKept)

  n.keep <- nrow(thetaKept)
  nm     <- c(keep, add)
  out    <- matrix(0, nrow = n.keep + n.add, ncol = n.keep + n.add,
                   dimnames = list(nm, nm))
  out[keep, keep] <- thetaKept
  out
}

#' Reliability‑Corrected Single‑Item SEM
#'
#' Replace (some of) the first‑order latent variables in a lavaan measurement
#' model by single composite indicators whose error variances are fixed from
#' Cronbach's \eqn{\alpha}. The function returns a modified lavaan model
#' syntax together with an augmented data set that contains the newly created
#' composite variables, so that you can fit the full SEM in a single step.
#'
#' The resulting object can be fed directly into \code{modsem} or \code{lavaan::sem}
#' by supplying \code{syntax = $syntax} and \code{data = $data}.
#'
#' @param syntax A character string containing lavaan model syntax.  Must at
#'   least include the measurement relations (\code{=~}).
#'
#' @param data A \code{data.frame}, \code{tibble} or object coercible to a data frame that
#'   holds the raw observed indicators.
#'
#' @param choose \emph{Optional.} Character vector with the names of the latent
#'   variables you wish to replace by single indicators.  Defaults to \strong{all}
#'   first‑order latent variables in \code{syntax}.
#'
#' @param scale.corrected Should reliability-corrected items be scale-corrected? If \code{TRUE}
#'   reliability-corrected single items are corrected for differences in factor loadings between
#'   the items. Default is \code{TRUE}.
#'
#' @param missing How should missing values be handled in the CFA? Default is \code{"fiml"}.
#'
#' @param warn.lav Should warnings from \code{lavaan::cfa} be displayed? If \code{FALSE}, they
#'   are suppressed.
#'
#' @param group Character. A variable name in the data frame defining the groups in a multiple
#'   group analysis
#'
#' @return An object of S3 class \code{ModsemRelcorr} (a named list) with elements:
#' \describe{
#'   \item{\code{syntax}}{Modified lavaan syntax string.}
#'   \item{\code{data}}{Data frame with additional composite indicator columns.}
#'   \item{\code{parTable}}{Parameter table corresponding to `syntax`.}
#'   \item{\code{reliability}}{Named numeric vector of reliabilities (one per latent
#'     variable).}
#'   \item{\code{AVE}}{Named numeric vector with Average Variance Extracted values.}
#'   \item{\code{lVs}}{Character vector of latent variables that were corrected.}
#'   \item{\code{single.items}}{Character vector with names for the generated reliability corrected items}
#' }
#'
#' @examples
#' \dontrun{
#' tpb_uk <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'  ATT =~ att3 + att2 + att1 + att4
#'  SN =~ sn4 + sn2 + sn3 + sn1
#'  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
#'  INT =~ int2 + int1 + int3 + int4
#'  BEH =~ beh3 + beh2 + beh1 + beh4
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'  INT ~ ATT + SN + PBC
#'  BEH ~ INT + PBC
#'  BEH ~ INT:PBC
#' "
#'
#' corrected <- relcorr_single_item(syntax = tpb_uk, data = TPB_UK)
#' print(corrected)
#'
#' syntax <- corrected$syntax
#' data   <- corrected$data
#'
#' est_dca <- modsem(syntax, data = data, method = "dblcent")
#' est_lms <- modsem(syntax, data = data, method="lms", nodes=32)
#' summary(est_lms)
#' }
#'
#' @export
relcorr_single_item <- function(syntax,
                                data,
                                choose = NULL,
                                scale.corrected = TRUE,
                                missing = "fiml",
                                warn.lav = TRUE,
                                group = NULL) {
  data <- as.data.frame(data)

  if (is.null(group)) {
    return(relcorrSingleItemGroup(
      syntax = syntax, data = data, choose = choose,
      scale.corrected = scale.corrected,
      missing = missing, warn.lav = warn.lav
    ))
  }

  mod_stopif(length(group) > 1L, "`group` must be a character vector of length 1!")
  mod_stopif(!group %in% colnames(data), sprintf("Unable to find `%s` in data!", group))

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = syntax,
    cov.syntax   = NULL,
    data         = data,
    group        = group
  )

  parTable      <- group.info$parTable
  group.label   <- as.character(group.info$group.levels)
  groups        <- getGroupsParTable(parTable)
  data[[group]] <- as.character(data[[group]])

  out <- list(parTable = NULL, syntax = NULL, data = NULL,
              relcorr.groups = vector("list", length = length(groups)))

  for (g in groups) {
    g.val      <- group.label[[g]]
    data.g     <- data[data[[group]] == g.val, , drop = FALSE]
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    parTable.g <- parTable.g[, colnames(parTable.g) != "group", drop = FALSE]

    relcorr.g <- relcorrSingleItemGroup(
      parTable        = parTable.g,
      data            = data.g,
      choose          = choose,
      scale.corrected = scale.corrected,
      missing         = missing,
      warn.lav        = warn.lav
    )

    out$relcorr.groups[[g]] <- relcorr.g

    newParTable.g <- relcorr.g$parTable
    newParTable.g$group <- g
    newData.g <- relcorr.g$data

    out$parTable <- rbind(out$parTable, newParTable.g)
    out$data     <- rbind(out$data, newData.g)
  }

  out$syntax <- parTableToSyntax(out$parTable)
  class(out) <- "MultiGroupModsemRelcorr"

  out
}


relcorrSingleItemGroup <- function(syntax = NULL,
                                   parTable = NULL,
                                   data,
                                   choose = NULL,
                                   scale.corrected = TRUE,
                                   missing = "fiml",
                                   warn.lav = TRUE) {
  data <- as.data.frame(data)

  if (is.null(parTable))
    parTable <- modsemify(syntax)

  intTerms      <- parTable$rhs[grepl(":", parTable$rhs)]

  parTableOuter <- parTable[parTable$op == "=~", ]
  parTableInner <- parTable[parTable$op != "=~", ]

  lVs            <- getLVs(parTable)
  allIndsLVs     <- getIndsLVs(parTable, lVs = lVs)
  allInds        <- unique(unlist(allIndsLVs))
  higherOrderLVs <- getHigherOrderLVs(parTable)
  firstOrderLVs  <- setdiff(lVs, higherOrderLVs)

  mod_stopif(any(higherOrderLVs %in% allInds),
         "Third order (or higher) latent variables are not allowed (yet)!")

  choose <- if (is.null(choose)) lVs else choose
  ignore <- setdiff(lVs, choose)

  mod_warnif(any(!choose %in% lVs), paste0("Could not find latent variables:\n",
         paste(choose[!choose %in% lVs], collapse = ", ")))
  choose <- choose[choose %in% lVs]

  ignoreRows <- parTableOuter$lhs %in% ignore

  parTableInner <- rbind(parTableInner, parTableOuter[ignoreRows, ])
  parTableOuter <- parTableOuter[!ignoreRows, ]

  mod_stopif(!NROW(parTableOuter), paste0("Cannot ignore all latent variables when creating ",
         "reliability corrected items!"))

  mod_warnif(any(!canBeNumeric(parTableOuter$mod, includeNA = TRUE)),
         paste0("Labels and modifiers in the measurement model used to create reliability ",
         "corrected single items will be ignored!\nThis may cause the model constraints ",
         "in the structural model to break!"))

  lVs  <- setdiff(firstOrderLVs, ignore)
  indsLVs <- getIndsLVs(parTableOuter, lVs = lVs)
  inds <- unlist(indsLVs)
  R <- stats::cov(data[inds], use = "pairwise.complete.obs")
  k <- length(lVs)

  # Add residual Covariances
  rescovind <- parTableInner$lhs %in% inds &
               parTableInner$rhs %in% inds &
               parTableInner$op == "~~"
  parTableOuter <- rbind(parTableOuter, parTableInner[rescovind, ])
  parTableInner <- parTableInner[!rescovind, ]

  # Check if there are any indicators in the structural model
  indsInInner <- parTableInner$lhs %in% inds | parTableInner$rhs %in% inds
  if (any(indsInInner)) {
    removing <- apply(
      X = parTableInner[indsInInner, c("lhs", "op", "rhs"), drop = FALSE],
      MARGIN = 1L, FUN = paste0, collapse = ""
    )

    mod_msg_warn(
      paste0("Removing expressions containing indicators!\n",
             paste0(removing, collapse = ", "))
    )

    parTableInner <- parTableInner[!indsInInner, ]
  }

  wrap      <- if (warn.lav) \(x) x else suppressWarnings
  cfaSyntax <- parTableToSyntax(parTableOuter)
  cfa       <- wrap(lavaan::cfa(cfaSyntax, data = data, se = "none", missing = missing))

  cov.lv      <- lavaan::lavInspect(cfa, "cov.lv")
  stdSolution <- lavaan::lavInspect(cfa, "std")
  lambda <- stdSolution$lambda
  theta  <- stdSolution$theta

  avefunc <- \(lV) calcAVE(lV, lambda.std = lambda)
  relfunc <- \(lV) calcChronbach(R = R, x = indsLVs[[lV]])

  rel.std <- vapply(lVs, FUN.VALUE = numeric(1L), relfunc)
  AVE.std <- vapply(lVs, FUN.VALUE = numeric(1L), avefunc)
  res.std <- 1 - rel.std
  singleInds <- getSingleIndNames(lVs)

  rchar <- c(letters, seq_len(9))
  while(any(singleInds %in% colnames(data))) {
    singleInds <- paste0(singleInds, sample(rchar, length(singleInds)))
  }

  varNewInds <- structure(numeric(length(lVs)), names=singleInds)
  newData <- data

  if (scale.corrected) {
    res <- stats::setNames(numeric(length(lVs)), nm = lVs)

    for (lV in lVs) {
      ITEM <- getScaleCorrectedItem(parTable = parTable, lV = lV,
                                    data = data, cfa = cfa)

      newIndName <- singleInds[[lV]]
      newInd     <- ITEM$item
      residual   <- ITEM$residual

      newData[[newIndName]] <- newInd
      res[[lV]] <- residual
    }

  } else {
    for (lV in lVs) {
      indsLV     <- indsLVs[[lV]]
      newIndName <- singleInds[[lV]]
      X          <- as.matrix(data[indsLV])
      newInd     <- rowMeans(data[indsLV], na.rm = TRUE)

      newData[[newIndName]]    <- newInd
      varNewInds[[newIndName]] <- stats::var(newInd, na.rm = TRUE)
    }

    rel <- rel.std * varNewInds
    res <- res.std * varNewInds
  }

  if (length(higherOrderLVs)) {
    for (lV in intersect(lVs, allInds)) {
      res[[lV]] <- res[[lV]] + getHigherOrderResidual(lV = lV, cfa = cfa)
    }
  }

  resCovMat <- getCompositeRVCOV(lVs = lVs, parTable = parTable, cfa = cfa,
                                 scale.corrected = scale.corrected)
  # Ensure diagonals match the already computed residual variances
  diag(resCovMat) <- as.numeric(res[lVs])

  # first order lVs
  newParTable <- parTableInner
  firstOrder  <- !lVs %in% allInds
  secondOrder <- lVs %in% allInds

  if (any(firstOrder)) {
    lVs1        <- lVs[firstOrder]
    singleInds1 <- singleInds[firstOrder]
    k1          <- length(lVs1)
    res1        <- format(res[firstOrder], scientific = FALSE)

    lhs <- unname(c(lVs1, singleInds1))
    rhs <- unname(c(singleInds1, singleInds1))
    op  <- c(rep("=~", k1), rep("~~", k1))
    mod <- unname(c(rep("1", k1), res1))

    newParTable <- rbind(
      newParTable, data.frame(lhs = lhs, op = op, rhs = rhs, mod = mod)
    )
  }

  if (any(secondOrder)) {
    mod_msg_note(
      paste0("Correcting first order items/lVs for higher order measurement model!",
      "Note that the first order LVs no longer can be used directly in the",
      "structural model!")
    )

    lVs2        <- lVs[secondOrder]
    singleInds2 <- singleInds[secondOrder]
    k2          <- length(lVs2)
    res2        <- res[secondOrder]

    for (hlV in higherOrderLVs) {
      lVs2h <- intersect(lVs2, allIndsLVs[[hlV]])
      singleInds2h <- singleInds2[lVs2h]
      res2h <- format(res2[lVs2h], scientific = FALSE)
      k2h   <- length(singleInds2h)

      lhs <- unname(c(rep(hlV, k2h), singleInds2h))
      rhs <- unname(c(singleInds2h, singleInds2h))
      op  <- c(rep("=~", k2h), rep("~~", k2h))
      mod <- unname(c(rep("", k2h), res2h))

      newParTable <- rbind(
        newParTable, data.frame(lhs = lhs, op = op, rhs = rhs, mod = mod)
      )
    }
  }

  if (length(lVs)) {
    lvPairs <- utils::combn(lVs, 2, simplify = FALSE)

    addRows <- lapply(lvPairs, function(pair) {
      lv_i <- pair[1]
      lv_j <- pair[2]
      val  <- resCovMat[lv_i, lv_j]

      if (abs(val) < 1e-12)
        return(NULL)

      data.frame(
        lhs = unname(singleInds[[lv_i]]),
        op  = "~~",
        rhs = unname(singleInds[[lv_j]]),
        mod = format(unname(val), scientific=FALSE)
      )
    })

    newRows <- do.call(rbind, Filter(Negate(is.null), addRows))

    if (!is.null(addRows) && NROW(newRows))
      newParTable <- rbind(newParTable, newRows)
  }

  newSyntax <- parTableToSyntax(newParTable)

  out <- list(syntax = newSyntax, data = newData,
              parTable = newParTable, reliability = rel.std,
              AVE = AVE.std, lVs = lVs, single.items = singleInds,
              intTerms = intTerms, residuals = res, res.std = res.std,
              cov.lv = cov.lv, residual_covariances = resCovMat)

  class(out) <- c("list", "ModsemRelcorr")

  out
}


getScaleCorrectedItem <- function(parTable, lV, data, cfa) {
  measr  <- parTable[parTable$lhs == lV & parTable$op == "=~", ]
  inds   <- unique(measr$rhs)

  matrices <- lavaan::lavInspect(cfa, what = "coef")

  theta  <- matrices$theta[inds, inds]
  lambda <- as.vector(matrices$lambda[inds, lV])

  I <- rep(1, length(inds))
  X <- as.matrix(data[, inds])

  # Ignore loadings for values in X which are missing
  L <- matrix(lambda, nrow = nrow(X), ncol = ncol(X), byrow = TRUE)
  L[is.na(X)] <- NA_real_

  item <- rowSums(X, na.rm = TRUE) / rowSums(L, na.rm = TRUE)

  residual <- (t(I) %*% theta %*% I) / sum(lambda)^2

  list(X = X, item = as.numeric(item), residual = as.numeric(residual))
}


getHigherOrderResidual <- function(lV, cfa) {
  parTable <- lavaan::parameterEstimates(cfa)
  inds     <- getInds(parTable)

  if (!lV %in% inds) return(0)

  residual <- parTable[parTable$lhs == lV &
                       parTable$op == "~~" &
                       parTable$rhs == lV, "est"]

  ifelse(!length(residual), yes = 0, no = residual[[1L]])
}


simulateCrossResCovRCS <- function(corrected,
                                   elemsInIntTerms,
                                   mc.reps = 3e4,
                                   parTable,
                                   include.normal.inds = FALSE) {
  if ( inherits(corrected, "ModsemRelcorr") &&
      !inherits(corrected, "MultiGroupModsemRelcorr")) {

    return(simulateCrossResCovRCS_Group(
      corrected           = corrected,
      elemsInIntTerms     = elemsInIntTerms,
      mc.reps             = mc.reps,
      parTable            = parTable,
      include.normal.inds = include.normal.inds
    ))
  }

  newRows <- NULL

  for (g in seq_along(corrected$relcorr.groups)) {
    corrected.g <- corrected$relcorr.groups[[g]]

    simCrossResCov.g <- simulateCrossResCovRCS_Group(
      corrected           = corrected.g,
      elemsInIntTerms     = elemsInIntTerms,
      mc.reps             = mc.reps,
      parTable            = parTable,
      include.normal.inds = include.normal.inds
    )

    newRows.g <- simCrossResCov.g$rows
    newRows.g$group <- g

    newRows <- rbind(newRows, newRows.g)
  }

  syntaxAdditions <- parTableToSyntax(newRows)

  list(syntax = syntaxAdditions, rows = newRows)
}


simulateCrossResCovRCS_Group <- function(corrected,
                                         elemsInIntTerms,
                                         mc.reps = 3e4,
                                         parTable,
                                         include.normal.inds = FALSE) {
  EMPTY <- list(syntax = "", rows = NULL)

  if (!length(elemsInIntTerms))
    return(EMPTY)

  intTerms        <- names(elemsInIntTerms)
  single.items    <- corrected$single.items
  residuals       <- corrected$residuals
  constructs      <- names(single.items)

  intTermIsRCS <- vapply(elemsInIntTerms, FUN.VALUE = logical(1L),
                         FUN = \(elems) all(elems %in% constructs))
  elemsInIntTerms <- elemsInIntTerms[intTermIsRCS]
  intTerms <- intTerms[intTermIsRCS]

  if (!length(elemsInIntTerms))
    return(EMPTY)

  Phi   <- corrected$cov.lv
  Theta <- diag(residuals)
  nm    <- names(residuals)
  Phi   <- Phi[nm, nm]
  dimnames(Theta) <- list(nm, nm)

  XI  <- mvtnorm::rmvnorm(mc.reps, sigma = Phi)
  EPS <- mvtnorm::rmvnorm(mc.reps, sigma = Theta)
  Y   <- XI + EPS

  CONSTRUCTS <- as.data.frame(XI)
  ITEMS      <- as.data.frame(Y)

  colnames(CONSTRUCTS) <- nm
  colnames(ITEMS)      <- single.items[nm]

  if (include.normal.inds) {
    RESIDUALS  <- as.data.frame(EPS)
    colnames(RESIDUALS)  <- single.items[nm]
  } else RESIDUALS <- list()

  for (intTerm in intTerms) {
    elems  <- elemsInIntTerms[[intTerm]]
    items  <- single.items[elems]
    nameXZ <- paste0(items, collapse = "")

    xz.construct <- multiplyIndicatorsCpp(CONSTRUCTS[elems])
    xz.item      <- multiplyIndicatorsCpp(ITEMS[items])

    CONSTRUCTS[[intTerm]] <- xz.construct
    ITEMS[[nameXZ]]      <- xz.item
  }

  Epsilon <- stats::cov(ITEMS) - stats::cov(CONSTRUCTS)
  newRows <- NULL

  for (i in seq_len(NROW(Epsilon))) {
    for (j in seq_len(i)) {
      mod <- format(Epsilon[i, j], scientific = FALSE)
      lhs <- rownames(Epsilon)[[i]]
      rhs <- colnames(Epsilon)[[j]]

      elems.lhs <- stringr::str_split_1(lhs, "composite_")
      elems.rhs <- stringr::str_split_1(rhs, "composite_")

      elems.lhs <- elems.lhs[elems.lhs != ""]
      elems.rhs <- elems.rhs[elems.rhs != ""]

      matching <- parTable[((parTable$lhs == lhs & parTable$rhs == rhs) |
                            (parTable$lhs == rhs & parTable$rhs == lhs)) &
                           parTable$op == "~~", ]

      if (NROW(matching) || !length(intersect(elems.lhs, elems.rhs)))
        next

      newRows <- rbind(
        newRows,
        createParTableRow(c(lhs, rhs), op = "~~", mod = mod)
      )
    }
  }

  syntaxAdditions <- parTableToSyntax(newRows)

  list(syntax = syntaxAdditions, rows = newRows)
}


#' @export
print.ModsemRelcorr <- function(x,...) {
  sep <- strrep(" ", 4)
  indent <- strrep(" ", 2)

  lVs <- x$lVs
  lVsf <- format(paste0(lVs, ":"), justify = "left")
  AVE_rel <- format(c(x$AVE, x$reliability), digits = 3, nsmall = 2, justify = "right")

  idxAVE <- seq_along(x$AVE)
  AVEs  <- AVE_rel[idxAVE]
  rels  <- AVE_rel[-idxAVE]

  cat("Average Variance Extracted:\n")
  for (i in seq_along(lVs)) {
    lV <- lVsf[[i]]
    AVE <- AVEs[[i]]
    cat(paste0(indent, lV, sep, AVE, "\n"))
  }

  cat(paste0("\nChronbach's Alpha:\n"))
  for (i in seq_along(lVs)) {
    lV <- lVsf[[i]]
    rel <- rels[[i]]
    cat(paste0(indent, lV, sep, rel, "\n"))
  }

  cat(paste0("\nGenerated Syntax:\n"))
  cat(padString(x$syntax, pad = indent))

  cat(paste0("Generated Items:\n"))
  datastr <- utils::capture.output(utils::str(x$data[x$single.items]))
  cat(padString(datastr, pad = indent))
}


getSingleIndNames <- function(lVs) {
  structure(paste0("composite_", lVs, "_"), names=lVs)
}


calcChronbach <- function(R, x) {
  Rx <- R[x, x]
  k <- length(x)

  if (k == 1) return(1) # should be treated like an observed variable

  mod_warnif(any(Rx < 0), paste0("Some item covariances are negative! Consider ",
         "recoding your items!"))

  Rx <- abs(Rx)
  varSum <- sum(diag(Rx))
  covSum <- sum(Rx[lower.tri(Rx)])

  (k / (k - 1)) * (1 - varSum / (varSum + 2 * covSum))
}


calcConstructRel <- function(lambda.std, lV) {
  lambda.std <- lambda.std[, lV, drop=FALSE]
  lambda.std <- lambda.std[lambda.std != 0]

  mod_warnif(any(lambda.std < 0), paste0("Some items have negative factor loadings!\n",
         "Consider recoding your items!"))

  sum(lambda.std) ^ 2 / (sum(lambda.std) ^ 2 + sum(1 - lambda.std ^ 2))
}


calcAVE <- function(lambda.std, lV) {
  inds <- rownames(lambda.std)

  lambda.std <- structure(lambda.std[, lV, drop=TRUE], names = inds)
  lambda.std <- lambda.std[lambda.std != 0]

  sum(lambda.std ^ 2) / length(lambda.std)
}


getCompositeRVCOV <- function(lVs, parTable, cfa, scale.corrected) {
  matrices <- lavaan::lavInspect(cfa, what = "coef")
  lambda   <- matrices$lambda
  theta    <- matrices$theta

  inds <- rownames(lambda)[apply(lambda, MARGIN = 1L, FUN = \(x) any(x != 0))]
  lambda <- lambda[inds, lVs, drop = FALSE]
  theta  <- theta[inds, inds, drop = FALSE]

  I <- lambda
  I[I!=0] <- 1

  T <- t(I) %*% theta %*% I

  FUN <- if (scale.corrected) sum else \(x) sum(x != 0)

  l <- apply(lambda, MARGIN = 2L, FUN = FUN)
  l.inv <- 1 / l
  l.inv[l <= 0] <- 1L

  if (length(l.inv) > 1) {
    L.inv <- diag(l.inv)
    dimnames(L.inv) <- list(names(l.inv), names(l.inv))

  } else {
    L.inv <- matrix(l.inv, nrow = 1, ncol = 1)
    dimnames(L.inv) <- list(names(l.inv), names(l.inv))
  }

  L.inv %*% T %*% L.inv
}

modsemMatrix <- function(mat, symmetric = !is.null(mat) && isSymmetric(mat), is.public = FALSE) {
  if (is.null(mat)) return(mat)

  if (is.public) {
    rn <- rownames(mat)
    cn <- colnames(mat)

    if (!is.null(rn)) {
      isTempRn <- startsWith(rn, prefix = TEMP_OV_PREFIX)
      rnClean <- stringr::str_remove_all(rn, pattern = TEMP_OV_PREFIX)
      isDupTempRn <- rnClean[isTempRn] %in% rnClean[!isTempRn]

      keepRn <- rep(TRUE, length(rn))
      keepRn[isTempRn][isDupTempRn] <- FALSE
      mat <- mat[keepRn, , drop = FALSE]

      rownames(mat) <- stringr::str_replace_all(
        string = rnClean[keepRn],
        pattern = OP_REPLACEMENTS_INV
      )
    }

    if (!is.null(cn)) {
      isTempCn <- startsWith(cn, prefix = TEMP_OV_PREFIX)
      cnClean <- stringr::str_remove_all(cn, pattern = TEMP_OV_PREFIX)
      isDupTempCn <- cnClean[isTempCn] %in% cnClean[!isTempCn]

      keepCn <- rep(TRUE, length(cn))
      keepCn[isTempCn][isDupTempCn] <- FALSE
      mat <- mat[ , keepCn, drop = FALSE]

      colnames(mat) <- stringr::str_replace_all(
        string = cnClean[keepCn],
        pattern = OP_REPLACEMENTS_INV
      )
    }
  }

  class(mat) <- unique(c("ModsemMatrix", class(mat)))

  if (symmetric)
    class(mat) <- unique(c("ModsemSymmetricMatrix", class(mat)))

  mat
}


#' @export
print.ModsemSymmetricMatrix <- function(x, digits = 3, sep = " ", ...) {
  if (nrow(x) == ncol(x)) x[upper.tri(x)] <- NA

  print.ModsemMatrix(x, digits = digits, sep = sep, ...)
}


#' @export
print.ModsemMatrix <- function(x, digits = 3, shift = 0L,
                               quote = FALSE, right = TRUE, ...) {
  y <- matrix(formatNumeric(x, digits = digits), nrow = nrow(x),
              ncol = ncol(x), dimnames = dimnames(x))
  attributes(y) <- attributes(x)

  y <- unclass(y)
  # Remove NAs
  y[is.na(x)] <- ""

  if (!is.null(colnames(x))) {
    colnames(y) <- abbreviate(colnames(x), minlength = digits + 3L) }
  if (shift > 0L) {
    empty.string <- rep(strrep(x = " ", times = shift), times = nrow(x))

    if (!is.null(rownames(x))) rownames(y) <- paste0(empty.string, rownames(x))
    else                       rownames(y) <- empty.string
  }

  print(y, ..., quote = quote, right = right)

  invisible(x)
}


modsemParTable <- function(parTable, is.public = FALSE) {
  if (is.null(parTable)) return(parTable)

  if (is.public) {
    paramCols <- c("lhs", "op", "rhs", "label")
    paramCols <- intersect(paramCols, colnames(parTable))

    for (col in paramCols) {
      parTable[[col]] <- stringr::str_replace_all(
        string = parTable[[col]],
        pattern = OP_REPLACEMENTS_INV
      )
    }
  }

  class(parTable) <- unique(c("ModsemParTable", class(parTable)))
  parTable
}


modsemVector <- function(vec, is.public = FALSE) {
  if (is.null(vec)) return(vec)

  if (is.public) {
    nm <- names(vec)
    isTemp <- startsWith(nm, prefix = TEMP_OV_PREFIX)

    if (any(isTemp)) {
      clean <- stringr::str_remove_all(nm, pattern = TEMP_OV_PREFIX)

      isDupTemp <- clean[isTemp] %in% clean[!isTemp]

      keep <- rep(TRUE, length(nm))
      keep[isTemp][isDupTemp] <- FALSE

      vec <- stats::setNames(vec[keep], nm = clean[keep])
    }

    names(vec) <- stringr::str_replace_all(
      string = names(vec),
      pattern = OP_REPLACEMENTS_INV
    )
  }

  class(vec) <- unique(c("ModsemVector", class(vec)))

  vec
}


#' @export
print.ModsemVector <- function(x, digits = 3L, ...) {
  y <- formatNumeric(x, digits = digits)
  print(y, quote = FALSE)
}


#' @export
print.ModsemParTable <- function(x, nd = 3L, ...) {
  row.names <- rownames(x)
  y <- lapply(x, \(x) if (is.numeric(x)) round(x, nd) else x) |>
    as.data.frame()

  rownames(y) <- row.names

  print(y, ...)

  invisible(x)
}

# Box-drawing characters live in BOX_CHARS (see zzz.R); the active set
# (unicode or ASCII fallback) is selected at load time in .onLoad().


#' Get the simple slopes of a SEM model
#'
#' This function calculates simple slopes (predicted values of the outcome variable)
#' at user-specified values of the focal predictor (\code{x}) and moderator (\code{z})
#' in a structural equation modeling (SEM) framework. It supports interaction terms
#' (\code{xz}), computes standard errors (SE), and optionally returns confidence or
#' prediction intervals for these predicted values. It also provides p-values for
#' hypothesis testing. This is useful for visualizing and interpreting moderation
#' effects or to see how the slope changes at different values of the moderator.
#'
#' @param x The name of the variable on the x-axis (focal predictor).
#' @param z The name of the moderator variable.
#' @param y The name of the outcome variable.
#' @param model An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}},
#'   \code{\link{modsem_mplus}}, or a \code{lavaan} object. This should be a fitted SEM
#'   model that includes paths for \code{y ~ x + z + x:z}.
#' @param vals_x Numeric vector of values of \code{x} at which to compute predicted
#'   slopes. Defaults to \code{-3:3}. If \code{rescale = TRUE}, these values are taken
#'   relative to the mean and standard deviation of \code{x}. A higher density of points
#'   (e.g., \code{seq(-3, 3, 0.1)}) will produce smoother curves and confidence bands.
#' @param vals_z Numeric vector of values of the moderator \code{z} at which to compute
#'   predicted slopes. Defaults to \code{-1:1}. If \code{rescale = TRUE}, these values
#'   are taken relative to the mean and standard deviation of \code{z}. Each unique value
#'   of \code{z} generates a separate regression line \code{y ~ x | z}.
#' @param rescale Logical. If \code{TRUE} (default), \code{x} and \code{z} are standardized
#'   according to their means and standard deviations in the model. The values in
#'   \code{vals_x} and \code{vals_z} are interpreted in those standardized units. The
#'   raw (unscaled) values corresponding to these standardized points will be displayed
#'   in the returned table.
#' @param ci_width A numeric value between 0 and 1 indicating the confidence (or
#'   prediction) interval width. The default is 0.95 (i.e., 95\% interval).
#' @param ci_type A string indicating whether to compute a \code{"confidence"} interval
#'   for the predicted mean (\emph{i.e.}, uncertainty in the regression line) or a
#'   \code{"prediction"} interval for individual outcomes. The default is
#'   \code{"confidence"}. If \code{"prediction"}, the residual variance is added to the
#'   variance of the fitted mean, resulting in a wider interval.
#' @param relative_h0 Logical. If \code{TRUE} (default), hypothesis tests for the
#'   predicted values (\code{predicted - h0}) assume \code{h0} is the model-estimated
#'   mean of \code{y}. If \code{FALSE}, the null hypothesis is \code{h0 = 0}.
#' @param standardized Should coefficients be standardized beforehand?
#' @param xz The name of the interaction term (\code{x:z}). If \code{NULL}, it will
#'   be created by combining \code{x} and \code{z} with a colon (e.g., \code{"x:z"}).
#'   Some backends may remove or alter the colon symbol, so the function tries to
#'   account for that internally.
#'
#' @param ... Additional arguments passed to lower-level functions or other internal
#'   helpers.
#'
#' @details
#' \strong{Computation Steps}
#' \enumerate{
#'    \item The function extracts parameter estimates (and, if necessary, their covariance
#'          matrix) from the fitted SEM model (\code{model}).
#'    \item It identifies the coefficients for \code{x}, \code{z}, and \code{x:z} in the model's
#'          parameter table, as well as the variance of \code{x}, \code{z} and the residual variance
#'          of \code{y}.
#'    \item If \code{xz} is not provided, it will be constructed by combining \code{x} and
#'          \code{z} with a colon (\code{":"}). Depending on the approach used to estimate the model,
#'          the colon may be removed or replaced internally; the function attempts to reconcile that.
#'
#'   \item A grid of \code{x} and \code{z} values is created from \code{vals_x} and
#'         \code{vals_z}. If \code{rescale = TRUE}, these values are transformed back into raw
#'         metric units for display in the output.
#'   \item For each point in the grid, a predicted value of \code{y} is computed via
#'         \code{(beta0 + beta_x * x + beta_z * z + beta_xz * x * z)} and, if included, a
#'         mean offset.
#'   \item The standard error (\code{std.error}) is derived from the covariance matrix of
#'         the relevant parameters, and if \code{ci_type = "prediction"}, adds the residual
#'         variance.
#'   \item Confidence (or prediction) intervals are formed using \code{ci_width} (defaulting
#'         to 95\%). The result is a table-like data frame with predicted values, CIs,
#'         standard errors, z-values, and p-values.
#' }
#'
#' @return A \code{data.frame} (invisibly inheriting class \code{"simple_slopes"})
#' with columns:
#' \itemize{
#'   \item \code{vals_x}, \code{vals_z}: The requested grid values of \code{x} and \code{z}.
#'   \item \code{predicted}: The predicted value of \code{y} at that combination of
#'         \code{x} and \code{z}.
#'   \item \code{std.error}: The standard error of the predicted value.
#'   \item \code{z.value}, \code{p.value}: The z-statistic and corresponding p-value
#'         for testing the null hypothesis that \code{predicted == h0}.
#'   \item \code{ci.lower}, \code{ci.upper}: Lower and upper bounds of the confidence
#'         (or prediction) interval.
#' }
#' An attribute \code{"variable_names"} (list of \code{x}, \code{z}, \code{y})
#' is attached for convenience.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' library(modsem)
#'
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#' # Inner model
#'   Y ~ X + Z + X:Z
#' "
#' est1 <- modsem(m1, data = oneInt)
#'
#' # Simple slopes at X in [-3, 3] and Z in [-1, 1], rescaled to the raw metric.
#' simple_slopes(x = "X", z = "Z", y = "Y", model = est1)
#'
#' # If the data or user wants unscaled values, set rescale = FALSE, etc.
#' simple_slopes(x = "X", z = "Z", y = "Y", model = est1, rescale = FALSE)
#' }
simple_slopes <- function(x,
                          z,
                          y,
                          model,
                          vals_x = -3:3,
                          vals_z = -1:1,
                          rescale = TRUE,
                          ci_width = 0.95,
                          ci_type = "confidence",
                          relative_h0 = TRUE,
                          standardized = FALSE,
                          xz = NULL,
                          ...) {
  mod_stopif(!isModsemObject(model) && !isLavaanObject(model), paste0("model must be of class ",
         "'modsem_pi', 'modsem_da', 'modsem_mplus' or 'lavaan'"))

  if (standardized) {
    parTable <- standardized_estimates(model, correction = TRUE)
  } else {
    parTable <- parameter_estimates(
      object = model,
      label.renamed.prod = TRUE
    )
  }

  parTable <- getMissingLabels(addMissingGroups(parTable))

  out <- list()
  group.label <- tryCatch(modsem_inspect(model, what = "group.label"), error = \(e) NULL)

  for (g in getGroupsParTable(parTable)) {
    label <- tryCatch(group.label[[g]], error = \(e) NULL)
    label <- if (is.null(label)) as.character(g) else label

    parTable.g <- parTable[parTable$group == g, , drop = FALSE]

    out[[label]] <- simpleSlopesGroup(
      x = x, z = z, y = y, parTable = parTable.g, model = model, vals_x = vals_x,
      vals_z = vals_z, rescale = rescale, ci_width = ci_width, ci_type = ci_type,
      relative_h0 = relative_h0, xz = xz, standardized = standardized, ...
    )
  }

  structure(out, class = "simple_slopes")
}


simpleSlopesGroup <- function(x, z, y, parTable, model, vals_x, vals_z, rescale,
                              ci_width, ci_type, relative_h0, xz,
                              standardized, ...) {
  if (is.null(xz))
    xz <- paste(x, z, sep = ":")

  checkInputsSimpleSlopes(x = x, z = z, y = y, xz = xz, parTable = parTable)

  xz <- c(xz, reverseIntTerm(xz))

  xx <- paste(x, x, sep = ":")
  xx <- c(xx, reverseIntTerm(xx))

  zz <- paste(z, z, sep = ":")
  zz <- c(zz, reverseIntTerm(zz))

  if (!inherits(model, c("modsem_da", "modsem_mplus")) &&
      !isLavaanObject(model)) {
    xz <- stringr::str_remove_all(xz, ":")
    xx <- stringr::str_remove_all(xx, ":")
    zz <- stringr::str_remove_all(zz, ":")
  }

  if (inherits(model, "lavaan")) {
    vcov <- lavaan::vcov
    coef <- lavaan::coef
    nobs <- lavaan::nobs
  }

  n     <- nobs(model)
  VCOV  <- getVcovSimpleSlopes(model, standardized = standardized)

  if (length(n) > 1) {
    # this won't work for multigroup models
    warning("simple_slopes/plot_interaction does not support multigroup models, be vary of results!")
    n <- n[1]
  }

  # Extract coefficients
  beta_x  <- parTable[parTable$lhs == y & parTable$rhs == x & parTable$op == "~", "est"]
  beta_z  <- parTable[parTable$lhs == y & parTable$rhs == z & parTable$op == "~", "est"]
  beta_xz <- parTable[parTable$lhs == y & parTable$rhs %in% xz & parTable$op == "~", "est"]
  beta_xx <- parTable[parTable$lhs == y & parTable$rhs %in% xx & parTable$op == "~", "est"]
  beta_zz <- parTable[parTable$lhs == y & parTable$rhs %in% zz & parTable$op == "~", "est"]
  beta0_y <- parTable[parTable$lhs == y & parTable$op == "~1", "est"]
  res_y   <- parTable[parTable$lhs == y & parTable$rhs == y & parTable$op == "~~", "est"]

  var_x   <- calcCovParTable(x, x, parTable)
  var_z   <- calcCovParTable(z, z, parTable)

  mod_stopif(is.na(var_x), "Variance for x not found in model")
  mod_stopif(is.na(var_z), "Variance for z not found in model")
  mod_stopif(length(beta_x) == 0, "Coefficient for x not found in model")
  mod_stopif(length(beta_z) == 0, "Coefficient for z not found in model")
  mod_warnif(length(beta_xz) == 0, "Coefficient for interaction term not found in model")

  if (!length(beta_xx)) beta_xx <- 0
  if (!length(beta_xz)) beta_xz <- 0
  if (!length(beta_zz)) beta_zz <- 0

  label_beta_x  <- parTable[parTable$lhs == y & parTable$rhs == x & parTable$op == "~", "label"]
  label_beta_z  <- parTable[parTable$lhs == y & parTable$rhs == z & parTable$op == "~", "label"]
  label_beta_xz <- parTable[parTable$lhs == y & parTable$rhs %in% xz & parTable$op == "~", "label"]
  label_beta_xx <- parTable[parTable$lhs == y & parTable$rhs %in% xx & parTable$op == "~", "label"]
  label_beta_zz <- parTable[parTable$lhs == y & parTable$rhs %in% zz & parTable$op == "~", "label"]
  label_beta0_y <- parTable[parTable$lhs == y & parTable$op == "~1", "label"]

  label_beta_x  <- ifelse(length(label_beta_x) == 0, "y~x", label_beta_x)
  label_beta_z  <- ifelse(length(label_beta_z) == 0, "y~z", label_beta_z)
  label_beta_xz <- ifelse(length(label_beta_xz) == 0, "y~xz", label_beta_xz)
  label_beta_xx <- ifelse(length(label_beta_xx) == 0, "y~xx", label_beta_xx)
  label_beta_zz <- ifelse(length(label_beta_zz) == 0, "y~zz", label_beta_zz)
  label_beta0_y <- ifelse(length(label_beta0_y) == 0, "y~1", label_beta0_y)

  labels <- c(label_beta0_y, label_beta_x, label_beta_z, label_beta_xz,
              label_beta_xx, label_beta_zz)
  VCOV   <- expandVCOV(VCOV, labels)

  mean_x <- getMean(x, parTable = parTable)
  mean_z <- getMean(z, parTable = parTable)
  mean_y <- getIntercept(y, parTable = parTable)
  h0     <- if (relative_h0) mean_y else 0

  if (rescale) {
    vals_x <- vals_x * sqrt(var_x) + mean_x
    vals_z <- vals_z * sqrt(var_z) + mean_z
  }

  alpha  <- 1 - ci_width
  ci.sig <- stats::qnorm(1 - alpha / 2) # two-tailed

  # creating margins
  df <- structure(expand.grid(x = vals_x, z = vals_z), names= c("vals_x", "vals_z"))

  df$predicted <-
    mean_y +
    beta_x * df$vals_x +
    beta_z * df$vals_z +
    beta_xx * df$vals_x ^ 2 +
    beta_zz * df$vals_z ^ 2 +
    beta_xz * df$vals_x * df$vals_z

  df$std.error <- calc_se(df, e = res_y, VCOV = VCOV, se_type = ci_type)
  df$z.value   <- (df$predicted - h0) / df$std.error # H0 = mean_y
  df$p.value   <- 2 * stats::pnorm(-abs(df$z.value))
  df$ci.upper  <- df$predicted + ci.sig * df$std.error
  df$ci.lower  <- df$predicted - ci.sig * df$std.error

  # significance test of slopes (not margins)
  k <- length(vals_z)
  slopeLabels <- c(label_beta_x, label_beta_xz)
  slopesVCOV <- VCOV[slopeLabels, slopeLabels]
  slopes <- beta_x + vals_z * beta_xz

  d_beta_x  <- rep(1, k) # d(beta_x)/d(beta_x) = 1
  d_beta_xz <- vals_z    # d(vals_z * beta_xz) / d(beta_xz) = vals_z
  jacobian <- matrix(c(d_beta_x, d_beta_xz),
                     nrow=k, ncol=2, byrow=FALSE)
  slopesVCOV <- jacobian %*% slopesVCOV %*% t(jacobian) # delta method
  std.error <- sqrt(diag(slopesVCOV))
  z.value <- slopes / std.error

  sig.slopes <- data.frame(
    param = paste0(y, "~", x),
    moderator = z,
    slope.predictor = slopes,
    value.moderator = vals_z,
    std.error = std.error,
    z.value = z.value,
    p.value = 2 * stats::pnorm(-abs(z.value)),
    ci.lower = slopes - ci.sig * std.error,
    ci.upper = slopes + ci.sig * std.error
  )

  # significance test min/max
  var_beta_xz <- VCOV[label_beta_xz, label_beta_xz]
  min_z <- min(vals_z)
  max_z <- max(vals_z)
  diff <- max_z * beta_xz - min_z * beta_xz
  grad_diff <- max_z - min_z # with respect to beta_xz
  std.error_diff <- sqrt(grad_diff^2 * var_beta_xz) # delta method
  z_diff <- diff / std.error_diff
  p_diff <- 2 * stats::pnorm(-abs(z_diff))

  sig.diff_min_max <- data.frame(
    min.z = min_z,
    max.z = max_z,
    diff = diff,
    std.error = std.error_diff,
    z.value = z_diff,
    p.value = p_diff,
    ci.lower = diff - ci.sig * std.error_diff,
    ci.upper = diff + ci.sig * std.error_diff
  )

  list(
    variable_names = c(x = x, z = z, y = y),
    margins = df,
    sig.slopes = sig.slopes,
    sig.diff_min_max = sig.diff_min_max
  )
}


calc_se <- function(df, e, VCOV, se_type = "confidence") {
  if (se_type == "prediction")
    return(sqrt(rep(e, nrow(df))))
  else if (se_type != "confidence")
    warning("se_type must be 'confidence' or 'prediction', using 'confidence'!")

  vals_x <- df$vals_x
  vals_z <- df$vals_z

  n <- nrow(df)
  i <- rep(1, n)
  X <- matrix(c(i, vals_x, vals_z, vals_x * vals_z, vals_x^2, vals_z^2), nrow=n)
  V <- calcSESimpleSlopes(X, VCOV)

  sqrt(as.vector(V))
}



printTable <- function(x) {
  if (!NROW(x)) return(NULL)

  box <- BOX_CHARS$chars

  header <- paste0(paste(x[1, ], collapse = paste0(" ", box$V_LINE, " ")), " ")
  header_vec <- unlist(stringr::str_split(header, ""))

  sep_thin_vec <- rep(box$H_LINE, nchar(header))
  sep_thin_vec[header_vec == box$V_LINE] <- box$D_CROSS
  sep_thin <- paste0(sep_thin_vec, collapse="")
  sep_thin <- paste0(box$LU_CORNER, sep_thin, box$RU_CORNER, "\n")

  sep_thick_vec <- rep(box$H_DLINE, nchar(header))
  sep_thick_vec[header_vec == box$V_LINE] <- box$F_DCROSS
  sep_thick <- paste0(sep_thick_vec, collapse="")
  sep_thick <- paste0(box$L_DCROSS, sep_thick, box$R_DCROSS, "\n")

  sep_bottom_vec <- sep_thin_vec
  sep_bottom_vec[header_vec == box$V_LINE] <- box$U_CROSS
  sep_bottom <- paste0(sep_bottom_vec, collapse="")
  sep_bottom <- paste0(box$LL_CORNER, sep_bottom, box$RL_CORNER, "\n")

  for (i in seq_len(nrow(x))) {
    str <- paste(x[i, ], collapse = paste0(" ", box$V_LINE, " "))
    str <- paste0(box$V_LINE, str, " ", box$V_LINE, "\n")

    if (i == 1) {
      cat(sep_thin, str, sep_thick, sep = "")
    } else cat(str)
  }

  cat(sep_bottom)
}


#' @export
print.simple_slopes <- function(x, digits = 2, scientific.p = FALSE, ...) {
  if (length(x) <= 1L)
    return(printSimpleSlopesGroup(x = x[[1L]], digits = digits, scientific.p = scientific.p, ...))

  box <- BOX_CHARS$chars
  for (g in names(x)) {
    printf("%s\n", strrep(box$H_LINE, getOption("width")))
    printf("  Group: %s\n", g)
    printf("%s\n", strrep(box$H_DLINE, getOption("width")))
    printSimpleSlopesGroup(x = x[[g]], digits = digits, scientific.p = scientific.p, ...)
  }

  invisible(x)
}


printSimpleSlopesGroup <- function(x, digits = 2, scientific.p = FALSE, ...) {
  fstring <- paste0("%.", digits, "f")
  digits.p <- max(3, digits)

  fformat <- \(x, justify = "right", ...) format(sprintf(fstring, x), justify = justify, ...)
  pformat <- \(x) formatPval(x, scientific = scientific.p, digits = digits.p)

  variables  <- x$variable_names
  margins    <- x$margins
  sig.slopes <- x$sig.slopes
  sig.diff_min_max <- x$sig.diff_min_max

  var_x <- variables[[1]]
  var_z <- variables[[2]]
  var_y <- variables[[3]]

  # Difference test ----------------------------------------------------------
  header <- c("Difference", "Std.Error",
              "z.value", "p.value", "Conf.Interval")
  ci.lower <- fformat(sig.diff_min_max$ci.lower)
  ci.upper <- fformat(sig.diff_min_max$ci.upper)

  X <- data.frame(
    diff      = fformat(sig.diff_min_max$diff),
    std.error = fformat(sig.diff_min_max$std.error),
    z.value   = fformat(sig.diff_min_max$z.value),
    p.value   = pformat(sig.diff_min_max$p.value),
    ci        = paste0("[", ci.lower, ", ", ci.upper, "]")
  )

  X1 <- matrix(header, nrow = 1)
  X2 <- padCharMatrix(as.matrix(X), n=1) # pad atleast one space to the left
  X <- apply(rbind(X1, X2), MARGIN = 2, format, digits = digits, justify = "right")

  cat(sprintf("\nDifference test of %s~%s|%s at %s = %s and %s:\n",
              var_y, var_x, var_z, var_z, fformat(sig.diff_min_max$min.z),
              fformat(sig.diff_min_max$max.z)))
  printTable(X)
  cat("\n")

  # Slope test -----------------------------------------------------------------
  header <- c(sig.slopes$param[1], sig.slopes$moderator[1], "Std.Error",
              "z.value", "p.value", "Conf.Interval")
  ci.lower <- fformat(sig.slopes$ci.lower)
  ci.upper <- fformat(sig.slopes$ci.upper)

  X <- data.frame(
    slope     = fformat(sig.slopes$slope.predictor),
    val_z     = fformat(sig.slopes$value.moderator),
    std.error = fformat(sig.slopes$std.error),
    z.value   = fformat(sig.slopes$z.value),
    p.value   = pformat(sig.slopes$p.value),
    ci        = paste0("[", ci.lower, ", ", ci.upper, "]")
  )

  X1 <- matrix(header, nrow = 1)
  X2 <- padCharMatrix(as.matrix(X), n=1) # pad atleast one space to the left
  X <- apply(rbind(X1, X2), MARGIN = 2, format, digits = digits, justify = "right")

  cat(sprintf("Effect of %s~%s|%s given values of %s:\n", var_y, var_x, var_z, var_z))
  printTable(X)
  cat("\n")

  # Margins --------------------------------------------------------------------
  predictors <- variables[1:2]
  outcome    <- variables[3]
  header     <- c(predictors[1], sprintf("Predicted %s", outcome), "Std.Error",
                  "z.value", "p.value", "Conf.Interval")
  ci.lower   <- fformat(margins$ci.lower)
  ci.upper   <- fformat(margins$ci.upper)
  cat_z      <- as.factor(fformat(round(margins$vals_z, digits)))

  X <- data.frame(vals_x    = fformat(margins$vals_x),
                  predicted = fformat(margins$predicted),
                  std.error = fformat(margins$std.error),
                  z.value   = fformat(margins$z.value),
                  p.value   = pformat(margins$p.value),
                  ci        = paste0("[", ci.lower, ", ", ci.upper, "]"))

  X1 <- matrix(header, nrow = 1)
  X2 <- padCharMatrix(as.matrix(X), n=1) # pad atleast one space to the left

  X <- apply(rbind(X1, X2), MARGIN = 2, format, digits = digits, justify = "right")

  for (z_i in unique(cat_z)) {
    Z1 <- X[1, ]
    Z2 <- X[2:nrow(X),][cat_z == z_i,]
    Z  <- rbind(Z1, Z2)

    printf("\nPredicted %s, given %s = %s:\n", outcome, predictors[2], z_i)
    printTable(Z)
    cat("\n")
  }

  invisible(x)
}


#' @export
as.data.frame.simple_slopes <- function(x, ...) {
  if (length(x) <= 1L)
    return(x[[1L]]$margins)

  purrr::list_rbind(
    lapply(X = names(x), FUN = \(g)
           cbind(x[[g]]$margins, data.frame(group = g)))
  )
}


checkInputsSimpleSlopes <- function(..., parTable) {
  varmask   <- parTable$op %in% c("=~", "~1", "~~", "~")
  variables <- c(parTable[varmask, "lhs"], parTable[varmask, "rhs"])
  variables <- unique(variables[variables != ""])

  checkLength <- \(x, nm) mod_stopif(length(x) != 1L, paste0(nm, " must be of length 1!"))
  checkExists <- \(x) mod_stopif(!x %in% variables && !grepl(":", x),
                             paste0("Unrecognized variable: ", x, "!"))

  params <- list(...)
  for (nm in names(params)) {
    checkLength(x = params[[nm]], nm = nm)
    checkExists(x = params[[nm]])
  }
}


getVcovSimpleSlopes <- function(model, standardized = FALSE) {
  if (inherits(model, "modsem_da")) {
    return(modsem_inspect(model, what = "vcov.all", standardized = standardized))

  } else if (inherits(model, "modsem_pi")) {
    what <- if (standardized) "vcov.std.all" else "vcov"
    return(modsem_inspect(model, what = what))

  } else if (inherits(model, "lavaan")) {
    what <- if (standardized) "vcov.std.all" else "vcov"
    return(lavaan::lavInspect(model, what = what))

  } else {
    mod_msg_warn(paste0("standardized vcov is not available for model of class: ",
             class(model)[[1L]]))
    return(vcov(model))
  }
}

#' compare model fit for \code{modsem} models
#'
#' @param est_h0 object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}} representing the
#' null hypothesis model (without interaction terms).
#'
#' @param est_h1 object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}} representing the
#' alternative hypothesis model (with interaction terms).
#'
#' @param ... additional arguments passed to the underlying comparison function. E.g.,
#' for \code{modsem_pi} models, this can be used to pass arguments to \code{lavaan::lavTestLRT}.
#' currently only used for \code{modsem_pi} models.
#'
#' @description Compare the fit of two models using the likelihood ratio test (LRT).
#' \code{est_h0} is the null hypothesis model, and \code{est_h1} the alternative hypothesis model.
#' Importantly, the function assumes that \code{est_h0} does not have more free parameters
#' (i.e., degrees of freedom) than \code{est_h1} (the alternative hypothesis model).
#' @rdname compare_fit
#' @export
#' @examples
#' \dontrun{
#' m1 <- "
#'  # Outer Model
#'  X =~ x1 + x2 + x3
#'  Y =~ y1 + y2 + y3
#'  Z =~ z1 + z2 + z3
#'
#'  # Inner model
#'  Y ~ X + Z + X:Z
#' "
#'
#' # LMS approach
#' est_h1 <- modsem(m1, oneInt, "lms")
#' est_h0 <- estimate_h0(est_h1, calc.se=FALSE) # std.errors are not needed
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#'
#' # Double centering approach
#' est_h1 <- modsem(m1, oneInt, method = "dblcent")
#' est_h0 <- estimate_h0(est_h1, oneInt)
#'
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#'
#' # Constrained approach
#' est_h1 <- modsem(m1, oneInt, method = "ca")
#' est_h0 <- estimate_h0(est_h1, oneInt)
#'
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#' }
#' @export
compare_fit <- function(est_h1, est_h0, ...) {
  UseMethod("compare_fit", est_h1)
}


#' @export
compare_fit.modsem_da <- function(est_h1, est_h0, ...) {
  mod_stopif(!inherits(est_h1, "modsem_da") || !inherits(est_h0, "modsem_da"),
         "Both est_h1 and est_h0 must be of class 'modsem_da'.")

  if (is.null(est_h0) || is.null(est_h1)) return(NULL) # should never trigger

  df <- length(coef(est_h1, type = "free")) - length(coef(est_h0, type = "free"))
  D <- -2 * (est_h0$logLik - est_h1$logLik)
  p <- stats::pchisq(D, df = df, lower.tail = FALSE, log.p = FALSE)
  list(D = D, df = df, p = p, diff.loglik = est_h1$logLik - est_h0$logLik)
}


#' @export
compare_fit.modsem_pi <- function(est_h1, est_h0, comparison = lavaan::lavTestLRT, ...) {
  mod_stopif(!inherits(est_h1, "modsem_pi") || !inherits(est_h0, "modsem_pi"),
         "Both est_h1 and est_h0 must be of class 'modsem_pi'.")

  est_h0 <- extract_lavaan(est_h0)
  est_h1 <- extract_lavaan(est_h1)
  comparison(est_h0, est_h1, ...)
}

devtools::load_all()

tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  # Covariances
  ATT ~~ SN + PBC
  PBC ~~ SN
  # Causal Relationsships
  INT ~ ATT + SN + PBC
  BEH ~ b * INT + PBC
  INT ~ a * PBC:SN
  BEH ~ c * PBC:SN

  IndirEffect := a * b
  TotalEffect := a * b + c
"

est_qml <- modsem(tpb, TPB, method = "qml")
standardized_estimates(est_qml)


est_lms <- modsem(tpb, TPB, method = "lms")
standardized_estimates(est_lms)

getParTableResCov <- function(relDf,
                              method,
                              pt = NULL,
                              explicit.zero = FALSE,
                              include.single.inds = FALSE,
                              setToZero = FALSE) {
  simple <- \()
    getParTableResCov.simple(relDf,
                             explicit.zero = explicit.zero,
                             include.single.inds = include.single.inds)
  switch(method,
    simple = simple(),
    simple.no.warn = suppressWarnings(simple()),
    ca = getParTableResCov.ca(relDf, pt = pt),
    equality = getParTableResCov.equality(relDf, setToZero = setToZero)
  )
}


getParTableResCov.simple <- function(relDf, explicit.zero = FALSE, include.single.inds = FALSE) {
  EMPTY <- data.frame(lhs = NULL, op = NULL, rhs = NULL, mod = NULL)
  attr(EMPTY, "OK") <- TRUE

  if (length(relDf) <= 1)
    return(EMPTY)

  OK <- TRUE

  if (include.single.inds) {
    allInds <- unique(unlist(relDf))
    relDf <- as.list(relDf)
    relDf <- c(relDf, as.list(stats::setNames(allInds, nm = allInds)))
  }

  prodNames <- sort(names(relDf))
  uniqueCombinations <- getUniqueCombos(prodNames)
  # Now we want to specify the covariance based on shared inds
  isShared <- vector("logical", length = nrow(uniqueCombinations))

  for (i in seq_len(nrow(uniqueCombinations))) {
    indsProd1 <- unlist(relDf[uniqueCombinations[i, "V1"]])
    indsProd2 <- unlist(relDf[uniqueCombinations[i, "V2"]])
    # Compare the Inds in prod1 and prod2, and convert to integer
    sharedValues <- as.integer(indsProd1 %in% indsProd2)

    # Sum the values
    numberShared <- sum(sharedValues)

    # if there is a difference in number of elems in `len.diff.ignore` they should be ignored...
    # See the simulation results in `tests/testthat/test_three_way.R`
    #> cov(x, z, w, xz, xw, zw, xm, zm, wm, xzw, xzm, xwm, zwm, xzwm)
    #>         x    z    w   xz   xw   zw   xm   zm   wm   xzw  xzm  xwm  zwm  xzwm
    #> x    1.20 0.70 0.80 0.00 0.00 0.00 0.00 0.00 0.00  1.84 0.64 1.04 0.78  0.00
    #> z    0.70 1.80 0.60 0.00 0.00 0.00 0.00 0.00 0.00  2.28 0.78 0.78 1.44  0.00
    #> w    0.80 0.60 1.40 0.00 0.00 0.00 0.00 0.00 0.00  1.94 0.78 1.24 1.14  0.00
    #> xz   0.00 0.00 0.00 2.65 1.28 1.86 0.50 0.57 0.36  0.00 0.00 0.00 0.00  3.36
    #> xw   0.00 0.00 0.00 1.28 2.32 1.46 0.88 0.54 0.76  0.00 0.00 0.00 0.00  3.25
    #> zw   0.00 0.00 0.00 1.86 1.46 2.88 0.66 1.26 0.78  0.00 0.00 0.00 0.00  4.09
    #> xm   0.00 0.00 0.00 0.50 0.88 0.66 2.44 1.46 1.72  0.00 0.00 0.00 0.00  4.54
    #> zm   0.00 0.00 0.00 0.57 0.54 1.26 1.46 3.69 1.38  0.00 0.00 0.00 0.00  5.56
    #> wm   0.00 0.00 0.00 0.36 0.76 0.78 1.72 1.38 3.16  0.00 0.00 0.00 0.00  4.96
    #> xzw  1.84 2.28 1.94 0.00 0.00 0.00 0.00 0.00 0.00 10.25 3.91 3.88 4.56  0.01
    #> xzm  0.64 0.78 0.78 0.00 0.00 0.00 0.00 0.00 0.00  3.91 6.98 4.69 5.79  0.01
    #> xwm  1.04 0.78 1.24 0.00 0.00 0.00 0.00 0.00 0.00  3.88 4.69 7.67 5.42  0.00
    #> zwm  0.78 1.44 1.14 0.00 0.00 0.00 0.00 0.00 0.00  4.56 5.79 5.42 8.90  0.00
    #> xzwm 0.00 0.00 0.00 3.36 3.25 4.09 4.54 5.56 4.96  0.01 0.01 0.00 0.00 28.76
    ignore <- length(indsProd1) %% 2 != length(indsProd2) %% 2

    if      (numberShared >= 1 && !ignore) isShared[[i]] <- TRUE
    else if (numberShared == 0 && !ignore) isShared[[i]] <- FALSE
  }

  if (all(isShared)) {
    mod_msg_warn(paste0("All residual covariances between product indicators were freed!\n",
             "The model will likely not be identifiable! Please try passing:\n",
             "  `res.cov.method = \"none\"` or `res.cov.method = \"equality\"`"))
    OK <- FALSE
  }

  prodsSharingInds    <- uniqueCombinations[isShared, c("V1", "V2")]
  prodsNotSharingInds <- uniqueCombinations[!isShared, c("V1", "V2")]

  # Syntax for oblique covariances
  if (nrow(prodsSharingInds) > 0) {
    syntaxOblique <- apply(prodsSharingInds,
                           MARGIN = 1,
                           FUN = createParTableRow,
                           op = "~~") |>
      purrr::list_rbind()

  } else syntaxOblique <- NULL

  if (explicit.zero && nrow(prodsNotSharingInds) > 0) {
    syntaxOrthogonal <- apply(prodsNotSharingInds,
                              MARGIN = 1,
                              FUN = createParTableRow,
                              op = "~~",
                              mod = "0") |>
      purrr::list_rbind()

  } else syntaxOrthogonal <- NULL

  out <- rbind(syntaxOrthogonal, syntaxOblique)

  if (is.null(out))
    return(EMPTY)

  attr(out, "OK") <- OK
  out
}



# Rescovs with same indicator constrained to equality
getParTableResCov.equality <- function(relDf, setToZero = FALSE) {
  EMPTY <- data.frame(lhs = NULL, op = NULL, rhs = NULL, mod = NULL)
  attr(EMPTY, "OK") <- TRUE

  if (length(relDf) <= 1)
    return(EMPTY)

  OK <- TRUE

  prodNames <- sort(names(relDf))
  sharedMatrix <- matrix("", nrow = length(prodNames), ncol = length(prodNames),
                         dimnames = list(prodNames, prodNames))
  # Now we want to specify the covariance based on shared inds
  for (i in prodNames) {
    for (j in prodNames) {
      sharedIndicators <- relDf[[i]][relDf[[i]] %in% relDf[[j]]]
      sharedMatrix[i, j] <- stringr::str_c(sharedIndicators,
                                           collapse = "_")
    }
  }
  labelMatrix <- sharedMatrix
  labelMatrix <- ifelse(labelMatrix == "", "0", paste0("share_", labelMatrix))
  labelMatrix[upper.tri(labelMatrix, diag = TRUE)] <- ""

  uniqueCombos <- getUniqueCombos(prodNames)
  uniqueCombos[["labels"]] <- vector("character", length = nrow(uniqueCombos))
  for (i in seq_len(nrow(uniqueCombos))) {
    uniqueCombos[["labels"]][[i]] <- labelMatrix[uniqueCombos[i, "V2"],
                                                 uniqueCombos[i, "V1"]]
  }

  parTable <- apply(uniqueCombos, MARGIN = 1,
                    FUN = function(x)
                      createParTableRow(x[c("V1", "V2")], op = "~~", mod = x[["labels"]])
                    ) |>
    purrr::list_rbind()
  if (!setToZero) parTable <- parTable[parTable$mod != 0, ]

  attr(parTable, "OK") <- OK
  parTable
}



# Constrained Approach ---------------------------------------------------------
getParTableResCov.ca <- function(relDf, pt) {
  if (nrow(relDf) > 2) {
    mod_msg_stop(paste0("Constrained approach for constraining residual covariances should ",
         "not be used with latent products with more than two components"))
  }
  if (ncol(relDf) <= 1) {
    return(NULL)
  }
  prodNames <- colnames(relDf)
  labelMatrix <- matrix("", nrow = length(prodNames), ncol = length(prodNames),
                        dimnames = list(prodNames, prodNames))
  # Now we want to specify the covariance based on shared inds
  parTable <- NULL
  for (i in 2:nrow(labelMatrix)) {
    rhs <- rownames(labelMatrix)[[i]]

    for (j in 1:(i-1)) {
      lhs <- colnames(relDf)[[j]]
      sharedIndicators <- relDf[[i]] %in% relDf[[j]]

      if (sum(sharedIndicators) >= 1) {
        labelMatrix[i, j] <- createLabelCov(rhs, lhs)

      } else if (sum(sharedIndicators) < 1) {
        labelMatrix[i, j] <- "0"
      }

      parTable <- rbind(parTable,
                        createParTableRow(c(rhs, lhs),
                                          op = "~~",
                                          mod = labelMatrix[i,j]))
    }

  }

  #apply eq constraints to those which are not set to zero
  if (length(parTable$mod[parTable$mod != "0"]) > 0) {
    eqConstraints <- apply(parTable[parTable$mod != "0", c("lhs", "rhs")],
                           MARGIN = 1,
                           FUN = function(vars, relDf)
                             getFormulaResCovProdInd(vars[["lhs"]],
                                                     vars[["rhs"]],
                                                     relDf, pt),
                           relDf = relDf) |>
      purrr::list_rbind()
  } else {
    eqConstraints <- NULL
  }
  rbind(parTable, eqConstraints)
}



getFormulaResCovProdInd <- function(indProd1, indProd2, relDf, pt) {
  if (is.null(indProd1) || is.null(indProd2)) {
    return(NULL)
  }
  cols <- c(indProd1, indProd2)
  rowShared <- relDf[relDf[[indProd1]] %in_paired% relDf[[indProd2]], cols]
  forceRowNames(rowShared) <- rownames(relDf)[relDf[[indProd1]] %in_paired% relDf[[indProd2]]]
  rowNotShared <- relDf[!(relDf[[indProd1]] %in_paired% relDf[[indProd2]]), cols]
  forceRowNames(rowNotShared) <- rownames(relDf)[!relDf[[indProd1]] %in_paired% relDf[[indProd2]]]

  latentNotShared <- rownames(rowNotShared)
  indShared <- rowShared[1, 1]
  indsNotShared <- unlist(rowNotShared[1, 1:2])
  lambdaShared <- createLabelLambda(indsNotShared, latentNotShared)
  varLatentNotShared <- trace_path(pt, latentNotShared, latentNotShared)
  varIndShared <- createLabelVar(indShared)

  rhs <- paste(lambdaShared[[1]], lambdaShared[[2]],
               varLatentNotShared, varIndShared, sep = " * ")
  lhs <- createLabelCov(indProd1, indProd2)
  createParTableRow(c(lhs, rhs), op = "==")
}

estQml <- function(model,
                   convergence = 1e-2,
                   verbose = FALSE,
                   max.iter = 500,
                   calc.se = TRUE,
                   FIM = "observed",
                   OFIM.hessian = FALSE,
                   EFIM.S = 3e4,
                   EFIM.parametric = TRUE,
                   robust.se = FALSE,
                   epsilon = 1e-6,
                   optimizer = "nlminb",
                   R.max = 1e6,
                   cr1s = TRUE,
                   ...) {
  data <- lapply(model$models, FUN = \(sub) sub$data)

  tryCatch({

    startTheta <- model$theta
    final <- mstepQml(model = model, theta = startTheta, max.iter = max.iter,
                      convergence = convergence, epsilon = epsilon,
                      verbose = verbose, optimizer = optimizer, ...)

    finalizeModelEstimatesDA(
      model             = model,
      theta             = final$par,
      method            = "qml",
      data              = data,
      logLik            = -final$objective,
      iterations        = final$iterations,
      converged         = (final$convergence == 0L) &&
                          (final$iterations <= max.iter),
      optimizer         = optimizer,
      calc.se           = calc.se,
      FIM               = FIM,
      OFIM.hessian      = OFIM.hessian,
      EFIM.S            = EFIM.S,
      EFIM.parametric   = EFIM.parametric,
      robust.se         = robust.se,
      epsilon           = epsilon,
      cr1s              = cr1s,
      R.max             = R.max,
      verbose           = verbose,
      includeStartModel = TRUE,
      startModel        = model
    )

  }, error = function(e) {
    mod_msg_warn(paste0(
      "Model estimation failed, returning starting values!\n",
      "Message: ", conditionMessage(e)
    ), newline. = verbose)

    finalizeModelEstimatesDA(
      model             = model,
      theta             = model$theta,   # start values
      method            = "qml",
      data              = data,
      logLik            = NA_real_,      # best-effort; unknown here
      iterations        = 0L,
      converged         = FALSE,
      optimizer         = optimizer,
      calc.se           = FALSE,         # <- ensure no SEs
      FIM               = FIM,
      OFIM.hessian      = OFIM.hessian,
      EFIM.S            = EFIM.S,
      EFIM.parametric   = EFIM.parametric,
      robust.se         = robust.se,
      epsilon           = epsilon,
      cr1s              = cr1s,
      R.max             = R.max,
      verbose           = verbose,
      includeStartModel = TRUE,
      startModel        = model
    )
  })
}

#' @export
#' @describeIn parameter_estimates Get parameter estimates of a \code{lavaan} object
parameter_estimates.lavaan <- function(object,
                                       colon.pi = NULL, # capture
                                       high.order.as.measr = NULL, # capture
                                       rm.tmp.ov = NULL, # capture argument
                                       label.renamed.prod = NULL, # capture argument
                                       is.public = NULL, # capture argument
                                       ...) {
  lavaan::parameterEstimates(object, ...)
}


isLavaanObject <- function(x) {
  inherits(x, "lavaan")
}


#' @export
#' @describeIn modsem_inspect Inspect a \code{lavaan} object
modsem_inspect.lavaan <- function(object, what = "free", ...) {
  lavaan::lavInspect(object, what = what, ...)
}


#' @describeIn centered_estimates Method for \code{lavaan} objects
#'
#' @param monte.carlo Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
#' standard errors; if \code{FALSE}, use the delta method (default).
#' @param mc.reps Number of Monte Carlo repetitions. Default is 10000.
#' @param tolerance.zero Threshold below which standard errors are set to \code{NA}.
#'
#' @export
centered_estimates.lavaan <- function(object,
                                      monte.carlo = FALSE,
                                      mc.reps = 10000,
                                      tolerance.zero = 1e-10,
                                      ...) {

  centering <- function(object, grouping = NULL) {

    solution <- centeredSolutionCOEFS(
      object      = object,
      monte.carlo = monte.carlo,
      mc.reps     = mc.reps,
      grouping    = grouping,
      ...
    )

    solution$parTable
  }

  transformEstimatesLavaan(
    object         = object,
    monte.carlo    = monte.carlo,
    mc.reps        = mc.reps,
    tolerance.zero = tolerance.zero,
    transformation = centering
  )
}


#' @describeIn standardized_estimates Method for \code{lavaan} objects
#'
#' @param monte.carlo Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
#' standard errors; if \code{FALSE}, use the delta method (default).
#' @param mc.reps Number of Monte Carlo repetitions. Default is 10000.
#' @param tolerance.zero Threshold below which standard errors are set to \code{NA}.
#'
#' @export
standardized_estimates.lavaan <- function(object,
                                          monte.carlo = FALSE,
                                          mc.reps = 10000,
                                          tolerance.zero = 1e-10,
                                          ...) {

  standardization <- function(object, grouping = NULL) {

    solution <- standardizedSolutionCOEFS(
      object      = object,
      monte.carlo = monte.carlo,
      mc.reps     = mc.reps,
      grouping    = grouping,
      ...
    )

    solution$parTable
  }

  transformEstimatesLavaan(
    object         = object,
    monte.carlo    = monte.carlo,
    mc.reps        = mc.reps,
    tolerance.zero = tolerance.zero,
    transformation = standardization
  )
}


transformEstimatesLavaan <- function(object,
                                     monte.carlo = FALSE,
                                     mc.reps = 10000,
                                     tolerance.zero = 1e-10,
                                     transformation) {
  parTable <- parameter_estimates(object)

  hiorder <- isHigherOrderParTable(parTable)
  cluster <- isClustered(object)

  mod_stopif(hiorder, "Transformations of higher-order models are not supported!")
  mod_stopif(cluster, "Transformations of clustered (multilevel) models is not supported!")

  applyTransformationByGrouping(
    parTable = parTable,
    FUN      = transformation,
    object   = object
  )
}

preCheckModel <- function(model, covModel = NULL, method = "lms", missing = "complete") {
  hasCovModel <- !is.null(model$covModel$matrices)

  checkCovModelVariables(covModel = covModel, modelXis = model$info$xis)

  checkZeroVariances(model = model, method = method)

  checkNodesLms(parTableMain = model$parTable,
                parTableCov  = covModel$parTable,
                nodes = model$quad$m, method = method,
                adaptive = model$quad$adaptive)

  checkOVsInStructuralModel(parTableMain = model$parTable,
                            parTableCov  = covModel$parTable)

  checkOverlappingIndicators(allIndsXis  = model$info$allIndsXis,
                             allIndsEtas = model$info$allIndsEtas,
                             method      = method)

  checkAConstraints(model = model, covModel = covModel, method = method)

  checkCovEtaXi(parTable = model$parTable, canBeCausedByCovModel = hasCovModel,
                method = method)
  checkCovEtaXi(parTable = model$covModel$parTable, canBeCausedByCovModel = FALSE,
                method = method)

  checkOmegaEtaXi(model = model, method = method)

  checkMissingMethod(method = method, missing = missing)
}


checkAConstraints <- function(model, covModel, method = "lms") {
  if (method != "lms" || !is.null(covModel$matrices)) return(NULL)

  An <- model$matrices$A
  Al <- model$labelMatrices$A

  isOKALabel <- all(Al == "")
  isOKANumeric <- all(is.na(An[lower.tri(An, diag = TRUE)])) ||
                 !any(is.na(An[lower.tri(An)]))

  mod_warnif_immediate(!isOKALabel || !isOKANumeric, paste0(
    "Variances and covariances of exogenous variables aren't truely ",
     "free parameters in the LMS approach.\n\n",
     "Using them in model constraints will likely not work as intended!\n\n",
     'To fix this you can pass an empty model to `cov.syntax`, for example: \n',
     '    `modsem(my_model, data = my_data, method = "lms", cov.syntax = "")`'
  ))
}


checkCovModelVariables <- function(covModel, modelXis, method = "lms") {
  if (is.null(covModel$matrices)) return(NULL) # nothing to check
  covModelEtas <- covModel$info$etas
  covModelXis  <- covModel$info$xis

  mod_stopif(!all(c(covModelXis, covModelEtas) %in% modelXis),
         paste0("All latent variables in the cov-model must be an ",
         "exogenous variable in the main model"))
  mod_stopif(!all(modelXis %in% c(covModelXis, covModelEtas)),
         paste0("All exogenous variables in main model must be ",
         "part of the cov-model"))
}


checkZeroVariances <- function(model, method = "lms") {
  if (method != "lms") return(NULL)

  nonLinearXis <- model$info$nonLinearXis
  inds <- model$info$indsXis[nonLinearXis]

  thetaDelta    <- model$matrices$thetaDelta
  thetaDeltaLab <- model$labelMatrices$thetaDelta

  message <- paste(
      "The variance of a moderating variable of integration",
      "has an indicator with zero residual variance!",
      "\nThis will likely not work with the LMS approach, see:",
      "\n   `vignette('observed_lms_qml', 'modsem')` for more information.",
      "\n\nThe following indicators have zero residual variance:"
  )

  m1 <- \(i) sprintf("  -> %s", i)

  width <- options()$width
  width <- if (is.null(width) || width == Inf) 30 else (width - 11) / 2

  error <- FALSE
  for (lv in nonLinearXis) for (ind in inds[[lv]]) {
    est <- thetaDelta[ind, ind]
    lab <- thetaDeltaLab[ind, ind]

    if (!is.na(est) && est == 0 && lab == "") {
      error <- TRUE
      message <- paste(message, m1(ind), sep = "\n")
    }
  }

  if (error) mod_msg_stop(message)
}


checkNodesLms <- function(parTableMain,
                          parTableCov,
                          nodes,
                          method = "lms",
                          adaptive = FALSE,
                          minNodesXiXi   = if (!adaptive) 16 else 15,
                          minNodesXiEta  = if (!adaptive) 32 else 24,
                          minNodesEtaEta = if (!adaptive) 48 else 32) {
  if (method != "lms") return(NULL)

  parTable <- rbind(parTableMain, parTableCov)

  etas     <- getEtas(parTable, isLV = TRUE)
  xis      <- getXis(parTable, etas = etas, isLV = TRUE)
  varsInts <- getVarsInts(getIntTermRows(parTable))

  nodesXiXi_ok   <- TRUE
  nodesXiEta_ok  <- TRUE
  nodesEtaEta_ok <- TRUE

  lapply(varsInts, FUN = function(x) {
    if      (all(x %in% xis))  nodesXiXi_ok   <<- nodes >= minNodesXiXi
    else if (all(x %in% etas)) nodesEtaEta_ok <<- nodes >= minNodesEtaEta
    else if (any(x %in% etas)) nodesXiEta_ok  <<- nodes >= minNodesXiEta
    else mod_msg_warn("Unable to classify latent variables in interaction terms")
  })

  mod_warnif_immediate(!nodesXiXi_ok,
    paste0("It is recommended that you have at least ",
           minNodesXiXi,  " nodes for interaction effects between ",
           "exogenous variables in the lms approach 'nodes = ", nodes, "'")
  )

  mod_warnif_immediate(!nodesXiEta_ok,
    paste0("It is recommended that you have at least ",
           minNodesXiEta, " nodes for interaction effects between exogenous ",
           "and endogenous variables in the lms approach 'nodes = ", nodes, "'")
  )

  mod_warnif_immediate(!nodesEtaEta_ok,
    paste0("It is recommended that you have at least ",
           minNodesEtaEta, " nodes for interaction effects between endogenous ",
           "variables in the lms approach 'nodes = ", nodes, "'")
  )
}


checkCovEtaXi <- function(parTable, canBeCausedByCovModel = FALSE,
                          method = "lms") {
  if (!NROW(parTable))
    return(NULL)

  if (method == "lms")
    return(NULL)

  etas <- getEtas(parTable, checkAny = FALSE, isLV = FALSE)
  xis  <- getXis(parTable, checkAny = FALSE, isLV = FALSE, etas = etas)

  isCov <- parTable$op == "~~"
  lxi   <- parTable$lhs %in% xis
  rxi   <- parTable$rhs %in% xis
  leta  <- parTable$lhs %in% etas
  reta  <- parTable$rhs %in% etas

  problematicRows <- isCov & ((lxi & reta) | (leta & rxi))

  if (canBeCausedByCovModel) {
    msgcov <- paste0(
      "\nThis may be because the model has been split into linear and non-linear parts!\n",
      "You can try passing `auto.split.syntax=FALSE` and `cov.syntax=NULL`..."
    )
  } else msgcov <- ""

  prob <- parTable[problematicRows, ]
  par <- paste0(prob$lhs, prob$op, prob$rhs)

  msg <- paste0(
    "Covariances between exogenous and endogenous variables are not available,\n",
    "and will be ignored! Consider trying `method=\"lms\"` instead. The problematic covariances are:\n",
    paste0(par, collapse = ", "), msgcov
  )

  mod_warnif_immediate(any(problematicRows), msg)
}


checkOmegaEtaXi <- function(model, method = "qml", zero.tol = 1e-10) {
  if (method != "qml") return(NULL)

  omegaEtaXi <- model$matrices$omegaEtaXi
  problematic <- any(is.na(omegaEtaXi) | abs(omegaEtaXi) > zero.tol)

  mod_warnif_immediate(problematic,
    paste0("Interactions between exogenous and enodgenous variables in the QML\n",
           "approach can be biased in some cases...\n",
           "You can try passing `auto.split.syntax=TRUE` and `cov.syntax=NULL`...")
  )
}


checkOVsInStructuralModel <- function(parTableMain, parTableCov) {
  parTable <- rbind(parTableMain, parTableCov)
  xisLVs   <- getXis(parTable, isLV = TRUE)
  xisAll   <- getXis(parTable, isLV = FALSE)

  mod_stopif(length(xisAll) != length(xisLVs) || !all(xisLVs %in% xisAll),
         paste0("Observed variables are not allowed in the structural model in LMS/QML directly. ",
         "Please redefine them as latent.\nSee:\n",
         "  vignette(\"observed_lms_qml\", \"modsem\")"))
}


checkMissingMethod <- function(method, missing) {
  method  <- tolower(method)
  missing <- tolower(missing)

  mod_stopif(method == "qml" && missing == "fiml",
         "Using FIML with QML is not available (yet), use LMS instead!")
}


checkOverlappingIndicators <- function(allIndsXis, allIndsEtas, method = "lms") {
  mod_stopif(any(allIndsXis %in% allIndsEtas) && method != "lms",
         paste0("The same indicator cannot be used for both an\n",
         "exogenous and endogenous variable, in the QML approach!\n",
         "Consider using the LMS approach instead.\n",
         "Overlapping indicators: ",
         paste(allIndsXis[allIndsXis %in% allIndsEtas], collapse = ", ")))
}


checkParTableDA <- function(parTable, method = "lms") {
  mod_stopif(isHigherOrderParTable(parTable) && method == "qml",
         paste0("Higher-order latent variables are not supported with `method=\"qml\"`.\n",
         "Try using `method=\"lms\"` or `rcs=TRUE`"))
}


checkResCovX_Y <- function(parTable, allIndsXis, allIndsEtas, method = "lms") {
  if (method == "lms") return(NULL)

  cond1 <- parTable$op == "~~"
  cond2 <- parTable$lhs %in% allIndsXis & parTable$rhs %in% allIndsEtas
  cond3 <- parTable$lhs %in% allIndsEtas & parTable$rhs %in% allIndsXis

  mod_stopif(any(cond1 & (cond2 | cond3)) && method == "qml",
         paste0("Residual covariances between indicators of endogenous lvs, and \n",
         "indicators of exogenous lvs are not allowed with `method=\"qml\"`.\n",
         "Try using `method=\"lms\"` instead!"))
}


checkVarsIntsDA <- function(varsInts, lVs) {
  for (xz in varsInts) {
    mod_stopif(!all(xz %in% lVs), paste0("Element in product term is not a latent variable: `",
           xz[!xz %in% lVs][[1]], "`!\n",
           "If it is an observed variable, please redefine it as a latent variable.\n",
           "See:\n  vignette(\"observed_lms_qml\", \"modsem\")"))
  }
}


postCheckModel <- function(model) {
  parTable <- model$parTable

  for (g in getGroupsParTable(parTable)) {
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]

    checkParTableEstimates(parTable.g)
    checkVariances(model$expected.matrices[[g]])
    checkCovMatrices(model$expected.matrices[[g]])
  }

  checkVCOV(vcov(model, type = "free"), calc.se = model$args$calc.se)
}


checkParTableEstimates <- function(parTable) {
  estimates <- parTable$est

  anyNonFinite <- any(is.na(estimates) | is.nan(estimates) | is.infinite(estimates))
  mod_warnif(anyNonFinite, "Some parameters could not be computed (NA, NaN or Inf)!")
}


checkVariances <- function(expected.matrices, rel.diff.tol = 1000) {
  variances.lv <- diag(expected.matrices$sigma.lv)
  variances.ov <- diag(expected.matrices$sigma.ov)
  residuals.lv <- expected.matrices$res.lv
  residuals.ov <- expected.matrices$res.ov

  check <- function(variances, type, rel.check = TRUE) {
    minVar <- min(variances, na.rm = TRUE) # should never be any NA, but just in case...
    maxVar <- max(variances, na.rm = TRUE)
    relDiff <- maxVar / minVar

    anyNeg <- any(variances < 0)
    mod_warnif(anyNeg, sprintf("Some estimated %s variances are negative!", type))

    mod_warnif(
      relDiff > rel.diff.tol && rel.check,
      sprintf("Some estimated %s variances are (at least) a factor %i times larger than others",
              type, rel.diff.tol)
    )
  }

  check(variances.lv, type = "lv")
  check(variances.ov, type = "ov")
  check(residuals.lv, type = "residual lv", rel.check = FALSE)
  check(residuals.ov, type = "residual ov", rel.check = FALSE)
}


checkCovMatrices <- function(expected.matrices) {
  cov.lv <- expected.matrices$sigma.lv
  cov.ov <- expected.matrices$sigma.ov

  check <- function(cov, type) {
    ok <- isPositiveDefinite(cov)
    type.long <- switch(type, lv = "latent variables", ov = "observed variables")

    mod_warnif(
      !ok,
      paste0(sprintf("Covariance matrix of %s is not positive definite!\n", type.long),
      sprintf("Use `modsem_inspect(fit, \"cov.%s\")` to investigate.", type))
    )
  }

  check(cov.lv, type = "lv")
  check(cov.ov, type = "ov")
}


checkVCOV <- function(vcov, calc.se = TRUE, tol.eigen = .Machine$double.eps ^ (3/4)) {
  if (!calc.se || is.null(vcov)) return(NULL) # checks not relevant

  eigenvalues <- tryCatch(
    eigen(vcov, only.values = TRUE)$values,
    error = \(e) rep(NA, min(1L, NCOL(vcov)))
  )

  if (all(is.na(eigenvalues))) {
    mod_msg_warn("Unable to compute eigenvalues of the variance-covariance matrix!")
    return(NULL)
  }

  minval <- min(eigenvalues, na.rm = TRUE) # should never be any NA, but just in case...
  if (minval < tol.eigen) {
    mod_warnif(minval >= 0,
           paste0("The variance-covariance matrix of the estimated parameters\n",
           "(vcov) does not appear to be positive definite! The smallest\n",
           sprintf("eigenvalue (= %e) is close to zero. This may\n", minval),
           "be a symptom that the model is not identified."))
    mod_warnif(minval < 0,
           paste0("The variance-covariance matrix of the estimated parameters\n",
           "(vcov) does not appear to be positive definite! The smallest\n",
           sprintf("eigenvalue (= %e) is smaller than zero. This may\n", minval),
           "be a symptom that the model is not identified."))
  }
}


isPositiveDefinite <- function(X, tol.eigen = .Machine$double.eps ^ (3/4)) {
  if (is.null(X) || !isSymmetric(X))
    return(FALSE)

  eigenvalues <- tryCatch(eigen(X, only.values = TRUE)$values,
                          error = \(e) rep(NA, min(1L, NCOL(X))))

  !any(is.na(eigenvalues)) & !any(eigenvalues <= tol.eigen)
}

transformedSolutionCOEFS <- function(object,
                                     monte.carlo = FALSE,
                                     mc.reps = 10000,
                                     tolerance.zero = 1e-10,
                                     delta.epsilon = 1e-8,
                                     grouping = NULL,
                                     center = TRUE,
                                     standardize = TRUE,
                                     ...) {
  mod_stopif(!inherits(object, c("modsem_da", "modsem_pi", "lavaan", "modsem_mplus")),
         "The model must be of class `modsem_da`, `modsem_mplus`, `modsem_pi` or `lavaan`!")

  isLav   <- inherits(object, "lavaan")
  isDA    <- inherits(object, "modsem_da")
  isMplus <- inherits(object, "modsem_mplus")

  if (isLav) {
    vcov <- lavaan::vcov # load vcov and coef from lavaan if dealing with a lavaan object
    coef <- lavaan::coef
  }

  parTable <- parameter_estimates(
    object = object,
    colon.pi = TRUE,
    high.order.as.measr = FALSE,
    rm.tmp.ov = FALSE,
    label.renamed.prod = TRUE
  )

  parTable <- subsetByGrouping(parTable, grouping = grouping) # if NULL no subsetting

  if (!"group" %in% names(parTable))
    parTable$group <- 1L

  if (!NROW(parTable)) return(NULL)

  if (isDA || isMplus) {
    cols.keep <- c("lhs", "op", "rhs", "label", "est", "std.error")
    if ("group" %in% names(parTable)) cols.keep <- c(cols.keep, "group")
    cols.keep <- intersect(cols.keep, names(parTable))
    parTable <- parTable[cols.keep]

  } else { # modsem_pi or lavaan
    if (!"label" %in% names(parTable)) parTable$label <- ""
    if (!"se"    %in% names(parTable)) parTable$se    <- NA

    cols.keep <- c("lhs", "op", "rhs", "label", "est", "se")
    if ("group" %in% names(parTable)) cols.keep <- c(cols.keep, "group")
    cols.keep <- intersect(cols.keep, names(parTable))
    parTable <- parTable[cols.keep]
    parTable <- rename(parTable, se = "std.error")
  }

  if (center && (isLav || isDA || isMplus)) { # not relevant for modsem_pi
    if (isDA || isMplus)
      parTable <- meanInteractions(parTable) # get means for interaction terms

    # if we had to center the solution, we have to replace the existing variances
    # if we're using LMS/QML/Mplus there aren't any variances
    isNonCentered <- isNonCenteredParTable(parTable)
    missingVars   <- !hasIntTermVariances(parTable)
    addVariances  <- (isNonCentered && isLav) || isMplus || isDA || missingVars

    if (addVariances) {
      mod_warnif(isLav,
        paste0("Replacing interaction (co-)", "variances when centering the model!\n")
      )

      # When replacing the covariance structure we might change the total variance
      # of the endogenous variables. Here we try to correct for this by
      # adding/subtracting the error to/from the residual variances
      parTableOld <- parTable
      etas        <- getSortedEtas(parTable)

      parTable <- var_interactions(parTable, ignore.means = TRUE, mc.reps = mc.reps)

      if (isLav) for (g in getGroupsParTable(parTable)) {
        maskNew.g <- parTable$group == g
        maskOld.g <- parTableOld$group == g

        if (!any(maskNew.g, na.rm = TRUE) || !any(maskOld.g, na.rm = TRUE)) next

        parTableNew.g <- centerInteractions(parTable[maskNew.g, , drop = FALSE])
        parTableOld.g <- parTableOld[maskOld.g, , drop = FALSE]

        for (eta in etas) {
          newVarEta  <- calcVarParTable(eta, parTable = parTableNew.g)
          oldVarEta  <- calcVarParTable(eta, parTable = parTableOld.g)
          diffVarEta <- newVarEta - oldVarEta

          cond <- (
            parTable$lhs == eta & parTable$lhs == parTable$rhs &
            parTable$op == "~~" & parTable$group == g
          )

          parTable[cond, "est"] <- parTable[cond, "est"] - diffVarEta
        }
      }
    }
  }

  originalLabels <- parTable$label
  labels         <- getParTableLabels(parTable, labelCol="label")
  labels.clean   <- getParTableLabels(parTable, labelCol="label", replace.dup = TRUE)
  parTable$label <- labels
  parTable$std.error <- NA

  groups <- getGroupsParTable(parTable)

  # Get vcov and coefs
  V     <- tryCatch(vcov(object), error = \(e) NULL)
  coefs <- structure(parTable$est, names = labels)

  if (is.null(V)) { # calc.se == FALSE
    k    <- length(coefs)
    pars <- names(coefs)
    V    <- matrix(0, nrow = k, ncol = k,
                   dimnames = list(pars, pars))
  }

  # Subset and expand based on unique labels
  labels.u <- unique(labels)
  V        <- expandVCOV(V, labels=labels.u)
  coefs    <- coefs[labels.u]

  # Replace labels with cleaned versions
  # E.g., if two parameters share the same label, we want to split them int two different labels
  parTable$label <- labels.clean
  legalNames     <- stringr::str_replace_all(labels.clean, OP_REPLACEMENTS)
  paramMapping   <- structure(labels, names = legalNames)

  V       <- V[labels, labels]
  coefs   <- coefs[labels]

  dimnames(V)  <- list(labels.clean, labels.clean)
  names(coefs) <- labels.clean

  if (monte.carlo) {
    COEFS <- as.data.frame(mvtnorm::rmvnorm(mc.reps, mean = coefs, sigma = V))
  } else { # delta method
    k <- length(coefs)
    COEFS <- matrix(coefs, nrow=k, ncol=k, byrow=TRUE)
    DeltaEpsilon <- diag(delta.epsilon, k)
    isFixed <- abs(diag(V)) < tolerance.zero
    DeltaEpsilon[isFixed, isFixed] <- 0
    COEFS_p <- as.data.frame(COEFS + DeltaEpsilon)
    COEFS_m <- as.data.frame(COEFS - DeltaEpsilon)
    COEFS <- rbind(COEFS_p, COEFS_m)
  }

  # Get legal parameter names
  names(COEFS) <- legalNames
  names(coefs) <- legalNames
  COEFS <- rbind(as.data.frame(as.list(coefs)), COEFS) # first row is the original values

  colnames(V) <- legalNames
  rownames(V) <- legalNames

  parTable$label <- stringr::str_replace_all(parTable$label, OP_REPLACEMENTS)
  if ("group" %in% names(parTable)) {
    parTable <- parTable[c("lhs", "op", "rhs", "label", "group")]
  } else {
    parTable <- parTable[c("lhs", "op", "rhs", "label")]
  }

  # Center interactions
  if (center) {
    for (g in groups) {
      mask.g <- parTable$group == g
      if (!any(mask.g, na.rm = TRUE)) next
      parTable.g <- parTable[mask.g, , drop = FALSE]
      COEFS <- centerInteractionsCOEFS(parTable.g, COEFS = COEFS) # re-estimate path-coefficients
    }
    parTable <- parTable[!(parTable$op %in% c("~1", "|") &
                           parTable$group %in% groups), , drop = FALSE]
  }

  # Unstandardized copies
  parTable.ustd  <- parTable
  COEFS.ustd     <- COEFS

  if (standardize) for (g in groups) {
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    if (!NROW(parTable.g)) next

    lVs.g        <- getLVs(parTable.g)
    intTerms.g   <- getIntTerms(parTable.g)
    etas.g       <- getSortedEtas(parTable.g, isLV = FALSE)
    xis.g        <- getXis(parTable.g, etas = etas.g, isLV = FALSE)
    indsLVs.g    <- getIndsLVs(parTable.g, lVs.g)
    allInds.g    <- unique(unlist(indsLVs.g))
    composites.g <- getComposites(parTable.g)
    factors.g    <- setdiff(lVs.g, composites.g)

    vars.g <- unique(c(allInds.g, lVs.g, intTerms.g, xis.g, etas.g))
    varianceEquations.g <- structure(
      getCovEqExprs(
        x = vars.g,
        y = vars.g,
        parTable = parTable.g,
        measurement.model = TRUE
      ),
      names = vars.g
    )
    variances.g <- lapply(varianceEquations.g, FUN = \(eq) eval(eq, envir = COEFS))

    # Factor Loadings
    for (lV in factors.g) {
      inds.lv <- indsLVs.g[[lV]]
      if (!length(inds.lv)) next

      for (ind in inds.lv) {
        selectRows <- (
          parTable.g$lhs == lV &
          parTable.g$op == "=~" &
          parTable.g$rhs == ind
        )

        if (!any(selectRows)) next

        label <- parTable.g[selectRows, "label"]

        var.lv <- variances.g[[lV]]
        var.ind <- variances.g[[ind]]
        if (is.null(var.lv) || is.null(var.ind)) next

        scalingCoef <- sqrt(var.lv) / sqrt(var.ind)
        lambda      <- COEFS[[label]] * scalingCoef

        COEFS[[label]] <- lambda
      }
    }
    
    # Composite Weights
    for (composite in composites.g) {
      inds.comp <- indsLVs.g[[composite]]
      if (!length(inds.comp)) next

      for (ind in inds.comp) {
        selectRows <- (
          parTable.g$lhs == composite &
          parTable.g$op == "<~" &
          parTable.g$rhs == ind
        )

        if (!any(selectRows)) next

        label <- parTable.g[selectRows, "label"]

        var.comp <- variances.g[[composite]]
        var.ind <- variances.g[[ind]]
        if (is.null(var.comp) || is.null(var.ind)) next

        scalingCoef <- sqrt(var.ind) / sqrt(var.comp)
        w <- COEFS[[label]] * scalingCoef

        COEFS[[label]] <- w 
      }
    }

    # Structural Coefficients
    selectStrucExprs <- parTable.g$op == "~" & parTable.g$lhs %in% etas.g

    for (eta in etas.g) {
      selectStrucExprsEta <- selectStrucExprs & parTable.g$lhs == eta
      structExprsEta      <- parTable.g[selectStrucExprsEta, ]

      for (xi in structExprsEta$rhs) {
        selectRows  <- selectStrucExprsEta & parTable.g$rhs == xi
        if (!any(selectRows)) next
        var.xi <- variances.g[[xi]]
        var.eta <- variances.g[[eta]]
        if (is.null(var.xi) || is.null(var.eta)) next

        scalingCoef <- sqrt(var.xi) / sqrt(var.eta)
        label       <- parTable.g[selectRows, "label"]
        gamma       <- COEFS[[label]] * scalingCoef

        COEFS[[label]] <- gamma
      }
    }

    # (Co-) Variances of xis
    selectCovXis <- parTable.g$op == "~~" &
      (parTable.g$lhs %in% c(xis.g, intTerms.g) |
         parTable.g$rhs %in% c(xis.g, intTerms.g))

    covRowsXis <- parTable.g[selectCovXis, , drop = FALSE]
    for (i in seq_len(nrow(covRowsXis))) {
      lhs <- covRowsXis$lhs[[i]]
      rhs <- covRowsXis$rhs[[i]]
      xis_pair <- c(lhs, rhs)
      selectRows  <- selectCovXis &
        parTable.g$lhs %in% xis_pair &
        parTable.g$rhs %in% xis_pair

      var.lhs <- variances.g[[lhs]]
      var.rhs <- variances.g[[rhs]]
      if (is.null(var.lhs) || is.null(var.rhs)) next

      scalingCoef <- sqrt(var.lhs) * sqrt(var.rhs)

      if (lhs != rhs) {
        selectRows <- selectRows & parTable.g$lhs != parTable.g$rhs
      }

      label <- parTable.g[selectRows, "label"]
      covs <- COEFS[[label]] / scalingCoef

      COEFS[[label]] <- covs
    }

    # Residual Variances and Covariances of etas
    etaCovariances <- parTable.g[
      parTable.g$lhs %in% etas.g &
      parTable.g$op == "~~" &
      parTable.g$rhs %in% etas.g, , drop = FALSE
    ]

    for (i in seq_len(NROW(etaCovariances))) {
      row   <- etaCovariances[i, , drop = FALSE]
      eta.x <- row$lhs
      eta.y <- row$rhs
      label <- row$label

      var.x <- variances.g[[eta.x]]
      var.y <- variances.g[[eta.y]]
      if (is.null(var.x) || is.null(var.y)) next

      COEFS[[label]] <- COEFS[[label]] / (sqrt(var.x) * sqrt(var.y))
    }

    # residual variances of indicators
    indicatorCovariances <- parTable.g[
      parTable.g$lhs %in% allInds.g &
      parTable.g$op == "~~" &
      parTable.g$rhs %in% allInds.g, , drop = FALSE
    ]

    for (i in seq_len(NROW(indicatorCovariances))) {
      row <- indicatorCovariances[i, , drop = FALSE]

      ind.x <- row$lhs
      ind.y <- row$rhs
      label <- row$label

      var.x <- variances.g[[ind.x]]
      var.y <- variances.g[[ind.y]]

      COEFS[[label]] <- COEFS[[label]] / (sqrt(var.x) * sqrt(var.y))
    }

    # Correct Scale of interaction terms
    COEFS <- correctStdSolutionCOEFS(
      parTable = parTable.g, # for generating equations
      COEFS.std = COEFS,
      COEFS.ustd = COEFS.ustd,
      variances = variances.g,
      intTerms = intTerms.g
    )
  }

  # recalculate custom parameters
  constrExprs <- sortConstrExprsFinalPt(parTable)
  for (i in seq_len(NROW(constrExprs))) {
    row <- constrExprs[i, , drop=FALSE]
    label <- row$label
    expr <- parse(text=constrExprs[i, "rhs"])
    newVals <- eval(expr, envir = COEFS)

    COEFS[[label]] <- newVals
  }

  nzeros <- round(log10(1 / tolerance.zero), 0)
  coefs <- unlist(COEFS[1, ])
  COEFS <- round(COEFS[2:(nrow(COEFS)), ],  nzeros) # skip first row

  if (monte.carlo) {
    vcov <- stats::cov(COEFS)
  } else {
    # delta method
    p <- 1:k
    m <- (k+1):(k + k)

    COEFS_p <- t(COEFS[p, ])
    COEFS_m <- t(COEFS[m, ])

    J <- (COEFS_p - COEFS_m) / (2 * delta.epsilon)
    vcov <- J %*% V %*% t(J)
  }

  # fill parTable
  std.errors <- suppressWarnings(sqrt(diag(vcov)))
  warnFunc <- function(type, row) {
    mod_msg_warn(paste0("Unable to calculate standardized ", type, " for: ",
             paste0(row$lhs, row$op, row$rhs)))
  }

  verboseLabels <- stringr::str_replace_all(parTable$label, OP_REPLACEMENTS)
  for (i in seq_len(nrow(parTable))) {
    row <- parTable[i, , drop = FALSE]
    label <- verboseLabels[[i]]

    if (!label %in% names(coefs)) {
      warnFunc("coefficient", row)
      est <- NA
    } else est <- coefs[[label]]

    if (!label %in% names(std.errors)) {
      warnFunc("std.error", row)
      se <- NA
    } else se <- std.errors[[label]]

    parTable[i, "est"] <- est
    parTable[i, "std.error"] <- se
  }

  # Fill in NA on zero-standard errors, and create vcov and coefs
  parTable[!is.na(parTable$std.error) &
           abs(parTable$std.error) < tolerance.zero, "std.error"] <- NA

  isFree <- !is.na(parTable$std.error)
  coefs <- structure(parTable$est[isFree], names = parTable$label[isFree])
  params <- intersect(names(coefs), rownames(vcov))

  coefs <- coefs[params]
  vcov <- vcov[params, params]

  cleanedParamLabels <- paramMapping[params]
  names(coefs) <- cleanedParamLabels
  colnames(vcov) <- rownames(vcov) <- cleanedParamLabels

  # Finalize parTable
  parTable$label    <- paramMapping[parTable$label]
  parTable$z.value  <- parTable$est / parTable$std.error
  parTable$p.value  <- 2 * stats::pnorm(-abs(parTable$z.value))
  parTable$ci.lower <- parTable$est - CI_WIDTH * parTable$std.error
  parTable$ci.upper <- parTable$est + CI_WIDTH * parTable$std.error

  # Remove added labels
  labelInOrig       <- parTable$label %in% originalLabels
  parTable[!labelInOrig, "label"] <- ""

  if (isDA) {
    indsHigherOrderLVs <- object$model$info$indsHigherOrderLVs
    parTable <- higherOrderStruct2Measr(parTable = parTable,
                                        indsHigherOrderLVs = indsHigherOrderLVs)
    parTable <- sortParTableDA(parTable, model = object$model)
  }

  # Reset index
  rownames(parTable) <- NULL

  list(parTable  = modsemParTable(parTable),
       coefs     = coefs,
       vcov      = vcov,
       COEFS     = COEFS)
}


correctStdSolutionCOEFS <- function(parTable,
                                    COEFS.std,
                                    COEFS.ustd,
                                    variances,
                                    intTerms) {
  for (XZ in intTerms) {
    elems <- stringr::str_split(XZ, pattern = ":")[[1L]]

    vars  <- as.data.frame(variances[elems])
    sds   <- sqrt(vars)

    rowsXZ <- parTable[parTable$rhs == XZ & parTable$op == "~", , drop = FALSE]
    Y <- rowsXZ$lhs[[1]]

    if (!length(Y)) {
      mod_msg_warn(paste0("No endogenous variable found for interaction term '", XZ, "'."))
      next
    }

    # Find the relevant exogenous variables
    struct <- parTable[parTable$lhs == Y & parTable$op == "~", , drop = FALSE]
    xis    <- struct$rhs

    # Unstandardized terms
    varY  <- variances[[Y]]
    sdY   <- sqrt(varY)
    B3    <- getCOEFS(y = Y, x = XZ, COEFS = COEFS.ustd, parTable = parTable)

    # Correlations
    combosXis <- getUniqueCombos(xis)
    combosXis <- combosXis[combosXis[[1]] == XZ |
                           combosXis[[2]] == XZ, , drop = FALSE]

    lxis <- combosXis[[1]]
    rxis <- combosXis[[2]]

    eqCorrs <- getCovEqExprs(x = lxis, y = rxis, parTable = parTable)
    corrs <- matrix(NA, nrow = NROW(COEFS.std), ncol = length(lxis))

    for (i in seq_len(ncol(corrs))) {
      eqCorr <- eqCorrs[[i]]
      corr <- eval(eqCorr, envir = COEFS.std)
      corr[is.na(corr)] <- 0 # in case there is no connection, it should be zero

      corrs[, i] <- corr
    }

    # Incorrectly standardized terms
    lcoefsIncorrect <- getCOEFS(x = lxis, y = Y, COEFS = COEFS.std, parTable = parTable)
    rcoefsIncorrect <- getCOEFS(x = rxis, y = Y, COEFS = COEFS.std, parTable = parTable)

    corrtermsIncorrect <- rowSums(2 * lcoefsIncorrect * rcoefsIncorrect * corrs)

    b3Incorrect <- getCOEFS(y = Y, x = XZ, COEFS = COEFS.std, parTable = parTable)
    projVarY_XZ <- b3Incorrect ^ 2 + corrtermsIncorrect # this should be the same
                                                        # for both the correctly, and
                                                        # incorrectly standardized terms
                                                        # and is the identity which makes it
                                                        # possible to standardize the terms correctly

    # Correctly standardized terms
    rowProds <- apply(sds, MARGIN = 1, FUN = prod)
    b3Correct <- B3 * abs(rowProds / sdY) # in case some variances are negative
                                           # we want to make sure we don't flip
                                           # the sign...

    lcoefsCorrect <- lcoefsIncorrect
    rcoefsCorrect <- rcoefsIncorrect
    if (any(lxis == XZ)) lcoefsCorrect[lxis == XZ] <- b3Correct
    if (any(rxis == XZ)) rcoefsCorrect[rxis == XZ] <- b3Correct
    corrtermsCorrect <- rowSums(2 * lcoefsCorrect * rcoefsCorrect * corrs)

    # Calculate the correct standard deviation of the interaction term
    # Using the identity:
    #   projVarY_XZ = b3Correct ^ 2 * sd(xz) ^ 2 + sd(xz) * corrterms
    # Solve for sd(xz) using the quadratic formula:
    #   sd(xz) = (- corrterms +/- sqrt(corrterms^2 + 4 * b3Correct ^ 2 * projVarY_XZ)) / 2 * b3Correct ^ 2
    numerator <- -corrtermsCorrect + sign(projVarY_XZ) *
      sqrt(corrtermsCorrect^2 + 4 * (b3Correct ^ 2) * projVarY_XZ)
    denominator <- 2 * (b3Correct ^ 2)
    sdXZ <- numerator / denominator # correctly standardized sd(xz)

    # Scaling factors
    scalefVar  <- (sdXZ^2) / 1
    scalefCoef <- b3Correct / b3Incorrect
    scalefCov  <- sdXZ

    lequalY  <- parTable$lhs == Y
    requalXZ <- parTable$rhs == XZ
    lequalXZ <- parTable$lhs == XZ
    isCov    <- parTable$op == "~~"
    isCoef   <- parTable$op == "~"

    covTerms <- parTable$label[xor(lequalXZ, requalXZ) & isCov]
    coefTerm <- parTable$label[lequalY & requalXZ & isCoef]
    varTerm  <- parTable$label[lequalXZ & requalXZ & isCov]

    COEFS.std[covTerms] <- COEFS.std[covTerms] * scalefCov
    COEFS.std[coefTerm] <- COEFS.std[coefTerm] * scalefCoef
    COEFS.std[varTerm]  <- COEFS.std[varTerm]  * scalefVar
  }

  COEFS.std
}


standardizedSolutionCOEFS <- function(object,
                                      monte.carlo = FALSE,
                                      mc.reps = 10000,
                                      tolerance.zero = 1e-10,
                                      delta.epsilon = 1e-8,
                                      grouping = NULL,
                                      center = TRUE,
                                      ...) {
  transformedSolutionCOEFS(
    object = object,
    monte.carlo = monte.carlo,
    mc.reps = mc.reps,
    tolerance.zero = tolerance.zero,
    delta.epsilon = delta.epsilon,
    grouping = grouping,
    standardize = TRUE,
    ...
  )
}


centeredSolutionCOEFS <- function(object,
                                  monte.carlo = FALSE,
                                  mc.reps = 10000,
                                  tolerance.zero = 1e-10,
                                  delta.epsilon = 1e-8,
                                  grouping = NULL,
                                  ...) {
  transformedSolutionCOEFS(
    object         = object,
    monte.carlo    = monte.carlo,
    mc.reps        = mc.reps,
    tolerance.zero = tolerance.zero,
    delta.epsilon  = delta.epsilon,
    standardize    = FALSE,
    center         = TRUE,
    ...
  )
}


getMeanFormula <- function(x, parTable, label.col = "label") {
  mod_stopif(length(x) > 1, "x must be a single string")

  meanY <- getIntercept(x, parTable = parTable, col = label.col)
  gamma <- parTable[parTable$lhs == x & parTable$op == "~", , drop = FALSE]

  if (NROW(gamma) == 0) return(meanY)
  for (i in seq_len(NROW(gamma))) {
    meanX <- getMeanFormula(gamma[i, "rhs"], parTable = parTable)
    meanY <- paste0("(", meanY, "+", gamma[i, label.col], "*", meanX, ")")
  }

  if (!length(meanY)) "0" else paste0("(", meanY, ")")
}


centerInteractionsCOEFS <- function(parTable, COEFS, center.means = TRUE,
                                    label.col = "label") {
  rows <- getIntTermRows(parTable)

  for (i in seq_len(NROW(rows))) {
    Y <- rows[i, "lhs"]
    XZ <- unlist(stringr::str_split(rows[i, "rhs"], ":"))
    X <- XZ[[1]]
    Z <- XZ[[2]]
    gamma <- parTable[parTable$lhs == Y & parTable$op == "~", , drop = FALSE]

    formulaMeanX <- parse(text = getMeanFormula(X, parTable = parTable))
    formulaMeanZ <- parse(text = getMeanFormula(Z, parTable = parTable))

    labelGammaXZ <- rows[i, label.col]
    labelGammaX  <- gamma[gamma$rhs == X, label.col] # length should always be 1, but just in case...
    labelGammaZ  <- gamma[gamma$rhs == Z, label.col]

    meanX   <- eval(formulaMeanX, envir = COEFS)
    meanZ   <- eval(formulaMeanZ, envir = COEFS)
    gammaXZ <- COEFS[[labelGammaXZ]]

    if (length(labelGammaX) == 1) {
      gammaX  <- COEFS[[labelGammaX]]
      COEFS[[labelGammaX]] <- gammaX + gammaXZ * meanZ
    }

    if (length(labelGammaZ) == 1) {
      gammaZ  <- COEFS[[labelGammaZ]]
      COEFS[[labelGammaZ]] <- gammaZ + gammaXZ * meanX
    }
  }

  if (center.means) {
    innerVars <- unique(unlist(parTable[parTable$op == "~", c("rhs", "lhs")]))
    interceptLabels <- parTable[parTable$lhs %in% innerVars &
                                parTable$op == "~1", label.col]

    for (label in interceptLabels)
      COEFS[[label]] <- 0
  }

  COEFS
}


addTransformedEstimatesPT <- function(parTable,
                                      FUN,
                                      pass.parTable = TRUE,
                                      values.to = "transformed",
                                      values.from = "est",
                                      merge.by = c("lhs", "op", "rhs", "group"),
                                      ...) {
    if (pass.parTable) parTable.transform <- FUN(parTable = parTable, ...)
    else               parTable.transform <- FUN(...)

    parTable.transform[[values.to]] <- parTable.transform[[values.from]]
    parTable.transform <- parTable.transform[c(merge.by, values.to)]

    leftJoin(left   = parTable,
             right  = parTable.transform,
             by     = merge.by)
}


applyTransformationByGrouping <- function(parTable,
                                          FUN,
                                          groupingcols = c("block", "group"),
                                          ...) {
  if (any(groupingcols %in% colnames(parTable))) {
    groupingcols <- intersect(groupingcols, colnames(parTable))
    categories   <- unique(parTable[groupingcols])

    parTable.out <- NULL

    for (i in seq_len(NROW(categories))) {
      grouping       <- structure(unlist(categories[i, ]),
                                  names = groupingcols)
      parTable.out_i <- FUN(..., grouping = grouping)

      if (is.null(parTable.out_i))
        next

      for (group in names(grouping))
        parTable.out_i[[group]] <- grouping[[group]]

      parTable.out <- rbind(parTable.out, parTable.out_i)
    }

    colblock1 <- c("lhs", "op", "rhs")
    colblock2 <- groupingcols
    colblock3 <- setdiff(colnames(parTable.out), c(colblock1, colblock2))

    parTable.out <- parTable.out[c(colblock1, colblock2, colblock3)]

  } else parTable.out <- FUN(...)

  parTable.out
}

createLabelCov <- function(x, y)
  paste("Cov", x, y, sep = "_")


createLabelVar <- function(x)
  paste("Var", x, sep = "_")


createLabelLambda <- function(ind, latent)
  paste("lambda", ind, latent, sep = "_")


createLabelLambdaSquared <- function(ind, latent)
  paste0(createLabelLambda(ind, latent), " ^ 2")


createLabelGamma <- function(x, y)
  paste("Gamma", x, y, sep = "_")


createLabelMean <- function(x)
  paste("Mean", x, sep = "_")


createLabelZeta <- function(x)
  paste("Zeta", x, sep = "_")

MODSEM_COLORS <- rlang::env(
  # user-configurable color names
  positive = NA_character_,
  negative = NA_character_,
  true     = NA_character_,
  false    = NA_character_,
  nan      = NA_character_,
  na       = NA_character_,
  inf      = NA_character_,
  string   = NA_character_,

  # run-time flags & helpers
  active   = FALSE,
  available = NULL,

  # pre-compiled styling functions (identity placeholders)
  f.positive = \(x) x,
  f.negative = \(x) x,
  f.true     = \(x) x,
  f.false    = \(x) x,
  f.nan      = \(x) x,
  f.na       = \(x) x,
  f.inf      = \(x) x,
  f.string   = \(x) x
)


PATTERN.pos    <- "(?<![-+A-Za-z0-9._~])([0-9]+(?:[.-][0-9]+)*(?:\\.[0-9]*)?(?:[eE][+-]?[0-9]+)?)(?![A-Za-z])"
PATTERN.neg    <- "(?<![-+A-Za-z0-9._~])(-[0-9]+(?:[.-][0-9]+)*(?:\\.[0-9]*)?(?:[eE][+-]?[0-9]+)?)(?![A-Za-z])"
PATTERN.TRUE   <- "(?<![A-Za-z0-9])TRUE(?![A-Za-z0-9])"
PATTERN.FALSE  <- "(?<![A-Za-z0-9])FALSE(?![A-Za-z0-9])"
PATTERN.NaN    <- "(?<![A-Za-z0-9])NaN(?![A-Za-z0-9])"
PATTERN.NA     <- "(?<![A-Za-z0-9])NA(?![A-Za-z0-9])"
PATTERN.Inf    <- "(?<![A-Za-z0-9])-?Inf(?![A-Za-z0-9])"
PATTERN.String <- "(\"(?:[^\"\\\\]|\\\\.)*\"|'(?:[^'\\\\]|\\\\.)*')"


setAvailableColors <- function() {
  MODSEM_COLORS$available <- grDevices::colors()
}


styleOrAsIs <- function(col) {
  if (is.na(col)) return(\(x) x)
  cli::make_ansi_style(col)
}


getColorizedASCII <- function(col) {
  if (is.na(col)) return("\\0") # no color -> leave as-is
  as.character(cli::make_ansi_style(col)("\\0"))
}


resetModsemColors <- function() {
  MODSEM_COLORS$active <- FALSE
  fields <- c("positive","negative","true","false","nan","na","inf","string")
  for (f in fields) MODSEM_COLORS[[f]] <- NA_character_

  funs <- grep("^f\\.", names(MODSEM_COLORS), value = TRUE)
  for (f in funs)  MODSEM_COLORS[[f]] <- \(x) x

  setAvailableColors()
}


captureOutput <- function (expr, envir = parent.frame(), split = FALSE) {
  # Capture output using a raw connections, as it is exponentially faster
  # for large outputs
  eval({
    file <- rawConnection(raw(0L), open = "w")

    on.exit(if (!is.null(file)) close(file))

    utils::capture.output(expr, file = file, split = split)

    res <- rawConnectionValue(file)
    close(file)
    file <- NULL

    rawToChar(res)

  }, envir = envir, enclos = envir)
}


#' Define or disable the color theme used by \code{modsem}
#'
#' All arguments are optional; omitted ones fall back to the defaults below.
#' Pass \code{active = FALSE} to turn highlighting off (and reset the palette).
#'
#' @param positive color of positive numbers.
#' @param negative color of negative numbers.
#' @param true color of \code{TRUE}.
#' @param false color of \code{FALSE}.
#' @param nan color of \code{NaN}.
#' @param na color of \code{NA}.
#' @param inf color of \code{-Inf} and \code{Inf}.
#' @param string color of quoted strings.
#' @param active Should color-theme be activated/deactived?
#'
#' @examples
#' set_modsem_colors(positive = "red3",
#'                   negative = "red3",
#'                   true = "darkgreen",
#'                   false = "red3",
#'                   na = "purple",
#'                   string = "darkgreen")
#'
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' "
#'
#' est <- modsem(m1, data = oneInt)
#' colorize_output(summary(est))
#' colorize_output(est) # same as colorize_output(print(est))
#' colorize_output(modsem_inspect(est, what = "coef"))
#'
#' \dontrun{
#' colorize_output(split = TRUE, {
#'   # Get live (uncolored) output
#'   # And print colored output at the end of execution
#'
#'   est_lms <- modsem(m1, data = oneInt, method = "lms")
#'   summary(est_lms)
#' })
#'
#' colorize_output(modsem_inspect(est_lms))
#' }
#' @return \code{TRUE} if colors are active afterwards, otherwise \code{FALSE}.
#' @export
set_modsem_colors <- function(positive = "green3",
                              negative = positive,
                              true     = "green3",
                              false    = "red",
                              nan      = "purple",
                              na       = "purple",
                              inf      = "purple",
                              string   = "cyan",
                              active   = TRUE) {
  if (is.null(MODSEM_COLORS$available)) setAvailableColors()

  cols <- c(positive, negative, true, false, nan, na, inf, string)
  if (!active || any(!cols %in% MODSEM_COLORS$available)) {
    resetModsemColors()
    return(FALSE)
  }

  MODSEM_COLORS$positive <- positive
  MODSEM_COLORS$negative <- negative
  MODSEM_COLORS$true     <- true
  MODSEM_COLORS$false    <- false
  MODSEM_COLORS$nan      <- nan
  MODSEM_COLORS$na       <- na
  MODSEM_COLORS$inf      <- inf
  MODSEM_COLORS$string   <- string

  MODSEM_COLORS$f.positive <- styleOrAsIs(positive)
  MODSEM_COLORS$f.negative <- styleOrAsIs(negative)
  MODSEM_COLORS$f.true     <- styleOrAsIs(true)
  MODSEM_COLORS$f.false    <- styleOrAsIs(false)
  MODSEM_COLORS$f.nan      <- styleOrAsIs(nan)
  MODSEM_COLORS$f.na       <- styleOrAsIs(na)
  MODSEM_COLORS$f.inf      <- styleOrAsIs(inf)
  MODSEM_COLORS$f.string   <- styleOrAsIs(string)

  MODSEM_COLORS$active <- TRUE

  MODSEM_COLORS$active
}

#' Capture, colorise, and emit console text
#' @param expr Expression or object with output which should be colorized.
#' @param positive color of positive numbers.
#' @param negative color of negative numbers.
#' @param true color of \code{TRUE}.
#' @param false color of \code{FALSE}.
#' @param nan color of \code{NaN}.
#' @param na color of \code{NA}.
#' @param inf color of \code{-Inf} and \code{Inf}.
#' @param string color of quoted strings.
#' @param append String appended after the colored output (default `\\n`).
#' @param split Should output be splitted? If \code{TRUE} the output is printed
#'  normally (in real time) with no colorization, and the colored output is printed
#'  after the code has finished executing.
#' @return Invisibly returns the *plain* captured text.
#'
#' @examples
#' set_modsem_colors(positive = "red3",
#'                   negative = "red3",
#'                   true = "darkgreen",
#'                   false = "red3",
#'                   na = "purple",
#'                   string = "darkgreen")
#'
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' "
#'
#' est <- modsem(m1, data = oneInt)
#' colorize_output(summary(est))
#' colorize_output(est) # same as colorize_output(print(est))
#' colorize_output(modsem_inspect(est, what = "coef"))
#'
#' \dontrun{
#' colorize_output(split = TRUE, {
#'   # Get live (uncolored) output
#'   # And print colored output at the end of execution
#'
#'   est_lms <- modsem(m1, data = oneInt, method = "lms")
#'   summary(est_lms)
#' })
#'
#' colorize_output(modsem_inspect(est_lms))
#' }
#' @export
colorize_output <- function(expr,
                            positive = MODSEM_COLORS$positive,
                            negative = MODSEM_COLORS$negative,
                            true     = MODSEM_COLORS$true,
                            false    = MODSEM_COLORS$false,
                            nan      = MODSEM_COLORS$nan,
                            na       = MODSEM_COLORS$na,
                            inf      = MODSEM_COLORS$inf,
                            string   = MODSEM_COLORS$string,
                            split    = FALSE,
                            append = "\n") {
  getFinalLine  <- \(x) stringr::str_split(x, pattern = "\r")
  getFinalLines <- \(x) vapply(getFinalLine(x), FUN = last,
                               FUN.VALUE = character(1L))

  lines <- captureOutput(expr, split = split)

  # Handle "\r"
  hasReplace <- grepl("\r", lines)
  lines[hasReplace] <- getFinalLines(lines[hasReplace])

  out <- stringr::str_c(lines, collapse = "\n")

  if (!MODSEM_COLORS$active) {
    cat(out, append)
    return(invisible(out))
  }

  # build mapping only for active colors
  mapping <- c()
  add <- function(pat, col) {
    if (!is.na(col))
      mapping <<- c(mapping, stats::setNames(getColorizedASCII(col), pat))
  }

  add(PATTERN.pos,    positive)
  add(PATTERN.neg,    negative)
  add(PATTERN.TRUE,   true)
  add(PATTERN.FALSE,  false)
  add(PATTERN.NaN,    nan)
  add(PATTERN.NA,     na)
  add(PATTERN.Inf,    inf)
  add(PATTERN.String, string)

  colored <- tryCatch(
    stringr::str_replace_all(out, mapping),
    error = function(e) {
      warning("colorization failed: ", conditionMessage(e), call. = FALSE)
      out
    }
  )

  cat(colored, append)
  invisible(out)
}

devtools::load_all()

active <- set_modsem_colors(positive = "red3",
                            negative = "red3",
                            true = "darkgreen",
                            false = "red3",
                            na = "purple",
                            string = "darkgreen")

testthat::expect_true(active)

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est <- modsem(m1, data = oneInt)
colorize_output(summary(est))

colorize_output(print(est))

colorize_output(modsem_inspect(est, "coef"))


colorize_output(split = TRUE, {
  # Get live (uncolored) output
  # And print colored output at the end of execution

  est_lms <- modsem(m1, data = oneInt, method = "lms")
  summary(est_lms)
})

colorize_output(modsem_inspect(est_lms))

computeFullIobs <- function(theta, model, P, louis = TRUE) {

  if (louis) {
    louisInfo <- observedInfoFromLouisLms(
      model = model,
      theta = theta,
      P = P
    )

    I <- louisInfo$I.obs

  } else {

    # negative hessian (sign = -1)
    I <- hessianObsLogLikLms(
      theta = theta,
      model = model,
      P     = P,
      sign  = -1
    )

  }

  # make sure it's symmetric
  0.5 * (I + t(I))
}


lmsIterationHistory <- function() {
  data.frame(
    iteration = integer(),
    mode = character(),
    loglik = numeric(),
    abs.change = numeric(),
    rel.change = numeric(),
    elapsed.time = numeric(),
    expected.remaining.iterations = numeric(),
    expected.final.loglik = numeric()
  )
}


# `history` is the running vector of observed log-likelihoods kept by `emLms()`,
# one entry per iteration
normalizeLmsLogLikHistory <- function(history) {
  logLik <- as.numeric(history)

  if (!length(logLik)) {
    return(data.frame(
      iteration  = integer(),
      logLik     = numeric(),
      deltaLL    = numeric(),
      relDeltaLL = numeric()
    ))
  }

  deltaLL <- c(NA_real_, diff(logLik))

  data.frame(
    iteration  = seq_along(logLik),
    logLik     = logLik,
    deltaLL    = deltaLL,
    relDeltaLL = c(NA_real_, deltaLL[-1L] / abs(logLik[-length(logLik)]))
  )
}


forecastLogLikLms <- function(history,
                              convergence.abs,
                              convergence.rel,
                              min.gains,
                              window) {
  empty <- function() {
    data.frame(
      expected.remaining.iterations = NA_real_,
      expected.final.loglik = NA_real_
    )
  }

  history <- normalizeLmsLogLikHistory(history)

  okLL <- is.finite(history$logLik)
  history <- history[okLL, , drop = FALSE]
  if (!NROW(history)) return(empty())

  currentLL <- history$logLik[NROW(history)]
  currentGain <- history$deltaLL[NROW(history)]
  gain <- history$deltaLL
  gainIteration <- history$iteration
  okGain <- is.finite(gain) & gain > 0
  gain <- gain[okGain]
  gainIteration <- gainIteration[okGain]

  if (length(gain) < min.gains) return(empty())

  if (is.finite(window) && length(gain) > window) {
    keep <- seq.int(length(gain) - window + 1L, length(gain))
    gain <- gain[keep]
    gainIteration <- gainIteration[keep]
  }

  fitData <- data.frame(iteration = gainIteration, logGain = log(gain))
  fit <- stats::lm(logGain ~ iteration, data = fitData)
  coefs <- stats::coef(fit)
  slope <- unname(coefs[["iteration"]])
  intercept <- unname(coefs[["(Intercept)"]])
  decay <- exp(slope)

  if (!is.finite(decay) || decay <= 0 || decay >= 1) {
    ratios <- gain[-1L] / gain[-length(gain)]
    ratios <- ratios[is.finite(ratios) & ratios > 0 & ratios < 1]
    if (!length(ratios)) return(empty())
    decay <- stats::median(ratios)
    expectedNextGain <- gain[[length(gain)]] * decay
  } else {
    expectedNextGain <- exp(intercept + slope * (history$iteration[NROW(history)] + 1L))
  }

  if (!is.finite(expectedNextGain) || expectedNextGain < 0) return(empty())

  decay <- max(0, min(decay, 1 - sqrt(.Machine$double.eps)))
  expectedRemainingGain <- if (decay == 0) expectedNextGain else expectedNextGain / (1 - decay)
  expectedFinalLogLik <- currentLL + expectedRemainingGain

  relTol <- convergence.rel * max(1, abs(currentLL), abs(expectedFinalLogLik), na.rm = TRUE)
  stopGain <- max(convergence.abs, relTol, na.rm = TRUE)
  currentAbsGain <- abs(currentGain)

  if (is.finite(currentAbsGain) && currentAbsGain < stopGain) {
    expectedRemainingIterations <- 0L
  } else if (expectedNextGain <= stopGain || decay == 0) {
    expectedRemainingIterations <- 1L
  } else {
    expectedRemainingIterations <- ceiling(log(stopGain / expectedNextGain) / log(decay) + 1)
  }

  data.frame(
    expected.remaining.iterations = expectedRemainingIterations,
    expected.final.loglik = expectedFinalLogLik
  )
}


appendLmsIterationHistory <- function(history, iteration, loglik.history,
                                      convergence.abs, convergence.rel,
                                      forecast.min.gains,
                                      forecast.window,
                                      mode,
                                      loglik,
                                      abs.change,
                                      rel.change,
                                      elapsed.time) {
  forecast <- forecastLogLikLms(
    history         = loglik.history,
    convergence.abs = convergence.abs,
    convergence.rel = convergence.rel,
    min.gains       = forecast.min.gains,
    window          = forecast.window
  )

  row <- data.frame(
    iteration = iteration,
    mode = as.character(mode),
    loglik = as.numeric(loglik),
    abs.change = as.numeric(abs.change),
    rel.change = as.numeric(rel.change),
    elapsed.time = as.numeric(elapsed.time)
  )

  rbind(history, cbind(row, forecast))
}


emLms <- function(model,
                  algorithm = c("EMA", "EM"),
                  em.control = list(),
                  verbose = FALSE,
                  convergence.abs = 1e-4,
                  convergence.rel = 1e-10,
                  max.iter = 500,
                  max.step = 1,
                  control = list(),
                  calc.se = TRUE,
                  FIM = "observed",
                  OFIM.hessian = FALSE,
                  EFIM.S = 3e4,
                  EFIM.parametric = TRUE,
                  robust.se = FALSE,
                  epsilon = 1e-6,
                  optimizer = "nlminb",
                  R.max = 1e6,
                  adaptive.quad = FALSE,
                  quad.range = -Inf,
                  adaptive.quad.tol = 1e-12,
                  nodes = 24,
                  cr1s = TRUE,
                  # NOTE: defaults are evaluated lazily in the frame of `emLms()`,
                  # so `em.control` overrides any of these which aren't passed
                  # explicitly by the caller
                  ema.min.iter            = em.control$ema.min.iter            %||% 10L,
                  ema.trend.min           = em.control$ema.trend.min           %||% 0,
                  ema.forecast.min.gains  = em.control$ema.forecast.min.gains  %||% 3L,
                  ema.forecast.window     = em.control$ema.forecast.window     %||% 10L,
                  qn.remaining.min        = em.control$qn.remaining.min        %||% 10,
                  qn.remaining.max        = em.control$qn.remaining.max        %||% 200,
                  fs.remaining.min        = em.control$fs.remaining.min        %||% 200,
                  fs.remaining.max        = em.control$fs.remaining.max        %||% 500,
                  fallback.max.step.min   = em.control$fallback.max.step.min   %||% 50,
                  fallback.max.step.mult  = em.control$fallback.max.step.mult  %||% 50,
                  qn.maxit.start          = em.control$qn.maxit.start          %||% 1L,
                  qn.maxit.mult           = em.control$qn.maxit.mult           %||% 2,
                  qn.maxit.max            = em.control$qn.maxit.max            %||% 25L,
                  qn.factr                = em.control$qn.factr                %||% NULL,
                  qn.pgtol                = em.control$qn.pgtol                %||% 0,
                  ema.final.qn            = em.control$ema.final.qn            %||% TRUE,
                  ema.final.qn.maxit      = em.control$ema.final.qn.maxit      %||% qn.maxit.max,
                  ema.final.qn.factr      = em.control$ema.final.qn.factr      %||% qn.factr,
                  ema.final.qn.rounds.max = em.control$ema.final.qn.rounds.max %||% 25L,
                  fs.maxit.start          = em.control$fs.maxit.start          %||% 1L,
                  ...) {

  algorithm <- toupper(match.arg(algorithm))
  ema.min.iter <- as.integer(ema.min.iter)
  ema.forecast.min.gains <- as.integer(ema.forecast.min.gains)
  qn.maxit.start <- as.integer(qn.maxit.start)
  qn.maxit.max <- as.integer(qn.maxit.max)
  ema.final.qn.maxit <- as.integer(ema.final.qn.maxit)
  ema.final.qn.rounds.max <- as.integer(ema.final.qn.rounds.max)
  fs.maxit.start <- as.integer(fs.maxit.start)

  default.optim.factr <- if (is.finite(convergence.rel) && convergence.rel > 0) {
    max(1, convergence.rel / .Machine$double.eps)
  } else {
    NULL
  }
  qn.factr <- if (is.null(qn.factr)) default.optim.factr else qn.factr
  ema.final.qn.factr <- if (is.null(ema.final.qn.factr)) qn.factr else ema.final.qn.factr
  do.final.qn <- isTRUE(ema.final.qn) && isTRUE(model$info$has.interaction)

  history <- lmsIterationHistory()
  loglik.history <- numeric()

  theta.lower  <- model$params$bounds$lower
  theta.upper  <- model$params$bounds$upper

  max.iter.fs <- fs.maxit.start
  max.iter.qn <- qn.maxit.start

  tryCatch({
    logLikNew <- -Inf
    thetaNew  <- model$theta

    lastQuad     <- NULL
    adaptiveQuad <- model$models[[1L]]$quad$adaptive # fixed across groups
    adaptiveFreq <- model$models[[1L]]$quad$adaptive.frequency

    n.em.iter  <- 0
    mode       <- "EM"
    iterations <- 0L
    run        <- TRUE
    ema.final.phase <- "none"
    ema.final.qn.rounds <- 0L

    testSimpleGradient <- !model$params$gradientStruct$useFDGradient

    history.start.time <- proc.time()[["elapsed"]]

    while (run) {
      iterations <- iterations + 1L
      logLikOld  <- logLikNew
      thetaOld   <- thetaNew
      recalcQuad <- adaptiveQuad && iterations %% adaptiveFreq == 0L
      tmp.max.step <- 0L

      mod_stopif(any(!is.finite(thetaOld)),
        "Missing/Non finite values in estimated parameters at iteration",
        paste0(iterations, "!")
      )

      # E-step at thetaOld
      P <- estepLms(
        model             = model,
        theta             = thetaOld,
        lastQuad          = lastQuad,
        recalcQuad        = recalcQuad,
        adaptive.quad.tol = adaptive.quad.tol,
        ...
      )

      if (testSimpleGradient) {
        tryCatch({
          gradientObsLogLikLms(theta = thetaNew, model = model, P = P)
        }, error = \(e) {
          mod_msg_warn_immediate(
            paste0("Optimized computation of gradient failed! Switching gradient type.\n",
                   "Message: ", conditionMessage(e)))
          model$params$gradientStruct$useFDGradient <<- TRUE
          model$params$gradientStruct$isNonLinear  <<- TRUE
        })
        testSimpleGradient <- FALSE
      }

      lastQuad <- P$quad

      logLikNew  <- P$obsLL
      deltaLL    <- logLikNew - logLikOld
      relDeltaLL <- if (is.finite(logLikOld)) deltaLL / abs(logLikOld) else Inf
      loglik.history <- c(loglik.history, logLikNew)

      updateStatusLog(iterations, mode, logLikNew, deltaLL, relDeltaLL, verbose)

      history <- appendLmsIterationHistory(
        history         = history,
        iteration       = iterations,
        loglik.history  = loglik.history,
        convergence.abs = convergence.abs,
        convergence.rel = convergence.rel,
        forecast.min.gains = ema.forecast.min.gains,
        forecast.window = ema.forecast.window,
        mode            = mode,
        loglik          = logLikNew,
        abs.change      = abs(deltaLL),
        rel.change      = abs(relDeltaLL),
        elapsed.time    = proc.time()[["elapsed"]] - history.start.time
      )

      converged <- (abs(deltaLL) < convergence.abs) ||
                   (abs(relDeltaLL) < convergence.rel)
      converged.em <- converged && mode == "EM"
      run.final.qn <- FALSE

      if (iterations >= max.iter) break
      if (algorithm == "EMA" && do.final.qn && ema.final.phase == "qn") {
        if (converged || ema.final.qn.rounds >= ema.final.qn.rounds.max) {
          ema.final.phase <- "em"
        } else {
          run.final.qn <- TRUE
        }

      } else if (converged.em) {
        if (algorithm == "EMA" && do.final.qn && ema.final.phase == "none") {
          ema.final.phase <- "qn"
          run.final.qn <- TRUE
        } else {
          break
        }
      }

      # If quadrature has just been recalculated, we should be less strict
      quad.err <- as.integer(recalcQuad) * P$quad.err
      if (deltaLL < -max(1e-8, quad.err, na.rm = TRUE)) {
        if (verbose) cat("\n")
        mod_msg_warn_immediate(
          sprintf("Loglikelihood decreased by %.2g!", deltaLL),
          "Increasing the number of nodes might help."
        )
      }

      if (mode != "EM") { # switch back to EM
        mode <- "EM"
        n.em.iter <- 0
      } else n.em.iter <- n.em.iter + 1L

      # EMA controller
      if (run.final.qn) {
        mode <- "QN"
        updateStatusLog(iterations, mode, logLikNew, deltaLL, relDeltaLL, verbose)

      } else if (ema.final.phase == "em") {
        mode <- "EM"

      } else if (algorithm == "EMA" && n.em.iter > ema.min.iter) {
        # Check if expected number of iterations is increasing
        change <- diff(utils::tail(history$expected.remaining.iterations, ema.min.iter))
       
        if (all(is.finite(change)) && mean(change) > ema.trend.min) {
          remaining <- utils::tail(history$expected.remaining.iterations, 1L)[[1L]]

          # Increase the max step in case remaining > 500 or 
          # QN/FS fails, and we fall back to an EM step
          tmp.max.step <- max(fallback.max.step.min,
                            fallback.max.step.mult * max.step,
                            na.rm = TRUE)

          if (remaining > fs.remaining.min && remaining <= fs.remaining.max) {
            mode <- "FS"
          } else if (remaining > qn.remaining.min && remaining <= qn.remaining.max) {
            mode <- "QN"
          }

          updateStatusLog(iterations, mode, logLikNew, deltaLL, relDeltaLL, verbose)
        }
      }

      .obsfn <- function(x) {
        ll <- obsLogLikLms(theta = x, model = model, P = P, sign = -1)
        if (!is.finite(ll)) .Machine$double.xmax^(1/10) else ll # optim doesn't handle Inf well
      }

      .obsgr <- function(x) {
        grad <- gradientObsLogLikLms(theta = x, model = model, P = P, sign = -1, epsilon = epsilon)
   
        non.finite <- !is.finite(grad)
        if (any(non.finite)) {
          # This might happen during a bad line search in optim()
          # as long as obsLogLikLms() is non finite as well, it should be ok,
          # as it will be rejected...
          ll <- obsLogLikLms(theta = x, model = model, P = P, sign = -1)
          mod_stopif(is.finite(ll),
            "Some gradient values are NaN, but objective is finite!"
          )

          dir <- sign(x - thetaOld)
          dir[!is.finite(dir)] <- 0

          grad[non.finite] <- dir[non.finite] * .Machine$double.xmax^(1/10)
        }

        grad
      }

      .obshs <- function(x, louis = TRUE) {
        computeFullIobs(theta = x, model = model, P = P, louis = louis)
      }

      .error <- function(e) {
        mod_msg_warn(mode, "mode failed! Message:", conditionMessage(e))

        mstepLms(
          model = model,
          P = P,
          theta = thetaOld,
          max.step = max(max.step, tmp.max.step, na.rm = TRUE),
          epsilon = epsilon,
          optimizer = optimizer,
          control = control,
          ...
        )
      }

      # accelerated step
      switch(mode,
        EM = {
          mstep <- mstepLms(
            model = model,
            P = P,
            theta = thetaOld,
            max.step = max(max.step, tmp.max.step, na.rm = TRUE),
            epsilon = epsilon,
            optimizer = optimizer,
            control = control,
            ...
          )

        },

        QN = {
          # Quasi Newton
          mstep <- tryCatch({
            qn.control <- list(maxit = max.iter.qn)
            if (!is.null(qn.factr)) qn.control$factr <- qn.factr
            if (!is.null(qn.pgtol)) qn.control$pgtol <- qn.pgtol

            if (ema.final.phase == "qn") {
              qn.control$maxit <- ema.final.qn.maxit
              if (!is.null(ema.final.qn.factr)) {
                qn.control$factr <- ema.final.qn.factr
              }
            }

            stats::optim(
              par = thetaOld,
              fn = .obsfn,
              gr = .obsgr,
              lower = theta.lower,
              upper = theta.upper,
              control = qn.control,
              method = "L-BFGS-B"
            )
          }, error = .error)

          if (ema.final.phase == "qn") {
            ema.final.qn.rounds <- ema.final.qn.rounds + 1L
          } else {
            max.iter.qn <- min(qn.maxit.max, qn.maxit.mult * max.iter.qn)
          }
        },

        FS = {
          # Fischer Scoring
          mstep <- tryCatch({
            fischerScoring(
              par = thetaOld,
              fn = .obsfn,
              gr = .obsgr,
              hs = .obshs,
              lower = theta.lower,
              upper = theta.upper,
              max.iter = max.iter.fs
            )
          }, error = .error)
        }

        )

      if (!is.null(mstep$par) &&
          length(mstep$par) == length(thetaOld) &&
          all(is.finite(mstep$par))) {
        thetaNew <- mstep$par
      } else {
        thetaNew <- thetaOld
      }
    } # while

    if (verbose) cat("\n")
    mod_warnif(iterations >= max.iter,
      "Maximum iterations reached!\n",
      "Consider tweaking these parameters:\n",
      formatParameters(
        convergence.abs, convergence.rel, algorithm,
        max.step, max.iter, nodes, adaptive.quad,
        adaptive.quad.tol, quad.range
      )
    )

    # final E-step
    P <- estepLms(
      model = model,
      theta = thetaNew,
      lastQuad = lastQuad,
      recalcQuad = FALSE,
      adaptive.quad.tol = adaptive.quad.tol,
      ...
    )

    fit <- finalizeModelEstimatesDA(
      model             = model,
      theta             = thetaNew,
      method            = "lms",
      data              = lapply(model$models, FUN = \(submodel) submodel$data),
      logLik            = P$obsLL,
      iterations        = iterations,
      converged         = iterations < max.iter,
      optimizer         = paste(algorithm, optimizer, sep = "-"),
      calc.se           = calc.se,
      FIM               = FIM,
      OFIM.hessian      = OFIM.hessian,
      EFIM.S            = EFIM.S,
      EFIM.parametric   = EFIM.parametric,
      robust.se         = robust.se,
      epsilon           = epsilon,
      cr1s              = cr1s,
      R.max             = R.max,
      verbose           = verbose,
      P                 = P,
      includeStartModel = TRUE,
      startModel        = model
    )
    fit$iteration.history <- history
    fit

  }, error = function(e) {
    if (verbose) cat("\n")
    mod_msg_warn(paste0(
      "Model estimation failed, returning starting values!\n",
      "Message: ", conditionMessage(e)
    ))

    P0 <- tryCatch(
      estepLms(model = model, theta = model$theta,
               lastQuad = NULL, recalcQuad = FALSE,
               adaptive.quad.tol = adaptive.quad.tol, ...),
      error = function(e2) NULL
    )
    ll0 <- if (!is.null(P0)) P0$obsLL else NA_real_

    fit <- finalizeModelEstimatesDA(
      model             = model,
      theta             = model$theta,
      method            = "lms",
      data              = lapply(model$models, FUN = \(submodel) submodel$data),
      logLik            = ll0,
      iterations        = 0L,
      converged         = FALSE,
      optimizer         = paste(algorithm, optimizer, sep = "-"),
      calc.se           = FALSE,
      FIM               = FIM,
      OFIM.hessian      = OFIM.hessian,
      EFIM.S            = EFIM.S,
      EFIM.parametric   = EFIM.parametric,
      robust.se         = robust.se,
      epsilon           = epsilon,
      cr1s              = cr1s,
      R.max             = R.max,
      verbose           = verbose,
      P                 = P0,
      includeStartModel = TRUE,
      startModel        = model
    )
    fit$iteration.history <- history
    fit
  })
}


fischerScoring <- function(par,
                           fn,
                           gr,
                           hs, ...,
                           lower = -Inf,
                           upper = Inf,
                           max.iter = 1L,
                           max.alpha = 1,
                           min.alpha = 1e-12,
                           abs.tol = 1e-4,
                           jitter.mult = sqrt(.Machine$double.eps)) {

  truncate <- function(x) {
    underflow <- x < lower
    overflow  <- x > upper 

    x[underflow] <- lower[underflow]
    x[overflow]  <- upper[overflow]

    x 
  }

  f0 <- fn(par, ...)

  for (i in seq_len(max.iter)) {
    g0 <- gr(par, ...)
    I  <- hs(par, ...)

    # conditioning & scaled jitter
    rc <- suppressWarnings(rcond(I))
    if (!is.finite(rc) || rc < 1e-10) {
      jj <- jitter.mult * max(1, max(diag(I), na.rm = TRUE))
      diag(I) <- diag(I) + jj
    }

    I <- 0.5 * (I + t(I))
    direction <- tryCatch(-solve(I, g0), error = function(e) NULL)

    if (is.null(direction)) {
      return(list(
        convergence = -1,
        par = par,
        objective = f0,
        iterations = i
      ))
    }
        
    alpha   <- max.alpha
    success <- FALSE

    while (alpha > min.alpha) {
      parTrial  <- truncate(par + alpha * direction)
      f1  <- suppressWarnings(fn(parTrial, ...))

      ok <- is.finite(f1) && f1 <= f0 
      if (ok) { success <- TRUE; break }
      alpha <- alpha / 2
    }

    # convergence
    if (success && abs(f1 - f0) <= abs.tol) {
      return(list(
        convergence = 0,
        iterations = i,
        par = parTrial,
        objective = f1
      ))

    } else if (success) {
      par <- parTrial
      f0 <- f1

    } else {
      return(list(
        convergence = 10,
        iterations = i,
        par = par,
        objective = f0
      ))
    }
  }

  list(
    convergence = 1,
    iterations  = i,
    par = par,
    objective = f0
  )
}

#' Summarize a parameter table from a \code{modsem} model.
#'
#' @param parTable A parameter table, typically obtained from a \code{\link{modsem}} model
#'  using \code{\link{parameter_estimates}} or \code{\link{standardized_estimates}}.
#'
#' @param scientific Logical, whether to print p-values in scientific notation.
#'
#' @param ci Logical, whether to include confidence intervals in the output.
#'
#' @param digits Integer, number of digits to round the estimates to
#'  (default is 3).
#'
#' @param loadings Logical, whether to include factor loadings in the output.
#'
#' @param regressions Logical, whether to include regression coefficients in the output.
#'
#' @param covariances Logical, whether to include covariance estimates in the output.
#'
#' @param intercepts Logical, whether to include intercepts in the output.
#'
#' @param variances Logical, whether to include variance estimates in the output.
#'
#' @param thresholds Logical, whether to include threshold estimates in the output.
#'
#' @return A summary object containing the parameter table and additional information.
#'
#' @examples
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#'   # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#' # Double centering approach
#' est_dca <- modsem(m1, oneInt)
#'
#' std <- standardized_estimates(est_dca, correction = TRUE)
#' summarize_partable(std)
#' @export
summarize_partable <- function(parTable,
                               scientific  = FALSE,
                               ci          = FALSE,
                               digits      = 3,
                               loadings    = TRUE,
                               regressions = TRUE,
                               covariances = TRUE,
                               intercepts  = TRUE,
                               variances   = TRUE,
                               thresholds  = TRUE) {
  if (!"label" %in% colnames(parTable)) parTable$label <- ""

  parTable <- rename(
    .X      = parTable,
    est.std = "est",
    se      = "std.error",
    pvalue  = "p.value",
    p       = "p.value",
    zvalue  = "z.value",
    z       = "z.value"
  )

  width.out <- getWidthPrintedParTable(
    parTable    = parTable,
    scientific  = scientific,
    ci          = ci,
    digits      = digits
  )

  info.names <- c("Number of model parameters",
                  "Number of latent variables",
                  "Number of observed variables")

  info.values <- c(nrow(parTable),
                   length(getLVs(parTable)),
                   length(getOVs(parTable)))

  out <- list(
    parTable    = parTable,
    scientific  = scientific,
    ci          = ci,
    digits      = digits,
    loadings    = loadings,
    regressions = regressions,
    covariances = covariances,
    intercepts  = intercepts,
    variances   = variances,
    thresholds  = thresholds,
    width.out   = width.out,
    info.names  = info.names,
    info.values = info.values
  )

  class(out) <- c("list", "modsem_partable_summary")

  out
}


#' @export
print.modsem_partable_summary <- function(x, ...) {
  printf("modsem (version %s)\n\n", PKG_INFO$version)

  pad <- "  "

  cat(allignLhsRhs(
    lhs = x$info.names,
    rhs = x$info.values,
    pad = pad,
    width.out = x$width.out
  ), "\n")

  printParTable(
    parTable    = x$parTable,
    scientific  = x$scientific,
    ci          = x$ci,
    digits      = x$digits,
    loadings    = x$loadings,
    regressions = x$regressions,
    covariances = x$covariances,
    intercepts  = x$intercepts,
    variances   = x$variances,
    thresholds  = x$thresholds
  )
}

devtools::load_all()

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + a * X:Z

  c := a
"

est <- modsem(m1, data = oneInt)
parTable <- standardized_estimates(est)
parTable <- standardized_estimates(est, correction = TRUE)
parTable <- standardized_estimates(est, correction = TRUE, std.errors = "delta")
summarize_partable(parTable)

standardized_estimates(est, correction = TRUE, std.errors = "monte.carlo", mc.reps=10000)

varXZ <- parTable[parTable$lhs == "XZ" & parTable$rhs == "XZ", "est"]
testthat::expect_true(varXZ > 1.02)
testthat::expect_equal(unname(calcVarParTable("Y", parTable)), 1)

a <- parTable[parTable$label == "a", "est"]
c <- parTable[parTable$label == "c", "est"]

testthat::expect_equal(a, c)

a <- parTable[parTable$label == "a", "est"]
c <- parTable[parTable$label == "c", "est"]

testthat::expect_equal(a, c)

m2 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z + X:X
"

est <- modsem(m2, data = oneInt)
parTable <- standardized_estimates(est, correction = TRUE)
parTable <- standardized_estimates(est, correction = TRUE, std.errors = "delta")
varXZ <- parTable[parTable$lhs == "XZ" & parTable$rhs == "XZ", "est"]
varXX <- parTable[parTable$lhs == "XX" & parTable$rhs == "XX", "est"]
testthat::expect_true(varXZ > 1.02)
testthat::expect_true(varXX > 1.7)
testthat::expect_equal(unname(calcVarParTable("Y", parTable)), 1)


m3 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + z1 + a * X:z1

  c := a
"

est <- modsem(m3, data = oneInt, method = "pind")
parTable <- standardized_estimates(est)
parTable <- standardized_estimates(est, correction = TRUE)
parTable <- standardized_estimates(est, correction = TRUE, std.errors = "delta")
summarize_partable(parTable)

varz1 <- parTable[parTable$lhs == "z1" & parTable$rhs == "z1", "est"]
varXz1 <- parTable[parTable$lhs == "Xz1" & parTable$rhs == "Xz1", "est"]
testthat::expect_true(varXz1 > 2)
testthat::expect_equal(varz1, 1)
testthat::expect_equal(unname(calcVarParTable("Y", parTable)), 1)


m4 <- "
# Outer Model
  X =~ x1 + x2 + x3
# Inner Model
  y1 ~ X + z1 + a * X:z1

  c := a
"

est <- modsem(m4, data = oneInt, method = "pind")
parTable <- standardized_estimates(est)
parTable <- standardized_estimates(est, correction = TRUE)
parTable <- standardized_estimates(est, correction = TRUE, std.errors = "delta")
summarize_partable(parTable)

varz1 <- parTable[parTable$lhs == "z1" & parTable$rhs == "z1", "est"]
varXz1 <- parTable[parTable$lhs == "Xz1" & parTable$rhs == "Xz1", "est"]
testthat::expect_true(varXz1 > 2)
testthat::expect_equal(varz1, 1)
testthat::expect_equal(unname(calcVarParTable("y1", parTable)), 1)


m5 <- "
  y1 ~ x1 + z1 + a * x1:z1
  c := a
"

est <- modsem(m5, data = oneInt, method = "dblcent")
parTable <- standardized_estimates(est)
parTable <- standardized_estimates(est, correction = TRUE)
parTable <- standardized_estimates(est, correction = TRUE, std.errors = "delta")
summarize_partable(parTable)

varz1 <- parTable[parTable$lhs == "z1" & parTable$rhs == "z1", "est"]
varx1z1 <- parTable[parTable$lhs == "x1z1" & parTable$rhs == "x1z1", "est"]
testthat::expect_true(varx1z1 > 1)
testthat::expect_equal(varz1, 1)
testthat::expect_equal(unname(calcVarParTable("y1", parTable)), 1)

PI_METHODS   = c("rca", "uca", "dblcent", "pind", "ca", "custom")
DA_METHODS   = c("lms", "qml")
MPLUS_METHOD = "mplus"


#' Estimate interaction effects in structural equation models (SEMs)
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param data dataframe
#'
#' @param method method to use:
#' \describe{
#'   \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
#'   \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
#'   \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
#'   \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
#'   \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
#'   \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
#'   \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
#'   \item{\code{"custom"}}{use parameters specified in the function call (passed to \code{lavaan}).}
#'   \item{\code{"mplus"}}{estimate model through \code{Mplus}.}
#' }
#'
#' @param ... arguments passed to other functions depending on the method (see \code{\link{modsem_pi}}, \code{\link{modsem_da}}, and \code{\link{modsem_mplus}})
#'
#' @return \code{modsem} object with class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, or \code{\link{modsem_mplus}}
#'
#' @export
#' @description
#' \code{modsem()} is a function for estimating interaction effects between latent variables
#' in structural equation models (SEMs).
#' Methods for estimating interaction effects in SEMs can basically be split into two frameworks:
#' \enumerate{
#'   \item Product Indicator (PI) based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"}, \code{"ca"}, \code{"pind"})
#'   \item Distributionally (DA) based approaches (\code{"lms"}, \code{"qml"}).
#' }
#'
#' For the product indicator-based approaches, \code{modsem()} is essentially a fancy wrapper for \code{lavaan::sem()} which generates the
#' necessary syntax and variables for the estimation of models with latent product indicators.
#'
#' The distributionally based approaches are implemented separately and are
#' not estimated using \code{lavaan::sem()}, but rather using custom functions (largely
#' written in \code{C++} for performance reasons). For greater control, it is advised that
#' you use one of the sub-functions (\link{modsem_pi}, \link{modsem_da}, \link{modsem_mplus}) directly,
#' as passing additional arguments to them via \code{modsem()} can lead to unexpected behavior.
#'
#' @examples
#' library(modsem)
#' # For more examples, check README and/or GitHub.
#' # One interaction
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#'
#' # Double centering approach
#' est1 <- modsem(m1, oneInt)
#' summary(est1)
#'
#' \dontrun{
#' # The Constrained Approach
#' est1_ca <- modsem(m1, oneInt, method = "ca")
#' summary(est1_ca)
#'
#' # LMS approach
#' est1_lms <- modsem(m1, oneInt, method = "lms")
#' summary(est1_lms)
#'
#' # QML approach
#' est1_qml <- modsem(m1, oneInt, method = "qml")
#' summary(est1_qml)
#' }
#'
#' # Theory Of Planned Behavior
#' tpb <- '
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC
#'   BEH ~ INT:PBC
#' '
#'
#' # Double centering approach
#' est_tpb <- modsem(tpb, data = TPB)
#' summary(est_tpb)
#'
#' \dontrun{
#' # The Constrained Approach
#' est_tpb_ca <- modsem(tpb, data = TPB, method = "ca")
#' summary(est_tpb_ca)
#'
#' # LMS approach
#' est_tpb_lms <- modsem(tpb, data = TPB, method = "lms", nodes = 32)
#' summary(est_tpb_lms)
#'
#' # QML approach
#' est_tpb_qml <- modsem(tpb, data = TPB, method = "qml")
#' summary(est_tpb_qml)
#' }
modsem <- function(model.syntax = NULL,
                   data = NULL,
                   method = "dblcent",
                   ...) {
  method <- tolower(method)

  if (is.null(model.syntax)) {
    mod_msg_stop("No model.syntax provided")
  } else if (!is.character(model.syntax)) {
    mod_msg_stop("The provided model syntax is not a string!")
  } else if (length(model.syntax) > 1) {
    mod_msg_stop("The provided model syntax is not of length 1")
  }

  if (is.null(data)) {
    mod_msg_stop("No data provided")
  } else {
    data <- as.data.frame(data)
  }

  if (method %in% PI_METHODS) {
    modsem_pi(model.syntax, data = data, method = method, ...)
  } else if (method %in% DA_METHODS) {
    modsem_da(model.syntax, data = data, method = method, ...)
  } else if (method == MPLUS_METHOD) {
    modsem_mplus(model.syntax, data = data, ...)
  } else {
    mod_msg_stop("Method not recognized")
  }
}

devtools::load_all()


m1 <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  Y ~ X + Z + X:Z
'


rthreshold <- \(k, offset = runif(1, min = -1, max = 1), sigma = 0.35) {
  t <- seq_len(k) - mean(seq_len(k)) + offset
  t <- t + runif(k, min = -sigma, max = sigma)
  c(-Inf, t, Inf)
}


cut_data <- function(data, k = 5, choose = NULL) {
  if (is.null(choose))
    choose <- colnames(data)

  standardize <- \(x) (x - mean(x)) / sd(x)

  for (var in choose) {
    x <- standardize(data[[var]])
    t <- rthreshold(k)
    data[[var]] <- cut(x, breaks = t, ordered_result = TRUE)
  }

  data
}


testthat::test_that("ordered LMS fits without conditions and returns stable interaction estimate", {
  choose <- colnames(oneInt)

  set.seed(2837290)
  data <- cut_data(oneInt, choose = choose)

  fit <- NULL
  testthat::expect_no_error({
    fit <- suppressWarnings(suppressMessages(
      modsem(
        m1,
        data,
        method = "lms",
        ordered = choose,
        mean.observed = FALSE,
        ordered.fixed.seed = FALSE,
        ordered.polyak.juditsky = TRUE,
        verbose = FALSE
      )
    ))
  })

  par_table <- parameter_estimates(fit)
  int_est <- par_table[
    par_table$lhs == "Y" &
      par_table$op == "~" &
      par_table$rhs == "X:Z",
    "est"
  ]

  testthat::expect_length(int_est, 1L)
  testthat::expect_gte(int_est, 0.440)
  testthat::expect_lte(int_est, 0.460)
})

devtools::load_all()

set.seed(123)
n <- 250
X <- rnorm(n, mean = 0, sd = sqrt(1.2))
Z <- rnorm(n, mean = 0, sd = sqrt(1.4))

Y1 <- 0.7 * X + 0.3 * Z + 0.5 * X * Z + rnorm(n, mean = 0, sd = sqrt(0.4))
Y2 <- 0.3 * X + 0.7 * Z + 0.1 * X * Z + rnorm(n, mean = 0, sd = sqrt(0.6))

x1 <-       X + 1.2 + rnorm(n, mean = 0, sd = sqrt(0.2))
x2 <- 0.8 * X + 0.9 + rnorm(n, mean = 0, sd = sqrt(0.2))

z1 <-       Z + 1.2 + rnorm(n, mean = 0, sd = sqrt(0.2))
z2 <- 0.8 * Z + 0.9 + rnorm(n, mean = 0, sd = sqrt(0.2))

y1 <-       Y1 + 1.2 + rnorm(n, mean = 0, sd = sqrt(0.2))
y2 <- 0.8 * Y1 + 0.9 + rnorm(n, mean = 0, sd = sqrt(0.2))

y3 <-       Y2 + 1.2 + rnorm(n, mean = 0, sd = sqrt(0.2))
y4 <- 0.8 * Y2 + 0.9 + rnorm(n, mean = 0, sd = sqrt(0.2))

data <- data.frame(x1, x2, z1, z2, y1, y2, y3, y4)

m_linear <- '
 X =~ x1 + x2
 Z =~ z1 + z2
 Y1 =~ y1 + y2
 Y2 =~ y3 + y4

 Y1 ~ X + Z
 Y2 ~ a * X + b * Z

 a < .33
 b > .721
 a > 0
 b < 1
'

lavaan::summary(lavaan::sem(m_linear, data))
testthat::expect_no_warning({
  lms_linear_obl <- modsem(m_linear, data = data, method = "lms")
  lms_linear_ort <- modsem(m_linear, data = data, method = "lms",
                           orthogonal.x = TRUE, orthogonal.y = TRUE)
  qml_linear_obl <- modsem(m_linear, data = data, method = "qml")
  qml_linear_ort <- modsem(m_linear, data = data, method = "qml",
                           orthogonal.x = TRUE, orthogonal.y = TRUE)
})

b.lower <- lms_linear_obl$model$params$bounds$lower
b.upper <- lms_linear_obl$model$params$bounds$upper

testthat::expect_equal(b.lower[["a"]], 0.000)
testthat::expect_equal(b.upper[["a"]], 0.330)
testthat::expect_equal(b.upper[["b"]], 1.000)
testthat::expect_equal(b.lower[["b"]], 0.721)

checkIter <- \(m, n) testthat::expect_equal(m$iterations, n)
checkIter(lms_linear_obl, 2L)
checkIter(lms_linear_ort, 2L)
checkIter(qml_linear_obl, 1L)
checkIter(qml_linear_ort, 1L)

m_nlinear <- '
 X =~ x1 + x2
 Z =~ z1 + z2
 Y1 =~ y1 + y2
 Y2 =~ y3 + y4

 Y1 ~ X + Z + X:Z
 Y2 ~ X + Z + X:Z
'

testthat::expect_no_warning({
  lms_nlinear_obl <- modsem(m_nlinear, data = data, method = "lms")
  lms_nlinear_ort <- modsem(m_nlinear, data = data, method = "lms",
                           orthogonal.x = TRUE, orthogonal.y = TRUE)
  qml_nlinear_obl <- modsem(m_nlinear, data = data, method = "qml")
  qml_nlinear_ort <- modsem(m_nlinear, data = data, method = "qml",
                           orthogonal.x = TRUE, orthogonal.y = TRUE)
})

devtools::load_all()


m1 <- '
# Outer Model
  X =~ x1 + lx2 * x2 + lx3 * x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z

# Constraints
  x1 ~~ vx1*x1
  x2 ~~ vx2*x2
  x3 ~~ vx3*x3

  proj_vx1 := 0.8
  proj_vx2 := (lx2 ^ 2) * 0.8
  proj_vx3 := (lx3 ^ 2) * 0.8

  vx1 == 1 - proj_vx1
  vx2 == 1 - proj_vx2
  vx3 == 1 - proj_vx3
'

lms.no.opt <- modsem(m1, oneInt, method = "lms",
                     optimize = FALSE,
                     standardize.data = TRUE,
                     mean.observed = FALSE,
                     convergence.abs = 1e-8)
lms.opt <- modsem(m1, oneInt, method = "lms",
                  optimize = TRUE,
                  standardize.data = TRUE,
                  mean.observed = FALSE,
                  convergence.abs = 1e-8)

testthat::expect_equal(coef(lms.no.opt),
                       coef(lms.opt), tol = 2e-4)

m2 <- '
# Outer Model
  X =~ x1 + lx2 * x2 + lx3 * x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z

# Constraints
  x1 ~~ vx1*x1
  x2 ~~ vx2*x2
  x3 ~~ vx3*x3

  X ~ proj_vx1 * 1
  Z ~ proj_vx2 * 1
  Y ~ proj_vx3 * 1

  x1 ~ mu_x1 * 1
  z1 ~ mu_z1 * 1
  y1 ~ mu_y1 * 1

  proj_vx1 == 0.8
  proj_vx2 == (lx2 ^ 2) * 0.8
  proj_vx3 == (lx3 ^ 2) * 0.8

  mu_x1 == -proj_vx1
  mu_z1 == -proj_vx2
  mu_y1 == -proj_vx3

  vx1 == 1 - proj_vx1
  vx2 == 1 - proj_vx2
  vx3 == 1 - proj_vx3
'

lms.no.opt <- modsem(m2, oneInt, method = "lms",
                     optimize = FALSE,
                     standardize.data = TRUE,
                     convergence.abs = 1e-8)
lms.opt <- modsem(m2, oneInt, method = "lms",
                  optimize = TRUE,
                  standardize.data = TRUE,
                  convergence.abs = 1e-8)

testthat::expect_equal(coef(lms.no.opt),
                       coef(lms.opt), tol = 3e-4)


m3 <- '
# Outer Model
  X =~ x1 + lx2 * x2 + lx3 * x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z

# Constraints
  x1 ~~ a*x1
  x2 ~~ aa*x2
  x3 ~~ aaa*x3

  X ~ aaa.a   * 1
  Z ~ aaa.aa  * 1
  Y ~ aaa.aaa * 1

  x1 ~ aaaaaa.a   * 1
  z1 ~ aaaaaa.aa  * 1
  y1 ~ aaaaaa.aaa * 1

  aaa.a   == 0.8
  aaa.aa  == (lx2 ^ 2) * 0.8
  aaa.aaa == (lx3 ^ 2) * 0.8

  aaaaaa.a   == -aaa.a
  aaaaaa.aa  == -aaa.aa
  aaaaaa.aaa == -aaa.aaa

  a   == 1 - aaa.a
  aa  == 1 - aaa.aa
  aaa == 1 - aaa.aaa
'

lms.no.opt <- modsem(m3, oneInt, method = "lms",
                     optimize = FALSE,
                     standardize.data = TRUE)
lms.opt <- modsem(m3, oneInt, method = "lms",
                  optimize = TRUE,
                  standardize.data = TRUE)

testthat::expect_equal(coef(lms.no.opt),
                       coef(lms.opt), tol = 3e-4)

devtools::load_all()

m0 <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z
'

set.seed(2934829)
oneInt2 <- oneInt
oneInt2$weights <- rnorm(NROW(oneInt2), mean = 1, sd = 0.2)
oneInt2$group   <- sample(2, NROW(oneInt2), replace = TRUE)

testthat::expect_no_condition({
  fit.lms.h0.std <- modsem(m0, oneInt2, method = "lms", sampling.weights = "weights", standardize.data = TRUE)
  fit.lms.h0 <- modsem(m0, oneInt2, method = "lms", sampling.weights = "weights")
  fit.qml.h0 <- modsem(m0, oneInt2, method = "qml", sampling.weights = "weights")
  fit.lms.h0.mg1 <- modsem(m0, oneInt2, method = "lms", group = "group", sampling.weights = "weights", sampling.weights.normalization = "total")
  fit.lms.h0.mg2 <- modsem(m0, oneInt2, method = "lms", group = "group", sampling.weights = "weights", sampling.weights.normalization = "group")
})

testthat::expect_true(fit.lms.h0$iterations == 2L)
testthat::expect_true(fit.qml.h0$iterations == 1L)
testthat::expect_true(fit.lms.h0.mg1$iterations == 2L)
testthat::expect_true(fit.lms.h0.mg2$iterations == 2L)

w.g1.1 <- fit.lms.h0.mg1$model$models[[1L]]$data$weights
w.g2.1 <- fit.lms.h0.mg1$model$models[[2L]]$data$weights
w.g1.2 <- fit.lms.h0.mg2$model$models[[1L]]$data$weights
w.g2.2 <- fit.lms.h0.mg2$model$models[[2L]]$data$weights

rsum <- \(..., digits = 5) round(sum(...), digits)
testthat::expect_true(rsum(w.g1.1) + rsum(w.g2.1) == length(w.g1.1) + length(w.g2.1))
testthat::expect_true(rsum(w.g1.1) != length(w.g1.1))
testthat::expect_true(rsum(w.g2.1) != length(w.g2.1))

testthat::expect_true(rsum(w.g1.2) + rsum(w.g2.2) == length(w.g1.2) + length(w.g2.2))
testthat::expect_true(rsum(w.g1.2) == length(w.g1.2))
testthat::expect_true(rsum(w.g2.2) == length(w.g2.2))

m1 <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
'

testthat::expect_no_condition({
  fit.lms.h1 <- modsem(m1, oneInt2, method = "lms", sampling.weights = "weights")
  fit.qml.h1 <- modsem(m1, oneInt2, method = "qml", sampling.weights = "weights")
})


testthat::expect_no_condition({
  bootstrap_modsem(fit.lms.h0, R = 5L, optimize = TRUE, type = "nonparametric")
  bootstrap_modsem(fit.qml.h0, R = 5L, optimize = TRUE, type = "nonparametric")
  bootstrap_modsem(fit.lms.h0, R = 5L, optimize = TRUE, type = "parametric")
  bootstrap_modsem(fit.qml.h0, R = 5L, optimize = TRUE, type = "parametric")
  bootstrap_modsem(fit.lms.h0.mg1, R = 5L, optimize = TRUE, type = "nonparametric")
  bootstrap_modsem(fit.lms.h0.mg1, R = 5L, optimize = TRUE, type = "parametric")
})


# Check missing sampling weights
oneInt3 <- oneInt2
oneInt3$weights[c(2, 3, 5, 18)] <- NA

testthat::expect_error(
  modsem(m1, oneInt3, method = "lms", sampling.weights = "weights"),
  regexp = ".*sampling.weights.*cannot have missing.*"
)

# Check negative sampling weights
oneInt3 <- oneInt2
oneInt3$weights[c(2, 3, 5, 18)] <- c(-1.2, -0.001, -0.2, -3.2)

testthat::expect_error(
  modsem(m1, oneInt3, method = "lms", sampling.weights = "weights"),
  regexp = ".*sampling.weights.*cannot have negative values.*"
)

devtools::load_all()

# no interaction
m1 <- '
level: 1
  fw =~ y1 + y2 + y3
  fw ~ x1 + x2 + x3
level: 2
  fb =~ y1 + y2 + y3
  fb ~ w1 + w2
'

fit <- modsem(model = m1, data = lavaan::Demo.twolevel, cluster = "cluster")

# with interaction
m2 <- '
level: 1
    X =~ x1 + x2 + x3
    Z =~ z1 + z2 + z3
    Y =~ y1 + y2 + y3

    Y ~ X + Z + X:Z

level: 2
    Y =~ y1 + y2 + y3
    Y ~ w1 + w2
'


# Generate data accoring to `m2`
set.seed(123)

# Parameters
n_clusters <- 100
n_per_cluster <- 20
n_total <- n_clusters * n_per_cluster

# Cluster-level predictors
w1 <- rnorm(n_clusters)
w2 <- rnorm(n_clusters)

# Level-2 latent Y
Y_b_lvl2 <- 0.7 * w1 + 0.5 * w2 + rnorm(n_clusters, sd = 0.5)

# Create base data frame
data <- data.frame(cluster = rep(1:n_clusters, each = n_per_cluster))

# Repeat cluster-level variables
data$w1 <- rep(w1, each = n_per_cluster)
data$w2 <- rep(w2, each = n_per_cluster)
data$Y_b_lvl2 <- rep(Y_b_lvl2, each = n_per_cluster)

# Level-1 latent variables
X <- rnorm(n_total)
Z <- rnorm(n_total)
XZ <- X * Z
Y_w <- 0.6 * X + 0.6 * Z + 0.5 * XZ + rnorm(n_total, sd = 0.5)
Y_total <- Y_w + data$Y_b_lvl2

# Generate cluster-level residuals for y1–y3 (random intercepts)
y_between_resid <- matrix(rnorm(n_clusters * 3, sd = 0.3), ncol = 3)  # residuals per cluster
colnames(y_between_resid) <- paste0("y_b_", 1:3)

# Expand residuals to full data
y_b_mat <- y_between_resid[data$cluster, ]

# Generate indicators of Y with both between and within residuals
Y_ind <- matrix(NA, n_total, 3)
colnames(Y_ind) <- paste0("y", 1:3)
for (j in 1:3) {
  within_resid <- rnorm(n_total, sd = 0.3)
  Y_ind[, j] <- Y_total + y_b_mat[, j] + within_resid
}

# Indicators for X and Z (no cluster-level component assumed here)
gen_indicators <- function(latent, sd = 0.3) {
  sapply(1:3, function(i) latent + rnorm(length(latent), sd = sd))
}
X_ind <- gen_indicators(X)
Z_ind <- gen_indicators(Z)
colnames(X_ind) <- paste0("x", 1:3)
colnames(Z_ind) <- paste0("z", 1:3)

# Combine all into final dataset
sim_data <- cbind(data, X_ind, Z_ind, Y_ind)

# Check structure
fit <- modsem(m2, data=sim_data, cluster="cluster")
testthat::expect_warning(summary(fit), regexp = "Comparative fit.*")
standardized_estimates(fit)
testthat::expect_error(standardized_estimates(fit, correction = TRUE),
                       regexp = "Correction of clustered .*not supported!")

# Test robust std.errors
mod <- '
# X1-3 are Level 1 variables
X1 =~ x1
X2 =~ x2
X3 =~ x3

# W1-2 are Level 2 variables
W1 =~ w1
W2 =~ w2

fw =~ y1 + y2 + y3
fw ~ X1 + X2 + X3 + W1 + W2
'

# Standard errors corrected for clustering
fit.rc <- modsem(mod, lavaan::Demo.twolevel, method = "lms", cluster = "cluster", robust.se = TRUE)
#> Regressions:
#>                  Estimate  Std.Error  z.value  P(>|z|)
#>   fw ~
#>     X1              0.517      0.027   19.224    0.000
#>     X2              0.384      0.029   13.388    0.000
#>     X3              0.174      0.026    6.677    0.000
#>     W1              0.207      0.089    2.331    0.020
#>     W2              0.158      0.075    2.111    0.035

# Standard errors not corrected for clustering
fit.r <- modsem(mod, lavaan::Demo.twolevel, method = "lms", robust.se = TRUE)
#> Regressions:
#>                  Estimate  Std.Error  z.value  P(>|z|)
#>   fw ~
#>     X1              0.517      0.029   18.068    0.000
#>     X2              0.384      0.028   13.520    0.000
#>     X3              0.174      0.028    6.286    0.000
#>     W1              0.207      0.031    6.737    0.000
#>     W2              0.158      0.029    5.455    0.000

fit.c <- modsem(mod, lavaan::Demo.twolevel, method = "lms", cluster = "cluster", robust.se = FALSE)

devtools::load_all()
m1 <- '
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3
Y ~ X + Z + X:X
'
methods <- c("rca", "ca", "dblcent", "lms", "qml")
ests <- vector("list", length(methods))
names(ests) <- methods


for (method in methods) {
  if (method == "lms") {
    ests[[method]] <- modsem(m1, data = oneInt, method = method,
                             convergence.abs = 1e-3)
  } else {
    ests[[method]] <- modsem(m1, data = oneInt, method = method)
  }
}

nlsemModel <- '
ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
CAREER =~ career1 + career2 + career3 + career4
SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
'

est_qml2 <- modsem(nlsemModel, data = jordan, method = "qml",
                   mean.observed = FALSE, OFIM.hessian = FALSE)
est_rca2 <- modsem(nlsemModel, data = jordan, method = "rca")
est_dblcent2 <- modsem(nlsemModel, data = jordan, method = "dblcent")

std <- standardized_estimates(est_qml2)
print(std)

x <- "ENJ:ENJ"
y <- "ENJ:SC"
covENJENJ_ENJSC <- std[std$lhs %in% c(x, y) &
                       std$rhs %in% c(x, y) &
                       std$rhs != std$lhs &
                       std$op == "~~", "est"]
testthat::expect_true(covENJENJ_ENJSC > 0.3) # check that covariances between product terms
                                           # are handled correctly
testthat::expect_equal(unname(calcVarParTable("CAREER", std)), 1)

# devtools::install_github("kss2k/modsem", ref = "higher-order-rcs")
# library(modsem)
devtools::load_all()
library(lavaan)
library(mvtnorm)


# Simulate data from two higher order variables X and Z
N <- 2000L
Sigma <- diag(c(1.2, 1.8))
Sigma[1, 2] <- Sigma[2, 1] <- 0.6
Mu <- rep(0, 2L)

set.seed(2394827)
XI <- rmvnorm(N, mean = Mu, sigma = Sigma)

disturbance <- \(eps) rnorm(N, mean = 0, sd = sqrt(eps))

# Second Order LVs
X <- XI[, 1]
Z <- XI[, 2]
Y <- 0.2 * X + 0.5 * Z + 0.5 * X * Z + disturbance(0.5)

# First Order LVs
X1 <- 1   * X + disturbance(0.2)
X2 <- 0.8 * X + disturbance(0.2)

Z1 <- 1   * Z + disturbance(0.2)
Z2 <- 0.8 * Z + disturbance(0.2)

# OV Indicators
x1 <- 1   * X1 + disturbance(0.2)
x2 <- 0.8 * X1 + disturbance(0.2)

x3 <- 1   * X2 + disturbance(0.2)
x4 <- 0.8 * X2 + disturbance(0.2)

z1 <- 1   * Z1 + disturbance(0.2)
z2 <- 0.8 * Z1 + disturbance(0.2)

z3 <- 1   * Z2 + disturbance(0.2)
z4 <- 0.8 * Z2 + disturbance(0.2)

Y1 <- 1   * Y + disturbance(0.2)
Y2 <- 0.8 * Y + disturbance(0.2)

y1 <- 1   * Y1 + disturbance(0.2)
y2 <- 0.8 * Y1 + disturbance(0.2)

y3 <- 1   * Y2 + disturbance(0.2)
y4 <- 0.8 * Y2 + disturbance(0.2)

data <- data.frame(x1, x2, x3, x4,
                   z1, z2, z3, z4,
                   y1, y2, y3, y4)


sem.syntax <- '
  LX1 =~ x1 + x2
  LX2 =~ x3 + x4

  LZ1 =~ z1 + z2
  LZ2 =~ z3 + z4

  LY1 =~ y1 + y2
  LY2 =~ y3 + y4

  LX =~ LX1 + LX2
  LZ =~ LZ1 + LZ2
  LY =~ LY1 + LY2

  LY ~ LX + LZ + LX:LZ
'

fit_lms_rcs <- modsem(sem.syntax, data, method = "lms", rcs = TRUE)
summary(fit_lms_rcs)
# n = 2000
#> Regressions:
#>                  Estimate  Std.Error  z.value  P(>|z|)
#>   LY ~
#>     LX              0.161      0.026    6.077    0.000
#>     LZ              0.487      0.023   21.349    0.000
#>     LX:LZ           0.491      0.018   27.764    0.000

runMplus <- tryCatch({
    MplusAutomation::detectMplus()
    TRUE
  },
  error = function(e) FALSE
)

if (runMplus) {
  fit_mplus <- modsem(sem.syntax, data = data, method = "mplus")
  summary(fit_mplus)
  #> Regressions:
  #>                  Estimate  Std.Error  z.value  P(>|z|)
  #>   LY ~
  #>     LX   (LY<-LX)   0.166      0.027    6.148    0.000
  #>     LZ   (LY<-LZ)   0.481      0.023   20.913    0.000
  #>     LX:LZ  (LY<-)   0.499      0.020   24.950    0.000
}

fit_lms <- modsem(sem.syntax, data = data, method = "lms")
standardized_estimates(fit_lms)

summary(fit_lms)
#> Regressions:
#>                  Estimate  Std.Error  z.value  P(>|z|)
#>   LY ~
#>     LX              0.166      0.027    6.144    0.000
#>     LZ              0.481      0.023   20.599    0.000
#>     LX:LZ           0.499      0.020   24.606    0.000

testthat::expect_error(modsem(sem.syntax, data = data, method = "qml"),
                       regexp = "Higher-order.*qml.*Try.*lms.*")

tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PB =~ pb1 + pb2 + pb3
  PC =~ pc1 + pc2 + pc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + SN
  PBC =~ PC + PB

  # Structural model
  BEH ~ PBC + INT + INT:PBC
'

est_lms2 <- modsem(tpb, data = TPB_2SO, method = "lms")
standardized_estimates(est_lms2)

# example 2
tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PBC =~ pbc1 + pbc2 + pbc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + PBC + SN

  # Structural model
  BEH ~ PBC + INT + INT:PBC
'

est_lms3 <- modsem(tpb, data = TPB_1SO, method = "lms", nodes = 32)
standardized_estimates(est_lms3)

devtools::load_all()

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + c * Z

  y1 ~~ 10000*y1
"

testthat::expect_warning(modsem(m1, oneInt, method = "lms", calc.se=FALSE),
                         regexp = "Some estimated ov variances.*")

m2 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ 0.0001 * y1 + y2 + y3

# Inner model
  Y ~ X + c * Z

  Y ~~ 10000*Y
"

testthat::expect_warning(modsem(m2, oneInt, method = "lms", calc.se=FALSE),
                         regexp = "Some estimated lv variances.*")

m3 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + c * Z
"

# we have to be a bit fancy in how we catch the errors here, since multiple
# are emitted...
quiet_modsem <- purrr::quietly(modsem)
results_lms <- quiet_modsem(m3, oneInt, method = "lms", auto.fix.first = FALSE)
pat1 <- "Standard errors for some .* could not be computed.*"
pat2 <- "The variance-covariance.*positive definite.*eigenvalue.*zero.*"
testthat::expect_true(any(grepl(pat1, results_lms$warnings)))
testthat::expect_true(any(grepl(pat2, results_lms$warnings)))

devtools::load_all()

tpb1 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:pbc1
'

testthat::expect_error(modsem(tpb1, data = TPB, method = "lms"),
                       regexp = "Element in product term .*pbc1.*")

tpb2 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC2
'

testthat::expect_error(modsem(tpb2, data = TPB, method = "lms"),
                       regexp = "Element in product term .*PBC2.*")

testthat::expect_error(modsem(tpb2, data = TPB, method = "dblcent"),
                       regexp = "Unknown element in product term: .*PBC2.*")

tpb3 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  INT:pbc1 ~ INT
'

testthat::expect_error(
  suppressWarnings(modsem(tpb3, data = TPB, method = "lms")),
  regexp = ".*Interaction terms.*dependent.*"
)

devtools::load_all()
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ 1.1*y1 + 1*y2 + y3

# Inner model
  Y ~ X + Z

  y2 ~ 1.3 * 1
"

est_lms <- modsem(m1, data = oneInt, method = "lms", calc.se=FALSE)
fit_mod <- fit_modsem_da(est_lms)

est_lavaan <- lavaan::sem(m1, data = oneInt, meanstructure = TRUE)
fit_lav <- as.list(lavaan::lavInspect(est_lavaan, "fit"))

expect_equal <- function(x, y, tolerance) {
  diff <- abs(x - y)
  ok <- diff <= tolerance
  cat(sprintf("x: %8.4f, y: %8.4f, y-x: %7.4f, y-x < tol: %s\n", x, y, y - x, ok))
  testthat::expect_equal(x, y, tolerance = tolerance)
}

compare_fits <- function(fit_mod, fit_lav) {
  expect_equal(fit_mod$chisq.value, fit_lav$chisq, tolerance = .2)
  expect_equal(fit_mod$chisq.df, fit_lav$df, tolerance = 0)
  expect_equal(fit_mod$RMSEA, fit_lav$rmsea, tolerance = .0001)
  expect_equal(fit_mod$RMSEA, fit_lav$rmsea, tolerance = .0001)
  expect_equal(fit_mod$RMSEA.pvalue, fit_lav$rmsea.pvalue, tolerance = .0001)
  expect_equal(fit_mod$RMSEA.lower, fit_lav$rmsea.ci.lower, tolerance = .0001)
  expect_equal(fit_mod$RMSEA.upper, fit_lav$rmsea.ci.upper, tolerance = .0001)
}

compare_fits(fit_mod, fit_lav)

m2 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ 1.1*y1 + 1*y2 + y3

# Inner model
  Y ~ X + Z
"
est_lms2 <- modsem(m2, data = oneInt, method = "lms", calc.se=FALSE,
                   mean.observed = FALSE)
est_lavaan2 <- lavaan::sem(m2, data = oneInt)

fit_lav2 <- as.list(lavaan::lavInspect(est_lavaan2, "fit"))
fit_mod2 <- fit_modsem_da(est_lms2)
compare_fits(fit_mod2, fit_lav2)

devtools::load_all()

get_se <- function(x) sqrt(diag(vcov(x)))

# example 1
tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PB =~ pb1 + pb2 + pb3
  PC =~ pc1 + pc2 + pc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + SN
  PBC =~ PC + PB

  # Higher order interaction
  INTxPBC =~ ATT:PC + ATT:PB + SN:PC + SN:PB

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

est_ca <- modsem(tpb, data = TPB_2SO, method = "ca")
testthat::expect_true(all(get_se(est_ca) < 0.15))

est_dblcent <- modsem(tpb, data = TPB_2SO, method = "dblcent")
testthat::expect_true(all(get_se(est_dblcent) < 0.15))

standardized_estimates(est_ca)
testthat::expect_error(
  standardized_estimates(est_ca, correction = TRUE, std.error = "delta"),
  regexp = "Correction of higher-order .*"
)

# example 2
tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PBC =~ pbc1 + pbc2 + pbc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + PBC + SN

  # Higher order interaction
  INTxPBC =~ ATT:PBC + SN:PBC + PBC:PBC

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

est_ca <- modsem(tpb, data = TPB_1SO, method = "ca")
testthat::expect_true(all(get_se(est_ca) < 0.1))

est_dblcent <- modsem(tpb, data = TPB_1SO, method = "dblcent")
testthat::expect_true(all(get_se(est_dblcent) < 0.1))


# example 3
tpb <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN  =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

  # Higher order constructs
  HIGH1 =~ PBC + ATT
  HIGH2 =~ INT + SN

  BEH ~ HIGH1 + HIGH2 + HIGH1:HIGH2
'


testthat::expect_error(modsem(tpb, TPB, method = "rca"),
                       regexp = "The `:` operator is not allowed for higher order *")

devtools::load_all()

library(mvtnorm)
set.seed(29347098)

var_X      <- 1
var_Z      <- 1
cov_X_Z    <- 0.2

gamma_Y_X  <- 0.3
gamma_Y_Z  <- 0.5
gamma_Y_XZ <- 0.3

zeta_Y     <- 0.8
lambda_1   <- 1
lambda_2   <- .7
lambda_3   <- .8

epsilon    <- 0.2
beta_1     <- 1.2
beta_2     <- 0.8
beta_3     <- 1.5
N          <- 2000

residual <- \(epsilon) rnorm(N, sd = sqrt(epsilon))
create_ind <- \(lv, beta, lambda, epsilon) beta + lambda * lv + residual(epsilon)

SXI <- matrix(c(var_X, cov_X_Z,
                cov_X_Z, var_Z), nrow = 2)
XI <- rmvnorm(N, sigma = SXI)

X <- XI[, 1]
Z <- XI[, 2]

Y <-
  gamma_Y_X * X +
  gamma_Y_Z * Z +
  gamma_Y_XZ * X * Z +
  residual(zeta_Y)

# Non-overlapping Indicators
x1 <- create_ind(X, beta_1, lambda_1, epsilon)
x3 <- create_ind(X, beta_3, lambda_3, epsilon)
z1 <- create_ind(Z, beta_1, lambda_1, epsilon)
z2 <- create_ind(Z, beta_2, lambda_2, epsilon)
z3 <- create_ind(Z, beta_3, lambda_3, epsilon)
y1 <- create_ind(Y, beta_1, lambda_1, epsilon)

# Overlapping Indicators
y2 <- beta_1 + lambda_2 * Y + lambda_3 * X + residual(epsilon)
y3 <- beta_2 + lambda_3 * Y + lambda_2 * X + residual(epsilon)
x2 <- beta_2 + lambda_2 * X + lambda_3 * Z + residual(epsilon)

data <- data.frame(
   x1, x2, x3,
   z1, z2, z3,
   y1, y2, y3
)

model <- '
  X =~ x1 + x2 + x3 + y2 + y3
  Z =~ z1 + z2 + z3 + x2
  Y =~ y1 + y2 + y3

  Y ~ X + Z + X:Z
'

testthat::expect_no_condition({
  fit <- modsem(model, data = data, method = "lms")
  summary(fit)
})

testthat::expect_error({
  modsem(model, data = data, method = "qml")
}, regexp = "The same indicator cannot .*")

devtools::load_all()

testthat::test_that("Residual covariances between xis and etas in the LMS approach", {
  set.seed(4201)
  n <- 2000L
  sigma <- matrix(c(1.00, 0.35, 0.20,
                    0.35, 0.50, 0.40,
                    0.20, 0.40, 0.60), nrow = 3L)
  latent <- matrix(stats::rnorm(n * 3L), ncol = 3L) %*% chol(sigma)
  x <- latent[, 1L]
  y <- x + latent[, 2L]
  z <- y + latent[, 3L]

  make_indicator <- function(v, loading) {
    loading * v + stats::rnorm(n, sd = sqrt(0.25))
  }

  data <- data.frame(
    x1 = make_indicator(x, 1.00),
    x2 = make_indicator(x, 0.85),
    x3 = make_indicator(x, 0.90),
    y1 = make_indicator(y, 1.00),
    y2 = make_indicator(y, 0.85),
    y3 = make_indicator(y, 0.90),
    z1 = make_indicator(z, 1.00),
    z2 = make_indicator(z, 0.85),
    z3 = make_indicator(z, 0.90)
  )

  syntax <- "
    X =~ x1 + x2 + x3
    Y =~ y1 + y2 + y3

    Y ~ 1*X
    X ~~ Y
  "

  fit.lav <- lavaan::sem(syntax, data, meanstructure = TRUE)
  fit.mod <- modsem(syntax, data, method = "lms", optimize = FALSE, convergence.abs = 1e-6)

  ebm.mod <- c(modsem_predict(fit.mod, method = "EBM"))
  ebm.lav <- c(lavaan::lavPredict(fit.lav, method = "EBM"))
  testthat::expect_equal(ebm.mod, ebm.lav, tol = 1e-4)

  ml.mod  <- c(modsem_predict(fit.mod, method = "ML"))
  ml.lav  <- c(lavaan::lavPredict(fit.lav, method = "ML"))
  testthat::expect_equal(ml.mod, ml.lav, tol = 1e-4)

  est.mod <- c(coef(fit.mod))
  est.lav <- lavaan::coef(fit.lav)[names(est.mod)]

  testthat::expect_equal(est.mod, est.lav, tol = 1e-4)

  syntax.int <- "
    X =~ x1 + x2 + x3
    Y =~ y1 + y2 + y3

    Y ~ 1*X + X:X
    X ~~ Y
  "

  y.int <- x + latent[,1L] + 0.2*x^2

  data.int <- data.frame(
    x1 = make_indicator(x, 1.00),
    x2 = make_indicator(x, 0.85),
    x3 = make_indicator(x, 0.90),
    y1 = make_indicator(y.int, 1.00),
    y2 = make_indicator(y.int, 0.85),
    y3 = make_indicator(y.int, 0.90)
  )

  fit.mod.int <- modsem(syntax.int, data, method = "lms")
  est.mod.int <- coef(fit.mod.int)
  testthat::expect_equal(
    est.mod.int[["X~~Y"]],
    est.mod[["X~~Y"]],
    tol = 1e-3
  )

  # Check that residual covariances work with cov.syntax models
  syntax.z.main <- "
    X =~ x1 + x2 + x3
    Y =~ y1 + y2 + y3
    Z =~ z1 + z2 + z3

    Z ~ 1*Y

    Y ~~ X
    Y ~~ Z
  "
  
  syntax.z.cov <- "
    Y ~ 1*X
  "

  fit.lav <- lavaan::sem(paste0(syntax.z.main, syntax.z.cov), data, meanstructure = TRUE)
  fit.mod <- modsem(syntax.z.main, data, method = "lms", cov.syntax = syntax.z.cov, optimize = FALSE, convergence.abs = 1e-8)

  est.mod <- c(coef(fit.mod))
  est.lav <- lavaan::coef(fit.lav)

  # reverse order of X~~Y
  names(est.lav)[names(est.lav)=="X~~Y"] <- "Y~~X"
  est.lav <- est.lav[names(est.mod)]

  testthat::expect_equal(est.mod, est.lav, tol = 1e-4)
})

devtools::load_all()


m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z + x1
'

testthat::expect_no_error(modsem(m1, oneInt, method = "lms"))


m3 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3 + jk

  # Inner model
  Y ~ X + Z + X:Z
'

testthat::expect_error(modsem(m3, oneInt, method = "lms"),
                       regexp = ".*Missing observed variables in data:.*jk.*")
testthat::expect_error(modsem(m3, oneInt, method = "qml"),
                       regexp = ".*Missing observed variables in data:.*jk.*")


m4 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3 + y1

  # Inner model
  Y ~ X + Z + X:Z
'

testthat::expect_error(modsem(m4, oneInt, method = "qml"),
                       regexp = ".*The same indicator cannot be used .* y1")


tpb_main <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ a1 * att1 + a2 * att2 + att3 + att4 + att5
  SN =~ s1 * sn1 + sn2
  PBC =~ p1 * pbc1 + pbc2 + pbc3
  INT =~ i1 * int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  ATT ~ SN
  BEH ~ INT + a * PBC
"

tpb_cov <- "
  INT ~ ATT + b * SN + c * PBC
"

testthat::expect_error(modsem(tpb_main, data = TPB, method = "lms",
                              calc.se=FALSE, cov.syntax = tpb_cov),
                       regexp = ".*All latent variables.*")

devtools::load_all()

set.seed(123)
n <- 200  # number of observations

# Step 1: Simulate latent variables
# We define latent variables: psc, SWE, decisions, work_engagement, irritation, stress, intention
# Means and variances can be adjusted as needed

latent_means <- rep(0, 6)  # [psc, SWE, decisions_not_influenced_latent, work_engagement, irritation, stress]
latent_cov <- matrix(c(
    1, 0.5,  0.4,   0,     0,   0,
  0.5,   1,  0.4,   0,     0,   0,
  0.4, 0.4,    1,   0,     0,   0,
    0,   0,    0,   1,   0.3, 0.3,
    0,   0,    0, 0.3,     1, 0.4,
    0,   0,    0, 0.3,   0.4,   1
), nrow = 6, byrow = TRUE)
lvs <- c("psc", "SWE", "dec", "eng", "irr", "strs")
dimnames(latent_cov) <- list(lvs, lvs)

latent_data <- as.data.frame(mvtnorm::rmvnorm(n = n, mean = latent_means, sigma = latent_cov))
names(latent_data) <- lvs

# Step 2: Simulate indicators with loadings and residual variances
# We assume loadings of ~0.8 and residuals to account for remaining variance

make_indicators <- function(latent, prefix, items, loading = 0.8) {
  residual_sd <- sqrt(1 - loading^2)
  indicators <- matrix(NA, nrow = n, ncol = items)
  for (i in 1:items) {
    indicators[, i] <- loading * latent + rnorm(n, 0, residual_sd)
  }
  indicators <- as.data.frame(indicators)
  colnames(indicators) <- paste0(prefix, "_", 1:items)
  return(indicators)
}

# Measurement model
PSC <- make_indicators(latent_data$psc, "PSC", 4)
SWE <- make_indicators(latent_data$SWE, "Innovation", 4)
ENG <- make_indicators(latent_data$eng, "engagement", 3)
STRS <- make_indicators(latent_data$strs, "stress", 3)
IRR <- make_indicators(latent_data$irr, "irritation", 3)

# Single-indicator latent: decisions and intention (use dummy for decisions)
decisions_not_influenced <- 0.8 * latent_data$dec + rnorm(n, 0, sqrt(1 - 0.8^2))
intention <- ifelse(latent_data$eng + rnorm(n) < 0, 0, 1)  # dummy outcome ~ work_engagement

# Step 3: Create dummy variable names matching your SEM model
colnames(PSC) <- c("PSC_address_problems", "PSC_nobody_intentional", "PSC_dare_risk", "PSC_skills_appreciated")
colnames(SWE) <- c("Innovation_push_through", "Innovation_inspire", "Innovation_innovation_potential", "Innovation_trust")
colnames(ENG) <- c("fit_energetic", "inspiring_work", "completely_absorbed")
colnames(STRS) <- c("too_little_time_tasks", "tasks_inadequately_fulfilled", "pointless_tasks")
colnames(IRR) <- c("thinking_about_work_issues_at_home", "irritable", "feeling_like_bundle_of_nerves")

# Final dataset
dummy_data <- data.frame(
  PSC,
  SWE,
  ENG,
  STRS,
  IRR,
  decisions_not_influenced = decisions_not_influenced,
  intention_to_change_leadership_binary = intention
)


# Job Demands-Resources (JDR) model for SEM
JDR_Model_modsem <- '
# Measurement model
  psc =~ PSC_address_problems + PSC_nobody_intentional + PSC_dare_risk + PSC_skills_appreciated  # social resource
  SWE =~ Innovation_push_through + Innovation_inspire + Innovation_innovation_potential + Innovation_trust  # personal resource
  work_engagement =~ fit_energetic + inspiring_work + completely_absorbed
  stress =~ too_little_time_tasks + tasks_inadequately_fulfilled + pointless_tasks
  irritation =~ thinking_about_work_issues_at_home + irritable + feeling_like_bundle_of_nerves  # Strain
  decisions_not_influenced_latent =~ decisions_not_influenced  # Job Demands
  intention_to_change_leadership_binary_latent =~ intention_to_change_leadership_binary  # Outcome

# Covariances observed
  Innovation_push_through ~~ Innovation_inspire
  inspiring_work ~~ completely_absorbed
  tasks_inadequately_fulfilled ~~ pointless_tasks

# Covariances
  psc ~~ SWE + decisions_not_influenced_latent
  SWE ~~ decisions_not_influenced_latent
  irritation ~~ work_engagement

# Structural model
  work_engagement ~ psc + SWE + decisions_not_influenced_latent
  irritation ~ psc + SWE + decisions_not_influenced_latent
  stress ~ work_engagement + irritation
  intention_to_change_leadership_binary_latent ~ work_engagement

# Interactions
  work_engagement ~ psc:decisions_not_influenced_latent
  work_engagement ~ SWE:decisions_not_influenced_latent
'


testthat::expect_error(
  modsem(model = JDR_Model_modsem, data = dummy_data, method = "lms"),
  regex = ".*zero residual variance.*"
)

testthat::expect_warning({
est_qml <- modsem(model = JDR_Model_modsem, data = dummy_data, method = "qml")
print(summary(est_qml))
}, regexp = ".*smallest eigenvalue.*")


JDR_Model_linear <- '
# Measurement model
  PSC =~ PSC_address_problems + PSC_nobody_intentional + PSC_dare_risk + PSC_skills_appreciated  # social resource
  SWE =~ Innovation_push_through + Innovation_inspire + Innovation_innovation_potential + Innovation_trust  # personal resource
  WE =~ fit_energetic + inspiring_work + completely_absorbed
  STRESS =~ too_little_time_tasks + tasks_inadequately_fulfilled + pointless_tasks
  IRRITATE =~ thinking_about_work_issues_at_home + irritable + feeling_like_bundle_of_nerves  # Strain
  NONINF =~ decisions_not_influenced  # Job Demands
  INT =~ intention_to_change_leadership_binary  # Outcome

# Structural model
  WE ~ PSC + SWE + NONINF
  IRRITATE ~ PSC + SWE + NONINF
  STRESS ~ WE + IRRITATE
  INT ~ WE
'

est_lin_lms <- modsem(JDR_Model_linear, dummy_data, "lms")
testthat::expect_equal(est_lin_lms$iterations, 2)
print(summary(est_lin_lms, H0 = FALSE))

printModsemPIHeader <- function(approach) {
  cat(paste0("modsem (version ", PKG_INFO$version, ", approach = ",
             approach, "):\n"))
}


#' summary for modsem objects
#'
#' @param object modsem object to summarized
#' @param H0 Should the baseline model be estimated, and used to produce
#'  comparative fit?
#' @param digits Number of digits for printed numerical values
#' @param scientific Should scientific format be used for p-values?
#' @param verbose Should messages be printed?
#' @param ... arguments passed to lavaan::summary()
#' @rdname summary
#' @export
summary.modsem_pi <- function(object,
                              H0 = is_interaction_model(object),
                              r.squared = TRUE,
                              adjusted.stat = FALSE,
                              digits = 3,
                              scientific = FALSE,
                              verbose = TRUE,
                              ...) {
  out         <- list()
  out$lavaan  <- lavaan::summary(extract_lavaan(object), ...)
  out$info    <- list(version = PKG_INFO$version, approach = attributes(object)$method)
  out$fit     <- tryCatch(lavaan::fitMeasures(extract_lavaan(object)), error = \(e) NA)
  out$N       <- lavaan::nobs(extract_lavaan(object))
  out$format  <- list(digits = digits, scientific = scientific, adjusted.stat = adjusted.stat)
  out$ngroups <- lavaan::lavInspect(extract_lavaan(object), what = "ngroups")

  # Check for interaction effect in the model
  parTable <- parameter_estimates(object)
  hasInteraction <- parTableHasInteraction(object$input$parTable)

  out$hasInteraction <- hasInteraction

  etas <- getEtas(parTable, isLV = FALSE)

  # If no interaction effect, skip H0, r.squared, and warn if H0 requested
  if (!hasInteraction) {
    mod_warnif(H0, "Comparative fit to H0 will not be calculated.")
    class(out) <- c("summary_modsem_pi", "list")
    return(out)
  }

  # Baseline (H0) model
  if (H0) tryCatch({
    if (verbose) cat("Estimating baseline model (H0)\n")
    est_h0 <- estimateH0Cached(object, ...)
    out$nullModel <- est_h0

    if (!is.null(est_h0)) {
      # Use compare_fit (modsem generic, which internally uses lavTestLRT/anova for modsem_pi)
      lrt <- compare_fit(object, est_h0)
      out$LRT <- lrt
      out$fitH0 <- tryCatch(lavaan::fitMeasures(extract_lavaan(est_h0)), error = \(e) NA)
    }
  }, error = \(e) mod_msg_warn(paste0("Baseline model could not be estimated: ", e$message)))

  # R-squared for latent endogenous variables only
  if (r.squared & length(etas)) tryCatch({
    r2All <- getRsqrPI(object)
    out$r.squared <- r2All[etas]

    if (H0 && !is.null(out$nullModel)) {
      r2AllH0 <- getRsqrPI(out$nullModel)

      out$r.squared.H0   <- r2AllH0[etas]
      out$r.squared.diff <- out$r.squared - out$r.squared.H0
    }
  }, error = \(e) mod_msg_warn(paste0("R-squared could not be computed: ", e$message)))

  class(out) <- c("summary_modsem_pi", "list")
  out
}


#' @export
print.summary_modsem_pi <- function(x, ...) {
  digits <- x$format$digits
  scientific <- x$format$scientific
  adjusted.stat <- x$format$adjusted.stat

  # Compute width for right justification based on lavaan output
  # lavaan always seems to use a width of 54 characters for the main
  # part of the first header, regardless of the widht of the coefficient
  # tables... So we don't need to compute it dynamically??
  # lavcat <- utils::capture.output(print(x$lavaan))
  # ncheck <- 10 # Check only the header
  # width.out <- max(nchar(lavcat[seq_len(ncheck)]))
  width.out <- 54

  # Helper for left/right align with indent of 2 spaces for names
  align_lavaan <- function(lhs, rhs, width, indent = 2) {
    lhs <- sprintf("%*s%s", indent, "", lhs)
    pad <- width - nchar(lhs) - nchar(rhs)
    pad <- ifelse(pad < 1, 1, pad)
    paste0(lhs, strrep(" ", pad), rhs)
  }

  printModsemPIHeader(x$info$approach)

  # Only print detailed output if there is an interaction effect
  if (!is.null(x$hasInteraction) && !x$hasInteraction) {
    # The warning is already given in summary, so just return after header and lavaan print.
    print(x$lavaan)
    return(invisible(x))
  }

  # Interaction Model Fit Measures (H1)
  cat("\nInteraction Model Fit Measures (H1):\n")
  fit <- x$fit
  namesH1 <- c("Loglikelihood",
               "Akaike (AIC)",
               "Bayesian (BIC)",
               "Chi-square",
               "Degrees of Freedom",
               "P-value (Chi-square)",
               "RMSEA",
               "CFI",
               "SRMR")

  valuesH1 <- c(
    tryFormatC(fit["logl"], digits = 2, format = "f"),
    tryFormatC(fit["aic"], digits = 2, format = "f"),
    tryFormatC(fit["bic"], digits = 2, format = "f"),
    tryFormatC(fit["chisq"], digits = 2, format = "f"),
    ifelse(is.na(fit["df"]), yes = "NA", no = as.character(fit["df"])),
    tryFormatC(fit["pvalue"], digits = digits, format = if (scientific) "e" else "f"),
    tryFormatC(fit["rmsea"], digits = 3, format = "f"),
    tryFormatC(fit["cfi"], digits = 3, format = "f"),
    tryFormatC(fit["srmr"], digits = 3, format = "f")
  )
  for(i in seq_along(namesH1)) {
    cat(align_lavaan(namesH1[i], valuesH1[i], width.out), "\n")
  }
  cat("\n")

  # Baseline Model (H0)
  if (!is.null(x$fitH0)) {
    cat("Fit Measures for Baseline Model (H0):\n")
    fitH0 <- x$fitH0
    valuesH0 <- c(
      tryFormatC(fitH0["logl"], digits = 2, format = "f"),
      tryFormatC(fitH0["aic"], digits = 2, format = "f"),
      tryFormatC(fitH0["bic"], digits = 2, format = "f"),
      tryFormatC(fitH0["chisq"], digits = 2, format = "f"),
      ifelse(is.na(fitH0["df"]), yes = "NA", no = as.character(fitH0["df"])),
      tryFormatC(fitH0["pvalue"], digits = digits,
              format = if (scientific) "e" else "f"),
      tryFormatC(fitH0["rmsea"], digits = 3, format = "f"),
      tryFormatC(fitH0["cfi"], digits = 3, format = "f"),
      tryFormatC(fitH0["srmr"], digits = 3, format = "f")
    )
    for(i in seq_along(namesH1)) {
      cat(align_lavaan(namesH1[i], valuesH0[i], width.out), "\n")
    }
    cat("\n")
  }

  # Comparative Fit (LRT)
  if (!is.null(x$LRT)) {
    cat("Comparative Fit to H0 (LRT test):\n")
    lrt <- x$LRT
    if (nrow(lrt) > 1) {
      lrt_row <- lrt[2, ]
      cat(align_lavaan("Chi-square diff",
                       tryFormatC(lrt_row[["Chisq diff"]], digits = 3, format = "f"),
                       width.out), "\n")
      cat(align_lavaan("Degrees of freedom diff",
                       as.character(lrt_row[["Df diff"]]), width.out), "\n")
      cat(align_lavaan("P-value (LRT)",
                       tryFormatC(lrt_row[["Pr(>Chisq)"]], digits = digits,
                               format = if (scientific) "e" else "f"),
                       width.out), "\n")
    }
    cat("\n")
  }

  # R-Squared
  if (!is.null(x$r.squared)) {
    k <- x$ngroups
    if (k > 1) printH <- \(...) printf("Total R-Squared %s (%s):\n", ...)
    else       printH <- \(...) printf("R-Squared %s (%s):\n", ...)

    printH("Interaction Model", "H1")
    for (i in seq_along(x$r.squared)) {
      cat(align_lavaan(names(x$r.squared)[i],
                       tryFormatC(x$r.squared[i], digits = 3, format = "f"),
                       width.out), "\n")
    }

    if (!is.null(x$r.squared.H0)) {
      printH("Baseline Model", "H0")
      for (i in seq_along(x$r.squared.H0)) {
        cat(align_lavaan(names(x$r.squared.H0)[i],
                         tryFormatC(x$r.squared.H0[i], digits = 3, format = "f"),
                         width.out), "\n")
      }

      printH("Change", "H1 - H0")
      for (i in seq_along(x$r.squared.diff)) {
        cat(align_lavaan(names(x$r.squared.diff)[i],
                         tryFormatC(x$r.squared.diff[i], digits = 3, format = "f"),
                         width.out), "\n")
      }
    }
    cat("\n")
  }

  # lavaan
  print(x$lavaan)
}



#' @describeIn parameter_estimates Get parameter estimates of a \code{\link{modsem_pi}} object
#' @param colon.pi Should colons (\code{:}) be added to the interaction terms (E.g., `XZ` -> `X:Z`)?
#' @export
parameter_estimates.modsem_pi <- function(object, colon.pi = FALSE,
                                          label.renamed.prod = FALSE,
                                          high.order.as.measr = NULL, # capture argument
                                          rm.tmp.ov = NULL, # capture argument
                                          is.public = NULL, # capture argument
                                          ...) {
  parTable <- lavaan::parameterEstimates(object$lavaan, ...)

  if (colon.pi) {
    elems <- object$elementsInProdNames

    if (length(elems) && !"label" %in% colnames(parTable))
      parTable$label <- ""

    if (label.renamed.prod)
      origLabels <- getParTableLabels(parTable, labelCol = "label")
    else
      origLabels <- parTable$label

    for (xz in names(elems)) {
      xzColon <- paste0(elems[[xz]], collapse = ":")
      rmatch <- parTable$rhs == xz
      lmatch <- parTable$lhs == xz

      parTable[rmatch | lmatch, "label"] <- origLabels[rmatch | lmatch]

      parTable[rmatch, "rhs"] <- xzColon
      parTable[lmatch, "lhs"] <- xzColon # shouldn't be necessary, but just in case
                                         # the user has done something weird...
    }
  }

  parTable
}


#' @describeIn standardized_estimates Method for \code{modsem_pi} objects
#'
#' @param correction Logical. Whether to apply a correction to the standardized estimates
#' of the interaction term. By default, \code{FALSE}, which standardizes the interaction term
#' such that \eqn{\sigma^2(XZ) = 1}, consistent with \code{lavaan}'s treatment of latent interactions.
#' This is usually wrong, as it does not account for the fact that the interaction term
#' is a product of two variables, which means that the variance of the interaction term
#' of standardized variables (usually) is not equal to 1.
#'
#' Hence the scale of the interaction effect changes, such that
#' the standardized interaction term does not correspond to one (standardized) unit
#' change in the moderating variables. If \code{TRUE}, a correction is applied by
#' computing the interaction term \eqn{b_3 = \frac{B_3 \sigma(X) \sigma(Z)}{\sigma(Y)}} (where
#' \eqn{B_3} is the unstandardized coefficient for the interaction term), and solving for \eqn{\sigma(XZ)}, which
#' is used to correct the variance of the interaction term, and its covariances.
#'
#' @param std.errors Character string indicating the method used to compute standard errors
#' when \code{correction = TRUE}. Options include:
#' \describe{
#'   \item{"rescale"}{Simply rescales the standard errors. Fastest option.}
#'   \item{"delta"}{Uses the delta method to approximate standard errors.}
#'   \item{"monte.carlo"}{Uses Monte Carlo simulation to estimate standard errors.}
#' }
#' Ignored if \code{correction = FALSE}.
#'
#' @param mc.reps Integer. Number of Monte Carlo replications to use if \code{std.errors = "monte.carlo"}.
#'
#' @param colon.pi Logical. If \code{TRUE}, the interaction terms in the output will be
#' will be formatted using \code{:} to seperate the elements in the interaction term. Default
#' is \code{FALSE}, using the default formatting from \code{lavaan}. Only relevant if
#' \code{std.errors != "rescale"} and \code{correction = TRUE}.
#'
#' @param ... Additional arguments passed on to \code{lavaan::standardizedSolution()}.
#'
#' @export
standardized_estimates.modsem_pi <- function(object,
                                             correction = FALSE,
                                             std.errors = c("rescale", "delta", "monte.carlo"),
                                             mc.reps = 10000,
                                             colon.pi = FALSE,
                                             ...) {
  std.errors  <- tolower(std.errors)
  std.errors  <- match.arg(std.errors)
  monte.carlo <- std.errors == "monte.carlo"

  uncorrected <- function(object) {
    rename(lavaan::standardizedSolution(object$lavaan, ...), est.std = "est")
  }

  cleanOutputParTable <- function(parTable) {
    hasColon <- any(grepl(":", parTable$rhs))

    if (colon.pi && !hasColon) {
      elems <- object$elementsInProdNames

      if (length(elems) && !"label" %in% colnames(parTable))
        parTable$label <- ""

      origLabels <- getParTableLabels(parTable, labelCol = "label")

      for (xz in names(elems)) {
        xzColon <- paste0(elems[[xz]], collapse = ":")
        rmatch <- parTable$rhs == xz
        lmatch <- parTable$lhs == xz

        parTable[rmatch | lmatch, "label"] <- origLabels[rmatch | lmatch]

        parTable[rmatch, "rhs"] <- xzColon
        parTable[lmatch, "lhs"] <- xzColon # shouldn't be necessary, but just in case
                                           # the user has done something weird...
      }

    } else if (!colon.pi) {
      rm <- \(x) stringr::str_remove_all(x, ":")
      parTable$rhs <- rm(parTable$rhs)
      parTable$lhs <- rm(parTable$lhs)

    }

    parTable
  }

  if (correction && std.errors == "rescale") {
    parTable.std.naive <- uncorrected(object)

    correction <- function(object) {
      correctStdSolutionPI(
        object       = object,
        parTable.std = parTable.std.naive
      )
    }

  } else if (correction) {
    correction <- function(object) {

      solution <- standardizedSolutionCOEFS(
        object      = object,
        monte.carlo = monte.carlo,
        mc.reps     = mc.reps,
        center      = FALSE,
        ...
      )

      solution$parTable
    }

  } else return(cleanOutputParTable(uncorrected(object)))

  parTable.ustd <- parameter_estimates(object)

  hiorder <- isHigherOrderParTable(parTable.ustd)
  cluster <- isClustered(object)
  rescale <- std.errors == "rescale"

  mod_stopif(hiorder && !rescale, paste0('Correction of higher-order models are ',
         'not supported at all with `std.errors = "rescale"`!'))
  mod_stopif(cluster, "Correction of clustered (multilevel) models is not supported!")
  mod_warnif(hiorder && rescale, "Correction of higher-order models will likely not work!")

  parTable.std <- correction(object)

  cleanOutputParTable(parTable.std)
}


#' @export
#' @describeIn modsem_inspect Inspect a \code{\link{modsem_pi}} object
modsem_inspect.modsem_pi <- function(object, what = "free", ...) {
  lavaan::lavInspect(object$lavaan, what = what, ...)
}


#' @export
#' @importFrom stats vcov
vcov.modsem_pi <- function(object, ...) {
  lavaan::vcov(object$lavaan, ...)
}


#' @export
#' @importFrom stats coef
coef.modsem_pi <- function(object, ...) {
  lavaan::coef(object$lavaan, ...)
}


#' @export
#' @importFrom stats coefficients
coefficients.modsem_pi <- function(object, ...) {
  lavaan::coef(object$lavaan, ...)
}


#' @export
#' @importFrom stats nobs
nobs.modsem_pi <- function(object, ...) {
  lavaan::nobs(object$lavaan, ...)
}


#' @export
print.modsem_pi <- function(x, ...) {
  printModsemPIHeader(attributes(x)$method)
  print(x$lavaan)
}


#' @describeIn modsem_predict Wrapper for \code{lavaan::predict}
#' @export
modsem_predict.modsem_pi <- function(object, ...) {
  lavaan::predict(extract_lavaan(object), ...)
}


#' @export
is_interaction_model.modsem_pi <- function(object) {
  isTRUE(object$has.interaction)
}

#' oneInt
#'
#' @name oneInt
#' @docType data
#' @description A simulated dataset based on the elementary interaction model.
#' @examples
#'
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' "
#'
#' est <- modsem(m1, data = oneInt)
#' summary(est)
NULL


#' TPB
#'
#' @name TPB
#' @docType data
#' @description A simulated dataset based on the Theory of Planned Behaviour
#' @examples
#'
#' tpb <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC + INT:PBC
#' "
#'
#' est <- modsem(tpb, data = TPB)
#' summary(est)
NULL


#' TPB_1SO
#'
#' @name TPB_1SO
#' @docType data
#' @description A simulated dataset based on the Theory of Planned Behaviour,
#' where INT is a higher order construct of ATT, SN, and PBC.
#' @examples
#' tpb <- '
#'   # First order constructs
#'   ATT =~ att1 + att2 + att3
#'   SN  =~ sn1 + sn2 + sn3
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   BEH =~ b1 + b2
#'
#'   # Higher order constructs
#'   INT =~ ATT + PBC + SN
#'
#'   # Higher order interaction
#'   INTxPBC =~ ATT:PBC + SN:PBC + PBC:PBC
#'
#'   # Structural model
#'   BEH ~ PBC + INT + INTxPBC
#' '
#'
#' \dontrun{
#' est_ca <- modsem(tpb, data = TPB_1SO, method = "ca")
#' summary(est_ca)
#'
#' est_dblcent  <- modsem(tpb, data = TPB_1SO, method = "dblcent")
#' summary(est_dblcent)
#' }
NULL


#' TPB_2SO
#'
#' @name TPB_2SO
#' @docType data
#' @description A simulated dataset based on the Theory of Planned Behaviour,
#' where INT is a higher order construct of ATT and SN, and PBC is a higher order
#' construct of PC and PB.
#' @examples
#' tpb <- '
#'   # First order constructs
#'   ATT =~ att1 + att2 + att3
#'   SN  =~ sn1 + sn2 + sn3
#'   PB =~ pb1 + pb2 + pb3
#'   PC =~ pc1 + pc2 + pc3
#'   BEH =~ b1 + b2
#'
#'   # Higher order constructs
#'   INT =~ ATT + SN
#'   PBC =~ PC + PB
#'
#'   # Higher order interaction
#'   INTxPBC =~ ATT:PC + ATT:PB + SN:PC + SN:PB
#'
#'   # Structural model
#'   BEH ~ PBC + INT + INTxPBC
#' '
#'
#' \dontrun{
#' est <- modsem(tpb, data = TPB_2SO, method = "ca")
#' summary(est)
#' }
NULL


#' TPB_UK
#'
#' @name TPB_UK
#' @docType data
#' @description A dataset based on the Theory of Planned Behaviour from a
#' UK sample. 4 variables with high communality were selected for each
#' latent variable (ATT, SN, PBC, INT, BEH), from two time points (t1 and t2).
#'
#' @source
#' Gathered from a replication study by Hagger et al. (2023).
#'
#' Obtained from \doi{10.23668/psycharchives.12187}
#'
#' @examples
#'
#' tpb_uk <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'  ATT =~ att3 + att2 + att1 + att4
#'  SN =~ sn4 + sn2 + sn3 + sn1
#'  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
#'  INT =~ int2 + int1 + int3 + int4
#'  BEH =~ beh3 + beh2 + beh1 + beh4
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'  # Causal Relationsships
#'  INT ~ ATT + SN + PBC
#'  BEH ~ INT + PBC
#'  BEH ~ INT:PBC
#' "
#'
#' est <- modsem(tpb_uk, data = TPB_UK)
#' summary(est)
NULL


#' Jordan subset of PISA 2006 data
#'
#' @name jordan
#' @docType data
#' @description The data stem from the large-scale assessment study PISA 2006
#' (Organisation for Economic Co-Operation and Development, 2009) where
#' competencies of 15-year-old students in reading, mathematics, and science
#' are assessed using nationally representative samples in 3-year cycles.
#' In this example, data from the student background questionnaire from the
#' Jordan sample of PISA 2006 were used. Only data of students with complete
#' responses to all 15 items (N = 6,038) were considered.
#'
#' @format
#' A data frame of fifteen variables and 6,038 observations:
#'
#' \describe{
#' \item{enjoy1}{indicator for enjoyment of science, item ST16Q01: I generally have fun when I am learning <broad science> topics.}
#' \item{enjoy2}{indicator for enjoyment of science, item ST16Q02: I like reading about <broad science>.}
#' \item{enjoy3}{indicator for enjoyment of science, item ST16Q03: I am happy doing <broad science> problems.}
#' \item{enjoy4}{indicator for enjoyment of science, item ST16Q04: I enjoy acquiring new knowledge in <broad science>.}
#' \item{enjoy5}{indicator for enjoyment of science, item ST16Q05: I am interested in learning about <broad science>.}
#' \item{academic1}{indicator for academic self-concept in science, item ST37Q01: I can easily understand new ideas in <school science>.}
#' \item{academic2}{indicator for academic self-concept in science, item ST37Q02: Learning advanced <school science> topics would be easy for me.}
#' \item{academic3}{indicator for academic self-concept in science, item ST37Q03: I can usually give good answers to <test questions> on <school science> topics.}
#' \item{academic4}{indicator for academic self-concept in science, item ST37Q04: I learn <school science> topics quickly.}
#' \item{academic5}{indicator for academic self-concept in science, item ST37Q05: <School science> topics are easy for me.}
#' \item{academic6}{indicator for academic self-concept in science, item ST37Q06: When I am being taught <school science>, I can understand the concepts very well.}
#' \item{career1}{indicator for career aspirations in science, item ST29Q01: I would like to work in a career involving <broad science>.}
#' \item{career2}{indicator for career aspirations in science, item ST29Q02: I would like to study <broad science> after <secondary school>.}
#' \item{career3}{indicator for career aspirations in science, item ST29Q03: I would like to spend my life doing advanced <broad science>.}
#' \item{career4}{indicator for career aspirations in science, item ST29Q04: I would like to work on <broad science> projects as an adult.}
#' }
#'
#' @source
#' This version of the dataset, as well as the description was gathered from the
#' documentation of the 'nlsem' package (https://cran.r-project.org/package=nlsem),
#' where the only difference is that the names of the variables were changed
#'
#' Originally the dataset was gathered by the Organisation for Economic Co-Operation and Development (2009).
#' Pisa 2006: Science competencies for tomorrow's world (Tech. Rep.).
#' Paris, France. Obtained from: https://www.oecd.org/pisa/pisaproducts/database-pisa2006.htm
#'
#' @examples
#' \dontrun{
#' m1 <- "
#'   ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
#'   CAREER =~ career1 + career2 + career3 + career4
#'   SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
#'   CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
#' "
#'
#' est <- modsem(m1, data = jordan, method = "qml")
#' summary(est)
#' }
NULL

calcCovParTable <- function(x, y, parTable, measurement.model = FALSE, maxlen = 100) {
  mod_stopif(length(x) != length(y), "x and y must be the same length")

  if (measurement.model) parTable <- redefineMeasurementModel(parTable)
  parTable <- parTable[!parTable$op %in% c(":=", "~1", "=~"), ]
  parTable <- prepParTable(parTable[c("lhs", "op", "rhs", "est")], paramCol = "est")

  tracePathsNumericCpp(x = x, y = y, parTable = parTable)
}


calcVarParTable <- function(x, parTable, measurement.model = FALSE, maxlen = 100) {
  var <- calcCovParTable(x = x, y = x, parTable = parTable,
                         measurement.model = measurement.model, maxlen = maxlen)
  structure(var, names = x)
}


reverseIntTerm <- function(xz) {
  mod_stopif(length(xz) > 1, "xz must be a single string")
  stringr::str_c(rev(stringr::str_split_1(xz, ":")), collapse = ":")
}


getEtas <- function(parTable, isLV = FALSE, checkAny = TRUE) {
  lVs <- unique(parTable[parTable$op %in% c("<~", "=~"), "lhs"])
  cond.lhs <- parTable$op == "~"
  cond.rhs <- parTable$op %in% c("<~", "=~") & parTable$rhs %in% lVs

  if (isLV) cond.lhs <- cond.lhs & parTable$lhs %in% lVs

  etas.lhs <- parTable[cond.lhs, "lhs"]
  etas.rhs <- parTable[cond.rhs, "rhs"]

  etas <- unique(c(etas.rhs, etas.lhs))
  mod_stopif(checkAny && !length(etas), "No etas found")

  etas
}


getSortedEtas <- function(parTable, isLV = FALSE, checkAny = TRUE) {
  unsortedEtas <- getEtas(parTable, isLV = isLV, checkAny = checkAny)

  cond1 <- parTable$op == "~"
  cond2 <- parTable$op %in% c("=~", "<~") & parTable$rhs %in% unsortedEtas

  structExprs <- parTable[cond1, , drop = FALSE]
  measrExprs  <- parTable[cond2, , drop = FALSE]

  if (NROW(measrExprs)) {
    measr2struct <- measrExprs
    measr2struct$lhs <- measrExprs$rhs
    measr2struct$op  <- "~"
    measr2struct$rhs <- measrExprs$lhs

    structExprs <- rbind(structExprs, measr2struct)
  }

  sortedEtas  <- character(0L)

  while (length(sortedEtas) < length(unsortedEtas) && nrow(structExprs) > 0) {
    mod_stopif(all(unique(structExprs$lhs) %in% structExprs$rhs), "Model is non-recursive")

    for (i in seq_len(nrow(structExprs))) {
      if ((eta <- structExprs[i, "lhs"]) %in% structExprs$rhs) next

      sortedEtas  <- c(eta, sortedEtas)
      structExprs <- structExprs[structExprs$lhs != eta, , drop = FALSE]
      break
    }
  }

  if (!all(unsortedEtas %in% sortedEtas) ||
      length(sortedEtas) != length(unsortedEtas)) {
      warning("unable to sort etas")
      return(unsortedEtas)
  }

  sortedEtas
}


getXis <- function(parTable, etas = NULL, isLV = TRUE, checkAny = TRUE) {
  if (is.null(etas)) etas <- getEtas(parTable, isLV = isLV)
  # add all LVs which are not etas
  xis <- unique(parTable[parTable$op %in% c("<~", "=~") & !parTable$lhs %in% etas, "lhs"])

  if (!isLV) { # add any other variabels found in structural expressions
    xis <- unique(c(xis, parTable[parTable$op == "~" &
                                  !parTable$rhs %in% etas, "rhs"]))
  }

  xis <- xis[!grepl(":", xis)] # remove interaction terms

  mod_stopif(checkAny && !length(xis), "No xis found")
  xis
}


getProdNames <- function(parTable) {
  unique(parTable[grepl(":", parTable$rhs) &
         parTable$op %in% c("~", "=~", "<~"), "rhs"])
}


getLVs <- function(parTable) {
  unique(parTable[parTable$op %in% c("=~", "<~"), "lhs"])
}


getComposites <- function(parTable) {
  unique(parTable[parTable$op == "<~", "lhs"])
}


getOVs <- function(parTable = NULL, model.syntax = NULL) {
  if (!is.null(model.syntax)) parTable <- modsemify(model.syntax)
  mod_stopif(is.null(parTable), "Missing parTable")

  lVs    <- getLVs(parTable)
  select <- parTable$op %in% c("<~", "=~", "~", "~~")
  vars   <- unique(c(parTable$lhs[select], parTable$rhs[select]))

  vars[!vars %in% lVs & !grepl(":", vars)]
}


getStructVars <- function(parTable) {
  struct <- parTable[parTable$op == "~", , drop = FALSE]
  unique(c(struct$lhs, struct$rhs))
}


getStructOVs <- function(parTable) {
  intersect(getStructVars(parTable), getOVs(parTable))
}


getHigherOrderLVs <- function(parTable) {
  lVs                  <- getLVs(parTable)
  isHigherOrder        <- logical(length(lVs))
  names(isHigherOrder) <- lVs

  for (lV in lVs) {
    inds <- parTable[parTable$lhs == lV & parTable$op %in% c("<~", "=~"), "rhs"] |>
      stringr::str_split(pattern = ":") |> unlist()

    if (any(inds %in% lVs)) isHigherOrder[[lV]] <- TRUE
  }

  if (!any(isHigherOrder)) NULL else lVs[isHigherOrder]
}


isHigherOrderParTable <- function(parTable) {
  length(getHigherOrderLVs(parTable)) > 0
}


isClustered <- function(object) {
  length(modsem_inspect(object, what = "cluster")) > 0
}


getIndsLVs <- function(parTable, lVs, isOV = FALSE, ovs = NULL) {
  if (!length(lVs)) return(NULL)

  measr <- parTable[parTable$op %in% c("<~", "=~") & parTable$lhs %in% lVs, ]
  mod_stopif(!NROW(measr), paste0("No measurement expressions found, for", lVs))

  if (isOV) .f <- \(lV) measr[measr$lhs == lV & measr$rhs %in% ovs, "rhs"]
  else      .f <- \(lV) measr[measr$lhs == lV, "rhs"]

  lapplyNamed(lVs, FUN = .f, names = lVs)
}


getInds <- function(parTable, ...) {
  unique(unlist(getIndsLVs(parTable, lVs = getLVs(parTable)), ...))
}


getIndicators <- function(parTable, observed=TRUE) {
  indicators <- unique(parTable[!grepl(":", parTable$rhs) &
                                parTable$op %in% c("<~", "=~"), "rhs"])

  if (observed) indicators <- indicators[!indicators %in% getLVs(parTable)]
  indicators
}


getReflectiveIndicators <- function(parTable) {
  unique(unlist(parTable[parTable$op == "=~", "rhs"]))
}


getCompositeIndicators <- function(parTable) {
  unique(parTable[parTable$op == "<~", "rhs"])
}


getIntTermRows <- function(parTable) {
  structExprs <- parTable[parTable$op == "~", ]
  structExprs[grepl(":", structExprs$rhs), ]
}


getIntTerms <- function(parTable, col = "rhs") {
  structExprs <- parTable[parTable$op == "~", ]
  unique(structExprs[grepl(":", structExprs[[col]]), col])
}


checkLhsIntTerms <- function(parTable) {
  bad <- getIntTerms(parTable, col = "lhs")
  mod_stopif(length(bad),
    "Interaction terms can (currently) not be treated as dependent variables!",
    "Independent variable interaction terms:", paste0(bad, collapse = ", ")
  )
}


getVarsInts <- function(intTerms, removeColonNames = TRUE) {
  if (removeColonNames) names <- stringr::str_remove_all(intTerms$rhs, ":")
  else names <- intTerms$rhs
  lapplyNamed(intTerms$rhs, FUN = stringr::str_split_1, pattern = ":",
              names = names)
}


maxchar <- function(x) {
  max(nchar(x), na.rm = TRUE)
}


fillColsParTable <- function(parTable) {
  colNames <- c("lhs", "op", "rhs", "label", "group", "est",
                "std.error", "z.value", "p.value", "ci.lower", "ci.upper")
  parTable[colNames[!colNames %in% colnames(parTable)]] <- NA
  parTable #[colNames]
}


#  function for getting unique combinations of two values in x
getUniqueCombos <- function(x, match = FALSE) {
  # Base case, x is 1 length long and there are no unique combos
  if (length(x) <= 1) return(NULL)

  rest <- getUniqueCombos(x[-1], match = FALSE) # this is not a mistake!
  combos <- data.frame(V1 = rep(x[[1]], length(x) - 1),
                       V2 = x[-1])
  if (match) combos <- rbind(data.frame(V1 = x, V2 = x), combos)
  rbind(combos, rest)
}


lastRow <- function(x) {
  x[nrow(x), ]
}


lapplyMatrix <- function(X, FUN, FUN.VALUE, ...) {
  matrix(vapply(X, FUN.VALUE = FUN.VALUE, FUN = FUN, ...),
         nrow = length(FUN.VALUE), ncol = length(X), ...)
}


# Wrapper of lapply where elements are names based on names argument, by default names
# are based on X
lapplyNamed <- function(X, FUN, ..., names = X) {
  structure(lapply(X, FUN, ...), names = names)
}


lapplyDf <- function(df, FUN, ...) {
  structure(lapply(df, FUN, ...),
            names = names(df),
            row.names = seq_len(nrow(df)),
            class = "data.frame")
}


isModsemObject <- function(x) {
  inherits(x, c("modsem_pi", "modsem_da", "modsem_mplus"))
}


getIntercept <- function(x, parTable, col = "est") {
  if (length(x) > 1) mod_msg_stop("x must be a single string")

  intercept <- parTable[parTable$lhs == x & parTable$op == "~1", col]

  if (length(intercept) == 0) return(0)
  intercept
}


getIntercepts <- function(x, parTable) {
  out <- vapply(x, FUN.VALUE = numeric(1L), FUN = function(x_i)
                getIntercept(x_i, parTable = parTable))
  names(out) <- x
  out
}


getMean <- function(x, parTable) {
  mod_stopif(length(x) > 1, "x must be a single string")

  meanY <- getIntercept(x, parTable = parTable)
  gamma <- parTable[parTable$lhs == x & parTable$op == "~", , drop = FALSE]

  if (NROW(gamma) == 0) return(meanY)
  for (i in seq_len(NROW(gamma))) {
    meanX <- getMean(gamma[i, "rhs"], parTable = parTable)
    meanY <- meanY + gamma[i, "est"] * meanX
  }

  meanY
}


getMeans <- function(x, parTable) {
  out <- vapply(x, FUN.VALUE = numeric(1L), FUN = function(x_i)
                getMean(x_i, parTable = parTable))
  names(out) <- x
  out
}


centerInteractions <- function(parTable, center.means = TRUE) {
  parTable <- meanInteractions(parTable)
  rows <- getIntTermRows(parTable)

  for (i in seq_len(NROW(rows))) {
    Y <- rows[i, "lhs"]
    XZ <- unlist(stringr::str_split(rows[i, "rhs"], ":"))
    X <- XZ[[1]]
    Z <- XZ[[2]]

    meanX   <- getMean(X, parTable)
    meanZ   <- getMean(Z, parTable)
    gammaXZ <- rows[i, "est"]

    isStructY    <- parTable$lhs == Y & parTable$op == "~"
    isStructRowX <- parTable$rhs == X & isStructY
    isStructRowZ <- parTable$rhs == Z & isStructY

    if (any(isStructRowX)) {
      gammaX <- parTable[isStructRowX, "est"] + gammaXZ * meanZ
      parTable[isStructRowX, "est"] <- gammaX
    }

    if (any(isStructRowZ)) {
      gammaZ <- parTable[isStructRowZ, "est"] + gammaXZ * meanX
      parTable[isStructRowZ, "est"] <- gammaZ
    }
  }

  if (center.means) {
    innerVars <- unique(unlist(parTable[parTable$op == "~", c("rhs", "lhs")]))
    parTable[parTable$lhs %in% innerVars & parTable$op == "~1", "est"] <- 0
  }

  parTable
}


meanInteractions <- function(parTable, ignore.means = FALSE) {
  out <- NULL

  for (g in getGroupsParTable(parTable)) {
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    out <- rbind(out, meanInteractionsGroup(parTable.g,
                                            ignore.means = ignore.means,
                                            group = g))
  }

  rbind(out, getZeroGroupParTable(parTable))
}


meanInteractionsGroup <- function(parTable, ignore.means = FALSE, group = 0L) {
  intTerms <- unique(parTable[grepl(":", parTable$rhs), "rhs"])

  # remove existing
  parTable <- parTable[!(parTable$op == "~1" & parTable$lhs %in% intTerms),
                       , drop = FALSE]

  present  <- colnames(parTable)
  newcols  <- c("lhs", "op", "rhs", "label", "group", "est")
  newcols  <- intersect(newcols, present)
  fillcols <- setdiff(present, newcols)

  for (intTerm in intTerms) {
    XZ <- unlist(stringr::str_split(intTerm, ":"))
    X <- XZ[[1]]
    Z <- XZ[[2]]

    meanX <- if (!ignore.means) getMean(X, parTable) else 0
    meanZ <- if (!ignore.means) getMean(Z, parTable) else 0

    covXZ <- calcCovParTable(x = X, y = Z, parTable = parTable)
    meanXZ <- meanX * meanZ + covXZ

    newRow <- data.frame(lhs = intTerm, op = "~1", rhs = "",
                         label = "", est = meanXZ, group = group)

    # Get correct cols
    newRow <- newRow[newcols]
    newRow[fillcols] <- NA

    parTable <- rbind(parTable, newRow)
  }

  parTable
}


getWarningWrapper <- function(silent = FALSE) { # function factory
  if (silent) return(suppressWarnings)
  function(x) x
}


isRowInParTable <- function(row, pt, ignore = NULL) {
  if (!is.null(ignore)) {
    row <- row[!names(row) %in% ignore]
    pt  <- pt[!colnames(pt) %in% ignore]
  }

  for (i in seq_len(nrow(pt))) {
    x <- unlist(row)
    y <- unlist(pt[i, ])
    if (all(x == y)) return(TRUE)
  }

  return(FALSE)
}


rename <- function(.X, ...) {
  newNames <- list(...)
  oldNames <- names(newNames)

  for (old in oldNames) {
    names(.X)[names(.X) == old] <- newNames[[old]]
  }

  .X
}


printf <- function(...) {
  cat(sprintf(...))
  utils::flush.console()
}


clearConsoleLine <- function() {
  printf(paste0("\r", strrep(" ", getOption("width", default=0L)), "\r"))
}


getDiffTwoMax <- function(x) {
  if (length(x) < 2) return(NA)
  y <- sort(x, decreasing = TRUE)
  abs.diff <- y[[1]] - y[[2]]
  rel.diff  <- (y[[1]] - y[[2]]) / y[[2]]

  list(rel = rel.diff, abs = abs.diff)
}


stripColonsParTable <- function(parTable) {
  parTable$lhs <- stringr::str_remove_all(parTable$lhs, ":")
  parTable$rhs <- stringr::str_remove_all(parTable$rhs, ":")
  parTable
}


getMissingLabels <- function(parTable) {
  if (!"label" %in% colnames(parTable)) parTable$label <- ""
  missing <- is.na(parTable$label) | parTable$label == ""
  group   <- parTable$group

  labels <- sprintf("%s%s%s", parTable$lhs, parTable$op, parTable$rhs)

  if (any(group > 1L))
    labels[group > 1L] <- paste0(labels[group > 1L], ".g", group[group > 1L])

  parTable[missing, "label"] <- labels[missing]
  parTable
}


strRemovIfString <- function(x, pattern) {
  if (is.character(x)) stringr::str_remove_all(x, pattern=pattern) else x
}


getParTableLabels <- function(parTable, labelCol="label", replace.dup = FALSE) {
  if (!labelCol %in% colnames(parTable))
    parTable[[labelCol]] <- ""

  if (replace.dup) {
		labels <- unique(parTable[[labelCol]][parTable[[labelCol]] != ""])

		for (label in labels) {
			match <- parTable[[labelCol]] == label
			first <- which.max(match)
			parTable[match, labelCol] <- ""
			parTable[first, labelCol] <- label
		}
  }

  custom <- parTable$op == ":="
  parTable[custom, c("op", "rhs")] <- ""

  group      <- parTable$group
  new.labels <- paste0(parTable$lhs, parTable$op, parTable$rhs)

  if (!is.null(group))
    new.labels <- ifelse(group > 1L,
                         yes = paste0(new.labels, ".g", group),
                         no = new.labels)

  ifelse(parTable[[labelCol]] == "", yes = new.labels, no = parTable[[labelCol]])
}


CI_WIDTH <- qnorm(.975)


padCharMatrix <- function(X, n=1) {
  pad <- strrep(" ", n)
  matrix(paste0(pad, X), nrow = nrow(X), ncol = ncol(X),
         dimnames = dimnames(X))
}


timeExpr <- function(expr) {
  start <- Sys.time()
  out <- expr
  end <- Sys.time()
  print(end - start)
  invisible(out)
}


formatParameters <- function(...) {
  pad <- "  "

  args <- as.character(substitute(list(...)))[-1]
  names <- format(args, justify = "left")
  values <- format(c(...), justify = "right")

  paste0(sprintf("%s%s = %s\n", pad, names, values),
         collapse = "")
}


padString <- function(x, pad) {
  s <- stringr::str_replace_all(x, pattern = "\n", replacement = paste0("\n", pad))
  paste0(pad, s, "\n")
}


getCoefsRows <- function(y, x, parTable) {
  row <- parTable[parTable$lhs %in% y &
                  parTable$rhs %in% x &
                  parTable$op == "~", , drop = FALSE]
}


getCoefs <- function(y, x, parTable, col = "est") {
  coefRows <- getCoefsRows(y, x, parTable)
  coefs <- structure(coefRows[[col]], names = coefRows$rhs)
  coefs[x]
}


getCOEFS <- function(y, x, COEFS, parTable = NULL) {
  if (is.null(parTable)) {
    y <- stringr::str_replace_all(y, ":", OP_REPLACEMENTS[[":"]])
    x <- stringr::str_replace_all(x, ":", OP_REPLACEMENTS[[":"]])
    labels <- paste0(y, OP_REPLACEMENTS[["~"]], x)

  } else {
    rows   <- getCoefsRows(y, x, parTable)
    names  <- paste0(rows$lhs, rows$op, rows$rhs)
    labels <- structure(rows$label, names = names)
    labels <- labels[paste0(y, "~", x)]
  }

  if (length(labels) == 1) COEFS[[labels]] else COEFS[labels]
}


parTableToEnv <- function(parTable, prefix = "") {
  lhs <- parTable$lhs
  rhs <- parTable$rhs
  op <- stringr::str_replace_all(parTable$op, OP_REPLACEMENTS)

  names <- paste0(prefix, lhs, op, rhs)
  values <- structure(parTable$est, names = names)

  list2env(values)
}


getConstrExprs <- function(parTable, parTableCov) {
  parTable <- unique(rbind(parTable, parTableCov))
  rows <- sortConstrExprs(parTable)
  if (is.null(rows)) return(NULL)

  fixedParams <- unique(rows$lhs)
  thetaFixedParams <- vector("list", length(fixedParams))
  names(thetaFixedParams) <- fixedParams

  exprs <- lapply(rows$rhs, function(expr) parse(text = expr))
  list(fixedParams = fixedParams, thetaFixedParams = thetaFixedParams,
       exprs = exprs)
}


sortConstrExprs <- function(parTable) {
  rows <- parTable[parTable$op %in% CONSTRAINT_OPS, ]
  if (NROW(rows) == 0) return(NULL)

  labelled <- splitCLabels(parTable$mod[parTable$mod != ""])
  isConst  <- canBeNumeric(rows$rhs)

  rows <- rows[!(isConst & rows$op %in% BOUNDUARY_OPS), ] # not relevant

  if (!all(rows$lhs %in% labelled)) {
    mod_msg_stop(paste0("Unknown labels in constraints: ", rows$lhs[!rows$lhs %in% labelled]))

  } else if (length(unique(rows$lhs)) != length(rows$lhs)) {
    mod_msg_stop(
      paste0("Duplicated labels in constraints:\n",
             paste0(rows$lhs[duplicated(rows$lhs)], collapse = ", "))
    )

  } else if (any(rows$op %in% BOUNDUARY_OPS)) {
    mod_msg_stop("Dynamic constraints with ('<', '>') are not implemented yet!")
  }

  definedLabels <- labelled[!labelled %in% rows$lhs]
  subRows <- rows
  sortedRows <- data.frame(lhs = NULL, op = NULL, lhs = NULL, mod = NULL)
  while (NROW(subRows) > 0) {
    matchedAny <- FALSE
    for (i in seq_len(nrow(subRows))) {
      labels_i <- getVarsExpr(subRows[i, "rhs"])
      if (length(labels_i) == 0 || all(labels_i %in% definedLabels)) {
        matchedAny <- TRUE
        sortedRows <- rbind(sortedRows, subRows[i, ])
        definedLabels <- c(definedLabels, subRows[i, "lhs"])
        subRows <- subRows[-i, ]
        break
      }
    }

    mod_stopif(!matchedAny, paste0("Unkown labels in constraints: ",
           labels_i[!labels_i %in% definedLabels]))
  }

  if (NROW(sortedRows) != NROW(rows)) {
    mod_msg_warn(paste0("Something went wrong when sorting and parsing constraint-expressions ",
             "attempting to estimate model with unsorted expressions"))
    return(rows)
  }

  sortedRows
}


sortConstrExprsFinalPt <- function(parTable) {
  if (!NROW(parTable)) return(NULL)

  # wrapr for sortConstrExprs meant to be used on the final output parTable
  # not for the input parTable (e.g., mod -> label)
  constrExprs <- sortConstrExprs(rename(parTable, label="mod"))

  if (is.null(constrExprs)) NULL else rename(constrExprs, mod="label")
}


subsetByGrouping <- function(df, grouping = NULL) {
  if (is.null(grouping)) return(df)
  grouping <- grouping[intersect(names(grouping), colnames(df))]
  df[apply(df[names(grouping)], MARGIN = 1, FUN = \(x) all(x == grouping)), ]
}


GINV <- function(X) {
  structure(MASS::ginv(X), dimnames = dimnames(X))
}


cov2cor <- function(vcov) {
  if (is.null(vcov))
    return(NULL)

  sd <- sqrt(abs(diag(vcov))) # use `abs()`, in case some variances are negative

  D <- diag(1 / sd)
  structure(D %*% vcov %*% D, dimnames = dimnames(vcov))
}


leftJoin <- function(left, right, by = intersect(colnames(left), colnames(right))) {
  ol <- "__orig_order_right__"
  or <- "__orig_order_left__"

  left[[ol]] <- seq_len(nrow(left))
  right[[or]] <- seq_len(nrow(right))

  # left join
  joined <- merge(left, right, by = by, all.x = TRUE)

  # order
  ordered <- joined[order(joined[[ol]], joined[[or]]), ]

  # return
  ordered[!colnames(ordered) %in% c(ol, or)]
}


eraseConsoleLines <- function(n = 1L) {
  if (n < 1L || !interactive() || !isOnUnix())
    return(invisible()) # if we're not in interactive mode, or on UNIX
                        # there is not point

  seq <- paste0(rep("\033[1A\033[2K", n), collapse = "")
  # bring cursor to start of the line we landed on
  cat(seq, "\033[1G", sep = "")
}


tryFormatC <- function(..., .onfail = "NA") {
  tryCatch(formatC(...), error = \(e) .onfail)
}


is.diag <- function(M, zero.tol = .Machine$double.eps) {
  diag(NROW(M)) > zero.tol
}


is.invertible <- function(M) {
  tryCatch({
    solve(M)
    TRUE
  }, error = \(e) FALSE)
}


isOnUnix <- function(onFail = FALSE) {
  tryCatch(tolower(.Platform$OS.type) == "unix",
           error = \(e) onFail)
}


isNonCenteredParTable <- function(parTable, tol = 1e-10) {
  intTerms <- unique(parTable[grepl(":", parTable$rhs), "rhs"])
  intVars  <- unique(unlist(stringr::str_split(intTerms, pattern = ":")))

  parTableProto <- parTable[parTable$op %in% c("~", "~1"), , drop = FALSE]

  if (!"est" %in% colnames(parTableProto))
    parTableProto$est <- 1

  if (!"group" %in% colnames(parTableProto))
    parTableProto$group <- 1L

  for (g in getGroupsParTable(parTableProto)) {
    parTableProto.g <- parTableProto
    means <- getMeans(intVars, parTableProto.g)

    if (any(abs(means) > tol))
      return(TRUE)
  }

  FALSE
}


hasIntTermVariances <- function(parTable) {
  # To pass the function, we need (at least) variances for all of the
  # interaction terms. Optimally we also have covariances
  intTerms <- unique(parTable[grepl(":", parTable$rhs), "rhs"])

  hasVariance <- \(xz)
    any(parTable$lhs == xz & parTable$rhs == xz & parTable$op == "~~")

  all(vapply(intTerms, FUN.VALUE = logical(1L), FUN = hasVariance))
}


splitCLabels <- function(labels) {
  if (!length(labels)) return(character())

  unlist(lapply(labels, function(label) {
    label_trim <- trimws(label)
    if (grepl("^c\\s*\\(", label_trim) && grepl("\\)$", label_trim)) {
      inside <- substr(label_trim,
                       start = regexpr("\\(", label_trim, perl = TRUE) + 1L,
                       stop = nchar(label_trim) - 1L)
      tokens <- trimws(strsplit(inside, ",", fixed = FALSE)[[1]])
      tokens[tokens != ""]
    } else {
      label_trim
    }
  }), use.names = FALSE)
}


MAX <- function(x) {
  if (!length(x)) 0 else max(x, na.rm = TRUE)
}


addMissingGroups <- function(parTable) {
  if ("group" %in% colnames(parTable))
    return(parTable)

  parTable$group <- 1L
  parTable[parTable$op %in% CONSTRAINT_OPS, "group"] <- 0L

  parTable
}


addNamedNullField <- function(x, field) {
  x[field] <- stats::setNames(list(field = NULL), nm = field)
  x
}


std1 <- function(v) {
  mu    <- mean(v, na.rm = TRUE)
  sigma <- stats::sd(v, na.rm = TRUE)

  if (!is.finite(sigma) || sigma == 0)
    sigma <- 1

  (v - mu) / sigma
}

colsOut <- c("lhs", "op", "rhs", "est", "std.error",
              "z.value", "p.value", "ci.lower", "ci.upper")
header <- c("Variable", "op", "Variable", "Estimate",
            "Std.Error", "z.value", "P(>|z|)", "ci.lower", "ci.upper")


shorten <- \(x, n) abbreviate(x, minlength = n, strict = TRUE)


formatParTable <- function(parTable,
                           digits = 3,
                           scientific = FALSE,
                           ci = FALSE,
                           width.min = 14,
                           pad.res = TRUE,
                           extra.cols = NULL,
                           shorten.lhs.header = FALSE) {
  parTable    <- fillColsParTable(parTable)
  parTable.in <- parTable # unformatted partable

  isCustom <- parTable$op == ":="
  isVar    <- parTable$op == "~~" & parTable$lhs == parTable$rhs
  isCov    <- parTable$op == "~~" & parTable$lhs != parTable$rhs
  isReg    <- parTable$op == "~"
  isIntr   <- parTable$op == "~1" # Intercept
  isMeasrF <- parTable$op == "=~" # Measurement of factors
  isMeasrC <- parTable$op == "<~" # Measurement of composites
  isThrs   <- parTable$op == "|"

  parTable[isCustom, "label"] <- ""
  parTable[isThrs, "lhs"] <- paste0(parTable[isThrs, "lhs"], "|",
                                    parTable[isThrs, "rhs"])

  isDoubleCol <- isReg  | isCov | isMeasrF | isMeasrC # Should output have double columns?
  isSingleCol <- isIntr | isVar | isCustom | isThrs

  if (shorten.lhs.header) {
    lhsDouble <- paste(shorten(parTable$lhs[isDoubleCol], n = width - 2L),
                       parTable$op[isDoubleCol])
  } else {
    lhsDouble <- paste(parTable$lhs[isDoubleCol], parTable$op[isDoubleCol])
  }

  width <- max(width.min, maxchar(lhsDouble))

  lhsSingle <- pasteLabels(parTable$lhs[isSingleCol],
                           parTable$label[isSingleCol],
                           width = width - pad.res)
  rhsDouble <- pasteLabels(parTable$rhs[isDoubleCol],
                           parTable$label[isDoubleCol],
                           width = width - pad.res)

  parTable$p.value <- formatPval(parTable$p.value, scientific = scientific)


  ftext <- \(text, n = width - pad.res) format(text, width = n, justify = "left")

  parTable[isDoubleCol, "lhs"] <- ftext(lhsDouble)
  parTable[isDoubleCol, "rhs"] <- ftext(rhsDouble)
  parTable[isSingleCol, "lhs"] <- ftext(lhsSingle)
  parTable[isSingleCol, "rhs"] <- ""

  if (pad.res) {
    etas    <- getEtas(parTable.in, isLV = FALSE, checkAny = FALSE)
    inds    <- getReflectiveIndicators(parTable.in)
    resvars <- c(etas, inds)

    isResVarCovOrInt <- (
      (isVar | isCov | isIntr) &
      (parTable.in$rhs %in% resvars | parTable.in$lhs %in% resvars)
    )

    padVarCov <- ifelse(isResVarCovOrInt, yes = ".", no = " ")

    parTable$lhs <- paste0(padVarCov, parTable$lhs)
    parTable$rhs <- paste0(padVarCov, parTable$rhs)
  }

  if (!ci) {
    ci.pattern <- "^ci\\.(lower|upper)$"
    header  <- header[!grepl(ci.pattern, header)]
    colsOut <- colsOut[!grepl(ci.pattern, colsOut)]
  }

  if (!is.null(extra.cols)) {
    header  <- c(header, extra.cols)
    colsOut <- c(colsOut, extra.cols)
  }

  parTable      <- parTable[colsOut]
  names(header) <- colsOut

  valueCols <- setdiff(colnames(parTable), c("lhs", "op", "rhs"))

  for (col in valueCols) { # skip lhs, op and rhs
    if (is.numeric(parTable[[col]]))
      parTable[[col]] <- round(parTable[[col]], digits)

    maxWidth        <- maxchar(c(header[[col]], parTable[[col]]))
    parTable[[col]] <- formatNumeric(parTable[[col]], width = maxWidth,
                                     digits = digits, justify = "right",
                                     scientific=FALSE)

    parTable[[col]] <- stringr::str_replace_all(parTable[[col]], "NA", "  ")
    header[[col]]   <- format(header[[col]], width = maxWidth, justify = "right")
  }

  list(parTable = parTable, header = header)
}


printParTable <- function(parTable,
                          scientific = FALSE,
                          ci = FALSE,
                          digits = 3,
                          loadings = TRUE,
                          regressions = TRUE,
                          covariances = TRUE,
                          intercepts = TRUE,
                          variances = TRUE,
                          thresholds = TRUE,
                          custom = TRUE,
                          extra.cols = NULL,
                          padWidth = 1, # we already pad lhs/rhs with "."/" "
                          padWidthLhs = 2,
                          spacing = 2) {
  formatted <- formatParTable(parTable, digits = digits,
                              ci = ci, scientific = scientific,
                              extra.cols = extra.cols)
  fParTable <- formatted$parTable
  header    <- formatted$header
  lhs       <- unique(fParTable$lhs)

  fullPadWidth <- padWidth + padWidthLhs + maxchar(fParTable$rhs) - 1
  pad          <- stringr::str_dup(" ", fullPadWidth)
  space        <- stringr::str_dup(" ", spacing)

  formattedHeader <- paste0(
    pad, stringr::str_c(header[-(1:3)], collapse = space), "\n"
  )

  # Latent Variables
  parTableLoadings <- fParTable[fParTable$op == "=~", ]
  if (loadings && NROW(parTableLoadings) > 0) {
    cat("Latent Variables:\n", formattedHeader)

    printParTableDouble(
      parTableLoadings,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Composites
  parTableComposites <- fParTable[fParTable$op == "<~", ]
  if (loadings && NROW(parTableComposites) > 0) {
    cat("Composites:\n", formattedHeader)

    printParTableDouble(
      parTableComposites,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Regressions
  parTableRegressions <- fParTable[parTable$op == "~", ]
  if (regressions && NROW(parTableRegressions) > 0) {
    cat("Regressions:\n", formattedHeader)

    printParTableDouble(
      parTableRegressions,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Intercepts
  parTableIntercepts <- fParTable[parTable$op == "~1", ]
  if (intercepts && NROW(parTableIntercepts) > 0) {
    cat("Intercepts:\n", formattedHeader)

    printParTableSingle(
      parTableIntercepts,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Covariances
  parTableCovariances <- fParTable[parTable$op == "~~" & parTable$lhs != parTable$rhs, ]
  if (covariances && NROW(parTableCovariances) > 0) {
    cat("Covariances:\n", formattedHeader)

    printParTableDouble(
      parTableCovariances,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Thresholds
  parTableThresholds <- fParTable[parTable$op == "|", ]
  if (thresholds && NROW(parTableThresholds) > 0) {
    cat("Thresholds:\n", formattedHeader)

    printParTableSingle(
      parTableThresholds,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Variances
  parTableVariances <- fParTable[parTable$op == "~~" & parTable$lhs == parTable$rhs, ]
  if (variances && NROW(parTableVariances) > 0) {
    cat("Variances:\n", formattedHeader)

    printParTableSingle(
      parTableVariances,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }

  # Defined parameters
  parTableCustom <- fParTable[parTable$op == ":=", ]
  if (custom && NROW(parTableCustom) > 0) {
    cat("Defined Parameters:\n", formattedHeader)

    printParTableSingle(
      parTableCustom,
      padWidth = padWidth,
      padWidthLhs = padWidthLhs,
      spacing = spacing
    )

    cat("\n")
  }
}


printParTableDouble <- function(parTable, padWidth = 2, padWidthLhs = 2,
                                spacing = 2) {
  lhs <- unique(parTable$lhs)
  pad <- stringr::str_dup(" ", padWidth)
  for (l in lhs) {
    cat(paste0(pad, l), "\n")
    printRowsParTable(lhs = parTable[parTable$lhs == l, "rhs", drop = FALSE],
                      rhs = parTable[parTable$lhs == l, -(1:3), drop = FALSE],
                      padWidthLhs = padWidthLhs + padWidth,
                      spacing = spacing)
  }
}


printParTableSingle <- function(parTable, padWidth = 2, padWidthLhs = 2,
                                spacing = 2) {
  lhs <- unique(parTable$lhs)
  pad <- stringr::str_dup(" ", padWidth)

  printRowsParTable(lhs = parTable[ , "lhs", drop = FALSE],
                    rhs = parTable[ , -(1:3), drop = FALSE],
                    padWidthLhs = padWidthLhs + padWidth,
                    spacing = spacing)
}


printRowsParTable <- function(lhs, rhs, padWidthLhs = 2,
                              spacing = 2) {
  rhs    <- lapplyDf(rhs, FUN = format)
  padLhs <- stringr::str_dup(" ", padWidthLhs)
  space  <- stringr::str_dup(" ", spacing)

  out <- ""
  for (i in seq_len(nrow(rhs))) {
    fStrRhs <- stringr::str_c(rhs[i, ], collapse = space)
    out <- paste0(out, padLhs, lhs[i, 1], fStrRhs, "\n")
  }
  cat(out)
}


pasteLabels <- function(vars, labels, width = 14) {
  initWidthVar   <- round(width / 2)
  widthLabel     <- min(maxchar(labels), width - initWidthVar - 3L) # space + ()
  widthVar       <- width - widthLabel - 3L

  mod_warnif(widthVar + widthLabel + 3L != width, "Mismatching widhts!")

  pasted <- paste0(vars, " (", labels, ")")
  widths <- nchar(pasted)

  wide         <- widths > width
  vars[wide]   <- shorten(vars[wide], n = widthVar)
  labels[wide] <- shorten(labels[wide], n = widthLabel)

  labels[labels != ""] <- paste0("(", labels[labels != ""], ")")

  for (i in seq_along(vars)) {
    ncharVar   <- nchar(vars[[i]])
    ncharLabel <- nchar(labels[[i]])
    sep        <- stringr::str_dup(" ", width - ncharVar - ncharLabel)

    vars[[i]] <- paste0(vars[[i]], sep, labels[[i]])
  }
  vars
}


allignLhsRhs <- function(lhs, rhs, pad = "", width.out = 50, rhs.scaled = NULL) {
  if (length(lhs) != length(rhs)) {
    warning("lhs and rhs must have the same length")
    if (length(lhs) > length(rhs)) lhs <- rhs[seq_along(lhs)]
    else rhs <- lhs[seq_along(rhs)]
  }

  out <- ""
  width.out <- width.out - nchar(pad)

  if (!is.null(rhs.scaled))
    rhs.scaled <- format(rhs.scaled, justify = "right")

  for (i in seq_along(lhs)) {
    ncharLhs <- nchar(lhs[[i]])
    ncharRhs <- nchar(rhs[[i]])

    sep  <- stringr::str_dup(" ", max(0, width.out - ncharLhs - ncharRhs))
    line <- paste0(pad, lhs[[i]], sep, rhs[[i]])

    if (!is.null(rhs.scaled))
      line <- paste0(line, pad, rhs.scaled[[i]], "\n")
    else
      line <- paste0(line, "\n")

    out  <- paste0(out, line)
  }

  out
}


# this is really ugly, but it is the easiest way to get the width of the
# printed table without splitting the function into multiple functions
# in a messy way
getWidthPrintedParTable <- function(parTable,
                                    scientific  = FALSE,
                                    ci          = FALSE,
                                    digits      = 3,
                                    padWidth    = 2,
                                    padWidthLhs = 2,
                                    spacing     = 2,
                                    extra.cols = NULL,
                                    ...) { # add ... for backwards compatability,
                                           # such that passing old (unused) arguments
  formatted <- formatParTable(             # doesn't throw an error
    parTable,
    digits = digits,
    ci = ci,
    scientific = scientific,
    extra.cols = extra.cols,
  )

  fParTable <- formatted$parTable
  header    <- formatted$header
  lhs       <- unique(fParTable$lhs)

  pad   <- stringr::str_dup(" ", padWidth + padWidthLhs +
                            maxchar(fParTable$rhs) - 1)
  space <- stringr::str_dup(" ", spacing)

  fStrHeader <- stringr::str_c(header[-(1:3)], collapse = space)
  formattedHeader <- paste0(pad, fStrHeader)
  nchar(formattedHeader)
}


formatPval <- function(p, scientific = TRUE, digits = 3) {
  if (scientific) return(format.pval(p))
  format(round(p, digits = digits), nsmall = digits)
}

# Functions for tracing paths in SEMs, used both for calculating (co-)variances,
# as well as calculating formulas for (co-)variances.
prepParTable <- function(parTable, addCovPt = TRUE, maxlen = 100, paramCol = "mod",
                         removeColon = TRUE) {
  # Remove any potential ':' from the model
  if (removeColon) parTable <- lapplyDf(parTable, strRemovIfString, pattern = ":")
  parTable <- parTable[c("lhs", "op", "rhs", paramCol)]

  # get relevant variables
  structuralVars <- parTable[parTable$op == "~", c("lhs", "rhs")] |>
    unlist() |> c(getLVs(parTable)) |> unique()

  parTable <- parTable[(parTable$lhs %in% structuralVars &
                       parTable$rhs %in% structuralVars) |
                       parTable$op == "~~", ]
  # redefine higher order (i.e.,`=~` -> `~`)
  if (any(parTable$op == "=~")) parTable <- redefineMeasurementModel(parTable)

  # add missing labels
  if (paramCol %in% c("mod", "label")) {
    labels <- apply(parTable[parTable[[paramCol]] == "", c("lhs", "op", "rhs")],
                    MARGIN = 1, FUN = stringr::str_c, collapse = "")
    parTable[parTable[[paramCol]] == "", paramCol] <- labels
  }

  if (addCovPt) {
    reversedCovPaths <- parTable[parTable$lhs != parTable$rhs & parTable$op == "~~", ]
    colnames(reversedCovPaths) <- c("rhs", "op", "lhs", paramCol)
    parTable <- rbind(parTable, reversedCovPaths)
  }

  parTable
}


cleanTracedPaths <- function(paths) {
  squaredExprs <- purrr::map_chr(paths, function(x) {
     reps <- as.data.frame(table(sort(x)))
     dplyr::if_else(reps[[2]] > 1,
                    true = paste(reps[[1]], reps[[2]], sep = " ^ "),
                    false = reps[[1]]) |>
     stringr::str_c(collapse = " * ")
  })
  reps <- as.data.frame(table(squaredExprs))
  dplyr::if_else(reps[[2]] > 1,
                 true = paste(reps[[2]], reps[[1]], sep = " * "),
                 false = reps[[1]]) |>
  stringr::str_c(collapse = " + ")
}


addMissingCovariances <- function(pt, paramCol = "mod") {
  pt <- pt[!pt$op %in% c("=~", "~1"), ]

  xis  <- getXis(pt, checkAny = FALSE, isLV = FALSE)
  xis  <- xis[!grepl(":", xis)]

  if (length(xis)) {
    cxis <- getUniqueCombos(xis, match = TRUE)
    phi  <- data.frame(lhs = cxis[[1]], op = "~~", rhs = cxis[[2]], mod = "")
  } else phi <- NULL

  etas <- getEtas(pt, checkAny = FALSE, isLV = FALSE)
  if (length(etas)) {
    psi  <- data.frame(lhs = etas, op = "~~", rhs = etas, mod = "")
  } else psi <- NULL

  if (is.null(psi) && is.null(phi)) return(pt)

  covs <- rbind(psi, phi)
  for (i in seq_len(nrow(covs))) {
    row <- covs[i, , drop = FALSE]
    if (isRowInParTable(row = row, pt = pt, ignore = paramCol)) next
    pt <- rbind(pt, row)
  }

  pt
}


#' Estimate formulas for (co-)variance paths using Wright's path tracing rules
#'
#' @param pt A data frame with columns \code{lhs}, \code{op}, \code{rhs}, and \code{mod}, from \link{modsemify}
#' @param x Source variable
#' @param y Destination variable
#' @param parenthesis If \code{TRUE}, the output will be enclosed in parenthesis
#' @param missing.cov If \code{TRUE}, covariances missing from the model syntax will be added
#' @param measurement.model If \code{TRUE}, the function will use the measurement model
#' @param maxlen Maximum length of a path before aborting
#' @param paramCol The column name in \code{pt} that contains the parameter labels
#' @param ... Additional arguments passed to \link{trace_path}
#'
#' @return A string with the estimated path (simplified if possible)
#' @export
#' @description
#' This function estimates the path from \code{x} to \code{y} using the path tracing rules.
#' Note that it only works with structural parameters, so \code{"=~"} are ignored, unless
#' \code{measurement.model = TRUE}.
#' If you want to use the measurement model,
#' \code{"~"} should be in the \code{mod} column of \code{pt}.
#' @examples
#' library(modsem)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#' pt <- modsemify(m1)
#' trace_path(pt, x = "Y", y = "Y", missing.cov = TRUE) # variance of Y
trace_path <- function(pt,
                       x,
                       y,
                       parenthesis = TRUE,
                       missing.cov = FALSE,
                       measurement.model = TRUE,
                       maxlen = 100,
                       paramCol = "mod",
                       ...) {
  pt <- pt[pt$op != "~1", ]
  if (measurement.model) pt <- redefineMeasurementModel(pt)
  else                   pt <- pt[pt$op != "=~"]


  if (missing.cov)
    pt <- addMissingCovariances(pt, paramCol = paramCol)

  generateSyntax(x = x, y = y, pt = pt, maxlen = maxlen,
                 parenthesis = parenthesis, paramCol = paramCol, ...)
}


generateSyntax <- function(x, y, pt, maxlen = 100, parenthesis = TRUE, paramCol = "mod",
                           removeColon = TRUE, ...) {
  pt    <- prepParTable(pt, paramCol = paramCol, removeColon = removeColon, ...)
  paths <- tracePathsCharacterCpp(x = x, y = y, parTable = pt, maxlen = maxlen, paramCol = paramCol)

  clean <- \(path) if (length(path)) cleanTracedPaths(path) else NA_character_
  cleaned <- vapply(paths, FUN.VALUE = character(1L), FUN = clean)

  ifelse(!is.na(cleaned) & parenthesis, yes = paste0("(", cleaned, ")"), no = cleaned)
}


getCovEqExprs <- function(x, y, parTable, paramCol = "label", removeColon = FALSE, ...) {
  paths <- trace_path(x = x, y = y, pt = parTable, paramCol = paramCol,
                      removeColon = removeColon, ...)

  lapply(paths, FUN = \(path) parse(text = path))
}

devtools::load_all()

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est <- modsem(m1, data = oneInt, method = "lms", mean.observed = FALSE)
est <- modsem(m1, data = oneInt, method = "qml", mean.observed = FALSE)
summary(est)


tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

est <- modsem(tpb, data = TPB, method = "lms", mean.observed = FALSE,
              nodes = 32)
summary(est)


tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 # Causal Relationsships
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

est <- modsem(tpb_uk, data = TPB_UK, method = "lms", mean.observed = FALSE,
              nodes = 32)
summary(est)

devtools::load_all()
library(lavaan)

m1 <- '
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3
Y ~ X + Z + X:Z
'

est <- sam(m1, oneInt)
parameter_estimates(est)

centered_estimates(est)
standardized_estimates(est)

plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z",
                 vals_z = c(-0.5, 0.5), model = est)
plot_jn(x = "X", z = "Z", y = "Y", model = est)
plot_surface(x = "X", z = "Z", y = "Y", model = est)


m2 <- '
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3
Y ~ X + Z + X:Z

X ~ 1
Z ~ 1
Y ~ 1
x1~0*1
z1~0*1
y1~0*1
'

est_nc <- sam(m2, oneInt)
wrap <- \(expr) (testthat::expect_warning(expr, regex = "Replacing.*"))
wrap(centered_estimates(est_nc))
wrap(standardized_estimates(est_nc))

inspectDA_Matrices <- c("lambda", "theta", "wmat", "tmat",
                        "psi", "beta", "nu", "alpha")


inspectDA_Optim <- c("coefficients.free", "vcov.free", "information",
                     "loglik", "iterations", "convergence")


modsem_inspect_da <- function(model,
                              what = "default",
                              standardized = FALSE,
                              is.public = TRUE) {
  # for brevity...
  .modsemVector <- \(...) modsemVector(..., is.public = is.public)
  .modsemMatrix <- \(...) modsemMatrix(..., is.public = is.public)

  mod_stopif(!length(what), "`what` is of length zero!")

  if (standardized)
    model <- standardize_model(model)

  finalModel <- model$model
  groupModels <- finalModel$models
  ovs      <- c(finalModel$info$allIndsXis, finalModel$info$allIndsEtas)
  n.groups <- length(groupModels)
  is.multi <- n.groups > 1L

  group.names <- names(groupModels)
  if (!length(group.names) || any(!nzchar(group.names))) {
    level.names <- finalModel$info$group.levels

    if (!is.null(level.names) && length(level.names) == n.groups)
      group.names <- level.names
    else
      group.names <- paste0("Group", seq_len(n.groups))
  }

  expected.raw <- model$expected.matrices
  expected.by.group <- vector("list", length = n.groups)
  names(expected.by.group) <- group.names

  if (is.list(expected.raw) && length(expected.raw)) {
    names.raw <- names(expected.raw)
    for (g in seq_len(n.groups)) {
      idx <- integer()
      if (length(names.raw)) {
        candidates <- unique(c(as.character(g), group.names[[g]]))
        idx <- match(candidates, names.raw, nomatch = NA_integer_)
        idx <- idx[!is.na(idx)]
      }
      if (!length(idx) && g <= length(expected.raw)) {
        idx <- g
      }
      if (length(idx)) expected.by.group[[g]] <- expected.raw[[idx[[1]]]]
    }
  } else if (!is.null(expected.raw)) {
    expected.by.group <- replicate(n.groups, expected.raw, simplify = FALSE)
    names(expected.by.group) <- group.names
  }

  buildExpectedPayload <- function(expected) {
    if (is.null(expected)) {
      return(list(
        cov.ov  = NULL,
        cov.lv  = NULL,
        cov.all = NULL,
        cor.ov  = NULL,
        cor.lv  = NULL,
        cor.all = NULL,
        mean.lv = NULL,
        mean.ov = NULL,
        mean.all = NULL,
        r2.all  = NULL,
        r2.lv   = NULL,
        r2.ov   = NULL,
        res.all = NULL,
        res.lv  = NULL,
        res.ov  = NULL
      ))
    }

    cov.ov  <- expected$sigma.ov
    cov.lv  <- expected$sigma.lv
    cov.all <- expected$sigma.all

    cor.ov  <- if (!is.null(cov.ov))  cov2cor(cov.ov)   else NULL
    cor.lv  <- if (!is.null(cov.lv))  cov2cor(cov.lv)   else NULL
    cor.all <- if (!is.null(cov.all)) cov2cor(cov.all)  else NULL

    list(
      cov.ov  = .modsemMatrix(cov.ov,  symmetric = TRUE),
      cov.lv  = .modsemMatrix(cov.lv,  symmetric = TRUE),
      cov.all = .modsemMatrix(cov.all, symmetric = TRUE),

      cor.ov  = .modsemMatrix(cor.ov,  symmetric = TRUE),
      cor.lv  = .modsemMatrix(cor.lv,  symmetric = TRUE),
      cor.all = .modsemMatrix(cor.all, symmetric = TRUE),

      mean.lv  = .modsemMatrix(expected$mu.lv),
      mean.ov  = .modsemMatrix(expected$mu.ov),
      mean.all = .modsemMatrix(expected$mu.all),

      r2.all  = .modsemVector(expected$r2.all),
      r2.lv   = .modsemVector(expected$r2.lv),
      r2.ov   = .modsemVector(expected$r2.ov),

      res.all = .modsemVector(expected$res.all),
      res.lv  = .modsemVector(expected$res.lv),
      res.ov  = .modsemVector(expected$res.ov)
    )
  }

  buildGroupPayload <- function(submodel, expected) {
    # matrices in lavaan notation
    matrices <- submodel$expected.matrices$matrices

    lambda <- matrices$lambda
    theta  <- matrices$theta
    wmat   <- matrices$wmat
    tmat   <- matrices$tmat
    beta   <- matrices$beta
    psi    <- matrices$psi
    alpha  <- matrices$alpha
    nu     <- matrices$nu

    if (!NROW(wmat)) {
      wmat <- lambda
      wmat[TRUE] <- 0
    }

    if (!NROW(tmat)) {
      tmat <- theta 
      tmat[TRUE] <- 0
    }

    if (!is.null(nu))    colnames(nu)   <- "~1"
    if (!is.null(alpha)) colnames(alpha) <- "~1"

    out <- c(
      list(
        N      = submodel$data$n,
        data   = submodel$data$data.full,
        lambda = .modsemMatrix(lambda),
        wmat   = .modsemMatrix(wmat),
        theta  = .modsemMatrix(theta, symmetric = TRUE),
        tmat   = .modsemMatrix(tmat),
        nu     = .modsemMatrix(nu),
        alpha  = .modsemMatrix(alpha),
        beta   = .modsemMatrix(beta),
        psi    = .modsemMatrix(psi, symmetric = TRUE)
      ),
      buildExpectedPayload(expected)
    )
  }

  group.payloads <- Map(buildGroupPayload, groupModels, expected.by.group)
  names(group.payloads) <- group.names

  collapseField <- function(field) {
    values <- lapply(group.payloads, `[[`, field)

    if (!is.multi)
      return(values[[1]])

    names(values) <- group.names
    values
  }

  N.val            <- collapseField("N")
  data.val         <- collapseField("data")
  lambda.val       <- collapseField("lambda")
  wmat.val         <- collapseField("wmat")
  tmat.val         <- collapseField("tmat")
  nu.val           <- collapseField("nu")
  theta.val        <- collapseField("theta")
  beta.val         <- collapseField("beta")
  psi.val          <- collapseField("psi")
  alpha.val        <- collapseField("alpha")
  cov.ov.val       <- collapseField("cov.ov")
  cov.lv.val       <- collapseField("cov.lv")
  cov.all.val      <- collapseField("cov.all")
  cor.ov.val       <- collapseField("cor.ov")
  cor.lv.val       <- collapseField("cor.lv")
  cor.all.val      <- collapseField("cor.all")
  mean.lv.val      <- collapseField("mean.lv")
  mean.ov.val      <- collapseField("mean.ov")
  mean.all.val     <- collapseField("mean.all")
  r2.all.val       <- collapseField("r2.all")
  r2.lv.val        <- collapseField("r2.lv")
  r2.ov.val        <- collapseField("r2.ov")
  res.all.val      <- collapseField("res.all")
  res.lv.val       <- collapseField("res.lv")
  res.ov.val       <- collapseField("res.ov")

  info <- list(
     N                 = N.val,
     vcov.all          = .modsemMatrix(model$vcov.all, symmetric = TRUE),
     vcov.free         = .modsemMatrix(model$vcov.free, symmetric = TRUE),
     information       = .modsemMatrix(model$FIM, symmetric = TRUE),
     data              = data.val,
     coefficients.all  = .modsemVector(model$coefs.all),
     coefficients.free = .modsemVector(model$coefs.free),
     partable          = parameter_estimates(model, is.public = is.public),
     partable.input    = model$originalParTable,
     loglik            = model$logLik,
     iterations        = model$iterations,
     convergence       = model$convergence,
     ovs               = ovs,

     ngroups     = model$model$info$n.groups,
     group       = model$args$group,
     group.label = if (!is.null(model$args$group)) group.names else NULL,

     lambda  = lambda.val,
     theta   = theta.val,
     wmat    = wmat.val,
     tmat    = tmat.val,
     nu      = nu.val,
     alpha   = alpha.val,
     beta    = beta.val,
     psi     = psi.val,

     cov.ov  = cov.ov.val,
     cov.lv  = cov.lv.val,
     cov.all = cov.all.val,

     cor.ov  = cor.ov.val,
     cor.lv  = cor.lv.val,
     cor.all = cor.all.val,

     mean.lv  = mean.lv.val,
     mean.ov  = mean.ov.val,
     mean.all = mean.all.val,

     r2.all  = r2.all.val,
     r2.lv   = r2.lv.val,
     r2.ov   = r2.ov.val,

     res.all = res.all.val,
     res.lv  = res.lv.val,
     res.ov  = res.ov.val
  )

  optim.fields <- inspectDA_Optim
  if (!is.null(model$iteration.history)) {
    info$iteration.history <- model$iteration.history
    optim.fields <- c(optim.fields, "iteration.history")
  }

  FIT <- \() {
    h0 <- estimate_h0(model, calc.se = FALSE)
    list(
      fit.h0 = fit_modsem_da(h0, chisq = TRUE),
      fit.h1 = fit_modsem_da(model, chisq = FALSE),
      comparative.fit = compare_fit(est_h1 = model, est_h0 = h0)
    )
  }

  if (length(what) > 1) {
    fields <- info[what]

  } else {
    fields <- switch(
      EXPR      = what,
      coef.all  = info[c("vcov.all", "coefficients.all")],
      coef      = info[c("vcov.all", "coefficients.all")],
      coef.free = info[c("vcov.free", "coefficients.free")],
      default   = info[names(info) != "data"],
      all       = info,
      matrices  = info[inspectDA_Matrices],
      optim     = info[optim.fields],
      fit       = FIT(),
      info[[what]]
    )
  }

  nullvalues <- vapply(fields, FUN.VALUE = logical(1L), FUN = is.null)
  okifnull   <- names(fields) %in% c("group", "group.label")

  mod_warnif(any(nullvalues & !okifnull),
         "Some fields in `modsem_inspect()` could not be retrieved!")

  fields
}

# function for selecting rows in a dataframe matching values on a specified column (type chr_vec)
selectRowsByCol <- function(value, df, column) {
  column <- as.character(column)
  out <- df[df[[column]] == value,]

  if (nrow(out) <= 0) return(NULL)
  else return(out)
}


# function for selecting values in the column of a dataframe matching values on a specified column (type chr_vec)
selectValuesByCol <- function(value, df, column.value, column.match) {
  column.value <- as.character(column.value)
  column.match <- as.character(column.match)

  # this should return a vector anyways, but here i force it
  out <- as.vector(df[[column.value]][df[[column.match]] == value])

  if (length(out) <= 0) return(NULL)
  else return(out)

}


# scale numeric vector is numeric
scaleIfNumeric <- function(x, scaleFactor = TRUE) {
  if (is.null(x)) {
    mod_msg_warn("x in scaleIfNumeric was NULL")
    return(NULL)
  }

  if (scaleFactor & is.factor(x))
    x <- as.numeric(x)

  if (is.numeric(x)) y <- (x - mean(x, na.rm = TRUE)) / stats::sd(x, na.rm = TRUE)
  else               y <- x

  y
}


centerIfNumeric <- function(x, scaleFactor = TRUE) {
  if (is.null(x)) {
    mod_msg_warn("x in centerIfNumeric was NULL")
    return(NULL)
  }

  if (scaleFactor & is.factor(x))
    x <- as.numeric(x)

  if (is.numeric(x)) y <- x - mean(x, na.rm = TRUE)
  else               y <- x

  y
}


# A fancy version of purrr::list_cbind()
  # This function will remove duplicate columns, before combining the
    # dataframes
  # note this function might be slow for large df's, I might create a C++
    # version of it
combineListDf <- function(listDf) {
  # This function should work recursively
  if (is.null(listDf) || length(listDf) < 1) {
    return(NULL)

  } else if (length(listDf) == 1)  {
    return(listDf[[1]])

    # This shouldnt really be necessary, but in the case that the function
      # recieves a dataframe, we should just return it
  } else if (is.data.frame(listDf))  {
    return(listDf)
  }

  # Basecase combine the first two columns of the df
    # check if there are matching colnames (the first df has priority)
  matchingColnames <- colnames(listDf[[2]]) %in% colnames(listDf[[1]])

  if (sum(as.integer(matchingColnames)) > 0) {
    duplicates <- stringr::str_c(colnames(listDf[[2]])[matchingColnames],
                                 collapse = ", ")
    mod_msg_warn(paste0("There were some duplicate product indicators, was this intended?\n",
             "The duplicates of these product indicators were removed: \n", duplicates, "\n"))
  }

  combinedDf <- cbind.data.frame(listDf[[1]], listDf[[2]][,!matchingColnames, drop = FALSE])

  combineListDf(c(list(combinedDf), listDf[-(1:2)]))
}


maxDepth <- function(list, max = 2, depth = 1) {

  if (is.null(list) | !is.list(list)) {
    return(depth)
  }

  mod_stopif(depth > max, "Incorrectly nested syntax")
  deepest <- 1
  for (i in seq_along(list)) {
    branchDepth <- maxDepth(list[[i]], max = max, depth + 1)
    if (branchDepth > deepest) {
      deepest <- branchDepth
    }
  }
  deepest

}


capturePrint <- function(x, ...) {
  paste(utils::capture.output(print(x, ...)), collapse = "\n")
}


rbindParTable <- function(parTable, newRows) {
  # Merges rows of two partables, and replaces duplicates in the lhs partable
  # check for duplicates in lhs, op & rhs
  if (is.null(newRows) || NROW(newRows) == 0 || NCOL(newRows) == 0) {
    return(parTable)
  }
  newParTableRows <- apply(newRows[c("lhs", "op", "rhs")],
                        MARGIN = 1,
                        FUN = function(row)
                          list(row)) |>
    purrr::list_flatten()

  duplicateRows <- apply(
    X = parTable[c("lhs", "op", "rhs")],
    MARGIN = 1,
    FUN = \(row, parTableRows) list(row) %in% parTableRows,
    parTableRows = newParTableRows
  )

  dupParTable <- parTable[duplicateRows, , drop = FALSE]
  dupPar <- paste0(dupParTable$lhs, dupParTable$op, dupParTable$rhs)

  mod_warnif(sum(as.integer(duplicateRows)) > 0,
    paste0("Some duplicates in the parTable was removed, have you accidentally ",
    "specified some of these in your syntax? \n", paste0(dupPar, collapse = ", "))
  )

  rbind(parTable[!duplicateRows, ], newRows)
}


greplRowDf <- function(col, df) {
  if (is.null(df)) return(FALSE)
  any(apply(df, MARGIN = 2, FUN = function(dfCol) all(sort(col) == sort(dfCol))))
}


`forceRowNames<-` <- function(df, value) {
  attr(df, "row.names") <- value
  df
}


# see functions for generating formulas for
# constrained approach
`%in_paired%` <- function(x, y) {
  out <- vector("logical", length(x))
  for (i in seq_along(x)) {
    matches <- y == x[[i]]
    if (any(matches)) {
      y <- y[!matches]
      out[[i]] <- TRUE
    } else {
      out[[i]] <- FALSE
    }
  }
  out
}


defineUndefinedLabels <- function(parTable.x, parTable.y) {
  # parTable.x = original parTable without new constraints
  # parTable.y = altered parTable with new constraints (and new labels)
  # goal: make sure user-specified labels are not overwritten

  parTable.o <- parTable.y # parTable out
  parTable.x <- stripColonsParTable(parTable.x)
  parTable.x <- rename(parTable.x, mod="mod.x")
  parTable.y <- rename(parTable.y, mod="mod.y")

  # means that we ha replaced a label
  parTable.z <- merge(parTable.y, parTable.x)
  parTable.z <- parTable.z[!parTable.z$mod.y %in% c("", "NA") &
                           !parTable.z$mod.x %in% c("", "NA") &
                           parTable.z$mod.x != parTable.z$mod.y &
                           parTable.z$op %in% c("=~", "~1", "~", "~~"), ]

  redefinedLabels <- getRedefinedLabels(parTable.z=parTable.z)

  rbind(redefinedLabels, parTable.o) # place redefinitions at the start, as to not confuse lavaan
}


getRedefinedLabels <- function(parTable.z) {
  redefinedLabels <- NULL
  for (i in seq_len(nrow(parTable.z))) {
    row <- parTable.z[i, , drop = FALSE]

    if (isLavLabelFunction(label = row$mod.x)) { # is function? then: skip and warn
      warnReplacingLabel(old = row$mod.x, new = row$mod.y, parTable.row = row)
      next
    }

    if (canBeNumeric(row$mod.x)) { # is numeric constant? then: constrain
      redef <- data.frame(lhs = row$mod.y, op = "==", rhs = row$mod.x, mod = "")
    } else { # is label? then: redefine using `:=`
      redef <- data.frame(lhs = row$mod.x, op = ":=", rhs = row$mod.y, mod = "")
    }

    redefinedLabels <- rbind(redefinedLabels, redef)
  }

  redefinedLabels
}


getVarsPI <- function(parTable) {
  unique(c(parTable$rhs[parTable$op %in% c("~", "=~")],
           parTable$lhs[parTable$op == "~"])) |>
  stringr::str_split(pattern = ":", simplify = FALSE) |>
  unlist() |> unique()
}


isLavLabelFunction <- function(label, context, warning = FALSE) {
  grepl("^(start|equal)\\(.*\\)$", label)
}


warnReplacingLabel <- function(old, new, parTable.row) {
  context <- paste(parTable.row$lhs, parTable.row$op,
                   paste0(old, "*", parTable.row$rhs))
  mod_msg_warn(paste0("Replacing `", old, "` with new label `", new, "` in: `",
           context, "`"))
}


checkHigherOrderInteractions <- function(elementsInProds, parTable) {
  higherOrderLVs <- getHigherOrderLVs(parTable)
  for (xz in elementsInProds) {
    mod_stopif(any(xz %in% higherOrderLVs), paste0("The `:` operator is not allowed ",
           "for higher order latent variables, please redefine the interaction term",
           "as a higher order latent variable using the `=~` operator.\n",
           "You can also try using `method=\"lms\"` or `rcs=TRUE`\n",
           "Run `vignette(\"higher_order_interactions\")` for more information"))
  }
}


checkElementsInProds <- function(elementsInProds, lVs, oVs) {
  for (xz in elementsInProds) {
    mod_stopif(!all(xz %in% c(lVs, oVs)), paste0("Unknown element in product term: `",
           xz[!xz %in% c(lVs, oVs)][[1]], "`!"))
  }
}


redefineMeasurementModel <- function(parTable) {
  lhs <- parTable$lhs
  rhs <- parTable$rhs
  parTable$lhs <- ifelse(parTable$op == "=~",  yes = rhs, no = lhs)
  parTable$rhs <- ifelse(parTable$op == "=~",  yes = lhs, no = rhs)
  parTable$op  <- ifelse(parTable$op == "=~", yes = "~", no = parTable$op)
  parTable
}


getRsqrPI <- function(model) {
  fit <- extract_lavaan(model)
  r2  <- lavaan::lavInspect(fit, what = "r2")

  if (!is.list(r2)) return(r2)

  R2 <- as.data.frame(r2)
  nobs <- lavaan::lavInspect(fit, what = "nobs")
  ntotal <- sum(nobs)

  apply(R2, MARGIN = 1, FUN = \(x) sum(x * nobs) / ntotal)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @name modsem
#' @importFrom Rcpp sourceCpp
#' @importFrom memoise memoise
#' @useDynLib modsem, .registration = TRUE
## usethis namespace: end
NULL

calcFIM_da <- function(model,
                       theta,
                       method = "lms",
                       calc.se = TRUE,
                       FIM = "observed",
                       robust.se = FALSE,
                       P = NULL,
                       hessian = FALSE,
                       EFIM.parametric = TRUE,
                       NA__ = -999,
                       EFIM.S = 3e4,
                       epsilon = 1e-8,
                       R.max = 1e6,
                       verbose = FALSE,
                       cr1s = TRUE) {
  if (!calc.se) return(list(FIM = NULL, vcov = NULL, vcov.sub = NULL, type = "none",
                            raw.labels = names(theta), n.additions = 0))
  if (verbose) printf("Calculating standard errors (%s)\n", FIM)

  collectCluster <- function(d) {
    if (is.null(d)) return(NULL)

    if (is.list(d) && !is.data.frame(d)) {
      clusters <- lapply(d, function(x) x$cluster)
      clusters <- clusters[lengths(clusters) > 0L]
      if (!length(clusters)) return(NULL)
      do.call(c, clusters)

    } else {
      d$cluster
    }
  }

  DATA <- lapply(model$models, FUN = \(sub) sub$data)
  cluster.vec <- if (robust.se) collectCluster(DATA) else NULL

  I <- switch(method,
     lms =
       switch(FIM,
          observed = calcOFIM_LMS(model, theta = theta, epsilon = epsilon,
                                  hessian = hessian, P = P, robust.se = robust.se,
                                  cluster = cluster.vec, cr1s = cr1s),
          expected = calcEFIM_LMS(model, theta = theta, epsilon = epsilon,
                                  S = EFIM.S, parametric = EFIM.parametric,
                                  verbose = verbose, R.max = R.max, P = P),
          mod_msg_stop("FIM must be either expected or observed")),
     qml =
       switch(FIM,
          observed = calcOFIM_QML(model, theta = theta, hessian = hessian,
                                  epsilon = epsilon, robust.se = robust.se,
                                  cluster = cluster.vec, cr1s = cr1s),
          expected = calcEFIM_QML(model, theta = theta, epsilon = epsilon,
                                  S = EFIM.S, parametric = EFIM.parametric,
                                  verbose = verbose, R.max = R.max),
          mod_msg_stop("FIM must be either expected or observed")),
     mod_msg_stop(paste0("Unrecognized method: ", method))
  )


  if (robust.se) {
    mod_warnif(hessian && FIM == "observed",
           paste0("`robust.se = TRUE` should not be paired with ",
                  "`OFIM.hessian = TRUE` and `FIM = \"observed\"`"))
    H <- calcHessian(model, theta = theta, method = method,
                     epsilon = epsilon, P = P)
    invH <- solveFIM(H, NA__ = NA__)

    vcov <- invH %*% I %*% invH

  } else {
    vcov <- solveFIM(I, NA__ = NA__)
  }

  vcov.all <- getVCOV_LabelledParams(vcov = vcov, model = model, theta = theta,
                                     method = method)

  nAdditions   <- ncol(vcov.all) - ncol(vcov)
  lavLabels    <- model$params$lavLabels
  subLavLabels <- lavLabels[colnames(vcov.all) %in% names(theta)]
  rawLabels    <- colnames(vcov.all)
  dimnames(vcov.all) <- list(lavLabels, lavLabels)
  dimnames(I) <- dimnames(vcov) <- list(subLavLabels, subLavLabels)

  list(FIM = I, vcov.all = vcov.all, vcov.free = vcov, type = FIM,
       raw.labels = rawLabels, n.additions = nAdditions)
}


fdHESS <- function(pars,
                   fun,
                   ...,
                   .relStep = .Machine$double.eps^(1/5),
                   .minAbsPar = 0,
                   .switch.size = 120L,
                   .mem.limit.bytes = 3 * 2 ^ 30) {
  stopifnot(is.numeric(pars))

  npar <- length(pars)
  if (npar == 0)
    return(matrix(0, 0, 0))

  # mirror C++ heuristic: quadratic fit becomes computationally expensive for
  # when there are many free parameters
  m.ls <- 1 + 2 * npar + (npar * (npar - 1)) / 2
  bytes.X <- m.ls * m.ls * 8
  force.full.fd <- npar >= .switch.size || bytes.X > .mem.limit.bytes

  computeFD <- function() {
    tryCatch(
      fdHessFullFD(
        pars = pars,
        fun = fun,
        ...,
        relStep = .relStep,
        minAbsPar = .minAbsPar
      ),
      error = function(e) {
        mod_msg_warn(paste0("Finite-difference Hessian calculation failed...\n  ", e$message))
        matrix(NA, nrow = npar, ncol = npar)
      }
    )
  }

  if (!force.full.fd) {
    tryCatch(
      nlme::fdHess(pars = pars, fun = fun, ..., .relStep = .relStep)$Hessian,
      error = function(e) {
        mod_msg_warn(paste0(
          "Switching to fallback finite-difference Hessian after nlme::fdHess error:\n  ",
          e$message
        ))
        computeFD()
      }
    )
  } else computeFD()
}


fdHessFullFD <- function(pars, fun, ..., relStep, minAbsPar) {
  npar <- length(pars)
  par_names <- names(pars)

  relStep <- rep(relStep, length.out = npar)
  minAbsPar <- rep(minAbsPar, length.out = npar)
  base <- as.numeric(pars)
  names(base) <- par_names

  incr <- pmax(abs(base), minAbsPar) * relStep
  incr[incr == 0] <- relStep[incr == 0]

  offset_template <- numeric(npar)
  eval_fun <- function(offset) {
    candidate <- base
    candidate[] <- base + offset
    fun(candidate, ...)
  }

  f0 <- eval_fun(offset_template)
  Hess <- matrix(0, npar, npar)

  for (i in seq_len(npar)) {
    step_plus <- offset_template
    step_minus <- offset_template
    step_plus[i] <- incr[i]
    step_minus[i] <- -incr[i]

    f_ip <- eval_fun(step_plus)
    f_im <- eval_fun(step_minus)

    hi <- incr[i]
    Hess[i, i] <- (f_ip + f_im - 2 * f0) / (hi * hi)
  }

  if (npar > 1) {
    for (i in seq_len(npar - 1)) {
      hi <- incr[i]
      for (j in seq.int(i + 1, npar)) {
        hj <- incr[j]

        step <- offset_template
        step[i] <- hi
        step[j] <- hj
        fpp <- eval_fun(step)

        step[j] <- -hj
        fpm <- eval_fun(step)

        step[i] <- -hi
        fmm <- eval_fun(step)

        step[j] <- hj
        fmp <- eval_fun(step)

        hij <- (fpp - fpm - fmp + fmm) / (4 * hi * hj)
        Hess[i, j] <- hij
        Hess[j, i] <- hij
      }
    }
  }

  if (!is.null(par_names))
    dimnames(Hess) <- list(par_names, par_names)

  Hess
}


calcHessian <- function(model, theta, method = "lms",
                        epsilon = 1e-8, P = NULL) {
  if (method == "lms") {
    if (is.null(P)) P <- estepLms(model, theta = theta)
    # negative hessian (sign = -1)
    fH <- \(model) observedInfoFromLouisLms(model = model, theta = theta,
                                            P = P)$I.obs

    H <- tryCatch(suppressWarnings(fH(model)), error = function(e) {
      mod_msg_warn(paste0("Optimized calculation of Hessian failed, attempting to switch!\n", e))
      model$params$gradientStruct$useFDGradient <- TRUE

      suppressWarnings(fH(model))
    })

  } else if (method == "qml") {
    # negative hessian (sign = -1)
    suppressWarnings({
      H <- hessianLogLikQml(theta = theta, model = model,
                            sign = -1, .relStep = .Machine$double.eps^(1/5))
    })
  }

  H
}


solveFIM <- function(H, NA__ = -999, use.ginv = FALSE) {
  tryCatch({
    if (use.ginv) GINV(H) else solve(H)

  }, error = function(e) {
     if (!use.ginv)
       return(solveFIM(H, NA__ = NA__, use.ginv = TRUE))

     H[TRUE] <- NA__
     H

  }, warning = function(w) {
     if (grepl("NaN", conditionMessage(w)))
       suppressWarnings(solve(H)) else solve(H)
  })
}


calcSE_da <- function(calc.se = TRUE, vcov, rawLabels, NA__ = -999) {
  if (!calc.se)
    return(stats::setNames(rep(NA__, length(rawLabels)), nm = rawLabels))

  if (is.null(vcov)) {
    mod_msg_warn(paste0("Fisher Information Matrix (FIM) was not calculated, ",
             "unable to compute standard errors"))
    return(rep(NA__, length(rawLabels)))
  }

  se <- suppressWarnings(sqrt(diag(vcov)))

  if (all(is.na(se))) {
    mod_msg_warn("Standard errors could not be computed, negative Hessian is singular.")
  } else if (any(is.nan(se))) {
    mod_msg_warn("Standard errors for some coefficients could not be computed.")
  }

  if (!is.null(names(se))) names(se) <- rawLabels
  se[is.na(se)] <- NA__
  se
}


calcOFIM_LMS <- function(model, theta, hessian = FALSE,
                         epsilon = 1e-6, P = NULL,
                         robust.se = FALSE,
                         cluster   = NULL,
                         cr1s      = TRUE) {
  if (is.null(P))
    P <- estepLms(model, theta = theta)

  if (hessian) {
    # negative hessian (sign = -1)
    I <- calcHessian(model, theta = theta,
                     method = "lms", epsilon = epsilon, P = P)
    return(I)
  }

  # S: N x k matrix of individual score contributions (OPG)
  S <- suppressWarnings(
    gradientObsLogLikLms_i(theta, model = model, P = P, sign = +1,
                           epsilon = epsilon)
  )

  if (!robust.se || is.null(cluster)) {
    # classic OFIM via outer product of gradients (BHHH)
    return(crossprod(S))
  }

  mod_stopif(length(cluster) != nrow(S),
         "Length of 'cluster' must equal the number of rows in the data / scores.")

  f <- as.factor(cluster)
  G <- nlevels(f)
  k <- ncol(S)

  # aggregate scores by cluster: s_g = sum_{i in g} s_i
  Sg <- matrix(0, nrow = G, ncol = k)
  lev <- levels(f)
  for (g in seq_len(G)) {
    idx <- which(f == lev[g])
    Sg[g, ] <- colSums(S[idx, , drop = FALSE])
  }

  B <- crossprod(Sg)  # meat = sum_g s_g s_g'

  # optional CR1S small-sample correction
  if (isTRUE(cr1s)) {
    N <- nrow(S); q <- ncol(S)
    if (G > 1 && N > q) {
      B <- B * (G / (G - 1)) * ((N - 1) / (N - q))
    }
  }

  B
}


calcEFIM_LMS <- function(model,
                         theta,
                         S          = 100,
                         parametric = TRUE,
                         epsilon    = 1e-6,
                         verbose    = FALSE,
                         R.max      = 1e6,
                         P          = NULL) {
  k <- length(theta) # number of free parameters
  N <- sum(vapply(model$models, FUN.VALUE = numeric(1L), FUN = \(sub) sub$data$n))
  G <- model$info$n.groups
  R <- min(R.max, N * S)
  R <- R - R %% G # make R divisble by the number of groups
  R.g <- R / G
  mod_warnif(R.max <= N, "R.max is less than N!")

  ovs <- colnames(model$models[[1L]]$data$data.full)

  if (parametric) {
    # final model (without SEs)
    finalModel <- getFinalModel(model = model, theta = theta, method = "lms")

    for (g in seq_len(model$info$n.groups)) {
      parTable.g <- modelToParTable(finalModel$models[[g]], method = "lms")
      sample.g   <- simulateDataParTable(parTable.g, N = R.g, colsOVs = ovs)$OV[[1L]]

      model$models[[g]]$data <- patternizeMissingDataFIML(sample.g)
    }

  } else for (g in seq_len(model$info$n.groups)) {
    data.g   <- model$models[[g]]$data$data.full
    sample.g <- data.g[sample(NROW(data.g), R.g, replace = TRUE), , drop = FALSE]
    model$models[[g]]$data <- patternizeMissingDataFIML(sample.g)
  }

  popEstep <- estepLms(model      = model,
                       theta      = theta,
                       recalcQuad = TRUE,
                       lastQuad   = if(!is.null(P)) P$quad else NULL)

  suppressWarnings({
    J <- gradientObsLogLikLms_i(theta   = theta,
                                model   = model,
                                P       = popEstep,
                                sign    = +1,
                                epsilon = epsilon)      # R × k matrix
  })

  I <- matrix(0, nrow = k, ncol = k)
  for (i in seq_len(S)) {
    if (R == N * S) {
      # non-overlapping split
      idx1 <- (i - 1) * N + 1
      sub  <- idx1:(idx1 + N - 1)
    } else {
      sub <- sample(R, N)
    }

    I <- I + crossprod(J[sub, , drop = FALSE])
  }

  I / S
}


calcEFIM_QML <- function(model, theta, data, S = 100,
                         parametric = TRUE, epsilon = 1e-8, verbose = FALSE,
                         R.max = 1e6) {
  k <- length(theta) # number of free parameters
  N <- sum(vapply(model$models, FUN.VALUE = numeric(1L), FUN = \(sub) sub$data$n))
  G <- model$info$n.groups
  R <- min(R.max, N * S)
  R <- R - R %% G # make R divisble by the number of groups
  R.g <- R / G

  mod_warnif(R.max <= N, "R.max is less than N!")

  ovs <- colnames(model$models[[1L]]$data$data.full)

  if (parametric) {
    # final model (without SEs)
    finalModel <- getFinalModel(model = model, theta = theta, method = "qml")

    for (g in seq_len(model$info$n.groups)) {
      parTable.g <- modelToParTable(finalModel$models[[g]], method = "qml")
      sample.g   <- simulateDataParTable(parTable.g, N = R.g, colsOVs = ovs)$OV[[1L]]

      model$models[[g]]$data <- patternizeMissingDataFIML(sample.g)
    }

  } else for (g in seq_len(model$info$n.groups)) {
    data.g   <- model$models[[g]]$data$data.full
    sample.g <- data.g[sample(NROW(data.g), R.g, replace = TRUE), , drop = FALSE]
    model$models[[g]]$data <- patternizeMissingDataFIML(sample.g)
  }

  for (g in seq_along(model$models)) {
    if (!is.null(model$models[[g]]$matrices$fullU)) {
      fullU     <- model$models[[g]]$matrices$fullU
      fullU_New <- fullU[rep(seq_len(NROW(fullU)), length.out = R.g), , drop = FALSE]

      model$models[[g]]$matrices$fullU <- fullU_New
    }
  }

  suppressWarnings({
    J <- gradientLogLikQml_i(theta = theta, model = model, sign = +1,
                             epsilon = epsilon)
  })

  I <- matrix(0, nrow = k, ncol = k)
  for (i in seq_len(S)) {
    if (R == N * S) {
      # non-overlapping split
      idx1 <- (i - 1) * N + 1
      sub  <- idx1:(idx1 + N - 1)
    } else {
      sub <- sample(R, N)
    }

    I <- I + crossprod(J[sub, , drop = FALSE])
  }

  I / S
}


calcOFIM_QML <- function(model, theta,
                         hessian = FALSE,
                         epsilon = 1e-8,
                         robust.se = FALSE,
                         cluster   = NULL,
                         cr1s      = TRUE) {
  if (hessian) {
    # negative hessian (sign = -1)
    I <- calcHessian(model = model, theta = theta,
                     method = "qml", epsilon = epsilon)
    return(I)
  }

  # S: N x k matrix of individual score contributions (sign = +1 => score)
  S <- suppressWarnings(
    gradientLogLikQml_i(theta, model = model, sign = +1, epsilon = epsilon)
  )

  if (!robust.se || is.null(cluster)) {
    # classic OFIM (BHHH / OPG)
    return(crossprod(S))
  }

  mod_stopif(length(cluster) != nrow(S),
         "Length of 'cluster' must equal the number of rows in the data / scores.")

  f <- as.factor(cluster)
  G <- nlevels(f)
  k <- ncol(S)

  # s_g = sum_{i in g} s_i
  Sg <- matrix(0, nrow = G, ncol = k)
  lev <- levels(f)
  for (g in seq_len(G)) {
    idx <- which(f == lev[g])
    Sg[g, ] <- colSums(S[idx, , drop = FALSE])
  }

  B <- crossprod(Sg)  # meat = sum_g s_g s_g'

  # Optional CR1S small-sample correction
  if (isTRUE(cr1s)) {
    N <- nrow(S); q <- ncol(S)
    if (G > 1 && N > q)
      B <- B * (G / (G - 1)) * ((N - 1) / (N - q))
  }

  B
}


getSE_Model <- function(model, se, method, n.additions) {
  params <- model$params

  for (g in seq_len(model$info$n.groups)) {
    SELECT_THETA_LAB  <- params$SELECT_THETA_LAB[[g]]
    SELECT_THETA_COV  <- params$SELECT_THETA_COV[[g]]
    SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN[[g]]

    SELECT_THETA_LAB  <- seq_len(MAX(SELECT_THETA_LAB) + n.additions)
    SELECT_THETA_COV  <- SELECT_THETA_COV  + n.additions
    SELECT_THETA_MAIN <- SELECT_THETA_MAIN + n.additions

    params$SELECT_THETA_LAB[[g]]  <- SELECT_THETA_LAB
    params$SELECT_THETA_COV[[g]]  <- SELECT_THETA_COV
    params$SELECT_THETA_MAIN[[g]] <- SELECT_THETA_MAIN
  }

  model$params <- params

  fillModel(replaceNonNaModelMatrices(model, value = -999),
            theta = se, method = method)
}

devtools::load_all()
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  # Covariances
  ATT ~~ SN + PBC
  PBC ~~ SN
  # Causal Relationsships
  INT ~ a * ATT + b * SN + c * PBC
  BEH ~ INT + PBC
  BEH ~ d * INT:PBC

  a_b := a * b * d # check that labels arent overwritten without redefining
'

method <- c("ca", "rca", "uca", "dblcent")
ests <- lapply(method, function(m) modsem(tpb, data = TPB, method = m))
estsMatch <- lapply(method, function(m) modsem(tpb, data = TPB, method = m,
                                               match = TRUE))

print(summary(ests[[1]]))

modsemOrderedMCCorrection <- function(model.syntax,
                                      data,
                                      method = "lms",
                                      ordered = NULL,
                                      calc.se = TRUE,
                                      group = NULL,
                                      verbose = interactive(),
                                      optimize = TRUE,
                                      start = NULL,
                                      mc.reps = NULL,
                                      ordered.mc.reps = mc.reps,
                                      ordered.min.iter = 20L,
                                      ordered.max.iter = 250L,
                                      ordered.tol = 1e-4,
                                      ordered.rng.seed = NULL,
                                      ordered.fixed.seed = FALSE,
                                      ordered.polyak.juditsky = TRUE,
                                      ordered.pj.extrapolate = TRUE,
                                      ordered.fn.args = list(),
                                      ordered.se = c("mixed", "delta", "penalized", "naive"),
                                      ordered.se.penalty = 0.5,
                                      ordered.delta.reps = ordered.mc.reps,
                                      ordered.delta.epsilon = 1e-2,
                                      ordered.boot.reps = 1000L,
                                      ordered.standardize = TRUE,
                                      standardize = NULL,
                                      standardize.out = NULL,
                                      cluster = NULL,
                                      sampling.weights = NULL,
                                      sampling.weights.normalization = NULL,
                                      mean.observed = NULL, # capture
                                      ...) {
  method <- tolower(method)
  ordered.se <- match.arg(ordered.se)
  mod_stopif(
    !method %in% c("lms", "qml"),
    "MC ordered correction is only available for LMS and QML."
  )

  if (is.null(verbose))
    verbose <- interactive()
  if (is.null(calc.se))
    calc.se <- TRUE
  if (is.null(ordered.mc.reps))
    ordered.mc.reps <- max(NROW(data), 2000L)
  if (is.null(ordered.delta.reps))
    ordered.delta.reps <- max(ordered.mc.reps, 10000L)

  ordered.mc.reps <- as.integer(ordered.mc.reps)
  ordered.delta.reps <- as.integer(ordered.delta.reps)
  ordered.boot.reps <- as.integer(ordered.boot.reps %||% 0L)

  mod_stopif(ordered.mc.reps < 2L, "`ordered.mc.reps` must be at least 2.")
  mod_stopif(ordered.boot.reps < 0L, "`ordered.boot.reps` must be non-negative.")
  mod_stopif(ordered.min.iter < 1L, "`ordered.min.iter` must be at least 1.")
  mod_stopif(ordered.max.iter < ordered.min.iter,
         "`ordered.max.iter` must be larger than `ordered.min.iter`.")
  mod_stopif(!is.list(ordered.fn.args), "`ordered.fn.args` must be a list.")

  if (ordered.fixed.seed && is.null(ordered.rng.seed)) {
    ordered.rng.seed <- floor(stats::runif(1L, min = 1, max = 9999999))
    if (verbose)
      mod_msg_note(sprintf("Using fixed MC ordered seed %i.", ordered.rng.seed))
  }

  parTable.in <- modsemify(model.syntax)
  ovs <- getOVs(parTable.in)

  data <- as.data.frame(data)
  cols.ordered <- mcOrderedCols(data = data, ovs = ovs, ordered = ordered)
  mod_stopif(!length(cols.ordered),
    "No ordered variables were found for MC ordered correction."
  )

  if (verbose) {
    mod_msg_note(
      sprintf(
        "Estimating Monte-Carlo corrected estimates (method=%s, R=%d).\n",
        toupper(method), ordered.mc.reps
      ),
      "This is an experimental feature!",
      "See `help(\"modsem_da\")` for more information.",
      "Consider using the (faster) MC-OrdPLSc estimator in the `plssem` package."
    )
  }

  observed <- mcPrepareOrderedData(
    data = data,
    ovs = ovs,
    ordered = cols.ordered,
    group = group,
    standardize = ordered.standardize
  )

  PROBS <- mcOrderedProbs(
    data = observed$raw, ordered = cols.ordered,
    group = group
  )

  fit0 <- mcFitModsemOrdered(
    model.syntax = model.syntax,
    data = observed$data,
    method = method,
    group = group,
    calc.se = isTRUE(calc.se),
    verbose = verbose,
    optimize = optimize,
    start = start,
    cluster = cluster,
    sampling.weights = sampling.weights,
    sampling.weights.normalization = sampling.weights.normalization,
    standardize = FALSE,
    standardize.out = FALSE,
    mean.observed = FALSE,
    ...
  )

  std.info <- mcStandardizedStateInfo(fit0)
  theta0 <- std.info$state
  p <- theta0
  if (!is.null(start) && length(start) == length(theta0))
    p[] <- start

  f <- function(p.i, reps = ordered.mc.reps, seed = ordered.rng.seed,
                verbose.fit = FALSE) {

    fit.sim <- mcFitSimulatedOrdinalStandardized(
      fit0 = fit0,
      state = p.i,
      std.info = std.info,
      model.syntax = model.syntax,
      method = method,
      group = group,
      ovs = ovs,
      ordered = cols.ordered,
      PROBS = PROBS,
      mc.reps = reps,
      seed = seed,
      standardize = ordered.standardize,
      sampling.weights = sampling.weights,
      sampling.weights.normalization = sampling.weights.normalization,
      mean.observed = FALSE,
      verbose = verbose.fit,
      ...
    )

    fit.sim$state - theta0
  }

  if (isTRUE(ordered.polyak.juditsky) && !isTRUE(ordered.pj.extrapolate)) {
    if (verbose) mod_msg_note("Warming up MC ordered solver.")

    warmup <- mcRobbinsMonro(
      p = p,
      f = f,
      tol = 10 * ordered.tol,
      min.iter = min(ordered.min.iter, 5L),
      max.iter = min(max(ordered.min.iter, 5L), 20L),
      verbose = verbose,
      polyak.juditsky = FALSE,
      fn.args = ordered.fn.args
    )

    p <- as.vector(warmup$root)
    names(p) <- names(theta0)
  }

  mcfit <- mcRobbinsMonro(
    p = p,
    f = f,
    tol = ordered.tol,
    min.iter = ordered.min.iter,
    max.iter = ordered.max.iter,
    verbose = verbose,
    polyak.juditsky = ordered.polyak.juditsky,
    pj.extrapolate = ordered.pj.extrapolate,
    fn.args = ordered.fn.args
  )

  theta.mc <- as.vector(mcfit$root)
  names(theta.mc) <- names(theta0)

  vcov.free <- NULL
  type.se <- if (isTRUE(calc.se)) fit0$type.se else "none"

  naive.se.label   <- "naive"
  penalized.se.label <- "penalized"
  delta.se.label   <- "mc-delta"
  mixed.se.label   <- "mixed[delta|penalized]"

  if (isTRUE(calc.se)) {
    vcov.free <- std.info$vcov.free

    if (!is.null(vcov.free) &&
        all(dim(vcov.free) == c(length(theta.mc), length(theta.mc)))) {

      if (identical(ordered.se, "delta")) {
        if (verbose)
          mod_msg_note("Calculating MC ordered delta-method standard errors.")

        vcov.free <- tryCatch({
          delta.seed <- ordered.rng.seed

          if (is.null(delta.seed)) {
            delta.seed <- floor(stats::runif(1L, min = 1, max = 9999999))

            if (verbose)
              mod_msg_note(sprintf("Using fixed MC ordered delta seed %i.", delta.seed))
          }

          vcov.delta <- mcAlignVcov(vcov.free, labels = names(theta.mc))

          H <- mcDeltaJacobian(
            p = theta.mc,
            f = f,
            reps = ordered.delta.reps,
            seed = delta.seed,
            epsilon = ordered.delta.epsilon,
            verbose = verbose
          )

          H.inv <- mcStableInverse(H)
          V <- H.inv %*% vcov.delta %*% t(H.inv)
          V <- mcStableVcov(V)

          dimnames(V) <- list(names(theta.mc), names(theta.mc))
          attr(V, "delta.jacobian") <- H
          attr(V, "delta.seed") <- delta.seed
          attr(V, "delta.inverse.method") <- attr(H.inv, "method")
          attr(V, "delta.condition") <- attr(H.inv, "condition")
          attr(V, "delta.psd.adjusted") <- attr(V, "psd.adjusted")

          # `tryCatch()` evaluates this expression directly in the function
          # frame, so a plain `<-` hits the local `type.se`; `<<-` would not.
          type.se <- delta.se.label

          V

        }, error = function(e) {
          mod_msg_warn(paste0(
            "Delta-method MC correction failed; using ordered MC penalized ",
            "standard errors instead. Message: ", conditionMessage(e)
          ))

          penalized <- mcOrderedPenaltyVcov(
            theta.mc = theta.mc,
            theta0 = theta0,
            vcov.free = std.info$vcov.free,
            lambda = ordered.se.penalty
          )

          type.se <<- penalized.se.label
          penalized
        })

      } else if (identical(ordered.se, "mixed")) {
        if (verbose)
          mod_msg_note("Calculating MC ordered mixed standard errors.")

        vcov.free <- tryCatch({
          ids.struct <- intersect(mcStructuralStateIds(std.info), names(theta.mc))
          mod_stopif(!length(ids.struct),
                     "No structural path coefficients found.")

          delta.seed <- ordered.rng.seed

          if (is.null(delta.seed)) {
            delta.seed <- floor(stats::runif(1L, min = 1, max = 9999999))

            if (verbose)
              mod_msg_note(sprintf("Using fixed MC ordered delta seed %i.", delta.seed))
          }

          vcov.delta <- mcAlignVcov(vcov.free, labels = names(theta.mc))

          H <- mcDeltaJacobian(
            p = theta.mc,
            f = f,
            reps = ordered.delta.reps,
            seed = delta.seed,
            epsilon = ordered.delta.epsilon,
            verbose = verbose,
            select = ids.struct
          )

          H.inv <- mcStableInverse(H)
          V.struct <- H.inv %*%
            vcov.delta[ids.struct, ids.struct, drop = FALSE] %*% t(H.inv)

          V <- mcOrderedPenaltyVcov(
            theta.mc = theta.mc,
            theta0 = theta0,
            vcov.free = std.info$vcov.free,
            lambda = ordered.se.penalty
          )

          V <- mcAlignVcov(V, labels = names(theta.mc))
          V[ids.struct, ids.struct] <- V.struct
          V <- mcStableVcov(V)

          dimnames(V) <- list(names(theta.mc), names(theta.mc))
          attr(V, "delta.jacobian") <- H
          attr(V, "delta.seed") <- delta.seed
          attr(V, "delta.inverse.method") <- attr(H.inv, "method")
          attr(V, "delta.condition") <- attr(H.inv, "condition")
          attr(V, "delta.parameters") <- ids.struct
          attr(V, "delta.psd.adjusted") <- attr(V, "psd.adjusted")

          type.se <- mixed.se.label

          V

        }, error = function(e) {
          mod_msg_warn(paste0(
            "Mixed delta-method MC correction failed; using ordered MC penalized ",
            "standard errors instead. Message: ", conditionMessage(e)
          ))

          penalized <- mcOrderedPenaltyVcov(
            theta.mc = theta.mc,
            theta0 = theta0,
            vcov.free = std.info$vcov.free,
            lambda = ordered.se.penalty
          )

          type.se <<- penalized.se.label
          penalized
        })

      } else if (identical(ordered.se, "naive")) {
        if (verbose)
          mod_msg_note("Using naive standard errors.")

        vcov.free <- mcOrderedNaiveScalingVcov(
          theta.mc = theta.mc,
          theta0 = theta0,
          vcov.free = vcov.free
        )

        type.se <- naive.se.label

      } else if (identical(ordered.se, "penalized")) {
        if (verbose)
          mod_msg_note("Using penalized standard errors.")

        vcov.free <- mcOrderedPenaltyVcov(
          theta.mc = theta.mc,
          theta0 = theta0,
          vcov.free = vcov.free,
          lambda = ordered.se.penalty
        )

        type.se <- penalized.se.label
      }
    }
  }

  fit.out <- mcOverwriteFitWithStdState(
    fit = fit0,
    state = theta.mc,
    vcov.free = vcov.free,
    std.info = std.info,
    calc.se = calc.se,
    type.se = type.se
  )

  sim.mc <- simulateDataParTableStandardized(
    parTable = mcStdParTableFromState(std.info = std.info, state = theta.mc),
    N = ordered.delta.reps,
    colsOVs = ovs,
    seed = ordered.rng.seed
  )

  thresholds <- mcOrderedThresholdsBoot(
    data = observed$raw,
    sim = sim.mc,
    ordered = cols.ordered,
    group = group,
    calc.se = calc.se,
    boot.reps = ordered.boot.reps,
    seed = ordered.rng.seed
  )

  fit.out <- mcAppendThresholds(
    fit = fit.out,
    thresholds = thresholds
  )

  fit.out$expected.matrices <- mcOrderedExpectedMatrices(
    fit = fit.out,
    std.info = std.info,
    state = theta.mc,
    observed = observed$data,
    ovs = ovs,
    ordered = cols.ordered,
    PROBS = PROBS,
    group = group,
    mc.reps = ordered.mc.reps,
    seed = ordered.rng.seed,
    standardize = ordered.standardize
  )

  fit.out$iterations <- mcfit$iter
  fit.out$convergence <- isTRUE(mcfit$converged)
  fit.out$convergence.msg <- getConvergenceMessage(
    fit.out$convergence, fit.out$iterations
  )

  fit.out$args$ordered <- cols.ordered
  fit.out$args$ordered.mc.reps <- ordered.mc.reps
  fit.out$args$ordered.min.iter <- ordered.min.iter
  fit.out$args$ordered.max.iter <- ordered.max.iter
  fit.out$args$ordered.tol <- ordered.tol
  fit.out$args$ordered.rng.seed <- ordered.rng.seed
  fit.out$args$ordered.polyak.juditsky <- ordered.polyak.juditsky
  fit.out$args$ordered.pj.extrapolate <- ordered.pj.extrapolate
  fit.out$args$ordered.fn.args <- ordered.fn.args
  fit.out$args$ordered.se <- ordered.se
  fit.out$args$ordered.se.penalty <- ordered.se.penalty
  fit.out$args$ordered.delta.reps <- ordered.delta.reps
  fit.out$args$ordered.delta.epsilon <- ordered.delta.epsilon
  fit.out$args$ordered.boot.reps <- ordered.boot.reps
  fit.out$args$optimize <- optimize
  fit.out$args$start <- start
  fit.out$ordered.mc <- mcfit

  fit.out$args$standardize <- TRUE
  fit.out$args$standardize.out <- TRUE

  fit.out
}


mcOrderedCols <- function(data, ovs, ordered = NULL) {
  ordered <- ordered %||% character(0L)
  cols <- ovs[ovs %in% colnames(data)]
  cols[cols %in% ordered | vapply(data[cols], is.ordered, logical(1L))]
}


mcPrepareOrderedData <- function(data, ovs, ordered, group = NULL,
                                 standardize = TRUE) {
  out <- as.data.frame(data)
  raw <- as.data.frame(data)

  if (is.null(group)) {
    out[ordered] <- lapply(out[ordered], mcOrderedInteger)
    if (standardize)
      out[ordered] <- lapply(out[ordered], std1)
  } else {
    group.vals <- out[[group]]
    for (col in ordered) {
      x.out <- numeric(NROW(out))
      for (g in unique(group.vals)) {
        idx <- which(group.vals == g)
        x <- mcOrderedInteger(out[[col]][idx])
        if (standardize) x <- std1(x)
        x.out[idx] <- x
      }
      out[[col]] <- x.out
    }
  }

  list(data = out, raw = raw[unique(c(ovs, group))])
}


mcOrderedInteger <- function(x) {
  if (is.ordered(x) || is.factor(x)) as.integer(as.ordered(x))
  else as.integer(as.ordered(x))
}


mcOrderedProbs <- function(data, ordered, group = NULL) {
  if (is.null(group)) {
    out <- list("1" = mcOrderedProbsGroup(data, ordered))
  } else {
    vals <- unique(data[[group]])
    out <- stats::setNames(vector("list", length(vals)), as.character(vals))
    for (g in vals)
      out[[as.character(g)]] <- mcOrderedProbsGroup(data[data[[group]] == g, ],
                                                    ordered)
  }

  out
}


mcOrderedProbsGroup <- function(data, ordered) {
  stats::setNames(lapply(ordered, function(col) {
    x <- as.ordered(data[[col]])
    tab <- as.numeric(table(x))
    p <- tab / sum(tab)
    p[p > 0]
  }), ordered)
}


mcFitModsemOrdered <- function(model.syntax, data, method, group, calc.se,
                               verbose, optimize, start, cluster = NULL,
                               sampling.weights = NULL,
                               sampling.weights.normalization = NULL,
                               standardize = FALSE,
                               standardize.out = FALSE, ...) {
  modsem_da(
    model.syntax = model.syntax,
    data = data,
    method = method,
    group = group,
    cluster = cluster,
    sampling.weights = sampling.weights,
    sampling.weights.normalization = sampling.weights.normalization,
    standardize = standardize,
    standardize.out = standardize.out,
    ordered = NULL,
    calc.se = calc.se,
    verbose = verbose,
    optimize = optimize,
    start = start,
    ...
  )
}


mcFitSimulatedOrdinalStandardized <- function(fit0,
                                              state,
                                              std.info,
                                              model.syntax,
                                              method,
                                              group,
                                              ovs,
                                              ordered,
                                              PROBS,
                                              mc.reps,
                                              seed,
                                              standardize = TRUE,
                                              sampling.weights = NULL,
                                              sampling.weights.normalization = NULL,
                                              verbose = FALSE,
                                              ...) {
  parTable <- mcStdParTableFromState(std.info = std.info, state = state)
  sim <- simulateDataParTableStandardized(
    parTable = parTable,
    N = mc.reps,
    colsOVs = ovs,
    seed = seed
  )

  data.sim <- mcOrdinalizeSimulatedData(
    sim = sim,
    ovs = ovs,
    ordered = ordered,
    PROBS = PROBS,
    group = group,
    standardize = standardize
  )

  if (!is.null(sampling.weights))
    data.sim[[sampling.weights]] <- 1

  fit.sim <- mcFitModsemOrdered(
    model.syntax = model.syntax,
    data = data.sim,
    method = method,
    group = group,
    calc.se = FALSE,
    verbose = verbose,
    optimize = FALSE,
    start = fit0$theta,
    cluster = NULL,
    sampling.weights = sampling.weights,
    sampling.weights.normalization = sampling.weights.normalization,
    ...
  )

  list(
    fit = fit.sim,
    state = mcExtractStandardizedState(fit.sim, std.info = std.info)
  )
}


mcOrdinalizeSimulatedData <- function(sim, ovs, ordered, PROBS, group = NULL,
                                      standardize = TRUE) {
  groups <- seq_along(sim$OV)
  out <- NULL

  for (g in groups) {
    data.g <- as.data.frame(sim$OV[[g]])[ovs]
    prob.g <- if (length(PROBS) == 1L) PROBS[[1L]] else PROBS[[g]]

    for (col in ordered) {
      data.g[[col]] <- mcOrdinalize(data.g[[col]], probs = prob.g[[col]])
      if (standardize)
        data.g[[col]] <- std1(data.g[[col]])
    }

    if (!is.null(group)) {
      group.name <- names(PROBS)[[g]]
      data.g[[group]] <- utils::type.convert(group.name, as.is = TRUE)
    }

    out <- rbind(out, data.g)
  }

  out
}


mcOrdinalize <- function(x, probs) {
  probs <- probs / sum(probs)
  if (length(probs) < 2L)
    return(rep(1L, length(x)))

  breaks <- stats::quantile(x, probs = cumsum(probs[-length(probs)]),
                            names = FALSE, type = 7, na.rm = TRUE)
  as.integer(findInterval(x, vec = breaks) + 1L)
}


mcParTableWithTheta <- function(fit, theta, method) {
  coefs <- getLavCoefs(model = fit$start.model, theta = theta, method = method)$all
  parTable <- as.data.frame(fit$parTable)
  key <- mcParTableKeys(parTable)
  idx <- match(key, names(coefs))
  hit <- !is.na(idx)
  parTable$est[hit] <- coefs[idx[hit]]
  parTable[parTable$op != "|", , drop = FALSE]
}


mcParTableKeys <- function(parTable) {
  key <- paste0(parTable$lhs, parTable$op, parTable$rhs)
  if ("group" %in% colnames(parTable)) {
    g <- parTable$group
    key <- ifelse(!is.na(g) & g > 1L, paste0(key, ".g", g), key)
  }

  key
}


mcRobbinsMonro <- function(p, f, tol, min.iter, max.iter, verbose = FALSE,
                           polyak.juditsky = FALSE, pj.extrapolate = TRUE,
                           fn.args = list(), k = 3L) {
  if (max.iter < min.iter) max.iter <- min.iter

  fn.a <- fn.args$fn.a
  if (is.null(fn.a))
    fn.a <- function(iter, a = 1, b = 1/2, c = 0, ...) a / (iter + c)^b
  fn.args$fn.a <- NULL

  history <- rbind(p, matrix(NA_real_, nrow = max.iter, ncol = length(p)))
  colnames(history) <- names(p)
  k.succ <- 0L
  pbar.last <- pbar <- p

  for (i in seq_len(max.iter)) {
    p.old <- p
    fp <- f(p)
    fp[!is.finite(fp)] <- 0
    a <- do.call(fn.a, c(list(iter = i), fn.args))
    p <- p - a * fp
    names(p) <- names(p.old)

    history[i + 1L, ] <- p
    change <- max(abs(p.old - p), na.rm = TRUE)

    if (isTRUE(polyak.juditsky)) {
      pbar.last <- pbar
      pbar <- mcPKAverage(history)
      names(pbar) <- names(p)
      change <- max(abs(pbar.last - pbar), na.rm = TRUE)
    }

    if (isTRUE(verbose)) {
      if (isTRUE(polyak.juditsky)) msg <- "\rMC iter %d; \u0394E(\u03B8) = %.4g"
      else msg <- "\rMC iter %d; \u0394\u03B8 = %.4g"

      printf(msg, i, change)
    }

    if (i >= min.iter && is.finite(change) && change < tol) {
      k.succ <- k.succ + 1L
      if (k.succ >= k) break
    } else {
      k.succ <- 0L
    }
  }

  history <- history[seq_len(i + 1L), , drop = FALSE]
  root <- if (isTRUE(polyak.juditsky)) pbar else p
  if (isTRUE(polyak.juditsky) && isTRUE(pj.extrapolate))
    root <- mcGetConvergencePoints(history)

  names(root) <- colnames(history)

  if (isTRUE(verbose))
    printf("\n")

  list(root = root, iter = i, converged = i < max.iter, history = history,
       polyak.juditsky = polyak.juditsky,
       pj.extrapolate = pj.extrapolate)
}


mcPKAverage <- function(history) {
  colMeans(history, na.rm = TRUE)
}


mcGetConvergencePoint <- function(y, t = seq_along(y)) {
  c.aitken <- mcAitkenAccelerate(y)
  span <- diff(range(y))

  if (length(y) < 4L || !is.finite(span) ||
      span < .Machine$double.eps^0.5)
    return(c.aitken)

  tryCatch({
    fit <- stats::nls(
      y ~ c + a * exp(-k * t),
      start = list(
        c = mean(utils::tail(y, 3L)),
        a = y[[1L]] - mean(utils::tail(y, 3L)),
        k = 0.1
      ),
      algorithm = "port",
      lower = c(c = -Inf, a = -Inf, k = 0)
    )

    c.nls <- stats::coef(fit)[["c"]]
    k.fit <- stats::coef(fit)[["k"]]

    bad.range <- c.nls < min(y) - span || c.nls > max(y) + span
    bad.k <- k.fit < sqrt(.Machine$double.eps)
    bad.agree <- abs(c.nls - c.aitken) > 0.5 * span

    if (bad.range || bad.k || bad.agree) c.aitken else c.nls
  }, error = function(e) c.aitken)
}


mcAitkenAccelerate <- function(y) {
  n <- length(y)
  if (n < 3L) return(mean(y, na.rm = TRUE))

  ests <- vapply(seq_len(n - 2L), FUN.VALUE = numeric(1L), FUN = function(i) {
    p0 <- y[[i]]
    p1 <- y[[i + 1L]]
    p2 <- y[[i + 2L]]

    denom <- p2 - 2 * p1 + p0
    if (abs(denom) < .Machine$double.eps^0.5)
      return(NA_real_)

    p0 - (p1 - p0)^2 / denom
  })

  valid <- ests[is.finite(ests)]
  if (!length(valid)) return(mean(y, na.rm = TRUE))
  stats::median(valid)
}


mcGetConvergencePoints <- function(history) {
  history <- history[stats::complete.cases(history), , drop = FALSE]
  out <- apply(history, MARGIN = 2L, FUN = mcGetConvergencePoint)
  names(out) <- colnames(history)
  out
}


mcDeltaJacobian <- function(p, f, reps, seed, epsilon = 1e-3,
                            verbose = interactive(), select = NULL) {
  select <- select %||% names(p)
  k <- length(select)
  J <- matrix(NA_real_, nrow = k, ncol = k,
              dimnames = list(select, select))

  for (j in seq_len(k)) {
    if (verbose) printf("\rCalculating Jacobian %d/%d...", j, k)

    id <- select[[j]]
    step <- epsilon * max(1, abs(p[[id]]))
    p.plus <- p
    p.minus <- p
    p.plus[[id]] <- p.plus[[id]] + step
    p.minus[[id]] <- p.minus[[id]] - step

    f.plus <- f(p.plus, reps = reps, seed = seed, verbose.fit = FALSE)
    f.minus <- f(p.minus, reps = reps, seed = seed, verbose.fit = FALSE)

    if (!all(is.finite(f.plus)) || !all(is.finite(f.minus)))
      mod_msg_stop("Non-finite finite-difference function value.")

    J[, j] <- (f.plus[select] - f.minus[select]) / (2 * step)
  }

  if (verbose)
    printf("\n")

  if (!all(is.finite(J)))
    mod_msg_stop("Non-finite finite-difference Jacobian.")

  J
}


mcStructuralStateIds <- function(std.info) {
  parTable <- std.info$template
  keep <- parTable$op == "~" & parTable$rhs != "1"
  unique(parTable$.state.id[keep])
}


mcAlignVcov <- function(vcov, labels) {
  if (is.null(vcov))
    mod_msg_stop("Missing variance-covariance matrix.")

  dn <- dimnames(vcov)
  if (is.null(dn) || is.null(dn[[1L]]) || is.null(dn[[2L]]))
    mod_msg_stop("Variance-covariance matrix must have dimnames.")

  missing <- setdiff(labels, rownames(vcov))
  if (length(missing))
    mod_msg_stop("Variance-covariance matrix is missing parameters: ",
         paste(missing, collapse = ", "))

  V <- vcov[labels, labels, drop = FALSE]
  V <- 0.5 * (V + t(V))
  dimnames(V) <- list(labels, labels)
  V
}


mcStableInverse <- function(X, rcond.tol = 1e-8) {
  if (NROW(X) != NCOL(X))
    mod_msg_stop("Jacobian must be square.")
  if (!all(is.finite(X)))
    mod_msg_stop("Jacobian contains non-finite values.")

  sv <- svd(X)
  d <- sv$d
  d.max <- max(d)
  d.min <- min(d)
  if (!is.finite(d.max) || d.max <= 0)
    mod_msg_stop("Jacobian is numerically zero.")

  rcond <- d.min / d.max
  if (is.finite(rcond) && rcond > rcond.tol) {
    out <- solve(X)
    method <- "solve"
    condition <- 1 / rcond
  } else {
    d.reg <- pmax(d, d.max * rcond.tol)
    out <- sv$v %*% (diag(1 / d.reg, nrow = length(d.reg)) %*% t(sv$u))
    method <- "svd-ridge"
    condition <- 1 / max(rcond, rcond.tol, na.rm = TRUE)
  }

  dimnames(out) <- rev(dimnames(X))
  attr(out, "method") <- method
  attr(out, "condition") <- condition
  out
}


mcStableVcov <- function(V, eigen.tol = .Machine$double.eps^0.5) {
  nms <- dimnames(V)
  V <- 0.5 * (V + t(V))
  if (!all(is.finite(V)))
    mod_msg_stop("Delta-method variance-covariance matrix contains non-finite values.")

  eig <- eigen(V, symmetric = TRUE)
  scale <- max(abs(eig$values), 1)
  lower <- -eigen.tol * scale
  adjusted <- any(eig$values < 0)
  materially.adjusted <- any(eig$values < lower)
  values <- pmax(eig$values, 0)

  if (adjusted) {
    V <- eig$vectors %*% (diag(values, nrow = length(values)) %*% t(eig$vectors))
    V <- 0.5 * (V + t(V))
  }

  dimnames(V) <- nms
  attr(V, "psd.adjusted") <- adjusted
  attr(V, "psd.materially.adjusted") <- materially.adjusted
  V
}


mcOrderedNaiveScalingVcov <- function(theta.mc, theta0, vcov.free,
                                      eps = .Machine$double.eps^0.5) {
  if (is.null(vcov.free)) return(NULL)

  ids <- intersect(names(theta.mc), intersect(names(theta0), rownames(vcov.free)))
  if (!length(ids)) return(vcov.free)

  scale <- rep(1, length(ids))
  names(scale) <- ids

  denom <- theta0[ids]
  numer <- theta.mc[ids]
  ok <- is.finite(denom) & is.finite(numer) & abs(denom) > eps
  scale[ok] <- numer[ok] / denom[ok]

  D <- diag(scale, nrow = length(scale))
  dimnames(D) <- list(ids, ids)
  V <- vcov.free[ids, ids, drop = FALSE]
  V <- D %*% V %*% D
  V <- 0.5 * (V + t(V))
  dimnames(V) <- list(ids, ids)
  expandVCOV(V, labels = rownames(vcov.free))
}


mcOrderedPenaltyVcov <- function(theta.mc, theta0, vcov.free, lambda = 1) {
  if (is.null(vcov.free)) return(NULL)

  ids <- intersect(names(theta.mc), intersect(names(theta0), rownames(vcov.free)))
  if (!length(ids)) return(vcov.free)

  d <- theta.mc[ids] - theta0[ids]
  V <- vcov.free[ids, ids, drop = FALSE]
  V <- V + lambda * diag(d^2, nrow = length(d))
  V <- 0.5 * (V + t(V))
  dimnames(V) <- list(ids, ids)
  expandVCOV(V, labels = rownames(vcov.free))
}


simulateDataParTableStandardized <- function(parTable, N, colsOVs = NULL,
                                             colsLVs = NULL, seed = NULL) {
  simulateDataParTable(
    parTable = parTable,
    N = N,
    colsOVs = colsOVs,
    colsLVs = colsLVs,
    seed = seed
  )
}


mcOrderedExpectedMatrices <- function(fit,
                                      std.info,
                                      state,
                                      observed,
                                      ovs,
                                      ordered,
                                      PROBS,
                                      group = NULL,
                                      mc.reps,
                                      seed = NULL,
                                      standardize = TRUE) {

  parTable <- mcStdParTableFromState(
    std.info = std.info, state = state
  )

  sim <- simulateDataParTableStandardized(
    parTable = parTable,
    N = mc.reps,
    colsOVs = ovs,
    seed = seed
  )

  sim.ord <- mcOrdinalizeSimulatedData(
    sim = sim,
    ovs = ovs,
    ordered = ordered,
    PROBS = PROBS,
    group = group,
    standardize = standardize
  )

  observed.blocks <- mcSplitOrderedBlocks(
    data = observed,
    ovs = ovs,
    group = group
  )

  expected.blocks <- mcSplitOrderedBlocks(
    data = sim.ord,
    ovs = ovs,
    group = group
  )

  out <- vector("list", length(observed.blocks))
  for (i in seq_along(observed.blocks)) {
    obs.g <- observed.blocks[[i]]
    exp.g <- expected.blocks[[min(i, length(expected.blocks))]]

    sigma.obs <- stats::cov(obs.g, use = "pairwise.complete.obs")
    sigma.exp <- stats::cov(exp.g, use = "pairwise.complete.obs")

    mu.obs <- matrix(
      colMeans(obs.g, na.rm = TRUE), ncol = 1,
      dimnames = list(colnames(obs.g), "~1")
    )

    mu.exp <- matrix(
      colMeans(exp.g, na.rm = TRUE), ncol = 1,
      dimnames = list(colnames(exp.g), "~1")
    )

    out[[i]] <- list(
      sigma.all = sigma.exp,
      sigma.lv = NULL,
      sigma.ov = sigma.exp,
      mu.all = mu.exp,
      mu.lv = NULL,
      mu.ov = mu.exp,
      r2.all = NULL,
      r2.lv = NULL,
      r2.ov = NULL,
      res.all = NULL,
      res.lv = NULL,
      res.ov = NULL,
      lambda = NULL,
      gammaXi = NULL,
      gammaEta = NULL,
      psi = NULL,
      phi = NULL,
      theta = NULL,
      sigma.ord.observed = sigma.obs,
      sigma.ord.expected = sigma.exp,
      mu.ord.observed = mu.obs,
      mu.ord.expected = mu.exp,
      sim.ov.ord = exp.g
    )
  }

  out
}


mcSplitOrderedBlocks <- function(data, ovs, group = NULL) {
  if (is.null(group)) {
    return(list(as.data.frame(data)[, ovs, drop = FALSE]))
  }

  vals <- unique(data[[group]])
  stats::setNames(
    lapply(vals, function(g) {
      as.data.frame(data[data[[group]] == g, ovs, drop = FALSE])
    }),
    as.character(vals)
  )
}


mcStandardizedStateInfo <- function(fit) {
  solution <- standardizedSolutionCOEFS(
    fit,
    monte.carlo = FALSE
  )
  parTable <- as.data.frame(solution$parTable)
  parTable$.state.id <- getParTableLabels(parTable, labelCol = "label")
  parTable$.state.redundant <- FALSE

  lhs <- parTable$lhs
  op  <- parTable$op
  rhs <- parTable$rhs

  has.residual.cov <- mcHasResidualCovariances(parTable)
  keep <- (
    op %in% c("=~", "<~", "~", "~~") &
    !(op == "~~" & (grepl(":", lhs) | grepl(":", rhs)))
  )

  if (!has.residual.cov) {
    redundant <- (
      keep & (op == "~~" & lhs == rhs)
    )

    parTable$.state.redundant[redundant] <- TRUE
    keep <- keep & !redundant
  }

  ids <- unique(parTable$.state.id[keep])
  est <- stats::setNames(parTable$est, parTable$.state.id)
  est <- est[ids]

  vcov.free <- solution$vcov
  if (!is.null(vcov.free) && length(vcov.free)) {
    vcov.free <- expandVCOV(vcov.free, labels = ids)
  } else {
    vcov.free <- NULL
  }

  list(
    template = parTable,
    state = est,
    vcov.free = vcov.free,
    reduced = !has.residual.cov,
    has.residual.cov = has.residual.cov
  )
}


mcHasResidualCovariances <- function(parTable) {
  is.cov <- (
    parTable$op == "~~" & parTable$lhs != parTable$rhs &
    !grepl(":", parTable$lhs) & !grepl(":", parTable$rhs)
  )

  if (!any(is.cov))
    return(FALSE)

  inds <- getIndicators(parTable)
  etas <- getEtas(parTable, checkAny = FALSE)

  covs <- parTable[is.cov, , drop = FALSE]

  any(
    (covs$lhs %in% inds & covs$rhs %in% inds) |
    covs$lhs %in% etas | covs$rhs %in% etas
  )
}


mcExtractStandardizedState <- function(fit, std.info) {
  solution <- standardizedSolutionCOEFS(fit, monte.carlo = FALSE)
  parTable <- as.data.frame(solution$parTable)

  parTable$.state.id <- getParTableLabels(parTable, labelCol = "label")
  est <- stats::setNames(parTable$est, parTable$.state.id)
  out <- est[names(std.info$state)]
  names(out) <- names(std.info$state)

  out
}


mcStdParTableFromState <- function(std.info, state) {
  parTable <- std.info$template
  idx <- match(parTable$.state.id, names(state))
  hit <- !is.na(idx)
  parTable$est[hit] <- state[idx[hit]]

  if (isTRUE(std.info$reduced))
    parTable <- mcStdRefreshRedundant(parTable)

  parTable
}


mcStdRefreshRedundant <- function(parTable) {
  # parTable <- var_interactions(parTable)

  groups <- getGroupsParTable(addMissingGroups(parTable))
  out <- parTable

  for (g in groups) {
    mask.g <- out$group == g
    parTable.g <- out[mask.g, , drop = FALSE]
    if (!NROW(parTable.g)) next

    lVs.g <- getLVs(parTable.g)
    intTerms.g <- getIntTerms(parTable.g)
    etas.g <- getSortedEtas(parTable.g, isLV = FALSE)
    xis.g <- getXis(parTable.g, etas = etas.g, isLV = FALSE)
    indsLVs.g <- getIndsLVs(parTable.g, lVs = lVs.g)
    allInds.g <- unique(unlist(indsLVs.g))

    exogenous <- unique(c(xis.g, intTerms.g))
    for (x in exogenous) {
      sel <- parTable.g$lhs == x & parTable.g$rhs == x & parTable.g$op == "~~"
      if (any(sel)) parTable.g[sel, "est"] <- 1
    }

    for (eta in etas.g) {
      sel <- parTable.g$lhs == eta & parTable.g$rhs == eta & parTable.g$op == "~~"
      if (!any(sel)) next
      total <- calcVarParTable(eta, parTable = parTable.g)
      parTable.g[sel, "est"] <- pmax(parTable.g[sel, "est"] + (1 - total), 1e-8)
    }

    for (ind in allInds.g) {
      sel <- parTable.g$lhs == ind & parTable.g$rhs == ind & parTable.g$op == "~~"
      if (!any(sel)) next
      total <- calcVarParTable(ind, parTable = parTable.g, measurement.model = TRUE)
      parTable.g[sel, "est"] <- pmax(parTable.g[sel, "est"] + (1 - total), 1e-8)
    }

    out[mask.g, ] <- parTable.g
  }

  out
}


mcOverwriteFitWithStdState <- function(fit, state, vcov.free, std.info, calc.se,
                                       type.se = fit$type.se, zero.tol = 1e-10) {
  parTable <- mcStdParTableFromState(std.info = std.info, state = state)

  se <- stats::setNames(rep(NA_real_, length(state)), names(state))
  if (isTRUE(calc.se) && !is.null(vcov.free)) {
    ids.v <- intersect(names(state), rownames(vcov.free))
    if (length(ids.v)) {
      se[ids.v] <- sqrt(pmax(diag(vcov.free[ids.v, ids.v, drop = FALSE]), 0))
      se[which(se <= zero.tol)] <- NA_real_
    }
  }

  parTable$std.error <- NA_real_
  idx.se <- match(parTable$.state.id, names(se))
  hit.se <- !is.na(idx.se)
  parTable$std.error[hit.se] <- se[idx.se[hit.se]]

  parTable <- addZStatsParTable(parTable)
  parTable$.state.id <- NULL
  parTable$.state.redundant <- NULL

  fit$parTable <- modsemParTable(sortParTableDA(parTable, model = fit$model))
  fit$coefs.all <- stats::setNames(parTable$est, getParTableLabels(parTable, labelCol = "label", replace.dup = TRUE))
  fit$coefs.free <- state
  fit$vcov.all <- if (!is.null(vcov.free))
    expandVCOV(vcov.free, labels = names(fit$coefs.all))
  else NULL
  fit$vcov.free <- vcov.free
  fit$type.se <- if (isTRUE(calc.se)) type.se else "none"
  fit$type.estimates <- "standardized"
  fit
}


mcReplaceFreeVcovAll <- function(fit, vcov.free) {
  if (is.null(fit$vcov.all) || is.null(dimnames(fit$vcov.all)))
    return(fit)

  lav <- mcFreeLavLabels(fit = fit, theta = fit$theta)
  keep <- lav %in% rownames(fit$vcov.all)
  if (!any(keep)) return(fit)

  V <- fit$vcov.all
  V[lav[keep], lav[keep]] <- vcov.free[keep, keep, drop = FALSE]
  fit$vcov.all <- V
  fit
}


mcSEFromVcov <- function(fit, theta, vcov.free, zero.tol = 1e-10) {
  if (is.null(vcov.free))
    return(stats::setNames(rep(NA_real_, length(fit$coefs.all)),
                           names(fit$coefs.all)))

  lav <- mcFreeLavLabels(fit = fit, theta = theta)

  se <- sqrt(pmax(diag(vcov.free), 0))
  se[se < zero.tol] <- NA_real_

  stats::setNames(se, lav)
}


mcFreeLavLabels <- function(fit, theta) {
  lav <- fit$start.model$params$lavLabels
  if (!is.null(names(lav))) {
    out <- lav[names(theta)]
  } else {
    out <- lav[seq_along(theta)]
  }
  unname(out)
}


# Thresholds for the ordered indicators are the empirical quantiles of the
# model-implied (simulated) continuous indicator distributions at the observed
# cumulative category proportions, matching the cut points used by
# `mcOrdinalize()`. No normality is assumed; indicators of endogenous latent
# variables are non-normal whenever interaction effects are present. The
# bootstrap resamples the category counts at the observed (group) sample size,
# holding the simulated reference distribution fixed, so it does not propagate
# uncertainty from the model parameter estimates.
mcOrderedThresholdsBoot <- function(data, sim, ordered, group = NULL,
                                    calc.se = TRUE, boot.reps = 1000L,
                                    seed = NULL, zero.tol = 1e-10) {
  if (!is.null(seed) && exists(".Random.seed")) .Random.seed.orig <- .Random.seed
  else                                          .Random.seed.orig <- NULL

  on.exit({
    if (!is.null(.Random.seed.orig)) .Random.seed <<- .Random.seed.orig
  })

  if (!is.null(seed))
    set.seed(seed)

  if (is.null(group)) {
    groups <- list("1" = data)
  } else {
    vals <- unique(data[[group]])
    groups <- stats::setNames(lapply(vals, function(g) data[data[[group]] == g, ]),
                              as.character(vals))
  }

  out <- NULL
  V.out <- NULL
  for (g in seq_along(groups)) {
    data.g <- groups[[g]]
    sim.g <- as.data.frame(sim$OV[[min(g, length(sim$OV))]])

    for (col in ordered) {
      x <- as.ordered(data.g[[col]])
      tab <- as.numeric(table(x))
      K <- length(tab)
      n <- sum(tab)
      if (K < 2L) next

      x.sim <- sim.g[[col]]
      tau <- mcThresholdsFromCounts(tab, x.sim = x.sim)

      labs <- paste0(col, "|t", seq_len(K - 1L))
      if (g > 1L) # match the group suffix convention in `getParTableLabels()`
        labs <- paste0(labs, ".g", g)

      se.tau <- rep(NA_real_, length(tau))

      if (isTRUE(calc.se) && boot.reps >= 2L) {
        counts.b <- stats::rmultinom(boot.reps, size = n, prob = tab / n)
        tau.b <- apply(counts.b, MARGIN = 2L, FUN = mcThresholdsFromCounts,
                       x.sim = x.sim)

        if (is.null(dim(tau.b)))
          tau.b <- matrix(tau.b, nrow = 1L)

        V.tau <- stats::cov(t(tau.b))
        dimnames(V.tau) <- list(labs, labs)
        se.tau <- sqrt(pmax(diag(V.tau), 0))
        se.tau[which(se.tau <= zero.tol)] <- NA_real_
        V.out <- diagPartitionedMat(V.out, V.tau)
      }

      out <- rbind(out, data.frame(
        lhs = col,
        op = "|",
        rhs = paste0("t", seq_len(K - 1L)),
        label = "",
        group = g,
        est = tau,
        std.error = se.tau,
        stringsAsFactors = FALSE,
        row.names = NULL
      ))
    }
  }

  if (!is.null(out)) {
    out <- addZStatsParTable(out)
    attr(out, "vcov.thresholds") <- V.out
  }
  out
}


# Same quantile computation as the break points in `mcOrdinalize()`.
mcThresholdsFromCounts <- function(counts, x.sim) {
  K <- length(counts)
  C <- cumsum(counts / sum(counts))[-K]
  eps <- .Machine$double.eps^0.5
  C <- pmin(pmax(C, eps), 1 - eps)
  stats::quantile(x.sim, probs = C, names = FALSE, type = 7, na.rm = TRUE)
}


mcAppendThresholds <- function(fit, thresholds) {
  if (is.null(thresholds) || !NROW(thresholds))
    return(fit)

  V.tau <- attr(thresholds, "vcov.thresholds")
  parTable <- as.data.frame(fit$parTable)
  missing.cols <- setdiff(colnames(parTable), colnames(thresholds))
  thresholds[missing.cols] <- NA
  thresholds <- thresholds[colnames(parTable)]

  fit$parTable <- modsemParTable(sortParTableDA(rbind(parTable, thresholds),
                                                model = fit$model))
  coef.t <- stats::setNames(thresholds$est, getParTableLabels(thresholds))
  fit$coefs.all <- c(fit$coefs.all, coef.t)

  if (!is.null(V.tau))
    fit$vcov.all <- diagPartitionedMat(fit$vcov.all, V.tau)

  fit
}


`%||%` <- function(x, y) if (is.null(x)) y else x

devtools::load_all()

tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

corrected <- relcorr_single_item(syntax = tpb_uk, data = TPB_UK,
                                 scale.corrected = TRUE)
print(corrected)
syntax <- corrected$syntax
data   <- corrected$data

est_dca <- modsem(syntax, data = data, method = "dblcent",
                  rcs.res.cov.xz = TRUE)
est_lms <- modsem(syntax, data = data, method="lms", nodes=32)

est_dca <- modsem(tpb_uk, data = TPB_UK, method = "dblcent", rcs = TRUE,
                  rcs.mc.reps = 3e4)
est_lms <- modsem(tpb_uk, data = TPB_UK, method = "lms", nodes = 32, rcs = TRUE)
summary(est_dca)
summary(est_lms)

est_dca <- modsem(tpb_uk, data = TPB_UK, method = "dblcent", rcs = TRUE,
                  rcs.choose = c("ATT", "SN", "PBC", "INT"))
est_ca <- modsem(tpb_uk, data = TPB_UK, method = "ca", rcs = TRUE)

if (FALSE) {
Data_C1 <- read.csv(file = '~/Downloads/Example_C1.csv')

# Specify the measurement model (Example.C1.Model.Measure)
Example.C1.Model.Measure <- '
  JDemand =~ JobD1 + JobD2 + JobD3
  JResource =~ JobRes1+ JobRes2 + JobRes3 + JobRes4 + JobRes5 + JobRes6
  HomeSick =~ HomeS1 + HomeS2 + HomeS3 + HomeS4 + HomeS5 + HomeS6 + HomeS7 + HomeS8 + HomeS9 + HomeS10 +
              HomeS11 + HomeS12 + HomeS13 + HomeS14 + HomeS15 + HomeS16 + HomeS17 + HomeS18 + HomeS19 + HomeS20
  EStability =~ EmoStab1 + EmoStab2 + EmoStab3 + EmoStab4 + EmoStab5 + EmoStab6
  Openness =~ Open1 + Open2 + Open3 + Open4 + Open5 + Open6
  Performance =~ TaskP1 + TaskP2 + TaskP3

  # Structural model #
  Performance ~ a1*JResource + HomeSick + EStability + Openness
  Performance ~ z1*JResource:HomeSick + w1a*JResource:EStability + HomeSick:EStability
  Performance ~ w1b*JResource:Openness + HomeSick:Openness

  # Control variable #
  Performance ~ JDemand
'
corrected <- relcorr_single_item(Example.C1.Model.Measure, Data_C1)
print(corrected)
#> Average Variance Extracted:
#>   JDemand:        0.663
#>   JResource:      0.781
#>   HomeSick:       0.919
#>   EStability:     0.781
#>   Openness:       0.778
#>   Performance:    0.658
#>
#> Construct Reliability:
#>   JDemand:        0.732
#>   JResource:      0.735
#>   HomeSick:       0.861
#>   EStability:     0.737
#>   Openness:       0.705
#>   Performance:    0.697
#>
#> Generated Syntax:
#>   Performance ~ a1*JResource
#>   Performance ~ HomeSick
#>   Performance ~ EStability
#>   Performance ~ Openness
#>   Performance ~ z1*JResource:HomeSick
#>   Performance ~ w1a*JResource:EStability
#>   Performance ~ HomeSick:EStability
#>   Performance ~ w1b*JResource:Openness
#>   Performance ~ HomeSick:Openness
#>   Performance ~ JDemand
#>   JDemand =~ 1*composite_JDemand_
#>   JResource =~ 1*composite_JResource_
#>   HomeSick =~ 1*composite_HomeSick_
#>   EStability =~ 1*composite_EStability_
#>   Openness =~ 1*composite_Openness_
#>   Performance =~ 1*composite_Performance_
#>   composite_JDemand_ ~~ 0.124747077738605*composite_JDemand_
#>   composite_JResource_ ~~ 0.0769133338087597*composite_JResource_
#>   composite_HomeSick_ ~~ 0.0238661172257718*composite_HomeSick_
#>   composite_EStability_ ~~ 0.254605663714866*composite_EStability_
#>   composite_Openness_ ~~ 0.287668887959575*composite_Openness_
#>   composite_Performance_ ~~ 0.0705913680595234*composite_Performance_
#>
#> Generated Items:
#>   'data.frame': 422 obs. of  6 variables:
#>     $ composite_JDemand_    : num  2.67 3.33 2.33 2.67 4 ...
#>     $ composite_JResource_  : num  2.67 3.5 3.17 3.5 3.33 ...
#>     $ composite_HomeSick_   : num  2.55 3 3.1 3 3.1 3.05 3.35 3.35 2.75 3.55 ...
#>     $ composite_EStability_ : num  5 5.5 4.5 5.17 4 ...
#>     $ composite_Openness_   : num  3.17 4.83 4.83 4.33 5.33 ...
#>     $ composite_Performance_: num  3 3.33 3 4 3.33 ...

est_dca <- modsem(corrected$syntax, data = corrected$data, method="dblcent")  # method = dblcent, rca, uca, ca, pind
est_lms <- modsem(corrected$syntax, data=corrected$data, method="lms")
}

devtools::load_all()


m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est_dca <- twostep(m1, data = oneInt, method = "dblcent")
est_lms <- twostep(m1, data = oneInt, method = "lms")
summary(est_dca)
summary(est_lms)


tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

tpb_lms <- twostep(tpb, data = TPB, method = "lms", nodes = 32)
tpb_dca <- suppressWarnings(twostep(tpb, data = TPB, method = "dblcent"))
summary(tpb_lms)
suppressWarnings(summary(tpb_dca))


tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 # Causal Relationsships
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

uk_lms <- twostep(tpb_uk, TPB_UK, method = "lms", nodes = 32, adaptive.quad = TRUE)
uk_qml <- twostep(tpb_uk, TPB_UK, method = "qml")
uk_dblcent <- twostep(tpb_uk, TPB_UK, method = "dblcent")
summary(uk_lms)
summary(uk_dblcent)

# Functions
specifyModelDA_Group <- function(syntax = NULL,
                                 data = NULL,
                                 method = "lms",
                                 m = 16,
                                 cov.syntax = NULL,
                                 double = FALSE,
                                 parTable = NULL,
                                 parTableCovModel = NULL,
                                 auto.fix.first = TRUE,
                                 auto.fix.single = TRUE,
                                 createTheta = TRUE,
                                 mean.observed = TRUE,
                                 standardize.inp = FALSE,
                                 standardize.out = FALSE,
                                 checkModel = TRUE,
                                 quad.range = Inf,
                                 adaptive.quad = FALSE,
                                 adaptive.frequency = 3,
                                 missing = "complete",
                                 orthogonal.x = FALSE,
                                 orthogonal.y = FALSE,
                                 cluster = NULL,
                                 sampling.weights = NULL,
                                 structovs = NULL,
                                 fix.composite.var = TRUE) {
  if (is.null(parTable) && !is.null(syntax)) parTable <- modsemify(syntax)
  mod_stopif(is.null(parTable), "No parTable found")

  checkParTableDA(parTable, method = method)
  # additions to lavaan-syntax for optimizer
  lavOptimizerSyntaxAdditions <- ""

  # General Information
  higherOrderLVs <- getHigherOrderLVs(parTable)
  indsHigherOrderLVs <- getIndsLVs(parTable, lVs = higherOrderLVs, isOV = FALSE)
  ovs <- getOVs(parTable)
  composites <- getComposites(parTable)

  # Check for higher order composites
  mod_stopif(any(composites %in% higherOrderLVs),
    "Higher order composites are not supported (yet)!"
  )

  # endogenous variables (etas)
  etas    <- getSortedEtas(parTable, isLV = TRUE, checkAny = TRUE)
  numEtas <- length(etas)

  indsEtas    <- getIndsLVs(parTable, lVs = etas, isOV = TRUE, ovs = ovs)
  numIndsEtas <- vapply(indsEtas, FUN.VALUE = vector("integer", 1L),
                        FUN = length)
  allIndsEtas    <- unique(unlist(indsEtas))
  numAllIndsEtas <- length(allIndsEtas)

  # exogenous variables (xis) and interaction terms
  checkLhsIntTerms(parTable)
  intTerms      <- getIntTermRows(parTable)
  varsInts      <- getVarsInts(intTerms)
  allVarsInInts <- unique(unlist(varsInts))
  xis           <- getXis(parTable, checkAny = TRUE)
  numXis        <- length(xis)

  omegaAndSortedXis <- sortXisConstructOmega(
    xis = xis, varsInts = varsInts, etas = etas, intTerms = intTerms,
    method = method, double = double, structovs = structovs,
    composites = composites
  )

  xis <- omegaAndSortedXis$sortedXis # get sorted xis according to interaction terms
  nonLinearXis <- omegaAndSortedXis$nonLinearXis

  indsXis    <- getIndsLVs(parTable, lVs = xis, isOV = TRUE, ovs = ovs)
  numIndsXis <- vapply(indsXis, FUN.VALUE = vector("integer", 1L),
                       FUN = length)
  allIndsXis    <- unique(unlist(indsXis))
  numAllIndsXis <- length(allIndsXis)

  # composite variables
  compositeXis  <- intersect(xis, composites)
  compositeEtas <- intersect(etas, composites)
  indsCompXis   <- indsXis[compositeXis]
  indsCompEtas  <- indsEtas[compositeEtas]

  # composite variables are only defined for lms
  mod_stopif(length(composites) && method == "qml",
    paste0("Composite constructs are not allowed with `method=\"qml\"` (yet)!\n",
    "Try `method=\"lms\"` instead.")
  )

  # clean data
  data.cleaned <- prepDataModsemDA(
    data             = data,
    allIndsXis       = allIndsXis,
    allIndsEtas      = allIndsEtas,
    missing          = missing,
    cluster          = cluster,
    sampling.weights = sampling.weights
  )

  # measurement model x
  if (method == "qml") {
    listLambdaX <- constructLambda(
      xis, indsXis, parTable = parTable,
      auto.fix.first = auto.fix.first
    )

  } else {
    listLambdaX <- constructLambda(
      c(xis, etas), c(indsXis, indsEtas),
      parTable = parTable,
      auto.fix.first = auto.fix.first
    )
  }

  lambdaX      <- listLambdaX$numeric
  labelLambdaX <- listLambdaX$label

  # Composite measurement model
  if (length(composites)) {

    listW <- constructLambda(
      c(xis, etas), c(indsXis, indsEtas),
      parTable = parTable,
      auto.fix.first = auto.fix.first,
      mode = "b"
    )

    listT <- constructT(c(xis, etas), c(indsXis, indsEtas),
      parTable = parTable, fix.composite.var = fix.composite.var,
      data = data.cleaned, missing = missing,
      sampling.weights = sampling.weights
    )

  } else {
    listW <- constructLambda(
      NULL, NULL, parTable = parTable,
      auto.fix.first = auto.fix.first
    )

    listT <- constructT(
      NULL, NULL, parTable = parTable
    )
  }

  W      <- listW$numeric
  T      <- listT$numeric
  labelW <- listW$label
  labelT <- listT$label

  # Intercepts
  if (method == "qml") {
    listTauX <- constructTau(
      xis, indsXis, parTable = parTable,
      mean.observed = mean.observed
    )

  } else {
    listTauX <- constructTau(
      c(xis, etas), c(indsXis, indsEtas),
      parTable = parTable,
      mean.observed = mean.observed
    )
  }

  tauX      <- listTauX$numeric
  labelTauX <- listTauX$label
  lavOptimizerSyntaxAdditions <- paste0(lavOptimizerSyntaxAdditions,
                                        listTauX$syntaxAdditions)

  checkResCovX_Y(parTable = parTable, allIndsXis = allIndsXis,
                 allIndsEtas = allIndsEtas, method = method)

  if (method == "qml") {
    listThetaDelta <- constructTheta(
      xis, indsXis, parTable = parTable,
      auto.fix.single = auto.fix.single
    )

  } else {
    listThetaDelta <- constructTheta(
      c(xis, etas), c(indsXis, indsEtas),
      parTable = parTable,
      auto.fix.single = auto.fix.single
    )
  }

  thetaDelta      <- listThetaDelta$numeric
  thetaLabelDelta <- listThetaDelta$label

  # measurement model y
  if (method == "qml") {
    listLambdaY <- constructLambda(
      etas, indsEtas, parTable = parTable,
      auto.fix.first = auto.fix.first
    )

  } else {
    listLambdaY <- constructLambda(
      NULL, NULL, parTable = parTable,
      auto.fix.first = auto.fix.first
    )
  }

  lambdaY      <- listLambdaY$numeric
  labelLambdaY <- listLambdaY$label

  if (method == "qml") {
    listTauY <- constructTau(
      etas, indsEtas, parTable = parTable,
      mean.observed = mean.observed
    )

  } else {
    listTauY <- constructTau(
      NULL, NULL, parTable = parTable,
      mean.observed = mean.observed
    )
  }

  tauY      <- listTauY$numeric
  labelTauY <- listTauY$label
  lavOptimizerSyntaxAdditions <- paste0(lavOptimizerSyntaxAdditions,
                                        listTauY$syntaxAdditions)

  if (method == "qml") {
    listThetaEpsilon <- constructTheta(
      etas, indsEtas, parTable = parTable,
      auto.fix.single = auto.fix.single
    )

  } else {
    listThetaEpsilon <- constructTheta(
      NULL, NULL, parTable = parTable,
      auto.fix.single = auto.fix.single
    )
  }

  thetaEpsilon      <- listThetaEpsilon$numeric
  thetaLabelEpsilon <- listThetaEpsilon$label

  # structural model
  Ieta         <- diag(numEtas) # used for (B^-1 = (Ieta - gammaEta)^-1)
  listGammaXi  <- constructGamma(etas, xis, parTable = parTable,
                                 auto.fix.first = auto.fix.first)
  gammaXi      <- listGammaXi$numeric
  labelGammaXi <- listGammaXi$label

  listGammaEta  <- constructGamma(etas, etas, parTable = parTable,
                                  auto.fix.first = auto.fix.first)
  gammaEta      <- listGammaEta$numeric
  labelGammaEta <- listGammaEta$label

  # covariance matrices
  listPsi  <- constructPsi(etas, parTable = parTable, orthogonal.y = orthogonal.y)
  psi      <- listPsi$numeric
  labelPsi <- listPsi$label

  listPhi <- constructPhi(xis, method = method, cov.syntax = cov.syntax,
                          parTableCovModel = parTableCovModel,
                          parTable = parTable, orthogonal.x = orthogonal.x)
  phi      <- listPhi$numeric
  labelPhi <- listPhi$label

  listA <- constructA(xis, method = method, cov.syntax = cov.syntax,
                      parTableCovModel = parTableCovModel,
                      parTable = parTable, orthogonal.x = orthogonal.x)
  A      <- listA$numeric
  labelA <- listA$label

  listCovZetaXi <- constructCovZetaXi(xis, etas, method = method,
                                       parTable = parTable)
  covZetaXi      <- listCovZetaXi$numeric
  labelCovZetaXi <- listCovZetaXi$label

  # mean etas
  listAlpha <- constructAlpha(etas, parTable = parTable,
                              mean.observed = mean.observed)
  alpha      <- listAlpha$numeric
  labelAlpha <- listAlpha$label

  # mean xis
  listBeta0 <- constructAlpha(xis, parTable = parTable,
                              mean.observed = mean.observed)
  beta0      <- listBeta0$numeric
  labelBeta0 <- listBeta0$label

  # quadratic terms
  listOmegaEtaXi  <- omegaAndSortedXis$omegaEtaXi
  omegaEtaXi      <- listOmegaEtaXi$numeric
  labelOmegaEtaXi <- listOmegaEtaXi$label

  listOmegaXiXi  <- omegaAndSortedXis$omegaXiXi
  omegaXiXi      <- listOmegaXiXi$numeric
  labelOmegaXiXi <- listOmegaXiXi$label

  # matrices for scaling variables in qml
  selectScalingY <- selectScalingY(lambdaY, method = method)
  selectBetaRows <- selectBetaRows(lambdaY, method = method)
  emptyR <- constructR(etas, indsEtas, lambdaY, method = method)
  fullR  <- constructFullR(etas, indsEtas, lambdaY, method = method)

  latentEtas <- getLatentEtasQml(indsEtas, method = method)
  colsU      <- getColsU(etas, indsEtas, lambdaY, method = method)

  fullL2      <- constructFullL2(colsU, etas = etas, method = method)
  selectSubL2 <- getSelectSubL2(fullL2, colsU = colsU, latentEtas = latentEtas,
                                method = method)
  fullSigma2ThetaEpsilon <- constructFullSigma2ThetaEpsilon(psi, method = method)
  selectSubSigma2ThetaEpsilon <-
    getSelectSubSigma2ThetaEpsilon(fullSigma2ThetaEpsilon, latentEtas = latentEtas,
                                   method = method)
  fullU <- constructFullU(fullL2 = fullL2, N = data.cleaned$n, etas = etas, method = method)

  scalingInds <- getScalingInds(indsEtas, R = emptyR, latentEtas = latentEtas,
                                method = method)

  # 1 = truly latent variables, 2 = latent variables with single indicators
  selectThetaEpsilon1 <- selectThetaEpsilon1(indsEtas, thetaEpsilon,
                                             scalingInds, method = method)
  selectThetaEpsilon2 <- selectThetaEpsilon2(indsEtas, thetaEpsilon,
                                             scalingInds, method = method)
  subThetaEpsilon1 <- constructSubThetaEpsilon1(indsEtas, thetaEpsilon,
                                                scalingInds, method = method)
  subThetaEpsilon2 <- constructSubThetaEpsilon2(indsEtas, thetaEpsilon,
                                                scalingInds, method = method)

  covModel <- covModel(
    cov.syntax, method = method, parTable = parTableCovModel,
    xis.main = xis, parTable.main = parTable,
    orthogonal.x = orthogonal.x, orthogonal.y = orthogonal.y
  )

  # list of matrices
  matrices <- list(
    lambdaX      = lambdaX,
    lambdaY      = lambdaY,
    gammaXi      = gammaXi,
    gammaEta     = gammaEta,
    thetaDelta   = thetaDelta,
    thetaEpsilon = thetaEpsilon,
    W            = W,
    T            = T,
    phi          = phi,
    A            = A,
    covZetaXi    = covZetaXi,
    Ieta         = Ieta,
    psi          = psi,
    tauX         = tauX,
    tauY         = tauY,
    alpha        = alpha,
    beta0        = beta0,
    omegaEtaXi   = omegaEtaXi,
    omegaXiXi    = omegaXiXi,

    selectScalingY      = selectScalingY,
    selectThetaEpsilon1 = selectThetaEpsilon1,
    selectThetaEpsilon2 = selectThetaEpsilon2,
    selectBetaRows      = selectBetaRows,

    emptyR = emptyR,
    fullR  = fullR,

    fullSigma2ThetaEpsilon      = fullSigma2ThetaEpsilon,
    selectSubSigma2ThetaEpsilon = selectSubSigma2ThetaEpsilon,

    fullL2      = fullL2,
    selectSubL2 = selectSubL2,

    fullU = fullU,
    colsU = colsU,
    colsR = colnames(emptyR),
    rowsR = rownames(emptyR),

    subThetaEpsilon1 = subThetaEpsilon1,
    subThetaEpsilon2 = subThetaEpsilon2
  )

  labelMatrices <- list(
    lambdaX      = labelLambdaX,
    lambdaY      = labelLambdaY,
    gammaXi      = labelGammaXi,
    gammaEta     = labelGammaEta,
    thetaDelta   = thetaLabelDelta,
    thetaEpsilon = thetaLabelEpsilon,
    W            = labelW,
    T            = labelT,

    phi   = labelPhi,
    A     = labelA,
    covZetaXi = labelCovZetaXi,
    psi   = labelPsi,
    tauX  = labelTauX,
    tauY  = labelTauY,
    alpha = labelAlpha,
    beta0 = labelBeta0,

    omegaEtaXi = labelOmegaEtaXi,
    omegaXiXi  = labelOmegaXiXi
  )

  k <- omegaAndSortedXis$k
  quad <- quadrature(m, k, quad.range = quad.range, adaptive = adaptive.quad,
                     adaptive.frequency = adaptive.frequency)

  model <- list(
    info = list(
      N             = data.cleaned$n,
      ncol          = data.cleaned$k,
      xis           = xis,
      etas          = etas,
      numXis        = numXis,
      numEtas       = numEtas,
      indsXis       = indsXis,
      indsEtas      = indsEtas,
      allIndsXis    = allIndsXis,
      allIndsEtas   = allIndsEtas,
      varsInts      = varsInts,
      latentEtas    = latentEtas,
      scalingInds   = scalingInds,
      kOmegaEta     = getK_NA(omegaEtaXi, labelOmegaEtaXi),
      nonLinearXis  = nonLinearXis,
      mean.observed = mean.observed,
      orthogonal.x  = orthogonal.x,
      orthogonal.y  = orthogonal.y,

      has.interaction     = NROW(intTerms) > 0L,
      higherOrderLVs      = higherOrderLVs,
      indsHigherOrderLVs  = indsHigherOrderLVs,
      hasComposites       = length(composites) > 0L,
      fixed.composite.var = fix.composite.var,

      lavOptimizerSyntaxAdditions = lavOptimizerSyntaxAdditions
    ),

    data          = data.cleaned,
    data.raw      = data,
    quad          = quad,
    matrices      = matrices,
    labelMatrices = labelMatrices,
    syntax        = syntax,
    cov.syntax    = cov.syntax,
    parTable      = parTable,
    covModel      = covModel,
    lavaan.fit    = NULL
  )

  if (checkModel)
    preCheckModel(model = model, covModel = covModel, method = method,
                  missing = missing)

  model
}


specifyModelDA <- function(..., group.info, createTheta = TRUE) {
  args <- list(...)

  n.groups <- group.info$n.groups
  mod_stopif(n.groups < 1L, "Invalid grouping structure supplied.")

  parTable    <- group.info$parTable
  parTableCov <- group.info$parTableCov
  structovs   <- group.info$structovs
  group.col   <- parTable$group

  mod_stopif(is.null(group.col) || max(group.col) != n.groups,
         "Number of group-specific parameter tables does not match number of groups.")

  submodels <- vector("list", length = n.groups)

  for (g in seq_len(n.groups)) {
    args.g <- args

    if (!is.null(group.info$data))
      args.g$data <- group.info$data[group.info$indices[[g]], , drop = FALSE]
    else
      args.g <- addNamedNullField(args.g, field = "data")

    args.g$parTable <- parTable[parTable$group == g, , drop = FALSE]
    args.g$structovs <- structovs

    if (!is.null(parTableCov))
      args.g$parTableCovModel <- parTableCov[parTableCov$group == g, , drop = FALSE]
    else
      args.g <- addNamedNullField(args.g, field = "parTableCovModel")

    submodel.g <- do.call(specifyModelDA_Group, args.g)
    submodel.g$info$group <- group.info$levels[[g]]
    submodel.g$info$n.groups <- 1L

    submodels[[g]] <- submodel.g
  }

  has.ov.interaction <- length(group.info$ovIntTerms) > 0

  model <- list(
    models   = submodels,
    syntax   = args$syntax,
    data.raw = group.info$data.raw,
    parTable = parTable,
    info     = list(
      n.groups      = n.groups,
      group.levels  = group.info$levels,
      group.info    = group.info,

      # Constants across groups
      xis           = submodels[[1L]]$info$xis,
      etas          = submodels[[1L]]$info$etas,
      numXis        = submodels[[1L]]$info$numXis,
      numEtas       = submodels[[1L]]$info$numEtas,
      indsXis       = submodels[[1L]]$info$indsXis,
      indsEtas      = submodels[[1L]]$info$indsEtas,
      allIndsXis    = submodels[[1L]]$info$allIndsXis,
      allIndsEtas   = submodels[[1L]]$info$allIndsEtas,
      varsInts      = submodels[[1L]]$info$varsInts,
      latentEtas    = submodels[[1L]]$info$latentEtas,
      scalingInds   = submodels[[1L]]$info$scalingInds,
      kOmegaEta     = submodels[[1L]]$info$kOmegaEta,
      nonLinearXis  = submodels[[1L]]$info$nonLinearXis,
      mean.observed = submodels[[1L]]$info$mean.observed,
      orthogonal.x  = submodels[[1L]]$info$orthogonal.x,
      orthogonal.y  = submodels[[1L]]$info$orthogonal.y,

      has.interaction     = submodels[[1L]]$info$has.interaction || has.ov.interaction,
      higherOrderLVs      = submodels[[1L]]$info$higherOrderLVs,
      indsHigherOrderLVs  = submodels[[1L]]$info$indsHigherOrderLVs,
      hasComposites       = submodels[[1L]]$info$hasComposites,
      fixed.composite.var = submodels[[1L]]$info$fixed.composite.var,

      lavOptimizerSyntaxAdditions = submodels[[1L]]$info$lavOptimizerSyntaxAdditions
    ),

    params = list()
  )

  # Currenlty we assume covModel has an uniform structure
  model$params$constrExprs <- getConstrExprs(parTable, model$models[[1L]]$covModel$parTable)

  if (createTheta) {
    params <- createTheta(model, parTable.in = parTable)
    model$params[names(params)] <- params

    # TODO: Remove occurences of `model$theta`, and replace them with `model$params$theta`
    model$theta <- params$theta # an ugly design decision, that was made at the very start

    model$params$bounds <- getParamBounds(model)
    model$params$gradientStruct <- getGradientStruct(
      model, theta = params$theta, method = args$method
    )
  }

  model
}


matrixToParTable <- function(matrixNA, matrixEst, matrixSE, matrixLabel,
                             op = "=~", rowsLhs = TRUE, symmetric = FALSE) {
  if (symmetric) {
    matrixNA[upper.tri(matrixNA)]       <- 0
    matrixLabel[upper.tri(matrixLabel)] <- ""
  }

  if (!rowsLhs) {
    matrixNA    <- t(matrixNA)
    matrixEst   <- t(matrixEst)
    matrixSE    <- t(matrixSE)
    matrixLabel <- t(matrixLabel)
  }

  parTable <- NULL
  for (lhs in rownames(matrixEst)) {
    for (rhs in colnames(matrixEst)) {
      if (!is.na(matrixNA[lhs, rhs]) && matrixLabel[lhs, rhs] == "") next
      newRow <- data.frame(lhs = lhs, op = op, rhs = rhs,
                           label = matrixLabel[lhs, rhs],
                           est = matrixEst[lhs, rhs],
                           std.error = matrixSE[lhs, rhs])
      parTable <- rbind(parTable, newRow)
    }
  }
  parTable
}


interceptsToParTable <- function(matrixNA, matrixEst, matrixSE, matrixLabel) {
  parTable <- matrixToParTable(matrixNA, matrixEst, matrixSE, matrixLabel,
                               op = "~1", rowsLhs = TRUE)
  if (!is.null(parTable)) parTable$rhs <- ""
  parTable
}


omegaToParTable <- function(omegaNA, omegaEst, omegaSE, omegaLabel, parTable.in = NULL) {
  rows <- rownames(omegaEst)
  cols <- colnames(omegaEst)

  C <- \(x, y) sprintf("%s:%s", x, y)
  getIntTerm <- function(lhs, rhs) {
    xz <- C(lhs, rhs)
    zx <- C(rhs, lhs)

    if (is.null(parTable.in))         xz
    else if (xz %in% parTable.in$rhs) xz
    else                              zx
  }

  parTable <- NULL
  for (row in rows) for (col in cols) {
    if (!is.na(omegaNA[row, col]) && omegaLabel[row, col] == "") next
    eta     <- getEtaRowLabelOmega(row)
    x       <- getXiRowLabelOmega(row)
    intTerm <- getIntTerm(lhs = x, rhs = col)

    newRow <- data.frame(lhs = eta, op = "~", rhs = intTerm,
                         label = omegaLabel[row, col], est = omegaEst[row, col],
                         std.error = omegaSE[row, col])
    parTable <- rbind(parTable, newRow)
  }
  parTable
}


mainModelToParTable <- function(finalModel, method = "lms") {
  matricesEst   <- finalModel$matrices
  matricesSE    <- finalModel$matricesSE
  matricesNA    <- finalModel$matricesNA
  matricesLabel <- finalModel$labelMatrices
  parTable.in   <- finalModel$parTable

  if (is.null(matricesSE)) matricesSE <- matricesNA

  etas     <- finalModel$info$etas
  numXis   <- finalModel$info$numXis
  parTable <- NULL

  # Coefficients Measurement Model
  newRows <- matrixToParTable(matricesNA$lambdaX,
                              matricesEst$lambdaX,
                              matricesSE$lambdaX,
                              matricesLabel$lambdaX,
                              op = "=~",
                              rowsLhs = FALSE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$lambdaY,
                              matricesEst$lambdaY,
                              matricesSE$lambdaY,
                              matricesLabel$lambdaY,
                              op = "=~",
                              rowsLhs = FALSE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$W,
                              matricesEst$W,
                              matricesSE$W,
                              matricesLabel$W,
                              op = "<~",
                              rowsLhs = FALSE)
  parTable <- rbind(parTable, newRows)

  # coefficients Structural Model
  newRows <- matrixToParTable(matricesNA$gammaXi,
                              matricesEst$gammaXi,
                              matricesSE$gammaXi,
                              matricesLabel$gammaXi,
                              op = "~",
                              rowsLhs = TRUE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$gammaEta,
                              matricesEst$gammaEta,
                              matricesSE$gammaEta,
                              matricesLabel$gammaEta,
                              op = "~",
                              rowsLhs = TRUE)
  parTable <- rbind(parTable, newRows)

  # interaction effects
  newRows <- omegaToParTable(matricesNA$omegaXiXi,
                             matricesEst$omegaXiXi,
                             matricesSE$omegaXiXi,
                             matricesLabel$omegaXiXi,
                             parTable.in = parTable.in)
  parTable <- rbind(parTable, newRows)

  newRows <- omegaToParTable(matricesNA$omegaEtaXi,
                             matricesEst$omegaEtaXi,
                             matricesSE$omegaEtaXi,
                             matricesLabel$omegaEtaXi,
                             parTable.in = parTable.in)
  parTable <- rbind(parTable, newRows)

  # Intercepts
  newRows <- interceptsToParTable(matricesNA$tauX,
                                  matricesEst$tauX,
                                  matricesSE$tauX,
                                  matricesLabel$tauX)
  parTable <- rbind(parTable, newRows)

  newRows <- interceptsToParTable(matricesNA$tauY,
                                  matricesEst$tauY,
                                  matricesSE$tauY,
                                  matricesLabel$tauY)
  parTable <- rbind(parTable, newRows)

  newRows <- interceptsToParTable(matricesNA$alpha,
                                  matricesEst$alpha,
                                  matricesSE$alpha,
                                  matricesLabel$alpha)
  parTable <- rbind(parTable, newRows)

  newRows <- interceptsToParTable(matricesNA$beta0,
                                  matricesEst$beta0,
                                  matricesSE$beta0,
                                  matricesLabel$beta0)
  parTable <- rbind(parTable, newRows)

  # Residual (co) variances Measurement Model
  newRows <- matrixToParTable(matricesNA$thetaDelta,
                              matricesEst$thetaDelta,
                              matricesSE$thetaDelta,
                              matricesLabel$thetaDelta,
                              op = "~~", rowsLhs = TRUE,
                              symmetric = TRUE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$thetaEpsilon,
                              matricesEst$thetaEpsilon,
                              matricesSE$thetaEpsilon,
                              matricesLabel$thetaEpsilon,
                              op = "~~", rowsLhs = TRUE,
                              symmetric = TRUE)
  parTable <- rbind(parTable, newRows)

  # Composite indicator (co-) variances
  newRows <- matrixToParTable(matricesNA$T,
                              matricesEst$T,
                              matricesSE$T,
                              matricesLabel$T,
                              op = "~~", rowsLhs = TRUE,
                              symmetric = TRUE)
  parTable <- rbind(parTable, newRows)

  # (Co) variances Structural Model
  if (method == "lms") {
    phiNA <- matricesNA$A
    phiEst <- matricesEst$phi
    phiSE <- matricesSE$A
    phiLabel <- matricesLabel$A
  } else if (method == "qml") {
    phiNA <- matricesNA$phi
    phiEst <- matricesEst$phi
    phiSE <- matricesSE$phi
    phiLabel <- matricesLabel$phi
  }

  newRows <- matrixToParTable(phiNA,
                              phiEst,
                              phiSE,
                              phiLabel,
                              op = "~~",
                              rowsLhs = FALSE,
                              symmetric = TRUE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$covZetaXi,
                              matricesEst$covZetaXi,
                              matricesSE$covZetaXi,
                              matricesLabel$covZetaXi,
                              op = "~~",
                              rowsLhs = TRUE)
  parTable <- rbind(parTable, newRows)

  newRows <- matrixToParTable(matricesNA$psi,
                              matricesEst$psi,
                              matricesSE$psi,
                              matricesLabel$psi,
                              op = "~~",
                              rowsLhs = FALSE,
                              symmetric = TRUE)

  parTable <- rbind(parTable, newRows)

  parTable <- lapplyDf(parTable, FUN = function(x) replace(x, x == -999, NA))
  parTable
}


customParamsToParTable <- function(model, coefs, se) {
  parTable <- model$parTable
  custom   <- parTable[parTable$op == ":=", ]

  if (!NROW(custom$lhs)) return(NULL)

  out <- NULL
  for (i in seq_len(NROW(custom))) {
    lhs <- custom[i, "lhs"]
    rhs <- custom[i, "rhs"]

    se.i <- tryCatch(se[[lhs]], error = \(e) NA)
    newRow <- data.frame(lhs = lhs, op = ":=", rhs = rhs, label = lhs,
                         group = 0L, est = coefs[[lhs]], std.error = se.i)
    out <- rbind(out, newRow)
  }

  out
}


modelToParTable <- function(model, coefs = NULL, se = NULL, method = "lms", calc.se = TRUE, group = 1L) {
  parTable <- rbind(covModelToParTable(model, method = method),
                    mainModelToParTable(model, method = method))

  colsOut <- c("lhs", "op", "rhs", "label", "group", "est", "std.error")
  parTable$group <- group
  parTable <- parTable[colsOut]

  if (!is.null(coefs) && !is.null(se) && !is.null(names(se))) {
    # this is ugly but should work...
    # due to how values are read from the matrices, std.errors are overwritten
    # by the custom parameter-values (e.g., 'X=~a*x1; a==1.2' results in a std.error of 1.2, when it should be 0)
    isLabelled <- parTable$label != ""
    labels     <- parTable[isLabelled, "label"]
    parTable[isLabelled, "std.error"] <- se[labels]
    # if the std.error of a labelled parameter is 0, it is invariant, and should be NA
    # NB: there is a very small chance that a std.error of 0 is caused by a rounding error
    zeroStdError <- parTable$std.error == 0 & !is.na(parTable$std.error)
    parTable[isLabelled & zeroStdError, "std.error"] <- NA
  }

  if (!calc.se) parTable$std.error <- NA  # when std.errors are not computed, static constraints
                                          # will get Non-NA std.errors, which is incorrect
                                          # this is naturally corrected for when calculating the
                                          # std.errors, but not when calc.se == FALSE
  parTable[!is.na(parTable$std.error) &
           parTable$std.error == -999, "std.error"] <- NA  # replace -999 with NA

  parTable
}


getConvergenceMessage <- function(converged, iterations) {
  pattern <- if (isTRUE(converged)) {
    "\nmodsem (%s) ended normally after %d iterations\n\n"
  } else {
    paste0(
           "modsem (%s) did NOT end normally after %d iterations\n",
           "** WARNING ** Estimates below are most likely unreliable\n"
    )
  }
  sprintf(pattern, PKG_INFO$version, iterations)
}


finalizeModelEstimatesDA <- function(model,
                                     theta,
                                     method = c("lms","qml"),
                                     data,
                                     logLik,
                                     iterations,
                                     converged,
                                     optimizer,
                                     # SE / FIM controls
                                     calc.se = TRUE,
                                     FIM = "observed",
                                     OFIM.hessian = FALSE,
                                     EFIM.S = 3e4,
                                     EFIM.parametric = TRUE,
                                     robust.se = FALSE,
                                     epsilon = 1e-6,
                                     cr1s = TRUE,
                                     R.max = 1e6,
                                     verbose = FALSE,
                                     # method-specific extras
                                     P = NULL,                 # only used by LMS
                                     includeStartModel = FALSE,
                                     startModel = NULL) {

  method <- match.arg(method)
  NA__ <- -999

  # coefficients and (final) filled model
  lavCoefs   <- getLavCoefs(model = model, theta = theta, method = method)

  # information matrix + SE
  typeSE <- if (!calc.se) "none" else if (robust.se) "robust" else "standard"

  fim.args <- list(model = model,
                   theta = theta,
                   method = method,
                   EFIM.S = EFIM.S,
                   hessian = OFIM.hessian,
                   calc.se = calc.se,
                   EFIM.parametric = EFIM.parametric,
                   verbose = verbose,
                   FIM = FIM,
                   robust.se = robust.se,
                   epsilon = epsilon,
                   cr1s = cr1s,
                   R.max = R.max,
                   NA__ = NA__)

  # only LMS uses the conditional P; QML ignores it
  if (!is.null(P) && method == "lms") fim.args$P <- P

  FIMo <- do.call(calcFIM_da, fim.args)

  SE <- calcSE_da(calc.se = calc.se,
                  vcov = FIMo$vcov.all,
                  rawLabels = FIMo$raw.labels,
                  NA__ = NA__)

  modelSE <- getSE_Model(model, se = SE, method = method,
                         n.additions = FIMo$n.additions)

  finalModel <- getFinalModel(model = model, theta = theta, method = method,
                              modelSE = modelSE)

  parTable <- NULL
  for (g in seq_len(model$info$n.groups)) {
    submodel <- finalModel$models[[g]]

    parTable.g <- modelToParTable(submodel,
                                  coefs = lavCoefs$all,
                                  se = SE,
                                  method = method,
                                  calc.se = calc.se,
                                  group = g)

    parTable <- rbind(parTable, parTable.g)
  }

  parTable <- rbind(parTable,
                    customParamsToParTable(model, coefs = lavCoefs$all, se = SE))

  parTable <- addZStatsParTable(parTable)
  parTable <- sortParTableDA(parTable = parTable, model = model)

  out <- list(
    model            = finalModel,
    method           = method,
    optimizer        = optimizer,
    data             = data,
    theta            = theta,
    coefs.all        = lavCoefs$all,
    coefs.free       = lavCoefs$free,
    parTable         = modsemParTable(parTable),
    originalParTable = model$parTable,
    logLik           = logLik,
    iterations       = iterations,
    convergence      = isTRUE(converged),
    convergence.msg  = getConvergenceMessage(converged, iterations),
    type.se          = typeSE,
    type.estimates   = "unstandardized",
    info.quad        = if (method == "lms") getInfoQuad(model$models[[1L]]$quad) else NULL,
    FIM              = FIMo$FIM,
    vcov.all         = FIMo$vcov.all,
    vcov.free        = FIMo$vcov.free,
    information      = FIMo$type
  )

  if (isTRUE(includeStartModel))
    out$start.model <- startModel

  out
}


addZStatsParTable <- function(parTable, se.col = "std.error", est.col = "est",
                              z.col = "z.value", p.col = "p.value",
                              ci.l = "ci.lower", ci.u = "ci.upper") {
  parTable[[z.col]] <- parTable[[est.col]] / parTable[[se.col]]
  parTable[[p.col]] <- 2 * stats::pnorm(-abs(parTable[[z.col]]))
  parTable[[ci.l]]  <- parTable[[est.col]] - CI_WIDTH * parTable[[se.col]]
  parTable[[ci.u]]  <- parTable[[est.col]] + CI_WIDTH * parTable[[se.col]]

  parTable
}


getFinalModel <- function(model, theta, method, modelSE = NULL) {
  finalModel <- fillModel(model, theta, fillPhi = method == "lms", method = method)

  # keep NA "skeletons" for printing and SE attachment
  emptyModel <- getEmptyModel(
    group.info       = model$info$group.info,
    cov.syntax       = model$models[[1L]]$cov.syntax,
    parTableCovModel = model$models[[1L]]$covModel$parTable,
    mean.observed    = model$info$mean.observed,
    method           = method,
    orthogonal.x     = model$info$orthogonal.x,
    orthogonal.y     = model$info$orthogonal.y
  )

  for (g in seq_along(finalModel$models)) {
    submodel <- finalModel$models[[g]]

    if (!is.null(modelSE)) {
      submodel$matricesSE <- modelSE$models[[g]]$matrices
      submodel$covModelSE <- modelSE$models[[g]]$covModel
    }

    submodel$matricesNA <- emptyModel$models[[g]]$matrices
    submodel$covModelNA <- emptyModel$models[[g]]$covModel

    finalModel$models[[g]] <- submodel
  }

  finalModel
}

parseLavaan <- function(model.syntax = NULL,
                        variableNames = NULL,
                        match = FALSE,
                        suppress.warnings.match = FALSE,
                        match.recycle = FALSE) {
  # Check if a model.syntax is provided, if not we should return an error
  if (is.null(model.syntax))
    mod_msg_stop("No model.syntax provided")
  else if (!is.character(model.syntax))
    mod_msg_stop("The provided model syntax is not a string!")
  else if (length(model.syntax) > 1)
    mod_msg_stop("The provided model syntax is not of length 1")

  parTable        <- modsemify(model.syntax)
  structuralExprs <- parTable[parTable$op == "~",]
  measureExprs    <- parTable[parTable$op %in% c("=~", "<~"), ]

  oVs  <- getOVs(parTable)
  lVs  <- getLVs(parTable)
  vars <- getVarsPI(parTable)
  prodNames <- getProdNames(parTable)
  prodNamesCleaned <- stringr::str_remove_all(prodNames, ":")

  # Get all the indicators in the model
  inds <- getIndicators(parTable, observed=TRUE)
  mod_stopif(!all(inds %in% variableNames),
    paste0("Unable to find observed variables in data: ",
    paste0(inds[!inds %in% variableNames], collapse = ", "))
  )

  # Are prods latent?
  elementsInProds <- lapplyNamed(prodNames,
                                 FUN = splitProdName,
                                 pattern = ":",
                                 names = prodNamesCleaned)
  prodLengths <- vapply(elementsInProds, FUN.VALUE = integer(1L), FUN = length)

  if (!length(prodLengths))
    prodLengths <- 0L

  checkElementsInProds(elementsInProds, lVs=lVs, oVs=oVs)
  checkHigherOrderInteractions(elementsInProds, parTable=parTable)

  # Identify composite (formative) variables and which product terms are fully composite
  composites <- getComposites(parTable)
  isCompositeProd <- vapply(
    elementsInProds,
    FUN.VALUE = logical(1L),
    FUN = function(elems) length(elems) > 0L && all(elems %in% composites)
  )

  # Inds belonging to latent variables which are specified in the syntax
  indsLatents <- structureLavExprs(measureExprs)

  if (length(prodNamesCleaned) > 0) {
    # Get inds belonging to latent variables, or if observed, just get the
    # observed variable in prod terms
    indsInLatentProds <-
      lapplyNamed(elementsInProds[prodNamesCleaned],
                  FUN = getIndsMultipleVariables,
                  indsLatents = indsLatents,
                  names = prodNamesCleaned)

    # Creating a relDF for prodTerms
    relDfs <- lapply(indsInLatentProds, FUN = createRelDf,
                     match = match, match.recycle = match.recycle,
                     suppress.warnings.match = suppress.warnings.match)

    # Get a list with all the inds in each interactionterm
    indsInLatentProds <- lapplyNamed(indsInLatentProds,
                                   FUN = function(x) unname(unlist(x)),
                                   names = prodNamesCleaned)
    # create the names for the indProds
    indProdNames <- lapplyNamed(relDfs,
                                FUN = colnames,
                                names = names(relDfs))
  } else { # in the case where ther is no interaction effects
    indsInLatentProds <- NULL
    relDfs <- NULL
    indProdNames <- NULL
  }

  # Return modelSpec
  modelSpec <- list(model.syntax = model.syntax,
                    parTable     = parTable,

                    oVs = oVs,
                    lVs = lVs,

                    has.interaction     = length(prodNames) > 0L,
                    prodNames           = prodNamesCleaned,
                    elementsInProdNames = elementsInProds,
                    relDfs = relDfs,

                    composites      = composites,
                    isCompositeProd = isCompositeProd,

                    indsInLatentProds = indsInLatentProds,
                    latentProds       = names(indsInLatentProds),
                    indProdNames      = indProdNames,
                    nways             = max(prodLengths)
  )
  modelSpec
}


# Function for structuring exprs in a parTable
structureLavExprs <- function(parTable = NULL) {
  # If empty return NULL
  if (is.null(parTable)) return(NULL)
  # Same if nrow <= 0
  else if (nrow(parTable) <= 0) return(NULL)

  # Get Dependents
  names <- unique(parTable$lhs)

  # see utils for definition of selectValuesByCol() and lapplyNamed()
  lapplyNamed(names, selectValuesByCol, parTable, "rhs" ,"lhs")
}


getIndsVariable <- function(varName = NULL,  indsLatents = NULL) {
  mod_stopif(is.null(varName), "Error in getIndsVariable(), varName is NULL")
  # Get the names of the latent variables in the model
  lVs <- names(indsLatents)

  # Check if our varName is a latent- or an observed variable
  if (!(varName %in% lVs)) {
    # If it is not a latent variable, we should just return the observed variable
    return(varName)
  } else if (varName %in% lVs) {
    # If it is a latent variable, we should return its inds
    return(indsLatents[[varName]])
  }

  # Error if it is neither a latent- nor an observerd variable
  mod_msg_stop(paste0("Something went wrong in getIndsVariable(), ",
        "varName neither observed not latent"))
}


# Function for getting inds for mutliple variables
getIndsMultipleVariables <- function(varNames = NULL, indsLatents = NULL) {
  # varNames = vector with variableNames
  # output should be a list
  # Check that arguements are supplied/not NULL
  mod_stopif(is.null(varNames), "Error in getIndsMultipleVariables(), varNames is NULL")
  # use getIndsVariable for each element in varNames, and return a named list
  lapplyNamed(varNames,
              FUN = getIndsVariable,
              indsLatents = indsLatents)
}


splitProdName <- function(prodName, pattern) {
  mod_stopif(is.null(prodName) ,"prodNames in splitProdName was NULL")
  mod_stopif(is.null(pattern) ,"pattern in splitProdName was NULL")
  stringr::str_split_1(prodName, pattern)
}


splitProdNamesVec <- function(prodNames, pattern) {
  mod_stopif(is.null(prodNames) ,"prodNames in splitProdNamesVec was NULL")
  mod_stopif(is.null(pattern) ,"pattern in splitProdNamesVec was NULL")
  unlist(stringr::str_split(prodNames, pattern))
}


fixProdNames <- function(prodName, pattern = NULL) {
  mod_stopif(is.null(pattern) ,"pattern in fixProdNames was NULL")
  stringr::str_remove_all(prodName, pattern)
}


fixLatentNamesSyntax <- function(model.syntax, pattern) {
  stringr::str_replace_all(model.syntax, pattern, "")
}


createRelDf <- function(indsProdTerm,
                        match = FALSE,
                        suppress.warnings.match = FALSE,
                        match.recycle = TRUE) {
  if (match) {
    lengths <- vapply(indsProdTerm, FUN.VALUE = integer(1L),
                      FUN = length)

    if ((shortest <- min(lengths)) != (longest <- max(lengths))) {
      mod_warnif(!suppress.warnings.match, paste0("Unequal number of indicators for latent variables ",
             "in product term, some indicators will be recycled!"))

      if (match.recycle)
        selector <- \(x) x[rep(seq_along(x), length.out = longest)]
      else
        selector <- \(x) x[seq_len(shortest)]

      indsProdTerm <- lapply(X = indsProdTerm, FUN = selector)
    }

    relDf <- t(as.data.frame(indsProdTerm))

  } else if (!match) {
    allCombos <- t(expand.grid(indsProdTerm))
    relDf <- NULL

    for (i in seq_len(ncol(allCombos))) {
      if (!greplRowDf(allCombos[, i], relDf))
        relDf <- cbind(relDf, allCombos[, i])
    }
  }

  names <- apply(relDf, MARGIN = 2, FUN = stringr::str_c, collapse = "")
  structure(as.data.frame(relDf),
            names = names,
            row.names = names(indsProdTerm))
}


# This should return just an observed variable, if it does not belong in a latent one
getIndsLatents <- function(names, latents) {
  lapplyNamed(names,
              FUN = function(name, latents) {
                if (name %in% names(latents)) {
                  latents[[name]]
                } else {
                  name
                }
              },
              latents = latents,
              names = names)
}

# Global variables
NAMES_PAR_MATRICES <- c("lambdaX", "lambdaY", "gammaXi", "gammaEta",
                      "thetaDelta", "thetaEpsilon", "W", "T", "phi", "A",
                      "covZetaXi", "psi", "tauX", "tauY", "alpha", "beta0",
                      "omegaEtaXi", "omegaXiXi")
NAMES_PAR_MATRICES_COV <- c("gamma", "psi")


createTheta <- function(model, start = NULL, parTable.in = NULL) {
  THETA_LAB  <- NULL
  THETA_COV  <- NULL
  THETA_MAIN <- NULL

  n.groups   <- model$info$n.groups
  if (is.null(n.groups)) n.groups <- model$info$n.groups
  if (is.null(n.groups)) n.groups <- length(model$models)

  LABELS_MAIN_GROUPS <- vector("list", length = n.groups)
  LABELS_COV_GROUPS  <- vector("list", length = n.groups)

  unionByNames <- function(x, y) {
    new <- setdiff(names(y), names(x))
    if (length(new)) c(x, y[new]) else x
  }

  # The custom labels must all be added, before computing THETA_LAB_ALL
  for (g in seq_len(n.groups)) {
    submodel     <- model$models[[g]]
    thetaLabel.g <- createThetaLabel(submodel$labelMatrices,
                                     submodel$covModel$labelMatrices,
                                     model$params$constrExprs)
    THETA_LAB <- unionByNames(THETA_LAB, thetaLabel.g)
  }
  THETA_LAB_ALL <- calcThetaLabel(THETA_LAB, model$params$constrExprs)
  LAV_LAB       <- names(THETA_LAB_ALL)

  for (g in seq_len(n.groups)) {
    submodel <- model$models[[g]]
    etas <- submodel$info$etas

    listThetaCov <- createThetaCovModel(submodel$covModel)
    thetaCov     <- listThetaCov$theta
    lavLabelsCov <- listThetaCov$lavLabels

    M            <- submodel$matrices
    lambdaX      <- as.vector(M$lambdaX)
    lambdaY      <- as.vector(M$lambdaY)
    thetaDelta   <- as.vector(M$thetaDelta)
    thetaEpsilon <- as.vector(M$thetaEpsilon)
    W            <- as.vector(M$W)
    T            <- as.vector(M$T)
    phi          <- as.vector(M$phi)
    A            <- as.vector(M$A)
    covZetaXi    <- as.vector(M$covZetaXi)
    psi          <- as.vector(M$psi)
    tauX         <- as.vector(M$tauX)
    tauY         <- as.vector(M$tauY)
    alpha        <- as.vector(M$alpha)
    beta0        <- as.vector(M$beta0)
    gammaXi      <- as.vector(M$gammaXi)
    gammaEta     <- as.vector(M$gammaEta)
    omegaXiXi    <- as.vector(M$omegaXiXi)
    omegaEtaXi   <- as.vector(M$omegaEtaXi)

    allModelValues <- c(
      lambdaX      = lambdaX,
      lambdaY      = lambdaY,
      tauX         = tauX,
      tauY         = tauY,
      thetaDelta   = thetaDelta,
      thetaEpsilon = thetaEpsilon,
      W            = W,
      T            = T,
      phi          = phi,
      A            = A,
      covZetaXi    = covZetaXi,
      psi          = psi,
      alpha        = alpha,
      beta0        = beta0,
      gammaXi      = gammaXi,
      gammaEta     = gammaEta,
      omegaXiXi    = omegaXiXi,
      omegaEtaXi   = omegaEtaXi
    )

    lavLabelsMain <- createLavLabels(M, subset = is.na(allModelValues),
                                     etas = etas, parTable.in = parTable.in)

    thetaMain <- allModelValues[is.na(allModelValues)]
    thetaMain <- fillThetaIfStartNULL(start = start, theta = thetaMain,
                                      lavlab = lavLabelsMain)

    allLabels <- names(c(thetaCov, thetaMain))
    lavLabels <- combineLavLabels(lavLabelsMain = lavLabelsMain,
                                  lavLabelsCov = lavLabelsCov,
                                  currentLabels = allLabels,
                                  g = g)

    if (g > 1L) {
      .newnames <- \(nm) sprintf("%s.g%d", nm, g)

      # thetaLabel is labelled across the submodels, so the names don't change!
      names(thetaCov)    <- .newnames(names(thetaCov))
      names(thetaMain)   <- .newnames(names(thetaMain))
    }

    LABELS_MAIN_GROUPS[[g]] <- if (length(thetaMain)) names(thetaMain) else character(0)
    LABELS_COV_GROUPS[[g]]  <- if (length(thetaCov)) names(thetaCov) else character(0)

    THETA_COV  <- unionByNames(THETA_COV, thetaCov)
    THETA_MAIN <- unionByNames(THETA_MAIN, thetaMain)
    LAV_LAB    <- union(LAV_LAB, lavLabels)
  }

  THETA <- c(THETA_LAB, THETA_COV, THETA_MAIN)

  SELECT_THETA_LAB  <- vector("list", length = n.groups)
  SELECT_THETA_COV  <- vector("list", length = n.groups)
  SELECT_THETA_MAIN <- vector("list", length = n.groups)

  for (g in seq_len(n.groups)) {
    labelsMain.g <- LABELS_MAIN_GROUPS[[g]]
    labelsCov.g  <- LABELS_COV_GROUPS[[g]]

    selectTL <- seq_along(THETA_LAB) # available to all sub models
    selectTC <- which(names(THETA_COV) %in% labelsCov.g)
    selectTM <- which(names(THETA_MAIN) %in% labelsMain.g)

    # The selections must be offset by their respective location THETA as a whole
    # only selectTL doesn't need to be shifted, as it's at the start
    selectTC <- selectTC + length(THETA_LAB)
    selectTM <- selectTM + length(THETA_LAB) + length(THETA_COV)

    SELECT_THETA_LAB[[g]]  <- selectTL
    SELECT_THETA_COV[[g]]  <- selectTC
    SELECT_THETA_MAIN[[g]] <- selectTM
  }

  list(theta             = THETA,
       freeParams        = length(THETA),
       SELECT_THETA_LAB  = SELECT_THETA_LAB,
       SELECT_THETA_COV  = SELECT_THETA_COV,
       SELECT_THETA_MAIN = SELECT_THETA_MAIN,
       lavLabels         = LAV_LAB)
}


createThetaCovModel <- function(covModel, start = NULL) {
  M <- covModel$matrices

  psi   <- as.vector(M$psi)
  gamma <- as.vector(M$gamma)
  thetaCov <- c(psi = psi, gamma = gamma)

  lavLabelsCov <- createLavLabelsCov(M, subset = is.na(thetaCov))
  thetaCov <- thetaCov[is.na(thetaCov)]
  thetaCov <- fillThetaIfStartNULL(start = start, theta = thetaCov,
                                   lavlab = lavLabelsCov)

  list(theta = thetaCov, lavLabels = lavLabelsCov)
}


fillThetaIfStartNULL <- function(start,
                                 theta,
                                 lavlab = NULL,
                                 var = 1,
                                 cov = 0,
                                 meas = 0.7,
                                 mean = 0,
                                 reg = 0) {
  if (!is.null(start)) {
    return(theta)

  } else if (!is.null(lavlab)) {
    tryCatch({
      OP <- "~~|<~|=~|~1|~"
      op <- stringr::str_extract(lavlab, pattern = OP)
      lr <- stringr::str_split_fixed(lavlab, pattern = OP, n = 2)

      lhs <- lr[, 1]
      rhs <- lr[, 2]
      op[is.na(op)] <- "~"

      theta.filled                          <- theta
      theta.filled[op == "~"]               <- reg
      theta.filled[op == "=~"]              <- meas
      theta.filled[op == "~1"]              <- mean
      theta.filled[op == "<~"]              <- meas
      theta.filled[op == "~~" & lhs == rhs] <- var
      theta.filled[op == "~~" & lhs != rhs] <- cov
      theta.filled[is.na(theta.filled)]     <- reg

      theta.filled
    }, error = \(e)
        fillThetaIfStartNULL(start = start, theta = theta, lavlab = NULL)
    )
  } else {
    vapply(theta, FUN = function(x) stats::runif(1),
           FUN.VALUE = numeric(1L))
  }
}


fillModel <- function(model, theta, fillPhi = FALSE, method = "lms") {
  params.utils <- model$params

  if (is.null(names(theta)))
    names(theta) <- names(params.utils$theta)

  # labeled parameters
  thetaLabel <- NULL

  if (length(params.utils$SELECT_THETA_LAB)) {
    thetaLabel <- theta[params.utils$SELECT_THETA_LAB[[1L]]] # same for all groups
    thetaLabel <- suppressWarnings(calcThetaLabel(thetaLabel, params.utils$constrExprs))
  }

  for (g in seq_len(model$info$n.groups)) {
    submodel <- model$models[[g]]

    # cov model
    thetaCov <- NULL
    if (length(params.utils$SELECT_THETA_COV[[g]]))
      thetaCov <- theta[params.utils$SELECT_THETA_COV[[g]]]

    thetaMain <- theta[params.utils$SELECT_THETA_MAIN[[g]]]

    submodel$covModel <- fillCovModel(submodel$covModel, thetaCov, thetaLabel)
    submodel$matrices <- fillMainModel(submodel, thetaMain, thetaLabel,
                                       fillPhi = fillPhi, method = method)

    model$models[[g]] <- submodel
  }

  model
}


fillMainModel <- function(model, theta, thetaLabel, fillPhi = FALSE,
                          method = "lms") {
  xis      <- model$info$xis
  numXis   <- model$info$numXis
  numEtas  <- model$info$numEtas
  M        <- model$matrices
  covModel <- model$covModel

  lMatrices <- model$labelMatrices[NAMES_PAR_MATRICES]
  pMatrices <- M[NAMES_PAR_MATRICES]
  M[NAMES_PAR_MATRICES] <- fillMatricesLabels(pMatrices, lMatrices, thetaLabel)

  if (!is.null(model$covModel$matrices)) {
    M$phi <- M$A <- expectedCovModel(covModel, method = method, sortedXis = xis)

  } else if (method == "lms") {
    M$A <- fillNA_Matrix(M$A, theta = theta, pattern = "^A([0-9]*)")

  } else if (method == "qml") {
    M$phi <- fillSymmetric(M$phi, fetch(theta, "^phi"))

  }

  M$lambdaX      <- fillNA_Matrix(M$lambdaX, theta = theta, pattern = "^lambdaX")
  M$lambdaY      <- fillNA_Matrix(M$lambdaY, theta = theta, pattern = "^lambdaY")
  M$thetaDelta   <- fillSymmetric(M$thetaDelta, fetch(theta, "^thetaDelta"))
  M$thetaEpsilon <- fillSymmetric(M$thetaEpsilon, fetch(theta, "thetaEpsilon"))
  M$W            <- fillNA_Matrix(M$W, theta = theta, pattern = "^W")
  M$T            <- fillSymmetric(M$T, fetch(theta, "^T"))
  M$covZetaXi    <- fillNA_Matrix(M$covZetaXi, theta = theta, pattern = "^covZetaXi")
  M$psi          <- fillSymmetric(M$psi, fetch(theta, "^psi"))
  M$tauX         <- fillNA_Matrix(M$tauX, theta = theta, pattern = "^tauX")
  M$tauY         <- fillNA_Matrix(M$tauY, theta = theta, pattern = "^tauY")
  M$alpha        <- fillNA_Matrix(M$alpha, theta = theta, pattern = "^alpha")
  M$beta0        <- fillNA_Matrix(M$beta0, theta = theta, pattern = "^beta0")
  M$gammaEta     <- fillNA_Matrix(M$gammaEta, theta = theta, pattern = "^gammaEta")
  M$gammaXi      <- fillNA_Matrix(M$gammaXi, theta = theta, pattern = "^gammaXi")
  M$omegaXiXi    <- fillNA_Matrix(M$omegaXiXi, theta = theta, pattern = "^omegaXiXi")
  M$omegaEtaXi   <- fillNA_Matrix(M$omegaEtaXi, theta = theta, pattern = "^omegaEtaXi")

  if (fillPhi) {
    M$phi <- M$A %*% t(M$A)
  }
  M
}


fillCovModel <- function(covModel, theta, thetaLabel) {
  if (is.null(names(theta))) names(theta) <- names(covModel$theta)
  if (is.null(covModel$matrices)) return(covModel)
  M <- covModel$matrices

  lMatrices <- covModel$labelMatrices[NAMES_PAR_MATRICES_COV]
  pMatrices <- M[NAMES_PAR_MATRICES_COV]
  M[NAMES_PAR_MATRICES_COV] <- fillMatricesLabels(pMatrices, lMatrices, thetaLabel)

  M$psi   <- fillSymmetric(M$psi, fetch(theta, "^psi"))
  M$gamma <- fillNA_Matrix(M$gamma, theta = theta, pattern = "^gamma")

  covModel$matrices <- M
  covModel
}


fillNA_Matrix <- function(X, theta, pattern) {
  idx <- is.na(X) & !is.nan(X)
  values <- fetch(theta, pattern)

  mod_stopif(length(values) && sum(idx) != length(values),
    paste0("Mismatch when filling matrix for pattern `", pattern, "`: expected ",
           sum(idx), " values but got ", length(values))
  )

  mod_stopif(sum(idx) > 0 && length(values) == 0,
    paste0("No values found in theta vector for pattern `", pattern, "`.")
  )

  X[idx] <- values
  X
}


fillSymmetric <- function(mat, values) {
  idx <- is.na(mat) & !is.nan(mat)
  if (sum(idx) > 0 && length(values) == 0) {
    mod_msg_stop("No values provided to fill symmetric matrix.")
  }
  if (length(values) && length(values) != sum(idx)) {
    mod_msg_stop(paste0("Mismatch when filling symmetric matrix: expected ", sum(idx),
          " values but got ", length(values)))
  }
  mat[idx] <- values
  mat[upper.tri(mat)] <- t(mat)[upper.tri(mat)]
  mat
}


getParamBounds <- function(model) {
  lower <- rep(-Inf, model$params$freeParams)
  upper <- rep(Inf, model$params$freeParams)
  names(lower) <- names(upper) <- names(model$theta)

  parTable <- model$parTable
  BOUNDS <- parTable[canBeNumeric(parTable$rhs) & parTable$op %in%
                     BOUNDUARY_OPS, , drop = FALSE]
  bound.param <- BOUNDS$lhs
  bound.type  <- BOUNDS$op
  bound.value <- as.numeric(BOUNDS$rhs)

  upper.type    <- bound.type == "<"
  lower.type <- bound.type == ">"

  bound.upper <- structure(bound.value[upper.type],
                           names = bound.param[upper.type])
  bound.lower <- structure(bound.value[lower.type],
                           names = bound.param[lower.type])

  upper[names(bound.upper)] <- bound.upper
  lower[names(bound.lower)] <- bound.lower

  list(lower = lower, upper = upper)
}


checkStartingParams <- function(start, model) {
  if (length(start) != length(model$theta)) {
    mod_msg_stop("The length of the starting parameters does not match the number of parameters in the model")
  }
  if (is.null(names(start))) {
    names(start) <- names(model$theta)
  }
  if (!all(names(start) %in% names(model$theta))) {
    mod_msg_stop("The names of the starting parameters do not match the names of the parameters in the model")
  }

  NULL
}


calcPhiTheta <- function(theta, model, method) {
  if (method != "lms") return(theta)

  modFilled <- fillModel(theta = theta, model = model, method = method,
                         fillPhi = TRUE)

  for (g in seq_len(model$info$n.groups)) {
    select <- c(model$params$SELECT_THETA_LAB[[g]],
                model$params$SELECT_THETA_COV[[g]],
                model$params$SELECT_THETA_MAIN[[g]])

    theta.g      <- theta[select]
    submodFilled <- modFilled$models[[g]]
    submodel     <- model$models[[g]]

    if (!is.null(submodel$covModel$matrices)) {
      matEst <- submodFilled$covModel$matrices
      matLab <- submodel$covModel$labelMatrices
      matNA  <- submodel$covModel$matrices
    } else {
      matEst <- submodFilled$matrices
      matNA  <- submodel$matrices
      matLab <- submodel$labelMatrices
    }

    vals   <- as.vector(matEst$phi[is.na(matNA$A) & !is.nan(matNA$A)])
    labels <- as.vector(matLab$A)

    if (any(labels != "")) {
      allVals <- as.vector(matEst$phi)
      labVals <- allVals[labels != ""]
      labels  <- labels[labels != ""]
      missing <- setdiff(labels, names(theta))
      mod_stopif(length(missing),
             paste0("Missing labelled parameters in theta vector: ",
             paste(missing, collapse = ", ")))

      theta[select][labels] <- labVals
    }

    theta[select][grepl("^A([0-9]+)?$", names(theta[select]))] <- vals
  }

  theta
}


DA_BLOCKS = list(
  lambdaX      = 0,
  lambdaY      = 1,
  tauX         = 2,
  tauY         = 3,
  thetaDelta   = 4,
  thetaEpsilon = 5,
  A            = 6,
  psi          = 7,
  alpha        = 8,
  beta0        = 9,
  gammaXi      = 10,
  gammaEta     = 11,
  omegaXiXi    = 12,
  omegaEtaXi   = 13,
  W            = 14,
  T            = 15,
  phi          = 16,  # QML: free parameter; LMS: derived from A (not free)
  covZetaXi    = 17
)


SYMMETRIC_BLOCKS_DA = c(
  thetaDelta   = 4,
  thetaEpsilon = 5,
  psi          = 7,
  T            = 15,
  phi          = 16
)


getParamNamesMatrix <- function(mat, matname) {
  if (is.character(mat)) {
    c(mat)

  } else {
    M <- list()
    M[[matname]] <- mat
    names(unlist(M))
  }
}


getParamLocationsMatrices <- function(matrices, isFree = is.na, g = 1L, ignore.g.label = FALSE) {
  matrices <- matrices[intersect(names(matrices), names(DA_BLOCKS))]
  locations <- data.frame(param = NULL, block = NULL, row = NULL, col = NULL)
  for (blockname in names(matrices)) {
    X <- matrices[[blockname]]
    n <- nrow(X)
    m <- ncol(X)

    if (!any(isFree(X))) next

    params <- getParamNamesMatrix(mat = X, matname = blockname)
    block  <- DA_BLOCKS[[blockname]]
    rowidx <- matrix(seq_len(n) - 1, nrow = n, ncol = m, byrow = FALSE)
    colidx <- matrix(seq_len(m) - 1, nrow = n, ncol = m, byrow = TRUE)

    params <- params[isFree(X)]
    rowidx <- rowidx[isFree(X)]
    colidx <- colidx[isFree(X)]

    if (g > 1L && !ignore.g.label)
      params <- sprintf("%s.g%d", params, g)

    locationsBlock <- data.frame(
      param = params,
      group = g,
      block = block,
      row   = rowidx,
      col   = colidx,
      symmetric = as.integer(block %in% SYMMETRIC_BLOCKS_DA)
    )

    locations <- rbind(locations, locationsBlock)
  }

  locations
}


getGradientStruct <- function(model, theta, method = "lms") {
  tryCatch(
    getGradientStructSimple(model = model, theta = theta, method = method),
    error = function(e) {
      mod_msg_warn(paste0("Failed to compute gradient structure: ", e$message))

      list(
        locations   = NULL,
        Jacobian    = NULL,
        nlinDerivs  = NULL,
        evalTheta   = NULL,
        useFDGradient = TRUE, # may not be true, but we should behave as if it is
        isNonLinear = TRUE  # may not be true, but we should behave as if it is
      )
    }
  )
}


# Evaluate the A matrix (lower-Cholesky of xi covariance) produced by the
# covModel for group g at a given theta.  Cheap: no integration, just matrix
# algebra inside expectedCovModel().
evalCovModelA <- function(theta, submodel, params, g, method = "lms") {
  thetaLabel <- NULL
  if (length(params$SELECT_THETA_LAB)) {
    thetaLabel <- suppressWarnings(
      calcThetaLabel(theta[params$SELECT_THETA_LAB[[1L]]], params$constrExprs)
    )
  }

  thetaCov <- if (length(params$SELECT_THETA_COV[[g]]))
    theta[params$SELECT_THETA_COV[[g]]] else NULL

  filledCov <- fillCovModel(submodel$covModel, thetaCov, thetaLabel)
  expectedCovModel(filledCov, method = method, sortedXis = submodel$info$xis)
}


# locations data.frame for the lower-triangular A entries controlled by covModel.
getCovModelALocations <- function(submodel, g, method = "lms") {
  numXis <- submodel$info$numXis
  xis    <- submodel$info$xis
  matname <- if (identical(method, "qml")) "phi" else "A"
  block <- if (identical(method, "qml")) DA_BLOCKS[["phi"]] else DA_BLOCKS[["A"]]

  cov_dim  <- matrix(0L, nrow = numXis, ncol = numXis, dimnames = list(xis, xis))
  free_mask <- lower.tri(cov_dim, diag = TRUE)
  params    <- getParamNamesMatrix(mat = cov_dim, matname = matname)[free_mask]
  rowidx    <- (row(cov_dim) - 1L)[free_mask]
  colidx    <- (col(cov_dim) - 1L)[free_mask]

  if (g > 1L) params <- sprintf("%s.g%d", params, g)

  data.frame(
    param     = params,
    group     = g,
    block     = block,
    row       = rowidx,
    col       = colidx,
    symmetric = as.integer(block %in% SYMMETRIC_BLOCKS_DA),
    stringsAsFactors = FALSE
  )
}


# Recompute the A-matrix (block 6) columns of Jacobian/Jacobian2 at the
# current theta.  expectedCovModel() involves nonlinear operations (Cholesky),
# so dA/dtheta is theta-dependent and must be refreshed at each gradient call.
# Jacobian2 is optional — pass NULL to skip the second-derivative update.
refreshCovModelJacobian <- function(theta, model, Jacobian, Jacobian2 = NULL,
                                    method = NULL) {
  if (is.null(method))
    method <- model$params$gradientStruct$covModelMethod %||% "lms"

  locations <- model$params$gradientStruct$locations
  cov.block <- if (identical(method, "qml")) DA_BLOCKS[["phi"]] else DA_BLOCKS[["A"]]
  cov.locs  <- locations[locations$block == cov.block, , drop = FALSE]
  if (!NROW(cov.locs)) return(list(J = Jacobian, J2 = Jacobian2))

  eps        <- (.Machine$double.eps)^(1/4)
  param.part <- rownames(Jacobian)

  for (g in seq_len(model$info$n.groups)) {
    submodel <- model$models[[g]]
    if (is.null(submodel$covModel$matrices)) next

    A.locs.g  <- cov.locs[cov.locs$group == g, , drop = FALSE]
    affecting <- c(model$params$SELECT_THETA_LAB[[1L]],
                   model$params$SELECT_THETA_COV[[g]])
    if (!length(affecting) || !NROW(A.locs.g)) next

    A0    <- evalCovModelA(theta, submodel, model$params, g, method = method)
    ri    <- A.locs.g$row + 1L
    ci    <- A.locs.g$col + 1L
    acols <- A.locs.g$param  # already carries the #N dedup suffix

    for (k.idx in affecting) {
      k_nm <- param.part[[k.idx]]
      tp <- theta; tp[[k.idx]] <- tp[[k.idx]] + eps
      tm <- theta; tm[[k.idx]] <- tm[[k.idx]] - eps
      Ap <- evalCovModelA(tp, submodel, model$params, g, method = method)
      Am <- evalCovModelA(tm, submodel, model$params, g, method = method)
      Jacobian[k_nm, acols] <- (Ap - Am)[cbind(ri, ci)] / (2 * eps)

      if (!is.null(Jacobian2))
        Jacobian2[k_nm, acols] <- (Ap - 2*A0 + Am)[cbind(ri, ci)] / eps^2
    }
  }

  list(J = Jacobian, J2 = Jacobian2)
}


getGradientStructSimple <- function(model, theta, method = "lms") {
  hasCovModel <- !is.null(model$models[[1L]]$covModel$matrices)

  parTable <- model$parTable
  parTable <- parTable[!parTable$op %in% BOUNDUARY_OPS, , drop = FALSE] # not relevant

  customParams <- parTable[parTable$op %in% c(":=", "=="), , drop = FALSE]
  for (i in seq_len(NROW(customParams))) {
    row <- customParams[i, , drop = FALSE]
    eq  <- sprintf("(%s)", row$rhs)
    pattern <- sprintf("(?<![A-z_\\.])%s(?![A-z_\\.])", row$lhs)

    mask <- parTable$op %in% CONSTRAINT_OPS
    parTable$rhs[mask] <- stringr::str_replace_all(
      parTable$rhs[mask], pattern = pattern, replacement = eq
    )
  }

  isConstraint <- parTable$op %in% CONSTRAINT_OPS & !canBeNumeric(parTable$rhs)
  constraints  <- parTable[isConstraint, ]
  restParTable <- parTable[!isConstraint, ]
  # TODO: find better way of removing unecessary constraints
  constraints  <- constraints[constraints$lhs %in% restParTable$mod, ]

  derivatives <- list()
  derivatives2 <- list()
  for (i in seq_len(NROW(constraints))) {
    constrVar <- constraints[i, "lhs"]
    constrEq  <- constraints[i, "rhs"]

    derivatives[[constrVar]] <- derivateConstraint(constrEq)
    derivatives2[[constrVar]] <- secondDerivateConstraint(constrEq)
  }

  isLinear <- vapply(
    X = derivatives, FUN.VALUE = logical(1L),
    FUN = \(X) all(vapply(X, FUN.VALUE = logical(1L), FUN = is.atomic))
  )

  linDerivs   <- derivatives[isLinear]
  nlinDerivs  <- derivatives[!isLinear]
  nlinDerivs2 <- derivatives2[!isLinear]
  evalTheta   <- \(theta) c(theta, suppressWarnings(calcThetaLabel(theta, model$constrExprs))) # This could be made a bit better

  locations <- NULL
  for (g in seq_len(model$info$n.groups)) {
    submodel <- model$models[[g]]

    locations <- rbind(
       locations,
       getParamLocationsMatrices(submodel$matrices, isFree = is.na, g = g),
       getParamLocationsMatrices(submodel$labelMatrices, isFree = \(x) x != "",
                                 g = g, ignore.g.label = TRUE) # labels don't change with g here
    )
  }

  k <- nrow(locations)
  m <- length(theta)

  param.full <- locations$param
  param.part <- names(theta)

  ordering <- structure(seq_along(theta), names = param.part)
  ordering <- ordering[param.full]

  locations  <- locations[order(ordering), ]
  param.full <- locations$param

  Jacobian <- matrix(0, nrow = m, ncol = k,
                     dimnames = list(param.part, param.full))
  Jacobian2 <- Jacobian

  for (j in seq_along(param.full)) {
    g <- locations$group[[j]]
    main_idx <- model$params$SELECT_THETA_MAIN[[g]]
    label_idx <- if (length(model$params$SELECT_THETA_LAB))
      model$params$SELECT_THETA_LAB[[1L]] else integer(0L)
    candidate_idx <- c(label_idx, main_idx)
    match.part <- candidate_idx[param.part[candidate_idx] == param.full[[j]]]
    if (length(match.part))
      Jacobian[match.part[[1L]], j] <- 1
  }

  for (dep in names(linDerivs)) {
    deriv <- linDerivs[[dep]]

    for (indep in names(deriv)) {
      match.full <- param.full == dep
      match.part <- param.part == indep
      Jacobian[match.part, match.full] <- deriv[[indep]]
    }
  }

  # When a covModel is present, the A matrix is entirely determined by
  # expectedCovModel(thetaCov) rather than being free parameters in thetaMain.
  # We extend the Jacobian with those A-entry columns and fill them via central
  # finite differences on expectedCovModel — cheap (matrix algebra only,
  # no integration) and exact enough for gradient/Hessian quality.
  if (hasCovModel) {
    cov.locs <- NULL
    for (g in seq_len(model$info$n.groups)) {
      if (!is.null(model$models[[g]]$covModel$matrices))
        cov.locs <- rbind(cov.locs, getCovModelALocations(model$models[[g]], g,
                                                           method = method))
    }

    n_cov  <- NROW(cov.locs)
    cov.J  <- matrix(0, nrow = m, ncol = n_cov,
                     dimnames = list(param.part, cov.locs$param))
    cov.J2 <- cov.J
    eps    <- (.Machine$double.eps)^(1/4) # good balance for both J and J2

    for (g in seq_len(model$info$n.groups)) {
      submodel <- model$models[[g]]
      if (is.null(submodel$covModel$matrices)) next

      A.locs.g  <- cov.locs[cov.locs$group == g, , drop = FALSE]
      affecting <- c(model$params$SELECT_THETA_LAB[[1L]],
                     model$params$SELECT_THETA_COV[[g]])
      if (!length(affecting) || !NROW(A.locs.g)) next

      A0 <- evalCovModelA(theta, submodel, model$params, g, method = method)
      ri <- A.locs.g$row + 1L  # 0-indexed -> 1-indexed
      ci <- A.locs.g$col + 1L

      for (k.idx in affecting) {
        k_nm <- param.part[[k.idx]]
        theta.p <- theta.m <- theta
        theta.p[[k.idx]] <- theta.p[[k.idx]] + eps
        theta.m[[k.idx]] <- theta.m[[k.idx]] - eps

        Ap <- evalCovModelA(theta.p, submodel, model$params, g, method = method)
        Am <- evalCovModelA(theta.m, submodel, model$params, g, method = method)

        cov.J [k_nm, A.locs.g$param] <- (Ap - Am)[cbind(ri, ci)] / (2 * eps)
        cov.J2[k_nm, A.locs.g$param] <- (Ap - 2*A0 + Am)[cbind(ri, ci)] / eps^2
      }
    }

    locations <- rbind(locations, cov.locs)
    Jacobian  <- cbind(Jacobian,  cov.J)
    Jacobian2 <- cbind(Jacobian2, cov.J2)
  }

  # In multigroup models we can have duplicated labels, that doesn't work
  # with out logic, so we must add some unique identifiers to the param.full
  # We also pass param.full as is, to map back and forth.
  addid <- \(nm) paste0(nm, "#", seq_along(nm))
  colnames(Jacobian)  <- addid(colnames(Jacobian))
  colnames(Jacobian2) <- addid(colnames(Jacobian2))

  locations$param <- addid(locations$param)

  list(
    param.full  = param.full,
    locations   = locations,
    Jacobian    = Jacobian,
    Jacobian2   = Jacobian2,
    nlinDerivs  = nlinDerivs,
    nlinDerivs2 = nlinDerivs2,
    evalTheta   = evalTheta,
    hasCovModel = hasCovModel,
    covModelMethod = method,
    useFDGradient = FALSE,
    isNonLinear = length(nlinDerivs) > 1
  )
}


derivateConstraint <- function(constr) {
  f <- stats::formula(paste0("~", constr))
  eq <- Deriv::Deriv(f, combine = "list", cache.exp = FALSE)

  vars <- all.vars(f)
  k    <- length(vars)

  if (k == 0) return(NULL)

  if (is.null(names(eq)) && k == 1L) {
    eq <- stats::setNames(list(eq), nm = vars)
  } else if (is.null(names(eq))) {
    names(eq) <- all.vars(f)
  } else {
    eq <- as.list(eq)[-1]
  }

  eq
}


secondDerivateConstraint <- function(constr) {
  f <- stats::formula(paste0("~", constr))
  eq <- Deriv::Deriv(f, nderiv = 2, combine = "list", cache.exp = FALSE)

  vars <- all.vars(f)
  k    <- length(vars)

  if (k == 0) return(NULL)

  out <- list()
  if (is.null(names(eq)) && k == 1L) {
    eq <- stats::setNames(list(eq), nm = vars)
    return(eq)

  } else if (is.null(names(eq))) {
    names(eq) <- all.vars(f)
    return(eq)

  } else {
    for (indep in all.vars(f)) {
      out[[indep]] <- eq[[indep]][[indep]]
    }
    return(out)
  }
}

# Functions for constructing matrices for LMS and QML.
EMPTY_MATSTRUCT <- list(numeric = matrix(nrow = 0, ncol = 0),
                        label = matrix(nrow = 0, ncol = 0))


setMatrixConstraints <- function(X, parTable, op, RHS, LHS, type, nonFreeParams) {
  fillConstExprs(X, parTable = parTable, op = op, RHS = RHS, LHS = LHS,
                 type = type, nonFreeParams = nonFreeParams) |>
    fillDynExprs(parTable = parTable, op = op, RHS = RHS, LHS = LHS, type = type)
}


fillConstExprs <- function(X, parTable, op, RHS, LHS, type, nonFreeParams = TRUE) {
  if (!NROW(parTable)) return(X)

  constExprs <- parTable[parTable$op %in% op &
                         parTable$rhs %in% RHS &
                         parTable$lhs %in% LHS &
                         canBeNumeric(parTable$mod, includeNA = !nonFreeParams), ]
  constExprs[constExprs$op == "~1", "rhs"] <- "1"

  setVal <- getSetValFunc(type)
  for (i in seq_len(NROW(constExprs))) {
    lhs <- constExprs[i, "lhs"]
    rhs <- constExprs[i, "rhs"]
    val <- as.numeric(constExprs[i, "mod"])
    X   <- setVal(X = X, rhs = rhs, lhs = lhs, val = val)
  }

  if (type == "symmetric") X[upper.tri(X)] <- 0
  X
}


fillDynExprs <- function(X, parTable, op, RHS, LHS, type) {
  # dynamic exprs need a corresponding matrix of labels
  labelX <- as.character.matrix(X, empty = TRUE)

  if (!NROW(parTable)) return(list(numeric = X, label = labelX))

  dynamicExprs <- parTable[parTable$op %in% op &
                           parTable$rhs %in% RHS &
                           parTable$lhs %in% LHS &
                           !canBeNumeric(parTable$mod,
                                         includeNA = TRUE), ]
  dynamicExprs[dynamicExprs$op == "~1", "rhs"] <- "1"

  setVal <- getSetValFunc(type)
  for (i in seq_len(NROW(dynamicExprs))) {
    lhs <- dynamicExprs[i, "lhs"]
    rhs <- dynamicExprs[i, "rhs"]
    mod <- dynamicExprs[i, "mod"]

    if (mod == "NA") {
      X      <- setVal(X = X, rhs = rhs, lhs = lhs, val = NA)
      labelX <- setVal(X = labelX, rhs = rhs, lhs = lhs, val = "")
    } else {
      X      <- setVal(X = X, rhs = rhs, lhs = lhs, val = 0)
      labelX <- setVal(X = labelX, rhs = rhs, lhs = lhs, val = mod)
    }
  }

  list(numeric = X, label = labelX)
}


getSetValFunc <- function(type) {
  switch(type,
         rhs       = setValRhsFirst,
         lhs       = setValLhsFirst,
         symmetric = setValSymmetric,
         stop("Unrecognized type, this is probably a bug!"))
}


setValRhsFirst <- function(X, rhs, lhs, val) {
  X[rhs, lhs] <- val
  X
}


setValLhsFirst <- function(X, rhs, lhs, val) {
  X[lhs, rhs] <- val
  X
}


setValSymmetric <- function(X, rhs, lhs, val) {
  X[lhs, rhs]     <- X[rhs, lhs] <- val
  X
}


getEmptyPhi <- function(phi) {
  labelPhi <- as.character.matrix(phi, empty = TRUE)
  list(numeric = phi, label = labelPhi)
}


notFilledLambda <- function(ind, lambda) {
  !any(is.na(lambda[ind, ])) && all(lambda[ind, ] == 0)
}


constructLambda <- function(lVs, indsLVs, parTable, auto.fix.first = TRUE, mode = "a") {
  numLVs        <- length(lVs)
  indsLVs       <- indsLVs[lVs] # make sure it is sorted
  allIndsLVs    <- unique(unlist(indsLVs))
  numAllIndsLVs <- length(allIndsLVs)
  firstVal      <- ifelse(auto.fix.first, 1, NA)

  lambda <- matrix(0, nrow = numAllIndsLVs, ncol = numLVs,
                   dimnames = list(allIndsLVs, lVs))

  for (lV in lVs) {
    firstFilled <- FALSE
    for (ind in indsLVs[[lV]]) {
      # TODO: make this work when duplicates appear seperately in lambdaY and lambdaX
      if (!firstFilled && auto.fix.first && notFilledLambda(ind, lambda)) {
        lambda[ind, lV] <- firstVal
        firstFilled <- TRUE
      } else lambda[ind, lV] <- NA
    }
  }

  switch(tolower(mode),
    a = {
      op <- "=~"
      composites <- getComposites(parTable)
      lambda[,intersect(colnames(lambda), composites)] <- 0
    },
    b = {
      op <- "<~"
      composites <- getComposites(parTable)
      lambda[,setdiff(colnames(lambda), composites)] <- 0
    },
    mod_msg_stop(paste0("Unsupported mode: ", mode, "!"))
  )

  setMatrixConstraints(X = lambda, parTable = parTable, op = op,
                       RHS = allIndsLVs, LHS = lVs, type = "rhs",
                       nonFreeParams = TRUE) # first params are by default set to 1
}


constructT <- function(lVs, indsLVs, parTable,
                       fix.composite.var = TRUE,
                       data = NULL, missing = "listwise",
                       sampling.weights = NULL) {
  indsLVs <- indsLVs[lVs] # make sure it is sorted
  allInds <- unique(unlist(indsLVs))
  k       <- length(allInds)

  T <- matrix(
    0, nrow = k, ncol = k,
    dimnames = list(allInds, allInds)
  )

  composites <- intersect(lVs, getComposites(parTable))
  compositeInds <- unique(unlist(indsLVs[composites]))

  for (composite in composites) {
    idx <- intersect(rownames(T), indsLVs[[composite]])

    if (fix.composite.var && !is.null(data$data.full)) {
      X <- data$data.full[,idx, drop = FALSE]

      S <- lavaan::lavCor(
        object           = X,
        missing          = missing,
        sampling.weights = sampling.weights,
        output           = "cov"
      )

      T[idx, idx] <- S[idx, idx]

    } else {
      T[idx, idx] <- NA

    }
  }

  setMatrixConstraints(X = T, parTable = parTable, op = "~~",
                       RHS = compositeInds, LHS = compositeInds,
                       type = "symmetric", nonFreeParams = FALSE)
}


constructTau <- function(lVs, indsLVs, parTable, mean.observed = TRUE) {
  indsLVs       <- indsLVs[lVs] # make sure it is sorted
  allIndsLVs    <- unique(unlist(indsLVs))
  numAllIndsLVs <- length(allIndsLVs)
  default       <- ifelse(mean.observed, NA, 0)
  lavOptimizerSyntaxAdditions <- ""

  tau <- matrix(default, nrow = numAllIndsLVs, ncol = 1,
                dimnames = list(allIndsLVs, "1"))
  for (lV in lVs) { # set first ind to 0, if lV has meanstructure
    subPT <- parTable[parTable$lhs == lV & parTable$op == "~1", ]
    indsLV <- indsLVs[[lV]]

    if (NROW(subPT) && length(indsLV)) {
      firstInd         <- indsLV[[1]]
      tau[firstInd, 1] <- 0
      lavOptimizerSyntaxAdditions <-
        getFixedInterceptSyntax(indicator = firstInd, parTable = parTable,
                                syntax = lavOptimizerSyntaxAdditions)
    }
  }

  c(setMatrixConstraints(X = tau, parTable = parTable, op = "~1",
                         RHS = "", LHS = allIndsLVs, type = "lhs",
                         nonFreeParams = FALSE),
    list(syntaxAdditions = lavOptimizerSyntaxAdditions))
}


constructTheta <- function(lVs, indsLVs, parTable, auto.fix.single = TRUE) {
  numLVs        <- length(lVs)
  indsLVs       <- indsLVs[lVs] # make sure it is sorted
  numIndsLVs    <- lapply(indsLVs, FUN = length)
  allIndsLVs    <- unique(unlist(indsLVs))
  numAllIndsLVs <- length(allIndsLVs)

  theta <- matrix(0, nrow = numAllIndsLVs, ncol = numAllIndsLVs,
                  dimnames = list(allIndsLVs, allIndsLVs))
  diag(theta) <- NA

  if (auto.fix.single) {
    for (lV in lVs) { # set to 0 if there is only a single indicator
      if (numIndsLVs[[lV]] != 1) next
      theta[indsLVs[[lV]], indsLVs[[lV]]] <- 0
    }
  }

  compositeIndicators <- getCompositeIndicators(parTable)
  isCompositeInd <- allIndsLVs %in% compositeIndicators
  theta[isCompositeInd, isCompositeInd] <- 0

  setMatrixConstraints(X = theta, parTable = parTable, op = "~~",
                       RHS = allIndsLVs, LHS = allIndsLVs, type = "symmetric",
                       nonFreeParams = FALSE)
}


constructGamma <- function(DVs, IVs, parTable, auto.fix.first = TRUE) {
  if (!length(DVs)) return(EMPTY_MATSTRUCT)

  notInt <- !grepl(":", parTable$rhs)
  exprsGamma1 <- parTable[parTable$op == "~" & notInt & parTable$lhs %in% DVs &
                          parTable$rhs %in% IVs, , drop = FALSE]
  exprsGamma2 <- parTable[parTable$op == "=~" & notInt & parTable$rhs %in% DVs &
                          parTable$lhs %in% IVs, , drop = FALSE]

  if (NROW(exprsGamma2)) {
    exprsGamma3 <- exprsGamma2
    exprsGamma3$lhs <- exprsGamma2$rhs
    exprsGamma3$op  <- "~"
    exprsGamma3$rhs <- exprsGamma2$lhs
  } else exprsGamma3 <- NULL

  exprsGamma <- rbind(exprsGamma1, exprsGamma3)


  numDVs <- length(DVs)
  numIVs <- length(IVs)
  gamma  <- matrix(0, nrow = numDVs, ncol = numIVs, dimnames = list(DVs, IVs))

  gamma <- setMatrixConstraints(X = gamma, parTable = exprsGamma, op = c("~", "=~"),
                                RHS = IVs, LHS = DVs, type = "lhs",
                                nonFreeParams = FALSE)

  higherOrderLVs <- unique(exprsGamma2[exprsGamma2$op == "=~", "lhs"])
  for (lVh in higherOrderLVs) {
    inds <- exprsGamma2[exprsGamma2$op == "=~" &
                        exprsGamma2$lhs == lVh, "rhs"]
    ind1 <- inds[[1L]]

    if (auto.fix.first &&
        is.na(gamma$numeric[ind1, lVh]) &&
        gamma$label[ind1, lVh] == "") {
      gamma$numeric[ind1, lVh] <- 1
    }
  }

  gamma
}


constructPsi <- function(etas, parTable,
                         orthogonal.y = FALSE,
                         orthogonal.x = FALSE, 
                         has.exo = FALSE) {
  if (!length(etas)) return(EMPTY_MATSTRUCT)

  numEtas   <- length(etas)
  psi       <- matrix(0, nrow = numEtas, ncol = numEtas,
                      dimnames = list(etas, etas))
  diag(psi) <- NA

  if (!orthogonal.y) {
    for (i in seq_len(NROW(psi))) for (j in seq_len(i - 1)) {
      eta_i <- etas[[i]]
      eta_j <- etas[[j]]

      if (isPureEta(eta_i, parTable) && isPureEta(eta_j, parTable))
        psi[i, j] <- NA
    }
  }

  if (has.exo && !orthogonal.x) {
    xis <- getXis(parTable, isLV = FALSE, checkAny = FALSE)

    for (i in seq_len(NROW(psi))) for (j in seq_len(i - 1)) {
      eta_i <- etas[[i]]
      eta_j <- etas[[j]]

      if (eta_i %in% xis && eta_j %in% xis)
        psi[i, j] <- NA
    }
  }

  setMatrixConstraints(X = psi, parTable = parTable, op = "~~",
                       RHS = etas, LHS = etas, type = "symmetric",
                       nonFreeParams = FALSE)
}


constructPhi <- function(xis, method = "lms", cov.syntax = NULL,
                         parTable, parTableCovModel = NULL,
                         orthogonal.x = FALSE) {
  if (!length(xis)) return(EMPTY_MATSTRUCT)

  numXis <- length(xis)
  phi    <- matrix(0, nrow = numXis, ncol = numXis,
                   dimnames = list(xis, xis))

  if (method != "lms" && is.null(cov.syntax) && is.null(parTableCovModel)) {
    if (!orthogonal.x) phi[lower.tri(phi, diag = TRUE)] <- NA
    else               diag(phi) <- NA

    setMatrixConstraints(X = phi, parTable = parTable, op = "~~",
                         RHS = xis, LHS = xis, type = "symmetric",
                         nonFreeParams = FALSE)
  } else getEmptyPhi(phi = phi)
}


constructA <- function(xis, method = "lms", cov.syntax = NULL,
                       parTable, parTableCovModel = NULL, orthogonal.x = FALSE) {
  if (!length(xis))
    return(EMPTY_MATSTRUCT)

  numXis <- length(xis)
  A      <- matrix(0, nrow = numXis, ncol = numXis,
                   dimnames = list(xis, xis))
  if (method == "lms" && is.null(cov.syntax) && is.null(parTableCovModel)) {
    if (!orthogonal.x) A[lower.tri(A, diag = TRUE)] <- NA
    else               diag(A) <- NA

    A <- setMatrixConstraints(X = A, parTable = parTable, op = "~~",
                              RHS = xis, LHS = xis, type = "symmetric",
                              nonFreeParams = FALSE)

    if (!any(is.na(A$numeric))) {
      Phi <- A$numeric
      Phi[upper.tri(Phi)] <- t(Phi)[upper.tri(Phi)]
      A$numeric <- t(chol(Phi))
    }

    A

  } else getEmptyPhi(phi = A)
}


constructCovZetaXi <- function(xis, etas, method = "lms", parTable) {
  if (method != "lms" || !length(xis) || !length(etas))
    return(EMPTY_MATSTRUCT)

  numXis <- length(xis)
  numEtas <- length(etas)

  covZetaXi <- matrix(
    0, nrow = numEtas, ncol = numXis,
    dimnames = list(etas, xis)
  )

  parTableResCov <- parTable[
    parTable$op == "~~" &
    ((parTable$lhs %in% etas & parTable$rhs %in% xis) |
     (parTable$lhs %in% xis & parTable$rhs %in% etas)),
    , drop = FALSE
  ]

  swap <- parTableResCov$lhs %in% xis
  lhs <- parTableResCov$lhs[swap]
  parTableResCov$lhs[swap] <- parTableResCov$rhs[swap]
  parTableResCov$rhs[swap] <- lhs

  setMatrixConstraints(
    X = covZetaXi, parTable = parTableResCov, op = "~~",
    RHS = c(xis, etas), LHS = c(xis, etas),
    type = "lhs", nonFreeParams = FALSE
  )
}


constructAlpha <- function(etas, parTable, mean.observed = TRUE) {
  if (!length(etas)) return(EMPTY_MATSTRUCT)

  numEtas <- length(etas)
  default <- if (mean.observed) 0 else NA

  alpha <- matrix(default, nrow = numEtas, ncol = 1,
                  dimnames = list(etas, "1"))

  if (!mean.observed) {
    for (i in seq_len(numEtas)) {
      eta <- etas[[i]]
      fill <- if (intTermsAffectLV(eta, parTable)) NA else 0
      alpha[eta, 1] <- fill
    }
  }

  setMatrixConstraints(X = alpha, parTable = parTable, op = "~1",
                       RHS = "", LHS = etas, type = "lhs",
                       nonFreeParams = FALSE)
}


selectScalingY <- function(lambdaY, method = "qml") {
  if (method != "qml") return(NULL)
  matrix(apply(lambdaY, MARGIN = 2, FUN = isScalingY),
         nrow = nrow(lambdaY), ncol = ncol(lambdaY),
         dimnames = dimnames(lambdaY))
}


selectBetaRows <- function(lambdaY, method = "qml") {
  if (method != "qml") return(NULL)
  scalingYs <- selectScalingY(lambdaY, method = "qml")
  matrix(apply(scalingYs, MARGIN = 1, FUN = function(x )!all(x)),
         nrow = nrow(lambdaY), ncol = 1,
         dimnames = list(rownames(lambdaY), "1"))
}


constructR <- function(etas, indsEtas, lambdaY, method = "qml") {
  if (method != "qml") return(NULL)
  hasMultipleInds <- vapply(indsEtas, FUN = function(x) length(x) > 1,
                            FUN.VALUE = logical(1L))
  etas <- names(indsEtas)[hasMultipleInds]
  indsEtas <- indsEtas[etas]
  if (length(etas) == 0) return(NULL)

  numEtas <- length(etas)
  numIndsEtas <- lapply(indsEtas, FUN = length)
  allIndsEtas <- unique(unlist(indsEtas))
  numAllIndsEtas <- length(allIndsEtas)
  selectBetaRows <- selectBetaRows(lambdaY, method = method)
  rowNamesR <- rownames(lambdaY)[selectBetaRows]

  R <- matrix(0, nrow = numAllIndsEtas - numEtas,
              ncol = numAllIndsEtas,
              dimnames = list(rowNamesR[seq_len(numAllIndsEtas - numEtas)],
                              allIndsEtas))

  lastRow <- lastCol <- 0
  for (i in seq_len(numEtas)) {
    nInds <- numIndsEtas[[etas[[i]]]] - 1
    if (nInds == 0) mod_msg_stop("Etas in QML must have at least two indicators")
    # free params
    R[seq_len(nInds) + lastRow, lastCol + 1] <- NA
    R[seq_len(nInds) + lastRow,
      seq_len(nInds) + lastCol + 1] <- diag(nInds)
    lastRow <- lastRow + nInds
    lastCol <- lastCol + nInds + 1
  }
  R
}


getLatentEtasQml <- function(indsEtas, method = "qml") {
  if (method != "qml") return(NULL)
  hasMultipleInds <- vapply(indsEtas, FUN = function(x) length(x) > 1,
                            FUN.VALUE = logical(1L))
  etas <- names(indsEtas)[hasMultipleInds]
  if (length(etas) == 0) return(NULL)
  etas
}


getColsU <- function(etas, indsEtas, lambdaY, method = "qml") {
  numEtas         <- length(etas)
  numIndsEtas     <- lapply(indsEtas, FUN = length)
  allIndsEtas     <- unique(unlist(indsEtas))
  numAllIndsEtas  <- length(allIndsEtas)
  selectBetaRows  <- selectBetaRows(lambdaY, method = "qml")
  rowNamesR       <- rownames(lambdaY)[selectBetaRows]

  hasMultipleInds <- vapply(indsEtas, FUN = function(x) length(x) > 1,
                            FUN.VALUE = logical(1L))
  if (sum(hasMultipleInds) == 0) return(NULL)

  colsU <- rowNamesR[seq_len(numAllIndsEtas - sum(hasMultipleInds))]
  colsU[is.na(colsU)] <- paste0("__FILL__ZERO__", seq_len(sum(is.na(colsU))))

  colsU
}


constructFullU <- function(fullL2, N, etas, method = "qml") {
  if (method != "qml" || N == 0) return(NULL)

  if (is.null(fullL2)) nCols <- length(etas)
  else nCols <- NCOL(fullL2)

  matrix(0, nrow = N, ncol = nCols, dimnames = list(NULL, colnames(fullL2)))
}


constructFullR <- function(etas, indsEtas, lambdaY, method = "qml") {
  if (method != "qml") return(NULL)

  numEtas <- length(etas)
  numIndsEtas <- lapply(indsEtas, FUN = length)
  allIndsEtas <- unlist(indsEtas)
  numAllIndsEtas <- length(allIndsEtas)
  selectBetaRows <- selectBetaRows(lambdaY, method = "qml")
  rowNamesR <- rownames(lambdaY)[selectBetaRows]
  hasMultipleInds <- vapply(indsEtas, FUN = function(x) length(x) > 1,
                            FUN.VALUE = logical(1L))
  if (sum(hasMultipleInds) == 0) return(NULL)

  matrix(0, nrow = numAllIndsEtas - sum(hasMultipleInds),
         ncol = numAllIndsEtas,
         dimnames = list(rowNamesR[seq_len(numAllIndsEtas - sum(hasMultipleInds))],
                                           allIndsEtas))

}


constructFullSigma2ThetaEpsilon <- function(psi, method = "qml") {
  if (method != "qml") return(NULL)
  matrix(0, nrow = nrow(psi), ncol = ncol(psi),
         dimnames = dimnames(psi))
}


getSelectSubSigma2ThetaEpsilon <- function(fullSigma2ThetaEpsilon,
                                           latentEtas, method = "qml") {
  if (method != "qml") return(NULL)
  select <- as.logical.matrix(fullSigma2ThetaEpsilon)
  select[TRUE] <- FALSE
  select[latentEtas, latentEtas] <- TRUE
  select
}


constructFullL2 <- function(colsU, etas, method = "qml") {
  if (method != "qml") return(NULL)

  if (is.null(colsU)) nCols <- length(etas)
  else nCols <- length(colsU)
  matrix(0, nrow = length(etas), ncol = nCols,
         dimnames = list(etas, colsU))
}


getSelectSubL2 <- function(fullL2, colsU, latentEtas, method = "qml") {
  if (method != "qml") return(NULL)
  select <- as.logical.matrix(fullL2)
  select[TRUE] <- FALSE
  select[latentEtas, !grepl("__FILL__ZERO__", colnames(select))] <- TRUE
  select
}


getScalingInds <- function(indsEtas, R, latentEtas, method = "qml") {
  if (method != "qml") return(NULL)
  allIndsEtas <- unique(unlist(indsEtas[latentEtas]))
  scalingInds <- allIndsEtas[!allIndsEtas %in% rownames(R)]
  scalingInds
}


selectThetaEpsilon1 <- function(indsEtas, thetaEpsilon, scalingInds,
                                method = "qml") {
  if (method != "qml") return(NULL)
  selectThetaEpsilon <- as.logical.matrix(thetaEpsilon)
  selectThetaEpsilon[TRUE] <- FALSE
  diag(selectThetaEpsilon)[scalingInds] <- TRUE
  selectThetaEpsilon
}


selectThetaEpsilon2 <- function(indsEtas, thetaEpsilon, scalingInds,
                                method = "qml") {
  if (method != "qml") return(NULL)
  selectThetaEpsilon <- as.logical.matrix(thetaEpsilon)
  selectThetaEpsilon[TRUE] <- FALSE

  allScalingInds <- vapply(indsEtas, FUN.VALUE = character(1L), FUN = \(x) x[1L])
  nonLatentScalingInds <- allScalingInds[!allScalingInds %in% scalingInds]

  diag(selectThetaEpsilon)[nonLatentScalingInds] <- TRUE
  selectThetaEpsilon
}


constructSubThetaEpsilon1 <- function(indsEtas, thetaEpsilon, scalingInds,
                                      method = "qml") {
  if (method != "qml") return(NULL)
  subThetaEpsilon <- matrix(0, nrow = length(scalingInds),
                            ncol = length(scalingInds),
                            dimnames = list(scalingInds, scalingInds))
  diag(subThetaEpsilon) <- NA
  subThetaEpsilon
}


constructSubThetaEpsilon2 <- function(indsEtas, thetaEpsilon, scalingInds,
                                      method = "qml") {
  if (method != "qml") return(NULL)

  allScalingInds <- vapply(indsEtas, FUN.VALUE = character(1L), FUN = \(x) x[1L])
  nonLatentScalingInds <- allScalingInds[!allScalingInds %in% scalingInds]

  subThetaEpsilon <- matrix(0, nrow = length(nonLatentScalingInds),
                            ncol = length(nonLatentScalingInds),
                            dimnames = list(nonLatentScalingInds,
                                            nonLatentScalingInds))

  diag(subThetaEpsilon) <- NA
  subThetaEpsilon
}


getScalingLambdaY <- function(lambdaY, indsEtas, etas, method = "qml") {
  if (method != "qml") return(NULL)
  hasMultipleInds <- vapply(indsEtas, FUN = function(x) length(x) > 1,
                            FUN.VALUE = logical(1L))
  latentEtas <- names(indsEtas)[hasMultipleInds]
  indsLatentEtas <- unlist(indsEtas[latentEtas])
  lambdaY[indsLatentEtas, latentEtas]
}


sortXisConstructOmega <- function(xis, varsInts, etas, intTerms,
                                  method = "lms", double = FALSE,
                                  structovs = NULL, composites = NULL) {
  checkVarsIntsDA(varsInts, lVs = c(etas, xis))

  listSortedXis <- sortXis(
    xis = xis, varsInts = varsInts, etas = etas, intTerms = intTerms,
    double = double, structovs = structovs, composites = composites
  )

  sortedXis    <- listSortedXis$sortedXis
  nonLinearXis <- listSortedXis$nonLinearXis

  omegaXiXi <- constructOmegaXiXi(xis = xis, etas = etas,
                                  sortedXis = sortedXis,
                                  nonLinearXis = nonLinearXis,
                                  varsInts = varsInts,
                                  intTerms = intTerms)
  omegaEtaXi <- constructOmegaEtaXi(xis = xis, etas = etas,
                                    sortedXis = sortedXis,
                                    nonLinearXis = nonLinearXis,
                                    varsInts = varsInts,
                                    intTerms = intTerms)

  list(sortedXis = sortedXis, nonLinearXis = nonLinearXis,
       omegaXiXi = omegaXiXi, omegaEtaXi = omegaEtaXi,
       k = length(nonLinearXis))
}


sortXis <- function(xis, varsInts, etas, intTerms, double, structovs = NULL,
                    composites = NULL) {
  # allVarsInInts should be sorted according to which variables
  # occur in the most interaction terms (makes it more efficient)
  allVarsInInts <- unique(unlist(varsInts))
  freqInIntTerms <- lapply(varsInts, FUN = unique) |> unlist() |>
    table() |> oneWayTableToDataFrame()

  sortedXis <- c(allVarsInInts, xis[!xis %in% allVarsInInts])
  nonLinearXis <- character(0L)

  for (interaction in varsInts) {
    if (any(interaction %in% nonLinearXis) && !double ||
        all(interaction %in% nonLinearXis) && double) next # no need to add it again

    mod_stopif(length(interaction) > 2, "Only interactions between two variables are allowed")
    mod_stopif(all(interaction %in% etas),
           paste0("Interactions between two endogenous variables are not allowed!",
           "You can try passing `auto.split.syntax=TRUE` to fix the issue.",
           "See \nvignette(\"interaction_two_etas\", \"modsem\")."))

    choice <- unique(interaction[which(!interaction %in% etas &
                                       !interaction %in% nonLinearXis)])

    if (length(choice) > 1L && !double) {
      freq <- freqInIntTerms[choice, "freq"]
      choice <- choice[whichIsMax(freq)]

      is.ov <- choice %in% structovs
      is.composite <- choice %in% composites

      if (any(is.ov) && !all(is.ov)) {
        # only possible if length(choice) > 1
        k <- which(!is.ov)

      } else if (any(is.composite) && !all(is.composite)) {
        # only possible if length(choice) > 1
        k <- which(!is.composite)

      } else {
        k <- 1L # pick first if both are equal

      }

      choice <- choice[[k]]
    }

    nonLinearXis <- c(nonLinearXis, choice)
  }

  linearXis <- xis[!xis %in% nonLinearXis]

  list(
    linearXis = linearXis,
    sortedXis = c(nonLinearXis, linearXis),
    nonLinearXis = nonLinearXis
  )
}


constructOmegaEtaXi <- function(xis, etas, sortedXis, nonLinearXis,
                                varsInts, intTerms) {
  omega      <- NULL
  labelOmega <- NULL

  for (eta in etas) {
    subOmega <- matrix(0, nrow = length(xis), ncol = length(etas),
                            dimnames = list(sortedXis, etas))
    subLabelOmega <- as.character.matrix(subOmega, empty = TRUE)

    for (row in varsInts[intTerms$lhs == eta]) {
      if (!any(row %in% etas) || all(row %in% etas)) next

      whichXi  <- which(!row %in% etas)
      whichEta <- which(row %in% etas)

      subOmega[row[[whichXi]], row[[whichEta]]] <-
        getFreeOrConstIntTerms(row, eta, intTerms)
      subLabelOmega[row[[whichXi]], row[[whichEta]]] <-
        getLabelIntTerms(row, eta, intTerms)
    }

    omega      <- rbind(omega, labelRowsOmega(subOmega, eta = eta))
    labelOmega <- rbind(labelOmega, labelRowsOmega(subLabelOmega, eta = eta))
  }
  list(numeric = omega, label = labelOmega)
}


constructOmegaXiXi <- function(xis, etas, sortedXis, nonLinearXis,
                               varsInts, intTerms) {
  omega      <- NULL
  labelOmega <- NULL
  for (eta in etas) {
    subOmega <- matrix(0, nrow = length(sortedXis), ncol = length(sortedXis),
                       dimnames = list(sortedXis, sortedXis))
    subLabelOmega <- as.character.matrix(subOmega, empty = TRUE)

    for (row in varsInts[intTerms$lhs == eta]) {
       if (!all(row %in% sortedXis)) next

       whichRow <- which(row %in% nonLinearXis)[[1]] # if quadratic term pick first
       whichCol <- ifelse(whichRow == 1, 2, 1)

       subOmega[row[[whichRow]], row[[whichCol]]] <-
         getFreeOrConstIntTerms(row, eta, intTerms)
       subLabelOmega[row[[whichRow]], row[[whichCol]]] <-
         getLabelIntTerms(row, eta, intTerms)
    }

    omega      <- rbind(omega, labelRowsOmega(subOmega, eta = eta))
    labelOmega <- rbind(labelOmega, labelRowsOmega(subLabelOmega, eta = eta))
  }

  list(numeric = omega, label = labelOmega)
}


labelRowsOmega <- function(X, eta) {
  rownames(X) <- paste0(eta, "~", rownames(X))
  X
}

ThreadEnv <- rlang::env(n.threads  = NULL,
                        n.blas     = NULL,
                        n.blas.old = NULL)


blas_set_num_threads <- function(n, .default = 1L) {
  tryCatch(
    purrr::quietly(RhpcBLASctl::blas_set_num_threads)(n),
    error = function(e) {
      mod_msg_warn(paste0("Could not set threads for OpenBLAS!\nMessage: ",
               conditionMessage(e)))

      .default # return default
    }
  )
}


blas_get_num_procs <- function(.default = 1L) {
  tryCatch(
    purrr::quietly(RhpcBLASctl::blas_get_num_procs)(),
    error = function(e) {
      mod_msg_warn(paste0("Could not get processors for OpenBLAS!\nMessage: ",
               conditionMessage(e)))

      .default # return default
    }
  )
}


setThreadEnv <- function(n.threads, n.blas = NULL) {
  ThreadEnv$n.threads <- n.threads

  if (!is.null(n.blas)) {
    ThreadEnv$n.blas.old <- blas_get_num_procs()
    ThreadEnv$n.blas     <- n.blas

    blas_set_num_threads(n.blas)

  } else if (!is.null(ThreadEnv$n.blas.old)) {
    # If we're not setting threads for BLAS, we should check if we need to
    # reset them back to some previous value

    blas_set_num_threads(ThreadEnv$n.blas.old)

    ThreadEnv$n.blas     <- NULL
    ThreadEnv$n.blas.old <- NULL
  }
}


resetThreadEnv <- function() {
  setThreadEnv(n.threads = NULL, n.blas = NULL)
}


setThreads <- function(n, n.blas = 1L) {
  resetThreadEnv() # reset

  if (is.null(n))
    k <- getDefaultThreads()

  else if (is.numeric(n))
    k <- getThreadsN(n)
  else if (is.character(n)) {
    k <- switch(n,
                default = getDefaultThreads(),
                max     = getMaxThreads(),
                min     = getMinThreads(),
                mod_msg_stop("Invalid string specifying number of threads"))

  } else mod_msg_stop("Invalid number of threads, must be integer, NULL, or character")

  setThreadEnv(n.threads = k, n.blas = n.blas)
}


resetThreads <- function() {
  resetThreadEnv()
}


getDefaultThreads <- function() {
  defaultCRAN <- 2
  mc.cores    <- options("mc.cores")[[1L]]
  default     <- if (!is.null(mc.cores)) mc.cores else defaultCRAN

  getThreadsN(default)
}


getMaxThreads <- function() {
  parallel::detectCores()
}


getMinThreads <- function() {
  getThreadsN(n = 1)
}


getThreadsN <- function(n) {
  ncores <- parallel::detectCores()
  ifelse(n >= ncores, ncores, n)
}

# Calculate weights and node points for mixture functions via Gauss-Hermite
# quadrature as defined in Klein & Moosbrugger (2000). Also stores information
# regarding adaptive quadrature.
quadrature <- function(m, k,
                       adaptive = FALSE,
                       quad.range = Inf,
                       adaptive.frequency = 3,
                       ...) {
  if (quad.range < 0) {
    mod_msg_warn("`quad.range` should be positive, using `-quad.range` instead!\n")
    quad.range <- -quad.range
  }

  a <- -quad.range
  b <- quad.range

  if (k == 0 || m == 0) {
    return(list(
      n = matrix(0),
      w = 1,
      k = 0,
      m = m,
      a = a,
      b = b,
      adaptive = FALSE,
      quad.range = quad.range,
      adaptive.frequency = adaptive.frequency
    ))
  }

  singleDimGauss <- fastGHQuad::gaussHermiteData(m)
  nodes <- singleDimGauss$x
  weights <- singleDimGauss$w

  nodes_list   <- replicate(k, nodes, simplify = FALSE)
  weight_list  <- replicate(k, weights, simplify = FALSE)
  nodes   <- do.call(expand.grid, nodes_list) |> as.matrix()
  weights <- Reduce(kronecker, rev(weight_list))

  nodes <- sqrt(2) * nodes
  weights <- weights * pi ^ (-k/2)

  list(
    n = nodes,
    w = weights,
    k = k,
    m = m,
    a = a,
    b = b,
    quad.range = quad.range,
    adaptive = adaptive,
    adaptive.frequency = adaptive.frequency
  )
}


adaptiveGaussQuadrature <- function(fun,
                                    a = -7,
                                    b = 7,
                                    m = 32,
                                    m.ceil = m + m / 2,
                                    k = 1,
                                    iter.max = 10,
                                    node.max = 500,
                                    tol = 1e-12,
                                    mdiff.tol = 2,
                                    secondary.pruning = TRUE,
                                    ...) {
  if (k == 0 || m == 0)
    return(list(n = matrix(0), w = 1, f = NA, m = 1, k = 1))

  mod_stopif(tol >= 1 || tol < 0,
         "`adaptive.quad.tol` must be in the boundary `[0, 1)`")

  if (is.null(m.ceil) || all(is.na(m.ceil)))
    m.ceil <- m + m / 2

  if (k <= 1) {
    out <- adaptiveGaussQuadratureK(
      fun = fun, a = a, b = b, m = m, m.ceil = m.ceil,
      k = 1, K = k, iter = 1, iter.max = iter.max,
      node.max = node.max, tol = tol, mdiff.tol = mdiff.tol, ...
    )

    return(out)
  }

  a      <- rep(a, length.out = k)
  b      <- rep(b, length.out = k)
  m.ceil <- rep(m.ceil, length.out = k)

  NODES     <- vector("list", k)
  WEIGHTS   <- vector("list", k)
  new.ceils <- integer(k)

  for (i in seq_len(k)) {
    QUAD <- adaptiveGaussQuadratureK(
      fun = fun, a = a[i], b = b[i], m = m, m.ceil = m.ceil[i],
      k = i, K = k, iter = 1, iter.max = iter.max,
      node.max = node.max, tol = tol, mdiff.tol = mdiff.tol, ...
    )

    NODES[[i]]     <- QUAD$n[, i]
    WEIGHTS[[i]]   <- QUAD$w
    new.ceils[[i]] <- QUAD$m.ceil
  }

  quadn <- as.matrix(do.call(expand.grid, NODES))
  quadw <- Reduce(kronecker, rev(WEIGHTS))
  quadf <- fun(quadn, ...)
  weighted <- sweep(quadf, 2, quadw, "*")
  integral <- quadratureLogLik(weighted)
  error <- 0
  error.abs <- 0
  error.rel <- 0

  if (secondary.pruning) {
    pruned <- pruneQuadratureNodes(
      quadw = quadw, quadn = quadn, quadf = quadf,
      a = a, b = b, tol = tol
    )

    quadw     <- pruned$quadw
    quadn     <- pruned$quadn
    quadf     <- pruned$quadf
    integral  <- pruned$I.cur
    error     <- pruned$I.err
    error.abs <- pruned$I.err.abs
    error.rel <- pruned$I.err.rel
  }

  list(
    n = quadn,
    w = quadw,
    F = quadf,
    m.ceil = new.ceils,
    error = error,
    error.abs = error.abs,
    error.rel = error.rel,
    integral = integral
  )
}


adaptiveGaussQuadratureK <- function(fun,
                                     a = -7,
                                     b = 7,
                                     m = 32,
                                     m.ceil = m + m / 2,
                                     k = 1,
                                     K = 1,
                                     iter = 1,
                                     iter.max = 10,
                                     node.max = 500,
                                     tol = 1e-12,
                                     mdiff.tol = 2,
                                     ...) {
  if (k == 0 || m == 0)
    return(list(n = matrix(0), w = 1, f = NA, m = 1, k = 1))

  if (is.null(m.ceil) || is.na(m.ceil) || m.ceil <= 0)
    m.ceil <- round(estMForNodesInRange(m, a = -5, b = 5))

  quad  <- quadrature(m = m.ceil, k = 1)

  if (K > 1) { # more computationally efficient/scalable
    quadn <- matrix(0, nrow = m.ceil, ncol = K)
    quadn[,k] <- quad$n
  } else quadn <- quad$n

  quadf <- fun(quadn, ...)
  quadw <- quad$w

  pruned <- pruneQuadratureNodes(
    quadw = quadw, quadn = quadn, quadf = quadf,
    a = a, b = b, tol = tol
  )

  quadw     <- pruned$quadw
  quadn     <- pruned$quadn
  quadf     <- pruned$quadf
  I.err     <- pruned$I.err
  I.full    <- pruned$I.full
  I.cur     <- pruned$I.cur
  I.err.abs <- pruned$I.err.abs
  I.err.rel <- pruned$I.err.rel

  lower <- min(quadn)
  upper <- max(quadn)

  diff.m <- NROW(quadn) - m

  # Calculate next ceiling
  if (abs(diff.m) > 2 * mdiff.tol) {
    # If the difference is large, use a rough estimate
    new.ceil <- round(estMForNodesInRange(k = m, a = lower, b = upper))

    # else do a step change
  } else new.ceil <- m.ceil - diff.m

  OKNextCeil <- new.ceil <= node.max && new.ceil != m.ceil
  converged <- abs(diff.m) <= mdiff.tol
  OKIter <- iter < iter.max

  if (!converged && OKIter && OKNextCeil) {

    return(adaptiveGaussQuadratureK(
      fun = fun,
      a = a,
      b = b,
      k = k,
      K = K,
      m = m,
      m.ceil = new.ceil,
      iter = iter + 1,
      tol = tol,
      iter.max = iter.max,
      node.max = node.max,
      mdiff.tol = mdiff.tol,
      ...
    ))
  }

  mod_warnif_immediate(
    iter >= iter.max,
    paste0("Max iterations reached fitting quasi-adaptive quadrature...\n",
           sprintf("Iter: %d, total: %d, target: %d, kept: %d, discarded: %d",
                   iter, m.ceil, m, NROW(quadn), m.ceil - NROW(quadn))),
    .newline = TRUE
  )

  list(
    n = quadn,
    w = quadw,
    F = quadf,
    k = k,
    m = nrow(quadn) ^ (1 / k),

    m.ceil   = m.ceil,
    iter     = iter,
    error    = I.err,
    error.abs = I.err.abs,
    error.rel = I.err.rel,
    integral = I.cur
  )
}


estGHNodesInRange <- function(m, a, b, scale = TRUE) {
  mod_stopif(!is.numeric(m) || length(m) != 1 || m <= 0,
         "'m' must be a single positive number")
  mod_stopif(!is.numeric(a) || length(a) != 1 || !is.numeric(b) || length(b) != 1,
         "'a' and 'b' must be numeric scalars")

  if (scale) {
    a <- a / sqrt(2)
    b <- b / sqrt(2)
  }
  # Ensure interval is ordered
  lo <- min(a, b)
  hi <- max(a, b)

  # Scaling factor from the semicircle law support
  scale <- sqrt(2 * m)

  # Normalize interval endpoints to t = x / sqrt(2m)
  t_lo <- lo / scale
  t_hi <- hi / scale

  # Clamp to the theoretical support of the density, [-1, 1]
  t_lo <- pmax(pmin(t_lo,  1), -1)
  t_hi <- pmax(pmin(t_hi,  1), -1)

  # Define the cumulative function F(t) = t*sqrt(1 - t^2) + arcsin(t)
  F <- \(t) t * sqrt(pmax(0, 1 - t^2)) + asin(t)

  (m / pi) * (F(t_hi) - F(t_lo))
}


estMForNodesInRange <- function(k, a, b,
                                lower = 1,
                                upper = NULL,
                                tol = 1e-6,
                                maxiter = 100,
                                scale = TRUE) {
  mod_stopif(!is.numeric(k) || length(k) != 1 || k <= 0,
         "'k' must be a single positive number")
  mod_stopif(!is.numeric(a) || length(a) != 1 || !is.numeric(b) || length(b) != 1,
         "'a' and 'b' must be numeric scalars")

  if (scale) {
    a <- a / sqrt(2)
    b <- b / sqrt(2)
  }

  # Ensure interval
  lo <- min(a, b)
  hi <- max(a, b)

  # Define f(m) = estimate_gh_nodes(m) - k
  f <- function(m) estGHNodesInRange(m, lo, hi, scale = FALSE) - k

  # Determine upper bound if not provided by doubling
  if (is.null(upper)) {
    upper <- lower * 2
    iter <- 0
    while (f(upper) < 0 && iter < maxiter) {
      upper <- upper * 2
      iter <- iter + 1
    }
    mod_stopif(f(upper) < 0, "Unable to bracket root: increase maxiter or provide a larger 'upper'.")
  }

  # Use uniroot for inversion
  root <- stats::uniroot(f, lower = lower, upper = upper, tol = tol)

  root$root
}


quadratureLogLik <- function(weighted) {
  rs <- rowSums(weighted)

  mod_warnif(any(rs < 0), "Found negative quadrature node contributions, this is likely a bug!")

  rs.safe <- pmax(rs, .Machine$double.xmin)
  sum(log(rs.safe))
}


quadratureRelativeError <- function(error, reference) {
  if (!is.finite(reference) || reference == 0)
    return(NA_real_)

  abs(error) / abs(reference)
}


pruneQuadratureNodes <- function(quadw, quadn, quadf, a, b, tol) {
  mod_stopif(!is.numeric(quadw), "`quadw` must be numeric.")

  weight_vec <- as.numeric(quadw)
  n.nodes <- NCOL(quadf)
  mod_stopif(length(weight_vec) != n.nodes,
         "`quadw` must have the same length as the number of quadrature nodes.")

  n.input <- NROW(quadn)

  # precompute weighted information to drop empty nodes early
  weighted <- sweep(quadf, 2, weight_vec, "*")
  I.full <- quadratureLogLik(weighted)

  zeroInfoNodes <- colSums(weighted) <= .Machine$double.xmin

  lower_vec <- rep(a, length.out = NCOL(quadn))
  upper_vec <- rep(b, length.out = NCOL(quadn))
  lower_mat <- matrix(lower_vec, nrow = NROW(quadn), ncol = NCOL(quadn), byrow = TRUE)
  upper_mat <- matrix(upper_vec, nrow = NROW(quadn), ncol = NCOL(quadn), byrow = TRUE)
  nodesOutside <- rowSums((quadn < lower_mat) | (quadn > upper_mat)) > 0

  isValidNode <- !(zeroInfoNodes | nodesOutside)

  quadn <- quadn[isValidNode, , drop = FALSE]
  quadf <- quadf[, isValidNode, drop = FALSE]
  weight_vec <- weight_vec[isValidNode]
  weighted <- weighted[, isValidNode, drop = FALSE]

  # Calculate per node contributions
  rs <- rowSums(weighted)

  # guard against log(0) / division by 0
  rs.safe <- pmax(rs, .Machine$double.xmin)

  # B[j,i] = log1p( - A[j,i] / rs[j] )
  B <- log1p(-sweep(weighted, 1, rs.safe, "/"))
  I.base <- quadratureLogLik(weighted)
  I.subvec <- I.base + colSums(B)

  contributions <- abs(abs(I.subvec) - abs(I.base))

  # identify nodes trivially safe to remove
  contrib.rank <- order(abs(contributions))
  cumulative   <- cumsum(contributions[contrib.rank])
  is.removable <- abs(cumulative) < tol * abs(I.base)

  # reverse ordering of is.removable
  is.removable <- is.removable[order(contrib.rank)]
  removable    <- which(is.removable)

  mod_warnif(sum(contributions[removable]) > tol * abs(I.base),
         paste0("Something went wrong when pruning nodes:\n",
         "More information than expected was lost!\n"), .newline = TRUE)

  mod_stopif(length(removable) >= NROW(quadn), "Cannot remove all nodes!")

  if (length(removable) > 0) {
    quadn <- quadn[-removable, , drop = FALSE]
    quadf <- quadf[, -removable, drop = FALSE]
    weighted <- weighted[, -removable, drop = FALSE]
    weight_vec <- weight_vec[-removable]
  }

  I.cur <- quadratureLogLik(weighted)
  I.err <- I.full - I.cur
  I.err.abs <- abs(I.err)
  I.err.rel <- quadratureRelativeError(I.err, I.full)

  n.output  <- NROW(quadn)
  n.removed <- n.input - n.output

  list(
    quadw   = weight_vec,
    quadn   = quadn,
    quadf   = quadf,
    I.full  = I.full,
    I.cur   = I.cur,
    I.err   = I.err,
    I.err.abs = I.err.abs,
    I.err.rel = I.err.rel,
    n.in    = n.input,
    n.out   = n.output,
    n.rm    = n.removed
  )
}

OP_REPLACEMENTS <- c("~~"  = "___COVARIANCE___",
                     "=~"  = "___MEASUREMENT___",
                     ":="  = "___CUSTOM___",
                     "<~"  = "___COMPOSITE___",
                     "~"   = "___REGRESSION___",
                     ":"   = "___INTERACTION___",
                     "<->" = "___MPLUS_COVARIANCE___",
                     "<-"  = "___MPLUS_REGRESSION___",
                     "\\|" = "___THRESHOLD___")

OP_OV_INT <- OP_REPLACEMENTS[[":"]]
OP_REPLACEMENTS_INV <- structure(names(OP_REPLACEMENTS), names = OP_REPLACEMENTS)


CONSTRAINT_OPS <- c("==", ">", "<", ":=")
BOUNDUARY_OPS <- c(">", "<")


TEMP_OV_PREFIX <- ".TEMP_OV__"


getFreeParams <- function(model) {
  model$freeParams
}


fetch <- function(x, pattern = ".*") {
  x[grepl(pattern, names(x))]
}


stripMatrices <- function(matrices, fill = -1) {
  lapply(matrices, function(mat) {
    mat[!is.na(mat)] <- fill
    mat
  })
}


removeInteractions <- function(model) {
  model$matrices$OmegaEtaXi[TRUE] <- 0
  model$matrices$OmegaXiXi[TRUE] <- 0
  model
}


# Faster version of mvtnorm::dmvnorm() given that sigma is positive
# there are some drawbacks to using mvnfast. In particular,
# its a little less consistent
dmvn <- function(X, mean, sigma, log = FALSE) {
  tryCatch({
    csigma <- suppressWarnings(chol(sigma))
    mvnfast::dmvn(X, mu=mean, sigma=csigma, log=log,
                  ncores = ThreadEnv$n.threads, isChol=TRUE)
  }, error = function(e) mvtnorm::dmvnorm(X, mean, sigma, log))
}


totalLogDmvn <- function(mu, sigma, nu, S, n, d) {
  # X: n x d matrix of observations (rows = samples, cols = features)
  # mu: d-dimensional mean vector
  # sigma: d x d covariance matrix
  if (any(diag(sigma) < 0))
    return(NaN)

  tryCatch({
    sigma_inv <- chol2inv(chol(sigma))

    log_det_sigma <- determinant(sigma, logarithm = TRUE)$modulus

    trace_term <- sum(sigma_inv * S)  # Efficient trace of product
    mean_diff <- matrix(nu - mu, nrow = 1)
    mahalanobis_term <- n * (mean_diff %*% sigma_inv %*% t(mean_diff))

    log_likelihood <- -0.5 * (n * d * log(2 * pi) + n * log_det_sigma + trace_term + mahalanobis_term)
    as.numeric(log_likelihood)

  }, error = function(e) NA)
}


totalLogDvmnW <- function(X, mu, sigma, nu, S, tgamma, n, d) {
  # X: n x d matrix of data
  # mu: d-dimensional mean vector
  # sigma: d x d covariance matrix
  # gamma: n-dimensional vector of weights
  # nu: weighted observed means
  # S: weighted observed covariance matrix
  if (any(diag(sigma) < 0))
    return(NaN)

  tryCatch({
    sigma_inv <- chol2inv(chol(sigma))

    log_det_sigma <- determinant(sigma, logarithm = TRUE)$modulus

    trace_term <- sum(sigma_inv * S)
    mean_diff <- matrix(nu - mu, nrow = 1)
    mahalanobis_term <- tgamma * (mean_diff %*% sigma_inv %*% t(mean_diff))

    log_likelihood <- -0.5 * (tgamma * d * log(2 * pi) + tgamma * log_det_sigma + trace_term + mahalanobis_term)
    as.numeric(log_likelihood)

  },  error = function(e) NA)
}


diagPartitionedMat <- function(X, Y) {
  isNULL <- \(x) is.null(x) || length(x) == 0

  if (isNULL(X)) return(Y) else if (isNULL(Y)) return(X)

  if (is.null(dimnames(X)) && is.null(dimnames(Y))) {
    rownames <- NULL
    colnames <- NULL

  } else if (is.null(dimnames(X))) {
    rownames <- c(rep("", NROW(X)), rownames(Y))
    colnames <- c(rep("", NCOL(X)), colnames(Y))

  } else if (is.null(dimnames(Y))) {
    rownames <- c(rownames(X), rep("", NROW(Y)))
    colnames <- c(colnames(X), rep("", NCOL(Y)))

  } else {
    rownames <- c(rownames(X), rownames(Y))
    colnames <- c(colnames(X), colnames(Y))
  }

  structure(rbind(cbind(X, matrix(0, nrow = NROW(X), ncol = NCOL(Y))),
                  cbind(matrix(0, nrow = NROW(Y), ncol = NCOL(X)), Y)),
            dimnames = list(rownames, colnames))
}


formatNumeric <- function(x, digits = 3, scientific = FALSE,
                          justify = "right", width = NULL) {
  digits_fmt <- if (is.finite(digits)) max(0L, as.integer(digits)) else 3L
  digits_fmt_fmt <- max(1L, digits_fmt)
  if (is.numeric(x)) {
    x_round <- round(x, digits_fmt)
    format(x_round, nsmall = digits_fmt, digits = digits_fmt_fmt,
           trim = FALSE, justify = justify, scientific = scientific,
           width = width)
  } else {
    format(x, trim = FALSE, justify = justify, scientific = scientific,
           width = width)
  }
}


oneWayTableToDataFrame <- function(table) {
  df <- as.data.frame(table)
  out <- data.frame(freq = df$Freq)
  rownames(out) <- df$Var1
  out
}


whichIsMax <- function(x) {
  which(x == max(x))
}


getK_NA <- function(omega, labelOmega) {
  na    <- apply(omega, MARGIN = 1, FUN = function(x) any(is.na(x)))
  label <- apply(labelOmega, MARGIN = 1, FUN = function(x) any(x != ""))
  sum(na | label)
}


canBeNumeric <- function(x, includeNA = FALSE) {
  if (includeNA) x[x == ""] <- 0
  !is.na(suppressWarnings(as.numeric(x)))
}


createDoubleIntTerms <- function(x, z = NULL, sep = ":") {
  if (is.null(z)) {
    z <- x[[2]]
    x <- x[[1]]
  }
  c(paste0(x, sep, z), paste0(z, sep, x))
}


getFreeOrConstIntTerms <- function(varsInInt, eta, intTerms) {
  expr <- intTerms[intTerms$lhs == eta & intTerms$rhs %in%
                   createDoubleIntTerms(varsInInt), "mod"]
  if (canBeNumeric(expr, includeNA = TRUE)) as.numeric(expr) else 0
}


getLabelIntTerms <- function(varsInInt, eta, intTerms) {
  expr <- intTerms[intTerms$lhs == eta & intTerms$rhs %in%
                   createDoubleIntTerms(varsInInt), "mod"]
  if (!canBeNumeric(expr)) return(expr)
  ""
}


getEmptyModel <- function(group.info, cov.syntax, parTableCovModel,
                          mean.observed = TRUE, method = "lms",
                          orthogonal.x, orthogonal.y) {
  group.info$parTable$mod <- ""
  group.info$parTable <- removeConstraintExpressions(group.info$parTable)
  group.info["data"]  <- list(data = NULL)

  if (!is.null(cov.syntax) && is.null(parTableCovModel)) {
    parTableCovModelOrig <- modsemify(cov.syntax)

    parTableCovModel <- expandParTableByGroup(
      parTableCovModelOrig, group.levels = group.info$group.levels
    )

    group.info$parTableCov.orig <- parTableCovModelOrig
    group.info$parTableCov      <- parTableCovModel

  } else if (NROW(parTableCovModel)) {
    parTableCovModel$mod <- ""
    parTableCovModel <- removeConstraintExpressions(parTableCovModel)

    group.info$parTableCov <- parTableCovModel
  }

  specifyModelDA(
    group.info       = group.info,
    method           = method,
    cov.syntax       = cov.syntax,
    mean.observed    = mean.observed,
    auto.fix.first   = FALSE,
    auto.fix.single  = FALSE,
    createTheta      = FALSE,
    checkModel       = FALSE,
    orthogonal.x     = orthogonal.x,
    orthogonal.y     = orthogonal.y
  )
}


#' @export
as.character.matrix <- function(x, empty = TRUE, ...) {
  if (empty) x[TRUE] <- ""
  matrix(as.character(x), nrow = NROW(x), ncol = NCOL(x),
          dimnames = dimnames(x))
}


replaceNonNaModelMatrices <- function(model, value = -999) {
  .fillna <- function(x) {
    x[!is.na(x)] <- value
    x
  }

  for (g in seq_along(model$models)) {
    model$models[[g]]$matrices <- lapply(model$models[[g]]$matrices, FUN = .fillna)

    if (!is.null(model$models[[g]]$covModel$matrices))
      model$models[[g]]$covModel$matrices <- lapply(model$models[[g]]$covModel$matrices, FUN = .fillna)
  }


  model
}


removeUnknownLabels <- function(parTable) {
  ops    <- c("==", ">", "<", ":=")
  labels <- unique(parTable$mod[parTable$mod != ""])
  parTable[!parTable$op %in% ops |
           (parTable$op %in% ops &
            parTable$lhs %in% parTable$mod &
            !constraintsContainUnmatchedLabels(parTable, labels)), ]
}


getLabeledParamsLavaan <- function(parTable, fixedParams = NULL) {
  if (is.null(parTable$label)) return(NULL)
  labelRows <- parTable[parTable$label != "" &
                        !parTable$label %in% fixedParams,
                        c("est", "label"), drop = FALSE] |>
    uniqueByVar("label")

  theta <- as.numeric(labelRows$est)
  names(theta) <- labelRows$label
  theta
}


uniqueByVar <- function(df, var) {
  vals <- unique(df[[var]])
  udf  <- NULL

  for (v in vals) {
    match <- df[df[[var]] == v, , drop = FALSE][1, , drop = FALSE]
    udf <- rbind(udf, match)
  }

  udf
}


removeInteractionVariances <- function(parTable) {
  parTable[!(parTable$op == "~~" &
             (grepl(":", parTable$lhs) | grepl(":", parTable$rhs))), ]
}


tr <- function(mat) sum(diag(mat))


traceOmegaXiXi <- function(omega, numEta, numXi) {
  lastRow <- 0
  lastCol <- 0
  trace <- numeric(numEta)
  for (i in seq_len(numEta)) {
    trace[[i]] <- tr(omega[seq_len(numXi) + (i - 1) * numXi,
                           seq_len(numXi) + (i - 1) * numXi])
  }
  trace
}


diagPartitioned <- function(matrix, length) {
  out <- matrix(0, nrow = length * nrow(matrix),
                ncol = length * ncol(matrix))
  nrows <- nrow(matrix)
  rows <- seq_len(nrows)
  ncols <- ncol(matrix)
  cols <- seq_len(ncols)
  for (i in seq_len(length)) {
    out[rows + (i - 1) * nrows,
        cols + (i - 1) * ncols] <- matrix
  }
  out
}


repPartitionedRows <- function(matrix, length = 1) {
  if (length <= 1) return(matrix)
  out <- matrix
  for (i in seq_len(length - 1)) {
    out <- rbind(out, matrix)
  }
  out
}


repPartitionedCols <- function(matrix, length = 1) {
  if (length <= 1) return(matrix)
  out <- matrix
  for (i in seq_len(length - 1)) {
    out <- cbind(out, matrix)
  }
  out
}


diagBindSquareMatrices <- function(X, Y) {
  XY <- matrix(0, nrow = NROW(X), ncol = NCOL(Y),
               dimnames = list(rownames(X), colnames(Y)))
  rbind(cbind(X, XY), cbind(t(XY), Y))
}


#' @export
as.logical.matrix <- function(x, ...) {
  structure(x != 0,
            dim = dim(x),
            dimnames = dimnames(x))
}


isScalingY <- function(x) {
  (seq_along(x) %in% which(x == 0)) | seq_along(x) %in% which.max(x == 1)
}


runningAverage <- function(x, n = 10) {
  if (length(x) < n) return(0)
  last <- length(x)
  if (last < n) first <- 1 else first <- last - n + 1
  mean(x[first:last])
}


nNegativeLast <- function(x, n = 10) {
  if (length(x) < n) return(0)
  last <- length(x)
  if (last < n) first <- 1 else first <- last - n + 1
  sum(x[first:last] < 0)
}


getDegreesOfFreedom <- function(p, coef, mean.structure = TRUE, model) {
  if (!length(p)) return(NA_real_)

  momentsPerGroup <- function(pp) {
    if (mean.structure) {
      pp * (pp + 3) / 2
    } else {
      pp * (pp + 1) / 2
    }
  }

  
  df.total <- sum(vapply(p, momentsPerGroup, numeric(1L)))
  df <- df.total - length(coef)

  if (isTRUE(model$model$info$fixed.composite.var)) {

    for (g in seq_along(model$model$models)) {
      M <- model$model$models[[g]]$matricesNA

      # This would be closer to lavaan...
      # df <- df - sum(apply(M$W, MARGIN = 2L, FUN = \(x) any(is.na(x))))
      df <- df - sum(is.na(M$T))
    }
  }

  df
}


nFreeInterceptsDA <- function(model) {
  parTable   <- getMissingLabels(parameter_estimates(model))
  intercepts <- parTable$label[parTable$op == "~1"]
  coefs <- coef(model, type = "free")

  sum(intercepts %in% names(coefs))
}


getInfoQuad <- function(quad) {
  list(dim = quad$k, nodes.dim = quad$m, nodes.total = quad$m ^ quad$k)
}


getFixedInterceptSyntax <- function(indicator, syntax, parTable) {
  if (is.null(indicator) || is.null(syntax) ||
      NROW(parTable[parTable$lhs == indicator &
           parTable$op == "~1", ])) return(syntax)
  else addition <-  paste0("\n", indicator, " ~ 0 * 1")
  paste0(syntax, addition)
}


getEtaRowLabelOmega <- function(label) {
  stringr::str_split_1(label, "~")[[1]]
}


getXiRowLabelOmega <- function(label) {
  stringr::str_split_1(label, "~")[[2]]
}


expandVCOV <- function(vcov, labels) {
  labels.vcov <- colnames(vcov)
  labels.vv <- intersect(labels, labels.vcov)
  labels.zz <- setdiff(labels, labels.vcov)

  m <- length(labels.vv)
  k <- length(labels.zz)

  Vvv <- vcov[labels.vv, labels.vv]
  Vzz <- matrix(0, nrow = k, ncol = k, dimnames = list(labels.zz, labels.zz))
  Vvz <- matrix(0, nrow = m, ncol = k, dimnames = list(labels.vv, labels.zz))

  V <- rbind(cbind(Vvv, Vvz), cbind(t(Vvz), Vzz))
  V[labels, labels] # sort
}


expandCoef <- function(coef, labels) {
  labels.coef <- names(coef)
  labels.c    <- intersect(labels, labels.coef)
  labels.z    <- setdiff(labels, labels.coef)

  m <- length(labels.c)
  k <- length(labels.z)

  coef.c <- coef[labels.c]
  coef.z <- stats::setNames(numeric(length(labels.z)), nm = labels.z)

  c(coef.c, coef.z)
}


getLabelVarXZ <- function(intTerm) {
  XZ   <- stringr::str_split_fixed(intTerm, ":", 2)
  X    <- XZ[[1]]
  Z    <- XZ[[2]]

  labelXZ <- paste0(X, OP_REPLACEMENTS[[":"]], Z)
  paste0(labelXZ, OP_REPLACEMENTS[["~~"]], labelXZ)
}


intTermsAffectLV <- function(lV, parTable, etas = NULL) {
  if (grepl(":", lV)) return(TRUE)
  else if (!lV %in% parTable[parTable$op == "~", "lhs"]) return(FALSE)

  gamma <- parTable[parTable$lhs == lV & parTable$op == "~", , drop = FALSE]

  if (NROW(gamma) == 0) return(FALSE)

  for (i in seq_len(NROW(gamma))) {
    rhs <- gamma$rhs[i]
    if (intTermsAffectLV(rhs, parTable)) return(TRUE)
  }

  FALSE
}


getLambdaParTable <- function(parTable, rows = NULL, cols = NULL, fill.missing = FALSE,
                              op = c("=~", "<~")) {
  lVs <- getLVs(parTable)

  indsLVs <- getIndsLVs(parTable, lVs = lVs)
  allInds <- unique(unlist(indsLVs))

  lambda <- matrix(
    0, nrow = length(allInds), ncol = length(lVs),
    dimnames = list(allInds, lVs)
  )

  for (lV in lVs) for (ind in indsLVs[[lV]]) {
    idx <- which(
      parTable$lhs == lV & parTable$rhs == ind & parTable$op %in% op
    )

    if (length(idx))
      lambda[ind, lV] <- parTable[idx[[1L]], "est"]
  }

  if (is.null(rows)) rows <- rownames(lambda)
  if (is.null(cols)) cols <- colnames(lambda)

  if (fill.missing) {
    missingCols <- setdiff(cols, colnames(lambda))
    missingRows <- setdiff(rows, rownames(lambda))

    if (length(missingCols)) {
      newCols <- matrix(0, nrow = NROW(lambda), ncol = length(missingCols),
                        dimnames = list(rownames(lambda), missingCols))
      lambda <- cbind(lambda, newCols)
    }

    if (length(missingRows) && fill.missing) {
      newRows <- matrix(0, nrow = length(missingRows), ncol = NCOL(lambda),
                        dimnames = list(missingRows, colnames(lambda)))
      lambda <- rbind(lambda, newRows)
    }
  }

  lambda <- lambda[rows, cols, drop = FALSE]
}


getLevelsParTable <- function(parTable) {
  parTable$lhs <- stringr::str_remove_all(parTable$lhs, pattern = ":")
  parTable$rhs <- stringr::str_remove_all(parTable$rhs, pattern = ":")

  xis <- getXis(parTable, isLV = FALSE)
  etas <- getEtas(parTable, isLV = FALSE)
  vars <- c(xis, etas)

  MAPPED <- c()

  getLevelVarParTable <- function(x) {
    if      (x %in% names(MAPPED)) return(MAPPED[[x]])
    else if (x %in% xis) {
      MAPPED[[x]] <<- 1L
      return(MAPPED[[x]])
    }

    if (!x %in% etas) stop("Unexpected type of variable: ", x)

    predictors <- parTable[parTable$op == "~" & parTable$lhs == x, , drop = FALSE]

    if (!NROW(predictors)) {
      MAPPED[[x]] <<- 1L
      return(MAPPED[[x]])
    }

    maxlevel <- 0L
    for (i in seq_len(NROW(predictors))) {
      predictor <- predictors$rhs[i]
      predlevel <- getLevelVarParTable(predictor)

      if (predlevel > maxlevel) maxlevel <- predlevel
    }

    MAPPED[[x]] <<- maxlevel + 1L
    MAPPED[[x]]
  }

  for (var in vars) {
    if (var %in% names(MAPPED)) next
    getLevelVarParTable(x = var)
  }

  MAPPED
}


isPureEta <- function(eta, parTable) {
  predictors <- unique(parTable[parTable$op == "~", "rhs"])
  indicators <- unique(parTable[parTable$op == "=~", "rhs"])
  dependents <- unique(parTable[parTable$op == "~", "lhs"])
  !eta %in% c(predictors, indicators) & eta %in% dependents
}


getCoefMatricesDA_LavRepresentation <- function(parTable,
                                                xis = NULL,
                                                etas = NULL,
                                                intTerms = NULL,
                                                centered = TRUE) {

  parTable <- removeInteractionVariances(parTable)

  if (centered) {
    parTable <- centerInteractions(parTable, center.means = FALSE) |>
      var_interactions(ignore.means = TRUE) |>
      meanInteractions(ignore.means = TRUE)
  } else {
    parTable <- var_interactions(parTable, ignore.means = FALSE) |>
      meanInteractions(ignore.means = FALSE)
  }

  if (is.null(intTerms))
    intTerms <- unique(parTable[grepl(":", parTable$rhs), "rhs"])
  if (is.null(etas))
    etas <- getEtas(parTable, isLV = TRUE)
  if (is.null(xis))
    xis  <- getXis(parTable, etas = etas, isLV = TRUE)

  xis <- unique(c(xis, intTerms)) # treat int-terms as normal xis
  lVs  <- c(xis, etas)

  indsLV <- getIndsLVs(parTable, lVs = lVs)
  inds <- unique(unlist(indsLV))

  createBeta <- function(vars) {
    beta <- matrix(0, nrow = length(vars), ncol = length(vars),
                   dimnames = list(vars, vars))

    betaRows <- parTable[
      parTable$lhs %in% vars &
      parTable$op == "~" &
      parTable$rhs %in% vars, ,
      drop = FALSE
    ]

    for (i in seq_len(NROW(betaRows))) {
      lhs <- betaRows$lhs[i]
      rhs <- betaRows$rhs[i]
      est <- betaRows$est[i]

      beta[rhs, lhs] <- est
    }

    beta
  }

  createCov <- function(vars) {
    cov <- matrix(0, nrow = length(vars), ncol = length(vars),
                  dimnames = list(vars, vars))
    covRows <- parTable[parTable$op == "~~" & parTable$lhs %in% vars &
                        parTable$rhs %in% vars, , drop = FALSE]
    for (i in seq_len(NROW(covRows))) {
      lhs <- covRows$lhs[i]
      rhs <- covRows$rhs[i]
      est <- covRows$est[i]

      if (lhs %in% vars && rhs %in% vars) {
        cov[lhs, rhs] <- cov[rhs, lhs] <- est
      } else warning("Unexpected type of variable in covariance: ", lhs, " and ", rhs)
    }

    cov
  }

  createTau <- function(var) {
    tau <- matrix(0, nrow = length(var), ncol = 1,
                   dimnames = list(var, "~1"))

    tauRows <- parTable[parTable$op == "~1" & parTable$lhs %in% var, , drop = FALSE]
    for (i in seq_len(NROW(tauRows))) {
      lhs <- tauRows$lhs[i]
      est <- tauRows$est[i]

      if (lhs %in% var) {
        tau[lhs, "~1"] <- est
      } else warning("Unexpected type of variable in beta: ", lhs)
    }

    tau 
  }

  lambda <- getLambdaParTable(
    parTable = parTable, rows = inds,
    cols = lVs, fill.missing = TRUE
  )

  psi   <- createCov(c(xis, etas))
  beta  <- createBeta(c(xis, etas))
  alpha <- createTau(c(xis, etas))
  nu    <- createTau(inds)
  theta <- createCov(inds)
  binv  <- solve(diag(nrow(beta)) - beta)

  composites <- getComposites(parTable)
  compositeInds <- getCompositeIndicators(parTable)

  if (length(composites) && length(compositeInds)) {
    factorInds <- setdiff(inds, compositeInds)
    factors    <- setdiff(c(xis, etas), composites)

    W <- lambda
    T <- theta

    W[,factors]    <- 0
    W[factorInds,] <- 0
    T[,factorInds] <- 0
    T[factorInds,] <- 0

    lambda[compositeInds,] <- 0
    lambda[,composites]    <- 0
    theta[compositeInds,]  <- 0
    theta[,compositeInds]  <- 0

    lambda.c <- T %*% W %*% GINV(t(W) %*% T %*% W)
    theta.c <- T - lambda.c %*% t(W) %*% T %*% W %*% t(lambda.c)

  } else {

    T <- NULL
    W <- NULL

    lambda.c <- lambda
    lambda.c[TRUE] <- 0

    theta.c <- theta 
    theta.c[TRUE] <- 0
  }

  list(
    beta = beta, binv = binv, psi = psi,
    theta = theta, alpha = alpha, nu = nu,
    lambda = lambda, inds = inds, xis = xis, etas = etas, lVs = lVs,
    lambda.c = lambda.c, theta.c = theta.c, tmat = T, wmat = W
  )
}


calcExpectedMatricesDA <- function(parTable, xis = NULL, etas = NULL, intTerms = NULL) {
  lapply(
    X = getGroupsParTable(parTable),
    FUN = function(g) {
      parTable.g <- parTable[parTable$group == g, , drop = FALSE]
      calcExpectedMatricesDA_Group(parTable.g, xis = xis, etas = etas, intTerms = intTerms)
    }
  )
}


calcExpectedMatricesDA_Group <- function(parTable, xis = NULL, etas = NULL, intTerms = NULL) {
  # Get matrices in lavaan (i.e., lisrel) representation
  matricesCentered <- getCoefMatricesDA_LavRepresentation(
    parTable = parTable, xis = xis, etas = etas,
    intTerms = intTerms, centered = TRUE
  )

  matricesNonCentered <- getCoefMatricesDA_LavRepresentation(
    parTable = parTable, xis = xis, etas = etas,
    intTerms = intTerms, centered = FALSE
  )

  lVs  <- matricesCentered$lVs
  xis  <- matricesCentered$xis
  etas <- matricesCentered$etas
  inds <- matricesCentered$inds

  # Sigma ----------------------------------------------------------------------
  # Uses centered solution
  beta      <- matricesCentered$beta
  psi       <- matricesCentered$psi
  binv      <- matricesCentered$binv
  nu        <- matricesCentered$nu
  alpha     <- matricesCentered$alpha
  lambda    <- matricesCentered$lambda
  lambda.c  <- matricesCentered$lambda.c
  theta     <- matricesCentered$theta
  theta.c   <- matricesCentered$theta.c

  # sigma.lv <- binv %*% psi %*% t(binv)
  # We have it the opposite way with the lavaan/lisrel representation of beta/binv
  sigma.lv <- t(binv) %*% psi %*% binv
  sigma.ov <- (
    (lambda + lambda.c) %*% sigma.lv %*% t(lambda + lambda.c) + theta + theta.c
  )

  # lower left corner cov-lv-ov
  sigma.lv.ov <- (lambda + lambda.c) %*% sigma.lv
  sigma.ov.lv <- t(sigma.lv.ov)

  sigma.all <- rbind(cbind(sigma.lv, sigma.ov.lv),
                     cbind(sigma.lv.ov, sigma.ov))

  # Residuals and R^2 ----------------------------------------------------------
  # Uses centered solution
  eta.all <- c(etas, inds)
  var.eta.all <- diag(sigma.all[eta.all, eta.all, drop = FALSE])
  res.eta.all <- c(diag(psi)[etas], diag(theta))

  r2.all <- (var.eta.all - res.eta.all) / var.eta.all
  r2.lv  <- r2.all[etas]
  r2.ov  <- r2.all[inds]

  res.all <- 1 - r2.all
  res.lv  <- res.all[etas]
  res.ov  <- res.all[inds]

  # Mu -------------------------------------------------------------------------
  # Uses uncentered solution
  betaNc     <- matricesNonCentered$beta
  psiNc      <- matricesNonCentered$psi
  binvNc     <- matricesNonCentered$binv
  nuNc       <- matricesNonCentered$nu
  lambdaNc   <- matricesNonCentered$lambda
  lambdaNc.c <- matricesNonCentered$lambda.c
  alphaNc    <- matricesNonCentered$alpha

  mu.lv <- t(binvNc) %*% alphaNc
  mu.ov  <- nuNc + (lambdaNc + lambdaNc.c) %*% mu.lv
  mu.all <- rbind(mu.lv, mu.ov)

  list(
    sigma.all  = sigma.all,
    sigma.lv   = sigma.lv,
    sigma.ov   = sigma.ov,
    mu.all     = mu.all,
    mu.lv      = mu.lv,
    mu.ov      = mu.ov,
    r2.all     = r2.all,
    r2.lv      = r2.lv,
    r2.ov      = r2.ov,
    res.all    = res.all,
    res.lv     = res.lv,
    res.ov     = res.ov,
    matrices   = matricesNonCentered,
    matrices.c = matricesCentered
  )
}


getInternalCoefsDA <- function(model) {
  model$theta
}


getXisModelDA <- function(model) {
  covModelEtas <- model$covModel$info$etas
  allXis <- unique(c(model$covModel$info$xis, model$info$xis))
  allXis[!allXis %in% covModelEtas]
}


getEtasModelDA <- function(model) {
  unique(c(model$covModel$info$etas, model$info$etas))
}


splitParTableEtas <- function(parTable, parTableCov = NULL, splitEtas, allEtas,
                              nonLinearEtas = NULL) {
  if (!length(splitEtas) || !NROW(parTable))
    return(list(parTable = parTable, parTableCov = parTableCov))

  for (eta in splitEtas) {
    isReg <- parTable$lhs == eta & parTable$op == "~"
    isCov <- (parTable$lhs == eta | parTable$rhs == eta) & parTable$op == "~~"
    isLinear <- !parTable$lhs %in% nonLinearEtas &
                !parTable$rhs %in% nonLinearEtas

    parTableEta <- parTable[isReg | isCov & isLinear, ]

    vars <- unique(c(parTableEta$rhs, parTableEta$lhs))
    downstreamEtas <- vars[vars != eta & vars %in% allEtas]

    parTable    <- parTable[!(isReg | isCov & isLinear), ]
    parTableCov <- rbind(parTableCov, parTableEta)

    split <- splitParTableEtas(parTable = parTable,
                               parTableCov = parTableCov,
                               splitEtas = downstreamEtas,
                               allEtas = allEtas,
                               nonLinearEtas = nonLinearEtas)

    parTable    <- split$parTable
    parTableCov <- split$parTableCov
  }

  list(parTable = parTable, parTableCov = parTableCov)
}


splitParTable <- function(parTable) {
  intTerms <- getIntTerms(parTable)
  empty <- list(parTable = parTable, parTableCov = NULL)
  if (!length(intTerms))
    return(empty)

  intVars  <- unique(unlist(stringr::str_split(intTerms, pattern = ":")))
  etas    <- getEtas(parTable)

  if (!length(intTerms))
    return(empty)

  isNonLinear <- vapply(
    X = etas,
    FUN.VALUE = logical(1L),
    FUN = intTermsAffectLV,
    parTable = parTable
  )

  nonLinearEtas <- etas[isNonLinear]
  linearEtas    <- etas[!isNonLinear]
  etasCovModel  <- intVars[intVars %in% linearEtas]

  if (!length(etasCovModel))
    return(empty)

  splitParTableEtas(
    parTable = parTable, allEtas = etas,
    splitEtas = etasCovModel,
    nonLinearEtas = nonLinearEtas
  )
}


sortParTableDA <- function(parTable, model) {
  parTable.input <- rbind(
    model$models[[1L]]$covModel$parTable, # by definition lower order etas
    model$models[[1L]]$parTable
  )

  etas.input <- getEtas(parTable.input)
  etas.model <- getEtasModelDA(model$models[[1L]]) # sorted to be lower-triangular in gammaEta
                                                   # and includes etas from cov-model
  etas <- unique(c(etas.input, etas.model))

  # xis <- getXisModelDA(model)
  # we don't want xis as they are sorted in the model
  # xis in the model are sorted to fit `OmegaXiXi`
  # and `OmegaEtaXi`, which doesn't match the sorting
  # in the model.syntax input. Instead we use getXis() on
  # parTable.input

  xis            <- getXis(parTable.input, checkAny = FALSE, etas = etas)
  indsXis        <- model$info$allIndsXis
  indsEtas       <- model$info$allIndsEtas
  higherOrderLVs <- model$info$higherOrderLVs

  isHigherOrderXi  <- xis  %in% higherOrderLVs
  isHigherOrderEta <- etas %in% higherOrderLVs

  xisLow   <- xis[!isHigherOrderXi]
  xisHigh  <- xis[isHigherOrderXi]
  etasLow  <- etas[!isHigherOrderEta]
  etasHigh <- etas[isHigherOrderEta]

  opOrder <- c("=~", "<~", "~", "~1", "~~", "|", ":=")
  varOrder <- unique(c(indsXis, indsEtas, xisLow, etasLow, xisHigh, xisLow))
  groupOrder <- c(getGroupsParTable(parTable), 0)

  getScore <- function(x, order.by) {
    order.by <- unique(c(order.by, x)) # ensure that all of x is in order.by
    mapping  <- stats::setNames(seq_along(order.by), nm = as.character(order.by))
    score    <- mapping[as.character(x)]

    if (length(score) != length(x)) {
      mod_msg_warn("Sorting of parameter estimates failed!\n")

      return(seq_along(x))
    }

    score
  }

  scoreGroup <- getScore(x = parTable$group, order.by = groupOrder)
  scoreLhs   <- getScore(x = parTable$lhs, order.by = varOrder)
  scoreOp    <- getScore(x = parTable$op,  order.by = opOrder)
  scoreRhs   <- getScore(x = parTable$rhs, order.by = varOrder)

  out <- parTable[order(scoreGroup, scoreOp, scoreLhs, scoreRhs), , drop = FALSE]
  rownames(out) <- NULL

  out
}


updateStatusLog <- function(iterations, mode, logLikNew, deltaLL, relDeltaLL, verbose = FALSE) {
  if (verbose) {
    clearConsoleLine()
    msg <- MSG_STRINGS$strings$updateStatusLog0
    printf(msg, iterations, mode, logLikNew, deltaLL, relDeltaLL)
  }
}


higherOrderStruct2Measr <- function(parTable, indsHigherOrderLVs) {
  if (is.null(indsHigherOrderLVs) || !length(indsHigherOrderLVs))
    return(parTable)

  higherOrderLVs <- names(indsHigherOrderLVs)

  for (lVh in higherOrderLVs) {
    inds <- indsHigherOrderLVs[[lVh]]
    mask <- parTable$rhs == lVh & parTable$op == "~" & parTable$lhs %in% inds

    if (!any(mask)) next

    lhs <- parTable[mask, "lhs"]
    rhs <- parTable[mask, "rhs"]

    parTable[mask, "lhs"] <- rhs
    parTable[mask, "op"]  <- "=~"
    parTable[mask, "rhs"] <- lhs
  }

  parTable
}


higherOrderMeasr2Struct <- function(parTable) {
  higherOrderLVs <- getHigherOrderLVs(parTable)

  if (is.null(higherOrderLVs) || !length(higherOrderLVs))
    return(parTable)

  for (lVh in higherOrderLVs) {
    mask <- parTable$lhs == lVh & parTable$op == "=~"

    if (!any(mask)) next

    lhs <- parTable[mask, "lhs"]
    rhs <- parTable[mask, "rhs"]

    parTable[mask, "lhs"] <- rhs
    parTable[mask, "op"]  <- "~"
    parTable[mask, "rhs"] <- lhs
  }

  parTable
}


recalcInterceptsY <- function(parTable.nlin, parTable.lin) {
  out <- NULL

  parTable.nlin <- addMissingGroups(parTable.nlin)
  parTable.lin  <- addMissingGroups(parTable.lin)

  for (g in getGroupsParTable(parTable.nlin)) {
    parTable.nlin.g <- parTable.nlin[parTable.nlin$group == g, , drop = FALSE]
    parTable.lin.g  <- parTable.lin[parTable.lin$group == g, , drop = FALSE]
    out <- rbind(out, recalcInterceptsY_Group(parTable.nlin = parTable.nlin.g,
                                              parTable.lin  = parTable.lin.g))
  }

  rbind(out, getZeroGroupParTable(parTable.nlin))
}


recalcInterceptsY_Group <- function(parTable.nlin, parTable.lin) {
  # fix intercept for indicators of endogenous variables, based on means
  # of interaction terms
  # intercepts are from a linear (H0) model, combined with a non-linear SAM
  # structural model. We want the mean structure to be coherent with those
  # from a full non-linear model
  parTable      <- parTable.nlin
  parTable.nlin <- meanInteractions(parTable.nlin, ignore.means = FALSE)

  for (eta in getEtas(parTable)) {
    meta.lin  <- getMean(eta, parTable = parTable.lin)
    meta.nlin <- getMean(eta, parTable = parTable.nlin)

    inds <- parTable[parTable$lhs == eta & parTable$op == "=~", "rhs"]
    inds <- unique(inds)

    ieta <- getIntercept(eta, parTable)

    if (ieta != 0) {
      cond <- parTable$op == "~1" & parTable$lhs == eta
      parTable[cond, "est"] <- parTable[cond, "est"] - (meta.nlin - meta.lin) # correct for the difference
      meta.lin  <- getMean(eta, parTable = parTable.nlin) # update for the rest of the code
    }

    for (ind in inds) {
      cond <- parTable$lhs == ind & parTable$op == "~1"
      lambda <- parTable[parTable$lhs == eta &
                         parTable$op == "=~" &
                         parTable$rhs == ind, "est"]

      if (!length(lambda) || !any(cond))
        next

      intercept <- parTable[cond, "est"]
      mu.lin  <- intercept + lambda * meta.lin
      mu.nlin <- intercept + lambda * meta.nlin
      parTable[cond, "est"] <- intercept - (mu.nlin - mu.lin) # correct for the difference
    }
  }

  parTable
}


recalcInterceptsComposites <- function(parTable, input) {
  out <- NULL

  parTable <- addMissingGroups(parTable)
  input    <- addMissingGroups(input)

  for (g in getGroupsParTable(parTable)) {
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    out <- rbind(out, recalcInterceptsComposites_Group(parTable.g, input))
  }

  rbind(out, getZeroGroupParTable(parTable))
}


recalcInterceptsComposites_Group <- function(parTable, input) {
  # Translate lavaan's composite mean convention to modsem's.
  # lavaan assigns composites the mean of their indicators (W' * mu_indicators).
  # modsem assigns composites zero mean by default (for composites without an
  # explicit ~1 in the user's syntax).
  # We reparameterize: set composite ~1 = 0 and route the composite mean
  # contribution (delta = gamma * mu_c) into either the eta intercept (if the
  # user specified eta ~1) or into the indicator intercepts (the default case,
  # where eta ~1 = 0 for identification), preserving E[ind] throughout.
  composites    <- getComposites(input)
  interceptVars <- unique(input[input$op == "~1", "lhs"])
  zeroMeanComps <- setdiff(composites, interceptVars)

  for (comp in zeroMeanComps) {
    muC <- getIntercept(comp, parTable)

    if (!length(muC) || muC == 0) next

    cond.comp <- parTable$op == "~1" & parTable$lhs == comp
    if (any(cond.comp))
      parTable[cond.comp, "est"] <- 0

    structRows <- parTable[parTable$op == "~" & parTable$rhs == comp, , drop = FALSE]
    for (i in seq_len(NROW(structRows))) {
      eta   <- structRows[i, "lhs"]
      gamma <- structRows[i, "est"]
      delta <- gamma * muC

      ieta <- getIntercept(eta, parTable)
      inds <- unique(parTable[parTable$lhs == eta & parTable$op == "=~", "rhs"])

      if (ieta != 0) {
        # eta has a user-specified intercept: absorb delta into eta ~1
        cond.eta <- parTable$op == "~1" & parTable$lhs == eta
        if (any(cond.eta))
          parTable[cond.eta, "est"] <- parTable[cond.eta, "est"] + delta
      } else {
        # eta ~1 is fixed to 0 for identification: push delta into indicators
        for (ind in inds) {
          lambda   <- parTable[parTable$lhs == eta &
                               parTable$op == "=~" &
                               parTable$rhs == ind, "est"]
          cond.ind <- parTable$op == "~1" & parTable$lhs == ind
          if (!length(lambda) || !any(cond.ind)) next
          parTable[cond.ind, "est"] <- parTable[cond.ind, "est"] + lambda * delta
        }
      }
    }
  }

  parTable
}


getGroupsParTable <- function(parTable) {
  sort(unique(parTable$group[parTable$group > 0L]))
}


getZeroGroupParTable <- function(parTable) {
  parTable[parTable$group <= 0L, , drop = FALSE]
}


prepareDataGroupDA <- function(group, data, sampling.weights, sampling.weights.normalization) {
  if (!is.null(sampling.weights)) {
    mod_stopif(length(sampling.weights) != 1L, "sampling.weights variable must be of length 1!")
    mod_stopif(!sampling.weights %in% names(data),
           sprintf("sampling.weights variable '%s' not found in `data`.", sampling.weights))

    weights <- data[[sampling.weights]]
    mod_stopif(any(is.na(weights)), "`sampling.weights` cannot have missing values!")
    mod_stopif(any(weights < 0), "`sampling.weights` cannot have negative values!")
  }

  if ((!is.null(sampling.weights) && tolower(sampling.weights.normalization) == "total")) {
    weights <- data[[sampling.weights]]
    data[[sampling.weights]] <- NROW(data) * weights / sum(weights)
  }

  if (is.null(group)) {
    if ((!is.null(sampling.weights) && tolower(sampling.weights.normalization) == "group")) {
      weights <- data[[sampling.weights]]
      data[[sampling.weights]] <- NROW(data) * weights / sum(weights)
    }

    return(list(
      has.groups = FALSE,
      group = NULL,
      group.raw = NULL,
      group.var = NULL,
      levels = NULL,
      n.groups = 1L,
      indices = list(seq_len(NROW(data))),
      data = data,
      data.raw = data
    ))
  }

  data.raw <- data

  mod_stopif(!is.data.frame(data), "`data` must be a data.frame when grouping is used.")
  n <- NROW(data)
  mod_stopif(n == 0L, "Grouping requires non-empty data.")

  group.raw <- NULL
  group.var <- NULL

  mod_stopif(length(group) != 1L, "Grouping variable must be of length 1!")
  mod_stopif(!group %in% names(data),
         sprintf("Grouping variable '%s' not found in `data`.", group))

  group.raw <- data[[group]]
  data <- data[, setdiff(names(data), group), drop = FALSE]
  group.var <- group

  if ((!is.null(sampling.weights) && tolower(sampling.weights.normalization) == "group")) {
    weights <- data[[sampling.weights]]

    for (g in group.raw) {
      cond      <- group.raw == g
      weights.g <- data[cond, sampling.weights]
      data[cond, sampling.weights] <- sum(cond) * weights.g / sum(weights.g)
    }
  }

  mod_stopif(length(group.raw) != n, "Length of `group` must match the number of rows in `data`!")
  mod_stopif(any(is.na(group.raw)), "`group` cannot contain missing values!")

  group.raw <- as.character(group.raw) # match lavaan behaviour, ignore factor levels
  levels_order <- unique(group.raw)
  group_factor <- factor(group.raw, levels = levels_order)

  group.levels <- levels(group_factor)
  n.groups <- length(group.levels)
  indices <- split(seq_len(n), group_factor)

  list(
    has.groups = n.groups > 1L,
    group = group_factor,
    group.raw = group.raw,
    group.var = group.var,
    levels = group.levels,
    n.groups = n.groups,
    indices = indices,
    data = data,
    data.raw = data.raw
  )
}


expandGroupModifier <- function(mod, n.groups) {
  if (n.groups <= 1L)
    return(rep(mod, n.groups))

  if (length(mod) == 0L || mod == "")
    return(rep("", n.groups))

  if (is.na(mod))
    return(rep(NA_character_, n.groups))

  mod_trim <- trimws(mod)
  if (!nzchar(mod_trim))
    return(rep("", n.groups))

  is_c_call <- grepl("^c\\s*\\(", mod_trim) && grepl("\\)$", mod_trim)
  if (!is_c_call)
    return(rep(mod_trim, n.groups))

  inside <- substr(mod_trim,
                   start = regexpr("\\(", mod_trim, perl = TRUE) + 1L,
                   stop = nchar(mod_trim) - 1L)
  tokens <- strsplit(inside, ",", fixed = FALSE)[[1]]
  tokens <- trimws(tokens)

  if (!length(tokens)) {
    tokens <- rep("", n.groups)
  } else if (length(tokens) == 1L) {
    tokens <- rep(tokens, n.groups)
  } else {
    mod_stopif(length(tokens) != n.groups,
           sprintf("Found %d modifiers but expected %d groups.", length(tokens), n.groups))
  }

  tokens
}


expandGroupModifiers <- function(mod, n.groups) {
  do.call(rbind, lapply(mod, FUN = expandGroupModifier, n.groups = n.groups))
}


expandParTableByGroup <- function(parTable, group.levels) {
  if (!NROW(parTable))
    return(parTable) # don't return NULL, as a zero-dim parTable has a special meaning
                     # for cov models

  n.groups <- length(group.levels)

  if (n.groups <= 1L) {
    if (!"mod" %in% names(parTable)) parTable$mod <- ""
    parTable$group <- 1L
    return(parTable[, c("lhs", "op", "rhs", "group", "mod")])
  }

  constraints <- parTable[parTable$op %in% CONSTRAINT_OPS, , drop = FALSE]
  baseTable   <- parTable[!parTable$op %in% CONSTRAINT_OPS, , drop = FALSE]

  if (!"mod" %in% names(baseTable)) baseTable$mod <- ""

  MOD <- expandGroupModifiers(mod = baseTable$mod, n.groups = n.groups)

  colsOut <- c("lhs", "op", "rhs", "group", "mod")
  parTableFull <- NULL
  for (i in seq_len(n.groups)) {
    parTable.g       <- baseTable
    parTable.g$mod   <- MOD[, i]
    parTable.g$group <- i

    parTable.g <- parTable.g[colsOut]
    parTableFull <- rbind(parTableFull, parTable.g)
  }

  if (NROW(constraints)) {
    if (!"mod" %in% names(constraints)) constraints$mod <- ""
    constraints$group <- 0L
    constraints <- constraints[colsOut]
  } else {
    constraints <- data.frame(lhs = character(),
                              op = character(),
                              rhs = character(),
                              group = integer(),
                              mod = character())
  }

  rbind(parTableFull, constraints)
}


parseModelArgumentsByGroupDA <- function(model.syntax, cov.syntax,
                                         method, data, group,
                                         auto.split.syntax = FALSE,
                                         sampling.weights = NULL,
                                         sampling.weights.normalization = "total",
                                         rng.indicators = 1) {
  parTable    <- modsemify(model.syntax)
  parTableCov <- modsemify(cov.syntax)

  # Check for observed (structural) variables
  structovs <- getStructOVs(rbind(parTable, parTableCov))
  ovs       <- getOVs(rbind(parTable, parTableCov))

  missing <- setdiff(ovs, colnames(data))
  mod_stopif(length(missing), paste0("Missing observed variables in data:\n  ",
         paste(missing, collapse = ", ")))

  varsInts  <- getVarsInts(getIntTermRows(parTable), removeColonNames = FALSE)
  isOV_Int  <- vapply(varsInts, FUN.VALUE = logical(1L), FUN = \(x) all(x %in% ovs))
  ovIntTerms <- names(varsInts)[isOV_Int]

  for (ovInt in ovIntTerms) {
    vars <- varsInts[[ovInt]]
    ovIntNew <- stringr::str_replace_all(
      string = ovInt, pattern = ":",
      replacement = OP_OV_INT
    )

    parTable[parTable$lhs == ovInt, "lhs"] <- ovIntNew
    parTable[parTable$rhs == ovInt, "rhs"] <- ovIntNew
    data[[ovIntNew]] <- apply(data[vars], MARGIN = 1L, FUN = prod)
    structovs <- c(structovs, ovIntNew)
  }

  for (ov in structovs) {
    tmp.ov <- paste0(TEMP_OV_PREFIX, ov)

    # Replace ov in measurement model with tmp.ov
    # parTable[parTable$op == "=~" & parTable$rhs == ov, "rhs"] <- tmp.ov

    data[[tmp.ov]] <- data[[ov]]
    parTable <- rbind(
      parTable,
      data.frame(lhs = ov, op = "=~", rhs = tmp.ov, mod = "1"),
      data.frame(lhs = tmp.ov, op = "~1", rhs = "", mod = "0")
    )

    if (!any(parTable$lhs == ov & parTable$op == "~1")) {
      parTable <- rbind(
        parTable,
        data.frame(lhs = ov, op = "~1", rhs = "", mod = "")
      )
    }
  }

  if (length(structovs))
    model.syntax <- parTableToSyntax(parTable)

  group.info <- prepareDataGroupDA(group = group, data = data, sampling.weights = sampling.weights,
                                   sampling.weights.normalization = sampling.weights.normalization)

  if (auto.split.syntax && is.null(parTableCov) && is.null(cov.syntax)) {
    split <- splitParTable(parTable)

    parTable    <- split$parTable
    parTableCov <- split$parTableCov

    syntax     <- parTableToSyntax(parTable)
    cov.syntax <- parTableToSyntax(parTableCov)
  }

  group.info$syntax     <- model.syntax
  group.info$cov.syntax <- cov.syntax

  group.info$parTable.orig    <- parTable
  group.info$parTableCov.orig <- parTableCov
  group.info$ovIntTerms       <- ovIntTerms
  group.info$structovs        <- structovs

  group.levels <- group.info$levels
  if (is.null(group.levels)) group.levels <- ""

  group.info$group.levels <- group.levels
  group.info$parTable    <- expandParTableByGroup(parTable, group.levels = group.levels)
  group.info$parTableCov <- expandParTableByGroup(parTableCov, group.levels = group.levels)

  group.info
}


removeTempOV_RowsParTable <- function(parTable) {
  tmp <- startsWith(parTable$lhs, TEMP_OV_PREFIX) | startsWith(parTable$rhs, TEMP_OV_PREFIX)
  parTable[!tmp, , drop = FALSE]
}

m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
est <- modsem_pi(m1, oneInt)
lav_est <- extract_lavaan(est)
syntax <- get_pi_syntax(m1)
data <- get_pi_data(m1, oneInt)
lavaan::sem(syntax, data)

createExprNode <- function(node, lhs = NULL, rhs = NULL,
                           priority = 1) {
  structure(list(node = node,
                 lhs  = lhs,
                 rhs  = rhs),

            class    = "exprNode",
            priority = priority)
}


getMinTokenPriority <- function(listTokens, min = NA) {
  if (is.null(listTokens) || length(listTokens) == 0) {
    mod_stopif(is.na(min), "Unable to find minimum priority for tokens")
    return(min)
  }

  tokenPriority <- getTokenPriority(listTokens[[1]])
  if (is.na(min) || (tokenPriority < min)) {
    min <- tokenPriority
  }
  getMinTokenPriority(listTokens[-1], min)
}


chooseToken <- function(listTokens, i = 1, chosenTokenIdx = NULL,
                        leftClosures = list()) {

  if (is.null(listTokens) || i > length(listTokens)) {
    mod_stopif(is.null(chosenTokenIdx) && length(leftClosures) > 0,
           paste0("Unmatched left bracket", last(leftClosures)))
    return(chosenTokenIdx)
  }
  token <- listTokens[[i]]

  if (is.LeftClosure(token)) {
    leftClosures <- appendToList(leftClosures, token)

  } else if (is.RightClosure(token)) {
    mod_stopif(length(leftClosures) == 0,
      paste0("Unmatched right bracket", highlightErrorToken(token)),
      only.format.header = TRUE
    )

    leftClosures <- leftClosures[-1]
    if (length(leftClosures) == 0) {
      return(i)
    }
  }

  if (length(leftClosures) == 0) {
    if (is.null(chosenTokenIdx)) {
      chosenTokenIdx <- i
    }
    if (getTokenPriority(token) < getTokenPriority(listTokens[[chosenTokenIdx]])) {
      chosenTokenIdx <- i
    }
  }
  chooseToken(listTokens, i + 1, chosenTokenIdx, leftClosures)
}


createSyntaxTreeLine <- function(listTokens, i = 1) {
  minPriority <- getMinTokenPriority(listTokens)
  if (i > length(listTokens) ||
        length(listTokens) == 1 &
          getTokenPriority(listTokens[[1]]) < 0) {
    return(NULL)
  }

  if (getTokenPriority(listTokens[[i]]) == minPriority) {
    if (i != 1) {
      lhs <- createSyntaxTreeLine(listTokens[1:(i-1)], i = 1)
    } else lhs <- NULL

    if (i != length(listTokens)) {
      rhs <- createSyntaxTreeLine(listTokens[(i + 1):length(listTokens)], i = 1)
    } else rhs <- NULL

    return(
      createExprNode(listTokens[[i]], lhs = lhs, rhs = rhs,
                   priority = getTokenPriority(listTokens[[i]]))
      )
  }
  createSyntaxTreeLine(listTokens, i + 1)
}


createSyntaxTreesSyntax <- function(syntax) {
  tokensLines <- tokenizeSyntax(syntax)
  lapply(tokensLines, createSyntaxTreeLine)
}


last <- function(x) {
  if (is.null(x) || length(x) == 0) {
    return(NULL)
  }
  x[[length(x)]]
}

#' summary for modsem objects
#'
#' @param object modsem object to summarized
#' @param scientific print p-values in scientific notation
#' @param standardized standardize estimates
#' @param ci print confidence intervals
#' @param digits number of digits to print
#' @param loadings print loadings
#' @param regressions print regressions
#' @param covariances print covariances
#' @param intercepts print intercepts
#' @param variances print variances
#' @param ... arguments passed to other functions
#' @rdname summary
#' @export
summary.modsem_mplus <- function(object,
                                 scientific = FALSE,
                                 standardized = FALSE,
                                 ci = FALSE,
                                 digits = 3,
                                 loadings = TRUE,
                                 regressions = TRUE,
                                 covariances = TRUE,
                                 intercepts = TRUE,
                                 variances = TRUE,
                                 ...) {
  if (standardized)
    object$parTable <- standardized_estimates(object)

  object$format <- list(digits = digits,
                        scientific = scientific,
                        ci = ci,
                        loadings = loadings,
                        regressions = regressions,
                        covariances = covariances,
                        intercepts = intercepts,
                        variances = variances)
  structure(object, class = "summary_mplus")
}


#' @export
print.summary_mplus <- function(x, ...) {
  cat("modsem: \nMethod =", attributes(x)$method, "\n")
  printParTable(x$parTable,
                scientific = x$format$scientific,
                ci = x$format$ci,
                digits = x$format$digits,
                loadings = x$format$loadings,
                regressions = x$format$regressions,
                covariances = x$format$covariances,
                intercepts = x$format$intercepts,
                variances = x$format$variances)
}


#' @export
print.modsem_mplus <- function(x, ...) {
  cat("modsem: \nMethod =", attributes(x)$method, "\n")
  print(x$parTable)
}


#' @export
#' @describeIn parameter_estimates Get parameter estimates of a \code{\link{modsem_mplus}} object
parameter_estimates.modsem_mplus <- function(object,
                                             colon.pi = NULL, # capture
                                             high.order.as.measr = NULL, # capture
                                             rm.tmp.ov = NULL, # capture argument
                                             label.renamed.prod = NULL, # capture argument
                                             is.public = NULL, # capture argument
                                             ...) {
  object$parTable
}


#' @export
var_interactions.modsem_mplus <- function(object, ...) {
  var_interactions.data.frame(parameter_estimates(object))
}


#' @describeIn standardized_estimates Retrieve standardized estimates from \code{\link{modsem_mplus}} object.
#' @param type Type of standardized estimates to retrieve. Can be one of: \code{"stdyx", "stdy", "std", "un", "modsem"}.
#' @export
standardized_estimates.modsem_mplus <- function(object, type = "stdyx", mc.reps = 1e6, ...) {
  useMplus <- type != "modsem"

  if (useMplus) {
    tryCatch({
      coefsTable <- coef(object$model, type = type)

      mplusTableToParTable(
        coefsTable    = coefsTable,
        intTerms      = object$info$intTerms,
        intTermsMplus = object$info$intTermsMplus,
        indicators    = object$info$indicators
      )
    }, error = \(e)
        standardized_estimates.modsem_mplus(object, type = "modsem")
    )

  } else {
    stdSolution <- standardizedSolutionCOEFS(
      object,
      mc.reps = mc.reps,
      ...
    )

    stdSolution$parTable
  }
}


#' @export
#' @importFrom stats nobs
nobs.modsem_mplus <- function(object, ...) {
  NROW(object$data)
}


#' @export
#' @importFrom stats vcov
vcov.modsem_mplus <- function(object, ...) {
  object$vcov
}


#' @export
#' @importFrom stats coef
coef.modsem_mplus <- function(object, ...) {
  object$coefs
}

#' Bootstrap a modsem Model
#'
#' A generic interface for parametric and non-parametric bootstrap procedures
#' for structural equation models estimated with the \pkg{modsem} ecosystem.
#' The function dispatches on the class of \code{model}; currently dedicated
#' methods exist for \code{\link{modsem_pi}} (product‑indicator approach) and
#' \code{\link{modsem_da}} (distributional‑analytic approach).
#'
#' @param model A fitted \code{modsem} object, or a function to be bootstrapped
#'   (e.g., \code{\link{modsem}}, \code{\link{modsem_da}} and \code{\link{modsem_pi}})
#' @param FUN  A function that returns the statistic of interest when applied to
#'   a fitted model.  The function must accept a single argument, the model
#'   object, and should ideally return a numeric vector; see Value.
#' @param ...  Additional arguments forwarded to \code{lavaan::bootstrapLavaan}
#'   for \code{\link{modsem_pi}} objects, or \code{\link{modsem_da}} for
#'   \code{\link{modsem_da}} objects.
#'
#' @return Depending on the return type of \code{FUN} either
#'   \describe{
#'     \item{numeric}{A matrix with \code{R} rows (bootstrap replicates) and as
#'       many columns as \code{length(FUN(model))}.}
#'     \item{other}{A list of length \code{R}; each element is the raw output of
#'       \code{FUN}. \strong{NOTE}: Only applies for \code{\link{modsem_da}} objects}
#'   }
#' @seealso \code{\link[lavaan]{bootstrapLavaan}}, \code{\link{modsem_pi}},
#'   \code{\link{modsem_da}}
#' @export
bootstrap_modsem <- function(model = modsem, FUN, ...) {
  UseMethod("bootstrap_modsem")
}


#' @describeIn bootstrap_modsem Bootstrap a \code{modsem_pi} model by delegating
#'   to \code{\link[lavaan]{bootstrapLavaan}}.
#'
#' @details A thin wrapper around \code{lavaan::bootstrapLavaan()} that performs the
#'   necessary book‑keeping so that \code{FUN} receives a fully‑featured
#'   \code{modsem_pi} object—rather than a bare \code{lavaan} fit—at every
#'   iteration.
#'
#' @examples
#'
#' m1 <- '
#'   X =~ x1 + x2
#'   Z =~ z1 + z2
#'   Y =~ y1 + y2
#'
#'   Y ~ X + Z + X:Z
#' '
#'
#' fit_pi <- modsem(m1, oneInt)
#' bootstrap_modsem(fit_pi, FUN = coef, R = 10L)
#'
#' @export
bootstrap_modsem.modsem_pi <- function(model, FUN = "coef", ...) {
  checkWarnRCS(model)

  wrapLavFit <- function(lavfit) {
    model$lavaan       <- lavfit
    model$coefParTable <- lavaan::parameterEstimates(lavfit)
    model$data         <- lavaan::lavInspect(lavfit, what = "data")
    model
  }

  fun.mod <- \(fit, ...) do.call(FUN, list(wrapLavFit(fit), ...))

  lavaan::bootstrapLavaan(extract_lavaan(model), FUN = fun.mod, ...)
}


#' @describeIn bootstrap_modsem Parametric or non-parametric bootstrap for
#'   \code{modsem_da} models.
#'
#' @param R number of bootstrap replicates.
#' @param P.max ceiling for the simulated population size.
#' @param type Bootstrap flavour, see Details.
#' @param verbose Should progress information be printed to the console?
#' @param calc.se Should standard errors for each replicate. Defaults to \code{FALSE}.
#' @param optimize Should starting values be re-optimized for each replicate. Defaults to \code{FALSE}.
#' @param algorithm Which algorithm should be used for the LMS approach? Defaults to \code{"EM"}, as
#'   opposed to \code{"EMA"}.
#' @param convergence.abs Absolute convergence criteria. Defaults to 10 times the original criteria.
#' @param convergence.rel Relative convergence criteria. Defaults to 10 times the original criteria.
#'
#' @details The function internally resamples the observed data (non-parametric
#'   case) or simulates from the estimated parameter table (parametric case),
#'   feeds the sample to \code{\link{modsem_da}}, evaluates \code{FUN} on the
#'   refitted object and finally collates the results.
#'
#' @examples
#'
#' m1 <- '
#'   X =~ x1 + x2
#'   Z =~ z1 + z2
#'   Y =~ y1 + y2
#'
#'   Y ~ X + Z + X:Z
#' '
#'
#' \dontrun{
#' fit_lms <- modsem(m1, oneInt, method = "lms")
#' bootstrap_modsem(fit_lms, FUN = coef, R = 10L)
#' }
#' @export
bootstrap_modsem.modsem_da <- function(model,
                                       FUN = "coef",
                                       R = 1000L,
                                       P.max = 1e5,
                                       type = c("nonparametric", "parametric"),
                                       verbose = interactive(),
                                       calc.se = FALSE,
                                       optimize = FALSE,
                                       convergence.abs = 10 * model$args$convergence.abs,
                                       convergence.rel = 10 * model$args$convergence.rel,
                                       algorithm = "EM",
                                       ...) {
  checkWarnRCS(model)

  type <- tolower(type)
  type <- match.arg(type)

  INSPECT  <- modsem_inspect(model, what = "all", is.public = FALSE)
  group    <- INSPECT$group

  if (!is.null(group)) {
    group.label <- INSPECT$group.label
    DATA        <- INSPECT$data

    sampling.weights <- NULL
    cluster          <- NULL
    data             <- NULL

    for (g in group.label) {
      data.g.mat         <- DATA[[g]]
      data.g             <- as.data.frame(data.g.mat)
      cluster.g          <- attr(data.g.mat, "cluster")
      sampling.weights.g <- attr(data.g.mat, "weights")
      data.g[[group]]    <- g

      data             <- rbind(data, data.g)
      cluster          <- c(cluster, cluster.g)
      sampling.weights <- c(sampling.weights, sampling.weights.g)
    }

  } else {
    data.mat         <- INSPECT$data
    data             <- as.data.frame(data.mat)
    cluster          <- attr(data.mat, "cluster")
    sampling.weights <- attr(data.mat, "weights")
  }

  allvars  <- colnames(data)
  ovs      <- INSPECT$ovs
  N        <- NROW(data)
  P        <- min(P.max, N * R)
  parTable <- parameter_estimates(model)

  mod_warnif(P.max <= N, "`P.max` is less than `N`!")

  mod_stopif(!is.null(cluster) && type != "nonparametric",
         "cluster bootstrap is only available with `type=\"nonparametric\"`!")

  if (!is.null(cluster) && !is.null(model$args$cluster))
    data[[model$args$cluster]] <- cluster

  if (!is.null(sampling.weights) && !is.null(model$args$sampling.weights))
    data[[model$args$sampling.weights]] <- sampling.weights

  population <- switch(type,
    parametric    = simulatedGroupsToDf(simulateDataParTable(parTable, N = P, colsOVs = ovs), type = "OV"),
    nonparametric = data,
    mod_msg_stop("Unrecognized type!\n")
  )

  # Use settings from the original model, but allow the user to
  # overrride the defaults
  userArgs <- list(...)
  argList  <- model$args
  argList[names(userArgs)] <- userArgs
 
  # Except these...
  argList$algorithm       <- algorithm
  argList$convergence.abs <- convergence.abs
  argList$convergence.rel <- convergence.rel
  argList$calc.se         <- calc.se
  argList$verbose         <- verbose
  argList$model.syntax    <- model$model$info$group.info$syntax
  argList$cov.syntax      <- model$model$info$group.info$cov.syntax
  argList$method          <- model$method
  argList$sampling.weights.normalization <- "none" # This has already been done

  if (type == "parametric" && !is.null(argList$sampling.weights)) {
    # By desing, the sampling weights aren't needed when simulating
    # parametric data, so we delete the argument.
    argList$sampling.weights <- NULL
  }

  if (!optimize) {
    argList$start        <- model$theta
    argList$optimize     <- FALSE
  } else {
    argList$start        <- NULL
    argList$optimize     <- TRUE
  }

  f0 <- do.call(FUN, args = list(model))
  FUN.VEC <- is.vector(f0) || inherits(f0, "ModsemVector")

  if (FUN.VEC) {
    out <- matrix(NA, nrow = R, ncol = length(f0),
                  dimnames = list(NULL, names(f0)))

  } else out <- vector("list", length = R)

  ERROR <- \(e) {mod_msg_warn(e); NULL}

  for (i in seq_len(R)) {
    printedLines <- utils::capture.output(split = TRUE, {
      if (verbose) printf("Bootstrap %d/%d...\n", i, R)

      sample_i  <- resample(population, n.out = N, cluster = cluster)
      argList_i <- c(argList, list(data = sample_i))

      fit_i <- tryCatch(
        do.call(modsem_da, args = argList_i),
        error = ERROR
      )

      if (is.null(fit_i)) {
        fi <- NA

      } else {
        fi <- tryCatch(do.call(FUN, args = list(fit_i)),
                       error = ERROR)
      }

      if (FUN.VEC) out[i, ] <- fi else out[[i]] <- fi
    })

    nprinted <- length(printedLines)
    if (i < R) eraseConsoleLines(nprinted)
  }

  out
}


#' @describeIn bootstrap_modsem Non-parametric bootstrap of \code{modsem} functions
#'
#' @param R number of bootstrap replicates.
#' @param verbose Should progress information be printed to the console?
#' @param data Dataset to be resampled.
#' @param FUN.args Arguments passed to \code{FUN}
#' @param cluster.boot Variable to cluster bootstrapping by
#'
#' @details This is a more general version of \code{boostrap_modsem} for
#'   bootstrapping \code{modsem} functions, not modsem objects.
#'   \code{model} is now a function to be boostrapped, and \code{...}
#'   are now passed to the function (\code{model}), not \code{FUN}. To
#'   pass arguments to \code{FUN} use \code{FUN.args}.
#'
#' @examples
#'
#' tpb <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC + INT:PBC
#' "
#'
#' \dontrun{
#' boot <- bootstrap_modsem(model = modsem,
#'                          model.syntax = tpb, data = TPB,
#'                          method = "dblcent", rcs = TRUE,
#'                          rcs.scale.corrected = TRUE,
#'                          FUN = "coef", R = 50L)
#' coef <- apply(boot, MARGIN = 2, FUN = mean, na.rm = TRUE)
#' se   <- apply(boot, MARGIN = 2, FUN = sd, na.rm = TRUE)
#'
#' cat("Parameter Estimates:\n")
#' print(coef)
#'
#' cat("Standard Errors: \n")
#' print(se)
#'
#' }
#' @export
bootstrap_modsem.function <- function(model = modsem,
                                      FUN = "coef",
                                      data,
                                      R = 1000L,
                                      verbose = interactive(),
                                      FUN.args = list(),
                                      cluster.boot = NULL,
                                      ...) {
  MODELFUN <- model # rename for convenience

  out  <- vector("list", length = R)
  args <- list(...)
  data <- as.data.frame(data)
  N    <- NROW(data)

  ERROR <- \(e) {mod_msg_warn(e); NULL}

  if (!is.null(cluster.boot)) {
    mod_stopif(length(cluster.boot) > 1L, "`cluster.boot` must be of length 1!")
    mod_stopif(!cluster.boot %in% colnames(data), "`cluster.boot` must be a column in `data`!")
    cluster.values <- data[[cluster.boot]]

  } else cluster.values <- NULL

  for (i in seq_len(R)) {
    printedLines <- utils::capture.output(split = TRUE, {
      if (verbose) printf("Bootstrap %d/%d...\n", i, R)

      sample_i  <- resample(data, n.out = NROW(data), cluster = cluster.values)
      argList_i <- c(list(data = sample_i), args)

      fit_i <- tryCatch(
        do.call(MODELFUN, args = argList_i),
        error = ERROR
      )

      if (is.null(fit_i)) {
        fi <- NA

      } else {
        fi <- tryCatch(do.call(FUN, args = c(list(fit_i), FUN.args)),
                       error = ERROR)
      }

      out[[i]] <- fi
    })

    nprinted <- length(printedLines)
    if (i < R) eraseConsoleLines(nprinted)
  }

  f1 <- out[[1L]]
  FUN.VEC <- is.vector(f1) || inherits(f1, c("ModsemVector", "lavaan.vector"))

  if (FUN.VEC) {
    MAT <- tryCatch(do.call("rbind", out), error = \(e) NULL)

    if (is.null(MAT)) return(out) else return(MAT)
  }

  out
}


resample <- function(df, n.out = NROW(df), cluster = NULL, replace = TRUE) {
  df.orig <- as.data.frame(df)

  if (is.null(cluster)) {
    idx <- sample(NROW(df.orig), size = n.out, replace = replace)
    return(df.orig[idx, , drop = FALSE])
  }

  mod_stopif(length(cluster) != NROW(df.orig), "Cluster must be of same lenght as data!")

  clusters <- unique(cluster)
  G <- length(clusters)

  clusters.sample <- sample(clusters, size = G, replace = replace)

  purrr::list_rbind(
    lapply(clusters.sample, FUN = \(ci) df.orig[cluster==ci, , drop=FALSE])
  )
}


checkWarnRCS <- function(model) {
  isRCS_Model <- attr(model, "isRCS_Model")

  mod_warnif(!is.null(isRCS_Model) && isRCS_Model,
         paste0("bootstrapping a model with estimated with `rcs=TRUE` directly, will\n",
         "generate naive standard errors! Use `bootstrap_modsem.function()` instead!"))
}

optimizeStartingParamsDA <- function(model,
                                     args = list(orthogonal.x = FALSE,
                                                 orthogonal.y = FALSE,
                                                 auto.fix.first = TRUE,
                                                 auto.fix.sinlge = TRUE,
                                                 robust.se = FALSE,
                                                 sampling.weights.normalization = "none"), # already fixed by modsem
                                     group = NULL,
                                     sampling.weights = NULL,
                                     engine = c("pi", "sam")) {
  engine <- tolower(engine)
  engine <- match.arg(engine)

  etas     <- model$info$etas
  indsEtas <- model$info$allIndsEtas
  xis      <- model$info$xis
  numXis   <- model$info$numXis
  indsXis  <- model$info$allIndsXis
  data     <- model$data.raw
  missing  <- tolower(args$missing)

  syntax     <- model$info$group.info$syntax
  cov.syntax <- model$info$group.info$cov.syntax

  robust.se       <- args$robust.se
  has.interaction <- model$info$has.interaction

  syntax <- paste(syntax, cov.syntax,
                  model$info$lavOptimizerSyntaxAdditions, sep = "\n")

  acceptable.missing <- c("listwise", "ml", "direct", "fiml")
  if (has.interaction && missing %in% c("ml", "fiml", "direct")) {
    # It is not worth the extra compute to use fiml (if fiml is set)
    # for getting good starting estiamtes for the model, for non-linear models.
    # FIML can be very slow if there are a lot of product indicators.
    # However, if the model is linear, the estimates should be the
    # exact same, and product indicators aren't an issue. Thus
    # we will save time.
    missing <- "listwise"
  } else if (!missing %in% acceptable.missing) missing <- "listwise"

  if (!has.interaction && robust.se) {
    estimator <- "MLR"

    # Not worth the extra compute for non-linear models
  } else estimator <- "ML"

  if (engine == "pi") {
    estPI <- modsem_pi(
      model.syntax     = syntax,
      data             = data,
      method           = "dblcent",
      estimator        = estimator,
      meanstructure    = TRUE,
      orthogonal.x     = args$orthogonal.x,
      orthogonal.y     = args$orthogonal.y,
      auto.fix.first   = args$auto.fix.first,
      auto.fix.single  = args$auto.fix.single,
      group            = group,
      sampling.weights = sampling.weights,
      res.cov.method   = "simple.no.warn",
      res.cov.across   = TRUE,
      match            = TRUE,
      match.recycle    = TRUE,
      missing          = missing,
      suppress.warnings.match = TRUE,
      suppress.warnings.lavaan = TRUE,
      sampling.weights.normalization = args$sampling.weights.normalization
    )
    parTable   <- parameter_estimates(estPI, colon.pi = TRUE)
    lavaan.fit <- extract_lavaan(estPI)

  } else if (engine == "sam") {
    fitSam <- parameterEstimatesLavSAM(
      syntax           = syntax,
      data             = data,
      estimator        = estimator,
      missing          = missing,
      meanstructure    = TRUE,
      orthogonal.x     = args$orthogonal.x,
      orthogonal.y     = args$orthogonal.y,
      auto.fix.first   = args$auto.fix.first,
      auto.fix.single  = args$auto.fix.single,
      group            = group,
      sampling.weights = sampling.weights,
      suppress.warnings.lavaan = TRUE,
      sampling.weights.normalization = args$sampling.weights.normalization
    )

    parTable   <- fitSam$parTable
    lavaan.fit <- fitSam$fit
  }

  mod_stopif(is.null(parTable), "lavaan failed!")

  if (isHigherOrderParTable(parTable))
    parTable <- higherOrderMeasr2Struct(parTable)

  params <- model$params
  SELECT_THETA_LAB  <- params$SELECT_THETA_LAB
  SELECT_THETA_COV  <- params$SELECT_THETA_COV
  SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN
  THETA <- params$theta

  thetaLabel <- getLabeledParamsLavaan(parTable, params$constrExprs$fixedParams)
  THETA[SELECT_THETA_LAB[[1L]]][names(thetaLabel)] <- thetaLabel

  for (g in seq_len(model$info$n.groups)) {
    submodel <- model$models[[g]]

    if ("group" %in% colnames(parTable))
      parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    else
      parTable.g <- parTable

    fillLabelsMatrix <- function(matNumeric, matLabel, symmetric = FALSE) {
      if (all(matLabel == ""))
        return(matNumeric)

      labels <- c(matLabel[matLabel != ""])
      labels <- labels[labels %in% names(thetaLabel)]

      for (label in labels)
        matNumeric[matLabel == label] <- thetaLabel[[label]]

      if (symmetric)
        matNumeric[upper.tri(matNumeric)] <- t(matNumeric)[upper.tri(matNumeric)]

      matNumeric
    }

    # Main Model
    matricesMain      <- submodel$matrices
    labelMatricesMain <- submodel$labelMatrices

    LambdaX <- findEstimatesParTable(matricesMain$lambdaX, parTable.g, op = "=~",
                                     rows.lhs = FALSE, fill = 0.7)
    LambdaY <- findEstimatesParTable(matricesMain$lambdaY, parTable.g, op = "=~",
                                     rows.lhs = FALSE, fill = 0.7)

    ThetaEpsilon <- findEstimatesParTable(matricesMain$thetaEpsilon, parTable.g,
                                          op = "~~", fill = 0.2)
    ThetaDelta   <- findEstimatesParTable(matricesMain$thetaDelta, parTable.g,
                                          op = "~~", fill = 0.2)

    W <- findEstimatesParTable(matricesMain$W, parTable.g, op = "<~",
                               fill = 1, rows.lhs = FALSE)

    Psi <- findEstimatesParTable(matricesMain$psi, parTable.g, op = "~~", fill = 0)
    Phi <- findEstimatesParTable(matricesMain$phi, parTable.g, op = "~~", fill = 0)
    A   <- findEstimatesParTable(matricesMain$A, parTable.g, op = "~~", fill = 0)
    covZetaXi <- findEstimatesParTable(matricesMain$covZetaXi, parTable.g, op = "~~", fill = 0)
    T   <- findEstimatesParTable(matricesMain$T, parTable.g, op = "~~", fill = 0)

    # Matrices which can be corrected to ensure viable starting parameters need to
    # get filled in using labels as well, just for the checks them selves
    Psi <- fillLabelsMatrix(Psi, labelMatricesMain$psi, symmetric = TRUE)
    Phi <- fillLabelsMatrix(Phi, labelMatricesMain$phi, symmetric = TRUE)
    A   <- fillLabelsMatrix(A, labelMatricesMain$A, symmetric = FALSE)
    covZetaXi <- fillLabelsMatrix(covZetaXi, labelMatricesMain$covZetaXi, symmetric = FALSE)
    T   <- fillLabelsMatrix(T, labelMatricesMain$T, symmetric = FALSE)
    W   <- fillLabelsMatrix(W, labelMatricesMain$W, symmetric = FALSE)

    ThetaEpsilon <- fillLabelsMatrix(ThetaEpsilon,
                                     labelMatricesMain$thetaEpsilon,
                                     symmetric = TRUE)

    ThetaDelta <- fillLabelsMatrix(ThetaDelta,
                                   labelMatricesMain$thetaDelta,
                                   symmetric = TRUE)

    correctDiag <- function(M, fill = 1, tol = 0) {
      M[M < tol & is.diag(M)] <- fill
      M
    }

    # Check for negative diagonals
    ThetaEpsilon <- correctDiag(ThetaEpsilon, tol = 0) # no negative values
    ThetaDelta   <- correctDiag(ThetaDelta, tol = 0) # no negative values
    Psi          <- correctDiag(Psi, tol = 0) # no negative values
    Phi          <- correctDiag(Phi, tol = 0) # no negative values
    A            <- correctDiag(A, tol = 0)
    T            <- correctDiag(T, tol = 0)

    as.I <- function(M) { # If Phi/A is non-invertible we want I instead
      I <- diag(NROW(M))
      dimnames(I) <- dimnames(M)
      I
    }

    if (!is.invertible(Phi)) Phi <- as.I(Phi)
    # Residuals don't need to be invertible...
    # if (!is.invertible(Psi))          Psi <- as.I(Psi)
    # if (!is.invertible(ThetaEpsilon)) ThetaEpsilon <- as.I(ThetaEpsilon)
    # if (!is.invertible(ThetaDelta))   ThetaDelta   <- as.I(ThetaDelta)

    A[upper.tri(A)] <- t(A)[upper.tri(A)]
    A <- t(tryCatch(chol(A), error = function(x) as.I(A)))

    beta0 <- findInterceptsParTable(matricesMain$beta0, parTable.g, fill = 0)
    alpha <- findInterceptsParTable(matricesMain$alpha, parTable.g, fill = 0)

    GammaEta <- findEstimatesParTable(matricesMain$gammaEta, parTable.g, op = "~", fill = 0)
    GammaXi  <- findEstimatesParTable(matricesMain$gammaXi, parTable.g, op = "~", fill = 0)

    OmegaEtaXi <- findInteractionEstimatesParTable(matricesMain$omegaEtaXi,
                                                   parTable = parTable.g, fill = 0)
    OmegaXiXi <- findInteractionEstimatesParTable(matricesMain$omegaXiXi,
                                                  parTable = parTable.g, fill = 0)
    tauX <- findInterceptsParTable(matricesMain$tauX, parTable.g, fill = 0)
    tauY <- findInterceptsParTable(matricesMain$tauY, parTable.g, fill = 0)

    thetaMain <- unlist(list(LambdaX[is.na(matricesMain$lambdaX)],
                             LambdaY[is.na(matricesMain$lambdaY)],
                             tauX[is.na(matricesMain$tauX)],
                             tauY[is.na(matricesMain$tauY)],
                             ThetaDelta[is.na(matricesMain$thetaDelta)],
                             ThetaEpsilon[is.na(matricesMain$thetaEpsilon)],
                             W[is.na(matricesMain$W)],
                             T[is.na(matricesMain$T)],
                             Phi[is.na(matricesMain$phi)],
                             A[is.na(matricesMain$A)],
                             covZetaXi[is.na(matricesMain$covZetaXi)],
                             Psi[is.na(matricesMain$psi)],
                             alpha[is.na(matricesMain$alpha)],
                             beta0[is.na(matricesMain$beta0)],
                             GammaXi[is.na(matricesMain$gammaXi)],
                             GammaEta[is.na(matricesMain$gammaEta)],
                             OmegaXiXi[is.na(matricesMain$omegaXiXi)],
                             OmegaEtaXi[is.na(matricesMain$omegaEtaXi)]))

    # Cov Model
    matricesCov      <- submodel$covModel$matrices
    labelMatricesCov <- submodel$covModel$labelMatrices

    if (!is.null(matricesCov)) {
      PsiCovModel <- findEstimatesParTable(matricesCov$psi, parTable.g, op = "~~", fill = 0)
      GammaCovModel <- findEstimatesParTable(matricesCov$gamma, parTable.g, op = "~", fill = 0)

      PsiCovModel <- fillLabelsMatrix(PsiCovModel, labelMatricesCov$psi, symmetric = TRUE)
      PsiCovModel <- correctDiag(PsiCovModel, tol = 0)

      if (!is.invertible(PsiCovModel)) PsiCovModel <- as.I(PsiCovModel)

      # NOTE: must match the order in createThetaCovModel()
      thetaCov <- unlist(list(PsiCovModel[is.na(matricesCov$psi)],
                              GammaCovModel[is.na(matricesCov$gamma)]))
    } else thetaCov <- NULL

    selectThetaMain <- SELECT_THETA_MAIN[[g]]
    selectThetaCov  <- SELECT_THETA_COV[[g]]

    if (length(selectThetaMain) == length(thetaMain))
      THETA[selectThetaMain] <- thetaMain

    if (length(selectThetaCov) == length(thetaCov) && length(thetaCov) > 0L)
      THETA[selectThetaCov] <- thetaCov
  }

  if (length(THETA) == length(model$theta)) {
    names(THETA) <- names(model$theta)
    model$theta <- THETA
  }

  model$lavaan.fit <- lavaan.fit
  model
}


findEstimatesParTable <- function(mat, parTable, op = NULL, rows.lhs = TRUE,
                                  fill = NULL) {
  if (is.null(op)) stop("Missing operator")
  for (row in rownames(mat)) {
    for (col in colnames(mat)) {
      if (is.na(mat[row, col]))
        mat[row, col] <- extractFromParTable(row = row, op = op, col = col,
                                             parTable = parTable,
                                             rows.lhs = rows.lhs, fill = fill)
    }
  }
  mat
}


findInterceptsParTable <- function(mat, parTable, fill = NULL) {
  for (row in rownames(mat)) {
    if (is.na(mat[row, ]))
      mat[row, ] <- extractFromParTable(row = row, op = "~1", col = "",
                                        parTable = parTable, rows.lhs = TRUE,
                                        fill = fill)
  }
  mat
}


findInteractionEstimatesParTable <- function(omega, parTable, fill = NULL) {
  rows <- rownames(omega)
  cols <- colnames(omega)

  for (row in rows) for (col in cols) {
    if (!is.na(omega[row, col])) next
    eta <- getEtaRowLabelOmega(row)
    x   <- getXiRowLabelOmega(row)
    xz  <- createDoubleIntTerms(x = x, z = col, sep = ":")
    omega[row, col] <- extractFromParTable(eta, "~", xz, parTable = parTable,
                                           rows.lhs = TRUE, fill = fill)
  }
  omega
}


extractFromParTable <- function(row, op, col, parTable, rows.lhs = TRUE, fill = NULL) {
  if (rows.lhs) {
    out <- parTable[parTable$lhs == row &
                    parTable$op == op &
                    parTable$rhs %in% col, "est"]
  } else {
    out <- parTable[parTable$lhs == col &
                    parTable$op == op &
                    parTable$rhs %in% row, "est"]
  }

  if (length(out) == 0 && op == "~~") {
    out <- parTable[parTable$lhs == col & parTable$op == op &
                    parTable$rhs %in% row, "est"]
  }

  if (length(out) == 0) {
    mod_stopif(is.null(fill), "No match found")
    out <- fill
  }

  mod_stopif(length(out) > 1, "Incorrect length of matches")

  out
}


sortParTable <- function(parTable, lhs, op, rhs) {
  out <- NULL
  for (l in lhs) {
    for (r in rhs) {
      row <- parTable[parTable$lhs == l & parTable$op == op & parTable$rhs == r, ]
      if (NROW(row) == 0) next
      out <- rbind(out, row)
    }
  }
  out$est
}


parameterEstimatesLavSAM <- function(syntax,
                                     data,
                                     estimator        = "ml",
                                     missing          = "listwise",
                                     meanstructure    = TRUE,
                                     orthogonal.x     = FALSE,
                                     orthogonal.y     = FALSE,
                                     auto.fix.first   = TRUE,
                                     auto.fix.single  = TRUE,
                                     group            = NULL,
                                     sampling.weights = NULL,
                                     suppress.warnings.lavaan = TRUE,
                                     sampling.weights.normalization = "total",
                                     ...) {
  parTable <- modsemify(syntax)
  higherOrderLVs <- getHigherOrderLVs(parTable)
  isHigherOrder  <- length(higherOrderLVs) > 0L
  hasComposites  <- any(parTable$op == "<~")
  isNonCentered  <- isNonCenteredParTable(parTable)
  lowerOrderInds <- unlist(getIndsLVs(parTable, lVs = higherOrderLVs,
                                      isOV = FALSE))

  if (suppress.warnings.lavaan) wrapper <- suppressWarnings
  else                          wrapper <- \(x) x # do nothing

  lavaanVersion           <- getPackageVersion("lavaan")
  optim.gradient.override <- NULL
  if (hasComposites && utils::compareVersion(lavaanVersion, "0.6-22") < 0)
    optim.gradient.override <- "numerical" # analytical does not work

  if (!any(grepl(":", parTable$rhs) | grepl(":", parTable$lhs))) {
    fitSem <- wrapper(lavaan::sem(
      model            = syntax,
      data             = data,
      meanstructure    = meanstructure,
      estimator        = estimator,
      missing          = missing,
      orthogonal.x     = orthogonal.x,
      orthogonal.y     = orthogonal.y,
      auto.fix.first   = auto.fix.first,
      auto.fix.single  = auto.fix.single,
      group            = group,
      se               = "none",
      sampling.weights = sampling.weights,
      optim.gradient   = optim.gradient.override,
      sampling.weights.normalization = sampling.weights.normalization,
      ...
    ))

    admissible <- lavaan::lavInspect(fitSem, what = "post.check")
    mod_stopif(!admissible, "The structural model is inadmissible!")

    output <- lavaan::parameterEstimates(fitSem)

    if (hasComposites)
      output <- recalcInterceptsComposites(output, input = parTable)

    return(list(fit = fitSem, parTable = output))
  }

  # Get SAM structural model with measurement model from a linear model
  lVs <- getLVs(parTable)

  getMeasrRows <- function(pt) {
    rhs <- pt$rhs
    lhs <- pt$lhs
    op  <- pt$op

    cond1 <- op %in% c("=~", "<~")
    cond2 <- op == "~1"
    cond3 <- op == "~~" & !lhs %in% lVs & !rhs %in% lVs

    # residual variances for higher order lvs are not returned from SAM
    # estimates
    cond4 <- op == "~~" & lhs %in% lowerOrderInds & rhs %in% lowerOrderInds

    pt[cond1 | cond2 | cond3 | cond4, , drop = FALSE]
  }

  getH0Rows <- function(pt) {
    ptH0 <- pt[!grepl(":", pt$lhs) & !grepl(":", pt$rhs), , drop = FALSE]
    removeUnknownLabels(ptH0)
  }

  parTableOuter <- getH0Rows(parTable)

  syntaxH0 <- parTableToSyntax(parTableOuter)

  fitH0 <- wrapper(lavaan::sem(
    model            = syntaxH0,
    data             = data,
    meanstructure    = meanstructure,
    estimator        = estimator,
    missing          = missing,
    orthogonal.x     = orthogonal.x,
    orthogonal.y     = orthogonal.y,
    auto.fix.first   = auto.fix.first,
    auto.fix.single  = auto.fix.single,
    group            = group,
    se               = "none",
    sampling.weights = sampling.weights,
    optim.gradient   = optim.gradient.override,
    sampling.weights.normalization = sampling.weights.normalization,
    ...
  ))

  admissible <- lavaan::lavInspect(fitH0, what = "post.check")
  mod_stopif(!admissible, "The measurement model is inadmissible!")

  if (isHigherOrder || hasComposites) {
    # use factor scores instead
    # using `sam.method="fsr"` doesn't work for this purpose (yet)
    # so we do it manually instead

    dataListSam <- tryCatch(
      lavaan::lavPredict(
        object = fitH0,
        transform = TRUE,
        append.data = TRUE,
        drop.list.single.group = FALSE,
      ),
      error = function(e) {
        lavaan::lavPredict(
          object = fitH0,
          transform = FALSE,
          append.data = TRUE,
          drop.list.single.group = FALSE,
        )
      }
    )

    if (hasComposites && any(parTableOuter$op == "=~") &&
        utils::compareVersion(lavaanVersion, "0.6-99") < 0) {
      # lavPredict doesn't handle composites very well at all,
      # and it seems to f**up all the factor scores. Here we define
      # a small submodel only of the latent variables, which seems to
      # work better. Here we try to remove the composites from the model

      parTableOuterReflective <- parTableOuter[
        parTableOuter$op == "=~", , drop = FALSE
      ]

      syntaxH0b <- parTableToSyntax(parTableOuterReflective)

      fitH0b <- wrapper(lavaan::cfa(
        model            = syntaxH0b,
        data             = data,
        meanstructure    = meanstructure,
        estimator        = estimator,
        missing          = missing,
        orthogonal.x     = orthogonal.x,
        orthogonal.y     = orthogonal.y,
        auto.fix.first   = auto.fix.first,
        auto.fix.single  = auto.fix.single,
        group            = group,
        se               = "none",
        sampling.weights = sampling.weights,
        sampling.weights.normalization = sampling.weights.normalization,
        ...
      ))

      admissible <- lavaan::lavInspect(fitH0b, what = "post.check")
      mod_stopif(!admissible, "The measurement model is inadmissible!")

      dataListSamReflective <- tryCatch(
        lavaan::lavPredict(
          object = fitH0b,
          transform = TRUE,
          append.data = FALSE,
          drop.list.single.group = FALSE,
        ),
        error = function(e) {
          lavaan::lavPredict(
            object = fitH0b,
            transform = FALSE,
            append.data = FALSE,
            drop.list.single.group = FALSE,
          )
        }
      )

      for (g in seq_along(dataListSamReflective)) {
        X <- dataListSam[[g]]
        Y <- dataListSamReflective[[g]]

        replace <- intersect(colnames(X), colnames(Y))
        X[,replace] <- Y[,replace]

        dataListSam[[g]] <- X
      }
    }

    if (hasComposites && utils::compareVersion(lavaanVersion, "0.6-99") < 0) {
      # Composites are not handled properly by lavPredict (yet)

      coefListH0 <- lavaan::lavInspect(
        object = fitH0,
        what = "coef",
        drop.list.single.group = FALSE
      )

      dataListH0 <- lavaan::lavInspect(
        object = fitH0,
        what = "data",
        drop.list.single.group = FALSE
      )

      composites <- getComposites(parTableOuter)

      # variables with non-zero intercepts
      interceptVars <- parTable[parTable$op == "~1", "lhs"]

      for (g in seq_along(dataListSam)) {
        Y.g <- dataListSam[[g]]
        X.g <- dataListH0[[g]]
        W.g <- coefListH0[[g]]$wmat

        composites.g <- intersect(colnames(Y.g), composites)
     
        if (is.null(W.g) || !length(composites.g))
          next

        F.g <- X.g %*% W.g
        Y.g[,composites.g] <- F.g[,composites.g]

        # We want to mean center composites where the mean structure has
        # been fixed to zero
        for (comp0 in setdiff(composites.g, interceptVars))
          Y.g[,comp0] <- Y.g[,comp0] - mean(Y.g[,comp0], na.rm = TRUE)

        dataListSam[[g]] <- Y.g
      }
    }

    if (!is.null(sampling.weights)) {
      mod_msg_warn(
        "Ignoring sampling weights when optimizing parameter estimates..."
      )
      sampling.weights <- NULL
    }

    if (length(group)) {
      mod_stopif(length(group) > 1L,
        paste0("Unable to optimize parameters for multigroup models with more\n",
        "than one grouping variable!")
      )

      if (!is.null(names(dataListSam))) groupings <- names(dataListSam)
      else groupings <- seq_along(dataListSam)

      for (g in groupings) {
        # Append grouping varible before rbind
        # Grouping variable may be a character, so we
        # convert to a data.frame before appending
        X <- as.data.frame(dataListSam[[g]])
        X[[group]] <- g

        dataListSam[[g]] <- X
      }
    }

    dataSam <- do.call(rbind, dataListSam)

    structvars <- unique(c(
      colnames(dataSam),
      parTable[grepl(":", parTable$lhs), "lhs"],
      parTable[grepl(":", parTable$rhs), "rhs"]
    ))

    parTableInner <- parTable[
      parTable$lhs %in% structvars &
      parTable$rhs %in% structvars &
      !parTable$op %in% c("=~", "<~"), , drop = FALSE
    ]

    syntaxSam <- parTableToSyntax(parTableInner)
    SAMFUN    <- lavaan::sem

  } else {
    syntaxSam <- parTableToSyntax(parTable)
    dataSam   <- data
    SAMFUN    <- lavaan::sam
  }

  fitSam <- wrapper(SAMFUN(
    model            = syntaxSam,
    data             = dataSam,
    se               = "none",
    estimator        = estimator,
    missing          = missing,
    orthogonal.x     = orthogonal.x,
    orthogonal.y     = orthogonal.y,
    auto.fix.first   = auto.fix.first,
    auto.fix.single  = auto.fix.single,
    group            = group,
    sampling.weights = sampling.weights,
    sampling.weights.normalization = sampling.weights.normalization,
    ... # don't need to set optim.gradient here, since we only have ovs
  ))

  admissible <- lavaan::lavInspect(fitSam, what = "post.check")
  mod_stopif(!admissible, "The structural model is inadmissible!")

  parTableH0 <- lavaan::parameterEstimates(fitH0)
  measr  <- getMeasrRows(parTableH0)
  struct <- lavaan::parameterEstimates(fitSam)

  addcol <- \(pt, col, val) if (!col %in% colnames(pt)) {pt[[col]] <- val; pt} else pt
  cols.x <- c("lhs", "op", "rhs", "group")
  cols.y <- c("label", "est")
  cols   <- c(cols.x, cols.y)

  measr  <- addcol(measr, col = "label", val = "")
  measr  <- addcol(measr, col = "group", val = 1L)
  struct <- addcol(struct, col = "label", val = "")
  struct <- addcol(struct, col = "group", val = 1L)

  parTableFull <- rbind(measr[cols], struct[cols])
  parTableFull <- parTableFull[!duplicated(parTableFull[cols.x]), , drop = FALSE]

  # if (!isNonCentered && !isHigherOrder) # if latent mean structure is not included
  parTableFull <- recalcInterceptsY(
    parTable.nlin = parTableFull,
    parTable.lin  = parTableH0
  )

  if (hasComposites) {
    parTableFull <- recalcInterceptsComposites(
      parTable = parTableFull,
      input = parTable
    )
  }

  list(fit = fitH0, parTable = parTableFull)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

calcSESimpleSlopes <- function(X, V) {
    .Call(`_modsem_calcSESimpleSlopes`, X, V)
}

muSigmaLmsCpp <- function(model, z) {
    .Call(`_modsem_muSigmaLmsCpp`, model, z)
}

completeScoresNodeAnalyticalLmsCpp <- function(modelR, dataR, z, block, row, col, symmetric, colidxR, n, npatterns = 1L, ncores = 1L) {
    .Call(`_modsem_completeScoresNodeAnalyticalLmsCpp`, modelR, dataR, z, block, row, col, symmetric, colidxR, n, npatterns, ncores)
}

completeLogLikLmsCpp <- function(modelR, P, quad, colidxR, n, d, npatterns = 1L) {
    .Call(`_modsem_completeLogLikLmsCpp`, modelR, P, quad, colidxR, n, d, npatterns)
}

gradLogLikLmsCpp <- function(modelR, P, block, row, col, symmetric, colidxR, n, d, npatterns = 1L, eps = 1e-6, ncores = 1L) {
    .Call(`_modsem_gradLogLikLmsCpp`, modelR, P, block, row, col, symmetric, colidxR, n, d, npatterns, eps, ncores)
}

gradObsLogLikLmsCpp <- function(modelR, dataR, colidxR, P, block, row, col, symmetric, n, eps = 1e-6, npatterns = 1L, ncores = 1L) {
    .Call(`_modsem_gradObsLogLikLmsCpp`, modelR, dataR, colidxR, P, block, row, col, symmetric, n, eps, npatterns, ncores)
}

observedLogLikLmsCpp <- function(modelR, dataR, colidxR, P, n, npatterns = 1L, ncores = 1L) {
    .Call(`_modsem_observedLogLikLmsCpp`, modelR, dataR, colidxR, P, n, npatterns, ncores)
}

hessObsLogLikLmsCpp <- function(modelR, dataR, P, block, row, col, symmetric, colidxR, n, npatterns = 1L, relStep = 1e-6, minAbs = 0.0, ncores = 1L) {
    .Call(`_modsem_hessObsLogLikLmsCpp`, modelR, dataR, P, block, row, col, symmetric, colidxR, n, npatterns, relStep, minAbs, ncores)
}

hessCompLogLikLmsCpp <- function(modelR, P, block, row, col, symmetric, colidxR, n, d, npatterns = 1L, relStep = 1e-6, minAbs = 0.0, ncores = 1L) {
    .Call(`_modsem_hessCompLogLikLmsCpp`, modelR, P, block, row, col, symmetric, colidxR, n, d, npatterns, relStep, minAbs, ncores)
}

densityMatrixLmsCpp <- function(modelR, V, dataR, colidxR, n, samplingWeights, npatterns = 1L, ncores = 1L) {
    .Call(`_modsem_densityMatrixLmsCpp`, modelR, V, dataR, colidxR, n, samplingWeights, npatterns, ncores)
}

estepSuffStatLmsCpp <- function(P, dataR, n, npatterns = 1L, ncores = 1L) {
    .Call(`_modsem_estepSuffStatLmsCpp`, P, dataR, n, npatterns, ncores)
}

muQmlCpp <- function(m, t, ncores = 1L) {
    .Call(`_modsem_muQmlCpp`, m, t, ncores)
}

sigmaQmlCpp <- function(m, t, ncores = 1L) {
    .Call(`_modsem_sigmaQmlCpp`, m, t, ncores)
}

calcKronXi <- function(m, t, ncores = 1L) {
    .Call(`_modsem_calcKronXi`, m, t, ncores)
}

calcBinvCpp <- function(m, t, ncores = 1L) {
    .Call(`_modsem_calcBinvCpp`, m, t, ncores)
}

dnormCpp <- function(x, mu, sigma, ncores = 1L) {
    .Call(`_modsem_dnormCpp`, x, mu, sigma, ncores)
}

varZCpp <- function(Omega, Sigma1, numEta) {
    .Call(`_modsem_varZCpp`, Omega, Sigma1, numEta)
}

logLikQmlCpp <- function(submodel, ncores = 1L) {
    .Call(`_modsem_logLikQmlCpp`, submodel, ncores)
}

analyticalObsGradQmlCpp <- function(submodel, block, row, col, symmetric) {
    .Call(`_modsem_analyticalObsGradQmlCpp`, submodel, block, row, col, symmetric)
}

analyticalGradQmlCpp <- function(submodel, block, row, col, symmetric) {
    .Call(`_modsem_analyticalGradQmlCpp`, submodel, block, row, col, symmetric)
}

gradLogLikQmlFDCpp <- function(submodel, block, row, col, symmetric, eps = 1e-6, ncores = 1L) {
    .Call(`_modsem_gradLogLikQmlFDCpp`, submodel, block, row, col, symmetric, eps, ncores)
}

gradObsLogLikQmlFDCpp <- function(submodel, block, row, col, symmetric, eps = 1e-6, ncores = 1L) {
    .Call(`_modsem_gradObsLogLikQmlFDCpp`, submodel, block, row, col, symmetric, eps, ncores)
}

hessLogLikQmlCpp <- function(submodel, block, row, col, symmetric, relStep = 1e-4, minAbs = 1.0, ncores = 1L) {
    .Call(`_modsem_hessLogLikQmlCpp`, submodel, block, row, col, symmetric, relStep, minAbs, ncores)
}

multiplyIndicatorsCpp <- function(df) {
    .Call(`_modsem_multiplyIndicatorsCpp`, df)
}

checkOpenMP_Cpp <- function() {
    .Call(`_modsem_checkOpenMP_Cpp`)
}

dmvnrmArmaMc <- function(x, mean, sigma, log = TRUE, ncores = 1L) {
    .Call(`_modsem_dmvnrmArmaMc`, x, mean, sigma, log, ncores)
}

repDmvnormCpp <- function(x, expected, sigma, t, ncores = 1L) {
    .Call(`_modsem_repDmvnormCpp`, x, expected, sigma, t, ncores)
}

totalDmvnWeighted <- function(mu, sigma, nu, S, tgamma, d) {
    .Call(`_modsem_totalDmvnWeighted`, mu, sigma, nu, S, tgamma, d)
}

dmvnfast <- function(X, mu, sigma, log, ncores, isChol) {
    .Call(`_modsem_dmvnfast`, X, mu, sigma, log, ncores, isChol)
}

modelMatrixCacheCpp <- function(matrices) {
    .Call(`_modsem_modelMatrixCacheCpp`, matrices)
}

impliedEtaFromZetaCpp <- function(zeta, xptr) {
    .Call(`_modsem_impliedEtaFromZetaCpp`, zeta, xptr)
}

impliedYFromEtaCpp <- function(eta, xptr) {
    .Call(`_modsem_impliedYFromEtaCpp`, eta, xptr)
}

logLikFromZetaMLCpp <- function(zeta, y, xptr, idx) {
    .Call(`_modsem_logLikFromZetaMLCpp`, zeta, y, xptr, idx)
}

gradLogLikFromZetaMLCpp <- function(zeta, y, xptr, idx, eps = 1e-4) {
    .Call(`_modsem_gradLogLikFromZetaMLCpp`, zeta, y, xptr, idx, eps)
}

hessLogLikFromZetaMLCpp <- function(zeta, y, xptr, idx, relStep = 1e-4, minAbsPar = 0.0) {
    .Call(`_modsem_hessLogLikFromZetaMLCpp`, zeta, y, xptr, idx, relStep, minAbsPar)
}

logLikFromZetaEBMCpp <- function(zeta, y, xptr, idx) {
    .Call(`_modsem_logLikFromZetaEBMCpp`, zeta, y, xptr, idx)
}

gradLogLikFromZetaEBMCpp <- function(zeta, y, xptr, idx, eps = 1e-4) {
    .Call(`_modsem_gradLogLikFromZetaEBMCpp`, zeta, y, xptr, idx, eps)
}

hessLogLikFromZetaEBMCpp <- function(zeta, y, xptr, idx, relStep = 1e-4, minAbsPar = 0.0) {
    .Call(`_modsem_hessLogLikFromZetaEBMCpp`, zeta, y, xptr, idx, relStep, minAbsPar)
}

tracePathsNumericCpp <- function(x, y, parTable, maxlen = 100L) {
    .Call(`_modsem_tracePathsNumericCpp`, x, y, parTable, maxlen)
}

tracePathsCharacterCpp <- function(x, y, parTable, paramCol = "mod", maxlen = 100L) {
    .Call(`_modsem_tracePathsCharacterCpp`, x, y, parTable, paramCol, maxlen)
}

devtools::load_all()

OK_OpenMP <- checkOpenMP_Cpp()
startupMessage <- purrr::quietly(devtools::load_all)()$messages[-1L]
startupMessage <- stringr::str_split(startupMessage, pattern = "\n")[[1L]] |>
  (\(x) (x[x != ""]))() |>
  stringr::str_pad(width = 4) |> paste0(collapse = "\n")


printSessionInfo <- function() {
  info <- sessionInfo()
  info$date <- tryCatch(as.character(Sys.Date()), error = "<NA>")
  info$openmp <- if (OK_OpenMP) "AVAILABLE" else "NOT AVAILABLE"

  show <- (sapply(info, FUN = length) == 1L) & names(info) != "locale"
  lhs <- names(info[show])
  rhs <- unlist(info[show])

  # cut large paths
  isLargePath <- stringr::str_count(rhs, pattern = "/") > 3L
  getstem <- \(path, i) stringr::str_split_i(path, pattern = "/", i = i)

  largePaths <- rhs[isLargePath]
  rhs[isLargePath] <- paste(
    "...",
    getstem(largePaths, -3L),
    getstem(largePaths, -2L),
    getstem(largePaths, -1L),
    sep = "/"
  )

  flhs <- format(lhs, justify = "left")
  frhs <- format(rhs, justify = "right")

  width <- options("width")[[1L]]
  space <- strrep(" ", max(2L, width - max(nchar(lhs)) - max(nchar(rhs))))
  sep1  <- paste0(strrep("\u2500", width), "\n")
  sep2  <- paste0(strrep("\u2550", width), "\n")

  cat("\n")
  cat(sep1)
  cat("RUNNING TESTS CASES\n")
  cat(sep2)

  for (i in seq_along(lhs))
    cat(flhs[i], space, frhs[i], "\n", sep = "")

  cat(sep1)
  cat("STARTUP MESSAGE:\n")
  cat(startupMessage, "\n")

  cat(sep2, "\n")
}

printSessionInfo()

OpenMP_Pattern <- "OpenMP is not available.*"

if (OK_OpenMP) {
  OK_OpenMP_Message <- !grepl(OpenMP_Pattern, startupMessage)
} else  {
  OK_OpenMP_Message <-  grepl(OpenMP_Pattern, startupMessage)
}
testthat::expect_true(OK_OpenMP_Message)

devtools::load_all()
set.seed(123)

m1 <- '
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3
Y ~ X + Z + a * X:Z

X ~~ varX * X # check that no warning is thrown
              # (one should be thrown for LMS but not QML)
'

est1 <- modsem(m1, data = oneInt, convergence.rel = 1e-2, method = "qml",
               robust.se = TRUE)
print(summary(est1, scientific = TRUE))
plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z", vals_z = c(-0.5, 0.5), model = est1)
plot_surface(x = "X", z = "Z", y = "Y", model = est1)

testthat::test_that("QML honors finite-difference gradient fallback", {
  model <- est1$start.model
  if (is.null(model)) model <- est1$model

  model$params$gradientStruct$useFDGradient <- TRUE
  model$params$gradientStruct$locations <- NULL
  model$params$gradientStruct$Jacobian <- NULL

  grad <- gradientLogLikQml(theta = model$theta, model = model, sum = TRUE)
  scores <- gradientLogLikQml(theta = model$theta, model = model, sum = FALSE)

  testthat::expect_length(grad, length(model$theta))
  testthat::expect_equal(ncol(scores), length(model$theta))
})

testthat::test_that("QML FD mapping includes composite variance blocks", {
  model <- est1$start.model
  if (is.null(model)) model <- est1$model

  submodel <- fillModel(theta = model$theta, model = model,
                        method = "qml")$models[[1L]]
  submodel$matrices$W <- matrix(1, 1, 1)
  submodel$matrices$T <- matrix(1, 1, 1)

  grad <- gradLogLikQmlFDCpp(
    submodel = submodel,
    block = c(14L, 15L),
    row = c(0L, 0L),
    col = c(0L, 0L),
    symmetric = c(0L, 1L)
  )

  testthat::expect_true(all(is.finite(grad)))
})


tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

  # Covariances
  ATT ~~ SN + PBC
  PBC ~~ SN
# Inner Model (Based on Steinmetz et al., 2011)
  BEH ~ b * INT + PBC
  INT ~ ATT + SN + a * PBC
  BEH ~ PBC:INT

  INT ~~ BEH
  VERY_LONG_LABEL_THAT_SHOULD_BE_SHORTENED := a * b
'

testthat::expect_warning(
est2 <- modsem(tpb, data = TPB, method = "qml",
               robust.se = TRUE,
               standardize = TRUE, convergence.rel = 1e-2),
regexp = "Covariances .*"
)

print(summary(est2, H0 = FALSE))
expect_warning(plot_jn(x = "INT", z = "PBC", y = "BEH", model = est2,
                       min_z = -1.5, max_z = -0.5),
               regex = "Truncating.*right.*left.*")
expect_warning(plot_jn(x = "INT", z = "PBC", y = "BEH", model = est2,
                       min_z = -1.5, max_z = 2.5) ,
               regex = "Truncating SD-range on the left!")
expect_warning(plot_jn(x = "INT", z = "PBC", y = "BEH", model = est2,
                       min_z = -2, max_z = 1) ,
               regex = "Truncating SD-range on the right!")

# check that standardizing twice does not change the estimates
deterministicCols <- c("lhs", "op", "rhs", "est") # std.error based cols are subject to rng
testthat::expect_equal(standardized_estimates(est2)[deterministicCols],
                       parameter_estimates(est2)[deterministicCols], tol = 1e-3)

# check correct standardization
calcCovParTable("BEH", "BEH", parameter_estimates(est2))[[1]] |>
  testthat::expect_equal(1)

# check that estimates are standardized the same way
# in vcov, coef and parameter_estimates
pt <- parameter_estimates(est2)
vcov_sd <- sqrt(vcov(est2)["BEH~PBC:INT", "BEH~PBC:INT"])
cond <- pt$lhs == "BEH" & pt$op == "~" & pt$rhs == "PBC:INT"
pt_sd <- pt[cond, "std.error"]
pt_est <- pt[cond, "est"]
coef_est <- coef(est2)[["BEH~PBC:INT"]]

expect_equal(vcov_sd, pt_sd)
expect_equal(pt_est, coef_est)

modsem_inspect(est2)
coefficients(est2)

# Observed Variables
m3 <- '
X =~ x1
Y =~ y1
Z =~ z1
Y ~ X + Z + X:Z
'

est3 <- modsem(m3, data = oneInt, convergence.rel = 1e-2, method = "qml",
               robust.se = TRUE)
print(summary(est3, scientific = TRUE))
plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z", vals_z = c(-0.5, 0.5), model = est3,
                 standardized = TRUE)

tpb2 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 #+ int2 + int3
  BEH =~ b1  + b2

# Inner Model (Based on Steinmetz et al., 2011)
  BEH ~ INT + PBC
  INT ~ ATT + SN + PBC
  BEH ~ PBC:INT
'

est4 <- modsem(tpb2, data = TPB, method = "qml",
               robust.se = TRUE,
               standardize = TRUE, convergence.rel = 1e-2)
print(summary(est4, H0 = FALSE))

testthat::expect_equal(standardized_estimates(est4)[deterministicCols],
                       parameter_estimates(est4)[deterministicCols])

calcCovParTable("BEH", "BEH", parameter_estimates(est4))[[1]] |>
  testthat::expect_equal(1)

vcov(est4)
modsem_inspect(est4)
coef(est4)
coefficients(est4)

testthat::test_that("simple QML analytical gradient matches C++ FD for supported blocks", {
  syntax <- '
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3
  Y ~ X + Z + X:Z
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = syntax,
    cov.syntax = NULL,
    method = "qml",
    data = oneInt,
    group = NULL,
    auto.split.syntax = FALSE
  )

  model <- specifyModelDA(
    group.info = group.info,
    method = "qml",
    m = 16,
    mean.observed = TRUE,
    double = FALSE,
    quad.range = Inf,
    adaptive.quad = FALSE,
    adaptive.frequency = 3,
    missing = "complete",
    orthogonal.x = FALSE,
    orthogonal.y = FALSE,
    auto.fix.first = TRUE,
    auto.fix.single = TRUE,
    fix.composite.var = TRUE,
    cluster = NULL,
    sampling.weights = NULL
  )

  model.filled <- fillModel(model = model, theta = model$theta, method = "qml")
  locations <- subset(
    model$params$gradientStruct$locations,
    group == 1 & block %in% c(0, 1, 2, 3, 4, 5, 7, 8, 9, 10, 12, 16)
  )

  grad.analytical <- analyticalGradQmlCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric
  )

  grad.fd <- gradLogLikQmlFDCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric,
    eps = 1e-7,
    ncores = 1L
  )

  testthat::expect_true(all(is.finite(grad.analytical)))
  testthat::expect_equal(grad.analytical, grad.fd, tolerance = 3e-3)

  tpb.linear <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = tpb.linear,
    cov.syntax = NULL,
    method = "qml",
    data = TPB,
    group = NULL,
    auto.split.syntax = FALSE
  )

  model <- specifyModelDA(
    group.info = group.info,
    method = "qml",
    m = 16,
    mean.observed = TRUE,
    double = FALSE,
    quad.range = Inf,
    adaptive.quad = FALSE,
    adaptive.frequency = 3,
    missing = "complete",
    orthogonal.x = FALSE,
    orthogonal.y = FALSE,
    auto.fix.first = TRUE,
    auto.fix.single = TRUE,
    fix.composite.var = TRUE,
    cluster = NULL,
    sampling.weights = NULL
  )

  model.filled <- fillModel(model = model, theta = model$theta, method = "qml")
  locations <- subset(
    model$params$gradientStruct$locations,
    group == 1 & block == 11
  )

  grad.analytical <- analyticalGradQmlCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric
  )

  grad.fd <- gradLogLikQmlFDCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric,
    eps = 1e-6,
    ncores = 1L
  )

  testthat::expect_true(all(is.finite(grad.analytical)))
  testthat::expect_equal(grad.analytical, grad.fd, tolerance = 2e-3)

  tpb.interaction <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
  BEH ~ ATT:PBC
  ATT ~ 1
  PBC ~ 1
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = tpb.interaction,
    cov.syntax = NULL,
    method = "qml",
    data = TPB,
    group = NULL,
    auto.split.syntax = FALSE
  )

  suppressWarnings({
    model <- specifyModelDA(
      group.info = group.info,
      method = "qml",
      m = 16,
      mean.observed = TRUE,
      double = FALSE,
      quad.range = Inf,
      adaptive.quad = FALSE,
      adaptive.frequency = 3,
      missing = "complete",
      orthogonal.x = FALSE,
      orthogonal.y = FALSE,
      auto.fix.first = TRUE,
      auto.fix.single = TRUE,
      fix.composite.var = TRUE,
      cluster = NULL,
      sampling.weights = NULL
    )
  })

  model.filled <- fillModel(model = model, theta = model$theta, method = "qml")
  locations <- subset(
    model$params$gradientStruct$locations,
    group == 1 & block %in% c(1, 2, 3, 5, 9, 10, 11, 12, 13)
  )

  grad.analytical <- analyticalGradQmlCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric
  )

  grad.fd <- gradLogLikQmlFDCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric,
    eps = 1e-6,
    ncores = 1L
  )

  testthat::expect_true(model.filled$models[[1]]$info$kOmegaEta > 0L)
  testthat::expect_equal(sort(unique(locations$block)),
                         c(1, 2, 3, 5, 9, 10, 11, 12, 13))
  testthat::expect_true(all(is.finite(grad.analytical)))
  testthat::expect_equal(grad.analytical, grad.fd, tolerance = 2e-3)
})

testthat::test_that("split QML uses hybrid gradient through covModel Jacobian", {
  tpb <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = tpb,
    cov.syntax = NULL,
    method = "qml",
    data = TPB,
    group = NULL,
    auto.split.syntax = TRUE
  )

  model <- specifyModelDA(
    group.info = group.info,
    method = "qml",
    m = 16,
    mean.observed = TRUE,
    double = FALSE,
    quad.range = Inf,
    adaptive.quad = FALSE,
    adaptive.frequency = 3,
    missing = "complete",
    orthogonal.x = FALSE,
    orthogonal.y = FALSE,
    auto.fix.first = TRUE,
    auto.fix.single = TRUE,
    fix.composite.var = TRUE,
    cluster = NULL,
    sampling.weights = NULL
  )

  grad.hybrid <- gradientLogLikQml(
    theta = model$theta,
    model = model,
    sign = 1,
    epsilon = 1e-6
  )

  objective <- function(theta) {
    model.filled <- fillModel(model = model, theta = theta, method = "qml")
    logLikQmlCpp(model.filled$models[[1L]], ncores = 1L)
  }

  f0 <- objective(model$theta)
  grad.fd <- numeric(length(model$theta))
  indices <- c(
    model$params$SELECT_THETA_COV[[1L]],
    model$params$SELECT_THETA_MAIN[[1L]]
  )

  for (i in indices) {
    theta.i <- model$theta
    theta.i[[i]] <- theta.i[[i]] + 1e-6
    grad.fd[[i]] <- (objective(theta.i) - f0) / 1e-6
  }

  testthat::expect_true(isTRUE(model$params$gradientStruct$hasCovModel))
  testthat::expect_true(all(is.finite(grad.hybrid)))
  testthat::expect_equal(grad.hybrid[indices], grad.fd[indices],
                         tolerance = 3e-3)
})

testthat::test_that("split QML uses hybrid observation scores through covModel Jacobian", {
  tpb <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = tpb,
    cov.syntax = NULL,
    method = "qml",
    data = TPB,
    group = NULL,
    auto.split.syntax = TRUE
  )

  suppressWarnings({
    model <- specifyModelDA(
      group.info = group.info,
      method = "qml",
      m = 16,
      mean.observed = TRUE,
      double = FALSE,
      quad.range = Inf,
      adaptive.quad = FALSE,
      adaptive.frequency = 3,
      missing = "complete",
      orthogonal.x = FALSE,
      orthogonal.y = FALSE,
      auto.fix.first = TRUE,
      auto.fix.single = TRUE,
      fix.composite.var = TRUE,
      cluster = NULL,
      sampling.weights = NULL
    )
  })

  scores.hybrid <- gradientLogLikQml_i(
    theta = model$theta,
    model = model,
    sign = 1,
    epsilon = 1e-6
  )

  obs_objective <- function(theta) {
    model.filled <- fillModel(model = model, theta = theta, method = "qml")
    logLikQmlGroup(model.filled$models[[1L]], sign = 1, sum = FALSE)
  }

  f0 <- obs_objective(model$theta)
  indices <- c(
    model$params$SELECT_THETA_COV[[1L]],
    model$params$SELECT_THETA_MAIN[[1L]]
  )
  indices <- indices[seq_len(min(12L, length(indices)))]

  scores.fd <- matrix(0, nrow = NROW(scores.hybrid), ncol = length(indices))
  colnames(scores.fd) <- names(model$theta)[indices]

  for (j in seq_along(indices)) {
    theta.j <- model$theta
    theta.j[[indices[[j]]]] <- theta.j[[indices[[j]]]] + 1e-6
    scores.fd[, j] <- (obs_objective(theta.j) - f0) / 1e-6
  }

  testthat::expect_true(isTRUE(model$params$gradientStruct$hasCovModel))
  testthat::expect_true(all(is.finite(scores.hybrid[, indices])))
  testthat::expect_equal(scores.hybrid[, indices], scores.fd,
                         tolerance = 1e-4)
})

testthat::test_that("unsplit QML uses analytical non-constant-Binv observation scores", {
  tpb <- '
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
  BEH ~ ATT:PBC
  ATT ~ 1
  PBC ~ 1
  '

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax = tpb,
    cov.syntax = NULL,
    method = "qml",
    data = TPB,
    group = NULL,
    auto.split.syntax = FALSE
  )

  suppressWarnings({
    model <- specifyModelDA(
      group.info = group.info,
      method = "qml",
      m = 16,
      mean.observed = TRUE,
      double = FALSE,
      quad.range = Inf,
      adaptive.quad = FALSE,
      adaptive.frequency = 3,
      missing = "complete",
      orthogonal.x = FALSE,
      orthogonal.y = FALSE,
      auto.fix.first = TRUE,
      auto.fix.single = TRUE,
      fix.composite.var = TRUE,
      cluster = NULL,
      sampling.weights = NULL
    )
  })

  model.filled <- fillModel(model = model, theta = model$theta, method = "qml")
  locations <- subset(
    model$params$gradientStruct$locations,
    group == 1 & block %in% c(2, 9, 11, 12, 13)
  )

  scores.analytical <- analyticalObsGradQmlCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric
  )

  scores.fd <- gradObsLogLikQmlFDCpp(
    submodel = model.filled$models[[1]],
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric,
    eps = 1e-6,
    ncores = 1L
  )

  testthat::expect_true(model.filled$models[[1]]$info$kOmegaEta > 0L)
  testthat::expect_equal(sort(unique(locations$block)),
                         c(2, 9, 11, 12, 13))
  testthat::expect_true(all(is.finite(scores.analytical)))
  testthat::expect_equal(scores.analytical, scores.fd, tolerance = 1e-4)
})

devtools::load_all()
m1 <- '
# Outer Model
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

# Inner model
  Y ~ b1 * X + b2 * Z + b3 * X:Z

  ccoef := (b1 + b2 * b3) / 2
  ccoef2 := b1 * b2

  b1 == b2
'

run <- tryCatch({
    MplusAutomation::detectMplus()
    TRUE
  },
  error = function(e) FALSE
)
if (run) {
  mplus <- modsem(m1, oneInt, method = "mplus", estimator = "MLR")
  standardized_estimates(mplus, type = "modsem")
  print(summary(mplus))
  print(vcov(mplus))
  plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z", vals_z = c(-0.5, 0.5), model = mplus)
}


tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

if (run) {
  mplus_tpb <- modsem(tpb, data = TPB, method = "mplus", rcs = TRUE)
  print(standardized_estimates(mplus_tpb, type = "modsem"))
  print(standardized_estimates(mplus_tpb, type = "stdyx"))
}

tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  SUBJECTIVE_NORMS =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  BEH ~ INT + PBC + INT:PBC:SUBJECTIVE_NORMS + INT:PBC + SUBJECTIVE_NORMS
"

if (run) {
  mplus_tpb_3way <- modsem(tpb, data = TPB[1:250, ], method = "mplus", rcs = TRUE,
                           integration = 8)
  standardized_estimates(mplus_tpb_3way)
}


# Test robust std.errors
mod <- '
# X1-3 are Level 1 variables
X1 =~ x1
X2 =~ x2
X3 =~ x3

# W1-2 are Level 2 variables
W1 =~ w1
W2 =~ w2

fw =~ y1 + y2 + y3
fw ~ X1 + X2 + X3 + W1 + W2
'

if (run) {
  # Standard errors corrected for clustering
  fit.rc <- modsem(mod, lavaan::Demo.twolevel, rcs = TRUE, method = "mplus",
                   cluster = "cluster", estimator = "MLR")
  summary(fit.rc)
}

#' Interaction between latent variables using LMS and QML approaches
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param data A dataframe with observed variables used in the model.
#'
#' @param group Character. A variable name in the data frame defining the groups in a multiple
#'   group analysis
#'
#' @param method method to use:
#' \describe{
#'   \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
#'   \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
#' }
#'
#' @param verbose should estimation progress be shown
#'
#' @param optimize should starting parameters be optimized
#'
#' @param nodes number of quadrature nodes (points of integration) used in \code{lms},
#'   increased number gives better estimates but slower computation. How many are needed depends on the complexity of the model.
#'   For simple models, somewhere between 16-24 nodes should be enough; for more complex models, higher numbers may be needed.
#'   For models where there is an interaction effect between an endogenous and exogenous variable,
#'   the number of nodes should be at least 32, but practically (e.g., ordinal/skewed data), more than 32 is recommended. In cases
#'   where data is non-normal, it might be better to use the \code{qml} approach instead.
#'   You can also consider setting \code{adaptive.quad = TRUE}.
#'
#' @param missing How should missing values be handled? If \code{"listwise"} (default) missing values
#'   are removed list-wise (alias: \code{"complete"} or \code{"casewise"}).
#'   If \code{impute} values are imputed using \code{Amelia::amelia}.
#'   If \code{"fiml"} (alias: \code{"ml"} or \code{"direct"}), full information maximum
#'   likelihood (FIML) is used. FIML can be (very) computationally intensive.
#'
#' @param convergence.abs Absolute convergence criterion.
#'   Lower values give better estimates but slower computation. Not relevant when
#'   using the QML approach. For the LMS approach the EM-algorithm stops whenever
#'   the relative or absolute convergence criterion is reached.
#'
#' @param convergence.rel Relative convergence criterion.
#'   Lower values give better estimates but slower computation.
#'   For the LMS approach the EM-algorithm stops whenever
#'   the relative or absolute convergence criterion is reached.
#'
#' @param optimizer optimizer to use, can be either \code{"nlminb"} or \code{"L-BFGS-B"}. For LMS, \code{"nlminb"} is recommended.
#'   For QML, \code{"L-BFGS-B"} may be faster if there is a large number of iterations, but slower if there are few iterations.
#'
#' @param center.data should data be centered before fitting model
#'
#' @param standardize.data should data be scaled before fitting model, will be overridden by
#'   \code{standardize} if \code{standardize} is set to \code{TRUE}.
#'
#' @param standardize.out should output be standardized (note will alter the relationships of
#'   parameter constraints since parameters are scaled unevenly, even if they
#'   have the same label). This does not alter the estimation of the model, only the
#'   output.
#'
#' \strong{NOTE}: It is recommended that you estimate the model normally and then standardize the output using
#' \code{\link{standardize_model}}, \code{\link{standardized_estimates}} or \code{summary(<modsem_da-object>, standardize=TRUE)}.
#'
#' @param mean.observed should the mean structure of the observed variables be estimated?
#'   This will be overridden by \code{standardize}, if \code{standardize} is set to \code{TRUE}.
#'
#' \strong{NOTE}: Not recommended unless you know what you are doing.
#'
#' @param standardize will standardize the data before fitting the model, remove the mean
#'   structure of the observed variables, and standardize the output. Note that \code{standardize.data},
#'   \code{mean.observed}, and \code{standardize.out} will be overridden by \code{standardize} if \code{standardize} is set to \code{TRUE}.
#'
#' \strong{NOTE}: It is recommended that you estimate the model normally and then standardize the output using
#'   \code{\link{standardized_estimates}}.
#'
#' @param double try to double the number of dimensions of integration used in LMS,
#' this will be extremely slow but should be more similar to \code{mplus}.
#'
#' @param cov.syntax model syntax for implied covariance matrix of exogenous latent variables
#'  (see \code{vignette("interaction_two_etas", "modsem")}).
#'
#' @param calc.se should standard errors be computed? \strong{NOTE}: If \code{FALSE}, the information matrix will not be computed either.
#'
#' @param FIM should the Fisher information matrix be calculated using the observed or expected values? Must be either \code{"observed"} or \code{"expected"}.
#'
#' @param EFIM.S if the expected Fisher information matrix is computed, \code{EFIM.S} selects the number of Monte Carlo samples. Defaults to 100.
#'   \strong{NOTE}: This number should likely be increased for better estimates (e.g., 1000), but it might drasticly increase computation time.
#'
#' @param OFIM.hessian Logical. If \code{TRUE} (default), standard errors are
#'   based on the negative Hessian (observed Fisher information).
#'   If \code{FALSE}, they come from the outer product
#'   of individual score vectors (OPG). For correctly specified models,
#'   these two matrices are asymptotically equivalent; yielding nearly identical
#'   standard errors in large samples. The Hessian usually shows smaller finite-sample
#'   variance (i.e., it's more consistent), and is therefore the default.
#'
#'   Note, that the Hessian is not always positive definite, and is more computationally
#'   expensive to calculate. The OPG should always be positive definite, and a lot
#'   faster to compute. If the model is correctly specified, and the sample size is large,
#'   then the two should yield similar results, and switching to the OPG can save a
#'   lot of time. Note, that the required sample size depends on the complexity of the model.
#'
#'   A large difference between Hessian and OPG suggests misspecification, and
#'   \code{robust.se = TRUE} should be set to obtain sandwich (robust) standard errors.
#'
#' @param EFIM.parametric should data for calculating the expected Fisher information matrix be
#'   simulated parametrically (simulated based on the assumptions and implied parameters
#'   from the model), or non-parametrically (stochastically sampled)? If you believe that
#'   normality assumptions are violated, \code{EFIM.parametric = FALSE} might be the better option.
#'
#' @param R.max Maximum population size (not sample size) used in the calculated of the expected
#'   fischer information matrix.
#'
#' @param robust.se should robust standard errors be computed, using the sandwich estimator?
#'
#' @param max.iter maximum number of iterations.
#'
#' @param max.step maximum steps for the M-step in the EM algorithm (LMS).
#'
#' @param start starting parameters.
#'
#' @param epsilon finite difference for numerical derivatives.
#'
#' @param quad.range range in z-scores to perform numerical integration in LMS using,
#'   when using quasi-adaptive Gaussian-Hermite Quadratures. By default \code{Inf}, such that \code{f(t)} is integrated from -Inf to Inf,
#'   but this will likely be inefficient and pointless at a large number of nodes. Nodes outside
#'   \code{+/- quad.range} will be ignored.
#'
#' @param adaptive.quad should a quasi adaptive quadrature be used? If \code{TRUE}, the quadrature nodes will be adapted to the data.
#'   If \code{FALSE}, the quadrature nodes will be fixed. Default is \code{FALSE}. The adaptive quadrature does not fit an adaptive
#'   quadrature to each participant, but instead tries to place more nodes where posterior distribution is highest. Compared with a
#'   fixed Gauss Hermite quadrature this usually means that less nodes are placed at the tails of the distribution.
#'
#' @param adaptive.frequency How often should the quasi-adaptive quadrature be calculated? Defaults to 3, meaning
#'   that it is recalculated every third EM-iteration.
#'
#' @param adaptive.quad.tol Relative error tolerance for quasi adaptive quadrature. Defaults to \code{1e-12}.
#'
#' @param n.threads number of threads to use for parallel processing. If \code{NULL}, it will use <= 2 threads.
#'   If an integer is specified, it will use that number of threads (e.g., \code{n.threads = 4} will use 4 threads).
#'   If \code{"default"}, it will use the default number of threads (2).
#'   If \code{"max"}, it will use all available threads, \code{"min"} will use 1 thread.
#'
#' @param algorithm algorithm to use for the EM algorithm. Can be either \code{"EM"} or \code{"EMA"}.
#'   \code{"EM"} is the standard EM algorithm. \code{"EMA"} is an
#'   accelerated EM procedure that uses Quasi-Newton and Fisher Scoring
#'   optimization steps when needed. Default is \code{"EM"}.
#'
#' @param em.control a list of control parameters for the EM algorithm. See \code{\link{default_settings_da}} for defaults.
#'
#' @param ordered Variables to be treated as ordered. Ordered indicators are handled
#'   with a Monte-Carlo correction for the LMS/QML estimator on standardized category
#'   scores. The fitted model is used to repeatedly simulate continuous indicators,
#'   ordinalize them to the observed cumulative category proportions, refit the
#'   standardized LMS/QML working model, and solve the resulting fixed-point problem.
#'   This follows the same general Monte-Carlo consistency logic as the MC-OrdPLSc
#'   algorithm described by Slupphaug, Mehmetoglu, and Mittner (2026,
#'   \doi{10.31234/osf.io/fwzj6_v1}).
#'
#'   Threshold standard errors are computed using a simple bootstrap of the category counts
#'   (see \code{ordered.boot.reps}), which does not propagate uncertainty from the
#'   model parameter estimates.
#'
#' @param ordered.mc.reps Integer. Monte-Carlo sample size used in each ordered MC
#'   correction step. Larger values reduce simulation noise but increase runtime.
#'
#' @param ordered.min.iter Integer. Minimum number of Robbins-Monro iterations for the
#'   ordered MC correction.
#'
#' @param ordered.max.iter Integer. Maximum number of Robbins-Monro iterations for the
#'   ordered MC correction.
#'
#' @param ordered.tol Convergence tolerance for the ordered MC Robbins-Monro updates.
#'
#' @param ordered.rng.seed Optional integer random seed used by the ordered MC
#'   correction.
#'
#' @param ordered.fixed.seed Logical. If \code{TRUE} and \code{ordered.rng.seed = NULL},
#'   a fixed seed is drawn once and reused throughout the ordered MC correction to
#'   improve numerical stability.
#'
#' @param ordered.polyak.juditsky Logical. Should Polyak-Juditsky averaging be used in
#'   the ordered MC Robbins-Monro solver?
#'
#' @param ordered.pj.extrapolate Logical. If \code{TRUE}, use extrapolation of the
#'   Polyak-Juditsky path to estimate the convergence point. If \code{FALSE}, the
#'   averaged iterate is used directly.
#'
#' @param ordered.se Character string selecting the ordered MC standard-error correction.
#'   \code{"delta"} (default) uses the delta method for all free parameters.
#'   \code{"penalized"} uses a conservative variance inflation
#'   based on the discrepancy between the naive and MC-corrected standardized estimates.
#'   \code{"naive"} uses the fast diagonal rescaling approximation.
#'   \code{"mixed"} uses the delta method for the structural path coefficients
#'   only, and penalized standard errors for the remaining parameters. This is
#'   considerably faster, and is a good option if you're only interested in
#'   the structural model.
#'
#' @param ordered.se.penalty Non-negative numeric multiplier used when
#'   \code{ordered.se = "penalized"}. The penalty adds
#'   \code{ordered.se.penalty * (theta_mc - theta_naive)^2} to the diagonal of the
#'   naive covariance matrix on the variance scale.
#'
#' @param ordered.delta.reps Integer. Monte-Carlo sample size used when approximating
#'   the ordered MC delta-method Jacobian. Only relevant if
#'   \code{ordered.se} is \code{"delta"} or \code{"mixed"}.
#'
#' @param ordered.delta.epsilon Finite-difference step size used for the ordered MC
#'   delta-method Jacobian. Only relevant if \code{ordered.se} is \code{"delta"}
#'   or \code{"mixed"}.
#'
#' @param ordered.boot.reps Integer. Number of bootstrap replications used to compute
#'   standard errors for the thresholds of ordered indicators. The bootstrap resamples
#'   the category counts at the observed sample size, holding the simulated
#'   (model-implied) reference distribution of the continuous indicators fixed. Note
#'   that this does not account for the uncertainty in the model parameter estimates,
#'   and should be seen as a rough approximation. Set to \code{0} to disable.
#'
#' @param ordered.standardize Logical. Should scored ordered indicators be standardized
#'   before the observed-data fit and after ordinalizing simulated indicators? This is
#'   recommended for numerical stability and is enabled by default.
#'
#' @param cluster Clusters used to compute standard errors robust to non-indepence of observations. Must be paired with
#'   \code{robust.se = TRUE}.
#'
#' @param cr1s Logical; if \code{TRUE}, apply the CR1S small-sample correction factor
#'   to the cluster-robust variance estimator. The CR1S factor is
#'   \eqn{(G / (G - 1)) \cdot ((N - 1) / (N - q))}, where \eqn{G} is the number of
#'   clusters, \eqn{N} is the total number of observations, and \eqn{q} is the number
#'   of free parameters. This adjustment inflates standard errors to reduce the
#'   small-sample downward bias present in the basic cluster-robust (CR0) estimator,
#'   especially when \eqn{G} is small. If \code{FALSE}, the unadjusted CR0 estimator
#'   is used. Defaults to \code{TRUE}. Only relevant if \code{cluster} is specified.
#'
#' @param sampling.weights A variable name in the data frame containing sampling weight information.
#'   Depending on the sampling.weights.normalization argument, these weights may be rescaled (or not)
#'   so that their sum equals the number of observations (total or per group)
#'
#' @param sampling.weights.normalization If \code{"none"}, the sampling weights (if provided) will not be
#'   transformed. If \code{"total"}, the sampling weights are normalized by dividing by the total sum
#'   of the weights, and multiplying again by the total sample size. If \code{"group"}, the sampling
#'   weights are normalized per group: by dividing by the sum of the weights (in each group), and
#'   multiplying again by the group size. The default is \code{"total"}.
#'
#' @param rcs Should latent variable indicators be replaced with reliability-corrected
#'   single item indicators instead? See \code{\link{relcorr_single_item}}.
#'
#' @param rcs.choose Which latent variables should get their indicators replaced with
#'   reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
#'   as the \code{choose} argument.
#'
#' @param rcs.scale.corrected Should reliability-corrected items be scale-corrected? If \code{TRUE}
#'   reliability-corrected single items are corrected for differences in factor loadings between
#'   the items. Default is \code{TRUE}.
#'
#' @param orthogonal.x If \code{TRUE}, all covariances among exogenous latent variables only are set to zero.
#'   Default is \code{FALSE}.
#'
#' @param orthogonal.y If \code{TRUE}, all covariances among endogenous latent variables only are set to zero.
#'   If \code{FALSE} residual covariances are added between pure endogenous variables;
#'   those that are predicted by no other endogenous variable in the structural model.
#'   Default is \code{FALSE}.
#'
#' @param orthogonal.y If \code{TRUE}, all covariances among endogenous latent variables only are set to zero.
#'   If \code{FALSE} residual covariances are added between pure endogenous variables;
#'   those that are predicted by no other endogenous variable in the structural model.
#'   Default is \code{FALSE}.
#'
#' @param auto.fix.first If \code{TRUE} the factor loading of the first indicator, for
#'   a given latent variable is fixed to \code{1}. If \code{FALSE} no loadings are fixed
#'   (automatically). Note that that this might make it such that the model no longer is
#'   identified. Default is \code{TRUE}. \strong{NOTE} this behaviour is overridden
#'   if the first loading is labelled, where it gets treated as a free parameter instead. This
#'   differs from the default behaviour in \code{lavaan}.
#'
#' @param auto.fix.single If \code{TRUE}, the residual variance of
#'   an observed indicator is set to zero if it is the only indicator of a latent variable.
#'   If \code{FALSE} the residual variance is not fixed to zero, and treated as a free parameter
#'   of the model. Default is \code{TRUE}. \strong{NOTE} this behaviour is overridden
#'   if the first loading is labelled, where it gets treated as a free parameter instead.
#'
#' @param fix.composite.var If \code{TRUE} (default) the block covariance structure
#'   of composite indiactors is fixed.
#'
#' @param auto.split.syntax Should the model syntax automatically be split into a
#'   linear and non-linear part? This is done by moving the structural model for
#'   linear endogenous variables (used in interaction terms) into the \code{cov.syntax}
#'   argument. This can potentially allow interactions between two endogenous variables
#'   given that both are linear (i.e., not affected by interaction terms). This is
#'   \code{FALSE} by default for the LMS approach.
#'   When using the QML approach interation effects between exogenous and endogenous
#'   variables can in some cases be biased, if the model is not split beforehand.
#'   The default is therefore \code{TRUE} for the QML approach.
#'
#' @param ... additional arguments to be passed to the estimation function.
#'
#' @return \code{modsem_da} object
#' @export
#'
#' @description
#' \code{modsem_da()} is a function for estimating interaction effects between latent variables
#' in structural equation models (SEMs) using distributional analytic (DA) approaches.
#' Methods for estimating interaction effects in SEMs can basically be split into
#' two frameworks:
#' 1. Product Indicator-based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"},
#' \code{"ca"}, \code{"pind"})
#' 2. Distributionally based approaches (\code{"lms"}, \code{"qml"}).
#'
#' \code{modsem_da()} handles the latter and can estimate models using both QML and LMS,
#' necessary syntax, and variables for the estimation of models with latent product indicators.
#'
#' \strong{NOTE}: Run \code{\link{default_settings_da}} to see default arguments.
#'
#' @references
#' Slupphaug, K., Mehmetoglu, M., and Mittner, M. (2026, March 21).
#' \emph{Consistent Estimates from Biased Estimators: Monte-Carlo Consistent Partial
#' Least Squares for Latent Interaction Models with Ordinal Indicators}. PsyArXiv.
#' \doi{10.31234/osf.io/fwzj6_v1}
#'
#' @examples
#' library(modsem)
#' # For more examples, check README and/or GitHub.
#' # One interaction
#' m1 <- "
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' "
#'
#' \dontrun{
#' # QML Approach
#' est_qml <- modsem_da(m1, oneInt, method = "qml")
#' summary(est_qml)
#'
#' # Theory Of Planned Behavior
#' tpb <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC
#'   BEH ~ INT:PBC
#' "
#'
#' # LMS Approach
#' est_lms <- modsem_da(tpb, data = TPB, method = "lms")
#' summary(est_lms)
#' }
modsem_da <- function(model.syntax = NULL,
                      data = NULL,
                      group = NULL,
                      method = "lms",
                      verbose = NULL,
                      optimize = NULL,
                      nodes = NULL,
                      missing = NULL,
                      convergence.abs = NULL,
                      convergence.rel = NULL,
                      optimizer = NULL,
                      center.data = NULL,
                      standardize.data = NULL,
                      standardize.out = NULL,
                      standardize = NULL,
                      mean.observed = NULL,
                      cov.syntax = NULL,
                      double = NULL,
                      calc.se = NULL,
                      FIM = NULL,
                      EFIM.S = NULL,
                      OFIM.hessian = NULL,
                      EFIM.parametric = NULL,
                      robust.se = NULL,
                      R.max = NULL,
                      max.iter = NULL,
                      max.step = NULL,
                      start = NULL,
                      epsilon = NULL,
                      quad.range = NULL,
                      adaptive.quad = NULL,
                      adaptive.frequency = NULL,
                      adaptive.quad.tol = NULL,
                      n.threads = NULL,
                      algorithm = NULL,
                      em.control = NULL,
                      ordered = NULL,
                      ordered.mc.reps = NULL,
                      ordered.min.iter = 20L,
                      ordered.max.iter = 250L,
                      ordered.tol = 1e-4,
                      ordered.rng.seed = NULL,
                      ordered.fixed.seed = FALSE,
                      ordered.polyak.juditsky = TRUE,
                      ordered.pj.extrapolate = TRUE,
                      ordered.se = c("delta", "penalized", "naive", "mixed"),
                      ordered.se.penalty = 0.25,
                      ordered.delta.reps = NULL,
                      ordered.delta.epsilon = 1e-2,
                      ordered.boot.reps = 1000L,
                      ordered.standardize = TRUE,
                      cluster = NULL,
                      cr1s = FALSE,
                      sampling.weights = NULL,
                      sampling.weights.normalization = NULL,
                      rcs = FALSE,
                      rcs.choose = NULL,
                      rcs.scale.corrected = TRUE,
                      orthogonal.x = NULL,
                      orthogonal.y = NULL,
                      auto.fix.first = NULL,
                      auto.fix.single = NULL,
                      auto.split.syntax = NULL,
                      fix.composite.var = NULL,
                      ...) {
  method <- tolower(method)

  if (is.null(model.syntax)) {
    mod_msg_stop("No model.syntax provided")
  } else if (!is.character(model.syntax)) {
    mod_msg_stop("The provided model syntax is not a string!")
  } else if (length(model.syntax) > 1) {
    mod_msg_stop("The provided model syntax is not of length 1")
  }

  ordered.se <- match.arg(ordered.se)

  if (length(ordered) || any(sapply(data, FUN = is.ordered))) {
    out <- modsemOrderedMCCorrection(
       model.syntax                   = model.syntax,
       data                           = data,
       method                         = method,
       verbose                        = verbose,
       optimize                       = optimize,
       nodes                          = nodes,
       missing                        = missing,
       convergence.abs                = convergence.abs,
       convergence.rel                = convergence.rel,
       optimizer                      = optimizer,
       center.data                    = center.data,
       standardize.data               = standardize.data,
       standardize.out                = standardize.out,
       standardize                    = standardize,
       mean.observed                  = mean.observed,
       cov.syntax                     = cov.syntax,
       double                         = double,
       calc.se                        = calc.se,
       FIM                            = FIM,
       EFIM.S                         = EFIM.S,
       OFIM.hessian                   = OFIM.hessian,
       EFIM.parametric                = EFIM.parametric,
       robust.se                      = robust.se,
       R.max                          = R.max,
       max.iter                       = max.iter,
       max.step                       = max.step,
       start                          = start,
       epsilon                        = epsilon,
       quad.range                     = quad.range,
       adaptive.quad                  = adaptive.quad,
       adaptive.frequency             = adaptive.frequency,
       adaptive.quad.tol              = adaptive.quad.tol,
       n.threads                      = n.threads,
       algorithm                      = algorithm,
       em.control                     = em.control,
       ordered                        = ordered,
       ordered.mc.reps                = ordered.mc.reps,
       ordered.min.iter               = ordered.min.iter,
       ordered.max.iter               = ordered.max.iter,
       ordered.tol                    = ordered.tol,
       ordered.rng.seed               = ordered.rng.seed,
       ordered.fixed.seed             = ordered.fixed.seed,
       ordered.polyak.juditsky        = ordered.polyak.juditsky,
       ordered.pj.extrapolate         = ordered.pj.extrapolate,
       ordered.se                     = ordered.se,
       ordered.se.penalty             = ordered.se.penalty,
       ordered.delta.reps             = ordered.delta.reps,
       ordered.delta.epsilon          = ordered.delta.epsilon,
       ordered.boot.reps              = ordered.boot.reps,
       ordered.standardize            = ordered.standardize,
       cluster                        = cluster,
       group                          = group,
       cr1s                           = cr1s,
       sampling.weights               = sampling.weights,
       sampling.weights.normalization = sampling.weights.normalization,
       rcs                            = rcs,
       rcs.choose                     = rcs.choose,
       rcs.scale.corrected            = rcs.scale.corrected,
       orthogonal.x                   = orthogonal.x,
       orthogonal.y                   = orthogonal.y,
       auto.fix.first                 = auto.fix.first,
       auto.fix.single                = auto.fix.single,
       auto.split.syntax              = auto.split.syntax,
       fix.composite.var              = fix.composite.var,
       ...)

    return(out)
  }

  if (is.null(data)) {
    mod_msg_stop("No data provided")
  } else {
    data <- as.data.frame(data)
  }

  if (rcs) { # use reliability-correct single items?
    corrected <- relcorr_single_item(
      syntax          = model.syntax,
      data            = data,
      group           = group,
      choose          = rcs.choose,
      scale.corrected = rcs.scale.corrected,
      warn.lav        = FALSE
    )

    model.syntax <- corrected$syntax
    data         <- corrected$data
  }

  if ("convergence" %in% names(list(...))) {
    convergence.rel <- list(...)$convergence
    mod_msg_warn("Argument 'convergence' is deprecated, use 'convergence.rel' instead.")
  }

  args <-
    getMethodSettingsDA(method,
      args =
        list(
          verbose                        = verbose,
          optimize                       = optimize,
          nodes                          = nodes,
          convergence.abs                = convergence.abs,
          convergence.rel                = convergence.rel,
          optimizer                      = optimizer,
          center.data                    = center.data,
          standardize.data               = standardize.data,
          standardize.out                = standardize.out,
          standardize                    = standardize,
          mean.observed                  = mean.observed,
          double                         = double,
          calc.se                        = calc.se,
          FIM                            = FIM,
          EFIM.S                         = EFIM.S,
          OFIM.hessian                   = OFIM.hessian,
          EFIM.parametric                = EFIM.parametric,
          robust.se                      = robust.se,
          R.max                          = R.max,
          max.iter                       = max.iter,
          max.step                       = max.step,
          epsilon                        = epsilon,
          quad.range                     = quad.range,
          adaptive.quad                  = adaptive.quad,
          adaptive.frequency             = adaptive.frequency,
          adaptive.quad.tol              = adaptive.quad.tol,
          n.threads                      = n.threads,
          algorithm                      = algorithm,
          em.control                     = em.control,
          missing                        = missing,
          orthogonal.x                   = orthogonal.x,
          orthogonal.y                   = orthogonal.y,
          auto.fix.first                 = auto.fix.first,
          auto.fix.single                = auto.fix.single,
          auto.split.syntax              = auto.split.syntax,
          cr1s                           = cr1s,
          group                          = group,
          sampling.weights               = sampling.weights,
          sampling.weights.normalization = sampling.weights.normalization,
          fix.composite.var              = fix.composite.var
        )
    )

  cont.cols <- setdiff(colnames(data), c(cluster, group, sampling.weights))

  if (args$center.data)
    data[cont.cols] <- lapply(data[cont.cols], FUN = centerIfNumeric, scaleFactor = FALSE)

  if (args$standardize.data)
    data[cont.cols] <- lapply(data[cont.cols], FUN = scaleIfNumeric, scaleFactor = FALSE)

  group.info <- parseModelArgumentsByGroupDA(
    model.syntax       = model.syntax,
    cov.syntax         = cov.syntax,
    method             = method,
    data               = data,
    group              = group,
    auto.split.syntax  = args$auto.split.syntax,
    sampling.weights   = sampling.weights,
    sampling.weights.normalization = args$sampling.weights.normalization
  )

  mod_stopif(!method %in% c("lms", "qml"), "Method must be either 'lms' or 'qml'")

  model <- specifyModelDA(
    group.info         = group.info,
    method             = method,
    m                  = args$nodes,
    mean.observed      = args$mean.observed,
    double             = args$double,
    quad.range         = args$quad.range,
    adaptive.quad      = args$adaptive.quad,
    adaptive.frequency = args$adaptive.frequency,
    missing            = args$missing,
    orthogonal.x       = args$orthogonal.x,
    orthogonal.y       = args$orthogonal.y,
    auto.fix.first     = args$auto.fix.first,
    auto.fix.single    = args$auto.fix.single,
    fix.composite.var  = args$fix.composite.var,
    cluster            = cluster,
    sampling.weights   = sampling.weights
  )

  if (args$optimize) {
    model <- tryCatch({
      .optimize <- purrr::quietly(optimizeStartingParamsDA)
      #.optimize <- \(...) list(result = optimizeStartingParamsDA(...))

      ops <- c(group.info$parTable$op, group.info$parTableCov$op)
      engine <- if (any(ops %in% c("<", ">", "=="))) "pi" else "sam"

      result <- .optimize(
        model            = model,
        args             = args,
        group            = group,
        sampling.weights = sampling.weights,
        engine           = engine
      )

      warnings  <- result$warnings

      if (length(warnings)) {
        fwarnings <- paste0(
          paste0(seq_along(warnings), ". ", warnings),
          collapse = "\n"
        )

        mod_msg_warn_immediate(
          paste0("warning when optimizing starting parameters:\n", fwarnings)
        )
      }

      result$result

    }, error = function(e) {
      mod_msg_warn_immediate(
        paste0("unable to optimize starting parameters:\n", e)
      )

      if (is.null(max.step) && args$max.step <= 1 && method == "lms") {
        # When we don't have optimized starting parameters with LMS, the algorithm
        # will likely not converge with the default max.step
        mod_msg_note("Increasing max step in EM-algorithm to 50")
        args$max.step <<- 50
      }

      model
    })
  }

  if (!is.null(start)) {
    checkStartingParams(start, model = model) # throws error if somethings wrong
    model$theta <- start
  }

  # We want to limit the number of threads available to OpenBLAS.
  # Depending on the OpenBLAS version, it might not be compatible with
  # OpenMP. If `n.blas > 1L` you might end up getting this message:
  #> OpenBLAS Warning : Detect OpenMP Loop and this application may hang.
  #>                    Please rebuild the library with USE_OPENMP=1 option.
  # We don't want to restrict OpenBLAS in any other setttings in other settings,
  # e.g., lavaan::sem, so we reset after the model has been estimated.
  setThreads(n = args$n.threads, n.blas = 1L)
  on.exit(resetThreads()) # clean up at end of function

  est <- tryCatch(switch(method,
    qml = estQml(model,
      verbose         = args$verbose,
      convergence     = args$convergence.rel,
      calc.se         = args$calc.se,
      FIM             = args$FIM,
      EFIM.S          = args$EFIM.S,
      OFIM.hessian    = args$OFIM.hessian,
      EFIM.parametric = args$EFIM.parametric,
      robust.se       = args$robust.se,
      max.iter        = args$max.iter,
      epsilon         = args$epsilon,
      optimizer       = args$optimizer,
      R.max           = args$R.max,
      cr1s            = args$cr1s,
      ...
    ),
    lms = emLms(model,
      verbose           = args$verbose,
      convergence.abs   = args$convergence.abs,
      convergence.rel   = args$convergence.rel,
      calc.se           = args$calc.se,
      FIM               = args$FIM,
      EFIM.S            = args$EFIM.S,
      OFIM.hessian      = args$OFIM.hessian,
      EFIM.parametric   = args$EFIM.parametric,
      robust.se         = args$robust.se,
      max.iter          = args$max.iter,
      max.step          = args$max.step,
      epsilon           = args$epsilon,
      optimizer         = args$optimizer,
      R.max             = args$R.max,
      em.control        = args$em.control,
      algorithm         = args$algorithm,
      adaptive.quad     = args$adaptive.quad,
      quad.range        = args$quad.range,
      adaptive.quad.tol = args$adaptive.quad.tol,
      nodes             = args$nodes,
      cr1s              = args$cr1s,
      ...
  )),
  error = function(e) {
    if (args$verbose) cat("\n")
    message <- paste0("modsem [%s]: Model estimation failed!\n",
                      "Message: %s")
    mod_msg_stop(sprintf(message, method, e$message))
  })

  # Finalize the model object
  # Expected means and covariances
  expected.matrices <- tryCatch(
    calcExpectedMatricesDA(
      parTable = est$parTable,
      xis  = getXisModelDA(model$models[[1L]]), # taking both the main model and cov model into account
      etas = getEtasModelDA(model$models[[1L]])  # taking both the main model and cov model into account
    ),
    error = function(e) {
      mod_msg_warn(paste0("Failed to calculate expected matrices: ", e$message))
      NULL
    }
  )

  if (!is.null(expected.matrices)) {
    est$expected.matrices <- expected.matrices

    for (g in seq_along(est$model$models)) # attach to submodels
      est$model$models[[g]]$expected.matrices <- expected.matrices[[g]]
  }


  # Arguments
  est$args <- args
  attr(est, "isRCS_Model") <- rcs
  class(est) <- c("modsem_da", "modsem")

  # Check the results
  postCheckModel(est)

  # Return
  if (args$standardize.out) standardize_model(est) else est
}

# Copied (almost) directly from lavaan/R/lav_msg.R
modsem_cache_env <- new.env(parent = emptyenv())

# Displays a message (... concatenated with spaces in between) with header
# 'modsem(function):', except when showheader == FALSE, and formatted to have
# a maximum line length of 'txt_width' while all but the first line start with
# 'indent' spaces. If a footer is specified it is appended to the formatted text
# 'as is'. The message is shown via R function 'message()'.
mod_msg_note <- function(..., showheader = TRUE, footer = "") {
  wat <- unlist(list(...), use.names = FALSE)
  if (!showheader) wat <- c("modsem NOTE: ___", wat)
  msg <- mod_msg(wat, showheader = showheader)
  if (footer != "") msg <- paste(msg, footer, sep = "\n")
  message(msg, domain = NA)
}

# Displays a message with header and formatted as
# above via R function 'warning()'.
mod_msg_warn <- function(..., footer = "", immediate. = FALSE) {
  wat <- unlist(list(...), use.names = FALSE)
  msg <- mod_msg(wat)
  if (footer != "") msg <- paste(msg, footer, sep = "\n")
  warning(msg, call. = FALSE, domain = NA, immediate. = immediate.)
}


# Warn immediately
mod_msg_warn_immediate <- function(..., footer = "") {
  mod_msg_warn(..., footer = "", immediate. = TRUE)
}

# Displays a message with header and formatted as
# above via R function 'stop()'.
mod_msg_stop <- function(..., footer = "", only.format.header = FALSE) {
  wat <- unlist(list(...), use.names = FALSE)
  msg <- mod_msg(wat)

  if (only.format.header) {
    # trust the formatting in ... already, only paste the header on top
    header <- attr(msg, "header")
    msg <- paste0(header, "\n", ...)
  }

  if (footer != "") msg <- paste(msg, footer, sep = "\n")
  stop(msg, call. = FALSE, domain = NA)
}

# Displays a message with header and formatted as
# above via R function 'stop()', where the message is prepended with "FIXME:",
# to indicate an internal error, e.g. an error condition which was supposed
# to be handled in the calling functions. Such error message do not have to
# be created by [n]gettext[f] because they don't have to be translated!!!
mod_msg_fixme <- function(...) {
  wat <- c("FIXME: ", unlist(list(...), use.names = FALSE))
  stop(mod_msg(wat), call. = FALSE, domain = NA)
}

# subroutine for above functions
mod_msg <- function(wat, txt_width = getOption("width", 80L),
                    indent = 3L, showheader = TRUE) {
  if (showheader) {
    ignore_in_stack <- c(
      "^eval$", "^try", "^doTryCatch", "^mod_msg", "^stop$", "^warning$",
      "^which$", "^unique$", "^as\\.", "^unlist$", "^message$",
      "^source$", "^withVisible$", "^tryCatch.W.E$", "^withCallingHandlers$",
      "^do.call$", "^paste", "^mod_stopif", "^mod_warnif", "^warn", "^stop",
      "^message", "^[0-9]+$" # anonymous functions?
    )
    sc <- sys.calls()
    sc_i <- length(sc)
    sc_naam <- ""
    while (sc_i > 0L) {
      x <- tryCatch(
        as.character(sc[[sc_i]][[1L]]),
        error = function(e) "unknown"
      )
      if (length(x) == 3L) {
        # needed if a function specified in namespace, e.g.
        # as.character(str2lang("modsem::sem(m, d)")[[1L]])
        x <- x[[3L]]
      }
      skip <- FALSE
      for (re in ignore_in_stack) {
        if (grepl(re, x)) {
          skip <- TRUE
          break
        }
      }
      if (!skip) {
        sc_naam <- x
        break
      }
      sc_i <- sc_i - 1L
    }
    if (sc_naam == "") {
      header <- "modsem: ___"
    } else {
      header <- paste0("modsem->", sc_naam, "(): ___")
    }
  } else {
    header <- ""
  }
  txt_width <- txt_width - indent # whitespace at the right
  # make sure we only have a single string
  txt <- paste(wat, collapse = " ")
  # split the txt in little chunks
  chunks <- strsplit(paste(header, txt), "\\s+", fixed = FALSE)[[1]]

  # chunk size (number of characters)
  chunk_size <- nchar(chunks)

  # remove empty chunk in position 1 (if txt starts with whitespace)
  if (chunk_size[1L] == 0L) {
    chunks <- chunks[-1L]
    chunk_size <- chunk_size[-1]
  }

  nstart <- 1L
  nstop <- 1L
  corr_line1 <- 7L # first line possibly contains "error: "
  while (nstart <= length(chunks)) {
    while (nstop < length(chunks) &&
           sum(chunk_size[seq.int(nstart, nstop + 1L)]) + corr_line1 +
           nstop - nstart + indent < txt_width && chunks[nstop + 1L] != "___") {
      nstop <- nstop + 1
    }
    corr_line1 <- 0L
    if (nstop < length(chunks) && chunks[nstop + 1L] == "___") {
      # forced line break
      chunks[nstop + 1L] <-  ""
      nstop <- nstop + 1L
    }
    if (nstop < length(chunks)) {
      chunks[nstop + 1L] <- paste0(
        "\n", strrep(" ", indent),
        chunks[nstop + 1L]
      )
    }
    nstart <- nstop + 1L
    nstop <- nstart
  }

  out <- paste(chunks, collapse = " ")
  attr(out, "header") <- stringr::str_replace_all(header, ": ___", ":")

  out
}


mod_stopif <- function(cond, ...) {
  if (length(cond) > 0 && !is.na(cond[[1L]]) && cond[[1L]])
    mod_msg_stop(...)
}


mod_warnif <- function(cond, ..., .newline = FALSE) {
  if (length(cond) > 0 && !is.na(cond[[1L]]) && cond[[1L]]) {
    if (.newline) cat("\n")
    mod_msg_warn(...)
  }
}


mod_warnif_immediate <- function(cond, ..., .newline = FALSE) {
  if (length(cond) > 0 && !is.na(cond[[1L]]) && cond[[1L]]) {
    if (.newline) cat("\n")
    mod_msg_warn_immediate(...)
  }
}


# Transforms a value to a character representation for use in messages
# if logsep = "array" (default), letters[1:3] is transformed to ("a", "b", "c")
# if logsep = "none", c("x", "y", "z") is transformed to "x", "y", "z"
# if logsep = "and", 1:3 is transformed to 1, 2 and 3
# if logsep = "or", c("a", "b", "c") is transformed to "a", "b" or "c"
# The type of quote can be modified via parameter qd (default = TRUE).
# If the object has names, the names will be prepended with a colon before the
# value, e.g. c(x = 2.3, y = 4) --> (x : 2.3, y : 4).
mod_msg_view <- function(x,
                         log_sep = c("array", "none", "and", "or"),
                         qd = TRUE) {
  if (missing(x)) {
    return("NULL")
  }
  log_sep <- match.arg(log_sep)
  xn <- names(x)
  if (is.list(x)) {
    xx <- sapply(x, mod_msg_view)
  } else {
    if (is.character(x)) {
      if (qd) {
        xx <- dQuote(x, q = FALSE)
      } else {
        xx <- sQuote(x, q = FALSE)
      }
    } else {
      xx <- as.character(x)
    }
    xx[is.na(x)] <- "NA"
  }
  if (!is.null(xn)) xx <- paste(xn, ":", xx)
  if (length(xx) == 1) {
    rv <- xx
  } else {
    if (log_sep == "array") rv <- paste0("(", paste(xx, collapse = ", "), ")")
    if (log_sep == "none") rv <- paste(xx, collapse = ", ")
    if (log_sep == "and") rv <- paste(paste(xx[-length(xx)], collapse = ", "),
     gettext("and"), xx[length(xx)])
    if (log_sep == "or") rv <- paste(paste(xx[-length(xx)], collapse = ", "),
     gettext("or"), xx[length(xx)])
  }
  rv
}
# Warning for deprecated functions
# Like base::.Deprecated but specialised for modsem
# parameter times specifies how many times the warning should be generated
# during one "modsem-package-session"
mod_deprecated <- function(new,
                           old = as.character(sys.call(sys.parent()))[1L],
                           times = 1L) {
  dprmsg <-  get0(paste0("dpr_", old), modsem_cache_env,
                  ifnotfound = as.integer(times))
  if (dprmsg <= 0L) return(invisible(NULL))
  assign(paste0("dpr_", old), dprmsg - 1L, modsem_cache_env)
  msg <- c(gettextf("'%s' is deprecated.\n", old),
         gettextf("Use '%s' instead.\n", new),
         gettext("See help(\"Deprecated\")"))
  msg <- paste(msg, collapse = "")
  warning(warningCondition(msg, old = old, new = new, package = NULL,
                           class = "deprecatedWarning"))
}

# Warning for deprecated arguments-parameter for another parameter in a function
# method_par specifies the name of the parameter used for a 'method'
# arg_par specifies the name of the deprecated parameter
# parameter times specifies how many times the warning should be generated
# during one "modsem-package-session"
mod_deprecated_args <- function(method_par, arg_par, times = 1L) {
  dprmsg <-  get0(paste0("dpr_", method_par, arg_par), modsem_cache_env,
                  ifnotfound = as.integer(times))
  if (dprmsg <= 0L) return(invisible(NULL))
  assign(paste0("dpr_", method_par, arg_par), dprmsg - 1L, modsem_cache_env)
  msg <- c(gettextf("Argument '%s' is deprecated.\n", arg_par),
      gettextf("The arguments for '%s' can now be provided in '%s' itself.\n",
      method_par, method_par))
  msg <- paste(msg, collapse = "")
  warning(warningCondition(msg, old = arg_par, new = method_par, package = NULL,
                           class = "deprecatedWarning"))
}

#  ---------------  examples of use ----------------------
# # warning if argument x is missing
#   mod_msg_warn(gettextf(
#     "argument %1$s is missing, using %2$s.",
#     x, mod_msg_view(usedvalue)
#   ))
#
# # warning if length of an argument x is greater then 1 and cannot be
#   mod_msg_warn(gettextf("%1$s argument should be a single character string.
#   Only the first one (%2$s) is used.", xname, x[[1]]))
#
# # error if argument is unknown (show value)
#   mod_msg_stop(gettextf(
#     "%1$s argument unknown: %2$s",
#     xname, mod_msg_view(xvalue)
#   ))
#
# # error if argument isn't one of the allowed values, show values allowed
#   if (length(allowed) == 2L) {
#     mod_msg_stop(gettextf(
#       "%1$s argument must be either %2$s",
#       x, mod_msg_view(allowed, "or")
#     ))
#   } else {
#     mod_msg_stop(gettextf(
#       "%1$s argument must be one of %2$s",
#       x, mod_msg_view(allowed, "or")
#     ))
#   }
#
# # error if argument isn't one of the allowed values (show invalid ones)
#   mod_msg_stop(sprintf(
#     ngettext(
#       length(invalids),
#       "invalid value in %1$s argument: %2$s.",
#       "invalid values in %1$s argument: %2$s."
#     ),
#     x, mod_msg_view(invalids, log_sep = "none")
#   ))

devtools::load_all()

m <- '
y1 ~ x1 + z1 + x1:z1
'

testthat::expect_no_error({
  fit_qml <- modsem(m, oneInt, method = "qml")
  fit_lms <- modsem(m, oneInt, method = "lms")
  pred <- modsem_predict(fit_lms)

  summary(fit_lms, standardized = TRUE)
  estimate_h0(fit_lms)
})

tpb <- '
int1 ~ att1 + sn1 + pbc1
b1 ~ int1 + pbc1 + int1:pbc1
'

testthat::expect_no_error({
  fit_qml <- modsem(tpb, TPB, method = "qml")
  fit_lms <- modsem(tpb, TPB, method = "lms")
  summary(fit_lms, standardized = TRUE)
})

m2 <- '
Z =~ z1 + z2
y1 ~ x1 + Z + x1:Z
'

testthat::expect_no_error({
  # test selection of integrating variable
  fit_lms <- modsem(m2, oneInt, method = "lms")
  head(modsem_predict(fit_lms))
})

OptimizerInfoQML <- rlang::env(eval = 0, logLiks = 0)


qmlNcores <- function() {
  ncores <- ThreadEnv$n.threads
  if (length(ncores) != 1L || is.null(ncores) || is.na(ncores)) 1L else ncores
}


resetOptimizerInfoQML <- function() {
  OptimizerInfoQML$eval    <- 0
  OptimizerInfoQML$logLiks <- -Inf
}


incrementIterations <- function(logLik) {
  eval       <- OptimizerInfoQML$eval + 1
  logLiks    <- c(OptimizerInfoQML$logLiks, logLik)
  diffLL     <- getDiffTwoMax(logLiks)
  deltaLL    <- diffLL$abs
  relDeltaLL <- diffLL$rel

  clearConsoleLine() # clear before printing
  msg <- MSG_STRINGS$strings$incrementIterations0
  printf(msg, eval, logLik, abs(deltaLL), abs(relDeltaLL))

  OptimizerInfoQML$eval    <- eval
  OptimizerInfoQML$logLiks <- logLiks
}


logLikQml <- function(theta, model, sum = TRUE, sign = -1, verbose = FALSE) {
  modelFilled <- fillModel(model, theta, method = "qml")

  ll <- 0
  for (g in seq_len(model$info$n.groups))
    ll <- ll + logLikQmlGroup(modelFilled$models[[g]], sum = sum, sign = sign)

  if (verbose)
    incrementIterations(sign * ll)

  ll
}


logLikQmlGroup <- function(submodel, sum = TRUE, sign = -1) {
  ncores <- qmlNcores()

  if (sum) {
    ll <- logLikQmlCpp(submodel, ncores = ncores)
    return(sign * ll)
  }

  # Per-observation path — used by robust SE / sandwich estimator only
  numXi       <- submodel$info$numXis
  numEta      <- submodel$info$numEtas
  kOmegaEta   <- submodel$info$kOmegaEta
  latentEtas  <- submodel$info$latentEtas

  m <- submodel$matrices
  m$numEta    <- numEta
  m$numXi     <- numXi
  m$kOmegaEta <- kOmegaEta

  m$tauX <- m$tauX + m$lambdaX %*% m$beta0
  m$x <- submodel$data$data.full[, submodel$info$allIndsXis, drop = FALSE]
  m$y <- submodel$data$data.full[, submodel$info$allIndsEtas, drop = FALSE]
  m$x <- centerIndicators(m$x, tau = m$tauX)
  m$y <- centerIndicators(m$y, tau = m$tauY)

  # Fill in residual variances for latent etas with only a single indicator
  # only needed if the resiudal variances are non-zero
  if (any(m$selectThetaEpsilon2)) {
    m$subThetaEpsilon2[is.na(m$subThetaEpsilon2)] <-
      m$thetaEpsilon[m$selectThetaEpsilon2]

    selectRows <- apply(m$selectSubSigma2ThetaEpsilon, MARGIN = 1, FUN = \(x) all(!x))
    selectCols <- apply(m$selectSubSigma2ThetaEpsilon, MARGIN = 2, FUN = \(x) all(!x))
    m$fullSigma2ThetaEpsilon[selectRows, selectCols] <- m$subThetaEpsilon2
  }

  t <- NROW(m$x)
  if (!is.null(m$emptyR)) {
    m$R <- m$emptyR
    m$R[is.na(m$R)] <- -m$lambdaY[!m$selectScalingY] # fill R with -Beta
    m$fullR[m$rowsR, m$colsR] <- m$R
    m$u <- m$y %*% t(m$fullR)
    m$fullU[, m$colsU] <- m$u
    m$Beta <- m$lambdaY[m$selectBetaRows, latentEtas]

    m$subThetaEpsilon1[is.na(m$subThetaEpsilon1)] <-
      m$thetaEpsilon[m$selectThetaEpsilon1]

    m$RER <- m$R %*% m$thetaEpsilon[m$colsR, m$colsR] %*% t(m$R)
    invRER <- solve(m$RER)
    m$L2 <- -m$subThetaEpsilon1 %*% t(m$Beta) %*% invRER
    m$fullL2[m$selectSubL2] <- m$L2

    m$Sigma2ThetaEpsilon <- m$subThetaEpsilon1 - m$subThetaEpsilon1 ^ 2 %*%
      t(m$Beta) %*% invRER %*% m$Beta

    m$fullSigma2ThetaEpsilon[m$selectSubSigma2ThetaEpsilon] <-
      m$Sigma2ThetaEpsilon
  }

  m$Sigma2ThetaEpsilon <- m$fullSigma2ThetaEpsilon
  m$L2     <- m$fullL2
  m$u      <- m$fullU
  m$LXPLX  <- m$lambdaX %*% m$phi %*% t(m$lambdaX) + m$thetaDelta
  invLXPLX <- solve(m$LXPLX)
  m$L1     <- m$phi %*% t(m$lambdaX) %*% invLXPLX
  m$Sigma1 <- m$phi - m$phi %*% t(m$lambdaX) %*% invLXPLX %*% m$lambdaX %*% m$phi
  m$kronXi <- calcKronXi(m, t, ncores = ncores)
  m$Binv   <- calcBinvCpp(m, t, ncores = ncores)

  Ey            <- muQmlCpp(m, t, ncores = ncores)
  sigmaEpsilon  <- sigmaQmlCpp(m, t, ncores = ncores)
  sigmaXU       <- calcSigmaXU(m)

  normalInds    <- colnames(sigmaXU)
  indsY         <- colnames(m$y)
  nonNormalInds <- indsY[!indsY %in% normalInds]

  # probability densities per observation
  f2i <- probf2(matrices = m, normalInds = normalInds, sigma = sigmaXU)
  f3i <- probf3(matrices = m, nonNormalInds = nonNormalInds, expected = Ey,
               sigma = sigmaEpsilon, t = t, numEta = numEta)
  lli <- f2i + f3i

  # Correct by sampling weights (if available)
  if (!is.null(submodel$data$weights))
    lli <- submodel$data$weights * lli

  sign * lli
}


calcSigmaXU <- function(matrices) {
  if (is.null(matrices$emptyR)) return(matrices$LXPLX)
  diagBindSquareMatrices(matrices$LXPLX, matrices$RER)
}


probf2 <- function(matrices, normalInds, sigma) {
  mu <- rep(0, ncol(sigma))
  X  <- cbind(matrices$x, matrices$u)[ , normalInds]

  dmvn(X, mean = mu, sigma = sigma, log = TRUE)
}


probf3 <- function(matrices, nonNormalInds, expected, sigma, t, numEta) {
  ncores <- qmlNcores()
  if (numEta == 1)
    p <- dnormCpp(matrices$y[, 1], mu = expected, sigma = sqrt(sigma),
                  ncores = ncores)
  else
    p <- repDmvnormCpp(matrices$y[, nonNormalInds], expected = expected,
                       sigma = sigma, t = t, ncores = ncores)
  p
}


centerIndicators <- function(X, tau) {
  for (i in seq_len(ncol(X))) X[, i] <- X[, i] - tau[[i]]
  X
}


refreshQmlGradientJacobian <- function(theta, model, Jacobian) {
  if (isTRUE(model$params$gradientStruct$hasCovModel))
    Jacobian <- refreshCovModelJacobian(theta, model, Jacobian,
                                        method = "qml")$J

  nlinDerivs <- model$params$gradientStruct$nlinDerivs
  if (length(nlinDerivs)) {
    evalTheta  <- model$params$gradientStruct$evalTheta
    param.full <- stringr::str_split_i(colnames(Jacobian), pattern = "#", i = 1L)
    param.part <- rownames(Jacobian)
    THETA      <- list2env(as.list(evalTheta(theta)))

    for (dep in names(nlinDerivs)) {
      derivs <- nlinDerivs[[dep]]
      for (indep in names(derivs)) {
        deriv <- eval(expr = derivs[[indep]], envir = THETA)
        Jacobian[param.part == indep, param.full == dep] <- deriv
      }
    }
  }

  Jacobian
}


qmlCanUseGradientStruct <- function(model) {
  gradientStruct <- model$params$gradientStruct

  !isTRUE(gradientStruct$useFDGradient) &&
    !is.null(gradientStruct$locations) &&
    !is.null(gradientStruct$Jacobian)
}


simpleGradientLogLikQml <- function(theta, model, sign = -1, epsilon = 1e-6) {
  modelFilled <- fillModel(model = model, theta = theta, method = "qml")
  locations   <- model$params$gradientStruct$locations
  Jacobian    <- model$params$gradientStruct$Jacobian
  Jacobian    <- refreshQmlGradientJacobian(theta, model, Jacobian)
  ncores      <- qmlNcores()

  grad <- matrix(0, nrow = NROW(locations), ncol = 1L,
                 dimnames = list(locations$param, NULL))

  for (g in seq_len(modelFilled$info$n.groups)) {
    locations.g <- locations[locations$group == g, , drop = FALSE]
    if (!NROW(locations.g)) next

    grad.g <- analyticalGradQmlCpp(
      submodel  = modelFilled$models[[g]],
      block     = locations.g$block,
      row       = locations.g$row,
      col       = locations.g$col,
      symmetric = locations.g$symmetric
    )

    bad <- !is.finite(grad.g)

    if (any(bad)) {
      grad.fd <- gradLogLikQmlFDCpp(
        submodel  = modelFilled$models[[g]],
        block     = locations.g$block[bad],
        row       = locations.g$row[bad],
        col       = locations.g$col[bad],
        symmetric = locations.g$symmetric[bad],
        eps       = epsilon,
        ncores    = ncores
      )

      grad.g[bad] <- grad.fd
    }

    grad[locations.g$param, ] <- grad.g
  }

  as.vector(sign * Jacobian %*% grad)
}


simpleObsGradientLogLikQml <- function(theta, model, sign = -1, epsilon = 1e-6) {
  modelFilled <- fillModel(model = model, theta = theta, method = "qml")
  locations   <- model$params$gradientStruct$locations
  Jacobian    <- model$params$gradientStruct$Jacobian
  Jacobian    <- refreshQmlGradientJacobian(theta, model, Jacobian)
  ncores      <- qmlNcores()

  N <- vapply(modelFilled$models, FUN.VALUE = integer(1L),
              FUN = \(sub) NROW(sub$data$data.full))
  N.start <- c(1L, cumsum(N)[-length(N)] + 1L)
  N.end   <- cumsum(N)

  scores <- matrix(0, nrow = sum(N), ncol = length(theta),
                   dimnames = list(NULL, names(theta)))

  for (g in seq_len(modelFilled$info$n.groups)) {
    locations.g <- locations[locations$group == g, , drop = FALSE]
    if (!NROW(locations.g)) next

    scores.g <- analyticalObsGradQmlCpp(
      submodel  = modelFilled$models[[g]],
      block     = locations.g$block,
      row       = locations.g$row,
      col       = locations.g$col,
      symmetric = locations.g$symmetric
    )

    bad <- colSums(!is.finite(scores.g)) > 0L

    if (any(bad)) {
      scores.fd <- gradObsLogLikQmlFDCpp(
        submodel  = modelFilled$models[[g]],
        block     = locations.g$block[bad],
        row       = locations.g$row[bad],
        col       = locations.g$col[bad],
        symmetric = locations.g$symmetric[bad],
        eps       = epsilon,
        ncores    = ncores
      )

      scores.g[, bad] <- scores.fd
    }

    J.g <- Jacobian[, locations.g$param, drop = FALSE]
    rows.g <- seq(N.start[[g]], N.end[[g]], by = 1L)
    scores[rows.g, ] <- sign * scores.g %*% t(J.g)
  }

  scores
}


gradientLogLikQml <- function(theta, model, epsilon = 1e-8, sign = -1,
                               .f = logLikQmlGroup, sum = TRUE, ...) {
  if (sum && qmlCanUseGradientStruct(model))
    return(simpleGradientLogLikQml(theta = theta, model = model,
                                    sign = sign, epsilon = epsilon))

  if (!sum && identical(.f, logLikQmlGroup) && qmlCanUseGradientStruct(model))
    return(simpleObsGradientLogLikQml(theta = theta, model = model,
                                      sign = sign, epsilon = epsilon))

  # Fallback: per-obs path — original R-level FD
  params <- model$params

  SELECT_THETA_LAB  <- params$SELECT_THETA_LAB
  SELECT_THETA_COV  <- params$SELECT_THETA_COV
  SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN

  N <- vapply(model$models, FUN.VALUE = numeric(1L), FUN = \(sub) sub$data$n)
  N.start <- c(1, cumsum(N)[-length(N)] + 1L)
  N.end   <- cumsum(N)

  n <- if (sum) 1L else sum(N)
  k <- length(theta)

  grad <- matrix(0, nrow = n, ncol = k, dimnames = list(NULL, names(theta)))

  .fg <- function(theta, g) {
    modFilled <- fillModel(theta = theta, model = model, method = "qml")
    .f(submodel = modFilled$models[[g]], sign = sign, sum = sum, ...)
  }

  for (g in seq_len(model$info$n.groups)) {
    f0 <- .fg(theta = theta, g = g)
    J  <- if (sum) 1L else seq(N.start[[g]], N.end[[g]], by = 1L)

    indices <- c(SELECT_THETA_LAB[[g]], SELECT_THETA_COV[[g]], SELECT_THETA_MAIN[[g]])

    for (i in indices) {
      theta_i    <- theta
      theta_i[i] <- theta_i[i] + epsilon
      fi         <- .fg(theta_i, g = g)
      grad[J, i] <- grad[J, i] + (fi - f0) / epsilon
    }
  }

  if (n == 1L) as.vector(grad) else grad
}


# Gradient function for per-observation QML scores.
gradientLogLikQml_i <- function(theta, model, sign = -1, epsilon = 1e-8) {
  gradientLogLikQml(theta = theta, model = model, sign = sign, epsilon = epsilon, sum = FALSE)
}


simpleHessianLogLikQml <- function(theta, model, sign = -1,
                                    .relStep = .Machine$double.eps ^ (1/5)) {
  modelFilled <- fillModel(model = model, theta = theta, method = "qml")
  locations   <- model$params$gradientStruct$locations
  Jacobian    <- model$params$gradientStruct$Jacobian
  Jacobian2   <- model$params$gradientStruct$Jacobian2
  jac         <- refreshCovModelJacobian(theta, model, Jacobian, Jacobian2,
                                         method = "qml")
  Jacobian    <- jac$J
  Jacobian2   <- jac$J2
  nlinDerivs  <- model$params$gradientStruct$nlinDerivs
  nlinDerivs2 <- model$params$gradientStruct$nlinDerivs2

  n.loc <- NROW(locations)
  nm    <- locations$param
  H     <- matrix(0.0, nrow = n.loc, ncol = n.loc, dimnames = list(nm, nm))
  grad  <- stats::setNames(numeric(n.loc), nm = nm)
  ncores <- qmlNcores()

  for (g in seq_len(modelFilled$info$n.groups)) {
    locations.g <- locations[locations$group == g, , drop = FALSE]
    if (!NROW(locations.g)) next

    HESS.g <- hessLogLikQmlCpp(
      submodel  = modelFilled$models[[g]],
      block     = locations.g$block,
      row       = locations.g$row,
      col       = locations.g$col,
      symmetric = locations.g$symmetric,
      relStep   = .relStep,
      minAbs    = 1.0,
      ncores    = ncores
    )

    nm.g <- locations.g$param
    dimnames(HESS.g$Hessian) <- list(nm.g, nm.g)
    names(HESS.g$gradient)   <- nm.g

    H[nm.g, nm.g] <- H[nm.g, nm.g] + HESS.g$Hessian
    grad[nm.g]    <- grad[nm.g] + HESS.g$gradient
  }

  if (length(nlinDerivs)) {
    evalTheta  <- model$params$gradientStruct$evalTheta
    param.full <- stringr::str_split_i(colnames(Jacobian), pattern = "#", i = 1L)
    param.part <- rownames(Jacobian)
    THETA      <- list2env(as.list(evalTheta(theta)))

    for (dep in names(nlinDerivs)) {
      derivs1 <- nlinDerivs[[dep]]
      derivs2 <- nlinDerivs2[[dep]]
      for (indep in names(derivs1)) {
        deriv1 <- eval(expr = derivs1[[indep]], envir = THETA)
        deriv2 <- eval(expr = derivs2[[indep]], envir = THETA)
        Jacobian [param.part == indep, param.full == dep] <- deriv1
        Jacobian2[param.part == indep, param.full == dep] <- deriv2
      }
    }
  }

  term1 <- Jacobian %*% H %*% t(Jacobian)
  term2 <- diag(drop(Jacobian2 %*% grad), nrow = nrow(Jacobian))
  sign * (term1 + term2)
}


hessianLogLikQml <- function(theta, model, sign = -1,
                              .relStep = .Machine$double.eps ^ (1/5),
                              .f = logLikQmlGroup) {
  if (qmlCanUseGradientStruct(model))
    return(simpleHessianLogLikQml(theta = theta, model = model,
                                   sign = sign, .relStep = .relStep))

  # Fallback: pure R-level FD Hessian when no usable gradient structure exists.
  params <- model$params

  SELECT_THETA_LAB  <- params$SELECT_THETA_LAB
  SELECT_THETA_COV  <- params$SELECT_THETA_COV
  SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN

  k <- length(theta)
  H <- matrix(0, nrow = k, ncol = k, dimnames = list(names(theta), names(theta)))

  for (g in seq_len(model$info$n.groups)) {
    indices <- c(SELECT_THETA_LAB[[g]], SELECT_THETA_COV[[g]], SELECT_THETA_MAIN[[g]])

    .fg <- function(theta.g) {
      theta[indices] <- theta.g
      modFilled <- fillModel(theta = theta, model = model, method = "qml")
      .f(submodel = modFilled$models[[g]], sign = sign, sum = TRUE)
    }

    theta.g <- theta[indices]
    Hg <- fdHESS(pars = theta.g, fun = .fg, .relStep = .Machine$double.eps^(1/5))
    H[indices, indices] <- H[indices, indices] + Hg
  }

  H
}


mstepQml <- function(model,
                     theta,
                     max.iter = 500,
                     verbose = FALSE,
                     convergence = 1e-6,
                     control = list(),
                     optimizer = "nlminb",
                     epsilon = 1e-8,
                     ...) {
  resetOptimizerInfoQML()

  gradient <- function(theta, model, sign, ...) # use ... to capture unused args
    gradientLogLikQml(theta = theta, model = model, epsilon = epsilon, sign = sign)

  if (optimizer == "nlminb") {
    control$iter.max <- max.iter
    control$eval.max <- max.iter * 2
    control$rel.tol  <- convergence

    est <- stats::nlminb(start = theta, objective = logLikQml, model = model,
                         gradient = gradient, sign = -1, verbose = verbose,
                         upper = model$params$bounds$upper,
                         lower = model$params$bounds$lower,
                         control = control, ...) |> suppressWarnings()

  } else if (optimizer == "L-BFGS-B") {
    control$factr <- convergence
    control$maxit <- max.iter

    est <- stats::optim(par = theta, fn = logLikQml, model = model,
                        gr = gradient, method = optimizer, sign = -1,
                        upper = model$params$bounds$upper,
                        lower = model$params$bounds$lower,
                        control = control, ...)

    est$objective  <- est$value
    est$iterations <- est$counts[["function"]]
  } else mod_msg_stop("Unrecognized optimizer, must be either 'nlminb' or 'L-BFGS-B'")

  if (verbose) cat("\n")

  est
}

#' Estimate interaction effects in structural equation models (SEMs) using a twostep procedure
#'
#' @param model.syntax \code{lavaan} syntax
#'
#' @param data dataframe
#'
#' @param method method to use:
#' \describe{
#'   \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
#'   \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
#'   \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
#'   \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
#'   \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
#'   \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
#'   \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
#'   \item{\code{"custom"}}{use parameters specified in the function call (passed to \code{lavaan}).}
#' }
#'
#' @param ... arguments passed to other functions depending on the method (see \code{\link{modsem_pi}} and \code{\link{modsem_da}})
#'
#' @return \code{modsem} object with class \code{\link{modsem_pi}} or \code{\link{modsem_da}}.
#'
#' @export
#' @description
#' Estimate an interaction model using a twostep procedure. For the PI approaches, the \code{lavaan::sam} function
#'   is used to optimize the models, instead of \code{lavaan::sem}. Note that the product indicators are still used,
#'   and not the newly developed SAM approach to estimate latent interactions. For the DA approaches (LMS and QML)
#'   the measurement model is estimated using a CFA (\code{lavaan::cfa}). The structural model is estimated using
#'   \code{\link{modsem_da}}, where the estimates in the measurement model are fixed, based on the CFA estimates.
#'   Note that standard errors are uncorrected (i.e., naive), and do not account for the uncertainty in the CFA estimates.
#'   \strong{NOTE, this is an experimental feature!}
#'
#' @examples
#' library(modsem)
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 +x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'
#'   # Inner model
#'   Y ~ X + Z + X:Z
#' '
#'
#' est_dblcent <- twostep(m1, oneInt, method = "dblcent")
#' summary(est_dblcent)
#'
#' \dontrun{
#' est_lms <- twostep(m1, oneInt, method = "lms")
#' summary(est_lms)
#'
#' est_qml <- twostep(m1, oneInt, method = "qml")
#' summary(est_qml)
#' }
#'
#' tpb_uk <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'  ATT =~ att3 + att2 + att1 + att4
#'  SN =~ sn4 + sn2 + sn3 + sn1
#'  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
#'  INT =~ int2 + int1 + int3 + int4
#'  BEH =~ beh3 + beh2 + beh1 + beh4
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'  # Causal Relationsships
#'  INT ~ ATT + SN + PBC
#'  BEH ~ INT + PBC
#'  BEH ~ INT:PBC
#' "
#'
#' uk_dblcent <- twostep(tpb_uk, TPB_UK, method = "dblcent")
#' summary(uk_dblcent)
#'
#' \dontrun{
#'   uk_qml <- twostep(tpb_uk, TPB_UK, method = "qml")
#'
#'   uk_lms <- twostep(tpb_uk, TPB_UK, method = "lms", nodes = 32, adaptive.quad = TRUE)
#'   summary(uk_lms)
#' }
#' @export
twostep <- function(model.syntax, data, method = "lms", ...) {
  if (method %in% PI_METHODS)      twostepfun <- twostepPI
  else if (method %in% DA_METHODS) twostepfun <- twostepDA
  else mod_msg_stop(paste0("Unsupported method: ", method))

  twostepfun(
     model.syntax = model.syntax,
     data         = data,
     method       = method,
     ...
  )
}


twostepPI <- function(..., LAVFUN = NULL) { # capture LAVFUN arg
  modsem_pi(..., LAVFUN = lavaan::sam)
}


twostepDA <- function(model.syntax, data, method = "lms", zero.tol = 1e-12,
                      fix.cov.xis = TRUE, ...) {
  mod_stopif(!method %in% DA_METHODS, paste0("Unsupported method: ", method))

  data <- as.data.frame(data)
  parTable      <- modsemify(model.syntax)
  parTableOuter <- parTable[parTable$op == "=~", ]
  parTableInner <- parTable[parTable$op != "=~", ]

  etas <- getEtas(parTable, isLV = FALSE)
  lvs  <- getLVs(parTable)
  syntaxCFA <- parTableToSyntax(parTableOuter)
  cfa       <- lavaan::cfa(syntaxCFA, data = data, meanstructure = TRUE)

  parTableCFA <- lavaan::parameterEstimates(cfa)

  fixCovVars <- if (fix.cov.xis) etas else lvs
  # Remove (co-) variances for etas (and optionally xis)
  parTableCFA <- parTableCFA[!(parTableCFA$op == "~~" &
                               parTableCFA$lhs %in% fixCovVars |
                               parTableCFA$rhs %in% fixCovVars), ]

  # Remove latent intercepts which are zero
  parTableCFA <- parTableCFA[!(parTableCFA$op == "~1" &
                               parTableCFA$lhs %in% lvs &
                               abs(parTableCFA$est) < zero.tol), ]

  parTableOuterFixed <- data.frame(
    lhs = parTableCFA$lhs,
    op = parTableCFA$op,
    rhs = parTableCFA$rhs,
    mod = format(parTableCFA$est, scientific = FALSE)
  )

  if ("label" %in% names(parTableCFA)) {
    labelledParams <- parTableCFA[parTableCFA$label != "", ]
    parTableCustom <- data.frame(
      lhs = labelledParams$label,
      op = ":=",
      rhs = as.character(labelledParams$est),
      mod = ""
    )
  } else parTableCustom <- NULL

  isNonLinEta <- vapply(
    X = etas,
    FUN.VALUE = logical(1L),
    FUN = intTermsAffectLV,
    parTable = parTable
  )

  if (any(isNonLinEta)) {
    etaIntercepts <- data.frame(
      lhs = etas[isNonLinEta],
      op = "~1",
      rhs = "",
      mod = ""
    )
  } else etaIntercepts <- NULL

  parTableTwoStep <- rbind(
    parTableCustom,
    parTableOuterFixed,
    parTableInner,
    etaIntercepts
  )

  syntaxTwoStep <- parTableToSyntax(parTableTwoStep)

  twostep.spec <- list(
    syntax        = syntaxTwoStep,
    cfa           = cfa,
    parTable      = parTableTwoStep,
    parTableOuter = parTableOuterFixed,
    parTableInner = parTableInner
  )

  fit <- modsem_da(
    model.syntax = syntaxTwoStep,
    data = data,
    method = method,
    ...
  )

  # prep partable from CFA
  parTableCFA <- addMissingGroups(rename(parTableCFA, se = "se.cfa"))
  parTableCFA <- parTableCFA[c("lhs", "op", "rhs", "group", "se.cfa")]
  parTableCFA$se.cfa[parTableCFA$se.cfa <= zero.tol] <- NA
  estParTable <- parameter_estimates(fit)

  # Add reversed order covariances, to ensure matching
  covrows <- parTableCFA$op == "~~" & parTableCFA$rhs != parTableCFA$lhs
  parTableCFACov     <- parTableCFA
  parTableCFACov$rhs <- parTableCFA$lhs
  parTableCFACov$lhs <- parTableCFA$rhs
  parTableCFA <- rbind(
    parTableCFA, parTableCFACov[covrows, ]
  )

  parTable.out <- leftJoin(left = estParTable, right = parTableCFA)

  parTable.out$std.error <- ifelse(
    is.na(parTable.out$std.error) & !is.na(parTable.out$se.cfa),
    yes = parTable.out$se.cfa,
    no  = parTable.out$std.error
  )

  parTable.out$z.value  <- parTable.out$est / parTable.out$std.error
  parTable.out$p.value  <- 2 * stats::pnorm(-abs(parTable.out$z.value))
  parTable.out$ci.lower <- parTable.out$est - CI_WIDTH * parTable.out$std.error
  parTable.out$ci.upper <- parTable.out$est + CI_WIDTH * parTable.out$std.error

  fit$parTable <- parTable.out

  parTabelLabelled <- getMissingLabels(parTable.out)

  # Get vcov
  vcov.11 <- lavaan::vcov(cfa)
  vcov.22 <- vcov(fit, type = "all")

  diff.pars <- setdiff(colnames(vcov.11), colnames(vcov.22))
  vcov.11   <- vcov.11[diff.pars, diff.pars]
  vcov.all  <- diagPartitionedMat(X = vcov.11, Y = vcov.22)

  # Get coef
  coef.1 <- lavaan::coef(cfa)
  coef.2 <- coef(fit, type = "all")

  diff.pars <- setdiff(names(coef.1), names(coef.2))
  coef.1    <- coef.1[diff.pars]
  coef.all  <- c(coef.1, coef.2)

  # Add to fit
  fit$vcov.all  <- vcov.all
  fit$coefs.all <- coef.all

  fit$twostep.spec <- twostep.spec

  fit
}

modsemOrderedScaleCorrection <- function(..., ordered.v2 = TRUE) {
  message("Transforming ordered variables to interval scale...\n",
          "This is an experimental feature, see `help(modsem_da)` for more information!")

  # v2 seems to be the best. Left here for testing purposes
  # The user can pass the `ordered.v2` argument to `modsem_da`
  # without it being publicly exposed as an option
  if (ordered.v2) FUN <- modsemOrderedScaleCorrectionV2
  else            FUN <- modsemOrderedScaleCorrectionV1

  FUN(...)
}


modsemOrderedScaleCorrectionV1 <- function(model.syntax,
                                           data,
                                           method = "lms",
                                           ordered = NULL,
                                           calc.se = TRUE,
                                           group = NULL,
                                           iter = 75L,
                                           warmup = 25L,
                                           N = max(NROW(data), 1e5), # 10,000 initally
                                           se = "simple",
                                           diff.theta.tol = 1e-10,
                                           ordered.mean.observed = FALSE,
                                           # Capture args
                                           verbose = interactive(),
                                           optimize = TRUE,
                                           start = NULL,
                                           mean.observed = NULL,
                                           probit.correction = FALSE,
                                           ...) {
  standardize <- \(x) (x - mean(x, na.rm = TRUE)) / stats::sd(x, na.rm = TRUE)

  stopif(!is.null(group), "Multigroup models do not work with: ",
         "`modsemOrderedScaleCorrectionV1()`!")

  if (is.null(verbose))
    verbose <- TRUE # default

  parTable.in <- modsemify(model.syntax)
  ovs <- getOVs(parTable.in)
  xis <- getXis(parTable.in, checkAny = FALSE)
  inds.xis <- unique(unlist(getIndsLVs(parTable.in, lVs = xis)))

  data <- as.data.frame(data)[ovs]
  cols <- ovs
  cols.ordered <- cols[cols %in% ordered | sapply(data, is.ordered)]
  cols.cont    <- cols[!(cols %in% cols.ordered) & vapply(data, is.numeric, TRUE)]

  ERROR <- \(e) {warning2(e, immediate. = FALSE); NULL}

  data.x <- data
  data.y <- data
  data.y[cols.ordered] <- lapply(data.y[cols.ordered], FUN = as.integer)

  stopif(iter <= warmup, "`ordered.iter` must be larger than `ordered.warmup`!")

  iter.keep <- max(1L, floor(iter - warmup))

  data_i <- data
  data_i[cols.ordered] <- lapply(data_i[cols.ordered],
                                 FUN = \(x) standardize(as.integer(x)))
  thresholds <- NULL
  theta_i    <- NULL

  pcorr <- polychor(vars = cols, data = data,
                    thresholds = NULL, # rescaled$thresholds,
                    ordered    = cols.ordered)

  for (i in seq_len(iter)) {
    printedLines <- utils::capture.output(split = TRUE, {
      j <- i - warmup
      mode <- if (j <= 0) "warmup" else "sampling"

      if (verbose)
        printf("Iterations %d/%d [%s]...\n", i, iter, mode)

      if (is.null(theta_i)) {
        optimize_i <- TRUE
        start_i    <- NULL
      } else {
        optimize_i <- FALSE
        start_i   <- theta_i
      }

      fit_i <- tryCatch(
        modsem_da(
          model.syntax     = model.syntax,
          data             = data_i,
          method           = method,
          start            = start_i,
          verbose          = verbose,
          optimize         = optimize_i,
          calc.se          = FALSE,
          mean.observed    = ordered.mean.observed,
          group            = group,
          ...
        ), error = \(e) {cat("\n"); print(e); NULL}
      )

      stopif(is.null(fit_i), "Model estimation failed!")

      theta_k <- theta_i
      theta_i <- fit_i$theta

      pars <- parameter_estimates(fit_i)

      sim_i <- simulateDataParTable(
        parTable = pars,
        N        = N
      )

      rescaled <- rescaleOrderedData(data = data.x,
                                     sim.ov = sim_i$oV,
                                     cols.ordered = cols.ordered,
                                     cols.cont = cols.cont,
                                     linear.ovs = inds.xis)

      if (probit.correction) data_j <- fitX2Cov(X = rescaled$data, S = pcorr)
      else                   data_j <- rescaled$data

      if (j >= 1L && !is.null(thresholds) && !is.null(data_i)) {
        lambda_j <- 1 / j
        lambda_i <- 1 - lambda_j

        for (col in cols.ordered)
          data_i[[col]] <- lambda_i * data_i[[col]] + lambda_j * data_j[[col]]

      } else {
        data_i     <- data_j
        thresholds <- rescaled$thresholds
      }
    })

    nprinted <- length(printedLines)
    if (i < iter) eraseConsoleLines(nprinted)

    epsilon    <- theta_i - theta_k
    diff.theta <- mean(abs(epsilon))

    if (diff.theta <= diff.theta.tol && j > 1L) {
      printf("Solution converged (iter = %d, warmup = %d, diff = %.2g)\n",
             i, warmup, diff.theta)

      break
    }
  }

  fit.out <- modsem_da(model.syntax     = model.syntax,
                       data             = data_i,
                       method           = method,
                       start            = theta_i,
                       verbose          = verbose,
                       optimize         = FALSE,
                       calc.se          = calc.se,
                       mean.observed    = ordered.mean.observed,
                       ...)

  rescaled <- rescaleOrderedData(data = data.x,
                                 sim.ov = sim_i$oV,
                                 cols.ordered = cols.ordered,
                                 cols.cont = cols.cont,
                                 linear.ovs = inds.xis,
                                 thresholds.vcov = TRUE)

  vcov.t <- NULL
  coef.t <- NULL
  for (col in cols.ordered) {
    vcov.t <- diagPartitionedMat(vcov.t, rescaled$vcov[[col]])
    coef.t <- c(coef.t, rescaled$thresholds[[col]])
  }

  parTable <- fit.out$parTable
  cols.t <- c("lhs", "op", "rhs", "label", "est", "std.error")
  cols.m <- setdiff(colnames(parTable), cols.t)

  lhs.t = stringr::str_split_i(names(coef.t), pattern = "\\|", i = 1L)
  rhs.t = stringr::str_split_i(names(coef.t), pattern = "\\|", i = 2L)

  parTable.t <- data.frame(lhs = lhs.t, op = "|", rhs = rhs.t, label = "",
                           est = coef.t, std.error = sqrt(diag(vcov.t)))
  parTable.t[cols.m] <- NA
  parTable.t <- addZStatsParTable(parTable.t)

  parTable <- sortParTableDA(rbind(parTable, parTable.t), model = fit.out$model)

  fit.out$coef.all <- c(fit.out$coef.all, coef.t)
  fit.out$vcov.all <- diagPartitionedMat(fit.out$vcov.all, vcov.t)
  fit.out$parTable <- parTable
  fit.out$args$optimize <- optimize
  fit.out$args$start    <- start

  fit.out
}


rescaleOrderedVariableMonteCarlo <- function(name,
                                             data,
                                             sim.ov,
                                             smooth_eps = 0,
                                             eps_expand = 1e-12,
                                             standardize = TRUE,
                                             thresholds.vcov = FALSE,
                                             vcov.boot = 250) {
  x   <- as.ordered(data[[name]])
  x.i <- as.integer(x)
  y   <- sim.ov[, name]

  # standardize for scale stability (no distributional assumption)
  y_mean <- mean(y, na.rm = TRUE)
  y_sd   <- stats::sd(y,  na.rm = TRUE)
  if (!is.finite(y_sd) || y_sd == 0) y_sd <- 1
  y <- (y - y_mean) / y_sd

  # observed category proportions (optional Laplace smoothing)
  tab <- as.numeric(table(x))
  K   <- length(tab)
  n   <- sum(tab)
  if (smooth_eps > 0) {
    p_obs <- (tab + smooth_eps) / (n + K * smooth_eps)
  } else {
    p_obs <- tab / n
  }

  # empirical CDF from data; quantile cutpoints on simulated y
  cdf_obs <- cumsum(p_obs[-length(p_obs)]) # drop last, we know that it sums to 1
  q <- stats::quantile(y, probs = cdf_obs, names = FALSE, type = 7)
  q <- c(-Inf, q, Inf) # [0, ..., last(p_obs)]

  # compute conditional means in each [q[i], q[i+1]) (last bin inclusive)
  mu <- numeric(K)
  for (i in seq_len(K)) {
    lo <- q[i]
    hi <- q[i + 1]

    # left-closed/right-open, last bin inclusive
    if (i < K) {
      idx <- (y >= lo & y < hi)
    } else {
      idx <- (y >= lo & y <= hi)
    }

    # guard for degenerate bins (ties / finite-N quantile artifacts)
    if (!any(idx)) {
      # expand a hair around the cutpoints to catch ties
      if (i < K) {
        idx <- (y >= (lo - eps_expand) & y < (hi + eps_expand))
      } else {
        idx <- (y >= (lo - eps_expand) & y <= (hi + eps_expand))
      }
      # still empty? fallback to midpoint
      if (!any(idx)) mu[i] <- 0.5 * (lo + hi)
    }

    if (any(idx)) mu[i] <- mean(y[idx])
  }

  if (thresholds.vcov) {
    Q <- matrix(NA_real_, nrow = vcov.boot, ncol = length(cdf_obs),
                dimnames = list(NULL, paste0(name, "|t", seq_along(cdf_obs))))

    for (i in seq_len(vcov.boot)) {
      N <- length(y)
      n <- length(x)

      replace  <- N <= 5L * n
      y.sample <- y[sample(N, n, replace = replace)]
      Q[i, ] <- stats::quantile(y.sample, probs = cdf_obs, names = FALSE, type = 7)
    }

    vcov <- stats::cov(Q)
  } else vcov <- NULL

  # exact centering to remove residual drift (weighted by observed category probs)
  mu <- mu - sum(p_obs * mu, na.rm = TRUE)

  # map each observed category to its conditional mean
  x.out <- rep(NA_real_, length(x))
  for (i in seq_len(K))
    x.out[x.i == i] <- mu[i]

  if (standardize)
    x.out <- std1(x.out)

  labels.t <- paste0(name, "|t", seq_len(sum(is.finite(q))))
  thresholds <- stats::setNames(q[is.finite(q)], labels.t)

  list(values = x.out, thresholds = thresholds, vcov = vcov)
}


rescaleOrderedVariableAnalytic <- function(name,
                                           data,
                                           sim.ov = NULL, # ignored
                                           smooth_eps = 0,
                                           standardize = TRUE,
                                           thresholds.vcov = FALSE,
                                           vcov.method = c("delta", "bootstrap"),
                                           vcov.boot = 250) {
  vcov.method <- match.arg(vcov.method)

  x   <- as.ordered(data[[name]])
  x.i <- as.integer(x)

  # observed category proportions (optional Laplace smoothing)
  tab <- as.numeric(table(x))
  K   <- length(tab)
  n   <- sum(tab)
  if (K < 2) stop("Need at least 2 ordered categories for '", name, "'.")

  if (smooth_eps > 0) {
    p_hat <- (tab + smooth_eps) / (n + K * smooth_eps)
  } else {
    p_hat <- tab / n
  }

  # cumulative probs for interior thresholds (K-1 of them)
  C <- cumsum(p_hat)[-K]  # drop last; sums to 1
  # numeric guards
  eps <- .Machine$double.eps^0.5
  C <- pmin(pmax(C, eps), 1 - eps)

  # thresholds on standard-normal scale
  t_int <- stats::qnorm(C)                 # length K-1
  q     <- c(-Inf, t_int, Inf)             # length K+1 for interval bounds

  # truncated-normal means for each category interval (a_i, b_i]
  # mu_i = (phi(a_i) - phi(b_i)) / (Phi(b_i) - Phi(a_i))
  a <- q[1:K]
  b <- q[2:(K+1)]
  Phi_a <- stats::pnorm(a)
  Phi_b <- stats::pnorm(b)
  phi_a <- stats::dnorm(a)
  phi_b <- stats::dnorm(b)
  denom <- pmax(Phi_b - Phi_a, eps)
  mu    <- (phi_a - phi_b) / denom

  # exact centering to remove small drift from smoothing/finite-n
  mu <- mu - sum(p_hat * mu)

  # map each observed category to its conditional mean
  x.out <- rep(NA_real_, length(x.i))
  for (i in seq_len(K)) x.out[x.i == i] <- mu[i]

  # optional standardization of the mapped scores (sample standardization)
  if (standardize)
    x.out <- std1(x.out)

  # labels for interior thresholds
  labels.t   <- paste0(name, "|t", seq_len(K - 1))
  thresholds <- stats::setNames(t_int, labels.t)

  vcov <- NULL
  if (thresholds.vcov && vcov.method == "delta") {
    # Delta method for t = qnorm(C), where C are cumulative probs from multinomial
    # Multinomial covariance of cell proportions: Var(p) = (diag(p) - pp^T)/n
    # Let C_j = sum_{i<=j} p_i, j=1..K-1.
    # J_p->C is lower-triangular with ones:
    J_pc <- matrix(0, nrow = K - 1, ncol = K)
    for (j in 1:(K - 1)) J_pc[j, 1:j] <- 1

    V_p  <- (diag(p_hat) - tcrossprod(p_hat, p_hat)) / n  # K x K
    V_C  <- J_pc %*% V_p %*% t(J_pc)                      # (K-1) x (K-1)

    # t = qnorm(C); dt/dC = 1 / dnorm(t) evaluated at t
    D_tC <- diag(1 / pmax(stats::dnorm(t_int), eps), nrow = K - 1)
    V_t  <- D_tC %*% V_C %*% D_tC

    colnames(V_t) <- rownames(V_t) <- labels.t
    vcov <- V_t

  } else if (thresholds.vcov && vcov.method == "bootstrap") {
    Q <- matrix(NA_real_, nrow = vcov.boot, ncol = K - 1,
                dimnames = list(NULL, labels.t))
    for (b_ix in seq_len(vcov.boot)) {
      # parametric bootstrap from Multinomial(n, p_hat)
      counts_b <- as.numeric(stats::rmultinom(1, size = n, prob = p_hat))
      p_b      <- counts_b / n
      C_b      <- cumsum(p_b)[-K]
      C_b      <- pmin(pmax(C_b, eps), 1 - eps)
      Q[b_ix,] <- stats::qnorm(C_b)
    }
    vcov <- stats::cov(Q)
  }

  list(values = x.out, thresholds = thresholds, vcov = vcov)
}


rescaleOrderedData <- function(data, sim.ov, cols.ordered, cols.cont,
                               thresholds.vcov = FALSE, linear.ovs = NULL,
                               ...) {
  if (!length(cols.ordered)) # nothing to do
    return(cols.ordered)

  # ensure column order consistent with data
  sim.mat <- as.data.frame(sim.ov)[, colnames(data), drop = FALSE]

  # make a full output copy right away, so continuous vars are preserved
  out <- as.data.frame(data, stringsAsFactors = FALSE)

  data.cat <- as.data.frame(data)
  data.cat[cols.ordered] <- lapply(data.cat[cols.ordered], as.integer)

  # thresholds + univariate conditional means for each ordered variable
  NamedList  <- \(nm) stats::setNames(vector("list", length(nm)), nm = nm)
  thresholds     <- NamedList(cols.ordered)
  univarEst      <- NamedList(cols.ordered)
  thresholdsVcov <- NamedList(cols.ordered)

  # standardized versions of the ordered cols in sim (these are the values we copy)
  sim.std.ord <- lapplyNamed(cols.ordered, \(col) std1(sim.mat[[col]]),
                             names = cols.ordered)
  sim.std.ord <- as.data.frame(sim.std.ord)

  for (col in cols.ordered) {
    if (col %in% linear.ovs) .f <- rescaleOrderedVariableAnalytic
    else                     .f <- rescaleOrderedVariableMonteCarlo

    scaling <- .f(name = col, data = data.cat, sim.ov = sim.mat,
                  thresholds.vcov = thresholds.vcov, ...)
    univarEst[[col]]      <- scaling$values
    thresholds[[col]]     <- scaling$thresholds
    thresholdsVcov[[col]] <- scaling$vcov
  }

  out[cols.ordered] <- as.data.frame(univarEst)[cols.ordered]
  # preserve original column order & types for continuous vars
  out <- out[, colnames(data), drop = FALSE]

  list(data = out, thresholds = thresholds, vcov = thresholdsVcov)
}


modsemOrderedScaleCorrectionV2 <- function(model.syntax,
                                           data,
                                           method = "lms",
                                           ordered = NULL,
                                           calc.se = TRUE,
                                           group = NULL,
                                           N = max(NROW(data), 1e5),
                                           se = "simple",
                                           diff.theta.tol = 1e-10,
                                           ordered.mean.observed = FALSE,
                                           verbose = interactive(),
                                           mean.observed = NULL,
                                           probit.correction = FALSE,
                                           ...) {
  standardize <- \(x) (x - mean(x, na.rm = TRUE)) / stats::sd(x, na.rm = TRUE)
  if (is.null(verbose)) verbose <- TRUE

  parTable.in <- modsemify(model.syntax)
  xis  <- getXis(parTable.in, checkAny = FALSE)
  inds <- getInds(parTable.in)

  parTable.cfa <- parTable.in[parTable.in$op == "=~" |
                              (parTable.in$lhs %in% inds &
                               parTable.in$rhs %in% inds), ,
                              drop = FALSE]

  data <- as.data.frame(data)
  cols <- getOVs(parTable.in)
  cols.ordered <- cols[cols %in% ordered | sapply(data[cols], is.ordered)]
  cols.cont    <- cols[!(cols %in% cols.ordered) & vapply(data[cols], is.numeric, TRUE)]
  ncat <- vapply(cols.ordered, FUN.VALUE = numeric(1L), FUN = \(x) length(unique(data[[x]])))

  warnif(any(ncat <= 2L) && !probit.correction,
         "Some ordered variables only have two categories.\n",
         "Consider passing `ordered.probit.correction=TRUE`")

  ordinalize <- function(x) {
    if   (is.ordered(x)) return(as.ordered(as.integer(x))) # re-order levels, e.g., [2, 3, 4] -> [1, 2, 3]
    else                 return(as.ordered(x))
  }

  data.x <- data
  data.x[cols.ordered] <- lapply(data.x[cols.ordered], FUN = ordinalize)

  if (verbose) printf("Estimating factor scores...\n")
  syntax.cfa <- parTableToSyntax(parTable.cfa)

  if (is.null(group)) group.vals <- rep(1L, NROW(data.x))
  else                group.vals <- data[[group]]

  data.s   <- NULL
  RESCALED <- list()

  for (g in unique(group.vals)) {
    data.xg <- data.x[group.vals == g, , drop = FALSE]

    fit.cfa <- lavaan::cfa(syntax.cfa, data = data.xg, ordered = cols.ordered, se = "none")
    fscores <- as.data.frame(lavaan::lavPredict(fit.cfa))

    rescaled.g <- rescaleOrderedData_OP(
      data            = data.xg,
      cols.ordered    = cols.ordered,
      parTable.in     = parTable.in,
      fscores         = fscores,
      thresholds.vcov = TRUE
    )

    RESCALED[[g]] <- rescaled.g
    data.sg <- rescaled.g$data[cols]

    pcorr <- polychor(vars = cols, data = data.xg,
                      thresholds = NULL, ordered = cols.ordered)

    if (probit.correction)
      data.sg <- fitX2Cov(X = data.sg, S = pcorr)

    data.sg <- as.data.frame(data.sg)

    if (!is.null(group))
      data.sg[[group]] <- g

    data.s <- rbind(data.s, data.sg)
  }

  fit.out <- modsem_da(model.syntax     = model.syntax,
                       data             = data.s,
                       method           = method,
                       verbose          = verbose,
                       calc.se          = calc.se,
                       mean.observed    = ordered.mean.observed,
                       group            = group,
                       ...)

  for (g in seq_along(RESCALED)) {
    rescaled.g <- RESCALED[[g]]

    vcov.t <- NULL
    coef.t <- NULL

    for (col in cols.ordered) {
      if (!is.null(rescaled.g$vcov[[col]]) && length(rescaled.g$vcov[[col]]))
        vcov.t <- diagPartitionedMat(vcov.t, rescaled.g$vcov[[col]])

      if (!is.null(rescaled.g$thresholds[[col]]) && length(rescaled.g$thresholds[[col]]))
        coef.t <- c(coef.t, rescaled.g$thresholds[[col]])
    }

    parTable <- fit.out$parTable
    cols.t <- c("lhs", "op", "rhs", "group", "label", "est", "std.error")
    cols.m <- setdiff(colnames(parTable), cols.t)

    if (length(coef.t)) {
      lhs.t <- stringr::str_split_i(names(coef.t), pattern = "\\|", i = 1L)
      rhs.t <- stringr::str_split_i(names(coef.t), pattern = "\\|", i = 2L)

      parTable.t <- data.frame(lhs = lhs.t, op = "|", rhs = rhs.t, label = "",
                               group = g, est = coef.t,
                               std.error = if (!is.null(vcov.t)) sqrt(diag(vcov.t)) else NA_real_)
      parTable.t[cols.m] <- NA

      parTable.t <- addZStatsParTable(parTable.t)
      parTable   <- sortParTableDA(rbind(parTable, parTable.t), model = fit.out$model)
      fit.out$coef.all <- c(fit.out$coef.all, coef.t)

      if (!is.null(vcov.t))
        fit.out$vcov.all <- diagPartitionedMat(fit.out$vcov.all, vcov.t)
    }
  }

  fit.out$parTable <- parTable
  fit.out
}


# Closed-form CE for ordered-probit (vectorized)
# y: integer categories 1..K, eta: linear predictor, tau: interior thresholds (length K-1)
calcCondExpOrdProbit <- function(y, eta, tau, eps = 1e-12) {
  K <- max(y, na.rm = TRUE)
  a <- c(-Inf, tau)[y]
  b <- c(tau,  Inf)[y]
  al <- a - eta
  bl <- b - eta
  num <- stats::dnorm(al) - stats::dnorm(bl)
  den <- pmax(stats::pnorm(bl) - stats::pnorm(al), eps)
  eta + num / den
}


imputeWithOrdProbit <- function(y.ord, X, standardize = TRUE, weights = NULL,
                                std.thresholds = TRUE,
                                vcov.method = c("delta", "simple", "none")) {
  vcov.method <- match.arg(vcov.method)

  # 1) Fit ordered probit with Hessian
  y.fac <- if (is.factor(y.ord) || is.ordered(y.ord)) y.ord else factor(y.ord, ordered = TRUE)
  df    <- as.data.frame(X)
  fit   <- suppressWarnings(MASS::polr(y.fac ~ ., data = df, weights = weights,
                                       method = "probit", Hess = TRUE))

  # 2) Extract coefficients and rebuild design as used by polr
  beta <- stats::coef(fit)                 # length P
  tau_probit <- as.numeric(fit$zeta)       # length Q = K-1
  K    <- length(levels(y.fac)); Q <- K - 1L
  yint <- as.integer(y.fac)

  Xmm_all <- stats::model.matrix(~ ., data = df)
  Xmm     <- if (length(beta)) Xmm_all[, names(beta), drop = FALSE] else
             matrix(0, nrow = nrow(df), ncol = 0) # degenerate
  eta     <- as.numeric(Xmm %*% beta)
  n       <- length(eta)

  # weights handling
  if (is.null(weights)) {
    w <- rep(1, n); wsum <- n
  } else {
    if (length(weights) != n) stop("weights must have length nrow(X).")
    w <- as.numeric(weights); wsum <- sum(w)
  }

  # 3) CE (conditional expectations) on probit scale; used only for imputed values
  a <- c(-Inf, tau_probit); b <- c(tau_probit,  Inf)
  eps <- 1e-12
  m <- matrix(NA_real_, nrow = n, ncol = K,
              dimnames = list(NULL, levels(y.fac)))
  for (j in seq_len(K)) {
    alpha <- (a[j] - eta)
    betaJ <- (b[j] - eta)
    numer <- stats::dnorm(alpha) - stats::dnorm(betaJ)
    denom <- stats::pnorm(betaJ) - stats::pnorm(alpha); denom <- pmax(denom, eps)
    m[, j] <- eta + numer / denom
  }

  # Aggregate to a single CE constant per observed category
  mu_cat <- numeric(K)
  for (j in seq_len(K)) {
    idx <- which(yint == j)
    if (!length(idx)) {
      pj <- stats::pnorm(b[j] - eta) - stats::pnorm(a[j] - eta)
      mu_cat[j] <- sum(m[, j] * pj * w) / max(sum(pj * w), eps)
    } else {
      wj <- w[idx]
      mu_cat[j] <- sum(m[idx, j] * wj) / sum(wj)
    }
  }
  names(mu_cat) <- levels(y.fac)
  y_ce <- mu_cat[yint]
  if (standardize) y_ce <- std1(y_ce)

  # 4) Thresholds: rescale from probit (Var eps =1) to SEM unit-Var(y*)
  #    y* variance = Var(eta) + 1 ; mean = E[eta]
  mu_eta <- sum(w * eta) / wsum
  # weighted variance (population form)
  var_eta <- sum(w * (eta - mu_eta)^2) / wsum
  s <- sqrt(var_eta + 1)

  tau.ystar <- if (std.thresholds) (tau_probit - mu_eta) / s else tau_probit

  # 5) Vcov for thresholds
  vcov.full <- tryCatch(stats::vcov(fit), error = function(e) NULL)
  vcov.tau.probit <- vcov.tau.ystar <- NULL
  if (!is.null(vcov.full) && Q > 0) {
    P <- length(beta)
    vcov.tau.probit <- as.matrix(vcov.full[P + seq_len(Q), P + seq_len(Q), drop = FALSE])

    if (vcov.method == "none") {
      vcov.tau.ystar <- NULL

    } else if (!std.thresholds || s <= 0) {
      # no rescale -> same vcov; if s invalid keep NULL
      vcov.tau.ystar <- vcov.tau.probit

    } else if (vcov.method == "simple") {
      # only scale by 1/s^2, ignore uncertainty in mu_eta and s
      vcov.tau.ystar <- vcov.tau.probit / (s^2)

    } else if (vcov.method == "delta") {
      # Full delta: tau' = (tau - mu_eta)/s, with mu_eta = E[X]^T beta,
      # s = sqrt(beta^T Sx beta + 1). Build J and compute J V J^T.
      EX <- if (P) colSums(Xmm * w) / wsum else numeric(0)
      # weighted covariance of X columns
      if (P) {
        Xc <- sweep(Xmm, 2, EX, FUN = "-")
        Sx <- (t(Xc * w) %*% Xc) / wsum   # P x P
        Sb <- as.numeric(Sx %*% beta)     # P-vector: Sx * beta
      } else {
        Sx <- matrix(0, 0, 0); Sb <- numeric(0)
      }

      J <- matrix(0, nrow = Q, ncol = P + Q)  # rows: thresholds; cols: [beta, tau]
      if (P) {
        # d tau'_k / d beta = -(EX)/s - (tau_k - mu_eta) * (Sx beta) / s^3
        for (k in seq_len(Q)) {
          J[k, 1:P] <- -(EX) / s - (tau_probit[k] - mu_eta) * (Sb) / (s^3)
        }
      }
      # d tau'_k / d tau_j = (1/s) if j==k else 0
      for (k in seq_len(Q)) J[k, P + k] <- 1 / s

      V <- as.matrix(vcov.full)
      vcov.tau.ystar <- J %*% V %*% t(J)
      # ensure symmetry
      vcov.tau.ystar <- 0.5 * (vcov.tau.ystar + t(vcov.tau.ystar))
    }
  }

  # Named outputs for thresholds (we’ll re-prefix with indicator name upstream)
  names(tau_probit) <- paste0("t", seq_len(Q))
  names(tau.ystar)  <- paste0("t", seq_len(Q))
  if (!is.null(vcov.tau.probit)) dimnames(vcov.tau.probit) <- list(names(tau_probit), names(tau_probit))
  if (!is.null(vcov.tau.ystar))  dimnames(vcov.tau.ystar)  <- list(names(tau.ystar),  names(tau.ystar))

  list(
    values          = y_ce,              # imputed y* CE values (you standardize again downstream if desired)
    mu_per_cat      = mu_cat,
    tau_probit      = tau_probit,        # thresholds on probit-error scale
    vcov.tau.probit = vcov.tau.probit,
    tau.ystar       = tau.ystar,         # thresholds on SEM y* unit-variance scale
    vcov.tau.ystar  = vcov.tau.ystar,    # vcov for tau.ystar (delta/simple/none)
    eta             = eta,
    mu_eta          = mu_eta,
    s_ystar         = s,                 # scaling used: s = sqrt(Var(eta)+1)
    fit             = fit
  )
}


.buildX_ForIndicator <- function(latent, rhsTerms, fscores) {
  X <- list()  # don't include latent self
  stopif(!length(rhsTerms), sprintf("expected predictors for eta (%s)!", latent))

  for (trm in rhsTerms) {
    if (grepl(":", trm, fixed = TRUE)) {
      parts <- strsplit(trm, ":", fixed = TRUE)[[1]]
      stopifnot(all(parts %in% colnames(fscores)))
      X[[trm]] <- Reduce(`*`, lapply(parts, function(p) fscores[[p]]))

    } else {
      stopifnot(trm %in% colnames(fscores))
      X[[trm]] <- fscores[[trm]]
    }
  }

  X <- as.data.frame(X)

  # drop duplicate columns if any (name collisions)
  X[, !duplicated(names(X)), drop = FALSE]
}


# Hybrid rescaler:
# - For ordered indicators of ENDOGENOUS latents: ordered-probit CE using fscores + interactions
# - For other ordered indicators: analytic (normal) mapping
rescaleOrderedData_OP <- function(data, cols.ordered, parTable.in, fscores,
                                  thresholds.vcov = FALSE, ...) {
  if (!length(cols.ordered)) return(list(data = data, thresholds = list(), vcov = list()))

  out <- as.data.frame(data, stringsAsFactors = FALSE)
  data.cat <- as.data.frame(data)
  data.cat[cols.ordered] <- lapply(data.cat[cols.ordered], as.integer)

  # maps: indicator -> latent (from loadings)
  meas    <- parTable.in[parTable.in$op == "=~", , drop = FALSE]
  ind2lat <- stats::setNames(as.character(meas$lhs), meas$rhs)

  # structural terms per endogenous latent
  struc       <- parTable.in[parTable.in$op == "~", , drop = FALSE]
  rhsByLatent <- tapply(struc$rhs, struc$lhs, c, simplify = FALSE)

  etas <- getEtas(parTable.in, checkAny = FALSE)
  ordered_endog <- intersect(cols.ordered, names(ind2lat)[ind2lat %in% etas])

  thresholds <- stats::setNames(vector("list", length(cols.ordered)), nm = cols.ordered)
  vcovs      <- stats::setNames(vector("list", length(cols.ordered)), nm = cols.ordered)

  for (col in cols.ordered) {
    if (col %in% ordered_endog) {
      # indicator belongs to an endogenous latent -> ordered-probit CE
      latent   <- ind2lat[[col]]
      rhsTerms <- rhsByLatent[[latent]]

      if (is.null(rhsTerms)) rhsTerms <- character(0)

      X   <- .buildX_ForIndicator(latent, rhsTerms, fscores)
      imp <- imputeWithOrdProbit(y.ord = data[[col]], X = X, standardize = TRUE)

      out[[col]] <- imp$values
      if (length(imp$tau.ystar)) {
        tn <- paste0(col, "|t", seq_along(imp$tau.ystar))
        thresholds[[col]] <- stats::setNames(imp$tau.ystar, tn)

        if (!is.null(imp$vcov.tau.ystar)) {
          dimnames(imp$vcov.tau.ystar) <- list(tn, tn)
          vcovs[[col]] <- imp$vcov.tau.ystar

        } else vcovs[[col]] <- NULL

      } else {
        thresholds[[col]] <- numeric(0); vcovs[[col]] <- NULL
      }

    } else {
      # exogenous or non-endogenous ordered indicator -> analytic normal mapping
      ana <- rescaleOrderedVariableAnalytic(name = col, data = data.cat,
                                            thresholds.vcov = thresholds.vcov, ...)
      out[[col]]        <- ana$values
      thresholds[[col]] <- ana$thresholds
      vcovs[[col]]      <- ana$vcov
    }
  }

  list(data = out, thresholds = thresholds, vcov = vcovs)
}


fitX2Cov <- function(X, S) {
  X <- as.matrix(X)
  n <- nrow(X)
  p <- ncol(X)

  stopif(!all(dim(S) == c(p, p)), "S must be a square matrix with dimensions equal to the number of columns in X")

  # Compute sample mean and covariance of X
  mu_X <- colMeans(X)
  X_centered <- sweep(X, 2, mu_X, FUN = "-")

  Sigma_X <- stats::cov(X)

  # Check positive definiteness
  if (any(eigen(Sigma_X, symmetric = TRUE)$values <= 0)) {
    warning("Sample covariance matrix of X is not positive-definite. Consider regularization.")
    return(X)
  }
  if (any(eigen(S, symmetric = TRUE)$values <= 0)) {
    warning("Target covariance matrix S must be positive-definite.")
    return(X)
  }

  # Matrix square root function using eigen-decomposition
  matrix_sqrt <- function(M) {
    eig <- eigen(M, symmetric = TRUE)
    eig$vectors %*% diag(sqrt(eig$values)) %*% t(eig$vectors)
  }

  # Compute inverse square root of Sigma_X
  Sigma_X_inv_sqrt <- solve(matrix_sqrt(Sigma_X))

  # Compute square root of S
  S_sqrt <- matrix_sqrt(S)

  # Transformation matrix A
  A <- Sigma_X_inv_sqrt %*% S_sqrt

  # Transform the centered data
  Y_centered <- X_centered %*% A

  # Add target mean back
  Y <- sweep(Y_centered, 2, mu_X, FUN = "+")

  colnames(Y) <- colnames(X)
  as.data.frame(Y)
}


polychor <- function(vars, data, thresholds = NULL, ordered = NULL) {
  combos <- getUniqueCombos(vars, match = FALSE)
  syntax <- paste0(sprintf("%s~~%s", combos$V1, combos$V2), collapse = "\n")

  for (var in names(thresholds)) {
    tau  <- thresholds[[var]]
    ftau <- format(tau[is.finite(tau)], scientific = FALSE)

    constr <- sprintf("%s|%s*t%d", var, ftau, seq_along(tau))
    syntax <- paste(syntax, paste0(constr, collapse = "\n"), sep = "\n")
  }

  lavaan::lavInspect(lavaan::sem(syntax, data, ordered = ordered, se = "none"),
                     what = "cov.ov")
}

#' Standardize a fitted \code{modsem_da} model
#'
#' \code{standardize_model()} post-processes the output of
#' \code{\link{modsem_da}()} (or of \code{\link{modsem}())} when \code{method = "lms"} /
#' \code{method = "qml"}), replacing the \emph{unstandardized} coefficient vector
#' (\code{$coefs}) and its variance–covariance matrix (\code{$vcov}) with \emph{fully
#' standardized} counterparts (including the measurement model).The procedure is purely
#' algebraic— \strong{no re-estimation is carried out} —and is therefore fast and
#' deterministic.
#'
#' @param model A fitted object of class \code{modsem_da}.
#'   Passing any other object triggers an error.
#'
#' @param monte.carlo Logical. If \code{TRUE}, the function will use Monte Carlo
#'   simulation to obtain the standard errors of the standardized estimates.
#'   If \code{FALSE}, the delta method is used.
#'   Default is \code{FALSE}.
#'
#' @param mc.reps Number of Monte Carlo replications. Default is 10,000.
#'   Ignored if \code{monte.carlo = FALSE}.
#'
#' @param ... Arguments passed on to other functions
#'
#' @return The same object (returned invisibly) with three slots overwritten
#' \describe{
#'   \item{\code{$parTable}}{Parameter table whose columns \code{est} and \code{std.error}
#'         now hold standardized estimates and their (delta-method)
#'         standard errors, as produced by \code{\link{standardized_estimates}()}.}
#'   \item{\code{$coefs}}{Named numeric vector of standardized coefficients.
#'         Intercepts (operator \code{~1}) are removed, because a standardized
#'         variable has mean 0 by definition.}
#'   \item{\code{$vcov}}{Variance–covariance matrix corresponding to the updated
#'         coefficient vector.  Rows/columns for intercepts are dropped, and
#'         the sub-matrix associated with rescaled parameters is adjusted
#'         so that its diagonal equals the squared standardized standard errors.}
#' }
#' The object keeps its class attributes, allowing seamless use by downstream
#' S3 methods such as \code{summary()}, \code{coef()}, or \code{vcov()}.
#'
#' Because the function merely transforms existing estimates, \emph{parameter
#' constraints imposed through shared labels remain satisfied}.
#'
#' @seealso
#' \describe{
#'    \item{\code{\link{standardized_estimates}()}}{Obtains the fully standardized
#'          parameter table used here.}
#'    \item{\code{\link{modsem}()}}{Fit model using LMS or QML approaches.}
#'    \item{\code{\link{modsem_da}()}}{Fit model using LMS or QML approaches.}
#' }
#' @examples
#' \dontrun{
#' # Latent interaction estimated with LMS and standardized afterwards
#' syntax <- "
#'   X =~ x1 + x2 + x3
#'   Y =~ y1 + y2 + y3
#'   Z =~ z1 + z2 + z3
#'   Y ~ X + Z + X:Z
#' "
#' fit  <- modsem_da(syntax, data = oneInt, method = "lms")
#' sfit <- standardize_model(fit, monte.carlo = TRUE)
#'
#' # Compare unstandardized vs. standardized summaries
#' summary(fit)  # unstandardized
#' summary(sfit) # standardized
#' }
#'
#' @export
standardize_model <- function(model, monte.carlo = FALSE, mc.reps = 10000, ...) {
  mod_stopif(!inherits(model, "modsem_da"), "The model must be of class 'modsem_da'.")

  solution <- standardizedSolutionCOEFS(model, monte.carlo = monte.carlo, mc.reps = mc.reps, ...)
  all.pars <- names(coef(model, type = "all"))

  coefs.all <- solution$coefs
  vcov.all  <- solution$vcov

  # Some parameters are lost as a function of the standardization procedure
  # E.g., intercept parameters. Here we add them back in, to avoid breaking
  # downstream functions appliead to the model object
  coefs.all <- expandCoef(coefs.all, labels = all.pars)
  vcov.all  <- expandVCOV(vcov.all, labels = all.pars)

  # It should be safe to use the same subset for both vcov.all and coefs.all
  # but in case something has gone wrong, we create seperate masks
  cnames.all <- names(coefs.all)
  vnames.all <- rownames(vcov.all)

  # intersect() shouldn't be neccessary, but just in case...
  cnames.free <- intersect(cnames.all, names(coef(model, type = "free")))
  vnames.free <- intersect(vnames.all, rownames(vcov(model, type = "free")))

  coefs.free <- coefs.all[cnames.free]
  vcov.free  <- vcov.all[vnames.free, vnames.free, drop = FALSE]

  parTable <- solution$parTable

  model$type.estimates <- "standardized"
  model$parTable       <- parTable

  model$coefs.all      <- coefs.all
  model$coefs.free     <- coefs.free

  model$vcov.all       <- vcov.all
  model$vcov.free      <- vcov.free

  model
}


rescaleVCOV <- function(vcov, new_se) {
  isZeroOrNA <- is.na(new_se) | new_se == 0
  S <- vcov[!isZeroOrNA, !isZeroOrNA]

  D_old_inv <- diag(1 / sqrt(diag(S)))
  D_new <- diag(new_se[!isZeroOrNA])

  R <- D_old_inv %*% S %*% D_old_inv
  Z <- D_new %*% R %*% D_new

  V <- vcov
  V[!isZeroOrNA, !isZeroOrNA] <- Z

  V
}

# Function for running the same model with multiple methods
# Used within tests
allMethods <- c("rca", "uca", "ca", "dblcent", "mplus", "pind")
allNativeMethods <- allMethods[allMethods != "mplus"]
fastMethods <- c("rca", "uca", "dblcent", "pind")

runMultipleMethods <- function(model.syntax,
                               data,
                               methods = allNativeMethods,
                               ...) {
  estimates <- structure(vector("list", length = length(methods)),
                         names = methods)
  for (method in methods) {
    estimates[[method]] <- tryCatch(
      modsem(model.syntax, data, method, ...),
      warning = function(w) {
        mod_msg_warn(paste0("Warning in ", method, "\n", conditionMessage(w), "\n"))
        modsem(model.syntax, data, method, ...)
      },
      error = function(e) {
        mod_msg_warn(paste0("Error in ", method, "\n", conditionMessage(e)))
        NA
      }
    )
  }
  estimates
}

#' Fit measures for QML and LMS models
#'
#' @param model model to be assessed
#' @description Calculates chi-sq test and p-value, as well as RMSEA for
#' the LMS and QML models. Note that the Chi-Square based fit measures should be calculated
#' for the baseline model, i.e., the model without the interaction effect
#'
#' @param model fitted model. Thereafter, you can use 'compare_fit()'
#' to assess the comparative fit of the models. If the interaction effect makes
#' the model better, and e.g., the RMSEA is good for the baseline model,
#' the interaction model likely has a good RMSEA as well.
#'
#' @param chisq should Chi-Square based fit-measures be calculated?
#'
#' @param lav.fit Should fit indices from the \code{lavaan} model used to optimize
#'   the starting parameters be included (if available)? This is usually only approprioate
#'   for linear models (i.e., no interaction effects), where the parameter estimates
#'   for LMS and QML are equivalent to ML estimates from lavaan.
#'
#' @param drop.list.single.group Logical. If FALSE, (some) results are returned as
#'   a list, where each element corresponds to a group (even if there is only a single group).
#'   If TRUE, the list will be unlisted if there is only a single group
#'
#' @export
fit_modsem_da <- function(model, chisq = TRUE, lav.fit = FALSE,
                          drop.list.single.group = TRUE) {
  parTable <- model$parTable
  if (isTRUE(chisq) && any(grepl(":", parTable$rhs), na.rm = TRUE)) {
    mod_msg_warn(paste0("Chi-Square based fit-measures for LMS and QML ",
             "should be calculated for baseline model ",
             "i.e., the model without the interaction effect"))
  }

  t      <- nFreeInterceptsDA(model)
  mean.s <- model$args$mean.observed || t > 0
  logLik <- model$logLik
  coef   <- coef(model, type = "free")
  npar   <- length(coef)

  finalModel <- model$model
  submodels  <- if (!is.null(finalModel$models)) finalModel$models else list()
  if (!length(submodels)) submodels <- list(list(data = model$data))

  n.groups <- max(1L, length(submodels))
  group.labels <- finalModel$info$group.levels

  if (!is.null(group.labels) && length(group.labels) != n.groups)
    group.labels <- NULL

  expected.list <- model$expected.matrices
  if (!is.list(expected.list) || length(expected.list) == 0)
    expected.list <- replicate(n.groups, expected.list, simplify = FALSE)

  if (length(expected.list) < n.groups) {
    expected.list <- c(expected.list,
                       rep(list(NULL), n.groups - length(expected.list)))
  }

  expected.list <- expected.list[seq_len(n.groups)]

  sampleCovBlocks <- vector("list", n.groups)
  expectedCovBlocks <- vector("list", n.groups)
  muObsBlocks <- vector("list", n.groups)
  muExpBlocks <- vector("list", n.groups)

  N_vec <- numeric(n.groups)
  p_vec <- numeric(n.groups)
  chisqParts <- rep(NA_real_, n.groups)
  srmrWeightNum <- 0
  srmrWeightDen <- 0

  for (g in seq_len(n.groups)) {
    submodel <- submodels[[g]]
    data.g   <- submodel$data

    data_full <- data.g$data.full
    if (is.null(data_full) || !NROW(data_full) || !NCOL(data_full)) {
      next
    }

    data_full <- as.matrix(data_full)
    N.g <- NROW(data_full)
    p.g <- NCOL(data_full)

    N_vec[g] <- N.g
    p_vec[g] <- p.g

    expected.g <- expected.list[[g]]
    has.ord <- !is.null(expected.g) &&
      !is.null(expected.g$sigma.ord.observed) &&
      !is.null(expected.g$sigma.ord.expected)

    if (has.ord) {
      O.g <- expected.g$sigma.ord.observed
      E.g <- expected.g$sigma.ord.expected
      mu.g <- if (!is.null(expected.g$mu.ord.observed)) expected.g$mu.ord.observed else
        matrix(0, nrow = nrow(O.g), ncol = 1, dimnames = list(rownames(O.g), "~1"))
      mu_hat.g <- if (!is.null(expected.g$mu.ord.expected)) expected.g$mu.ord.expected else
        matrix(0, nrow = nrow(E.g), ncol = 1, dimnames = list(rownames(E.g), "~1"))
      p.g <- NCOL(O.g)
    } else {
      mu.g_vec <- apply(data_full, 2, mean, na.rm = TRUE)
      mu.g <- matrix(mu.g_vec, ncol = 1,
                     dimnames = list(colnames(data_full), "~1"))
      O.g <- stats::cov(data_full, use = "pairwise.complete.obs")
      E.g <- if (!is.null(expected.g) && !is.null(expected.g$sigma.ov)) expected.g$sigma.ov else NULL
      mu_hat.g <- if (!is.null(expected.g) && !is.null(expected.g$mu.ov)) expected.g$mu.ov else NULL
    }

    if (!mean.s || is.null(mu_hat.g))
      mu_hat.g <- mu.g

    if (!is.null(E.g)) {
      order_vars <- colnames(data_full)
      E.g <- E.g[order_vars, order_vars, drop = FALSE]
      mu_hat.g <- mu_hat.g[order_vars, , drop = FALSE]
    }

    sampleCovBlocks[[g]] <- modsemMatrix(O.g, symmetric = TRUE)
    muObsBlocks[[g]] <- modsemMatrix(mu.g)

    if (!is.null(E.g)) {
      expectedCovBlocks[[g]] <- modsemMatrix(E.g, symmetric = TRUE)
      muExpBlocks[[g]] <- modsemMatrix(mu_hat.g)
    } else {
      expectedCovBlocks[[g]] <- NULL
      muExpBlocks[[g]] <- NULL
    }

    if (chisq && !is.null(E.g)) {
      chi.g <- tryCatch(
        calcChiSqr(O = O.g, E = E.g, N = N.g, p = p.g, mu = mu.g, mu.hat = mu_hat.g),
        error = function(e) {
          mod_msg_warn(paste0("Failed to compute chi-square contribution for group ",
                   g, ": ", conditionMessage(e)))
          NA_real_
        }
      )
      chisqParts[[g]] <- chi.g

      if (!is.na(chi.g)) {
        srmr.g <- tryCatch(
          calcSRMR_Mplus(S = O.g, M = mu.g, Sigma.hat = E.g, Mu.hat = mu_hat.g,
                         mean.structure = mean.s),
          error = function(e) NA_real_
        )

        if (!is.na(srmr.g)) {
          weight <- max(N.g, 1)
          srmrWeightNum <- srmrWeightNum + weight * srmr.g^2
          srmrWeightDen <- srmrWeightDen + weight
        }
      }
    }
  }

  names(sampleCovBlocks)   <- group.labels
  names(expectedCovBlocks) <- group.labels
  names(muObsBlocks)       <- group.labels
  names(muExpBlocks)       <- group.labels

  N_total <- sum(N_vec)
  df <- getDegreesOfFreedom(
    p = p_vec[p_vec > 0], coef = coef, mean.structure = mean.s,
    model = model
  )

  if (chisq) {
    if (all(!is.na(chisqParts))) {
      chisqValue <- sum(chisqParts)

      if (!is.na(df) && df > 0)
        chisqP <- stats::pchisq(chisqValue, df, lower.tail = FALSE)
      else
        chisqP <- NA_real_

      if (!is.na(df) && df > 0 && N_total > 0) {
        RMSEA <- calcRMSEA(chisqValue, df, N_total)

      } else {
        RMSEA <- list(rmsea = NA_real_, lower = NA_real_, upper = NA_real_,
                      ci.level = NA_real_, pvalue = NA_real_, close.h0 = NA_real_)
      }

      SRMR       <- if (srmrWeightDen > 0) sqrt(srmrWeightNum / srmrWeightDen) else NA_real_

    } else {
      chisqValue <- NA_real_
      chisqP     <- NA_real_
      RMSEA      <- list(rmsea = NA_real_, lower = NA_real_, upper = NA_real_,
                         ci.level = NA_real_, pvalue = NA_real_, close.h0 = NA_real_)
      SRMR       <- NA_real_
    }

  } else {
    chisqValue     <- NULL
    chisqP         <- NULL
    df             <- NULL
    sigma_expected <- NULL
    mu_expected    <- NULL
    SRMR           <- NULL
    RMSEA          <- list(
      rmsea = NULL,
      lower = NULL,
      upper = NULL,
      ci.level = NULL,
      pvalue = NULL,
      close.h0 = NULL
    )
  }

  AIC  <- calcAIC(logLik, k = npar)
  AICc <- calcAdjAIC(logLik, k = npar, N = N_total)
  BIC  <- calcBIC(logLik, k = npar, N = N_total)
  aBIC <- calcAdjBIC(logLik, k = npar, N = N_total)

  if (lav.fit) {
    lavfit <- tryCatch(
      lavaan::fitMeasures(model$model$lavaan.fit),
      error = function(e) {
        mod_msg_warn(paste0("Unable to retrieve fit measures for lavaan model!\n",
                 "Message: ", conditionMessage(e)))
        NULL
      }
    )
  } else lavfit <- NULL

  if (drop.list.single.group && n.groups <= 1L) {
    sampleCovBlocks   <- sampleCovBlocks[[1L]]
    expectedCovBlocks <- expectedCovBlocks[[1L]]
    muObsBlocks       <- muObsBlocks[[1L]]
    muExpBlocks       <- muExpBlocks[[1L]]
  }

  list(
    lav.fit = lavfit,

    sigma.observed = sampleCovBlocks,
    sigma.expected = expectedCovBlocks,
    mu.observed    = muObsBlocks,
    mu.expected    = muExpBlocks,

    npar         = npar,
    chisq.value  = chisqValue,
    chisq.pvalue = chisqP,
    chisq.df     = df,
    chisq.group  = stats::setNames(chisqParts, group.labels),

    AIC  = AIC,
    AICc = AICc,
    BIC  = BIC,
    aBIC = aBIC,
    SRMR = SRMR,

    RMSEA          = RMSEA$rmsea,
    RMSEA.lower    = RMSEA$lower,
    RMSEA.upper    = RMSEA$upper,
    RMSEA.ci.level = RMSEA$ci.level,
    RMSEA.pvalue   = RMSEA$pvalue,
    RMSEA.close.h0 = RMSEA$close.h0
  )
}


fitModsemDA_Internal <- function(...) {
  fit_modsem_da(..., drop.list.single.group = FALSE)
}


calcChiSqr <- function(O, E, N, p, mu, mu.hat) {
  diff_mu <- mu - mu.hat
  Einv    <- solve(E)
  as.vector(
    (N - 1) * (t(diff_mu) %*% Einv %*% diff_mu +
               tr(O %*% Einv) - log(det(O %*% Einv)) - p)
  )
}


tryCatchUniroot <- function(f, lower, upper, errorVal = NA) {
  tryCatch(
    stats::uniroot(f, lower = lower, upper = upper)$root,
    error = function(e) errorVal
  )
}


calcRMSEA <- function(chi.sq, df, N, ci.level = 0.90, close.h0=0.05) {
  alpha  <- 1 - ci.level

  fLower <- \(lambda) stats::pchisq(chi.sq, df, ncp=lambda) - (1 - alpha/2)
  fUpper <- \(lambda) stats::pchisq(chi.sq, df, ncp=lambda) - (    alpha/2)
  fRMSEA <- \(lambda) sqrt(max(lambda, 0) / (df * (N - 1)))

  point <- chi.sq - df
  lower <- tryCatchUniroot(fLower, lower=0, upper=chi.sq, errorVal=0)
  upper <- tryCatchUniroot(fUpper, lower=0, upper=10*chi.sq, errorVal=df * (N - 1)) # i.e., RMSEA.upper = 1

  rmseaLower  <- fRMSEA(lower)
  rmseaUpper  <- fRMSEA(upper)
  rmseaHat    <- fRMSEA(point)
  rmseaPvalue <- 1 - stats::pchisq(chi.sq, df=df, ncp=df*(N-1)*close.h0^2)

  list(rmsea = rmseaHat, lower = rmseaLower, upper = rmseaUpper,
       ci.level = ci.level, pvalue = rmseaPvalue, close.h0 = close.h0)
}


calcAIC <- function(logLik, k) {
  2 * k - 2 * logLik
}


calcAdjAIC <- function(logLik, k, N) {
  AIC <- calcAIC(logLik, k)
  AIC + (2 * k ^ 2 + 2 * k) / (N - k - 1)
}


calcBIC <- function(logLik, k, N) {
  log(N) * k - 2 * logLik
}


calcAdjBIC <- function(logLik, k, N) {
  log((N + 2) / 24) * k - 2 * logLik
}


calcSRMR_Mplus <- function(S, M, Sigma.hat, Mu.hat, mean.structure = TRUE) {
  # Bollen approach: simply using cov2cor ('correlation residuals')
  S.cor <- cov2cor(S)
  Sigma.cor <- cov2cor(Sigma.hat)
  R.cor <- (S.cor - Sigma.cor)
  nvar  <- NCOL(Sigma.cor)

  # meanstructure
  if (mean.structure) {
    # standardized residual mean vector
    R.cor.mean <- M / sqrt(diag(S)) - Mu.hat / sqrt(diag(Sigma.hat))

    e <- nvar * (nvar + 1) / 2 + nvar
    srmr.mplus <-
      sqrt((sum(R.cor[lower.tri(R.cor, diag = FALSE)]^2) +
            sum(R.cor.mean^2) +
            sum(((diag(S) - diag(Sigma.hat)) / diag(S))^2)) / e)

    e <- nvar * (nvar + 1) / 2
    srmr.mplus.nomean <-
      sqrt((sum(R.cor[lower.tri(R.cor, diag = FALSE)]^2) +
            sum(((diag(S) - diag(Sigma.hat)) / diag(S))^2)) / e)
  } else {
    e <- nvar * (nvar + 1) / 2
    srmr.mplus.nomean <- srmr.mplus <-
      sqrt((sum(R.cor[lower.tri(R.cor, diag = FALSE)]^2) +
            sum(((diag(S) - diag(Sigma.hat)) / diag(S))^2)) / e)
  }

  attr(srmr.mplus, "nomean") <- srmr.mplus.nomean
  srmr.mplus
}

#' Estimation latent interactions through \code{Mplus}
#'
#' @param model.syntax lavaan/modsem syntax
#' @param data dataset
#' @param estimator estimator argument passed to \code{Mplus}.
#' @param cluster cluster argument passed to \code{Mplus}.
#' @param categorical categorical argument passed to \code{Mplus}.
#' @param type type argument passed to \code{Mplus}.
#' @param algorithm algorithm argument passed to \code{Mplus}.
#' @param processors processors argument passed to \code{Mplus}.
#' @param integration integration argument passed to \code{Mplus}.
#'
#' @param rcs Should latent variable indicators be replaced with reliability-corrected
#'   single item indicators instead? See \code{\link{relcorr_single_item}}.
#'
#' @param rcs.choose Which latent variables should get their indicators replaced with
#'   reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
#'   as the \code{choose} argument.
#'
#' @param rcs.scale.corrected Should reliability-corrected items be scale-corrected? If \code{TRUE}
#'   reliability-corrected single items are corrected for differences in factor loadings between
#'   the items. Default is \code{TRUE}.
#'
#' @param output.std Should \code{STANDARDIZED} be added to \code{OUTPUT}?
#'
#' @param cleanup Should the \code{Mplus} files (\code{.inp}, \code{.dat} and \code{.out})
#'   created during estimation be deleted when \code{modsem_mplus()} exits? Default is
#'   \code{FALSE}, meaning the files are kept in the working directory.
#'
#' @param ... arguments passed to other functions
#'
#' @return modsem_mplus object
#' @export
#'
#' @examples
#' # Theory Of Planned Behavior
#' m1 <- '
#' # Outer Model
#'   X =~ x1 + x2
#'   Z =~ z1 + z2
#'   Y =~ y1 + y2
#'
#' # Inner model
#'   Y ~ X + Z + X:Z
#' '
#'
#' \dontrun{
#' # Check if Mplus is installed
#' run <- tryCatch({MplusAutomation::detectMplus(); TRUE},
#'                 error = \(e) FALSE)
#'
#' if (run) {
#'   est_mplus <- modsem_mplus(m1, data = oneInt)
#'   summary(est_mplus)
#' }
#' }
#'
modsem_mplus <- function(model.syntax,
                         data,
                         estimator = "ml",
                         cluster = NULL,
                         type = ifelse(is.null(cluster), yes = "random", no = "complex"),
                         algorithm = "integration",
                         processors = 2,
                         integration = 15,
                         rcs = FALSE,
                         rcs.choose = NULL,
                         rcs.scale.corrected = TRUE,
                         output.std = TRUE,
                         categorical = NULL,
                         cleanup = FALSE,
                         ...) {
  if (!requireNamespace("MplusAutomation", quietly = TRUE))
    stop("modsem_mplus() requires the 'MplusAutomation' package; please install it.")

  if (rcs) { # use reliability-correct single items?
    corrected <- relcorr_single_item(
      syntax          = model.syntax,
      data            = data,
      choose          = rcs.choose,
      scale.corrected = rcs.scale.corrected,
      warn.lav        = FALSE
    )

    model.syntax <- corrected$syntax
    data         <- corrected$data
  }

  parTable <- modsemify(model.syntax)

  # Abbreviate variable names
  names <- unique(c(parTable$rhs, parTable$lhs))
  names.xz   <- names[grepl(":", names)]
  names.nlin <- unique(unlist(stringr::str_split(names.xz, pattern = ":")))

  names     <- names[!grepl(":", names)]
  names.lin <- setdiff(names, names.nlin)

  # we need some part of both x and z to be available in the intTerm name
  # so variables in int terms must be less than width = 8
  abbreviated.lin  <- abbreviate(names.lin, minlength = 8L, strict = TRUE)
  abbreviated.nlin <- abbreviate(names.nlin, minlength = 6L, strict = TRUE)
  abbreviated <- c(abbreviated.lin, abbreviated.nlin)

  # Abbreviate names in intTerms
  intTerms <- unique(parTable$rhs[grepl(":", parTable$rhs)])
  abbrevIntTerm <- function(xz)
    paste0(abbreviated[stringr::str_split(xz, pattern = ":")[[1]]], collapse = ":")

  newIntTerms <- vapply(intTerms, FUN.VALUE = character(1L), FUN = abbrevIntTerm)
  names(newIntTerms) <- intTerms
  abbreviated <- c(abbreviated, newIntTerms)

  # Fix names in parTable
  rmask <- parTable$rhs != "" & !parTable$op %in% c(":=", "==", "<", ">")
  lmask <- parTable$lhs != ""
  parTable$rhs[rmask] <- abbreviated[parTable$rhs[rmask]]
  parTable$lhs[lmask] <- abbreviated[parTable$lhs[lmask]]

  # Fix names in data
  data  <- as.data.frame(data)
  catCols <- colnames(data)[sapply(data, is.factor)]
  categorical <- intersect(union(catCols, categorical), colnames(data))

  dmask <- colnames(data) %in% names(abbreviated)
  dmaskCat <- categorical %in% names(abbreviated)

  colnames(data)[dmask] <- abbreviated[colnames(data)[dmask]]
  categorical[dmaskCat] <- abbreviated[categorical[dmaskCat]]

  indicators <- unique(parTable[parTable$op == "=~", "rhs", drop = TRUE])
  intTerms <- unique(getIntTermRows(parTable)$rhs)
  intTermsMplus <- stringr::str_remove_all(intTerms, ":") |>
    stringr::str_to_upper()

  # Set OUTPUT
  OUTPUT <- "TECH1;\nTECH3;"

  getIntTermLength <- \(xz) length(stringr::str_split(xz, pattern = ":")[[1]])
  kway <- max(0, vapply(intTerms, FUN.VALUE = integer(1L), FUN = getIntTermLength))

  if (output.std && kway <= 2)
    OUTPUT <- paste(OUTPUT, "STANDARDIZED;", sep = "\n")

  # Cluster
  if (!is.null(cluster)) {
    VARIABLE <- sprintf("CLUSTER = %s;", cluster)

  } else VARIABLE <- NULL

  # Categorical variables
  categorical <- intersect(categorical, indicators)
  if (length(categorical)) {
    VARIABLE <- paste0(
      VARIABLE, sprintf("\nCATEGORICAL = %s;",
                        paste0(categorical, collapse = " "))
    )
  }

  usevariables <- intersect(c(cluster, indicators), colnames(data))

  # Estimate model
  model <- MplusAutomation::mplusObject(
    TITLE = "Running Model via Mplus",
    VARIABLE = VARIABLE,
    usevariables = usevariables,
    ANALYSIS = paste0(
      paste(paste("estimator =", estimator),
            paste("type =", type),
            paste("algorithm =", algorithm),
            paste("processors =", processors),
            paste("integration = ", integration),
            sep = ";\n"), ";\n"), # add final ";"
    MODEL = parTableToMplusModel(parTable, ignoreConstraints = TRUE, ...),
    MODELCONSTRAINT = parTableToMplusModelConstraints(parTable, ...),
    OUTPUT = OUTPUT,
    rdata = data[usevariables],
  )

  fprefix <- getMplusFilePrefix()
  if (cleanup)
    on.exit(cleanupMplusFiles(fprefix), add = TRUE)

  results <- MplusAutomation::mplusModeler(
    model,
    modelout = paste0(fprefix, ".inp"),
    run = 1L,
    writeData = "always",
    hashfilename = FALSE
  )

  coefsTable    <- coef(results)
  mplusParTable <- mplusTableToParTable(
    coefsTable,
    intTerms = intTerms,
    intTermsMplus = intTermsMplus,
    indicators = indicators,
    parTable.in = parTable
  )

  # coef and vcov
  TECH1 <- MplusAutomation::get_results(results, element = "tech1")
  pars.tech1 <- getOrderedParameterLabelsMplus(mplusParTable,
                                               TECH1 = TECH1,
                                               intTerms = intTerms,
                                               intTermsMplus = intTermsMplus)
  labels.tech1 <- names(pars.tech1)

  isfree <- coefsTable$pval != 999
  labels <- stringr::str_remove_all(coefsTable$Label[isfree], pattern = " ")
  coefs  <- structure(coefsTable$est[isfree], names = labels)
  coefs  <- coefs[labels.tech1[labels.tech1 %in% labels]]

  vcov <- tryCatch({
    # vcov <- MplusAutomation::get_tech3(results)$paramCov
    # `MplusAutomation::get_tech3` seems very unstable...
    # I often get this error:
    # `Error in get_results(x = results, element = "tech3"): could not find function "get_results"`
    # `MplusAutomation::get_results` seems to work better...
    vcov <- MplusAutomation::get_results(results, element = "tech3")$paramCov
    vcov[upper.tri(vcov)] <- t(vcov)[upper.tri(vcov)]
    dimnames(vcov) <- list(labels.tech1, labels.tech1)
    vcov

  }, error = function(e) {
    mod_msg_warn(paste0("Unable to retrive `tech3` from `Mplus` results\n",
             "Message: ", e))

    k <- length(pars.tech1)
    vcov <- matrix(NA, nrow = k, ncol = k,
                   dimnames = list(labels.tech1, labels.tech1))
  })

  std <- tryCatch({
    MplusAutomation::get_results(results, element = "standardized")
  }, error = function(e) {
    mod_msg_warn(paste0("Unable to retrive `standardized` from `Mplus` results\n",
             "Message: ", e))
    std <- NULL
  })

  modelSpec <- list(
    parTable = mplusParTable,
    model    = results,
    data     = data,
    coefs    = modsemVector(coefs),
    vcov     = modsemMatrix(vcov, symmetric = TRUE),
    std      = std,
    info     = list(indicators    = indicators,
                    intTerms      = intTerms,
                    intTermsMplus = intTermsMplus,
                    pars.tech1    = pars.tech1)
  )

  structure(modelSpec,
            class = "modsem_mplus",
            method = "Mplus")
}


xwith <- function(elems) {
  if (length(elems) <= 1) return(NULL)
  lhs <- paste0(elems[[1]], elems[[2]])
  rhs <- paste(elems[[1]], "XWITH", elems[[2]])

  elems2 <- c(lhs, elems[-c(1, 2)])

  unique(rbind(
   createParTableRow(c(lhs, rhs), op = ":"),
   xwith(elems2)
  ))
}


parTableToMplusModel <- function(parTable, ignoreConstraints = FALSE) {
  if (ignoreConstraints) {
    parTable <- parTable[!parTable$op %in% c(":=", "==", "<", ">"),
                         , drop = FALSE]
  }

  # Interaction expressions
  interactions <- parTable[grepl(":", parTable$rhs), "rhs"]
  elemsInInts <- stringr::str_split(interactions, ":")
  newRows <- lapply(elemsInInts, FUN = xwith) |>
    purrr::list_rbind()
  parTable <- rbind(parTable, newRows)

  # Identify variances
  isVar <- parTable$op == "~~" & parTable$lhs == parTable$rhs

  parTable$op <- replaceLavOpWithMplus(parTable$op)
  parTable[isVar, c("op", "lhs")] <- ""

  out <- ""
  for (i in seq_len(nrow(parTable))) {
    mod   <- parTable[["mod"]][i]
    empty <- mod == ""
    islab <- !canBeNumeric(mod)

    if (!empty && islab)       modifier <- paste0(" (", mod,")")
    else if (!empty && !islab) modifier <- paste0("@", mod)
    else                       modifier <- ""

    line <- paste0(parTable[["lhs"]][[i]], " ",
                   parTable[["op"]][[i]], " ",
                   parTable[["rhs"]][[i]],
                   modifier, ";" ,"\n")
    out <- paste0(out, line)
  }
  stringr::str_remove_all(out, ":")
}


parTableToMplusModelConstraints <- function(parTable) {
  constraints <- parTable[parTable$op %in% c(":=", "==", "<", ">"),
                          , drop = FALSE]

  if (!NROW(constraints))
    return(NULL)

  out <- ""
  for (i in seq_len(NROW(constraints))) {
    lhs <- constraints$lhs[i]
    op  <- constraints$op[i]
    rhs <- constraints$rhs[i]

    if (op == ":=")
      out <- paste0(out, "NEW(", lhs, ");\n")

    mop <- ifelse(op %in% c(":=", "=="), yes = "=", no = op)
    out <- paste0(out, sprintf("%s %s %s;\n", lhs, mop, rhs))
  }

  out
}


replaceLavOpWithMplus <- function(op) {
  vapply(op,
         FUN = switchLavOpToMplus,
         FUN.VALUE = character(1L))
}


switchLavOpToMplus <- function(op) {
  switch(op,
         "=~" = "BY",
         "~" = "ON",
         "~~" = "WITH",
         ":" = "|",
         mod_msg_stop(paste0("Operator not supported for use in Mplus: ", op, "\n")))
}


mplusTableToParTable <- function(coefsTable,
                                 indicators,
                                 intTerms,
                                 intTermsMplus,
                                 parTable.in = NULL) {

  coefsTable <- rename(coefsTable, Label = "label",
                       se = "std.error", pval = "p.value")
  coefsTable$label <- stringr::str_remove_all(coefsTable$label, pattern = " ")

  indicatorsCaps <- stringr::str_to_upper(indicators)
  patternMeas <-
    paste0("(", stringr::str_c(indicatorsCaps, collapse = "|"), ")") |>
    paste0("<-(?!>|(Intercept|Thresholds))")
  measCoefNames <- grepl(patternMeas, coefsTable$label, perl = TRUE)

  # Mplus has lhs/rhs in reversed order for the measurement model,
  # compared to lavaan,
  measRhs <- stringr::str_split_i(coefsTable$label[measCoefNames],
                                  "<-", i = 1)
  measLhs <- stringr::str_split_i(coefsTable$label[measCoefNames],
                                  "<-", i = 2)
  measModel <- data.frame0(lhs = measLhs, op = "=~", rhs = measRhs) |>
    cbind0(coefsTable[measCoefNames, ])

  # Structural Model
  sub <- coefsTable[!measCoefNames, , drop = FALSE]
  patternStruct <- "<-(?!>|(Intercept|Thresholds))"
  structCoefNames <- grepl(patternStruct, sub$label, perl = TRUE)

  structLhs <- stringr::str_split_i(sub$label[structCoefNames],
                                    "<-", i = 1)
  structRhs <- stringr::str_split_i(sub$label[structCoefNames],
                                    "<-", i = 2)
  structModel <- data.frame0(lhs = structLhs, op = "~", rhs = structRhs) |>
    cbind0(sub[structCoefNames, ])

  maxCharMplus <- maxchar(c(structModel$rhs, structModel$lhs))
  for (i in seq_along(intTerms)) {
    xzMplus  <- substr(intTermsMplus[[i]], start = 1, stop = maxCharMplus)
    xzModsem <- intTerms[[i]]

    structModel[structModel$rhs == xzMplus, "rhs"] <- xzModsem
    structModel[structModel$lhs == xzMplus, "lhs"] <- xzModsem
  }

  # Variances and Covariances
  sub <- sub[!structCoefNames, , drop = FALSE]
  patternCovVar <- "<->"
  covVarCoefNames <- grepl(patternCovVar, sub$label, perl = TRUE)
  covVarLhs <- stringr::str_split_i(sub$label[covVarCoefNames],
                                    "<->", i = 1)
  covVarRhs <- stringr::str_split_i(sub$label[covVarCoefNames],
                                    "<->", i = 2)
  covVarModel <- data.frame0(lhs = covVarLhs, op = "~~", rhs = covVarRhs) |>
    cbind0(sub[covVarCoefNames, ])

  # Intercepts
  sub <- sub[!covVarCoefNames, , drop = FALSE]
  patternIntercept <- "<-Intercept"
  interceptNames <- grepl(patternIntercept, sub$label, perl = TRUE)
  interceptLhs <- stringr::str_split_i(sub$label[interceptNames],
                                       "<-", i = 1)
  interceptModel <- data.frame0(lhs = interceptLhs, op = "~1", rhs = "") |>
    cbind0(sub[interceptNames, ])

  # Thresholds
  sub <- sub[!interceptNames, , drop = FALSE]
  patternThreshold <- "<-Thresholds"
  thresholdNames <- grepl(patternThreshold, sub$label, perl = TRUE)
  thresholdLhsPart <- stringr::str_split_i(sub$label[thresholdNames],
                                       "<-", i = 1)

  thresholdLhsSplit <- stringr::str_split_fixed(thresholdLhsPart, pattern = "\\$", n = 2)
  thresholdLhs <- thresholdLhsSplit[,1L]
  thresholdRhs <- paste0("t", thresholdLhsSplit[,2L])

  thresholdModel <- data.frame0(lhs = thresholdLhs, op = "|", rhs = thresholdRhs) |>
    cbind0(sub[thresholdNames, ])

  # Custom / Remaining
  sub <- sub[!thresholdNames, , drop = FALSE]

  if (NROW(sub)) {
    customModel <- cbind0(data.frame0(lhs = sub$label, op = ":=", rhs = ""), sub)

    if (!is.null(parTable.in)) {
      lrCustom <- parTable.in[parTable.in$op == ":=", c("lhs", "rhs"), drop = FALSE]
      lrCustom$lhs <- toupper(lrCustom$lhs)

      notRhs   <- colnames(customModel) != "rhs"
      customModel <- merge(x = customModel[notRhs],
                           y = lrCustom, by = "lhs",
                           all.x = TRUE, all.y = FALSE)
    }

  } else customModel <- NULL

  # Combine
  mplusParTable <- rbind(measModel, structModel, covVarModel, interceptModel,
                         thresholdModel, customModel)
  mplusParTable [c("lhs", "rhs")] <-
    lapplyDf(mplusParTable[c("lhs", "rhs")], function(x)
             stringr::str_remove_all(x, " "))

  if (!is.null(parTable.in) && any(parTable.in$mod != "")) {
    LABELS <- parTable.in[parTable.in$mod != "", , drop = FALSE]
    LABELS <- rename(LABELS[c("lhs", "op", "rhs", "mod")], mod = "label.lav")

    mplusParTable$order <- seq_len(NROW(mplusParTable))
    mplusParTable <- merge(x = mplusParTable,
                           y = LABELS,
                           by = c("lhs", "op", "rhs"),
                           all.x = TRUE, all.y = FALSE)

    match <- !is.na(mplusParTable$label.lav)
    mplusParTable[match, "label"] <- mplusParTable[match, "label.lav"]

    # clean up
    mplusParTable <- mplusParTable[colnames(mplusParTable) != "label.lav"]
    mplusParTable <- mplusParTable[order(mplusParTable$order), , drop = FALSE]
  }

  mplusParTable$ci.lower <- mplusParTable$est - CI_WIDTH * mplusParTable$std.error
  mplusParTable$ci.upper <- mplusParTable$est + CI_WIDTH * mplusParTable$std.error
  mplusParTable$p.value[mplusParTable$p.value == 999] <- NA
  mplusParTable$z.value <- mplusParTable$est / mplusParTable$std.error
  mplusParTable$z.value[is.infinite(mplusParTable$z.value)] <- NA

  modsemParTable(mplusParTable)
}


getOrderedParameterLabelsMplus <- function(parTable, TECH1, intTerms, intTermsMplus) {
  spec   <- TECH1$parameterSpecification

  CPAR <- "THE.ADDITIONAL.PARAMETERS"
  if (CPAR %in% names(spec)) custom <- spec[[CPAR]]$new_additional
  else                       custom <- NULL

  if ("X" %in% names(spec)) specX <- spec$X
  else                      specX <- spec

  tau    <- specX$tau
  nu     <- specX$nu
  alpha  <- specX$alpha
  lambda <- specX$lambda
  beta   <- specX$beta
  psi    <- specX$psi
  theta  <- specX$theta

  setLabel <- function(out, label, id) {
    if (!length(label)) {
      mod_msg_warn(paste0("Unable to find label for parameter ", id, "!"))
      out[as.character(id)] <- id

    } else out[label] <- id

    out
  }

  getReg <- function(M, row.lhs = TRUE, op = "~", try.op = NULL) {
    if (is.null(M))
      return(NULL)

    out <- c()

    rows <- rownames(M)
    cols <- colnames(M)
    for (i in seq_len(NROW(M))) for (j in seq_len(NCOL(M))) {
      id <- as.integer(M[i, j])

      if (is.na(id) || id <= 0L)
        next

      if (row.lhs) {
        lhs <- rows[i]
        rhs <- cols[j]
      } else {
        lhs <- cols[j]
        rhs <- rows[i]
      }

      label <- parTable[parTable$op == op &
                        parTable$lhs == lhs &
                        parTable$rhs == rhs, "label"]

      if (!length(label) && !is.null(try.op)) {
        # can either be =~, or ~, either way we must also
        # switch from order of lhs and rhs
        label <- parTable[parTable$op == try.op &
                          parTable$lhs == rhs &
                          parTable$rhs == lhs, "label"]
      }

      out <- setLabel(out = out, label = label, id = id)
    }

    out
  }

  getThreshold <- function(T) {
    if (is.null(T))
      return(NULL)

    vars <- colnames(T)
    split <- stringr::str_split_fixed(vars, pattern = "\\$", n = 2)

    T <- c(T)
    out <- c()

    for (i in seq_along(T)) {
      id <- as.integer(T[i])

      if (is.na(id) || id <= 0L)
        next

      lhs <- split[i, 1]
      rhs <- paste0("t", split[i, 2])

      label <- parTable[parTable$op == "|" &
                        parTable$lhs == lhs &
                        parTable$rhs == rhs, "label"]

      out <- setLabel(out = out, label = label, id = id)
    }

    out
  }

  getIntercept <- function(T) {
    if (is.null(T))
      return(NULL)

    vars <- colnames(T)
    T <- c(T)
    out <- c()

    for (i in seq_along(T)) {
      id <- as.integer(T[i])

      if (is.na(id) || id <= 0L)
        next

      lhs <- vars[i]

      label <- parTable[parTable$op == "~1" &
                        parTable$lhs == lhs, "label"]

      out <- setLabel(out = out, label = label, id = id)
    }

    out
  }


  getCovariance <- function(M, op = "~~") {
    if (is.null(M))
      return(NULL)

    out <- c()

    rows <- rownames(M)
    cols <- colnames(M)

    for (i in seq_len(NROW(M))) for (j in seq_len(i)) {
      id <- as.integer(M[i, j])

      if (is.na(id) || id <= 0L)
        next

      lhs <- cols[i]
      rhs <- rows[j]

      label <- parTable[parTable$op == op &
                        ((parTable$lhs == lhs & parTable$rhs == rhs) |
                         (parTable$lhs == rhs & parTable$rhs == lhs)), "label"]

      out <- setLabel(out = out, label = label, id = id)
    }

    out
  }

  getCustom <- function(M, op = ":=") {
    if (is.null(M) || NROW(M) == 0L) return(NULL)

    # assumes M has nrows(M)
    mod_warnif(NROW(M) > 1L,
           paste0("Expected parameter matrix for additional pars\n",
           "to have a single row!"))

    cols <- colnames(M)
    out  <- c()

    for (i in seq_len(NCOL(M))) {
      id <- as.integer(M[1L, i])

      if (is.na(id) || id <= 0L)
        next

      lhs <- cols[i]

      label <- parTable[parTable$op == op & parTable$lhs == lhs, "lhs"]
      out <- setLabel(out = out, label = label, id = id)
    }

    out
  }

  for (i in seq_along(intTerms)) {
    xzMplus  <- substr(intTermsMplus[[i]], start = 1L, stop = 8L)
    xzModsem <- intTerms[[i]]

    parTable[parTable$lhs == xzModsem, "lhs"] <-  xzMplus
    parTable[parTable$rhs == xzModsem, "rhs"] <-  xzMplus
  }

  out <- c(
    getThreshold(tau),
    getIntercept(nu),
    getIntercept(alpha),
    getReg(lambda, row.lhs = FALSE, op = "=~"),
    getReg(beta, row.lhs = TRUE, op = "~", try.op = "=~"), # include "=~" for higher order models
    getCovariance(psi),
    getCovariance(theta),
    getCustom(custom)
  )

  out <- out[!duplicated(out)] # unique() removes labels
  sort(out)
}


any0 <- function(x) {
  any(vapply(x, FUN.VALUE = logical(1L), FUN = \(x) length(x) <= 0))
}


data.frame0 <- function(...) {
  args <- list(...)

  if (any0(args))
    return(NULL)

  data.frame(...)
}


cbind0 <- function(...) {
  args <- list(...)

  if (any0(args))
    return(NULL)

  cbind(...)
}


mplusFilePrefixExists <- function(fprefix) {
  inp <- paste0(fprefix, ".inp")
  dat <- paste0(fprefix, ".dat")
  out <- paste0(fprefix, ".out")
  file.exists(inp) || file.exists(dat) || file.exists(out)
}


cleanupMplusFiles <- function(fprefix) {
  files <- paste0(fprefix, c(".inp", ".dat", ".out"))
  unlink(files[file.exists(files)])
}


getMplusFilePrefix <- function(max.iter = 20, base = "mpresults") {
  randid <- generateRandomCharId(n=12)
  fprefix <- paste0(base, randid)

  if (!mplusFilePrefixExists(fprefix))
    return(fprefix)

  for (i in seq_len(max.iter)) {
    id <- generateRandomCharId(2)
    fprefix <- paste0(fprefix, id)

    if (!mplusFilePrefixExists(fprefix))
      return(fprefix)
  }

  mod_msg_warn(
    "Unable to create a unique name for Mplus files!",
    "Previous results might get overwritten..."
  )

  fprefix
}


generateRandomCharId <- function(n = 36) {
  # use only lower letters, as Mplus sometimes seem to convert upper case
  # characters in the .inp file to lower characters in the .out file. In
  # particular this seems to happen on Mplus 8.11 on Windows
  chars <- c(letters, as.character(0:9))
  paste0(sample(chars, size = n, replace = TRUE), collapse = "")
}

#' Plot Interaction Effects in a SEM Model
#'
#' This function creates an interaction plot of the outcome variable (\code{y}) as a function
#' of a focal predictor (\code{x}) at multiple values of a moderator (\code{z}). It is
#' designed for use with structural equation modeling (SEM) objects (e.g., those from
#' \code{\link{modsem}}). Predicted means (or predicted individual values) are calculated
#' via \code{\link{simple_slopes}}, and then plotted with \code{ggplot2} to display
#' multiple regression lines and confidence/prediction bands.
#'
#' @param x A character string specifying the focal predictor (x-axis variable).
#' @param z A character string specifying the moderator variable.
#' @param y A character string specifying the outcome (dependent) variable.
#' @param vals_x A numeric vector of values at which to compute and plot the focal
#'   predictor \code{x}. The default is \code{seq(-3, 3, .001)}, which provides a
#'   relatively fine grid for smooth lines. If \code{rescale=TRUE}, these values
#'   are in standardized (mean-centered and scaled) units, and will be converted back
#'   to the original metric in the internal computation of predicted means.
#' @param vals_z A numeric vector of values of the moderator \code{z} at which to draw
#'   separate regression lines. Each distinct value in \code{vals_z} defines a separate
#'   group (plotted with a different color). If \code{rescale=TRUE}, these values
#'   are also assumed to be in standardized units.
#' @param model An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}},
#'   \code{\link{modsem_mplus}}, or possibly a \code{lavaan} object. Must be a fitted
#'   SEM model containing paths for \code{y ~ x + z + x:z}.
#' @param alpha_se A numeric value in \eqn{[0, 1]} specifying the transparency of
#'   the confidence/prediction interval ribbon. Default is \code{0.15}.
#' @param digits An integer specifying the number of decimal places to which the
#'   moderator values (\code{z}) are rounded for labeling/grouping in the plot.
#' @param ci_width A numeric value in \eqn{(0,1)} indicating the coverage of the
#'   confidence (or prediction) interval. The default is \code{0.95} for a 95\%
#'   interval.
#' @param ci_type A character string specifying whether to compute
#'   \code{"confidence"} intervals (for the mean of the predicted values, default)
#'   or \code{"prediction"} intervals (which include residual variance).
#' @param rescale Logical. If \code{TRUE} (default), \code{vals_x} and \code{vals_z}
#'   are interpreted as standardized units, which are mapped back to the raw scale
#'   before computing predictions. If \code{FALSE}, \code{vals_x} and \code{vals_z}
#'   are taken as raw-scale values directly.
#' @param standardized Should coefficients be standardized beforehand?
#' @param xz A character string specifying the interaction term (\code{x:z}).
#'   If \code{NULL}, the term is created automatically as \code{paste(x, z, sep = ":")}.
#'   Some SEM backends may handle the interaction term differently (for instance, by
#'   removing or modifying the colon), and this function attempts to reconcile that
#'   internally.
#' @param greyscale Logical. If \code{TRUE} the plot is plotted in greyscale.
#' @param ... Additional arguments passed on to \code{\link{simple_slopes}}.
#'
#' @details
#' \strong{Computation Steps:}
#' \enumerate{
#'   \item Calls \code{\link{simple_slopes}} to compute the predicted values of \code{y}
#'         at the specified grid of \code{x} and \code{z} values.
#'   \item Groups the resulting predictions by unique \code{z}-values (rounded to
#'         \code{digits}) to create colored lines.
#'   \item Plots these lines using \code{ggplot2}, adding ribbons for confidence
#'         (or prediction) intervals, with transparency controlled by \code{alpha_se}.
#' }
#'
#' \strong{Interpretation:}
#' Each line in the plot corresponds to the regression of \code{y} on \code{x} at
#' a given level of \code{z}. The shaded region around each line (ribbon) shows
#' either the confidence interval for the predicted mean (if \code{ci_type =
#' "confidence"}) or the prediction interval for individual observations (if
#' \code{ci_type = "prediction"}). Where the ribbons do not overlap, there is
#' evidence that the lines differ significantly over that range of \code{x}.
#'
#' @return A \code{ggplot} object that can be further customized (e.g., with
#'   additional \code{+ theme(...)} layers). By default, it shows lines for each
#'   moderator value and a shaded region corresponding to the interval type
#'   (confidence or prediction).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' library(modsem)
#'
#' # Example 1: Interaction of X and Z in a simple SEM
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' "
#' est1 <- modsem(m1, data = oneInt)
#'
#' # Plot interaction using a moderate range of X and two values of Z
#' plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z",
#'                  vals_x = -3:3, vals_z = c(-0.2, 0), model = est1)
#'
#' # Example 2: Interaction in a theory-of-planned-behavior-style model
#' tpb <- "
#' # Outer Model
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN  =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC
#'   BEH ~ PBC:INT
#' "
#' est2 <- modsem(tpb, data = TPB, method = "lms", nodes = 32)
#'
#' # Plot with custom Z values and a denser X grid
#' plot_interaction(x = "INT", z = "PBC", y = "BEH",
#'                  xz = "PBC:INT",
#'                  vals_x = seq(-3, 3, 0.01),
#'                  vals_z = c(-0.5, 0.5),
#'                  model = est2)
#' }
plot_interaction <- function(x, z, y, model, vals_x = seq(-3, 3, .001),
                             vals_z, alpha_se = 0.15, digits = 2,
                             ci_width = 0.95, ci_type = "confidence",
                             rescale = TRUE, standardized = FALSE, xz = NULL,
                             greyscale = FALSE,
                             ...) {
  slopes <- simple_slopes(x = x, z = z, y = y, model = model, vals_x = vals_x,
                          vals_z = vals_z, rescale = rescale, ci_width = ci_width,
                          ci_type = ci_type, standardized = standardized, xz = xz, ...)
  df <- as.data.frame(slopes)
  df$cat_z <- as.factor(round(df$vals_z, digits))

  # declare within the scope, to not get notes in R CMD check
  std.error <- df$std.error
  predicted <- df$predicted
  cat_z     <- df$cat_z
  vals_x    <- df$vals_x
  ci.lower  <- df$ci.lower
  ci.upper  <- df$ci.upper

  # plotting margins
  p <- ggplot2::ggplot(df, ggplot2::aes(x = vals_x, y = predicted, colour = cat_z, group = cat_z)) +
    ggplot2::geom_line() +
    ggplot2::geom_ribbon(ggplot2::aes(ymin = ci.lower, ymax = ci.upper, fill = cat_z),
                         alpha = alpha_se, linewidth = 0, linetype = "blank") +
    ggplot2::labs(x = x, y = y, colour = z, fill = z) +
    ggplot2::ggtitle(sprintf("Marginal Effects of %s on %s, Given %s", x, y, z)) +
    ggplot2::theme_bw()

  if (length(unique(df$group)) > 1L) {
    group <- NULL # stop R CMD check from complaining about `~group`
    p <- p + ggplot2::facet_wrap(~group)
  }

  if (greyscale)
    p <- suppressMessages(p + ggplot2::scale_colour_grey() + ggplot2::scale_fill_grey())

  p
}



#' Plot Surface for Interaction Effects
#'
#' Generates a 3D surface plot to visualize the interaction effect of two variables (\code{x} and \code{z})
#' on an outcome (\code{y})
#' using parameter estimates from a supported model object (e.g., \code{lavaan} or \code{modsem}).
#' The function allows specifying ranges for \code{x} and \code{z} in standardized z-scores, which are then transformed
#' back to the original scale based on their means and standard deviations.
#'
#' @param x A character string specifying the name of the first predictor variable.
#' @param z A character string specifying the name of the second predictor variable.
#' @param y A character string specifying the name of the outcome variable.
#'
#' @param model A model object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, \code{\link{modsem_mplus}}, or \code{lavaan}. The model should
#'   include paths for the predictors (\code{x}, \code{z}, and \code{xz}) to the outcome (\code{y}).
#'
#' @param min_x Numeric. Minimum value of \code{x} in z-scores. Default is -3.
#' @param max_x Numeric. Maximum value of \code{x} in z-scores. Default is 3.
#' @param min_z Numeric. Minimum value of \code{z} in z-scores. Default is -3.
#' @param max_z Numeric. Maximum value of \code{z} in z-scores. Default is 3.
#'
#' @param standardized Should coefficients be standardized beforehand?
#'
#' @param detail Numeric. Step size for the grid of \code{x} and \code{z} values, determining the resolution of the surface.
#'   Smaller values increase plot resolution. Default is \code{1e-2}.
#'
#' @param xz Optional. A character string or vector specifying the interaction term between \code{x} and \code{z}.
#'   If \code{NULL}, the interaction term is constructed as \code{paste(x, z, sep = ":")} and adjusted for specific model classes.
#'
#' @param colorscale Character or list. Colorscale used to color the surface.
#'   - Default is \code{"Viridis"}, which matches the classic Plotly default.
#'   - Can be a built-in palette name (e.g., \code{"Greys"}, \code{"Plasma"}, \code{"Turbo"}),
#'     or a custom two-column list with numeric stops (\code{0}–\code{1}) and color codes.
#'   Example custom scale:
#'   \code{list(c(0, "white"), c(1, "black"))} for a black-and-white gradient.
#'
#' @param reversescale Logical. If \code{TRUE}, reverses the color mapping so that
#'   low values become high colors and vice versa.
#'   Default is \code{FALSE}.
#'
#' @param showscale Logical. If \code{TRUE}, displays the colorbar legend
#'   alongside the plot.
#'   Default is \code{TRUE}.
#'
#' @param cmin Numeric or \code{NULL}. The minimum value for the colorscale mapping.
#'   If \code{NULL}, the minimum of the data (\code{proj_y}) is used automatically.
#'   Use this to standardize the color range across multiple plots.
#'
#' @param cmax Numeric or \code{NULL}. The maximum value for the colorscale mapping.
#'   If \code{NULL}, the maximum of the data (\code{proj_y}) is used automatically.
#'   Use this to standardize the color range across multiple plots.
#'
#' @param surface_opacity Numeric (0–1). Controls the opacity of the surface.
#'   - \code{1} = fully opaque (default)
#'   - \code{0} = fully transparent
#'   Useful when overlaying multiple surfaces or highlighting gridlines.
#'
#' @param grid Logical. If \code{TRUE}, draws gridlines (wireframe)
#'   directly on the surface using Plotly's contour features.
#'   Default is \code{FALSE}.
#'
#' @param grid_nx Integer. Approximate number of gridlines to draw along the
#'   **x-axis** direction when \code{grid = TRUE}.
#'   Higher values create a denser grid.
#'   Default is \code{12}.
#'
#' @param grid_ny Integer. Approximate number of gridlines to draw along the
#'   **y-axis** direction when \code{grid = TRUE}.
#'   Higher values create a denser grid.
#'   Default is \code{12}.
#'
#' @param grid_color Character. Color of the gridlines drawn on the surface.
#'   Must be a valid CSS color string, including \code{rgba()} for transparency.
#'   - Default is \code{"rgba(0,0,0,0.45)"} (semi-transparent black).
#'   Example: \code{"rgba(255,255,255,0.8)"} for semi-transparent white lines.
#'
#' @param group Which group to create surface plot for. Only relevant for multigroup
#'   models. Must be an integer index, representing the nth group.
#'
#' @param ... Additional arguments passed to \code{plotly::plot_ly}.
#'
#' @details
#' The input \code{min_x}, \code{max_x}, \code{min_z}, and \code{max_z} define the range of \code{x} and \code{z} values in z-scores.
#' These are scaled by the standard deviations and shifted by the means of the respective variables, obtained
#' from the model parameter table. The resulting surface shows the predicted values of \code{y} over the grid of \code{x} and \code{z}.
#'
#' The function supports models of class \code{modsem} (with subclasses \code{modsem_pi}, \code{modsem_da}, \code{modsem_mplus}) and \code{lavaan}.
#' For \code{lavaan} models, it is not designed for multigroup models, and a warning will be issued if multiple groups are detected.
#'
#' @return A \code{plotly} surface plot object displaying the predicted values of \code{y} across the grid of \code{x} and \code{z} values.
#'   The color bar shows the values of \code{y}.
#'
#' @note
#' The interaction term (\code{xz}) may need to be manually specified for some models. For non-\code{lavaan} models,
#' interaction terms may have their separator (\code{:}) removed based on circumstances.
#'
#' @examples
#' m1 <- "
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#' # Inner model
#'   Y ~ X + Z + X:Z
#' "
#' est1 <- modsem(m1, data = oneInt)
#' plot_surface("X", "Z", "Y", model = est1)
#'
#' \dontrun{
#' tpb <- "
#' # Outer Model (Based on Hagger et al., 2007)
#'   ATT =~ att1 + att2 + att3 + att4 + att5
#'   SN =~ sn1 + sn2
#'   PBC =~ pbc1 + pbc2 + pbc3
#'   INT =~ int1 + int2 + int3
#'   BEH =~ b1 + b2
#'
#' # Inner Model (Based on Steinmetz et al., 2011)
#'   INT ~ ATT + SN + PBC
#'   BEH ~ INT + PBC
#'   BEH ~ PBC:INT
#' "
#'
#' est2 <- modsem(tpb, TPB, method = "lms", nodes = 32)
#' plot_surface(x = "INT", z = "PBC", y = "BEH", model = est2)
#' }
#'
#' @export
plot_surface <- function(x, z, y, model,
                         min_x = -3,
                         max_x = 3,
                         min_z = -3,
                         max_z = 3,
                         standardized = FALSE,
                         detail = 1e-2,
                         xz = NULL,
                         colorscale = "Viridis",
                         reversescale = FALSE,
                         showscale = TRUE,
                         cmin = NULL,
                         cmax = NULL,
                         surface_opacity = 1,
                         grid = FALSE,
                         grid_nx = 12,
                         grid_ny = 12,
                         grid_color = "rgba(0,0,0,0.45)",
                         group = NULL,
                         ...) {

  mod_stopif(!isModsemObject(model) && !isLavaanObject(model), paste0("model must be of class ",
         "'modsem_pi', 'modsem_da', 'modsem_mplus' or 'lavaan'"))

  if (standardized) {
    parTable <- standardized_estimates(model, correction = TRUE)
  } else {
    parTable <- parameter_estimates(
      object = model,
      label.renamed.prod = TRUE
    )
  }

  parTable <- addMissingGroups(parTable)
  groups   <- getGroupsParTable(parTable)

  if (length(groups) > 1L && is.null(group)) {
    mod_msg_warn(paste0("Plotting of surface plots for multiple groups is not implemented yet!\n",
             "You can choose which group to plot using the `group` argument.\n",
             "Plotting surface plot for the first group..."))
  }

  if (is.null(group))
    group <- 1L

  parTable <- parTable[parTable$group == group, , drop = FALSE]

  if (is.null(xz))
    xz <- paste(x, z, sep = ":")

  checkInputsSimpleSlopes(x = x, z = z, y = y, xz = xz, parTable = parTable)

  xz <- c(xz, reverseIntTerm(xz))

  xx <- paste(x, x, sep = ":")
  xx <- c(xx, reverseIntTerm(xx))

  zz <- paste(z, z, sep = ":")
  zz <- c(zz, reverseIntTerm(zz))

  if (!inherits(model, c("modsem_da", "modsem_mplus")) &&
      !isLavaanObject(model)) {
    xz <- stringr::str_remove_all(xz, ":")
    xx <- stringr::str_remove_all(xx, ":")
    zz <- stringr::str_remove_all(zz, ":")
  }

  if (isLavaanObject(model)) {
    nobs <- unlist(model@Data@nobs)
    mod_warnif(length(nobs) > 1, "plot_interaction is not intended for multigroup models")
    n <- nobs[[1]]
  } else {
    n <- nrow(model$data)
  }

  lVs <- c(x, z, y, xz, xx, zz)
  coefs <- parTable[parTable$op == "~" & parTable$rhs %in% lVs & parTable$lhs == y, ]
  gamma_x  <- coefs[coefs$rhs == x, "est"]
  sd_x     <- sqrt(calcCovParTable(x, x, parTable))
  gamma_z  <- coefs[coefs$rhs == z, "est"]
  sd_z     <- sqrt(calcCovParTable(z, z, parTable))
  gamma_xz <- coefs[coefs$rhs %in% xz, "est"]
  gamma_xx <- coefs[coefs$rhs %in% xx, "est"]
  gamma_zz <- coefs[coefs$rhs %in% zz, "est"]

  mod_stopif(!length(gamma_x),  "coefficient for x not found in model")
  mod_stopif(!length(sd_x),    "variance of x not found in model")
  mod_stopif(!length(gamma_z),  "coefficient for z not found in model")
  mod_stopif(!length(sd_z),    "variance of z not found in model")
  mod_warnif(!length(gamma_xz), "coefficient for xz not found in model")

  if (!length(gamma_xz)) gamma_xz <- 0
  if (!length(gamma_xx)) gamma_xx <- 0
  if (!length(gamma_zz)) gamma_zz <- 0

  # offset by mean
  intercept_y <- getIntercept(y, parTable = parTable)
  mean_x <- getMean(x, parTable = parTable)
  mean_z <- getMean(z, parTable = parTable)
  vals_x <- sd_x * seq(min_x, max_x, by = detail) + mean_x
  vals_z <- sd_z * seq(min_z, max_z, by = detail) + mean_z

  calcExpectedY <- function(x, z) {
    intercept_y +
      gamma_x  * x +
      gamma_z  * z +
      gamma_xz * x * z +
      gamma_xx * x ^ 2 +
      gamma_zz * z ^ 2
  }

  proj_y <- t(outer(vals_x, vals_z, FUN = calcExpectedY))

  # gridline setup: use surface-contours drawn along x and y
  # We specify regular spacing by "size". Plotly draws these as lines on the surface.
  nx <- max(1L, as.integer(grid_nx))
  ny <- max(1L, as.integer(grid_ny))
  size_x <- (max(vals_x) - min(vals_x)) / nx
  size_y <- (max(vals_z) - min(vals_z)) / ny

  contour_spec <- list(
    x = if (grid) list(
      show  = TRUE,
      start = min(vals_x), end = max(vals_x), size = size_x,
      color = grid_color
    ) else list(show = FALSE),
    y = if (grid) list(
      show  = TRUE,
      start = min(vals_z), end = max(vals_z), size = size_y,
      color = grid_color
    ) else list(show = FALSE),
    z = list(show = FALSE)  # keep z-contours off unless desired later
  )

  # hoverover box
  # We could use a monospace font, and use format() such that all of the
  # numerical values are aligned properly, but for now we keep it simple...
  hovertemplate <- paste0(
    x, " %{x:.2f}<br>",
    z, " %{y:.2f}<br>",
    y, " %{z:.2f}",
    "<extra></extra>"
  )

  if (grid || tolower(colorscale) != "viridis") {
    # for some reason this doesn't render with pkgdown this is a
    # temporary workaround, until I figure out what is going on...
    plotly::plot_ly(
      x = ~vals_x,
      y = ~vals_z,
      z = ~proj_y,
      type = "surface",
      surfacecolor = ~proj_y,
      colorscale = colorscale,
      reversescale = reversescale,
      showscale = showscale,
      opacity = surface_opacity,
      contours = contour_spec,
      colorbar = list(title = y),
      cmin = cmin,
      cmax = cmax,
      hovertemplate = hovertemplate,
      ...
    ) |>
      plotly::layout(
        title = sprintf("Surface Plot of Interaction Effect between %s and %s, on %s", x, z, y),
        scene = list(
          xaxis = list(title = x),
          zaxis = list(title = y),
          yaxis = list(title = z)
        )
      )

  } else {
    plotly::plot_ly(
      z = ~proj_y,
      x = ~vals_x,
      y = ~vals_z,
      type = "surface",
      colorbar = list(title = y),
      hovertemplate = hovertemplate
    ) |>
    plotly::layout(title = sprintf("Surface Plot of Interaction Effect between %s and %s, on %s", x, z, y),
                   scene = list(xaxis = list(title = x),
                                zaxis = list(title = y),
                                yaxis = list(title = z)))
  }
}


# Helper function to reverse interaction term
reverseIntTerm <- function(xz) {
  if (grepl(":", xz)) {
    vars <- strsplit(xz, ":")[[1]]
    rev_xz <- paste(rev(vars), collapse = ":")
  } else {
    rev_xz <- xz
  }
  rev_xz
}

devtools::load_all()


HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9
              visual ~ c(v_t1, v_t2) * textual + speed
              interaction := v_t2 - v_t1'

fit_lavaan <- lavaan::sem(HS.model,
                          data = lavaan::HolzingerSwineford1939,
                          group = "school")

# using modsem
fit_modsem <- modsem(HS.model,
                     data = lavaan::HolzingerSwineford1939,
                     group = "school")


lavaanEst <- lavaan::parameterEstimates(fit_lavaan)
lavaanEst[is.na(lavaanEst)] <- -999
modsemEst <- lavaan::parameterEstimates(fit_modsem$lavaan)
modsemEst[is.na(modsemEst)] <- -999
testthat::expect_equal(lavaanEst, modsemEst)

set.seed(123)
oneIntMG <- oneInt
oneIntMG$group <- sample(c("G1", "G2"), nrow(oneInt), replace = TRUE)

m1 <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  Y ~ X + Z + X:Z
'

est.mg <- modsem(m1, oneIntMG, method = "lms", group = "group", robust.se = TRUE)
est.mg.rcs <- modsem(m1, oneIntMG, method = "lms", group = "group", robust.se = TRUE, rcs = TRUE)
standardized_estimates(est.mg, correction = TRUE)
standardized_estimates(est.mg, correction = TRUE, std.errors = "delta")
summary(standardize_model(est.mg))

est.dca.mg <- modsem(m1, oneIntMG, method = "dblcent", group = "group")
est.dca.rcs.mg <- modsem(m1, oneIntMG, method = "dblcent", group = "group", rcs = TRUE)

m2 <- '
  X =~ x1 + lx2 * x2 + lx3 * x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  Y ~ c(a1, a2) * X + Z + X:Z

  a1==a2
'

est <- modsem(m2, oneIntMG, method = "lms", group = "group")
standardized_estimates(est, correction = TRUE)
standardized_estimates(est, correction = TRUE, std.errors = "delta")
summary(est, standardized = TRUE, center = TRUE)
plot_jn(x = "X", z = "Z", y = "Y", model = est)
plot_jn(x = "X", z = "Z", y = "Y", model = est, plot.jn.points = FALSE)
plot_interaction(x = "X", z = "Z", y = "Y", model = est, vals_z = c(1, 0))
testthat::expect_warning(
  plot_surface(x = "X", z = "Z", y = "Y", model = est),
  regexp = "Plotting of surface.*"
)

oneIntMG2 <- oneIntMG
oneIntMG2$group[c(2, 388, 291, 1502)] <- NA

testthat::expect_error(
  modsem(m1, oneIntMG2, method = "lms", group = "group"),
  regexp = ".*group.* cannot contain missing values.*"
)

devtools::load_all()

m1 <- "
# Outer Model
  X =~ x1
  X =~ x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + Z + X:Z
"

est1 <- modsem(m1, data = oneInt)

simple_slopes(x = "X", z = "Z", y = "Y", model = est1) # default of 2 digits
print(simple_slopes(x = "X", z = "Z", y = "Y", model = est1), digits = 3)
print(simple_slopes(x = "X", z = "Z", y = "Y", model = est1), digits = 4)

plot_interaction(x = "X", z = "Z", y = "Y",
                 vals_z = c(1, 0), model = est1, greyscale = TRUE)
plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z",
                 vals_z = c(1, 0), model = est1, ci_type = "prediction")
plot_jn(x = "X", z = "Z", y = "Y", model = est1, greyscale = TRUE)
plot_jn(x = "X", z = "Z", y = "Y", model = est1, greyscale = TRUE, plot.jn.points = FALSE)

plot_surface(x = "X", z = "Z", y = "Y", model = est1, colorscale = "Greys", grid = TRUE, grid_color = "black")
# check input length validation
checkInputValidation <- function(x.default = "X",
                                 z.default = "Z",
                                 y.default = "Y",
                                 xz.default = NULL,
                                 FUN = simple_slopes,
                                 debug = FALSE,
                                 ...) {
  params <- list(x  = x.default,
                 z  = z.default,
                 y  = y.default,
                 xz = xz.default)

  for (par in names(params)) {
    if (is.null(params[[par]]))
      next

    if (debug) {
      printf("  |> Checking %s, xz.default = %s\n", par,
             ifelse(is.null(xz.default), yes = "NULL", no = xz.default))
    }

    argsLong <- c(params, list(...))
    argsLong[[par]] <- c(argsLong[[par]], argsLong[[par]])

    testthat::expect_error(
      do.call(FUN, argsLong),
      regexp = sprintf("%s must be of length 1.*", par)
    )

    if (!par != "xz") { # xz is not validated the same way...
      argsWrong <- c(params, list(...))
      argsWrong[[par]] <- "I_DO_NOT_EXIST"
      testthat::expect_error(
         do.call(FUN, argsWrong),
         regexp = sprintf(".*Unrecognized variable: I_DO_NOT_EXIST.*")
      )
    }
  }
}


debug <- TRUE
funcs <- list(simple_slopes, plot_interaction, plot_jn, plot_surface)
args  <- list(list(model = est1, vals_z = c(0, 1)),
              list(model = est1, vals_z = c(0, 1)),
              list(model = est1), list(model = est1))
for (i in seq_along(funcs)) {
  FUN <- funcs[[i]]
  if (debug) printf("Checking func %d...\n", i)
  for (xz_null in c(TRUE, FALSE)) {
    args_i <- c(args[[i]], list(FUN = FUN, debug = debug,
                                xz.default = if (xz_null) NULL else "X:Z"))

    do.call(checkInputValidation, args_i)
  }
}

estepLms <- function(model, theta, lastQuad = NULL, recalcQuad = FALSE,
                     adaptive.quad.tol = 1e-12, ...) {
  modFilled <- fillModel(model = model, theta = theta, method = "lms")

  P <- list(P_GROUPS = vector("list", length = model$info$n.groups),
            quad = NULL, obsLL = NULL)

  for (g in seq_len(model$info$n.groups)) {
    lastQuad.g <- if (!is.null(lastQuad)) lastQuad[[g]] else NULL
    submodel   <- modFilled$models[[g]]

    P$P_GROUPS[[g]] <- estepLmsGroup(
      submodel          = submodel,
      lastQuad          = lastQuad.g,
      recalcQuad        = recalcQuad,
      adaptive.quad.tol = adaptive.quad.tol,
      ...
    )
  }

  P$quad.err <- sum(vapply(
    X = P$P_GROUPS,
    FUN.VALUE = numeric(1L),
    FUN = function(P) {
      err <- P$quad$error.abs
      if (!length(err) || !is.finite(err[[1L]])) return(0.0)
      abs(err[[1L]])
    }
  ))

  P$quad  <- lapply(
    X = P$P_GROUPS,
    FUN = \(P) P$quad
  )

  P$obsLL <- sum(vapply(
    X = P$P_GROUPS,
    FUN.VALUE = numeric(1L),
    FUN = \(P) P$obsLL
  ))

  P
}


estepLmsGroup <- function(submodel, lastQuad = NULL, recalcQuad = FALSE,
                          adaptive.quad.tol = 1e-12, ...) {
  data             <- submodel$data
  sampling.weights <- data$weights

  if (submodel$quad$adaptive && (recalcQuad || is.null(lastQuad))) {
    m <- submodel$quad$m
    a <- submodel$quad$a
    b <- submodel$quad$b
    m <- submodel$quad$m
    k <- submodel$quad$k

    if (!is.null(lastQuad)) m.ceil <- lastQuad$m.ceil
    else if (k > 1) m.ceil <- m
    else m.ceil <- round(estMForNodesInRange(m, a = -5, b = 5))

    quad <- tryCatch({
        adaptiveGaussQuadrature(
          fun = densityLms, modFilled = submodel,
          data = data, a = a, b = b, m = m,
          k = k, m.ceil = m.ceil, tol = adaptive.quad.tol,
        )
      }, error = function(e) {
        mod_msg_warn_immediate(
          paste0("Calculation of adaptive quadrature failed!\n", e),
          .newline = TRUE
        )
        NULL
      }
    )

    if (is.null(quad)) {
      estep.fixed <- estepLmsGroup(
        submodel = submodel,
        lastQuad = if (!is.null(lastQuad)) lastQuad else submodel$quad,
        recalcQuad = FALSE,
        ...
      )

      return(estep.fixed)
    }

    V <- quad$n
    w <- quad$w
    P <- sweep(quad$F, MARGIN = 2, STATS = w, FUN = "*")

  } else {
    quad <- if (submodel$quad$adaptive) lastQuad else submodel$quad
    V    <- quad$n
    w    <- quad$w
    densityVals <- densityLms(V, modFilled = submodel, data = data)
    P    <- sweep(densityVals, MARGIN = 2, STATS = w, FUN = "*")
  }

  density <- rowSums(P)
  P <- P / density
  if (is.null(sampling.weights)) observedLogLik <- sum(log(density))
  else observedLogLik <- sum(sampling.weights * log(density))


  # Sampling weights multiply each observation's log likelihood and posterior
  # contribution; they must not alter the posterior node probabilities.
  if (!is.null(sampling.weights))
    P <- sampling.weights * P

  stats <- estepSuffStatLmsCpp(P = P, dataR = data$data.split,
                               n = data$n.pattern, npatterns = data$p,
                               ncores = ThreadEnv$n.threads)

  # Create a vector for sampling weights, needed in some C++ code
  if (!is.null(sampling.weights)) sampling.weights.vec <- sampling.weights
  else                            sampling.weights.vec <- rep(1, NROW(P))

  list(P = P, mean = stats$mean, cov = stats$cov, tgamma = stats$tgamma,
       V = V, w = w, obsLL = observedLogLik, quad = quad,
       sampling.weights = sampling.weights.vec)
}


# Maximization step of EM-algorithm (see Klein & Moosbrugger, 2000)
mstepLms <- function(theta, model, P,
                     max.step,
                     verbose = FALSE,
                     control = list(),
                     optimizer = "nlminb",
                     optim.method = "L-BFGS-B",
                     epsilon = 1e-6,
                     ...) {
  thetaInput <- theta

  gradient <- function(theta) {
    grad <- gradientCompLogLikLms(
      theta = theta,
      model = model,
      P = P,
      sign = -1,
      epsilon = epsilon
    )

    non.finite <- !is.finite(grad)
    if (any(non.finite)) {
      # This might happen during a bad line search in optim()
      # as long as obsLogLikLms() is non finite as well, it should be ok,
      # as it will be rejected...
      ll <- compLogLikLms(
        theta = theta,
        model = model,
        P = P, sign = -1,
        epsilon = epsilon
      )

      mod_warnif(is.finite(ll),
        "Some gradient values are NaN, but objective is finite!"
      )

      dir <- sign(theta - thetaInput)
      dir[!is.finite(dir)] <- 0

      grad[non.finite] <- dir[non.finite] * .Machine$double.xmax^(1/10)
    }

    grad
  }

  objective <- function(theta) {
    ll <- compLogLikLms(
      theta = theta,
      model = model,
      P = P, sign = -1,
      epsilon = epsilon
    )

    if (!is.finite(ll)) .Machine$double.xmax^(1/10) else ll # optim doesn't handle Inf well
  }

  if (optimizer == "nlminb") {
    if (is.null(control$iter.max)) {
      # max.step is usually 1
      control$iter.max <- max.step
    }

    if (is.null(control$eval.max)) {
      # defaults to 200, so we keep the default unless iter.max > 100
      control$eval.max <- max(200L, 2 * control$iter.max)
    }

    est <- suppressWarnings(stats::nlminb(
      start     = theta,
      objective = objective,
      gradient  = gradient,
      upper     = model$params$bounds$upper,
      lower     = model$params$bounds$lower,
      control   = control,
      ...
    ))

  } else if (optimizer == "L-BFGS-B") {
    if (is.null(control$maxit)) {
      # max.step is usually 1
      control$maxit <- max.step
    }

    est <- stats::optim(
      par    = theta,
      fn     = objective,
      gr     = gradient,
      method = optim.method, control = control,
      lower  = model$params$bounds$lower,
      upper  = model$params$bounds$upper,
      ...
    )

    est$objective  <- est$value
    est$iterations <- est$counts[["function"]]

  } else {
    mod_msg_stop("Unrecognized optimizer, must be either 'nlminb' or 'L-BFGS-B'")
  }

  est
}


compLogLikLms <- function(theta, model, P, sign = -1, ...) {
  tryCatch({
    modFilled <- fillModel(model = model, theta = theta, method = "lms")

    ll <- 0
    for (g in seq_len(model$info$n.groups)) {
      submodel <- modFilled$models[[g]]
      data <- submodel$data

      ll <- ll + completeLogLikLmsCpp(
        modelR=submodel, P=P$P_GROUPS[[g]], quad=P$quad[[g]],
        colidxR = data$colidx0, n = data$n.pattern,
        d = data$d.pattern, npatterns = data$p
      )
    }

    sign * ll

  }, error = \(e) NA)
}


compLogLikLmsGroup <- function(submodel, P, sign = -1, ...) {
  tryCatch({
    data <- submodel$data

    ll <- completeLogLikLmsCpp(
      modelR = submodel, P = P, quad = P$quad,
      colidxR = data$colidx0, n = data$n.pattern,
      d = data$d.pattern, npatterns = data$p
    )

    sign * ll

  }, error = \(e) NA)
}


obsLogLikLms <- function(theta, model, P, sign = -1, ...) {
  tryCatch({
    modFilled <- fillModel(model = model, theta = theta, method = "lms")
    ll <- 0

    for (g in seq_len(model$info$n.groups)) {
      ll <- ll + obsLogLikLmsGroup(
        submodel = modFilled$models[[g]], P = P$P_GROUPS[[g]], sign = 1
      )
    }

    sign * ll
  }, error = \(e) NA_real_)
}


obsLogLikLmsGroup <- function(submodel, P, sign = -1, ...) {
  tryCatch({
    data <- submodel$data
    sign * observedLogLikLmsCpp(
      modelR = submodel, dataR = data$data.split, colidxR = data$colidx0,
      P = P, n = data$n.pattern, npatterns = data$p,
      ncores = ThreadEnv$n.threads
    )
  }, error = \(e) NA_real_)
}



gradientCompLogLikLms <- function(theta, model, P, sign = -1, epsilon = 1e-6) {
  gradientAllLogLikLms(theta = theta, model = model, P = P, sign = sign,
                       epsilon = epsilon, FGRAD = gradLogLikLmsCpp, FOBJECTIVE = compLogLikLmsGroup)
}


gradientObsLogLikLms <- function(theta, model, P, sign = -1, epsilon = 1e-6) {
  FGRAD <- function(modelR, P, block, row, col, symmetric, colidxR,
                    n, d, npatterns, eps, ncores) {
    gradObsLogLikLmsCpp(
      modelR = modelR, dataR = modelR$data$data.split, colidxR = colidxR,
      P = P, block = block, row = row, col = col, symmetric = symmetric,
      n = n, eps = eps, npatterns = npatterns, ncores = ncores
    )
  }

  gradientAllLogLikLms(
    theta = theta, model = model, P = P, sign = sign, epsilon = epsilon,
    FGRAD = FGRAD, FOBJECTIVE = obsLogLikLmsGroup
  )
}


gradientAllLogLikLms <- function(theta, model, P, sign = -1, epsilon = 1e-6,
                                 FGRAD, FOBJECTIVE) {
  useFDGradient <- model$params$gradientStruct$useFDGradient

  if (useFDGradient) gradient <- \(...) complicatedGradientAllLogLikLms(..., FOBJECTIVE = FOBJECTIVE)
  else               gradient <- \(...) simpleGradientAllLogLikLms(..., FGRAD = FGRAD)

  c(gradient(theta = theta, model = model, P = P, sign = sign, epsilon = epsilon))
}


complicatedGradientAllLogLikLms <- function(theta, model, P, sign = -1, epsilon = 1e-4, FOBJECTIVE, sum = TRUE, ...) {
  params <- model$params

  SELECT_THETA_LAB  <- params$SELECT_THETA_LAB
  SELECT_THETA_COV  <- params$SELECT_THETA_COV
  SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN

  N <- vapply(model$models, FUN.VALUE = numeric(1L), FUN = \(sub) sub$data$n)
  N.start <- c(1, cumsum(N)[-length(N)] + 1L)
  N.end   <- cumsum(N)

  if (sum) n <- 1L
  else     n <- sum(N)

  k  <- length(theta)

  grad <- matrix(0, nrow = n, ncol = k, dimnames = list(NULL, names(theta)))

  FOBJECTIVE_GROUP <- function(theta, g) {
    modFilled <- fillModel(theta = theta, model = model, method = "lms")
    FOBJECTIVE(submodel = modFilled$models[[g]], P = P$P_GROUPS[[g]], sign = sign, ...)
  }

  for (g in seq_len(model$info$n.groups)) {
    f0 <- FOBJECTIVE_GROUP(theta = theta, g = g)

    if (sum) J <- 1L
    else     J <- seq(N.start[[g]], N.end[[g]], by = 1L)

    indices <- c(
      SELECT_THETA_LAB[[g]],
      SELECT_THETA_COV[[g]],
      SELECT_THETA_MAIN[[g]]
    )

    for (i in indices) {
      theta_i <- theta
      theta_i[i] <- theta_i[i] + epsilon

      fi <- FOBJECTIVE_GROUP(theta_i, g = g)
      grad[J, i] <- grad[J, i] + (fi - f0) / epsilon
    }
  }

  if (n == 1L) as.vector(grad) else grad
}


simpleGradientAllLogLikLms <- function(theta, model, P, sign = -1, epsilon = 1e-6, FGRAD, N.FGRAD = 1L) {
  # simple gradient which should work if constraints are well-behaved functions
  # which can be derivated by Deriv::Deriv, and there is no covModel
  modelR     <- fillModel(model=model, theta=theta, method="lms")
  locations  <- model$params$gradientStruct$locations
  Jacobian   <- model$params$gradientStruct$Jacobian
  Jacobian   <- refreshCovModelJacobian(theta, model, Jacobian)$J
  nlinDerivs <- model$params$gradientStruct$nlinDerivs

  # grad <- stats::setNames(numeric(NROW(locations)), nm = locations$param)
  grad <- matrix(0, nrow = NROW(locations), ncol = N.FGRAD,
                 dimnames = list(locations$param, NULL))

  for (g in seq_len(modelR$info$n.groups)) {
    submodelR   <- modelR$models[[g]]
    locations.g <- locations[locations$group == g, , drop = FALSE]

    data      <- submodelR$data
    block     <- locations.g$block
    row       <- locations.g$row
    col       <- locations.g$col
    param     <- locations.g$param
    symmetric <- locations.g$symmetric

    grad.g <- FGRAD(modelR    = submodelR,
                    P         = P$P_GROUPS[[g]],
                    block     = block,
                    row       = row,
                    col       = col,
                    symmetric = symmetric,
                    colidxR   = data$colidx0,
                    npatterns = data$p,
                    n         = data$n.pattern,
                    d         = data$d.pattern,
                    eps       = epsilon,
                    ncores    = ThreadEnv$n.threads)

    grad[locations.g$param, ] <- grad.g
  }

  if (length(nlinDerivs)) {
    evalTheta  <- model$params$gradientStruct$evalTheta
    param.full <- stringr::str_split_i(colnames(Jacobian), pattern = "#", i = 1L) # non-unique
    param.part <- rownames(Jacobian)
    THETA      <- list2env(as.list(evalTheta(theta)))

    for (dep in names(nlinDerivs)) {
      derivs  <- nlinDerivs[[dep]]

      for (indep in names(derivs)) {
        deriv <- eval(expr = derivs[[indep]], envir = THETA)

        match.full <- param.full == dep
        match.part <- param.part == indep
        Jacobian[match.part, match.full] <- deriv
      }
    }
  }

  sign * Jacobian %*% grad
}


obsLogLikLmsGroup_i <- function(submodel, P, sign = 1) {
  data <- submodel$data
  V  <- P$V
  w  <- P$w
  m  <- nrow(V)
  px <- numeric(data$n)

  for (i in seq_len(m)) {
    z_i        <- V[i, ]
    ms_i       <- muSigmaLmsCpp(model = submodel, z = z_i)
    mu_i    <- ms_i$mu
    sigma_i <- ms_i$sigma

    dens_i <- numeric(data$n)
    offset <- 1L

    for (id in data$ids) {
      n.pattern <- data$n.pattern[[id]]
      colidx    <- data$colidx[[id]]

      end <- offset + n.pattern - 1L
      dens_i[offset:end] <- dmvn(data$data.split[[id]],
                                 mean = mu_i[colidx],
                                 sigma = sigma_i[colidx, colidx],
                                 log = FALSE)
      offset <- end + 1L
    }

    px <- px + w[i] * dens_i
  }

  logdens <- log(px)

  sampling.weights <- data$weights
  if (!is.null(sampling.weights))
    logdens <- sampling.weights * logdens

  sign * logdens
}


gradientObsLogLikLms_i <- function(theta, model, P, sign = 1, epsilon = 1e-4) {
  complicatedGradientAllLogLikLms(theta = theta, model = model, P = P,
                                  sign = sign, epsilon = epsilon, sum = FALSE,
                                  FOBJECTIVE = obsLogLikLmsGroup_i)
}


densitySingleLms <- function(z, modFilled, data) {
  ms    <- muSigmaLmsCpp(model = modFilled, z = z)
  mu    <- ms$mu
  sigma <- ms$sigma

  density <- numeric(data$n)

  offset <- 1L
  for (id in data$ids) { # go along patterns
    n.pattern <- data$n.pattern[[id]]

    colidx <- data$colidx[[id]]
    dataid <- data$data.split[[id]]

    end <- offset + n.pattern - 1L
    density[offset:end] <- dmvn(data$data.split[[id]],
                                mean = mu[colidx],
                                sigma = sigma[colidx, colidx])
    offset <- end + 1L
  }

  density
}


densityLms <- function(z, modFilled, data) {
  if (is.null(dim(z))) z <- matrix(z, ncol = modFilled$quad$k)
  densityMatrixLmsCpp(modelR = modFilled, V = z,
                      dataR = data$data.split, colidxR = data$colidx0,
                      n = data$n.pattern, samplingWeights = numeric(0L),
                      npatterns = data$p, ncores = ThreadEnv$n.threads)
}


hessianAllLogLikLms <- function(theta, model, P, sign = -1,
                                FHESS, FOBJECTIVE, .relStep = .Machine$double.eps ^ (1/5)) {
  useFDGradient <- model$params$gradientStruct$useFDGradient

  if (useFDGradient) hessian <- \(...) complicatedHessianAllLogLikLms(..., FOBJECTIVE = FOBJECTIVE, .relStep = .relStep)
  else               hessian <- \(...) simpleHessianAllLogLikLms(..., FHESS = FHESS, .relStep = .relStep)

  hessian(theta = theta, model = model, P = P, sign = sign)
}


complicatedHessianAllLogLikLms <- function(theta, model, P, sign = -1,
                                          .relStep = .Machine$double.eps ^ (1/6),
                                          FOBJECTIVE) {
  params <- model$params

  SELECT_THETA_LAB  <- params$SELECT_THETA_LAB
  SELECT_THETA_COV  <- params$SELECT_THETA_COV
  SELECT_THETA_MAIN <- params$SELECT_THETA_MAIN

  k <- length(theta)
  H <- matrix(0, nrow = k, ncol = k, dimnames = list(names(theta), names(theta)))

  for (g in seq_len(model$info$n.groups)) {
    indices <- c(
      SELECT_THETA_LAB[[g]],
      SELECT_THETA_COV[[g]],
      SELECT_THETA_MAIN[[g]]
    )

    FOBJECTIVE_GROUP <- function(theta.g) {
      theta[indices] <- theta.g # local copy of theta
      modFilled <- fillModel(theta = theta, model = model, method = "lms")
      FOBJECTIVE(submodel = modFilled$models[[g]], sign = sign, P = P$P_GROUPS[[g]])
    }

    theta.g <- theta[indices]
    Hg <- fdHESS(pars = theta.g, fun = FOBJECTIVE_GROUP, .relStep = .Machine$double.eps^(1/5))

    H[indices, indices] <- H[indices, indices] + Hg
  }

  H
}


simpleHessianAllLogLikLms <- function(theta, model, P, sign = -1,
                                      .relStep = .Machine$double.eps ^ (1/5),
                                      FHESS) {
  # simple Hessian which should work if constraints are well-behaved functions
  # which can be derivated by Deriv::Deriv, and there is no covModel
  modelR      <- fillModel(model = model, theta = theta, method = "lms")
  locations   <- model$params$gradientStruct$locations
  Jacobian    <- model$params$gradientStruct$Jacobian
  Jacobian2   <- model$params$gradientStruct$Jacobian2
  jac         <- refreshCovModelJacobian(theta, model, Jacobian, Jacobian2)
  Jacobian    <- jac$J
  Jacobian2   <- jac$J2
  nlinDerivs  <- model$params$gradientStruct$nlinDerivs
  nlinDerivs2 <- model$params$gradientStruct$nlinDerivs2

  n.loc <- NROW(locations)
  nm <- locations$param
  H <- matrix(0.0, nrow = n.loc, ncol = n.loc, dimnames = list(nm, nm))
  grad <- stats::setNames(numeric(n.loc), nm = nm)

  for (g in seq_len(modelR$info$n.groups)) {
    locations.g <- locations[
      locations$group == g, , drop = FALSE
    ]

    if (!NROW(locations.g)) next

    submodelR <- modelR$models[[g]]
    data.g    <- submodelR$data

    HESS.g <- FHESS(
      modelR    = submodelR,
      P         = P$P_GROUPS[[g]],
      block     = locations.g$block,
      row       = locations.g$row,
      col       = locations.g$col,
      colidxR   = data.g$colidx0,
      n         = data.g$n.pattern,
      d         = data.g$d.pattern,
      npatterns = data.g$p,
      symmetric = locations.g$symmetric,
      .relStep  = .relStep,
      ncores    = ThreadEnv$n.threads
    )

    H.g    <- HESS.g$Hessian
    grad.g <- HESS.g$gradient
    nm.g   <- locations.g$param

    dimnames(H.g) <- list(nm.g, nm.g)
    names(grad.g) <- nm.g

    H[nm.g, nm.g] <- H[nm.g, nm.g] + H.g
    grad[nm.g] <- grad[nm.g] + grad.g
  }

  if (length(nlinDerivs)) {
    evalTheta  <- model$params$gradientStruct$evalTheta
    param.full <- stringr::str_split_i(colnames(Jacobian), pattern = "#", i = 1L) # non-unique
    param.part <- rownames(Jacobian)
    THETA      <- list2env(as.list(evalTheta(theta)))

    for (dep in names(nlinDerivs)) {
      derivs1 <- nlinDerivs[[dep]]
      derivs2 <- nlinDerivs2[[dep]]

      for (indep in names(derivs1)) {
        deriv1 <- eval(expr = derivs1[[indep]], envir = THETA)
        deriv2 <- eval(expr = derivs2[[indep]], envir = THETA)

        match.full <- param.full == dep
        match.part <- param.part == indep

        Jacobian[match.part, match.full] <- deriv1
        Jacobian2[match.part, match.full] <- deriv2
      }
    }
  }

  term1 <- Jacobian %*% H %*% t(Jacobian)
  term2 <- diag(drop(Jacobian2 %*% grad), nrow = nrow(Jacobian))

  sign * (term1 + term2)
}


hessianCompLogLikLms <- function(theta, model, P, sign = -1,
                                 .relStep = .Machine$double.eps ^ (1/5)) {

  FHESS <- function(modelR, P, block, row, col, symmetric, eps, .relStep, colidxR,
                    n, d, npatterns, ncores) {
    hessCompLogLikLmsCpp(modelR = modelR, P = P, block = block,
                         row = row, col = col, symmetric = symmetric,
                         colidxR = colidxR, n = n, d = d, relStep = .relStep,
                         npatterns = npatterns, minAbs = 0.0, ncores = ncores)
  }

  hessianAllLogLikLms(theta = theta, model = model, P = P, sign = sign,
                      FHESS = FHESS, FOBJECTIVE = compLogLikLmsGroup,
                      .relStep = .relStep)
}


hessianObsLogLikLms <- function(theta, model, P, sign = -1,
                                .relStep = .Machine$double.eps ^ (1/5)) {
  FHESS <- function(modelR, P, block, row, col, symmetric, eps, .relStep,
                    colidxR, n, d, npatterns, ncores) {
    hessObsLogLikLmsCpp(
      modelR = modelR, dataR = modelR$data$data.split, P = P,
      block = block, row = row, col = col, symmetric = symmetric,
      colidxR = colidxR, n = n, npatterns = npatterns,
      relStep = .relStep, minAbs = 0.0, ncores = ncores
    )
  }

  hessianAllLogLikLms(
    theta = theta, model = model, P = P, sign = sign,
    FHESS = FHESS, FOBJECTIVE = obsLogLikLmsGroup, .relStep = .relStep
  )
}


logdensAllObsNode <- function(theta, model, data, z, group = NULL) {
  modFilled <- fillModel(model = model, theta = theta, method = "lms")
  submodel <- if (!is.null(modFilled$models)) {
    if (is.null(group)) modFilled$models[[1L]] else modFilled$models[[group]]
  } else modFilled
  # densitySingleLms returns densities for all rows at given z (not log)
  dens <- densitySingleLms(z = z, modFilled = submodel, data = data)
  # guard against underflow/zeros
  log(pmax(dens, .Machine$double.xmin))
}


activeThetaIndicesLms <- function(model, group, p) {
  select_lab  <- model$params$SELECT_THETA_LAB[[group]]
  select_cov  <- model$params$SELECT_THETA_COV[[group]]
  select_main <- model$params$SELECT_THETA_MAIN[[group]]

  if (is.null(select_lab) && is.null(select_cov) && is.null(select_main)) {
    return(seq_len(p))
  }

  idx <- c(select_lab,
           select_cov,
           select_main)

  idx <- sort(unique(as.integer(idx)))
  idx <- idx[idx >= 1L & idx <= p]

  if (length(idx)) idx else seq_len(p)
}


# per-node, per-observation complete-data score via finite difference
# Returns an n x p matrix S_j with row i = s_{ij}^T = grad_theta log p(y_i, z_j | theta)
completeScoresNodeFD <- function(theta, model, data, z,
                                 epsilon = 1e-6, scheme = c("forward","central"),
                                 group = NULL, active = NULL) {
  scheme <- match.arg(scheme)
  p <- length(theta)

  if (is.null(active)) {
    active <- seq_len(p)
  } else {
    active <- unique(as.integer(active))
    active <- active[active >= 1L & active <= p]
  }

  n <- data$n
  n_active <- length(active)

  if (!n_active) {
    return(matrix(0.0, nrow = n, ncol = 0L))
  }

  col_names <- names(theta)
  if (!length(col_names)) col_names <- rep("", p)

  S <- matrix(0.0, nrow = n, ncol = n_active,
              dimnames = list(NULL, col_names[active]))

  if (scheme == "forward") {
    f0 <- logdensAllObsNode(theta, model, data, z, group = group)
    for (pos in seq_len(n_active)) {
      k <- active[pos]
      th1 <- theta; th1[k] <- th1[k] + epsilon
      f1 <- logdensAllObsNode(th1, model, data, z, group = group)
      S[, pos] <- (f1 - f0) / epsilon
    }
  } else { # central
    for (pos in seq_len(n_active)) {
      k <- active[pos]
      thp <- theta; thp[k] <- thp[k] + epsilon
      thm <- theta; thm[k] <- thm[k] - epsilon
      fp <- logdensAllObsNode(thp, model, data, z, group = group)
      fm <- logdensAllObsNode(thm, model, data, z, group = group)
      S[, pos] <- (fp - fm) / (2 * epsilon)
    }
  }
  S
}


lmsFirstDerivativeJacobian <- function(theta, model) {
  gradientStruct <- model$params$gradientStruct
  Jacobian <- refreshCovModelJacobian(
    theta, model, gradientStruct$Jacobian
  )$J
  nlinDerivs <- gradientStruct$nlinDerivs

  if (length(nlinDerivs)) {
    evalTheta  <- gradientStruct$evalTheta
    param.full <- stringr::str_split_i(
      colnames(Jacobian), pattern = "#", i = 1L
    )
    param.part <- rownames(Jacobian)
    THETA <- list2env(as.list(evalTheta(theta)))

    for (dep in names(nlinDerivs)) {
      for (indep in names(nlinDerivs[[dep]])) {
        deriv <- eval(expr = nlinDerivs[[dep]][[indep]], envir = THETA)
        Jacobian[param.part == indep, param.full == dep] <- deriv
      }
    }
  }

  Jacobian
}


completeScoresNodeAnalytical <- function(submodel, data, z, locations,
                                         Jacobian, active) {
  rawScores <- completeScoresNodeAnalyticalLmsCpp(
    modelR = submodel,
    dataR = data$data.split,
    z = c(z),
    block = locations$block,
    row = locations$row,
    col = locations$col,
    symmetric = locations$symmetric,
    colidxR = data$colidx0,
    n = data$n.pattern,
    npatterns = data$p,
    ncores = ThreadEnv$n.threads
  )

  mapping <- Jacobian[active, locations$param, drop = FALSE]
  rawScores %*% t(mapping)
}

# I_obs = I_com - I_mis using Louis' identity
observedInfoFromLouisLms <- function(model,
                                     theta,
                                     P = NULL,
                                     recompute.P = is.null(P),
                                     adaptive.quad.tol = 1e-12,
                                     fd.epsilon = 1e-6,
                                     fd.scheme = c("forward","central"),
                                     score.method = c("auto", "analytical", "finite.difference"),
                                     symmetrize = TRUE,
                                     jitter = 0.0,
                                     ...) {
  fd.scheme <- match.arg(fd.scheme)
  score.method <- match.arg(score.method)
  if (score.method == "auto") {
    score.method <- if (isTRUE(model$params$gradientStruct$useFDGradient)) {
      "finite.difference"
    } else {
      "analytical"
    }
  }

  # E-step (if needed)
  if (recompute.P) {
    P <- estepLms(model = model, theta = theta,
                  lastQuad = NULL, recalcQuad = FALSE,
                  adaptive.quad.tol = adaptive.quad.tol, ...)
  }

  p   <- length(theta)
  lbl <- names(theta)

  n_total   <- sum(vapply(model$models, function(sub) sub$data$n, numeric(1L)))

  Icom <- hessianCompLogLikLms(theta = theta, model = model, P = P, sign = -1)

  if (score.method == "analytical") {
    modelFilled <- fillModel(model = model, theta = theta, method = "lms")
    locations <- model$params$gradientStruct$locations
    Jacobian <- lmsFirstDerivativeJacobian(theta, model)
  }

  total_M <- matrix(0.0, p, p, dimnames = list(lbl, lbl))
  Sbar    <- matrix(0.0, n_total, p)

  row_offset <- 0L
  for (g in seq_len(model$info$n.groups)) {
    submodel <- model$models[[g]]
    data.g   <- submodel$data
    P.g      <- P$P_GROUPS[[g]]
    rows     <- seq_len(data.g$n) + row_offset

    active_idx <- activeThetaIndicesLms(model, g, p)
    if (score.method == "analytical") {
      submodelFilled <- modelFilled$models[[g]]
      locations.g <- locations[locations$group == g, , drop = FALSE]
    }
    sampling_weights <- P.g$sampling.weights
    mod_stopif(length(sampling_weights) != data.g$n,
               "Invalid sampling-weight vector in LMS posterior object.")
    sqrt_weights <- sqrt(pmax(sampling_weights, 0))
    Jg <- length(P.g$w)
    for (j in seq_len(Jg)) {
      z_j <- P.g$V[j, , drop = FALSE]
      S_j <- if (score.method == "analytical") {
        completeScoresNodeAnalytical(
          submodel = submodelFilled, data = data.g, z = z_j,
          locations = locations.g, Jacobian = Jacobian, active = active_idx
        )
      } else {
        completeScoresNodeFD(
          theta, model, data.g, z_j,
          epsilon = fd.epsilon, scheme = fd.scheme,
          group = if (model$info$n.groups > 1L) g else NULL,
          active = active_idx
        )
      }
      # P.g$P contains w_i * gamma_ij. This is directly suitable for
      # E_w[s s']; for E_w[s]E_w[s]' we need sqrt(w_i) * E[s], not
      # w_i * E[s], otherwise crossprod() squares the sampling weights.
      r_j <- P.g$P[, j]
      mean_score_weights <- numeric(length(r_j))
      positive_weights <- sqrt_weights > 0
      mean_score_weights[positive_weights] <-
        r_j[positive_weights] / sqrt_weights[positive_weights]

      Rhalf <- sqrt(pmax(r_j, 0))
      if (NROW(S_j) && NCOL(S_j)) {
        X <- S_j * Rhalf
        block <- total_M[active_idx, active_idx, drop = FALSE]
        block <- block + crossprod(X)
        total_M[active_idx, active_idx] <- block

        Sbar_block <- Sbar[rows, active_idx, drop = FALSE]
        Sbar_block <- Sbar_block + (S_j * mean_score_weights)
        Sbar[rows, active_idx] <- Sbar_block
      }
    }

    row_offset <- row_offset + data.g$n
  }

  sbar_outer <- crossprod(Sbar)
  Imis <- total_M - sbar_outer
  Iobs <- Icom - Imis

  if (symmetrize) {
    sym <- function(A) 0.5 * (A + t(A))
    Icom <- sym(Icom)
    Imis <- sym(Imis)
    Iobs <- sym(Iobs)
  }

  if (jitter > 0) {
    Icom <- Icom + diag(jitter, p)
    Imis <- Imis + diag(jitter, p)
    Iobs <- Iobs + diag(jitter, p)
  }

  list(I.obs = Iobs, I.com = Icom, I.mis = Imis, P = P)
}

devtools::load_all()
m1 <- "
 # Outer Model
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Z =~ z1 + z2 + z3

 # Inner model
 Y ~ X + Z + X:Z
"


# modsem_da
est_h1 <- modsem(m1, oneInt, "lms")
est_h0 <- estimate_h0(est_h1, calc.se=FALSE) # std.errors are not needed
compare_fit(est_h1 = est_h1, est_h0 = est_h0)

modsem_inspect(est_h1, what="fit")


# modsem_pi
est_h1 <- modsem(m1, oneInt, method = "dblcent")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)

est_h1 <- modsem(m1, oneInt, method = "ca")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)

createThetaLabel <- function(labelMatrices, labelMatricesCov,
                             constrExprs, start = NULL) {
  matrices <- c(labelMatrices, labelMatricesCov)
  labels <- lapply(matrices, FUN = function(x) {
    if (NCOL(x) == 0 || NROW(x) == 0) return(NULL)

    select <- apply(x, MARGIN = 2, FUN = function(z)
                    !canBeNumeric(z, includeNA = TRUE))
    as.vector(x[select])
  }) |> unlist() |> unique()

  if (!is.null(constrExprs)) {
    labels <- labels[!labels %in% constrExprs$fixedParams]
  }

  if (is.null(start)) {
    start <- vapply(labels, FUN.VALUE = numeric(1L),
                    FUN = function(x) stats::runif(1))
  }

  names(start) <- labels
  start
}


calcThetaLabel <- function(theta, constrExprs) {
  if (is.null(constrExprs)) return(theta)
  Theta <- c(as.list(theta), constrExprs$thetaFixedParams)

  for (i in seq_along(constrExprs$fixedParams)) {
    Theta[constrExprs$fixedParams[[i]]] <-
      eval(constrExprs$exprs[[i]], envir = Theta)
  }

  unlist(Theta)
}


fillMatricesLabels <- function(matrices, labelMatrices, thetaLabel,
                               constraintExprs = NULL) {
  if (is.null(thetaLabel)) return(matrices)
  labels <- names(thetaLabel)

  isOK  <- \(M, lM) all(dim(M) > 0L) && !all(lM == "")
  valid <- unlist(purrr::map2(.x = matrices, .y = labelMatrices, .f = isOK))

  if (!any(valid))
    return(matrices)

  fill <- function(M, L) {
    lapply(labels, FUN = \(label) M[L == label] <<- thetaLabel[[label]])
    M
  }

  matrices[valid] <- purrr::map2(
    .x = matrices[valid], .y = labelMatrices[valid], .f = fill
  )

  matrices
}


parTableLabelsToList <- function(parTable) {
  parTable <- parTable[parTable$label != "", ]

  labelList <- list()
  for (i in seq_len(NROW(parTable))) {
    val   <- parTable[i, "est"]
    label <- parTable[i, "label"]

    labelList[[label]] <- val
  }

  labelList
}


getVarsExpr <- function(expr) {
  tokens <- createTokensLine(getCharsLine(expr))
  tokens[vapply(tokens, FUN.VALUE = logical(1L), FUN = is.LavName)] |>
    lapply(as.character) |> unlist()
}


removeConstraintExpressions <- function(parTable) {
  if (!NROW(parTable)) return(parTable)
  parTable[!parTable$op %in% c("==", "<", ">", ":="), ]
}


getVCOV_LabelledParams <- function(vcov, model, theta, method, epsilon = 1e-8) {
  J <- getJacobianLabelledParams(model = model, theta = theta, epsilon = epsilon,
                                 method = method)
  if (is.null(J)) return(vcov)
  J %*% vcov %*% t(J)
}


getJacobianLabelledParams <- function(model, theta, method, epsilon = 1e-8) {
  constrExprs <- model$constrExprs

  g <- getTransformationsTheta(model, theta = theta, method = method)
  J <- matrix(0, nrow = length(g), ncol = length(theta),
              dimnames = list(names(g), names(theta)))

  for (i in seq_along(theta)) {
    theta_i      <- theta
    theta_i[[i]] <- theta[[i]] + epsilon
    g_i <- getTransformationsTheta(model, theta = theta_i, method = method)

    J[, i] <- (g_i - g) / epsilon
  }

  J
}


getTransformationsTheta <- function(model, theta, method) {
  thetaFull <- calcPhiTheta(theta = theta, model = model, method = method)

  constrExprs    <- model$params$constrExprs
  selectThetaLab <- model$params$SELECT_THETA_LAB[[1L]] # fixed across models

  if (!length(selectThetaLab) && !length(constrExprs$exprs))
    return(thetaFull)

  zeroLen <- length(selectThetaLab) <= 0
  thetaLabel <- if (zeroLen) NULL else thetaFull[selectThetaLab]
  thetaRest  <- if (zeroLen) thetaFull else thetaFull[-selectThetaLab]

  thetaLabel <- suppressWarnings(calcThetaLabel(thetaLabel, constrExprs))
  c(thetaLabel, thetaRest)
}


constraintsContainUnmatchedLabels <- function(parTable, labels) {
  vapply(seq_len(NROW(parTable)), FUN.VALUE = logical(1L), FUN = function(i) {
    if (!parTable[i, "op"] %in% c("==", "<", ">", ":=")) return(FALSE)
    x <- getVarsExpr(parTable[i, "rhs"])
    !all(x %in% labels)
  })
}

#' Extract parameterEstimates from an estimated model
#'
#' @param object An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, or \code{\link{modsem_mplus}}
#' @param ... Additional arguments passed to other functions
#'
#' @examples
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#'   # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#' # Double centering approach
#' est_dca <- modsem(m1, oneInt)
#'
#' pars <- parameter_estimates(est_dca) # no correction
#'
#' # Pretty summary
#' summarize_partable(pars)
#'
#' # Only print the data.frame
#' pars
#' @export
parameter_estimates <- function(object, ...) {
  UseMethod("parameter_estimates")
}


#' Extract or modify parTable from an estimated model with estimated variances of interaction terms
#'
#' @param object An object of class \code{\link{modsem_da}},  \code{\link{modsem_mplus}},
#' or a parTable of class \code{\link{data.frame}}
#' @param ... Additional arguments passed to other functions
#' @export
var_interactions <- function(object, ...) {
  UseMethod("var_interactions")
}


#' Get Standardized Estimates
#'
#' Computes standardized estimates of model parameters for various types of \code{\link{modsem}} objects.
#'
#' @param object An object of class \code{\link{modsem_da}}, \code{\link{modsem_mplus}},
#' \code{\link{modsem_pi}}, or a parameter table (\code{parTable}) of class \code{data.frame}.
#' @param ... Additional arguments passed to underlying methods. See specific method
#' documentation for supported arguments, including:
#' \describe{
#'   \item{\code{correction}}{(Logical) Applies only to \code{modsem_pi} objects. Whether to correct
#'   standardized estimates for the interaction term by computing \code{var(xz)} based on
#'   variances and covariance of \code{x} and \code{z}. Default is \code{FALSE}.}
#'   \item{\code{std.errors}}{(Character) Specifies the method for computing standard errors when
#'   \code{correction = TRUE}. Options are \code{"rescale"}, \code{"delta"}, and \code{"monte.carlo"}.
#'   See method \code{standardized_estimates.modsem_pi()} for details.}
#' }
#'
#' @details
#' Standard errors can either be calculated using the delta method, or a monte.carlo simulation
#'   (\code{monte.carlo} is not available for \code{\link{modsem_pi}} objects if \code{correction == FALSE}.).
#'   \strong{NOTE} that the standard errors of the standardized paramters change the assumptions of the
#'   model, and will in most cases yield different z and p-values, compared to the unstandardized solution.
#'   In almost all cases, significance testing should be based on the unstandardized solution. Since,
#'   the standardization process changes the model assumptions, it also changes what the p-statistics measure.
#'   I.e., the test statistics for the standardized and unstandardized solutions belong to different sets of
#'   hypothesis, which are not exactly equivalent to each other.
#'
#' For \code{modsem_da} and \code{modsem_mplus} objects, the interaction term is not a formal
#'  variable in the model and therefore lacks a defined variance. Under assumptions of normality
#'  and zero-mean variables, the interaction variance is estimated as:
#'  \deqn{var(xz) = var(x) * var(z) + cov(x, z)^2}
#'  This means the standardized estimate for the interaction differs from approaches like
#'  \code{lavaan}, which treats the interaction as a latent variable with unit variance.
#'
#' For \code{modsem_pi} objects, the interaction term is standardized by default assuming
#'  \code{var(xz) = 1}, but this can be overridden using the \code{correction} argument.
#'
#' \strong{NOTE:} Standardized estimates are always placed in the \strong{\code{est}} column,
#' not \code{est.std}, regardless of model type.
#'
#' @return A \code{data.frame} with standardized estimates in the \code{est} column.
#'
#' @examples
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#'   # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#' # Double centering approach
#' est_dca <- modsem(m1, oneInt)
#'
#' std1 <- standardized_estimates(est_dca) # no correction
#' summarize_partable(std1)
#'
#' std2 <- standardized_estimates(est_dca, correction = TRUE) # apply correction
#' summarize_partable(std2)
#'
#' \dontrun{
#' est_lms <- modsem(m1, oneInt, method = "lms")
#' standardized_estimates(est_lms) # correction not relevant for lms
#' }
#'
#' @export
standardized_estimates <- function(object, ...) {
  UseMethod("standardized_estimates")
}


#' Get Centered Interaction Term Estimates
#'
#' Computes centered estimates of model parameters. This is relevant when there is an
#'  interaction term in the model, as the simple main effects depend upon the mean structure
#'  of the structural model. Currenlty only available for
#'  \code{\link{modsem_da}} and \code{lavaan} object.
#'  It is not relevant for the PI approaches (excluding the "pind" method, which is not recommended),
#'  since the indicators are centered before computing the product terms.
#'  The centering can be applied to observed variable interactions in \code{lavaan} models
#'  and latent interactions estimated using the \code{sam} function.
#'
#' @param object An object of class \code{\link{modsem_da}}
#' @param ... Additional arguments passed to underlying methods. See specific method
#' documentation for supported arguments, including:
#' @return A \code{data.frame} with centered estimates in the \code{est} column.
#'
#' @examples
#' m1 <- '
#'   # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#'   # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#' \dontrun{
#' est_lms <- modsem(m1, oneInt, method = "lms")
#' centered_estimates(est_lms)
#' }
#'
#' @export
centered_estimates <- function(object, ...) {
  UseMethod("centered_estimates")
}


#' @export
standardized_estimates.data.frame <- function(object, intercepts = FALSE, ...) {
  parTable <- object[c("lhs", "op", "rhs", "label", "est", "std.error")]
  parTable <- centerInteractions(parTable) # re-estimate path-coefficients
                                          # when intercepts are zero

  if (!intercepts) {
    parTable <- parTable[parTable$op != "~1", ]
  } else {
    parTable[parTable$op == "~1", "est"]       <- 0
    parTable[parTable$op == "~1", "std.error"] <- NA
  }

  parTable <- var_interactions(parTable)
  lVs      <- getLVs(parTable)
  intTerms <- getIntTerms(parTable)
  etas     <- getSortedEtas(parTable, isLV = TRUE)
  xis      <- getXis(parTable, etas = etas, isLV = TRUE)
  indsLVs  <- getIndsLVs(parTable, lVs)
  allInds  <- unique(unlist(indsLVs))

  # get variances
  variancesInds <- calcVarParTable(allInds, parTable, measurement.model = TRUE)
  variancesLVs  <- calcVarParTable(lVs, parTable, measurement.model = FALSE)
  variancesIntTerms <- structure(numeric(length(intTerms)), names = intTerms)

  for (xz in intTerms) {
    variancesIntTerms[[xz]] <- parTable[parTable$lhs == xz &
                                        parTable$rhs == xz &
                                        parTable$op == "~~", "est"]
  }

  variances <- c(variancesInds, variancesLVs, variancesIntTerms)

  # Factor Loadings
  lambda     <- NULL
  selectRows <- NULL
  selectCols <- c("est", "std.error")

  for (lV in lVs) {
    for (ind in indsLVs[[lV]]) {
      selectRows  <- parTable$lhs == lV & parTable$op == "=~" & parTable$rhs == ind
      scalingCoef <- sqrt(variances[[lV]]) / sqrt(variances[[ind]])
      lambda      <- parTable[selectRows, selectCols]

      parTable[selectRows, selectCols] <- lambda * scalingCoef
    }
  }

  # Structural Coefficients
  gamma               <- NULL
  selectStrucExprsEta <- NULL
  structExprsEta      <- NULL
  selectStrucExprs    <- parTable$op == "~" & parTable$lhs %in% etas

  for (eta in etas) {
    selectStrucExprsEta <- selectStrucExprs & parTable$lhs == eta
    structExprsEta      <- parTable[selectStrucExprsEta, ]

    for (xi in structExprsEta$rhs) { # xi can be an interaction term
      selectRows  <- selectStrucExprsEta & parTable$rhs == xi
      scalingCoef <- sqrt(variances[[xi]]) / sqrt(variances[[eta]])
      gamma       <- parTable[selectRows, selectCols]

      parTable[selectRows, selectCols] <- gamma * scalingCoef
    }
  }

  # (Co-) Variances of xis
  selectCovXis <- parTable$op == "~~" & parTable$lhs %in% xis
  selectRows   <- NULL
  combosXis    <- getUniqueCombos(xis, match = TRUE)

  for (i in seq_len(nrow(combosXis))) {
    xis         <- combosXis[i, , drop = TRUE]
    selectRows  <- selectCovXis & parTable$lhs %in% xis & parTable$rhs %in% xis
    scalingCoef <- sqrt(variances[[xis[[1]]]]) * sqrt(variances[[xis[[2]]]])

    if (xis[[1]] != xis[[2]]) {
      selectRows <- selectRows & parTable$lhs != parTable$rhs
    }

    covXis <- parTable[selectRows, selectCols]
    parTable[selectRows, selectCols] <- covXis / scalingCoef
  }

  # Residual Variances etas
  selectRows <- NULL
  residual   <- NULL

  for (eta in etas) {
    selectRows <- parTable$lhs == eta & parTable$op == "~~" & parTable$rhs == eta
    residual   <- parTable[selectRows, selectCols]

    parTable[selectRows, selectCols] <- residual / variances[[eta]]
  }

  # residual variances inds
  for (ind in allInds) {
    selectRows <- parTable$lhs == ind & parTable$op == "~~" & parTable$rhs == ind
    residual   <- parTable[selectRows, selectCols]

    parTable[selectRows, selectCols] <- residual / variances[[ind]]
  }

  # recalculate variance of interaction terms
  # and rescale coefficients for interaction terms
  parTable <- var_interactions(parTable)
  for (xz in intTerms) {
    selectRows <- parTable$rhs == xz & parTable$op == "~"
    varXZ      <- parTable[parTable$lhs == xz & parTable$op == "~~" &
                           parTable$rhs == xz, "est"]
    gamma      <- parTable[selectRows, selectCols]

    parTable[selectRows, selectCols] <- gamma / sqrt(varXZ) # unstandardizing, since varXZ != 1 | cov(X, Z) != 0
  }

  # recalculate custom parameters
  constrExprs <- sortConstrExprsFinalPt(parTable)
  parTable <- parTable[parTable$op != ":=", ]

  for (i in seq_len(NROW(constrExprs))) {
    row <- constrExprs[i, , drop=FALSE]

    expr      <- parse(text=constrExprs[i, "rhs"])
    labelList <- parTableLabelsToList(parTable) # must be updated for each iteration

    oldVal <- row$est
    newVal <- eval(expr, envir = labelList)

    ratio  <- newVal / oldVal
    values <- row[ , selectCols]

    row[, selectCols] <- values * ratio

    parTable <- rbind(parTable, row)
  }

  parTable$z.value  <- parTable$est / parTable$std.error
  parTable$p.value  <- 2 * stats::pnorm(-abs(parTable$z.value))
  parTable$ci.lower <- parTable$est - CI_WIDTH * parTable$std.error
  parTable$ci.upper <- parTable$est + CI_WIDTH * parTable$std.error

  modsemParTable(parTable)
}


#' Inspect model information
#'
#' @param object fitted model to inspect
#' @param what what to inspect
#' @param ... Additional arguments passed to other functions
#' @description function used to inspect fitted object. Similar to \code{lavaan::lavInspect}
#' argument \code{what} decides what to inspect
#' @details For \code{\link{modsem_pi}} objects, it is just a wrapper for \code{lavaan::lavInspect}.
#'  For \code{\link{modsem_da}} objects an internal function is called, which takes different
#'  keywords for the \code{what} argument.
#' @export
modsem_inspect <- function(object, what = NULL, ...) {
  UseMethod("modsem_inspect")
}



#' Wrapper for vcov
#'
#' @param object fitted model to inspect
#' @param ... additional arguments
#' @description wrapper for \code{vcov}, to be used with \code{modsem::modsem_vcov}, since
#' \code{vcov} is not in the namespace of \code{modsem}, but \code{stats}.
#' @export
modsem_vcov <- function(object, ...) {
  vcov(object, ...)
}


#' Wrapper for coef
#'
#' @param object fitted model to inspect
#' @param ... additional arguments
#' @description wrapper for \code{coef}, to be used with \code{modsem::modsem_coef}, since
#' \code{coef} is not in the namespace of \code{modsem}, but \code{stats}.
#' @export
modsem_coef <- function(object, ...) {
  coef(object, ...)
}


#' Wrapper for nobs
#'
#' @param object fitted model to inspect
#' @param ... additional arguments
#' @description wrapper for \code{nobs}, to be used with \code{modsem::modsem_nobs}, since
#' \code{nobs} is not in the namespace of \code{modsem}, but \code{stats}.
#' @export
modsem_nobs <- function(object, ...) {
  nobs(object, ...)
}


#' Predict From \code{modsem} Models
#'
#' A generic function (and corresponding methods) that produces predicted
#' values or factor scores from \code{\link{modsem}} models.
#'
#' @param object An object of class \code{modsem_pi} or \code{modsem_da},
#'   respectively.
#' @param newdata Optional \code{data.frame} or matrix. If supplied, factor
#'   scores are computed for these observations instead of the data used during
#'   model estimation. Must contain all observed-variable columns required by
#'   the model; missing values are handled via FIML. Currently only supported
#'   by the \code{\link{modsem_da}} method.
#' @param method Character. Scoring method. One of \code{"EBM"} (Empirical
#'   Bayes Modal; MAP estimate of the posterior), \code{"ML"} (maximum
#'   likelihood), \code{"Regression"} (alias for \code{"EBM"}), or
#'   \code{"Bartlett"} (alias for \code{"ML"}). For \code{\link{modsem_da}}
#'   objects this selects the optimisation-based scoring algorithm. For
#'   \code{\link{modsem_pi}} objects the argument is forwarded to
#'   \code{lavaan::lavPredict}, which accepts the same \code{method} values.
#' @param type Character. Which scores to return: \code{"lv"} for latent
#'   variable scores only (default), \code{"ov"} for model-implied observed
#'   variable scores, or \code{"all"} for both combined. For
#'   \code{\link{modsem_da}} objects this is handled directly; for
#'   \code{\link{modsem_pi}} objects it is forwarded to
#'   \code{lavaan::lavPredict}, which supports the same \code{type} values.
#' @param standardized Logical. If \code{TRUE}, each score column is
#'   standardised (mean 0, SD 1) before returning. Currently only used by the
#'   \code{\link{modsem_da}} method.
#' @param se Logical. If \code{TRUE}, attach standard-error and asymptotic
#'   covariance attributes for \code{\link{modsem_da}} predictions. Standard
#'   errors are returned in the \code{"se"} attribute and covariance matrices in
#'   the \code{"acov"} attribute.
#' @param drop.list.single.group Logical. If \code{TRUE}, covariance attributes
#'   for single-group \code{\link{modsem_da}} predictions drop the outer group
#'   list, matching \code{lavaan::lavPredict}.
#' @param ... Further arguments passed to \code{lavaan::predict} for
#'   \code{\link{modsem_pi}} objects; currently ignored by the
#'   \code{\link{modsem_da}} method.
#'
#' @return
#' * For \code{\link{modsem_pi}}: whatever \code{lavaan::predict()} returns,
#'   which is usually a matrix of factor scores.
#' * For \code{\link{modsem_da}}: a numeric matrix with one row per
#'   observation. Columns depend on \code{type}: latent variable scores
#'   (\code{"lv"}), model-implied observed-variable scores (\code{"ov"}), or
#'   both (\code{"all"}). Columns are optionally standardised if
#'   \code{standardized = TRUE}.
#'
#' @examples
#' m1 <- '
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#'
#' est_dca <- modsem(m1, oneInt, method = "dblcent")
#' head(modsem_predict(est_dca))
#'
#' \dontrun{
#' est_lms <- modsem(m1, oneInt, method = "lms")
#' head(modsem_predict(est_lms))
#' }
#'
#' @export
modsem_predict <- function(object, ...) {
  UseMethod("modsem_predict")
}


#' Check if model object has interaction terms
#'
#' @param object An object of class \code{modsem_pi} or \code{modsem_da},
#'   respectively.
#'
#' @return Logical. \code{TRUE} if the model has an interaction term,
#'   otherwise it returns \code{FALSE}.
#'
#' @examples
#' m1 <- '
#' # Outer Model
#'   X =~ x1 + x2 + x3
#'   Z =~ z1 + z2 + z3
#'   Y =~ y1 + y2 + y3
#'
#' # Inner Model
#'   Y ~ X + Z + X:Z
#' '
#'
#' est_dca <- modsem(m1, oneInt, method = "dblcent")
#' is_interaction_model(est_dca)
#'
#' \dontrun{
#' est_lms <- modsem(m1, oneInt, method = "lms")
#' is_interaction_model(est_lms)
#' }
#'
#' @export
is_interaction_model <- function(object) {
  UseMethod("is_interaction_model")
}

devtools::load_all()

tpb.lin <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT <~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH <~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
'

testthat::expect_no_condition({
  fit_lav <- lavaan::sem(tpb.lin, TPB, optim.gradient = "numerical")
  fit_mod <- modsem(tpb.lin, TPB, method = "lms", optimize = FALSE, max.step = 10000, convergence.abs = 1e-6)
  head(modsem_predict(fit_mod))
  head(modsem_predict(fit_mod, newdata = TPB[1:100, ]))
})

coef_lav <- lavaan::coef(fit_lav)
coef_mod <- coef(fit_mod)[names(coef_lav)]

msr_lav <- lavaan::fitMeasures(fit_lav)
msr_mod <- fit_modsem_da(fit_mod)

testthat::expect_equal(c(coef_lav), c(coef_mod), tol = 1e-3)
testthat::expect_equal(msr_lav[["chisq"]], msr_mod[["chisq.value"]], tol = 1e-3)
testthat::expect_equal( # modsem does not fix composite variances like lavaan
  msr_mod[["chisq.df"]] + msr_mod[["npar"]] - msr_lav[["npar"]] - 15, # -15 for mean structure
  msr_lav[["df"]]
)

# Test standardization of indicator covariances
std <- standardized_estimates(fit_mod)
att <- c("att1", "att2", "att3", "att4", "att5")
S <- cor(TPB[att])

for (i in seq_along(att)) {
  x <- att[[i]]

  for (j in seq_len(i)) {
    y <- att[[j]]

    observed <- S[x, y]
    expected <- std[
      (std$lhs == x & std$op == "~~" & std$rhs == y) |
      (std$lhs == y & std$op == "~~" & std$rhs == x), "est"
    ]

    testthat::expect_equal(observed, expected)
  }
}


tpb.int <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT <~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH <~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
'

testthat::expect_no_condition(
  modsem(tpb.int, TPB, method = "lms", nodes = 32)
)

testthat::expect_error(
  modsem(tpb.int, TPB, method = "qml"),
  regexp = "Composite constructs .*"
)

testthat::expect_true(!is.null(modsem_inspect(fit_mod)$wmat))

devtools::load_all()
m1 <- "
# Outer Model
  X =~ x1
  Z =~ z1
  x1 ~~ 0.1 * x1
  Y =~ y1

# Inner model
  Y ~~ Y

  Y ~ a * X + c * Z + b * X:Z

  a == b ^ 2 + 0.01 * c

  a2 := sqrt(a - 0.0001)^2
  coef := sqrt(a * 2 + b^2)
"

est1 <- modsem(m1, oneInt, method = "lms", optimize = TRUE, verbose = TRUE,
               convergence.abs = 1e-2, R.max = 50000)
plot_interaction("X", "Z", "Y", -3:3, c(-0.5, 0.5), model = est1, standardized=TRUE)
print(summary(est1, adjusted.stat = TRUE))
plot_surface(x = "X", z = "Z", y = "Y", model = est1)
std_est1_mc <- standardize_model(est1, monte.carlo=TRUE)
std_est1_delta <- standardize_model(est1, monte.carlo=FALSE)
mdata <- modsem_predict(est1, center.data = FALSE)
ndata <- modsem_predict(est1, newdata = oneInt[1:100, ], center.data = FALSE)

# test constraints and custom parameters
coefs <- coef(est1)
a <- coefs[["a"]]
b <- coefs[["b"]]
c <- coefs[["c"]]
testthat::expect_equal(a, b ^ 2 + 0.01 * c)

test_coefs <- function(est) {
  coefs <- coef(est)
  a <- coefs[["a"]]
  b <- coefs[["b"]]
  c <- coefs[["coef"]]
  testthat::expect_equal(c, sqrt(a * 2 + b^2))
}

test_coefs(est1)
test_coefs(std_est1_delta)
test_coefs(std_est1_mc)

m1 <- "
# Outer Model
  X =~ x1
  Z =~ z1
  x1 ~~ 0.1 * x1
  Y =~ y1

# Inner model
  Y ~ a * X + a * Z
  X ~~ varX * X
  Y ~~ Y
  Y ~ b * X:Z + 0.05 * X:X
  b == a * 1.2

  coef := sqrt(a * 2 + b^2)
"

testthat::expect_warning(
  modsem(m1, oneInt, method = "lms"),
  regexp = "Variances and covariances .*"
)

testthat::expect_no_warning(
  modsem(m1, oneInt, method = "lms", cov.syntax = "")
)

# PROBLEM:
#   I have no clue why, but changing the ordering of how the interaction terms
#   are specified, ends up changing the number of iterations (and results ever
#   so slightly) -- even though the matrices are exactly the same. This can be
#   seen through the fact that the starting loglikelihoods are the same (if optimized)
#   indicating that the matrices are the same (i.e,. produce the same results, when
#   given the same values).
# ANSWER:
#   slightly different results from lavaan, giving slightly different
#   starting parameters
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  LATENT_VAR_ATT =~ a1 * att1 + a2 * att2 + att3 + att4 + att5
  SN =~ s1 * sn1 + sn2
  PBC =~ p1 * pbc1 + pbc2 + pbc3
  INT =~ i1 * int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  # Covariances
  LATENT_VAR_ATT ~~ SN + PBC
  PBC ~~ SN
  # Causal Relationsships
  INT ~ gamma_int_att * LATENT_VAR_ATT + b * SN + b * PBC
  BEH ~ 0.2 * INT + a * PBC
  BEH ~ PBC:INT
  gamma_int_att == a
  p1 == 1 + 0.1
  a2 == 1
  s1 == 1
  i1 == 1
  my_custom_parameter := a * 2
"

covModel <- '
PBC ~ a * LATENT_VAR_ATT + SN
'

testthat::expect_warning({
  est2 <- modsem(tpb, TPB, method = "lms", verbose = TRUE, convergence.abs = 0.001,
                 cov.syntax = covModel, nodes = 16, robust.se = TRUE)
}, regexp = "between endogenous variables")

ust_pt <- parameter_estimates(est2)
std_pt <- standardized_estimates(est2, monte.carlo=FALSE)
testthat::expect_true(all(is.na(ust_pt[ust_pt$label == "p1", "std.error"])))

# Check constraints
label <- "my_custom_parameter"
u_est <- ust_pt[ust_pt$label == label, "est"]
s_est <- std_pt[std_pt$label == label, "est"]
u_a   <- unique(ust_pt[ust_pt$label == "a", "est"])
s_a   <- unique(std_pt[std_pt$label == "a", "est"])
testthat::expect_true(length(s_a) > 1)
testthat::expect_true(length(u_a) == 1)
testthat::expect_equal(u_est, 2 * u_a)
testthat::expect_equal(s_est, 2 * s_a[1])
testthat::expect_equal(
  unname(calcVarParTable(c("PBC", "INT", "BEH"), std_pt)), rep(1, 3)
)

plot_interaction(x = "INT", z = "PBC", y = "BEH", vals_z = c(-0.5, 0.5), model = est2)
print(summary(est2))
var_interactions(est2)
print(vcov(est2)[8:14, 8:14])
modsem_inspect(est2)
print(coef(est2))
coefficients(est2)
cat("Number of observations using 'nobs()':", nobs(est2), "\n")

tpb2 <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  # Covariances
  ATT ~~ SN + PBC
  PBC ~~ SN
  # Causal Relationsships
  INT ~ a * ATT + b * SN + c * PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

testthat::expect_warning(modsem(tpb, TPB, method = "lms", convergence.abs = 1000,
                                nodes = 16, calc.se = FALSE),
                         regexp = "between exogenous and endogenous variables")
modsem_predict(est2)


testthat::expect_true({
  parTable <- summary(est2, H0 = FALSE, standardized = TRUE,
                      intercepts = TRUE)$parTable
  all(is.na(parTable[parTable$op == "~1", "std.all"]))
})


testthat::expect_true({
  parTable <- summary(est2, H0 = FALSE, standardized = TRUE)$parTable
  all(is.na(parTable[parTable$op == "~1", "std.all"]))
})

tpb_uk <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att3 + att2 + att1 + att4
  SN =~ sn4 + sn2 + sn3 + sn1
  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
  INT =~ int2 + int1 + int3 + int4
  BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
# Causal Relationships
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

testthat::expect_warning(
modsem(tpb_uk, TPB_UK, method = "lms", nodes=32, max.iter=10, calc.se=FALSE)
)


m1 <- "
# Outer Model
  X =~ x1
  Z =~ z1
  x1 ~~ 0.1 * x1
  Y =~ y1

# Inner model
  Y ~ X + Z
  Y ~ X:Z
"

testthat::expect_warning(
  modsem(m1, oneInt, method = "lms", optimizer = "ssjj"),
  regexp = "Model estimation failed"
)

m1 <- "
# Outer Model
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

# Inner model
  Y ~ X + Z + X:Z + X:X
"

est3 <- modsem(m1, oneInt, method = "lms")
std <- standardized_estimates(est3)
testthat::expect_equal(unname(calcVarParTable("Y", std)), 1)
summarize_partable(std)


m1 <- "
# Outer Model
  X =~ x1 + x2
  Z =~ z1 + z2
  .Y =~ y1 + y2

# Inner model
  .Y ~ c * X + Z + X:Z

  X~~ a*X + b*Z

  a == 1
  b == .2
  c > 0.7
"

est_m1 <- modsem(m1, oneInt, method = "lms", cov.syntax = "")
coef_m1 <- coef(est_m1)
summary(est_m1)
print(modsem_inspect(est_m1, what = c("vcov.all", "theta", "lambda", "psi", "information", "coefficients.all")))

# test constraints
testthat::expect_equal(coef_m1[["a"]], 1)
testthat::expect_equal(coef_m1[["b"]], .2)
testthat::expect_equal(coef_m1[["c"]], .7)
testthat::expect_equal(est_m1$model$params$bounds$lower[["c"]], 0.7)

m2 <- "
# Outer Model
  X =~ x1 + x2
  Z =~ z1 + z2
  .Y =~ y1 + y2

# Inner model
  .Y ~ c * X + Z + X:Z

  X~~ a*X + b*Z

  a == 1
  b == .2
  c > a * 2
"

testthat::expect_error(
  modsem(m2, oneInt, method = "lms", cov.syntax = ""),
  regexp = "Dynamic constraints .*"
)


m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ NA*y1 + y2 + y3
  Z =~ z1 + z2 + z3
  Y ~~ 1*Y

  # Inner model
  Y ~ X + Z
'

summary(modsem(m1, oneInt, method = "lms"), H0 = FALSE)

devtools::load_all()
library(mvtnorm)


cov_xz <- 0.8
cov_mx <- 0.7
cov_mz <- 0.4
var_x <- 1.2
var_z <- 1.8
var_m <- 0.9

sigma <- matrix(
 c( var_x, cov_xz, cov_mx,
   cov_xz,  var_z, cov_mz,
   cov_mx, cov_mz,  var_m),
  nrow = 3, ncol = 3
)

set.seed(139427)
N <- 1e7
XI <- rmvnorm(N, mean = rep(0, 3), sigma = sigma)

XI
X <- XI[, 1]
Z <- XI[, 2]
M <- XI[, 3]

XX <- X * X
XZ <- X * Z
ZZ <- Z * Z
MX <- M * X
MZ <- M * Z

pt.cov <- data.frame(
  lhs = c("X", "X", "X", "Z", "Z", "M"),
  op = "~~",
  rhs = c("X", "Z", "M", "Z", "M", "M"),
  est = c(var_x, cov_xz, cov_mx, var_z, cov_mz, var_m)
)
names <- c("X", "Z", "M", "X:X", "X:Z", "Z:Z", "M:X", "M:Z")

pt.int <- data.frame(
  lhs = "Y",
  op = "~",
  rhs = names,
  est = 0
)

pt <- rbind(pt.cov, pt.int)
pt.var <- var_interactions(pt)


INT <- data.frame(X, Z, M, XX, XZ, ZZ, MX, MZ)
colnames(INT) <- names
observed <- cov(INT)
expected <- observed
expected[TRUE] <- NA
for (x in colnames(observed)) {
  for (y in rownames(observed)) {
    match <- pt.var[pt.var$op == "~~" & pt.var$lhs == x & pt.var$rhs == y, "est"]

    if (length(match))
      expected[y, x] <- expected[x, y] <- match
  }
}

testthat::expect_equal(round(expected, 2), round(observed, 2))

#' @export
var_interactions.data.frame <- function(object, ignore.means = FALSE,
                                        monte.carlo = FALSE, mc.reps = 1e6, ...) {

  # Preparation
  parTable <- addMissingGroups(object)
  parTable <- removeInteractionVariances(fillColsParTable(parTable))

  ## interaction (and square) terms explicitly declared in the model
  out <- NULL
  for (g in getGroupsParTable(parTable)) {
    parTable.g <- parTable[parTable$group == g, , drop = FALSE]
    out <- rbind(out, varInteractionsGroup(parTable.g, ignore.means = ignore.means,
                                           monte.carlo = monte.carlo,
                                           mc.reps = mc.reps, group = g, ...))
  }

  rbind(out, getZeroGroupParTable(parTable))
}


varInteractionsGroup <- function(parTable, ignore.means = FALSE,
                                 monte.carlo = FALSE, mc.reps = 1e6,
                                 group = group, ...) {
  intTerms <- unique(parTable$rhs[parTable$op == "~" & grepl(":", parTable$rhs)])
  if (length(intTerms) == 0L)
    return(modsemParTable(parTable))                 # nothing to do

  getIntTermLength <- \(xz) length(stringr::str_split(xz, pattern = ":")[[1L]])
  nway <- max(vapply(intTerms, FUN.VALUE = integer(1L), FUN = getIntTermLength))

  if (nway <= 2L && !monte.carlo) {
    varInteractionsAnalytic(
      parTable = parTable,
      intTerms = intTerms,
      ignore.means = ignore.means,
      group = group,
      ...
    )

  } else {
    varInteractionsMonteCarlo(
      parTable = parTable,
      intTerms = intTerms,
      ignore.means = ignore.means,
      mc.reps = mc.reps,
      group = group,
      ...
    )

  }
}


varInteractionsAnalytic <- function(parTable, intTerms, ignore.means = FALSE, group = 0L, ...) {
  # Gather first‑ and second‑order moments for the base variables --
  vars <- sort(unique(unlist(strsplit(intTerms, ":", fixed = TRUE))))

  if (ignore.means) mu <- stats::setNames(numeric(length(vars)), vars)
  else              mu <-  sapply(vars, getMean, parTable = parTable)

  getCov <- function(a, b) calcCovParTable(a, b, parTable)
  Sigma  <- outer(vars, vars, getCov)
  dimnames(Sigma) <- list(vars, vars)

  # Helper functions for the Gaussian fourth‑moment identity
  cov_PX <- function(i, j, k)
    mu[i] * Sigma[j, k] + mu[j] * Sigma[i, k]

  cov_PP <- function(i, j, k, l)
    Sigma[i, k] * Sigma[j, l] + Sigma[i, l] * Sigma[j, k] +
    mu[i] * mu[j] * Sigma[k, l] + mu[k] * mu[l] * Sigma[i, j] +
    mu[i] * mu[k] * Sigma[j, l] + mu[j] * mu[k] * Sigma[i, l] +
    mu[i] * mu[l] * Sigma[j, k] + mu[j] * mu[l] * Sigma[i, k]

  # Split product names once to avoid repetition
  prodPairs  <- lapply(intTerms, function(s) sort(strsplit(s, ":", fixed = TRUE)[[1L]]))
  names(prodPairs) <- intTerms                # keep original order/labels

  # Assemble rows to append
  rows <- vector("list", 0L)

  for (P in names(prodPairs)) {

    ij <- prodPairs[[P]];  i <- ij[1L];  j <- ij[2L]

    ## variance of the product term itself
    varP <- cov_PP(i, j, i, j)
    rows[[length(rows) + 1L]] <-
      data.frame(lhs = P, op = "~~", rhs = P, label = "",
                 group = group, est = varP, std.error = NA,
                 z.value = NA, p.value = NA, ci.lower = NA,
                 ci.upper = NA)

    ## covariances with the *original* variables
    for (k in vars) {
      rows[[length(rows) + 1L]] <-
        data.frame(lhs = k, op = "~~", rhs = P, label = "", group = group,
                   est = cov_PX(i, j, k), std.error = NA, z.value = NA,
                   p.value = NA, ci.lower = NA, ci.upper = NA)
    }
  }

  # Covariances among the declared product terms
  if (length(prodPairs) > 1L) {
    comb <- utils::combn(names(prodPairs), 2L)
    for (cc in seq_len(ncol(comb))) {
      P1 <- comb[1L, cc];  P2 <- comb[2L, cc]
      ij <- prodPairs[[P1]];  kl <- prodPairs[[P2]]
      rows[[length(rows) + 1L]] <-
        data.frame(lhs = P1, op = "~~", rhs = P2, label = "",
                   group = group, est = cov_PP(ij[1L], ij[2L], kl[1L], kl[2L]),
                   std.error = NA, z.value = NA, p.value = NA,
                   ci.lower = NA, ci.upper = NA)
    }
  }

  # Bind & return
  parTable <- rbind(parTable, do.call(rbind, rows))
  modsemParTable(parTable)
}


varInteractionsMonteCarlo <- function(parTable, intTerms, ignore.means = FALSE,
                                      mc.reps = 1e6, group = 0L, ...) {
  # Gather first‑ and second‑order moments for the base variables --
  intVars  <- unique(unlist(strsplit(intTerms, ":", fixed = TRUE)))
  xis      <- unique(parTable[parTable$op == "~", "rhs"])
  vars     <- sort(unique(c(intVars, xis)))

  if (ignore.means) mu <- stats::setNames(numeric(length(vars)), vars)
  else              mu <-  sapply(vars, getMean, parTable = parTable)

  getCov <- function(a, b) calcCovParTable(a, b, parTable)
  Sigma  <- outer(vars, vars, getCov)
  dimnames(Sigma) <- list(vars, vars)

  XI <- as.data.frame(mvtnorm::rmvnorm(mc.reps, mean = mu, sigma = Sigma))

  for (intTerm in intTerms) {
    elems <- stringr::str_split(intTerm, pattern = ":")[[1L]]

    XI[[intTerm]] <- multiplyIndicatorsCpp(XI[elems])
  }

  Sigma.xz <- stats::cov(XI)
  isXZ <- \(x) x %in% intTerms
  isXi <- \(x) x %in% xis
  isOK <- \(x, z)  (isXZ(x) && isXi(z)) || (isXZ(z) && isXi(x))

  for (i in seq_len(nrow(Sigma.xz))) for (j in seq_len(i)) {
    x   <- rownames(Sigma.xz)[[i]]
    z   <- colnames(Sigma.xz)[[j]]
    est <- Sigma.xz[x, z]

    if (!isOK(x, z))
      next

    newRow <- data.frame(lhs = x, op = "~~", rhs = z, label = "",
                         group = group, est = est, std.error = NA, z.value = NA,
                         p.value = NA, ci.lower = NA, ci.upper = NA)

    parTable <- rbind(parTable, newRow)
  }

  parTable
}

#' Estimate baseline model for \code{modsem} models
#'
#' @param object An object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}}.
#' @param warn_no_interaction Logical. If `TRUE`, a warning is issued if no interaction terms are found in the model.
#' @param ... Additional arguments passed to the `modsem_da` function, overriding
#'   the arguments in the original model.
#' @description Estimates a baseline model (H0) from a given model (H1).
#' The baseline model is estimated by removing all interaction terms from the model.
#' @rdname estimate_h0
#' @export
#' @examples
#' \dontrun{
#' m1 <- "
#'  # Outer Model
#'  X =~ x1 + x2 + x3
#'  Y =~ y1 + y2 + y3
#'  Z =~ z1 + z2 + z3
#'
#'  # Inner model
#'  Y ~ X + Z + X:Z
#' "
#'
#' # LMS approach
#' est_h1 <- modsem(m1, oneInt, "lms")
#' est_h0 <- estimate_h0(est_h1, calc.se=FALSE) # std.errors are not needed
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#'
#' # Double centering approach
#' est_h1 <- modsem(m1, oneInt, method = "dblcent")
#' est_h0 <- estimate_h0(est_h1, oneInt)
#'
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#'
#' # Constrained approach
#' est_h1 <- modsem(m1, oneInt, method = "ca")
#' est_h0 <- estimate_h0(est_h1, oneInt)
#'
#' compare_fit(est_h1 = est_h1, est_h0 = est_h0)
#' }
#' @export
estimate_h0 <- function(object, warn_no_interaction = TRUE, ...) {
  UseMethod("estimate_h0", object)
}


estimateH0Cached <- memoise::memoise(estimate_h0)


#' @describeIn estimate_h0 Estimate baseline model for \code{\link{modsem_da}} objects
#' @export
estimate_h0.modsem_da <- function(object, warn_no_interaction = TRUE, ...) {
  argList    <- object$args
  parTable   <- object$model$info$group.info$parTable
  data       <- object$model$data.raw
  method     <- object$method
  cov.syntax <- object$model$info$group.info$cov.syntax

  newArgList <- list(...)
  newArgNames <- intersect(names(argList), names(newArgList))
  argList[newArgNames] <- newArgList[newArgNames]

  tryCatch({
    ovints <- stringr::str_replace_all(
      string = object$model$info$group.info$ovIntTerms,
      pattern = ":", replacement = OP_OV_INT
    )

    isInteractionTerm <-
      parTable$op == "~" &
      (parTable$rhs %in% ovints | grepl(":", parTable$rhs))

    constrainedParTable <- redefineIntTermLabels(
      parTable = parTable,
      isInteractionTerm = isInteractionTerm
    )

    # TODO: Do something similar for cov.syntax

    islvint  <- grepl(":", constrainedParTable$rhs)
    haslvint <- any(islvint)
    isovint  <- constrainedParTable$rhs %in% ovints & constrainedParTable$op == "~"
    hasovint <- length(ovints)

    constrainedParTable[isovint, "mod"] <- "0"
    constrainedParTable <- constrainedParTable[!islvint, , drop = FALSE]

    if (!haslvint && !hasovint) {
      mod_warnif(warn_no_interaction, paste0("No interaction terms found in the model. ",
             "The baseline model is identical to the original model, ",
             "and won't be estimated!"))
      return(NULL)
    }

    syntax <- parTableToSyntax(constrainedParTable)
    argList <- c(
      list(model.syntax = syntax, data = data, method = method,
           cov.syntax = cov.syntax), argList
    )

    do.call(modsem_da, args = argList)
  },
  error = function(e) {
    mod_msg_warn(paste0("Null model could not be estimated. ",
             "Error message: ", e$message))
    NULL
  })
}


#' @describeIn estimate_h0 Estimate baseline model for \code{\link{modsem_pi}} objects
#' @param reduced Should the baseline model be a reduced version of the model?
#'   If \code{TRUE}, the latent product term and its (product) indicators are kept in the model,
#'   but the interaction coefficients are constrained to zero. If \code{FALSE}, the
#'   interaction terms are removed completely from the model. Note that the models will no longer be
#'   nested, if the interaction terms are removed from the model completely.
#' @export
estimate_h0.modsem_pi <- function(object, warn_no_interaction = TRUE,
                                  reduced = TRUE, ...) {
  pars       <- parameter_estimates(object)
  input      <- object$input
  method     <- object$method
  syntax     <- input$syntax
  data       <- input$data
  parTable   <- input$parTable

  # Get arguments
  lavArgs    <- input$lavArgs
  modsemArgs <- input$modsemArgs
  lavArgs$meanstructure <- any(pars$op == "~1")
  argList    <- c(modsemArgs, lavArgs)

  newArgList <- list(...)
  newArgNames <- intersect(names(argList), names(newArgList))
  argList[newArgNames] <- newArgList[newArgNames]

  if (!is.null(lavArgs$cluster)) {
    mod_msg_warn("Baseline model cannot (yet) be estimated with clustering!")
    return(NULL)
  }

  tryCatch({
    if (!parTableHasInteraction(parTable)) {
      mod_warnif(warn_no_interaction,
             paste0("No interaction terms found in the model. ",
             "The baseline model is identical to the original model, ",
             "and won't be estimated!"))
      return(NULL)

    } else if (parTableHasHigherOrderInteraction(parTable)) {
      mod_msg_warn("Unable to estimate baseline model for models with higher-order interaction terms!")
      return(NULL)
    }

    isInteractionTerm <- grepl(":", parTable$rhs)
    constrainedParTable <- redefineIntTermLabels(
      parTable = parTable,
      isInteractionTerm = isInteractionTerm
    )

    # We don't want to remove the interaction term, since we want the baseline model
    # to have the same variables as the original model (inluding the product indicators).
    # However, we want to constrain the interaction terms to zero.

    isInteractionTerm <- grepl(":", constrainedParTable$rhs) # recompute as parTable might have changed
    if (reduced) constrainedParTable[isInteractionTerm, "mod"] <- "0"
    else         constrainedParTable <- constrainedParTable[!isInteractionTerm, ]

    syntax <- parTableToSyntax(constrainedParTable)
    argList <- c(list(model = syntax, data = data, method = method), argList)

    fit <- do.call(modsem_pi, args = argList)

    if (is.null(extract_lavaan(fit))) {
      mod_msg_warn("`lavaan` failed when estimating the model!")
      return(NULL)
    }

    fit
  },
  error = function(e) {
    mod_msg_warn(paste0("Baseline model could not be estimated: ", e$message))
    NULL
  })
}


parTableHasInteraction <- function(parTable) {
  any(grepl(":", parTable$rhs) & parTable$mod != "0")
}


parTableHasHigherOrderInteraction <- function(parTable) {
  any(grepl(":", parTable$rhs) & parTable$op == "=~")
}


redefineIntTermLabels <- function(parTable, isInteractionTerm = NULL) {
  if (is.null(isInteractionTerm))
    isInteractionTerm <- grepl(":", parTable$rhs)

  constrainedParTable <- parTable

  # If there are any label modifiers, we must not remove them completely
  # from the model, as they might be used elsewhere in the model
  preExistingMod <- constrainedParTable[isInteractionTerm, "mod"]
  preExistingMod <- preExistingMod[preExistingMod != ""]
  preExistingLab <- unique(preExistingMod[!canBeNumeric(preExistingMod)])

  # Redine the labels as zero, if they exist
  if (length(preExistingLab)) {
    redefinedLabels <- data.frame(lhs = preExistingLab, op  = ":=", rhs = "0",
                                  mod = preExistingLab, group = 0)
    cols <- intersect(colnames(redefinedLabels), colnames(parTable))
    constrainedParTable <- rbind(redefinedLabels[cols], constrainedParTable[cols])
  }

  constrainedParTable
}

# function for finding index for matching ) to a function call
findFunctionEnd <- function(listTokens, i = 1) {
  if (i > length(listTokens)) {
    mod_msg_stop("No matching parenthesis for function call")

  } else if (listTokens[[i]] == ")") {
    return(i)

  } else {
    findFunctionEnd(listTokens, i + 1)
  }
}


LavEqual <- function(string) {
  if (!is.character(string)) {
    mod_msg_stop(paste0("Expected argument in equal() to be string, got: ", string))
  } else if (length(string) > 1) {
    mod_msg_stop(paste0("Expected a single string in equal(), got: ", string))
  }
  paste0("equal(\"", string, "\")")
}


LavStart <- function(number) {
  if (!is.numeric(number)) {
    mod_msg_stop(paste0("Expected argument in start() to be string, got: ", number))
  } else if (length(number) > 1) {
    mod_msg_stop(paste0("Expected a single number in start(), got: ", number))
  }
  paste0("start(", number, ")")
}


LavConcat <- function(...) {
  as.character(substitute(expression(c(...))))[[2]]
}
  

PlsWithin  <- function(x) {
  paste0("within(", as.character(substitute(x)), ")")
}


PlsBetween <- function(x) {
  paste0("between(", as.character(substitute(x)), ")")
}


modEnv <- rlang::env(
  equal   = LavEqual,
  start   = LavStart,
  c       = LavConcat,
  within  = PlsWithin,
  between = PlsBetween
)

---
title: "non-centered interaction terms (LMS and QML)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{non-centered interaction terms (LMS and QML)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Non-centered interaction terms
Using the LMS and QML approaches it is possible to estimate interaction terms where the
means of the latent variables are not centered (i.e., they have non-zero means).

Here we can see an example using the `TPB` dataset:
```{r}
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC

# Adding Latent Intercepts
  INT ~ 1
  BEH ~ 1
  PBC ~ 1
  SN  ~ 1
  ATT ~ 1
'

est <- modsem(tpb, TPB, method = "lms", nodes = 32)
summary(est)
```

Comparing this to the estimates we get when `PBC` and `INT` have zero means,
we see that the coefficients `BEH~PBC` and `BEH~INT` are drastically changed.
This is not a bug, and is a function of the interaction effect rescaling the
coefficients, when not centered at zero. When using the `standardized_estimates`
function, or `summary(est, standardized = TRUE)` the interaction effect is
centered, and we can see that the coefficients `BEH~PBC` and `BEH~INT` are
rescaled once again.

```{r}
summary(est, standardized = TRUE, centered = TRUE)
```
It is also possible to get the centered solution using the `centered_estimates()` function.
Note, that `centered_estimates()` removes the mean structure of the model all together.

```{r}
centered_estimates(est)
```

---
title: "quadratic effects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{quadratic effects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```
# Quadratic Effects and Interaction Effects

Quadratic effects are essentially a special case of interaction effects—where a variable interacts with itself. As such, all of the methods in `modsem` can also be used to estimate quadratic effects.

Below is a simple example using the `LMS` approach.

```{r}
library(modsem)
m1 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner model
Y ~ X + Z + Z:X + X:X
'

est1_lms <- modsem(m1, data = oneInt, method = "lms")
summary(est1_lms)
```

In this example, we have a simple model with two quadratic effects and one interaction effect. We estimate the model using both the `QML` and double-centering approaches, with data from a subset of the PISA 2006 dataset.

```{r}
m2 <- '
ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
CAREER =~ career1 + career2 + career3 + career4
SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
'

est2_dca <- modsem(m2, data = jordan)
est2_qml <- modsem(m2, data = jordan, method = "qml")
summary(est2_qml)
```

**NOTE**: We can also use the LMS approach to estimate this model, but it will be a lot
slower, since we have to integrate along both `ENJ` and `SC`. In the first example
it is sufficient to only integrate along `X`, but the addition of the `SC:SC` term
means that we have to explicitly model `SC` as a moderator. This means that
we (by default) have to integrate along `24^2=576` nodes. This both affects the
the optimization process, but also dramatically affects the computation time of
the standard errors. To make the estimation process it is possible to reduce the number
of quadrature nodes. Additionally, we can
also pass `mean.observed = FALSE`, constraining the intercepts of the indicators to zero.

```{r}
m2 <- '
ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
CAREER =~ career1 + career2 + career3 + career4
SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
'

est2_lms <- modsem(m2, data = jordan, method = "lms",
                   nodes = 15, mean.observed = FALSE)
summary(est2_lms)
```

---
title: "higher order interactions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{higher order interactions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# LMS approach
As of version `1.0.13`, the `modsem` function supports the estimation of higher order interaction effects.
A general implementation is however only available with the LMS approach. Currently,
estimation of higher order interaction effects using the QML approach is not available,
and only partially available for the product indicator approaches (i.e., `modsem_pi()`).

## Interaction between two higher order latent variables
In `modsem` there are two datasets which are variants of the Theory of Planned Behaviour (`TPB`) dataset.
The `TPB_2SO` contains two second order latent variables,
`INT` (intention) which is a second order latent variable of `ATT` (attitude) and `SN` (subjective norm),
and `PBC` (perceived behavioural control) which is a second order latent variable of
`PC` (perceived control) and `PB` (perceived behaviour).

```{r}
tpb_2so <- '
  # First order latent variables
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PB =~ pb1 + pb2 + pb3
  PC =~ pc1 + pc2 + pc3
  BEH =~ b1 + b2

  # Higher order latent variables
  INT =~ ATT + SN
  PBC =~ PC + PB

  # Structural model
  BEH ~ PBC + INT + INT:PBC
'

est_lms_2so <- modsem(tpb_2so, data = TPB_2SO, method = "lms")
summary(est_lms_2so)
```

## Interaction between a first order and a higher order latent variable 
In the `TPB_1SO` dataset, the `INT` latent variable is a second order latent variable of `ATT`, `SN` and `PBC`.
In this example, we will estimate the interaction effect between the `INT` (higher order latent variable) and `PBC` (first order latent variable).

```{r}
tpb_1so <- '
  # First order latent variables
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PBC =~ pbc1 + pbc2 + pbc3
  BEH =~ b1 + b2

  # Higher order latent variables
  INT =~ ATT + PBC + SN

  # Structural model
  BEH ~ PBC + INT + INT:PBC
'

est_lms_1so <- modsem(tpb_1so, data = TPB_1SO, method = "lms", nodes = 32)
summary(est_lms_1so)
```


# Product Indicator Approaches
As of yet, the `modsem` package does not support using the interaction operator `:` between two higher order latent variables,
when using one of the product indicator approaches (i.e., using `modsem_pi()`).
However, you can still attempt to estimate the interaction effect between two higher order latent variables by specifying the interaction term in
models using the product indicator approaches.

## Interaction between two higher order latent variables
**WARNING:** Please note that the literature on higher order
interactions in product indicator approaches is virtually non-existant, and you
will likely need to experiment with different approaches to find one that works.
As well as experiment with adding constraints to the model.

Here we see the same example as for the LMS approach, where ther is an interaction
effect between two higher order latent variables.

```{r}
tpb_2so <- '
  # First order latent variables
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PB =~ pb1 + pb2 + pb3
  PC =~ pc1 + pc2 + pc3
  BEH =~ b1 + b2

  # Higher order latent variables
  INT =~ ATT + SN
  PBC =~ PC + PB

  # Higher order interaction
  INTxPBC =~ ATT:PC + ATT:PB + SN:PC + SN:PB

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

est_ca <- modsem(tpb_2so, data = TPB_2SO, method = "ca")
summary(est_ca)

est_dblcent <- modsem(tpb_2so, data = TPB_2SO, method = "dblcent")
summary(est_dblcent)
```

## Interaction between a first order and a higher order latent variable 
Here we see the same example as for the LMS approach, where ther is an interaction
effect between a higher order latent variable, and a first order latent variable.

```{r}
tpb_1so <- '
  # First order latent variables
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PBC =~ pbc1 + pbc2 + pbc3
  BEH =~ b1 + b2

  # Higher order latent variables
  INT =~ ATT + PBC + SN

  # Higher order interaction
  INTxPBC =~ ATT:PBC + SN:PBC + PBC:PBC

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

est_ca <- modsem(tpb_1so, data = TPB_1SO, method = "ca")
summary(est_ca)

est_dblcent  <- modsem(tpb_1so, data = TPB_1SO, method = "dblcent")
summary(est_dblcent)
```

---
title: "interactions with ordered data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{interactions with ordered data}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
set.seed(2938472)
```

`modsem` implements a Monte-Carlo correction for LMS and QML
models with ordinal data. Here we refer to these informally as
`MC-LMS-ORD` and `MC-QML-ORD`.

The `MC-LMS-ORD` and `MC-QML-ORD` algorithms are based on 
[Slupphaug, Mehmetoglu, and Mittner (2026)](https://osf.io/preprints/psyarxiv/fwzj6_v1)
For a more direct implementation of the original algorithm, we
recommend checking out the [`plssem` package](https://CRAN.R-project.org/package=plssem).

# Example

Here we ordinalize the data in the `oneInt` dataset.

```{r}
ordinalize <- function(x, probs = c(0, 0.35, 0.7, 1)) {
  x <- (x - mean(x)) / sd(x)
  cut(
    x,
    breaks = stats::quantile(x, probs = probs),
    include.lowest = TRUE,
    ordered_result = TRUE
  )
}

oneIntOrd <- as.data.frame(lapply(oneInt, ordinalize))
```

Now we can estimate our model, indicating which variables are ordinal, using
the `ordered=` argument.

```{r}
model <- "
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  Y ~ X + Z + X:Z
"

fit_lms_ord <- modsem(
  model,
  data = oneIntOrd,
  method = "lms",
  ordered = colnames(oneIntOrd)
)

summary(fit_lms_ord)
```

---
title: "observed variables in the LMS- and QML approach"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{observed variables in the LMS- and QML approach}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- TRUE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# The Latent Moderated Structural Equations (LMS) and the Quasi Maximum Likelihood (QML) Approach

As of version `1.0.16`, observed variables are supported (in most circumstances) for both the LMS and QML approaches.
Here we can see an example using only observed variables.

```{r}
library(modsem)

fit <- modsem('y1 ~ x1 + z1 + x1:z1', data = oneInt, method = "lms")
summary(fit, standardized = TRUE)
```

If you're using an older version of `modsem`, the rest of the vignette details how to handle observed variables
in the LMS and QML approaches. There might also be some cases where where converting observed variables to
latent ones might be a more reliable option. For example, the LMS approach might struggle if it has to integrate
along an observed variable.

## The LMS Approach

For the `LMS` approach, you can use the above-mentioned method in almost all cases, except when using an observed variable as a moderating variable. In the `LMS` approach, you typically select one variable in an interaction term as the moderator. The interaction effect is then estimated via numerical integration at `n` quadrature nodes of the moderating variable. However, this process requires that the moderating variable has an error term, as the distribution of the moderating variable is modeled as \( X \sim N(Az, \varepsilon) \), where \( Az \) is the expected value of \( X \) at quadrature point `k`, and \( \varepsilon \) is the error term. If the error term is zero, the probability of observing a given value of \( X \) will not be computable.

In most instances, the first variable in the interaction term is chosen as the moderator. For example, if the interaction term is `"X:Z"`, `"X"` will usually be chosen as the moderator. Therefore, if only one of the variables is latent, you should place the latent variable first in the interaction term. If both variables are observed, you must specify a measurement error (e.g., `"x1 ~~ 0.1 * x1"`) for the indicator of the first variable in the interaction term.

```{r}
# Interaction effect between a latent and an observed variable
m1 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 + z2 # Z is latent
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ Z:X
'

lms1 <- modsem(m1, oneInt, method = "lms")

# Interaction effect between two observed variables
m2 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed
  x1 ~~ 0.1 * x1 # Specify a variance for the measurement error

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

lms2 <- modsem(m2, oneInt, method = "lms")
summary(lms2)
```
If you forget to specify a measurement error for the indicator of the first variable in the interaction term, you will receive an error message.

```{r, error = TRUE}
m2 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

lms3 <- modsem(m2, oneInt, method = "lms")
```
**Note:** You only get an error message for `X`/`x1`, since `Z` is not modelled as a moderating variable in this example.

## The QML Approach

The estimation process for the `QML` approach differs from the `LMS` approach, and you do not encounter the same issue as in the `LMS` approach. Therefore, you don't need to specify a measurement error for moderating variables.

```{r}
m3 <- '
# Outer Model
  X =~ x1 # X is observed
  Z =~ z1 # Z is observed
  Y =~ y1 # Y is observed

# Inner model
  Y ~ X + Z
  Y ~ X:Z
'

qml3 <- modsem(m3, oneInt, method = "qml")
summary(qml3)
```

---
title: "customizing the output from `plot_interaction()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{customizing the output from `plot_interaction()}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
```

Since `plot_interaction()` just returns a `ggplot` object, the output be customized using a combination of different `ggplot2` functions.

## Baseline Plot

Here we fit a simple model, and create a baseline plot, which we'll customize.

```{r model}
library(modsem)
library(ggplot2)

m <- '
X =~ x1 + x2 + x3
Z =~ z1 + z2 + z3
Y =~ y1 + y2 + y3

Y ~ X + Z + X:Z
'

fit <- modsem(m, data = oneInt, method = "lms")
p <- plot_interaction(x = "X", z = "Z", y = "Y", vals_z = c(-1, 1), model = fit)
print(p)
```

## Changing Titles and Labels

### Title

The title of the plot can be changed using the `ggtitle()` function.

```{r plot-title, echo=FALSE}
p + ggtitle("This is a new title")
```

### Axis Labels

The axis labels can be changed using the `xlab()` and `ylab()` functions.

```{r axis-labels, echo=FALSE}
p + xlab("This is the x-axis") + ylab("This is the y-axis")
```

### Legend Title

Changing the legend title is a bit trickier, but it can be done using the
`guides()` function. Since `plot_interaction()` uses both the `fill` and `color` aesthetic, both must be specified, with the same title.

```{r}
p + guides(fill=guide_legend(title = "New Legend Title"),
           color=guide_legend(title = "New Legend Title"))
```

## Changing Colors

The colors of both the `fill` and `color` aesthetic can be changed using different functions.

### Greyscale

Here we replace the `fill` aesthetic with a greyscale.

```{r}
p + scale_fill_grey()
```
Here we do the same for the `color` aesthetic.
```{r}
p + scale_color_grey()
```

Combining both, we get a greyscale version of the plot.

```{r}
p + scale_color_grey() + scale_fill_grey()
```

**NOTE**: As of version `1.0.14` there is a separate `greyscale` argument, for using a greyscale theme,
simplifying the process.

### Brewer Color palettes

`ggplot2` comes with a few different color palettes, which we can use to the replace the default color palette. Here we have numerous options (you can also download additional packages), but here we can see a few examples using some of the `brewer` palettes.

To see the available palettes you can run this snippet.

```{r, fig.height = 18, fig.width = 8}
# You might need to install the `RColorBrewer` package
RColorBrewer::display.brewer.all()
```

Here we have an example using the `Dark2` palette.
```{r}
# We need to apply the color scale to both the fill and color aesthetics
p + scale_fill_brewer(palette = "Dark2") +
    scale_color_brewer(palette = "Dark2")
```

Here we have an example using the `Accent` palette.
```{r}
# We need to apply the color scale to both the fill and color aesthetics
p + scale_fill_brewer(palette = "Accent") +
    scale_color_brewer(palette = "Accent")
```

### Custom Color Palettes

It is also possible to use custom/manual color palettes, but this is more cumbersome. Here we see an example.

```{r}
custom_colors = c("#E69F00", "#56B4E9")

p + scale_fill_manual(values = custom_colors) +
    scale_color_manual(values = custom_colors)
```

## Changing Line types

Changing the linetypes is a more compliacted than changing the colors in the plot. This is because we have to add a whole new aesthetic to the plot. This can done using the `aes()` function. Behind the scenes, `modsem` uses a variable `cat_z` to form the groups in the output from `plot_interaction`, which is why we pass `linetype = cat_z`.

```{r}
p + aes(linetype = cat_z)
```
Here we can see that we had an unfortunate side effect, where we created a second Legend in our plot. This can however be fixed by renaming the Legend title (see the [Legend Title] section).

```{r}
p + aes(linetype = cat_z) +
    guides(linetype = guide_legend("Z"))
```

## Combinations

Here we can see an example combining the different options.

```{r}
p +
  ggtitle("Customized Plot") + # New Title
  xlab("Values of X") + # New X-axis label
  ylab("Predicted values of Y") + # New y-axis label
  scale_color_grey() + # Add Grey theem
  scale_fill_grey() +
  aes(linetype = cat_z) + # Add linetypes
  guides(color    = guide_legend("Values of Z"), # Change Legend Title
         fill     = guide_legend("Values of Z"),
         linetype = guide_legend("Values of Z"))
```

---
title: "reliability-corrected single items"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{reliability-corrected single items}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

## Reliablity-Corrected Single Items

If wanted, indicators for latent variables can be replaced with reliablity
corrected single items, using Chronbach's $\alpha$. This can either be done
using the `relcorr_single_item` function, returning the altered model syntax
and data, or via the `rcs` argument in `modsem`. Here we can see an example
using the `relcorr_single_item` function:

```{r}
tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

corrected <- relcorr_single_item(syntax = tpb_uk, data = TPB_UK)
corrected
```
Here we can see that `relcorr_single_item` returns a new model syntax, and a
new `data.frame` containing the generated items. Additionally, it also
returns the Chronbach's $\alpha$ and *average variance extraced* (AVE)
for the different constructs in the model. The syntax and data can be
extracted using the `$` operator, and used to estimate the model.

```{r}
syntax <- corrected$syntax
data   <- corrected$data

est_dca <- modsem(syntax, data = data, method = "dblcent")
est_lms <- modsem(syntax, data = data, method="lms", nodes=32)
summary(est_lms)
```

The easiest approach however, is to use the `rcs` argument in the `modsem` function
to call `relcorr_single_item` before estimating the model.

```{r}
est_dca <- modsem(tpb_uk, data = TPB_UK, method = "dblcent", rcs = TRUE)
est_lms <- modsem(tpb_uk, data = TPB_UK, method = "lms", rcs = TRUE)
```

# Choosing Variables
If you don't want to use reliablity-corrected single items for all of the
latent variables in the model, you can use the `choose` argument in
`relcorr_single_item` (or`rcs.choose` in `modsem`) to select which
set of indicators to replace.

```{r}
relcorr_single_item(syntax = tpb_uk, data = TPB_UK,
                    choose = c("ATT", "SN", "PBC", "INT"))

est_dca <- modsem(tpb_uk, data = TPB_UK, method = "dblcent", rcs = TRUE,
                  rcs.choose = c("ATT", "SN", "PBC", "INT", "INT:PBC"))
summary(est_dca)
```

---
title: "simple slopes analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{simple slopes analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Simple Slopes Analysis

Simple slope effects can be plotted using the included `plot_interaction()` function. This function takes a fitted model object and the names of the two variables that are interacting. The function will plot the interaction effect of the two variables, where:

- The x-variable is plotted on the x-axis.
- The y-variable is plotted on the y-axis.
- The z-variable determines at which points the effect of x on y is plotted.

The function will also plot the 95% confidence interval for the interaction effect.
Note that the `vals_z` argument (as well as the values of `x`) are scaled by the
mean and standard deviation of the variables. Unless the `rescale` argument is set to `FALSE`.

Here is a simple example using the double-centering approach:

```{r}
m1 <- "
# Outer Model
  X =~ x1
  X =~ x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)
plot_interaction(x = "X", z = "Z", y = "Y", vals_z = c(-1, 1), model = est1)
```

If you want to see the numerical values of the simple slopes, you can use the `simple_slopes()` function:

```{r}
m1 <- "
# Outer Model
  X =~ x1
  X =~ x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"

est1 <- modsem(m1, data = oneInt)
simple_slopes(x = "X", z = "Z", y = "Y", vals_z = c(-1, 1), model = est1)
```

The `simple_slopes()` function returns a `simple_slopes` object.
It only has two methods/generics:
`print.simple_slopes()`, which prints the simple slopes in a easy-to-read format
and `as.data.frame.simple_slopes()`. The `print()` method will not only
print the predicted values, but also significance tests for the difference between
the slope at the lowest value of `z` and the slope at the highest value of `z`,
as well as significance tests for the slope of `x` at the
different values of `vals_z`.

In the example above, we can see that there is a significant
difference between the slope at at `-1 * sd(Z)` and `+1 * sd(Z)`. Note that
by default `vals_z` is rescaled by the mean and standard deviation of the variable, unless `rescale = FALSE` is set.
This means that the values of `vals_z` are interpreted as standard deviations from the mean of `Z`.

If you want to extract the simple slopes as a `data.frame`, you can use the `as.data.frame()` function:

```{r}
m1 <- "
# Outer Model
  X =~ x1
  X =~ x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"

est1 <- modsem(m1, data = oneInt)
slopes <- simple_slopes(x = "X", z = "Z", y = "Y",
                        vals_z = c(0, 1), model = est1)
as.data.frame(slopes)
```

---
title: "methods"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```
There are a number of approaches for estimating interaction effects in SEM.
In `modsem()`, the `method = "method"` argument allows you to choose which to use.
Different approaches can be categorized into two groups:
Product Indicator (PI) and Distribution Analytic (DA) approaches.

## Product Indicator (PI) Approaches:
- `"ca"` = constrained approach (Algina & Moulder, 2001)
    - Note that constraints can become quite complicated for complex models,
      particularly when there is an interaction including enodgenous variables.
      The method can therefore be quite slow.
- `"uca"` = unconstrained approach (Marsh, 2004)
- `"rca"` = residual centering approach (Little et al., 2006)
- `"dblcent"` = double centering approach (Marsh., 2013)
  - default
- `"pind"` = basic product indicator approach (not recommended)

## Distribution Analytic (DA) Approaches
- `"lms"` = The Latent Moderated Structural equations (LMS) approach, see the [vignette](https://modsem.org/articles/lms_qml.html)
- `"qml"` = The Quasi Maximum Likelihood (QML) approach, see the [vignette](https://modsem.org/articles/lms_qml.html)
- `"mplus"`
  - estimates model through Mplus, if it is installed

```{r, eval = FALSE}

m1 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner model
Y ~ X + Z + X:Z
'

# Product Indicator Approaches
modsem(m1, data = oneInt, method = "ca")
modsem(m1, data = oneInt, method = "uca")
modsem(m1, data = oneInt, method = "rca")
modsem(m1, data = oneInt, method = "dblcent")

# Distribution Analytic Approaches
modsem(m1, data = oneInt, method = "mplus")
modsem(m1, data = oneInt, method = "lms")
modsem(m1, data = oneInt, method = "qml")
```

---
title: "using lavaan functions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{using lavaan functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```
If you're using one of the product indicator approaches, you might want to use some lavaan
functions on top of the estimated lavaan-object.
To do so you can extract the lavaan-object using the `extract_lavaan()` function.

```{r}
library(lavaan)

m1 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner model
Y ~ X + Z + X:Z
'

est <- modsem(m1, oneInt)
lav_object <- extract_lavaan(est)
bootstrap <- bootstrapLavaan(lav_object, R = 500)
```

---
title: "interaction effects between endogenous variables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{interaction effects between endogenous variables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

## The Problem

Interaction effects between two endogenous (i.e., dependent) variables work as you would expect for the product indicator methods (`"dblcent"`, `"rca"`, `"ca"`, `"uca"`). However, for the `lms` and `qml` approaches, it is not as straightforward.

The `lms` and `qml` approaches can (by default) handle interaction effects between endogenous and exogenous (i.e., independent) variables, but they do not natively support interaction effects between two endogenous variables. When an interaction effect exists between two endogenous variables, the equations cannot easily be written in "reduced" form, meaning that normal estimation procedures won't work.

## The Solution

Despite these limitations, there is a workaround for both the `lms` and `qml` approaches. Essentially, the model can be split into two parts: one linear and one non-linear. You can replace the covariance matrix used in the estimation of the non-linear model with the model-implied covariance matrix from a linear model. This allows you to treat an endogenous variable as if it were exogenous—provided that it can be expressed in a linear model.

## Example

Let's consider the theory of planned behavior (TPB), where we wish to estimate the quadratic effect of `INT` on `BEH` (`INT:INT`). We can use the following model:

```{r}
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:INT
'
```

Since `INT` is an endogenous variable, its quadratic term (i.e., an interaction effect with itself) would involve two endogenous variables. As a result, we would normally not be able to estimate this model using the `lms` or `qml` approaches. However, we can split the model into two parts: one linear and one non-linear.

While `INT` is an endogenous variable, it can be expressed in a linear model since it is not affected by any interaction terms:

```{r}
tpb_linear <- 'INT ~ PBC + ATT + SN'
```

We can then remove this part from the original model, giving us:

```{r}
tpb_nonlinear <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  BEH ~ INT + PBC
  BEH ~ INT:INT
'
```

Now, we can estimate the non-linear model since `INT` is treated as an exogenous variable. However, this would not incorporate the structural model for `INT`. To address this, we can instruct `modsem` to replace the covariance matrix (`phi`) of (`INT`, `PBC`, `ATT`, `SN`) with the model-implied covariance matrix from the linear model while estimating both models simultaneously. To achieve this, we use the `cov.syntax` argument in `modsem`:

```{r}
est_lms <- modsem(tpb_nonlinear, data = TPB, cov.syntax = tpb_linear,
                  method = "lms")
summary(est_lms)

est_qml <- modsem(tpb_nonlinear, data = TPB, cov.syntax = tpb_linear,
                  method = "qml")
summary(est_qml)
```

It is also possible to make `modsem` attempt to split up the model syntax automatically using the
`auto.split.syntax` argument:
```{r}
est_lms <- modsem(tpb, data = TPB, method = "lms", auto.split.syntax = TRUE)
summary(est_lms)
```

---
title: "composite constructs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{composite constructs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Composite Constructs
As of version `1.0.20`, the `modsem` function supports the estimation of models
with composite constructs, when using `method="lms"`. The approach is based
on [Tamara Schamberger, Florian Schuberth, Jörg Henseler & Yves Rosseel, 2015](https://arxiv.org/abs/2508.06112).
Depending on your `lavaan` version (`>=0.6-20`), composite constructs can also
be used with the product indicator approaches (e.g., `method="dblcent"`).

Here we can see a simple example, using the LMS approach with the `TPB` dataset.

```{r}
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  # Latent Variables
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3

  # Composites
  ATT <~ att1 + att2 + att3 + att4 + att5
  BEH <~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
'

fit <- modsem(tpb, TPB, method = "lms", nodes = 32)
summary(fit)
```

---
title: "fit measures for LMS and QML"
output: rmarkdown::html_vignette
vignette: |
  %\VignetteIndexEntry{fit measures for LMS and QML}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Introduction

This vignette demonstrates how to evaluate and compare model fit for latent interaction models estimated via

- The Latent Moderated Structural Equations (LMS) approach.
- The Quasi-Maximum Likelihood (QML) approach.

using the `modsem` package (v≥1.0.8). Because standard Chi-square statistics are not available under LMS/QML, we assess fit by:

1. Examining fit indices for the baseline (no-interaction) model.
2. Conducting a likelihood-ratio difference test to compare the baseline and interaction models (Klein & Moosbrugger, 2000; Klein & Múthen, 2007).

If the baseline model shows acceptable fit and adding the latent interaction significantly improves fit, the interaction model can also be deemed well-fitting.

# Example

We define a model with three latent variables (`X`, `Y`, `Z`) and their interaction (`X:Z`):

```{r}
m1 <- "
# Outer (measurement) model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner (structural) model
Y ~ X + Z + X:Z
"

# Estimate the full (H1) model via LMS
est_h1 <- modsem(m1, oneInt, method = "lms")

# Estimate the baseline (H0) model without interaction
est_h0 <- estimate_h0(est_h1, calc.se = FALSE) # std.errors are not needed
```
# Fit measures baseline model

To get fit measures for the baseline model you can use the `fit_modsem_da()` function.
```{r}
fit_modsem_da(est_h0)
```
It can also be used to get fit measures for the full model, but should be pared with
`chisq = FALSE` to avoid the Chi-square test. If it is set to `TRUE` it will
calculate the Chi-square test while ignoring the interaction terms in the model.

```{r}
fit_modsem_da(est_h1, chisq = FALSE)
```

# Difference Test of Fit

Compare H0 vs. H1 using a log-likelihood ratio test:

```{r}
compare_fit(est_h1 = est_h1, est_h0 = est_h0)
```

A **significant** p-value indicates the latent interaction term significantly improves model fit.

# Inspecting Fit Indices
For convenience, you can also use the `modsem_inspect()` function with
`what = "fit"` to get fit indices for both models, and comparative fit in one go.

```{r}
modsem_inspect(est_h1, what = "fit")
```

# References
    Klein, A., & Moosbrugger, H. (2000).
      <doi:10.1007/BF02296338>.
      "Maximum likelihood estimation of latent interaction effects with the LMS method."
    Klein, A. G., & Muthén, B. O. (2007).
      <doi:10.1080/00273170701710205>.
      "Quasi-maximum likelihood estimation of structural equation models with multiple interaction and quadratic effects."

---
title: "customizing interaction terms"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{customizing interaction terms}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup, eval = TRUE}
library(modsem)
```

By default, `modsem()` creates product indicators for you based on the interaction specified in your model. Behind the scenes, `modsem()` generates a total of 9 variables (product indicators) that are used as the indicators for your latent product.

```{r}
m1 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner model
Y ~ X + Z + X:Z
'

est1 <- modsem(m1, oneInt)
cat(est1$syntax)
```

While this is often sufficient, you might want more control over how these indicators are created. In general, `modsem()` offers two mechanisms for controlling the creation of product indicators:
1. By specifying the measurement model for your latent product yourself.
2. By using the `mean()` and `sum()` functions, collectively known as parceling operations.

## Specifying the Measurement Model

By default, `modsem()` creates all possible combinations of product indicators. However, another common approach is to match the indicators by order. For example, let's say you have an interaction between the latent variables `X` and `Z`: `X =~ x1 + x2` and `Z =~ z1 + z2`. By default, you would get `XZ =~ x1z1 + x1z2 + x2z1 + x2z2`. If you prefer to use the *matching approach*, you would expect `XZ =~ x1z1 + x2z2` instead. To achieve this, you can use the `match = TRUE` argument.

```{r}
m2 <- '
# Outer Model
X =~ x1 + x2
Y =~ y1 + y2
Z =~ z1 + z2

# Inner model
Y ~ X + Z + X:Z
'

est2 <- modsem(m2, oneInt, match = TRUE)
summary(est2)
```

## More Complicated Models

If you want even more control, you can use the `get_pi_syntax()` and `get_pi_data()` functions to extract the modified syntax and data from `modsem()`, allowing you to modify them as needed. This can be particularly useful in cases where you want to estimate a model using a feature in `lavaan` that isn't available in `modsem()`.

For example, the syntax for ordered and multigroup models (as of now) isn't as flexible in `modsem()` as it is in `lavaan`. You can modify the auto-generated syntax (along with the altered dataset) from `modsem()` to suit your needs.

```{r}
m3 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner model
Y ~ X + Z + X:Z
'
syntax <- get_pi_syntax(m3)
cat(syntax)
```

```{r}
data <- get_pi_data(m3, oneInt)
head(data)
```

The generated syntax and data can then be used to estimate a model using `lavaan`, giving the
user more control over arguments passed onto `lavaan::sem()`, and potentially the ability to
modify the the generated syntax and data.

```{r}
library(lavaan)
fit <- sem(syntax, data = data, fixed.x = FALSE,
           std.lv = TRUE, auto.fix.first = TRUE,
           auto.var = TRUE, auto.cov.lv.x = TRUE)
summary(fit)
```

---
title: "plotting interaction effects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{plotting interaction effects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
```

```{r setup}
library(modsem)
```

# Plotting Interaction Effects

Interaction effects can be plotted using the included `plot_interaction()` function. This function takes a fitted model object and the names of the two variables that are interacting. The function will plot the interaction effect of the two variables, where:

- The x-variable is plotted on the x-axis.
- The y-variable is plotted on the y-axis.
- The z-variable determines at which points the effect of x on y is plotted.

The function will also plot the 95% confidence interval for the interaction effect.
Note that the `vals_z` argument (as well as the values of `x`) are scaled by the
mean and standard deviation of the variables. Unless the `rescale` argument is set to `FALSE`.

Here is a simple example using the double-centering approach:

```{r}
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)
plot_interaction(x = "X", z = "Z", y = "Y",
                 vals_z = -3:3, model = est1)
```

Here is a different example using the `lms` approach in the theory of planned behavior model:

```{r}
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "lms")
plot_interaction(x = "INT", z = "PBC", y = "BEH",
                 vals_z = c(-0.5, 0.5), model = est2)
```

# Plotting Johnson-Neyman Regions

The `plot_jn()` function can be used to plot Johnson-Neyman regions for a given interaction effect.
This function takes a fitted model object, the names of the two variables that are interacting,
and the name of the interaction effect. The function will plot the Johnson-Neyman regions for the interaction effect.

The `plot_jn()` function will also plot the 95% confidence interval for the interaction effect.

`x` is the name of the x-variable, `z` is the name of the z-variable, and `y` is the name of the y-variable.
`model` is the fitted model object. The argument `min_z` and `max_z` are used to specify the range of values for the
moderating variable.

Here is an example using the `ca` approach in the Holzinger-Swineford (1939) dataset:

```{r}
m1 <-  '
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9

  visual ~ speed + textual + speed:textual
'

est1 <- modsem(m1, data = lavaan::HolzingerSwineford1939, method = "ca")
plot_jn(x = "speed", z = "textual", y = "visual", model = est1, max_z = 6)
```

Here is another example using the `qml` approach in the theory of planned behavior model:
```{r}
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "qml")
plot_jn(x = "INT", z = "PBC", y = "BEH", model = est2,
        min_z = -1.5, max_z = -0.5)
```

# Plotting (3D) Surface Plots
The `plot_surface()` function can be used to plot 3D surface plots for a given interaction effect.
This function takes a fitted model object, the names of the two variables that are interacting,
and the name of the dependent variable. The function will plot the 3D surface plot for the interaction effect.

```{r, eval = EVAL_DEFAULT}
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "qml")
plot_surface(x = "INT", z = "PBC", y = "BEH", model = est2)
```

# Quadratic Effects and Response Surface Models
As of version `1.0.13`, `modsem` also handles quadratic effects when plotting interaction using the `plot_interaction()`
and `plot_surface()` functions. Here we can see an example using some generated data for a
model with two quadratic effects, and a single interaction effect. The model is equivalent
to a typical response surface regression.

## Generating Data
```{r}
set.seed(2934269)

var_X      <- 1
var_Z      <- 1
cov_X_Z    <- 0.2

zeta_Y     <- 0.4

gamma_Y_X  <-  0
gamma_Y_Z  <-  0
gamma_Y_XZ <-  1
gamma_Y_ZZ <-  3 # exclude for now
gamma_Y_XX <- -1 # exclude for now


lambda_1   <- 1
lambda_2   <- .7
lambda_3   <- .8


epsilon    <- 0.2
beta_1     <- 1.2
beta_2     <- 0.8
beta_3     <- 1.5
N          <- 1500

residual <- \(epsilon) rnorm(N, sd = sqrt(epsilon))
create_ind <- \(lv, beta, lambda, epsilon) beta + lambda * lv + residual(epsilon)


Phi <- matrix(c(var_X, cov_X_Z,
                cov_X_Z, var_Z), nrow = 2)
XI <- mvtnorm::rmvnorm(N, sigma = Phi)

X <- XI[, 1]
Z <- XI[, 2]

Y <-
  gamma_Y_X * X +
  gamma_Y_Z * Z +
  gamma_Y_XZ * X * Z +
  gamma_Y_XX * X * X +
  gamma_Y_ZZ * Z * Z +
  residual(zeta_Y)

x1 <- create_ind(X, beta_1, lambda_1, epsilon)
x2 <- create_ind(X, beta_2, lambda_2, epsilon)
x3 <- create_ind(X, beta_3, lambda_2, epsilon)

z1 <- create_ind(Z, beta_1, lambda_1, epsilon)
z2 <- create_ind(Z, beta_2, lambda_2, epsilon)
z3 <- create_ind(Z, beta_3, lambda_2, epsilon)

y1 <- create_ind(Y, beta_1, lambda_1, epsilon)
y2 <- create_ind(Y, beta_2, lambda_2, epsilon)
y3 <- create_ind(Y, beta_2, lambda_2, epsilon)


data.sim <- data.frame(
   x1, x2, x3,
   z1, z2, z3,
   y1, y2, y3
)
```

## Fitting Model
Here we fit the model using the QML approach.

```{r, eval = EVAL_DEFAULT}
model <- '
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  Y ~ X + Z + X:X + Z:Z + X:Z
'

est3 <- modsem(model, data = data.sim, method = "qml")
```

## 2D Plot
Here we can see a two dimensional plot, where the slope of `X` no longer is constant
at a given value of `Z`, since `X` has an interaction effect with itself.
```{r, eval = EVAL_DEFAULT}
plot_interaction(x = "X", z = "Z", y = "Y", model = est3, vals_z = c(-1, 0, 1))
```

## 3D (Response Surface) Plot
Here we can see a three-dimensional surface plot.

```{r, eval = EVAL_DEFAULT}
plot_surface(x = "X", z = "Z", y = "Y", model = est3)
```

---
title: "LMS and QML approaches"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{LMS and QML approaches}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# The Latent Moderated Structural Equations (LMS) and the Quasi Maximum Likelihood (QML) Approach

Both the `LMS` and `QML` approaches work on most models, but interaction effects with endogenous variables can be tricky to estimate (see the [vignette](https://modsem.org/articles/interaction_two_etas.html)). Both approaches, particularly the `LMS` approach, are computationally intensive and are partially implemented in C++ (using `Rcpp` and `RcppArmadillo`). Additionally, starting parameters are optimized, to generate good starting parameters for faster convergence. If you want to monitor the progress of the estimation process, you can use `verbose = TRUE`.

## A Simple Example

Here is an example of the `LMS` approach for a simple model. By default, the `summary()` function calculates fit measures compared to a null model (i.e., the same model without an interaction term).

```{r}
library(modsem)
m1 <- '
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
'

lms1 <- modsem(m1, oneInt, method = "lms")
# Standardized estimates
summary(lms1, standardized = TRUE)
```

Here is the same example using the `QML` approach:

```{r}
qml1 <- modsem(m1, oneInt, method = "qml", robust.se = TRUE) # get robust standard errors
summary(qml1, fit = TRUE, adjusted = TRUE) # get additional fit indices
```

## A More Complicated Example

Below is an example of a more complex model based on the theory of planned behavior (TPB), which includes two endogenous variables and an interaction between an endogenous and exogenous variable. When estimating more complex models with the `LMS` approach, it is recommended to increase the number of nodes used for numerical integration. By default, the number of nodes is set to 24, but this can be increased using the `nodes` argument. The `nodes` argument has no effect on the `QML` approach.

When there is an interaction effect between an endogenous and exogenous variable, it is recommended to use at least 32 nodes for the `LMS` approach. You can also obtain robust standard errors by setting `robust.se = TRUE` in the `modsem()` function.

**NOTE**: If you want the `LMS` approach to produce results as similar as possible to `Mplus`,
you should consider increasing the number of nodes using the `nodes` argument, and or
using a quasi-adaptive quadrature by setting `adaptive.quad=TRUE`.

```{r}
# ATT = Attitude
# PBC = Perceived Behavioral Control
# INT = Intention
# SN = Subjective Norms
# BEH = Behavior
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

lms2 <- modsem(tpb, TPB, method = "lms", nodes = 32)
summary(lms2)

qml2 <- modsem(tpb, TPB, method = "qml")
summary(qml2, standardized = TRUE) # Standardized estimates
```

---
title: "estimation alternatives using the LMS approach"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{estimation alternatives using the LMS approach}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Accelerated EM and Adaptive Quadrature
By default (as of `v1.0.9`) the LMS approach uses an accelerated EM procedure (`"EMA"`) that uses Quasi-Newton and Fisher
Scoring optimization steps when needed. If desireable, this can be switched to the standard Expectation-Maximization (EM)
algorithm, by setting `algorithm = "EM"`.

By default the LMS approach also uses a fixed Gauss-Hermite quadrature, to estimate a numerical
approximation of the log-likelihood. Instead of a fixed quadrature, it is possible to use a
quasi-adaptive quadrature instead. Due to performance reasons, the adaptive quadrature does not fit an
individual quadrature to each participant, but instead one for the entire sample (at each EM iteration),
based on the whole sample densities of the likelihood function. It essentially works by
removing irrelevant nodes which don't contribute to the integral, and increasing the number of nodes
which actually contribute to the integral. This usually means that more nodes are placed towards the
center of the distribution, compared to a standard fixed Gauss-Hermite quadrature.
Using the EMA and adaptive quadrature might yield estimates that are closer to results from `Mplus`.

If the model struggles to converge, you can try changing the EM procedure by setting `algorithm = "EMA"`,
or `algorithm = "EM"`, and `adaptive.quad = TRUE` in the `modsem()` function. Additionally it is possible to tweak these parameters:

- `max.iter`: Maximum number of iterations for the EM algorithm (default is 500).
- `max.step`: Maximum number of steps used in the Maximization step of the EM algorithm (default is 1).
- `convergence.rel`: Relative convergence criterion for the EM algorithm.
- `convergence.abs`: Absolute convergence criterion for the EM algorithm.
- `nodes`: Number of nodes for numerical integration (default is 24). Increasing this number can improve the accuracy of the estimates, especially for complex models.
- `quad.range`: Integration range for quadrature. Smaller ranges means that the integral is more focused. Applies to only when using a quasi-adaptive quadrature.
- `adaptive.frequency`: How often should the quasi-adaptive quadrature be calculated? Defaults to every third EM iteration.
- `adaptive.quad.tol`: Relative error tolerance when calculating the quasi-adaptive quadrature.

Here we can see an example using the `TPB_UK` dataaset, which is more troublesome to
estimate than the simulated `TPB` dataset.

```{r}
tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att3 + att2 + att1 + att4
  SN =~ sn4 + sn2 + sn3 + sn1
  PBC =~ pbc2 + pbc1 + pbc3 + pbc4
  INT =~ int2 + int1 + int3 + int4
  BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
"

fit <- modsem(tpb_uk,
              data = TPB_UK,
              method = "lms",
              nodes = 32, # Number of nodes for numerical integration
              adaptive.quad = TRUE, # Use quasi-adaptive quadrature
              adaptive.frequency = 3, # Update the quasi-adaptive quadrature every third EM-iteration
              adaptive.quad.tol = 1e-12, # Relative error tolerance for quasi-adaptive quad
              algorithm ="EMA", # Use accelerated EM algorithm (Default)
              convergence.abs = 1e-4, # Relative convergence criterion
              convergence.rel = 1e-10, # Relative convergence criterion
              max.iter = 500, # Maximum number of iterations
              max.step = 1) # Maximum number of steps in the maximization step
summary(fit)
```

---
title: "missing data in the LMS and QML approaches"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{missing data in the LMS and QML approaches}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# Options for Handling Missing Values
By default, missing values for both the LMS and QML are removed list-wise
(`missing="listwise"`), such that only complete observations are kept.
It is also possible to set `missing="impute"` to impute missing values,
or to use the `modsem_mimpute()` function to perform multiple imputation.
For the LMS approach it is also possible to use
*Full Information Maximum Likelihood* (FIML)

## Full Information Maximum Likelihood (FIML)

If you're using the LMS approach, it is possible to estimate your model using
*Full Information Maximum Likelihood* (FIML), by setting `missing="fiml"`. Here
is an example, where we generate some missing data using the `oneInt` dataset.

```{r}
set.seed(2834027) # set seed for reproducibility
n.missing <- 200L # generate 200 missing values
oneInt2 <- as.matrix(oneInt)
oneInt2[sample(length(oneInt2), n.missing)] <- NA

m1 <- '
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
'

lms_fiml <- modsem(m1, oneInt2, method = "lms", missing = "fiml")
summary(lms_fiml)
```

## (Multiple) Imputation

If you're using the QML approach it is not possible (yet) to use FIML, and FIML
can also be very computationally intensive when using the LMS approach. Another
option is to use (multiple) imputation.

If `missing="impute"` is set, single imputation is performed.

```{r}
qml_impute <- modsem(m1, oneInt2, method = "qml", missing = "impute")
summary(qml_impute)
```

A better option than single imputation, is multiple imputation, which can be
performed both for the LMS and QML approaches, using the `modsem_mimpute()` function.

```{r}
lms_mimpute <- modsem_mimpute(m1, oneInt2, method = "lms")
summary(lms_mimpute)
```

---
title: "modsem"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{modsem}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
EVAL_DEFAULT <- FALSE
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = EVAL_DEFAULT
)
```

```{r setup}
library(modsem)
```

# The Basic Syntax

`modsem` introduces a new feature to the `lavaan` syntax—the semicolon operator (`:`). The semicolon operator works the same way as in the `lm()` function. To specify an interaction effect between two variables, you join them by `Var1:Var2`.

Models can be estimated using one of the product indicator approaches (`"ca"`, `"rca"`, `"dblcent"`, `"pind"`) or by using the latent moderated structural equations approach (`"lms"`) or the quasi maximum likelihood approach (`"qml"`). The product indicator approaches are estimated via `lavaan`, while the `lms` and `qml` approaches are estimated via `modsem` itself.

## A Simple Example

Here is a simple example of how to specify an interaction effect between two latent variables in `lavaan`.

```{r}
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner Model
  Y ~ X + Z + X:Z
'

est1 <- modsem(m1, oneInt)
summary(est1)
```

By default, the model is estimated using the `"dblcent"` method. If you want to use another method, you can change it using the `method` argument.

```{r}
est1 <- modsem(m1, oneInt, method = "lms")
summary(est1)
```

## Interactions Between Two Observed Variables

`modsem` allows you to estimate interactions between not only latent variables but also observed variables. Below, we first run a regression with only observed variables, where there is an interaction between `x1` and `z2`, and then run an equivalent model using `modsem()`.

### Using a Regression

```{r}
reg1 <- lm(y1 ~ x1*z1, oneInt)
summary(reg1)
```

### Using `modsem`

`modsem` can also be used to to estimate interaction effects between observed variables.
Interaction effects with observed variables are not supported by the `LMS` and `QML` approaches.
In most cases, you can define a latent variable with a single indicator to estimate the interaction
effect between two observed variables in the `LMS` and `QML` approaches.

```{r}
est2 <- modsem('y1 ~ x1 + z1 + x1:z1', data = oneInt, method = "dblcent")
summary(est2)
```

## Interactions Between Latent and Observed Variables

`modsem` also allows you to estimate interaction effects between latent and observed variables.
To do so, simply join a latent and an observed variable with a colon (e.g., `'latent:observed'`).
In most of the product indicator approaches the residuals of product indicators with common variables (e.g., `x1z1` and `x1z2`)
are allowed to covary freely. This is problematic when there is an interaction term between an observed variable
(or a latent variable with a single indicator) and a latent variables (with multiple indicators), since
all of the product indicators share a common element.

In the example below, all the generated product indicators
(`x1z1`, `x2z1` and `x3z1`) share `z1`. If all the indicator residuals of
a latent variabel are allowed to covary freely, the model
will (in general) be unidenfified. The simplest way to fix this issue, is to constrain the residual covariances
to be zero, by using the `res.cov.method` argument.

```{r}
m3 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3

  # Inner Model
  Y ~ X + z1 + X:z1
'

est3 <- modsem(m3, oneInt, method = "dblcent", res.cov.method = "none")
summary(est3)
```

## Quadratic Effects

Quadratic effects are essentially a special case of interaction effects. Thus, `modsem` can also be used to estimate quadratic effects.

```{r}
m4 <- '
# Outer Model
X =~ x1 + x2 + x3
Y =~ y1 + y2 + y3
Z =~ z1 + z2 + z3

# Inner Model
Y ~ X + Z + Z:X + X:X
'

est4 <- modsem(m4, oneInt, method = "qml")
summary(est4)
```

## More Complicated Examples

Here is a more complex example using the theory of planned behavior (TPB) model.

```{r}
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
'

# The double-centering approach
est_tpb <- modsem(tpb, TPB)

# Using the LMS approach
est_tpb_lms <- modsem(tpb, TPB, method = "lms")
summary(est_tpb_lms)
```

Here is an example that includes two quadratic effects and one interaction effect, using the `jordan` dataset. The dataset is a subset of the PISA 2006 dataset.

```{r}
m2 <- '
ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
CAREER =~ career1 + career2 + career3 + career4
SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
'

est_jordan <- modsem(m2, data = jordan)
est_jordan_qml <- modsem(m2, data = jordan, method = "qml")
summary(est_jordan_qml)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multiple_imputation_da.R
\name{modsem_mimpute}
\alias{modsem_mimpute}
\title{Estimate a \code{modsem} model using multiple imputation}
\usage{
modsem_mimpute(
  model.syntax,
  data,
  method = "lms",
  m = 25,
  verbose = interactive(),
  se = c("simple", "full"),
  ...
)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{A dataframe with observed variables used in the model.}

\item{method}{Method to use:
\describe{
  \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
  \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
}}

\item{m}{Number of imputations to perform. More imputations will yield better estimates
but can also be (a lot) slower.}

\item{verbose}{Should progress be printed to the console?}

\item{se}{How should corrected standard errors be computed? Alternatives are:
\describe{
  \item{\code{"simple"}}{Uncorrected standard errors are only calculated once, in the first imputation.
                         The standard errors are thereafter corrected using the distribution of the
                         estimated coefficients from the different imputations.}
  \item{\code{"full"}}{Uncorrected standard errors are calculated and aggregated for each imputation.
                       This can give more accurate results, but can be (a lot) slower.
                       The standard errors are thereafter corrected using the distribution of the
                       estimated coefficients from the different imputations.}
}}

\item{...}{Arguments passed to \code{\link{modsem}}.}
}
\description{
Estimate a \code{modsem} model using multiple imputation
}
\details{
\code{modsem_impute} is currently only available for the DA approaches
 (LMS and QML). It performs multiple imputation using \code{Amelia::amelia}
 and returns aggregated coefficients from the multiple imputations, along with
 corrected standard errors.
}
\examples{

m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'

oneInt2 <- oneInt

set.seed(123)
k <- 200
I <- sample(nrow(oneInt2), k, replace = TRUE)
J <- sample(ncol(oneInt2), k, replace = TRUE)
for (k_i in seq_along(I)) oneInt2[I[k_i], J[k_i]] <- NA

\dontrun{
est <- modsem_mimpute(m1, oneInt2, m = 25)
summary(est)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/generics_lavaan.R,
%   R/generics_modsem_da.R, R/generics_modsem_pi.R
\name{modsem_inspect}
\alias{modsem_inspect}
\alias{modsem_inspect.lavaan}
\alias{modsem_inspect.modsem_da}
\alias{modsem_inspect.modsem_pi}
\title{Inspect model information}
\usage{
modsem_inspect(object, what = NULL, ...)

\method{modsem_inspect}{lavaan}(object, what = "free", ...)

\method{modsem_inspect}{modsem_da}(object, what = NULL, ...)

\method{modsem_inspect}{modsem_pi}(object, what = "free", ...)
}
\arguments{
\item{object}{A fitted object of class \code{"modsem_da"}.}

\item{what}{Character scalar selecting what to return (see \emph{Details}).
If \code{NULL} the value \code{"default"} is used.}

\item{...}{Passed straight to \code{modsem_inspect_da()}.}
}
\value{
A named list with the extracted information. If a single piece of information is returned,
 it is returned as is; not as a named element in a list.
}
\description{
function used to inspect fitted object. Similar to \code{lavaan::lavInspect}
argument \code{what} decides what to inspect

\code{modsem_inspect.modsem_da} lets you pull lavaan-style matrices,
optimiser diagnostics, expected moments, or fit measures from a
\code{\link{modsem_da}} object.
}
\details{
For \code{\link{modsem_pi}} objects, it is just a wrapper for \code{lavaan::lavInspect}.
 For \code{\link{modsem_da}} objects an internal function is called, which takes different
 keywords for the \code{what} argument.

Below is a list of possible values for the \code{what} argument,
organised in several sections.  Keywords are \emph{case-sensitive}.

\strong{Presets}

\describe{
  \item{\code{"default"}}{Everything in \emph{Sample information}, \emph{Optimiser diagnostics}
  \emph{Parameter tables}, \emph{Model matrices}, and \emph{Expected-moment matrices} except
  the raw \code{data} slot}
  \item{\code{"coef"}}{Coefficients and variance-covariance matrix of both free and constrained parameters (same as \code{"coef.all"}).}
  \item{\code{"coef.all"}}{Coefficients and variance-covariance matrix of both free and constrained parameters (same as \code{"coef"}).}
  \item{\code{"coef.free"}}{Coefficients and variance-covariance matrix of the free parameters.}
  \item{\code{"all"}}{All items listed below, including \code{data}.}
  \item{\code{"matrices"}}{The lavaan-style model matrices.}
  \item{\code{"optim"}}{Only the items under \emph{Optimiser diagnostics}}.
  \item{\code{"fit"}}{A list with \code{fit.h0}, \code{fit.h1}, comparative.fit}
}

\strong{Sample information:}

\describe{
\item{\code{"N"}}{Number of analysed rows (integer).}
\item{\code{"ngroups"}}{Number of groups in model (integer).}
\item{\code{"group"}}{Group variable in model (character).}
\item{\code{"group.label"}}{Group labels (character).}
\item{\code{"ovs"}}{Observed variables used in model (character).}
}

\strong{Parameter estimates and standard errors:}

\describe{
  \item{\code{"coefficients.free"}}{Free parameter values.}
  \item{\code{"coefficients.all"}}{Both free and constrained parameter values.}
  \item{\code{"vcov.free"}}{Variance–covariance of free coefficients only.}
  \item{\code{"vcov.all"}}{Variance–covariance of both free and constrained coefficients.}
}

\strong{Optimiser diagnostics:}

\describe{
  \item{\code{"coefficients.free"}}{Free parameter values.}
  \item{\code{"vcov.free"}}{Variance–covariance of free coefficients only.}
  \item{\code{"information"}}{Fisher information matrix.}
  \item{\code{"loglik"}}{Log-likelihood.}
  \item{\code{"iterations"}}{Optimiser iteration count.}
  \item{\code{"convergence"}}{\code{TRUE}/\code{FALSE} indicating whether the model converged.}
  \item{\code{"iteration.history"}}{LMS iteration history with mode, log-likelihood, changes, elapsed time, and forecast columns, if available.}
}

\strong{Parameter tables:}

\describe{
  \item{\code{"partable"}}{Parameter table with estimated parameters.}
  \item{\code{"partable.input"}}{Parsed model syntax.}
}

\strong{Model matrices:}

\describe{
  \item{\code{"lambda"}}{Factor loadings.}
  \item{\code{"theta"}}{Residual covariance matrix for indicators.}
  \item{\code{"wmat"}}{Composite loading matrix for observed indicators, if present.}
  \item{\code{"tmat"}}{Composite residual matrix for indicators, if present.}
  \item{\code{"psi"}}{Residual covariance matrix for latent variables.}
  \item{\code{"beta"}}{Structural regression matrix among latent variables.}
  \item{\code{"nu"}}{Intercepts for observed variables.}
  \item{\code{"alpha"}}{Intercepts for latent variables.}
}
\strong{Model-implied matrices:}

\describe{
  \item{\code{"cov.ov"}}{Model-implied covariance of observed variables.}
  \item{\code{"cov.lv"}}{Model-implied covariance of latent variables.}
  \item{\code{"cov.all"}}{Joint covariance of observed + latent variables.}
  \item{\code{"cor.ov"}}{Correlation counterpart of \code{"cov.ov"}.}
  \item{\code{"cor.lv"}}{Correlation counterpart of \code{"cov.lv"}.}
  \item{\code{"cor.all"}}{Correlation counterpart of \code{"cov.all"}.}
  \item{\code{"mean.ov"}}{Expected means of observed variables.}
  \item{\code{"mean.lv"}}{Expected means of latent variables.}
  \item{\code{"mean.all"}}{Joint mean vector.}
}

\strong{R-squared and standardized residual variances:}

\describe{
  \item{\code{"r2.all"}}{R-squared values for both observed (i.e., indicators) and latent endogenous variables.}
  \item{\code{"r2.lv"}}{R-squared values for latent endogenous variables.}
  \item{\code{"r2.ov"}}{R-squared values for observed (i.e., indicators) variables.}
  \item{\code{"res.all"}}{Standardized residuals (i.e., \code{1 - R^2}) for both observed (i.e., indicators) and latent endogenous variables.}
  \item{\code{"res.lv"}}{Standardized residuals (i.e., \code{1 - R^2}) for latent endogenous variables.}
  \item{\code{"res.ov"}}{Standardized residuals (i.e., \code{1 - R^2}) for observed variables (i.e., indicators).}
}

\strong{Interaction-specific caveats:}

\itemize{
  \item If the model contains an \emph{uncentred} latent interaction term it is centred
    internally before any \code{cov.*}, \code{cor.*}, or \code{mean.*} matrices are
    calculated.
  \item These matrices should not be used to compute fit-statistics (e.g.,
    chi-square and RMSEA) if there is an interaction term in the model.
}
}
\section{Methods (by class)}{
\itemize{
\item \code{modsem_inspect(lavaan)}: Inspect a \code{lavaan} object

\item \code{modsem_inspect(modsem_da)}: Inspect a \code{\link{modsem_da}} object

\item \code{modsem_inspect(modsem_pi)}: Inspect a \code{\link{modsem_pi}} object

}}
\examples{
\dontrun{
m1 <- "
 # Outer Model
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Z =~ z1 + z2 + z3

 # Inner model
 Y ~ X + Z + X:Z
"

est <- modsem(m1, oneInt, "lms")

modsem_inspect(est) # everything except "data"
modsem_inspect(est, what = "optim")
modsem_inspect(est, what = "matrices")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorize_output.R
\name{colorize_output}
\alias{colorize_output}
\title{Capture, colorise, and emit console text}
\usage{
colorize_output(
  expr,
  positive = MODSEM_COLORS$positive,
  negative = MODSEM_COLORS$negative,
  true = MODSEM_COLORS$true,
  false = MODSEM_COLORS$false,
  nan = MODSEM_COLORS$nan,
  na = MODSEM_COLORS$na,
  inf = MODSEM_COLORS$inf,
  string = MODSEM_COLORS$string,
  split = FALSE,
  append = "\\n"
)
}
\arguments{
\item{expr}{Expression or object with output which should be colorized.}

\item{positive}{color of positive numbers.}

\item{negative}{color of negative numbers.}

\item{true}{color of \code{TRUE}.}

\item{false}{color of \code{FALSE}.}

\item{nan}{color of \code{NaN}.}

\item{na}{color of \code{NA}.}

\item{inf}{color of \code{-Inf} and \code{Inf}.}

\item{string}{color of quoted strings.}

\item{split}{Should output be splitted? If \code{TRUE} the output is printed
normally (in real time) with no colorization, and the colored output is printed
after the code has finished executing.}

\item{append}{String appended after the colored output (default `\\n`).}
}
\value{
Invisibly returns the *plain* captured text.
}
\description{
Capture, colorise, and emit console text
}
\examples{
set_modsem_colors(positive = "red3",
                  negative = "red3",
                  true = "darkgreen",
                  false = "red3",
                  na = "purple",
                  string = "darkgreen")

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est <- modsem(m1, data = oneInt)
colorize_output(summary(est))
colorize_output(est) # same as colorize_output(print(est))
colorize_output(modsem_inspect(est, what = "coef"))

\dontrun{
colorize_output(split = TRUE, {
  # Get live (uncolored) output
  # And print colored output at the end of execution

  est_lms <- modsem(m1, data = oneInt, method = "lms")
  summary(est_lms)
})

colorize_output(modsem_inspect(est_lms))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trace_paths_wright.R
\name{trace_path}
\alias{trace_path}
\title{Estimate formulas for (co-)variance paths using Wright's path tracing rules}
\usage{
trace_path(
  pt,
  x,
  y,
  parenthesis = TRUE,
  missing.cov = FALSE,
  measurement.model = TRUE,
  maxlen = 100,
  paramCol = "mod",
  ...
)
}
\arguments{
\item{pt}{A data frame with columns \code{lhs}, \code{op}, \code{rhs}, and \code{mod}, from \link{modsemify}}

\item{x}{Source variable}

\item{y}{Destination variable}

\item{parenthesis}{If \code{TRUE}, the output will be enclosed in parenthesis}

\item{missing.cov}{If \code{TRUE}, covariances missing from the model syntax will be added}

\item{measurement.model}{If \code{TRUE}, the function will use the measurement model}

\item{maxlen}{Maximum length of a path before aborting}

\item{paramCol}{The column name in \code{pt} that contains the parameter labels}

\item{...}{Additional arguments passed to \link{trace_path}}
}
\value{
A string with the estimated path (simplified if possible)
}
\description{
This function estimates the path from \code{x} to \code{y} using the path tracing rules.
Note that it only works with structural parameters, so \code{"=~"} are ignored, unless
\code{measurement.model = TRUE}.
If you want to use the measurement model,
\code{"~"} should be in the \code{mod} column of \code{pt}.
}
\examples{
library(modsem)
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
pt <- modsemify(m1)
trace_path(pt, x = "Y", y = "Y", missing.cov = TRUE) # variance of Y
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/generics_lavaan.R,
%   R/generics_modsem_da.R
\name{centered_estimates}
\alias{centered_estimates}
\alias{centered_estimates.lavaan}
\alias{centered_estimates.modsem_da}
\title{Get Centered Interaction Term Estimates}
\usage{
centered_estimates(object, ...)

\method{centered_estimates}{lavaan}(
  object,
  monte.carlo = FALSE,
  mc.reps = 10000,
  tolerance.zero = 1e-10,
  ...
)

\method{centered_estimates}{modsem_da}(
  object,
  monte.carlo = FALSE,
  mc.reps = 10000,
  tolerance.zero = 1e-10,
  ...
)
}
\arguments{
\item{object}{An object of class \code{\link{modsem_da}}}

\item{...}{Additional arguments passed to underlying methods. See specific method
documentation for supported arguments, including:}

\item{monte.carlo}{Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
standard errors; if \code{FALSE}, use the delta method (default).}

\item{mc.reps}{Number of Monte Carlo repetitions. Default is 10000.}

\item{tolerance.zero}{Threshold below which standard errors are set to \code{NA}.}
}
\value{
A \code{data.frame} with centered estimates in the \code{est} column.
}
\description{
Computes centered estimates of model parameters. This is relevant when there is an
 interaction term in the model, as the simple main effects depend upon the mean structure
 of the structural model. Currenlty only available for
 \code{\link{modsem_da}} and \code{lavaan} object.
 It is not relevant for the PI approaches (excluding the "pind" method, which is not recommended),
 since the indicators are centered before computing the product terms.
 The centering can be applied to observed variable interactions in \code{lavaan} models
 and latent interactions estimated using the \code{sam} function.
}
\section{Methods (by class)}{
\itemize{
\item \code{centered_estimates(lavaan)}: Method for \code{lavaan} objects

\item \code{centered_estimates(modsem_da)}: Method for \code{modsem_da} objects

}}
\examples{
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  # Inner Model
  Y ~ X + Z + X:Z
'
\dontrun{
est_lms <- modsem(m1, oneInt, method = "lms")
centered_estimates(est_lms)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem-package.R, R/modsem.R
\name{modsem}
\alias{modsem}
\title{Estimate interaction effects in structural equation models (SEMs)}
\usage{
modsem(model.syntax = NULL, data = NULL, method = "dblcent", ...)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{dataframe}

\item{method}{method to use:
\describe{
  \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
  \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
  \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
  \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
  \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
  \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
  \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
  \item{\code{"custom"}}{use parameters specified in the function call (passed to \code{lavaan}).}
  \item{\code{"mplus"}}{estimate model through \code{Mplus}.}
}}

\item{...}{arguments passed to other functions depending on the method (see \code{\link{modsem_pi}}, \code{\link{modsem_da}}, and \code{\link{modsem_mplus}})}
}
\value{
\code{modsem} object with class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, or \code{\link{modsem_mplus}}
}
\description{
\code{modsem()} is a function for estimating interaction effects between latent variables
in structural equation models (SEMs).
Methods for estimating interaction effects in SEMs can basically be split into two frameworks:
\enumerate{
  \item Product Indicator (PI) based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"}, \code{"ca"}, \code{"pind"})
  \item Distributionally (DA) based approaches (\code{"lms"}, \code{"qml"}).
}

For the product indicator-based approaches, \code{modsem()} is essentially a fancy wrapper for \code{lavaan::sem()} which generates the
necessary syntax and variables for the estimation of models with latent product indicators.

The distributionally based approaches are implemented separately and are
not estimated using \code{lavaan::sem()}, but rather using custom functions (largely
written in \code{C++} for performance reasons). For greater control, it is advised that
you use one of the sub-functions (\link{modsem_pi}, \link{modsem_da}, \link{modsem_mplus}) directly,
as passing additional arguments to them via \code{modsem()} can lead to unexpected behavior.
}
\examples{
library(modsem)
# For more examples, check README and/or GitHub.
# One interaction
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'

# Double centering approach
est1 <- modsem(m1, oneInt)
summary(est1)

\dontrun{
# The Constrained Approach
est1_ca <- modsem(m1, oneInt, method = "ca")
summary(est1_ca)

# LMS approach
est1_lms <- modsem(m1, oneInt, method = "lms")
summary(est1_lms)

# QML approach
est1_qml <- modsem(m1, oneInt, method = "qml")
summary(est1_qml)
}

# Theory Of Planned Behavior
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

# Double centering approach
est_tpb <- modsem(tpb, data = TPB)
summary(est_tpb)

\dontrun{
# The Constrained Approach
est_tpb_ca <- modsem(tpb, data = TPB, method = "ca")
summary(est_tpb_ca)

# LMS approach
est_tpb_lms <- modsem(tpb, data = TPB, method = "lms", nodes = 32)
summary(est_tpb_lms)

# QML approach
est_tpb_qml <- modsem(tpb, data = TPB, method = "qml")
summary(est_tpb_qml)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{jordan}
\alias{jordan}
\title{Jordan subset of PISA 2006 data}
\format{
A data frame of fifteen variables and 6,038 observations:

\describe{
\item{enjoy1}{indicator for enjoyment of science, item ST16Q01: I generally have fun when I am learning <broad science> topics.}
\item{enjoy2}{indicator for enjoyment of science, item ST16Q02: I like reading about <broad science>.}
\item{enjoy3}{indicator for enjoyment of science, item ST16Q03: I am happy doing <broad science> problems.}
\item{enjoy4}{indicator for enjoyment of science, item ST16Q04: I enjoy acquiring new knowledge in <broad science>.}
\item{enjoy5}{indicator for enjoyment of science, item ST16Q05: I am interested in learning about <broad science>.}
\item{academic1}{indicator for academic self-concept in science, item ST37Q01: I can easily understand new ideas in <school science>.}
\item{academic2}{indicator for academic self-concept in science, item ST37Q02: Learning advanced <school science> topics would be easy for me.}
\item{academic3}{indicator for academic self-concept in science, item ST37Q03: I can usually give good answers to <test questions> on <school science> topics.}
\item{academic4}{indicator for academic self-concept in science, item ST37Q04: I learn <school science> topics quickly.}
\item{academic5}{indicator for academic self-concept in science, item ST37Q05: <School science> topics are easy for me.}
\item{academic6}{indicator for academic self-concept in science, item ST37Q06: When I am being taught <school science>, I can understand the concepts very well.}
\item{career1}{indicator for career aspirations in science, item ST29Q01: I would like to work in a career involving <broad science>.}
\item{career2}{indicator for career aspirations in science, item ST29Q02: I would like to study <broad science> after <secondary school>.}
\item{career3}{indicator for career aspirations in science, item ST29Q03: I would like to spend my life doing advanced <broad science>.}
\item{career4}{indicator for career aspirations in science, item ST29Q04: I would like to work on <broad science> projects as an adult.}
}
}
\source{
This version of the dataset, as well as the description was gathered from the
documentation of the 'nlsem' package (https://cran.r-project.org/package=nlsem),
where the only difference is that the names of the variables were changed

Originally the dataset was gathered by the Organisation for Economic Co-Operation and Development (2009).
Pisa 2006: Science competencies for tomorrow's world (Tech. Rep.).
Paris, France. Obtained from: https://www.oecd.org/pisa/pisaproducts/database-pisa2006.htm
}
\description{
The data stem from the large-scale assessment study PISA 2006
(Organisation for Economic Co-Operation and Development, 2009) where
competencies of 15-year-old students in reading, mathematics, and science
are assessed using nationally representative samples in 3-year cycles.
In this example, data from the student background questionnaire from the
Jordan sample of PISA 2006 were used. Only data of students with complete
responses to all 15 items (N = 6,038) were considered.
}
\examples{
\dontrun{
m1 <- "
  ENJ =~ enjoy1 + enjoy2 + enjoy3 + enjoy4 + enjoy5
  CAREER =~ career1 + career2 + career3 + career4
  SC =~ academic1 + academic2 + academic3 + academic4 + academic5 + academic6
  CAREER ~ ENJ + SC + ENJ:ENJ + SC:SC + ENJ:SC
"

est <- modsem(m1, data = jordan, method = "qml")
summary(est)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/generics_lavaan.R,
%   R/generics_modsem_da.R, R/generics_modsem_mplus.R, R/generics_modsem_pi.R
\name{parameter_estimates}
\alias{parameter_estimates}
\alias{parameter_estimates.lavaan}
\alias{parameter_estimates.modsem_da}
\alias{parameter_estimates.modsem_mplus}
\alias{parameter_estimates.modsem_pi}
\title{Extract parameterEstimates from an estimated model}
\usage{
parameter_estimates(object, ...)

\method{parameter_estimates}{lavaan}(
  object,
  colon.pi = NULL,
  high.order.as.measr = NULL,
  rm.tmp.ov = NULL,
  label.renamed.prod = NULL,
  is.public = NULL,
  ...
)

\method{parameter_estimates}{modsem_da}(
  object,
  high.order.as.measr = TRUE,
  is.public = TRUE,
  rm.tmp.ov = is.public,
  label.renamed.prod = NULL,
  ...
)

\method{parameter_estimates}{modsem_mplus}(
  object,
  colon.pi = NULL,
  high.order.as.measr = NULL,
  rm.tmp.ov = NULL,
  label.renamed.prod = NULL,
  is.public = NULL,
  ...
)

\method{parameter_estimates}{modsem_pi}(
  object,
  colon.pi = FALSE,
  label.renamed.prod = FALSE,
  high.order.as.measr = NULL,
  rm.tmp.ov = NULL,
  is.public = NULL,
  ...
)
}
\arguments{
\item{object}{An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, or \code{\link{modsem_mplus}}}

\item{...}{Additional arguments passed to other functions}

\item{colon.pi}{Should colons (\code{:}) be added to the interaction terms (E.g., `XZ` -> `X:Z`)?}

\item{high.order.as.measr}{Should higher order measurement model be
denoted with the \code{=~} operator? If \code{FALSE} the \code{~}
operator is used.}

\item{rm.tmp.ov}{Should temporary (hidden) variables be removed?}

\item{label.renamed.prod}{Should renamed product terms keep their old (implicit) labels?}

\item{is.public}{Should public version of parameter table be returned?
If \code{FALSE}, the internal version of the parameter table is returned.}
}
\description{
Extract parameterEstimates from an estimated model
}
\section{Methods (by class)}{
\itemize{
\item \code{parameter_estimates(lavaan)}: Get parameter estimates of a \code{lavaan} object

\item \code{parameter_estimates(modsem_da)}: Get parameter estimates of a \code{\link{modsem_da}} object

\item \code{parameter_estimates(modsem_mplus)}: Get parameter estimates of a \code{\link{modsem_mplus}} object

\item \code{parameter_estimates(modsem_pi)}: Get parameter estimates of a \code{\link{modsem_pi}} object

}}
\examples{
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  # Inner Model
  Y ~ X + Z + X:Z
'
# Double centering approach
est_dca <- modsem(m1, oneInt)

pars <- parameter_estimates(est_dca) # no correction

# Pretty summary
summarize_partable(pars)

# Only print the data.frame
pars
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_pi.R
\name{extract_lavaan}
\alias{extract_lavaan}
\title{extract lavaan object from modsem object estimated using product indicators}
\usage{
extract_lavaan(object)
}
\arguments{
\item{object}{modsem object}
}
\value{
lavaan object
}
\description{
extract lavaan object from modsem object estimated using product indicators
}
\examples{
library(modsem)
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
est <- modsem_pi(m1, oneInt)
lav_est <- extract_lavaan(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method_settings_pi.R
\name{default_settings_pi}
\alias{default_settings_pi}
\title{default arguments for product indicator approaches}
\usage{
default_settings_pi(method = c("rca", "uca", "pind", "dblcent", "ca"))
}
\arguments{
\item{method}{which method to get the settings for}
}
\value{
list
}
\description{
This function returns the default settings for the product indicator approaches
}
\examples{
library(modsem)
default_settings_pi()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standardized_solution_da.R
\name{standardize_model}
\alias{standardize_model}
\title{Standardize a fitted \code{modsem_da} model}
\usage{
standardize_model(model, monte.carlo = FALSE, mc.reps = 10000, ...)
}
\arguments{
\item{model}{A fitted object of class \code{modsem_da}.
Passing any other object triggers an error.}

\item{monte.carlo}{Logical. If \code{TRUE}, the function will use Monte Carlo
simulation to obtain the standard errors of the standardized estimates.
If \code{FALSE}, the delta method is used.
Default is \code{FALSE}.}

\item{mc.reps}{Number of Monte Carlo replications. Default is 10,000.
Ignored if \code{monte.carlo = FALSE}.}

\item{...}{Arguments passed on to other functions}
}
\value{
The same object (returned invisibly) with three slots overwritten
\describe{
  \item{\code{$parTable}}{Parameter table whose columns \code{est} and \code{std.error}
        now hold standardized estimates and their (delta-method)
        standard errors, as produced by \code{\link{standardized_estimates}()}.}
  \item{\code{$coefs}}{Named numeric vector of standardized coefficients.
        Intercepts (operator \code{~1}) are removed, because a standardized
        variable has mean 0 by definition.}
  \item{\code{$vcov}}{Variance–covariance matrix corresponding to the updated
        coefficient vector.  Rows/columns for intercepts are dropped, and
        the sub-matrix associated with rescaled parameters is adjusted
        so that its diagonal equals the squared standardized standard errors.}
}
The object keeps its class attributes, allowing seamless use by downstream
S3 methods such as \code{summary()}, \code{coef()}, or \code{vcov()}.

Because the function merely transforms existing estimates, \emph{parameter
constraints imposed through shared labels remain satisfied}.
}
\description{
\code{standardize_model()} post-processes the output of
\code{\link{modsem_da}()} (or of \code{\link{modsem}())} when \code{method = "lms"} /
\code{method = "qml"}), replacing the \emph{unstandardized} coefficient vector
(\code{$coefs}) and its variance–covariance matrix (\code{$vcov}) with \emph{fully
standardized} counterparts (including the measurement model).The procedure is purely
algebraic— \strong{no re-estimation is carried out} —and is therefore fast and
deterministic.
}
\examples{
\dontrun{
# Latent interaction estimated with LMS and standardized afterwards
syntax <- "
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3
  Y ~ X + Z + X:Z
"
fit  <- modsem_da(syntax, data = oneInt, method = "lms")
sfit <- standardize_model(fit, monte.carlo = TRUE)

# Compare unstandardized vs. standardized summaries
summary(fit)  # unstandardized
summary(sfit) # standardized
}

}
\seealso{
\describe{
   \item{\code{\link{standardized_estimates}()}}{Obtains the fully standardized
         parameter table used here.}
   \item{\code{\link{modsem}()}}{Fit model using LMS or QML approaches.}
   \item{\code{\link{modsem_da}()}}{Fit model using LMS or QML approaches.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_da.R
\name{modsem_da}
\alias{modsem_da}
\title{Interaction between latent variables using LMS and QML approaches}
\usage{
modsem_da(
  model.syntax = NULL,
  data = NULL,
  group = NULL,
  method = "lms",
  verbose = NULL,
  optimize = NULL,
  nodes = NULL,
  missing = NULL,
  convergence.abs = NULL,
  convergence.rel = NULL,
  optimizer = NULL,
  center.data = NULL,
  standardize.data = NULL,
  standardize.out = NULL,
  standardize = NULL,
  mean.observed = NULL,
  cov.syntax = NULL,
  double = NULL,
  calc.se = NULL,
  FIM = NULL,
  EFIM.S = NULL,
  OFIM.hessian = NULL,
  EFIM.parametric = NULL,
  robust.se = NULL,
  R.max = NULL,
  max.iter = NULL,
  max.step = NULL,
  start = NULL,
  epsilon = NULL,
  quad.range = NULL,
  adaptive.quad = NULL,
  adaptive.frequency = NULL,
  adaptive.quad.tol = NULL,
  n.threads = NULL,
  algorithm = NULL,
  em.control = NULL,
  ordered = NULL,
  ordered.mc.reps = NULL,
  ordered.min.iter = 20L,
  ordered.max.iter = 250L,
  ordered.tol = 1e-04,
  ordered.rng.seed = NULL,
  ordered.fixed.seed = FALSE,
  ordered.polyak.juditsky = TRUE,
  ordered.pj.extrapolate = TRUE,
  ordered.se = c("delta", "penalized", "naive", "mixed"),
  ordered.se.penalty = 0.25,
  ordered.delta.reps = NULL,
  ordered.delta.epsilon = 0.01,
  ordered.boot.reps = 1000L,
  ordered.standardize = TRUE,
  cluster = NULL,
  cr1s = FALSE,
  sampling.weights = NULL,
  sampling.weights.normalization = NULL,
  rcs = FALSE,
  rcs.choose = NULL,
  rcs.scale.corrected = TRUE,
  orthogonal.x = NULL,
  orthogonal.y = NULL,
  auto.fix.first = NULL,
  auto.fix.single = NULL,
  auto.split.syntax = NULL,
  fix.composite.var = NULL,
  ...
)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{A dataframe with observed variables used in the model.}

\item{group}{Character. A variable name in the data frame defining the groups in a multiple
group analysis}

\item{method}{method to use:
\describe{
  \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
  \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
}}

\item{verbose}{should estimation progress be shown}

\item{optimize}{should starting parameters be optimized}

\item{nodes}{number of quadrature nodes (points of integration) used in \code{lms},
increased number gives better estimates but slower computation. How many are needed depends on the complexity of the model.
For simple models, somewhere between 16-24 nodes should be enough; for more complex models, higher numbers may be needed.
For models where there is an interaction effect between an endogenous and exogenous variable,
the number of nodes should be at least 32, but practically (e.g., ordinal/skewed data), more than 32 is recommended. In cases
where data is non-normal, it might be better to use the \code{qml} approach instead.
You can also consider setting \code{adaptive.quad = TRUE}.}

\item{missing}{How should missing values be handled? If \code{"listwise"} (default) missing values
are removed list-wise (alias: \code{"complete"} or \code{"casewise"}).
If \code{impute} values are imputed using \code{Amelia::amelia}.
If \code{"fiml"} (alias: \code{"ml"} or \code{"direct"}), full information maximum
likelihood (FIML) is used. FIML can be (very) computationally intensive.}

\item{convergence.abs}{Absolute convergence criterion.
Lower values give better estimates but slower computation. Not relevant when
using the QML approach. For the LMS approach the EM-algorithm stops whenever
the relative or absolute convergence criterion is reached.}

\item{convergence.rel}{Relative convergence criterion.
Lower values give better estimates but slower computation.
For the LMS approach the EM-algorithm stops whenever
the relative or absolute convergence criterion is reached.}

\item{optimizer}{optimizer to use, can be either \code{"nlminb"} or \code{"L-BFGS-B"}. For LMS, \code{"nlminb"} is recommended.
For QML, \code{"L-BFGS-B"} may be faster if there is a large number of iterations, but slower if there are few iterations.}

\item{center.data}{should data be centered before fitting model}

\item{standardize.data}{should data be scaled before fitting model, will be overridden by
\code{standardize} if \code{standardize} is set to \code{TRUE}.}

\item{standardize.out}{should output be standardized (note will alter the relationships of
  parameter constraints since parameters are scaled unevenly, even if they
  have the same label). This does not alter the estimation of the model, only the
  output.

\strong{NOTE}: It is recommended that you estimate the model normally and then standardize the output using
\code{\link{standardize_model}}, \code{\link{standardized_estimates}} or \code{summary(<modsem_da-object>, standardize=TRUE)}.}

\item{standardize}{will standardize the data before fitting the model, remove the mean
  structure of the observed variables, and standardize the output. Note that \code{standardize.data},
  \code{mean.observed}, and \code{standardize.out} will be overridden by \code{standardize} if \code{standardize} is set to \code{TRUE}.

\strong{NOTE}: It is recommended that you estimate the model normally and then standardize the output using
  \code{\link{standardized_estimates}}.}

\item{mean.observed}{should the mean structure of the observed variables be estimated?
  This will be overridden by \code{standardize}, if \code{standardize} is set to \code{TRUE}.

\strong{NOTE}: Not recommended unless you know what you are doing.}

\item{cov.syntax}{model syntax for implied covariance matrix of exogenous latent variables
(see \code{vignette("interaction_two_etas", "modsem")}).}

\item{double}{try to double the number of dimensions of integration used in LMS,
this will be extremely slow but should be more similar to \code{mplus}.}

\item{calc.se}{should standard errors be computed? \strong{NOTE}: If \code{FALSE}, the information matrix will not be computed either.}

\item{FIM}{should the Fisher information matrix be calculated using the observed or expected values? Must be either \code{"observed"} or \code{"expected"}.}

\item{EFIM.S}{if the expected Fisher information matrix is computed, \code{EFIM.S} selects the number of Monte Carlo samples. Defaults to 100.
\strong{NOTE}: This number should likely be increased for better estimates (e.g., 1000), but it might drasticly increase computation time.}

\item{OFIM.hessian}{Logical. If \code{TRUE} (default), standard errors are
  based on the negative Hessian (observed Fisher information).
  If \code{FALSE}, they come from the outer product
  of individual score vectors (OPG). For correctly specified models,
  these two matrices are asymptotically equivalent; yielding nearly identical
  standard errors in large samples. The Hessian usually shows smaller finite-sample
  variance (i.e., it's more consistent), and is therefore the default.

  Note, that the Hessian is not always positive definite, and is more computationally
  expensive to calculate. The OPG should always be positive definite, and a lot
  faster to compute. If the model is correctly specified, and the sample size is large,
  then the two should yield similar results, and switching to the OPG can save a
  lot of time. Note, that the required sample size depends on the complexity of the model.

  A large difference between Hessian and OPG suggests misspecification, and
  \code{robust.se = TRUE} should be set to obtain sandwich (robust) standard errors.}

\item{EFIM.parametric}{should data for calculating the expected Fisher information matrix be
simulated parametrically (simulated based on the assumptions and implied parameters
from the model), or non-parametrically (stochastically sampled)? If you believe that
normality assumptions are violated, \code{EFIM.parametric = FALSE} might be the better option.}

\item{robust.se}{should robust standard errors be computed, using the sandwich estimator?}

\item{R.max}{Maximum population size (not sample size) used in the calculated of the expected
fischer information matrix.}

\item{max.iter}{maximum number of iterations.}

\item{max.step}{maximum steps for the M-step in the EM algorithm (LMS).}

\item{start}{starting parameters.}

\item{epsilon}{finite difference for numerical derivatives.}

\item{quad.range}{range in z-scores to perform numerical integration in LMS using,
when using quasi-adaptive Gaussian-Hermite Quadratures. By default \code{Inf}, such that \code{f(t)} is integrated from -Inf to Inf,
but this will likely be inefficient and pointless at a large number of nodes. Nodes outside
\code{+/- quad.range} will be ignored.}

\item{adaptive.quad}{should a quasi adaptive quadrature be used? If \code{TRUE}, the quadrature nodes will be adapted to the data.
If \code{FALSE}, the quadrature nodes will be fixed. Default is \code{FALSE}. The adaptive quadrature does not fit an adaptive
quadrature to each participant, but instead tries to place more nodes where posterior distribution is highest. Compared with a
fixed Gauss Hermite quadrature this usually means that less nodes are placed at the tails of the distribution.}

\item{adaptive.frequency}{How often should the quasi-adaptive quadrature be calculated? Defaults to 3, meaning
that it is recalculated every third EM-iteration.}

\item{adaptive.quad.tol}{Relative error tolerance for quasi adaptive quadrature. Defaults to \code{1e-12}.}

\item{n.threads}{number of threads to use for parallel processing. If \code{NULL}, it will use <= 2 threads.
If an integer is specified, it will use that number of threads (e.g., \code{n.threads = 4} will use 4 threads).
If \code{"default"}, it will use the default number of threads (2).
If \code{"max"}, it will use all available threads, \code{"min"} will use 1 thread.}

\item{algorithm}{algorithm to use for the EM algorithm. Can be either \code{"EM"} or \code{"EMA"}.
\code{"EM"} is the standard EM algorithm. \code{"EMA"} is an
accelerated EM procedure that uses Quasi-Newton and Fisher Scoring
optimization steps when needed. Default is \code{"EM"}.}

\item{em.control}{a list of control parameters for the EM algorithm. See \code{\link{default_settings_da}} for defaults.}

\item{ordered}{Variables to be treated as ordered. Ordered indicators are handled
  with a Monte-Carlo correction for the LMS/QML estimator on standardized category
  scores. The fitted model is used to repeatedly simulate continuous indicators,
  ordinalize them to the observed cumulative category proportions, refit the
  standardized LMS/QML working model, and solve the resulting fixed-point problem.
  This follows the same general Monte-Carlo consistency logic as the MC-OrdPLSc
  algorithm described by Slupphaug, Mehmetoglu, and Mittner (2026,
  \doi{10.31234/osf.io/fwzj6_v1}).

  Threshold standard errors are computed using a simple bootstrap of the category counts
  (see \code{ordered.boot.reps}), which does not propagate uncertainty from the
  model parameter estimates.}

\item{ordered.mc.reps}{Integer. Monte-Carlo sample size used in each ordered MC
correction step. Larger values reduce simulation noise but increase runtime.}

\item{ordered.min.iter}{Integer. Minimum number of Robbins-Monro iterations for the
ordered MC correction.}

\item{ordered.max.iter}{Integer. Maximum number of Robbins-Monro iterations for the
ordered MC correction.}

\item{ordered.tol}{Convergence tolerance for the ordered MC Robbins-Monro updates.}

\item{ordered.rng.seed}{Optional integer random seed used by the ordered MC
correction.}

\item{ordered.fixed.seed}{Logical. If \code{TRUE} and \code{ordered.rng.seed = NULL},
a fixed seed is drawn once and reused throughout the ordered MC correction to
improve numerical stability.}

\item{ordered.polyak.juditsky}{Logical. Should Polyak-Juditsky averaging be used in
the ordered MC Robbins-Monro solver?}

\item{ordered.pj.extrapolate}{Logical. If \code{TRUE}, use extrapolation of the
Polyak-Juditsky path to estimate the convergence point. If \code{FALSE}, the
averaged iterate is used directly.}

\item{ordered.se}{Character string selecting the ordered MC standard-error correction.
\code{"delta"} (default) uses the delta method for all free parameters.
\code{"penalized"} uses a conservative variance inflation
based on the discrepancy between the naive and MC-corrected standardized estimates.
\code{"naive"} uses the fast diagonal rescaling approximation.
\code{"mixed"} uses the delta method for the structural path coefficients
only, and penalized standard errors for the remaining parameters. This is
considerably faster, and is a good option if you're only interested in
the structural model.}

\item{ordered.se.penalty}{Non-negative numeric multiplier used when
\code{ordered.se = "penalized"}. The penalty adds
\code{ordered.se.penalty * (theta_mc - theta_naive)^2} to the diagonal of the
naive covariance matrix on the variance scale.}

\item{ordered.delta.reps}{Integer. Monte-Carlo sample size used when approximating
the ordered MC delta-method Jacobian. Only relevant if
\code{ordered.se} is \code{"delta"} or \code{"mixed"}.}

\item{ordered.delta.epsilon}{Finite-difference step size used for the ordered MC
delta-method Jacobian. Only relevant if \code{ordered.se} is \code{"delta"}
or \code{"mixed"}.}

\item{ordered.boot.reps}{Integer. Number of bootstrap replications used to compute
standard errors for the thresholds of ordered indicators. The bootstrap resamples
the category counts at the observed sample size, holding the simulated
(model-implied) reference distribution of the continuous indicators fixed. Note
that this does not account for the uncertainty in the model parameter estimates,
and should be seen as a rough approximation. Set to \code{0} to disable.}

\item{ordered.standardize}{Logical. Should scored ordered indicators be standardized
before the observed-data fit and after ordinalizing simulated indicators? This is
recommended for numerical stability and is enabled by default.}

\item{cluster}{Clusters used to compute standard errors robust to non-indepence of observations. Must be paired with
\code{robust.se = TRUE}.}

\item{cr1s}{Logical; if \code{TRUE}, apply the CR1S small-sample correction factor
to the cluster-robust variance estimator. The CR1S factor is
\eqn{(G / (G - 1)) \cdot ((N - 1) / (N - q))}, where \eqn{G} is the number of
clusters, \eqn{N} is the total number of observations, and \eqn{q} is the number
of free parameters. This adjustment inflates standard errors to reduce the
small-sample downward bias present in the basic cluster-robust (CR0) estimator,
especially when \eqn{G} is small. If \code{FALSE}, the unadjusted CR0 estimator
is used. Defaults to \code{TRUE}. Only relevant if \code{cluster} is specified.}

\item{sampling.weights}{A variable name in the data frame containing sampling weight information.
Depending on the sampling.weights.normalization argument, these weights may be rescaled (or not)
so that their sum equals the number of observations (total or per group)}

\item{sampling.weights.normalization}{If \code{"none"}, the sampling weights (if provided) will not be
transformed. If \code{"total"}, the sampling weights are normalized by dividing by the total sum
of the weights, and multiplying again by the total sample size. If \code{"group"}, the sampling
weights are normalized per group: by dividing by the sum of the weights (in each group), and
multiplying again by the group size. The default is \code{"total"}.}

\item{rcs}{Should latent variable indicators be replaced with reliability-corrected
single item indicators instead? See \code{\link{relcorr_single_item}}.}

\item{rcs.choose}{Which latent variables should get their indicators replaced with
reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
as the \code{choose} argument.}

\item{rcs.scale.corrected}{Should reliability-corrected items be scale-corrected? If \code{TRUE}
reliability-corrected single items are corrected for differences in factor loadings between
the items. Default is \code{TRUE}.}

\item{orthogonal.x}{If \code{TRUE}, all covariances among exogenous latent variables only are set to zero.
Default is \code{FALSE}.}

\item{orthogonal.y}{If \code{TRUE}, all covariances among endogenous latent variables only are set to zero.
If \code{FALSE} residual covariances are added between pure endogenous variables;
those that are predicted by no other endogenous variable in the structural model.
Default is \code{FALSE}.}

\item{auto.fix.first}{If \code{TRUE} the factor loading of the first indicator, for
a given latent variable is fixed to \code{1}. If \code{FALSE} no loadings are fixed
(automatically). Note that that this might make it such that the model no longer is
identified. Default is \code{TRUE}. \strong{NOTE} this behaviour is overridden
if the first loading is labelled, where it gets treated as a free parameter instead. This
differs from the default behaviour in \code{lavaan}.}

\item{auto.fix.single}{If \code{TRUE}, the residual variance of
an observed indicator is set to zero if it is the only indicator of a latent variable.
If \code{FALSE} the residual variance is not fixed to zero, and treated as a free parameter
of the model. Default is \code{TRUE}. \strong{NOTE} this behaviour is overridden
if the first loading is labelled, where it gets treated as a free parameter instead.}

\item{auto.split.syntax}{Should the model syntax automatically be split into a
linear and non-linear part? This is done by moving the structural model for
linear endogenous variables (used in interaction terms) into the \code{cov.syntax}
argument. This can potentially allow interactions between two endogenous variables
given that both are linear (i.e., not affected by interaction terms). This is
\code{FALSE} by default for the LMS approach.
When using the QML approach interation effects between exogenous and endogenous
variables can in some cases be biased, if the model is not split beforehand.
The default is therefore \code{TRUE} for the QML approach.}

\item{fix.composite.var}{If \code{TRUE} (default) the block covariance structure
of composite indiactors is fixed.}

\item{...}{additional arguments to be passed to the estimation function.}
}
\value{
\code{modsem_da} object
}
\description{
\code{modsem_da()} is a function for estimating interaction effects between latent variables
in structural equation models (SEMs) using distributional analytic (DA) approaches.
Methods for estimating interaction effects in SEMs can basically be split into
two frameworks:
1. Product Indicator-based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"},
\code{"ca"}, \code{"pind"})
2. Distributionally based approaches (\code{"lms"}, \code{"qml"}).

\code{modsem_da()} handles the latter and can estimate models using both QML and LMS,
necessary syntax, and variables for the estimation of models with latent product indicators.

\strong{NOTE}: Run \code{\link{default_settings_da}} to see default arguments.
}
\examples{
library(modsem)
# For more examples, check README and/or GitHub.
# One interaction
m1 <- "
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
"

\dontrun{
# QML Approach
est_qml <- modsem_da(m1, oneInt, method = "qml")
summary(est_qml)

# Theory Of Planned Behavior
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
"

# LMS Approach
est_lms <- modsem_da(tpb, data = TPB, method = "lms")
summary(est_lms)
}
}
\references{
Slupphaug, K., Mehmetoglu, M., and Mittner, M. (2026, March 21).
\emph{Consistent Estimates from Biased Estimators: Monte-Carlo Consistent Partial
Least Squares for Latent Interaction Models with Ordinal Indicators}. PsyArXiv.
\doi{10.31234/osf.io/fwzj6_v1}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_interaction.R
\name{plot_surface}
\alias{plot_surface}
\title{Plot Surface for Interaction Effects}
\usage{
plot_surface(
  x,
  z,
  y,
  model,
  min_x = -3,
  max_x = 3,
  min_z = -3,
  max_z = 3,
  standardized = FALSE,
  detail = 0.01,
  xz = NULL,
  colorscale = "Viridis",
  reversescale = FALSE,
  showscale = TRUE,
  cmin = NULL,
  cmax = NULL,
  surface_opacity = 1,
  grid = FALSE,
  grid_nx = 12,
  grid_ny = 12,
  grid_color = "rgba(0,0,0,0.45)",
  group = NULL,
  ...
)
}
\arguments{
\item{x}{A character string specifying the name of the first predictor variable.}

\item{z}{A character string specifying the name of the second predictor variable.}

\item{y}{A character string specifying the name of the outcome variable.}

\item{model}{A model object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}}, \code{\link{modsem_mplus}}, or \code{lavaan}. The model should
include paths for the predictors (\code{x}, \code{z}, and \code{xz}) to the outcome (\code{y}).}

\item{min_x}{Numeric. Minimum value of \code{x} in z-scores. Default is -3.}

\item{max_x}{Numeric. Maximum value of \code{x} in z-scores. Default is 3.}

\item{min_z}{Numeric. Minimum value of \code{z} in z-scores. Default is -3.}

\item{max_z}{Numeric. Maximum value of \code{z} in z-scores. Default is 3.}

\item{standardized}{Should coefficients be standardized beforehand?}

\item{detail}{Numeric. Step size for the grid of \code{x} and \code{z} values, determining the resolution of the surface.
Smaller values increase plot resolution. Default is \code{1e-2}.}

\item{xz}{Optional. A character string or vector specifying the interaction term between \code{x} and \code{z}.
If \code{NULL}, the interaction term is constructed as \code{paste(x, z, sep = ":")} and adjusted for specific model classes.}

\item{colorscale}{Character or list. Colorscale used to color the surface.
- Default is \code{"Viridis"}, which matches the classic Plotly default.
- Can be a built-in palette name (e.g., \code{"Greys"}, \code{"Plasma"}, \code{"Turbo"}),
  or a custom two-column list with numeric stops (\code{0}–\code{1}) and color codes.
Example custom scale:
\code{list(c(0, "white"), c(1, "black"))} for a black-and-white gradient.}

\item{reversescale}{Logical. If \code{TRUE}, reverses the color mapping so that
low values become high colors and vice versa.
Default is \code{FALSE}.}

\item{showscale}{Logical. If \code{TRUE}, displays the colorbar legend
alongside the plot.
Default is \code{TRUE}.}

\item{cmin}{Numeric or \code{NULL}. The minimum value for the colorscale mapping.
If \code{NULL}, the minimum of the data (\code{proj_y}) is used automatically.
Use this to standardize the color range across multiple plots.}

\item{cmax}{Numeric or \code{NULL}. The maximum value for the colorscale mapping.
If \code{NULL}, the maximum of the data (\code{proj_y}) is used automatically.
Use this to standardize the color range across multiple plots.}

\item{surface_opacity}{Numeric (0–1). Controls the opacity of the surface.
- \code{1} = fully opaque (default)
- \code{0} = fully transparent
Useful when overlaying multiple surfaces or highlighting gridlines.}

\item{grid}{Logical. If \code{TRUE}, draws gridlines (wireframe)
directly on the surface using Plotly's contour features.
Default is \code{FALSE}.}

\item{grid_nx}{Integer. Approximate number of gridlines to draw along the
**x-axis** direction when \code{grid = TRUE}.
Higher values create a denser grid.
Default is \code{12}.}

\item{grid_ny}{Integer. Approximate number of gridlines to draw along the
**y-axis** direction when \code{grid = TRUE}.
Higher values create a denser grid.
Default is \code{12}.}

\item{grid_color}{Character. Color of the gridlines drawn on the surface.
Must be a valid CSS color string, including \code{rgba()} for transparency.
- Default is \code{"rgba(0,0,0,0.45)"} (semi-transparent black).
Example: \code{"rgba(255,255,255,0.8)"} for semi-transparent white lines.}

\item{group}{Which group to create surface plot for. Only relevant for multigroup
models. Must be an integer index, representing the nth group.}

\item{...}{Additional arguments passed to \code{plotly::plot_ly}.}
}
\value{
A \code{plotly} surface plot object displaying the predicted values of \code{y} across the grid of \code{x} and \code{z} values.
  The color bar shows the values of \code{y}.
}
\description{
Generates a 3D surface plot to visualize the interaction effect of two variables (\code{x} and \code{z})
on an outcome (\code{y})
using parameter estimates from a supported model object (e.g., \code{lavaan} or \code{modsem}).
The function allows specifying ranges for \code{x} and \code{z} in standardized z-scores, which are then transformed
back to the original scale based on their means and standard deviations.
}
\details{
The input \code{min_x}, \code{max_x}, \code{min_z}, and \code{max_z} define the range of \code{x} and \code{z} values in z-scores.
These are scaled by the standard deviations and shifted by the means of the respective variables, obtained
from the model parameter table. The resulting surface shows the predicted values of \code{y} over the grid of \code{x} and \code{z}.

The function supports models of class \code{modsem} (with subclasses \code{modsem_pi}, \code{modsem_da}, \code{modsem_mplus}) and \code{lavaan}.
For \code{lavaan} models, it is not designed for multigroup models, and a warning will be issued if multiple groups are detected.
}
\note{
The interaction term (\code{xz}) may need to be manually specified for some models. For non-\code{lavaan} models,
interaction terms may have their separator (\code{:}) removed based on circumstances.
}
\examples{
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)
plot_surface("X", "Z", "Y", model = est1)

\dontrun{
tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"

est2 <- modsem(tpb, TPB, method = "lms", nodes = 32)
plot_surface(x = "INT", z = "PBC", y = "BEH", model = est2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_fit.R
\name{compare_fit}
\alias{compare_fit}
\title{compare model fit for \code{modsem} models}
\usage{
compare_fit(est_h1, est_h0, ...)
}
\arguments{
\item{est_h1}{object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}} representing the
alternative hypothesis model (with interaction terms).}

\item{est_h0}{object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}} representing the
null hypothesis model (without interaction terms).}

\item{...}{additional arguments passed to the underlying comparison function. E.g.,
for \code{modsem_pi} models, this can be used to pass arguments to \code{lavaan::lavTestLRT}.
currently only used for \code{modsem_pi} models.}
}
\description{
Compare the fit of two models using the likelihood ratio test (LRT).
\code{est_h0} is the null hypothesis model, and \code{est_h1} the alternative hypothesis model.
Importantly, the function assumes that \code{est_h0} does not have more free parameters
(i.e., degrees of freedom) than \code{est_h1} (the alternative hypothesis model).
}
\examples{
\dontrun{
m1 <- "
 # Outer Model
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Z =~ z1 + z2 + z3

 # Inner model
 Y ~ X + Z + X:Z
"

# LMS approach
est_h1 <- modsem(m1, oneInt, "lms")
est_h0 <- estimate_h0(est_h1, calc.se=FALSE) # std.errors are not needed
compare_fit(est_h1 = est_h1, est_h0 = est_h0)

# Double centering approach
est_h1 <- modsem(m1, oneInt, method = "dblcent")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)

# Constrained approach
est_h1 <- modsem(m1, oneInt, method = "ca")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method_settings_da.R
\name{default_settings_da}
\alias{default_settings_da}
\title{default arguments fro LMS and QML approach}
\usage{
default_settings_da(method = c("lms", "qml"))
}
\arguments{
\item{method}{which method to get the settings for}
}
\value{
list
}
\description{
This function returns the default settings for the LMS and QML approach.
}
\examples{
library(modsem)
default_settings_da()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_pi.R
\name{modsem_pi}
\alias{modsem_pi}
\title{Interaction between latent variables using product indicators}
\usage{
modsem_pi(
  model.syntax = NULL,
  data = NULL,
  method = "dblcent",
  match = NULL,
  match.recycle = NULL,
  standardize.data = FALSE,
  center.data = FALSE,
  first.loading.fixed = FALSE,
  center.before = NULL,
  center.after = NULL,
  residuals.prods = NULL,
  residual.cov.syntax = NULL,
  constrained.prod.mean = NULL,
  constrained.loadings = NULL,
  constrained.var = NULL,
  res.cov.method = NULL,
  res.cov.across = NULL,
  composite.int.res.cov = FALSE,
  auto.scale = "none",
  auto.center = "none",
  estimator = "ML",
  group = NULL,
  cluster = NULL,
  run = TRUE,
  na.rm = FALSE,
  suppress.warnings.lavaan = FALSE,
  suppress.warnings.match = FALSE,
  rcs = FALSE,
  rcs.choose = NULL,
  rcs.res.cov.xz = rcs,
  rcs.mc.reps = 1e+05,
  rcs.scale.corrected = TRUE,
  LAVFUN = lavaan::sem,
  ...
)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{dataframe}

\item{method}{method to use:
\describe{
  \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
  \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
  \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
  \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
  \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
}}

\item{match}{should the product indicators be created by using the match-strategy}

\item{match.recycle}{should the indicators be recycled when using the match-strategy? I.e.,
if one of the latent variables have fewer indicators than the other, some indicators
are recycled to match the latent variable with the most indicators.}

\item{standardize.data}{should data be scaled before fitting model}

\item{center.data}{should data be centered before fitting model}

\item{first.loading.fixed}{Should the first factor loading in the latent product be fixed to one? Defaults to \code{FALSE}, as
this already happens in \code{lavaan} by default. If \code{TRUE}, the first factor loading in the latent product is fixed to one.
Manually in the generated syntax (e.g., \code{XZ =~ 1*x1z1}).'}

\item{center.before}{should indicators in products be centered before computing products.}

\item{center.after}{should indicator products be centered after they have been computed?}

\item{residuals.prods}{should indicator products be centered using residuals.}

\item{residual.cov.syntax}{should syntax for residual covariances be produced.}

\item{constrained.prod.mean}{should syntax for product mean be produced.}

\item{constrained.loadings}{should syntax for constrained loadings be produced.}

\item{constrained.var}{should syntax for constrained variances be produced.}

\item{res.cov.method}{method for constraining residual covariances. Options are
\describe{
  \item{"simple"}{Residuals of product indicators with variables in common are allowed to covary freely. Defualt for most approches.}
  \item{"ca"}{Residual covariances of product indicators are constrained according to the constrained approach.}
  \item{"equality"}{Residuals of product indicators with variables in common are constrained to have equal covariances".
                    Can be useful for models where the model is unidentifiable using \code{res.cov.method == "simple"},
                    (e.g., when there is an interaction between an observed and a latent variable).}
  \item{"none"}{Residual covariances between product indicators are not specificed (i.e., constrained to zero).
                Produces the same results as \code{constrained.cov.syntax = FALSE}.
                Can be useful for models where the model is unidentifiable using \code{res.cov.method == "simple"},
                (e.g., when there is an interaction between an observed and a latent variable).}
}}

\item{res.cov.across}{Should residual covariances be specified/freed across different interaction terms.
For example if you have two interaction terms \code{X:Z} and \code{X:W} the residuals of the
generated product indicators \code{x1:z1} and \code{x1:w1} may be correlated. If \code{TRUE}
residual covariances are allowed across different latent interaction terms. If \code{FALSE}
residual covariances are only allowed between product indicators which belong to the same
latent interaction term.}

\item{composite.int.res.cov}{Should residual covariance syntax be generated for interaction
terms that are fully composite (i.e., all component variables defined with \code{<~})?
Defaults to \code{FALSE} because for fully composite interaction terms the residual
covariances of the product indicators are not needed. Set to \code{TRUE} to force
residual covariance syntax for composite interaction terms (useful for testing).}

\item{auto.scale}{methods which should be scaled automatically (usually not useful)}

\item{auto.center}{methods which should be centered automatically (usually not useful)}

\item{estimator}{estimator to use in \code{lavaan}}

\item{group}{group variable for multigroup analysis}

\item{cluster}{cluster variable for multilevel models}

\item{run}{should the model be run via \code{lavaan}, if \code{FALSE} only modified syntax and data is returned}

\item{na.rm}{should missing values be removed (case-wise)? Defaults to FALSE. If \code{TRUE}, missing values are removed case-wise.
If \code{FALSE} they are not removed.}

\item{suppress.warnings.lavaan}{should warnings from \code{lavaan} be suppressed?}

\item{suppress.warnings.match}{should warnings from \code{match} be suppressed?}

\item{rcs}{Should latent variable indicators be replaced with reliability-corrected
single item indicators instead? See \code{\link{relcorr_single_item}}.}

\item{rcs.choose}{Which latent variables should get their indicators replaced with
reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
as the \code{choose} argument.}

\item{rcs.res.cov.xz}{Should the residual (co-)variances of the product indicators
created from the reliability-corrected single items (created if \code{rcs = TRUE})
be specified and constrained before estimating the model? If \code{TRUE} the estimates
for the constraints are approximated using a monte carlo simulation (see the \code{rcs.mc.reps} argument).
If \code{FALSE} the residual variances are not specified, which usually mean that all
are constrained to zero.}

\item{rcs.mc.reps}{Sample size used in monte-carlo simulation, when approximating the
the estimates of the residual (co-)variances between the product indicators formed
by reliabiliyt-corrected single items (see the \code{rcs.res.cov.xz} argument).}

\item{rcs.scale.corrected}{Should reliability corrected items be scale-corrected? If \code{TRUE}
reliability-corrected single items are corrected for differences in factor loadings between
the items. Default is \code{TRUE}.}

\item{LAVFUN}{Function used to estimate the model. Defaults to \code{lavaan::sem}.}

\item{...}{arguments passed to \code{LAVFUN}}
}
\value{
\code{modsem} object
}
\description{
\code{modsem_pi()} is a function for estimating interaction effects between latent variables,
in structural equation models (SEMs), using product indicators.
Methods for estimating interaction effects in SEMs can basically be split into
two frameworks:
1. Product Indicator based approaches (\code{"dblcent"}, \code{"rca"}, \code{"uca"},
\code{"ca"}, \code{"pind"}), and
2. Distributionally based approaches (\code{"lms"}, \code{"qml"}).
\code{modsem_pi()} is essentially a fancy wrapper for \code{lavaan::sem()} which generates the
necessary syntax and variables for the estimation of models with latent product indicators.
Use \code{default_settings_pi()} to get the default settings for the different methods.
}
\examples{
library(modsem)
# For more examples, check README and/or GitHub.
# One interaction
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'

# Double centering approach
est <- modsem_pi(m1, oneInt)
summary(est)

\dontrun{
# The Constrained Approach
est_ca <- modsem_pi(m1, oneInt, method = "ca")
summary(est_ca)
}

# Theory Of Planned Behavior
tpb <- '
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  # Covariances
  ATT ~~ SN + PBC
  PBC ~~ SN
  # Causal Relationships
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ INT:PBC
'

# Double centering approach
est_tpb <- modsem_pi(tpb, data = TPB)
summary(est_tpb)

\dontrun{
# The Constrained Approach
est_tpb_ca <- modsem_pi(tpb, data = TPB, method = "ca")
summary(est_tpb_ca)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics_modsem_da.R,
%   R/generics_modsem_mplus.R, R/generics_modsem_pi.R
\name{summary.modsem_da}
\alias{summary.modsem_da}
\alias{summary.modsem_mplus}
\alias{summary.modsem_pi}
\title{summary for modsem objects}
\usage{
\method{summary}{modsem_da}(
  object,
  H0 = is_interaction_model(object),
  verbose = interactive(),
  r.squared = TRUE,
  fit = FALSE,
  adjusted.stat = FALSE,
  digits = 3,
  scientific = FALSE,
  ci = FALSE,
  standardized = FALSE,
  centered = FALSE,
  monte.carlo = FALSE,
  mc.reps = 10000,
  loadings = TRUE,
  regressions = TRUE,
  covariances = TRUE,
  intercepts = TRUE,
  variances = TRUE,
  thresholds = TRUE,
  var.interaction = FALSE,
  ...
)

\method{summary}{modsem_mplus}(
  object,
  scientific = FALSE,
  standardized = FALSE,
  ci = FALSE,
  digits = 3,
  loadings = TRUE,
  regressions = TRUE,
  covariances = TRUE,
  intercepts = TRUE,
  variances = TRUE,
  ...
)

\method{summary}{modsem_pi}(
  object,
  H0 = is_interaction_model(object),
  r.squared = TRUE,
  adjusted.stat = FALSE,
  digits = 3,
  scientific = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{modsem object to summarized}

\item{H0}{Should the baseline model be estimated, and used to produce
comparative fit?}

\item{verbose}{Should messages be printed?}

\item{r.squared}{Calculate R-squared.}

\item{fit}{Print additional fit measures.}

\item{adjusted.stat}{Should sample size corrected/adjustes AIC and BIC be reported?}

\item{digits}{Number of digits for printed numerical values}

\item{scientific}{Should scientific format be used for p-values?}

\item{ci}{print confidence intervals}

\item{standardized}{standardize estimates}

\item{centered}{Print mean centered estimates.}

\item{monte.carlo}{Should Monte Carlo bootstrapped standard errors be used? Only
relevant if \code{standardized = TRUE}. If \code{FALSE} delta method is used instead.}

\item{mc.reps}{Number of Monte Carlo repetitions. Only relevant if \code{monte.carlo = TRUE},
and \code{standardized = TRUE}.}

\item{loadings}{print loadings}

\item{regressions}{print regressions}

\item{covariances}{print covariances}

\item{intercepts}{print intercepts}

\item{variances}{print variances}

\item{thresholds}{Print thresholds.}

\item{var.interaction}{If FALSE variances for interaction terms will be removed
from the output.}

\item{...}{arguments passed to lavaan::summary()}
}
\description{
summary for modsem objects

summary for modsem objects

summary for modsem objects
}
\examples{
\dontrun{
m1 <- "
 # Outer Model
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Z =~ z1 + z2 + z3

 # Inner model
 Y ~ X + Z + X:Z
"

est1 <- modsem(m1, oneInt, "qml")
summary(est1, ci = TRUE, scientific = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{is_interaction_model}
\alias{is_interaction_model}
\title{Check if model object has interaction terms}
\usage{
is_interaction_model(object)
}
\arguments{
\item{object}{An object of class \code{modsem_pi} or \code{modsem_da},
respectively.}
}
\value{
Logical. \code{TRUE} if the model has an interaction term,
  otherwise it returns \code{FALSE}.
}
\description{
Check if model object has interaction terms
}
\examples{
m1 <- '
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
'

est_dca <- modsem(m1, oneInt, method = "dblcent")
is_interaction_model(est_dca)

\dontrun{
est_lms <- modsem(m1, oneInt, method = "lms")
is_interaction_model(est_lms)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_mplus.R
\name{modsem_mplus}
\alias{modsem_mplus}
\title{Estimation latent interactions through \code{Mplus}}
\usage{
modsem_mplus(
  model.syntax,
  data,
  estimator = "ml",
  cluster = NULL,
  type = ifelse(is.null(cluster), yes = "random", no = "complex"),
  algorithm = "integration",
  processors = 2,
  integration = 15,
  rcs = FALSE,
  rcs.choose = NULL,
  rcs.scale.corrected = TRUE,
  output.std = TRUE,
  categorical = NULL,
  cleanup = FALSE,
  ...
)
}
\arguments{
\item{model.syntax}{lavaan/modsem syntax}

\item{data}{dataset}

\item{estimator}{estimator argument passed to \code{Mplus}.}

\item{cluster}{cluster argument passed to \code{Mplus}.}

\item{type}{type argument passed to \code{Mplus}.}

\item{algorithm}{algorithm argument passed to \code{Mplus}.}

\item{processors}{processors argument passed to \code{Mplus}.}

\item{integration}{integration argument passed to \code{Mplus}.}

\item{rcs}{Should latent variable indicators be replaced with reliability-corrected
single item indicators instead? See \code{\link{relcorr_single_item}}.}

\item{rcs.choose}{Which latent variables should get their indicators replaced with
reliability-corrected single items? It is passed to \code{\link{relcorr_single_item}}
as the \code{choose} argument.}

\item{rcs.scale.corrected}{Should reliability-corrected items be scale-corrected? If \code{TRUE}
reliability-corrected single items are corrected for differences in factor loadings between
the items. Default is \code{TRUE}.}

\item{output.std}{Should \code{STANDARDIZED} be added to \code{OUTPUT}?}

\item{categorical}{categorical argument passed to \code{Mplus}.}

\item{cleanup}{Should the \code{Mplus} files (\code{.inp}, \code{.dat} and \code{.out})
created during estimation be deleted when \code{modsem_mplus()} exits? Default is
\code{FALSE}, meaning the files are kept in the working directory.}

\item{...}{arguments passed to other functions}
}
\value{
modsem_mplus object
}
\description{
Estimation latent interactions through \code{Mplus}
}
\examples{
# Theory Of Planned Behavior
m1 <- '
# Outer Model
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

# Inner model
  Y ~ X + Z + X:Z
'

\dontrun{
# Check if Mplus is installed
run <- tryCatch({MplusAutomation::detectMplus(); TRUE},
                error = \(e) FALSE)

if (run) {
  est_mplus <- modsem_mplus(m1, data = oneInt)
  summary(est_mplus)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{var_interactions}
\alias{var_interactions}
\title{Extract or modify parTable from an estimated model with estimated variances of interaction terms}
\usage{
var_interactions(object, ...)
}
\arguments{
\item{object}{An object of class \code{\link{modsem_da}},  \code{\link{modsem_mplus}},
or a parTable of class \code{\link{data.frame}}}

\item{...}{Additional arguments passed to other functions}
}
\description{
Extract or modify parTable from an estimated model with estimated variances of interaction terms
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{oneInt}
\alias{oneInt}
\title{oneInt}
\description{
A simulated dataset based on the elementary interaction model.
}
\examples{

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est <- modsem(m1, data = oneInt)
summary(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize_partable.R
\name{summarize_partable}
\alias{summarize_partable}
\title{Summarize a parameter table from a \code{modsem} model.}
\usage{
summarize_partable(
  parTable,
  scientific = FALSE,
  ci = FALSE,
  digits = 3,
  loadings = TRUE,
  regressions = TRUE,
  covariances = TRUE,
  intercepts = TRUE,
  variances = TRUE,
  thresholds = TRUE
)
}
\arguments{
\item{parTable}{A parameter table, typically obtained from a \code{\link{modsem}} model
using \code{\link{parameter_estimates}} or \code{\link{standardized_estimates}}.}

\item{scientific}{Logical, whether to print p-values in scientific notation.}

\item{ci}{Logical, whether to include confidence intervals in the output.}

\item{digits}{Integer, number of digits to round the estimates to
(default is 3).}

\item{loadings}{Logical, whether to include factor loadings in the output.}

\item{regressions}{Logical, whether to include regression coefficients in the output.}

\item{covariances}{Logical, whether to include covariance estimates in the output.}

\item{intercepts}{Logical, whether to include intercepts in the output.}

\item{variances}{Logical, whether to include variance estimates in the output.}

\item{thresholds}{Logical, whether to include threshold estimates in the output.}
}
\value{
A summary object containing the parameter table and additional information.
}
\description{
Summarize a parameter table from a \code{modsem} model.
}
\examples{
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  # Inner Model
  Y ~ X + Z + X:Z
'
# Double centering approach
est_dca <- modsem(m1, oneInt)

std <- standardized_estimates(est_dca, correction = TRUE)
summarize_partable(std)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{modsem_coef}
\alias{modsem_coef}
\title{Wrapper for coef}
\usage{
modsem_coef(object, ...)
}
\arguments{
\item{object}{fitted model to inspect}

\item{...}{additional arguments}
}
\description{
wrapper for \code{coef}, to be used with \code{modsem::modsem_coef}, since
\code{coef} is not in the namespace of \code{modsem}, but \code{stats}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_pi.R
\name{get_pi_data}
\alias{get_pi_data}
\title{Get data with product indicators for different approaches}
\usage{
get_pi_data(model.syntax, data, method = "dblcent", match = FALSE, ...)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{data to create product indicators from}

\item{method}{method to use:
\code{"rca"} = residual centering approach,
\code{"uca"} = unconstrained approach,
\code{"dblcent"} = double centering approach,
\code{"pind"} = prod ind approach, with no constraints or centering,
\code{"custom"} = use parameters specified in the function call}

\item{match}{should the product indicators be created by using the match-strategy}

\item{...}{arguments passed to other functions (e.g., \link{modsem_pi})}
}
\value{
\code{data.frame}
}
\description{
\code{get_pi_data()} is a function for creating a dataset with product indiactors used for estimating
latent interaction models using one of the product indicator approaches.
}
\examples{
library(modsem)
library(lavaan)
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
syntax <- get_pi_syntax(m1)
data <- get_pi_data(m1, oneInt)
est <- sem(syntax, data)
summary(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem-package.R
\docType{package}
\name{modsem-package}
\alias{modsem-package}
\title{modsem: Latent Interaction (and Moderation) Analysis in Structural Equation Models (SEM)}
\description{
Estimation of interaction (i.e., moderation) effects between latent variables in structural equation models (SEM). The supported methods are: The constrained approach (Algina & Moulder, 2001). The unconstrained approach (Marsh et al., 2004). The residual centering approach (Little et al., 2006). The double centering approach (Lin et al., 2010). The latent moderated structural equations (LMS) approach (Klein & Moosbrugger, 2000). The quasi-maximum likelihood (QML) approach (Klein & Muthén, 2007) The constrained- unconstrained, residual- and double centering- approaches are estimated via 'lavaan' (Rosseel, 2012), whilst the LMS- and QML- approaches are estimated via 'modsem' it self. Alternatively model can be estimated via 'Mplus' (Muthén & Muthén, 1998-2017). References: Algina, J., & Moulder, B. C. (2001). \doi{10.1207/S15328007SEM0801_3}. "A note on estimating the Jöreskog-Yang model for latent variable interaction using 'LISREL' 8.3." Klein, A., & Moosbrugger, H. (2000). \doi{10.1007/BF02296338}. "Maximum likelihood estimation of latent interaction effects with the LMS method." Klein, A. G., & Muthén, B. O. (2007). \doi{10.1080/00273170701710205}. "Quasi-maximum likelihood estimation of structural equation models with multiple interaction and quadratic effects." Lin, G. C., Wen, Z., Marsh, H. W., & Lin, H. S. (2010). \doi{10.1080/10705511.2010.488999}. "Structural equation models of latent interactions: Clarification of orthogonalizing and double-mean-centering strategies." Little, T. D., Bovaird, J. A., & Widaman, K. F. (2006). \doi{10.1207/s15328007sem1304_1}. "On the merits of orthogonalizing powered and product terms: Implications for modeling interactions among latent variables." Marsh, H. W., Wen, Z., & Hau, K. T. (2004). \doi{10.1037/1082-989X.9.3.275}. "Structural equation models of latent interactions: evaluation of alternative estimation strategies and indicator construction." Muthén, L.K. and Muthén, B.O. (1998-2017). "'Mplus' User’s Guide. Eighth Edition." \url{https://www.statmodel.com/}. Rosseel Y (2012). \doi{10.18637/jss.v048.i02}. "'lavaan': An R Package for Structural Equation Modeling."
}
\seealso{
Useful links:
\itemize{
  \item \url{https://modsem.org}
  \item \url{https://github.com/kss2k/modsem}
}

}
\author{
\strong{Maintainer}: Kjell Solem Slupphaug \email{slupphaugkjell@gmail.com} (\href{https://orcid.org/0009-0005-8324-2834}{ORCID})

Authors:
\itemize{
  \item Kjell Solem Slupphaug \email{slupphaugkjell@gmail.com} (\href{https://orcid.org/0009-0005-8324-2834}{ORCID})
}

Other contributors:
\itemize{
  \item Mehmet Mehmetoglu \email{mehmetm@ntnu.no} (\href{https://orcid.org/0000-0002-6092-8551}{ORCID}) [contributor]
  \item Matthias Mittner \email{matthias.mittner@uit.no} (\href{https://orcid.org/0000-0003-0205-7353}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/generics_lavaan.R,
%   R/generics_modsem_da.R, R/generics_modsem_mplus.R, R/generics_modsem_pi.R
\name{standardized_estimates}
\alias{standardized_estimates}
\alias{standardized_estimates.lavaan}
\alias{standardized_estimates.modsem_da}
\alias{standardized_estimates.modsem_mplus}
\alias{standardized_estimates.modsem_pi}
\title{Get Standardized Estimates}
\usage{
standardized_estimates(object, ...)

\method{standardized_estimates}{lavaan}(
  object,
  monte.carlo = FALSE,
  mc.reps = 10000,
  tolerance.zero = 1e-10,
  ...
)

\method{standardized_estimates}{modsem_da}(
  object,
  monte.carlo = FALSE,
  mc.reps = 10000,
  tolerance.zero = 1e-10,
  rm.tmp.ov = TRUE,
  ...
)

\method{standardized_estimates}{modsem_mplus}(object, type = "stdyx", mc.reps = 1e+06, ...)

\method{standardized_estimates}{modsem_pi}(
  object,
  correction = FALSE,
  std.errors = c("rescale", "delta", "monte.carlo"),
  mc.reps = 10000,
  colon.pi = FALSE,
  ...
)
}
\arguments{
\item{object}{An object of class \code{\link{modsem_da}}, \code{\link{modsem_mplus}},
\code{\link{modsem_pi}}, or a parameter table (\code{parTable}) of class \code{data.frame}.}

\item{...}{Additional arguments passed on to \code{lavaan::standardizedSolution()}.}

\item{monte.carlo}{Logical. If \code{TRUE}, use Monte Carlo simulation to estimate
standard errors; if \code{FALSE}, use the delta method (default).}

\item{mc.reps}{Integer. Number of Monte Carlo replications to use if \code{std.errors = "monte.carlo"}.}

\item{tolerance.zero}{Threshold below which standard errors are set to \code{NA}.}

\item{rm.tmp.ov}{Should temporary (hidden) variables be removed?}

\item{type}{Type of standardized estimates to retrieve. Can be one of: \code{"stdyx", "stdy", "std", "un", "modsem"}.}

\item{correction}{Logical. Whether to apply a correction to the standardized estimates
of the interaction term. By default, \code{FALSE}, which standardizes the interaction term
such that \eqn{\sigma^2(XZ) = 1}, consistent with \code{lavaan}'s treatment of latent interactions.
This is usually wrong, as it does not account for the fact that the interaction term
is a product of two variables, which means that the variance of the interaction term
of standardized variables (usually) is not equal to 1.

Hence the scale of the interaction effect changes, such that
the standardized interaction term does not correspond to one (standardized) unit
change in the moderating variables. If \code{TRUE}, a correction is applied by
computing the interaction term \eqn{b_3 = \frac{B_3 \sigma(X) \sigma(Z)}{\sigma(Y)}} (where
\eqn{B_3} is the unstandardized coefficient for the interaction term), and solving for \eqn{\sigma(XZ)}, which
is used to correct the variance of the interaction term, and its covariances.}

\item{std.errors}{Character string indicating the method used to compute standard errors
when \code{correction = TRUE}. Options include:
\describe{
  \item{"rescale"}{Simply rescales the standard errors. Fastest option.}
  \item{"delta"}{Uses the delta method to approximate standard errors.}
  \item{"monte.carlo"}{Uses Monte Carlo simulation to estimate standard errors.}
}
Ignored if \code{correction = FALSE}.}

\item{colon.pi}{Logical. If \code{TRUE}, the interaction terms in the output will be
will be formatted using \code{:} to seperate the elements in the interaction term. Default
is \code{FALSE}, using the default formatting from \code{lavaan}. Only relevant if
\code{std.errors != "rescale"} and \code{correction = TRUE}.}
}
\value{
A \code{data.frame} with standardized estimates in the \code{est} column.
}
\description{
Computes standardized estimates of model parameters for various types of \code{\link{modsem}} objects.
}
\details{
Standard errors can either be calculated using the delta method, or a monte.carlo simulation
  (\code{monte.carlo} is not available for \code{\link{modsem_pi}} objects if \code{correction == FALSE}.).
  \strong{NOTE} that the standard errors of the standardized paramters change the assumptions of the
  model, and will in most cases yield different z and p-values, compared to the unstandardized solution.
  In almost all cases, significance testing should be based on the unstandardized solution. Since,
  the standardization process changes the model assumptions, it also changes what the p-statistics measure.
  I.e., the test statistics for the standardized and unstandardized solutions belong to different sets of
  hypothesis, which are not exactly equivalent to each other.

For \code{modsem_da} and \code{modsem_mplus} objects, the interaction term is not a formal
 variable in the model and therefore lacks a defined variance. Under assumptions of normality
 and zero-mean variables, the interaction variance is estimated as:
 \deqn{var(xz) = var(x) * var(z) + cov(x, z)^2}
 This means the standardized estimate for the interaction differs from approaches like
 \code{lavaan}, which treats the interaction as a latent variable with unit variance.

For \code{modsem_pi} objects, the interaction term is standardized by default assuming
 \code{var(xz) = 1}, but this can be overridden using the \code{correction} argument.

\strong{NOTE:} Standardized estimates are always placed in the \strong{\code{est}} column,
not \code{est.std}, regardless of model type.
}
\section{Methods (by class)}{
\itemize{
\item \code{standardized_estimates(lavaan)}: Method for \code{lavaan} objects

\item \code{standardized_estimates(modsem_da)}: Method for \code{modsem_da} objects

\item \code{standardized_estimates(modsem_mplus)}: Retrieve standardized estimates from \code{\link{modsem_mplus}} object.

\item \code{standardized_estimates(modsem_pi)}: Method for \code{modsem_pi} objects

}}
\examples{
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

  # Inner Model
  Y ~ X + Z + X:Z
'
# Double centering approach
est_dca <- modsem(m1, oneInt)

std1 <- standardized_estimates(est_dca) # no correction
summarize_partable(std1)

std2 <- standardized_estimates(est_dca, correction = TRUE) # apply correction
summarize_partable(std2)

\dontrun{
est_lms <- modsem(m1, oneInt, method = "lms")
standardized_estimates(est_lms) # correction not relevant for lms
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/est_h0.R
\name{estimate_h0}
\alias{estimate_h0}
\alias{estimate_h0.modsem_da}
\alias{estimate_h0.modsem_pi}
\title{Estimate baseline model for \code{modsem} models}
\usage{
estimate_h0(object, warn_no_interaction = TRUE, ...)

\method{estimate_h0}{modsem_da}(object, warn_no_interaction = TRUE, ...)

\method{estimate_h0}{modsem_pi}(object, warn_no_interaction = TRUE, reduced = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{\link{modsem_da}} or \code{\link{modsem_pi}}.}

\item{warn_no_interaction}{Logical. If `TRUE`, a warning is issued if no interaction terms are found in the model.}

\item{...}{Additional arguments passed to the `modsem_da` function, overriding
the arguments in the original model.}

\item{reduced}{Should the baseline model be a reduced version of the model?
If \code{TRUE}, the latent product term and its (product) indicators are kept in the model,
but the interaction coefficients are constrained to zero. If \code{FALSE}, the
interaction terms are removed completely from the model. Note that the models will no longer be
nested, if the interaction terms are removed from the model completely.}
}
\description{
Estimates a baseline model (H0) from a given model (H1).
The baseline model is estimated by removing all interaction terms from the model.
}
\section{Methods (by class)}{
\itemize{
\item \code{estimate_h0(modsem_da)}: Estimate baseline model for \code{\link{modsem_da}} objects

\item \code{estimate_h0(modsem_pi)}: Estimate baseline model for \code{\link{modsem_pi}} objects

}}
\examples{
\dontrun{
m1 <- "
 # Outer Model
 X =~ x1 + x2 + x3
 Y =~ y1 + y2 + y3
 Z =~ z1 + z2 + z3

 # Inner model
 Y ~ X + Z + X:Z
"

# LMS approach
est_h1 <- modsem(m1, oneInt, "lms")
est_h0 <- estimate_h0(est_h1, calc.se=FALSE) # std.errors are not needed
compare_fit(est_h1 = est_h1, est_h0 = est_h0)

# Double centering approach
est_h1 <- modsem(m1, oneInt, method = "dblcent")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)

# Constrained approach
est_h1 <- modsem(m1, oneInt, method = "ca")
est_h0 <- estimate_h0(est_h1, oneInt)

compare_fit(est_h1 = est_h1, est_h0 = est_h0)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorize_output.R
\name{set_modsem_colors}
\alias{set_modsem_colors}
\title{Define or disable the color theme used by \code{modsem}}
\usage{
set_modsem_colors(
  positive = "green3",
  negative = positive,
  true = "green3",
  false = "red",
  nan = "purple",
  na = "purple",
  inf = "purple",
  string = "cyan",
  active = TRUE
)
}
\arguments{
\item{positive}{color of positive numbers.}

\item{negative}{color of negative numbers.}

\item{true}{color of \code{TRUE}.}

\item{false}{color of \code{FALSE}.}

\item{nan}{color of \code{NaN}.}

\item{na}{color of \code{NA}.}

\item{inf}{color of \code{-Inf} and \code{Inf}.}

\item{string}{color of quoted strings.}

\item{active}{Should color-theme be activated/deactived?}
}
\value{
\code{TRUE} if colors are active afterwards, otherwise \code{FALSE}.
}
\description{
All arguments are optional; omitted ones fall back to the defaults below.
Pass \code{active = FALSE} to turn highlighting off (and reset the palette).
}
\examples{
set_modsem_colors(positive = "red3",
                  negative = "red3",
                  true = "darkgreen",
                  false = "red3",
                  na = "purple",
                  string = "darkgreen")

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
# Inner Model
  Y ~ X + Z + X:Z
"

est <- modsem(m1, data = oneInt)
colorize_output(summary(est))
colorize_output(est) # same as colorize_output(print(est))
colorize_output(modsem_inspect(est, what = "coef"))

\dontrun{
colorize_output(split = TRUE, {
  # Get live (uncolored) output
  # And print colored output at the end of execution

  est_lms <- modsem(m1, data = oneInt, method = "lms")
  summary(est_lms)
})

colorize_output(modsem_inspect(est_lms))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parser.R
\name{modsemify}
\alias{modsemify}
\title{Generate parameter table for \code{lavaan} syntax}
\usage{
modsemify(syntax, parentheses.as.string = FALSE)
}
\arguments{
\item{syntax}{model syntax}

\item{parentheses.as.string}{Should parentheses be read parsed as literal strings?}
}
\value{
\code{data.frame} with columns \code{lhs, op, rhs, mod}
}
\description{
Generate parameter table for \code{lavaan} syntax
}
\examples{
library(modsem)
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
modsemify(m1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simple_slopes.R
\name{simple_slopes}
\alias{simple_slopes}
\title{Get the simple slopes of a SEM model}
\usage{
simple_slopes(
  x,
  z,
  y,
  model,
  vals_x = -3:3,
  vals_z = -1:1,
  rescale = TRUE,
  ci_width = 0.95,
  ci_type = "confidence",
  relative_h0 = TRUE,
  standardized = FALSE,
  xz = NULL,
  ...
)
}
\arguments{
\item{x}{The name of the variable on the x-axis (focal predictor).}

\item{z}{The name of the moderator variable.}

\item{y}{The name of the outcome variable.}

\item{model}{An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}},
\code{\link{modsem_mplus}}, or a \code{lavaan} object. This should be a fitted SEM
model that includes paths for \code{y ~ x + z + x:z}.}

\item{vals_x}{Numeric vector of values of \code{x} at which to compute predicted
slopes. Defaults to \code{-3:3}. If \code{rescale = TRUE}, these values are taken
relative to the mean and standard deviation of \code{x}. A higher density of points
(e.g., \code{seq(-3, 3, 0.1)}) will produce smoother curves and confidence bands.}

\item{vals_z}{Numeric vector of values of the moderator \code{z} at which to compute
predicted slopes. Defaults to \code{-1:1}. If \code{rescale = TRUE}, these values
are taken relative to the mean and standard deviation of \code{z}. Each unique value
of \code{z} generates a separate regression line \code{y ~ x | z}.}

\item{rescale}{Logical. If \code{TRUE} (default), \code{x} and \code{z} are standardized
according to their means and standard deviations in the model. The values in
\code{vals_x} and \code{vals_z} are interpreted in those standardized units. The
raw (unscaled) values corresponding to these standardized points will be displayed
in the returned table.}

\item{ci_width}{A numeric value between 0 and 1 indicating the confidence (or
prediction) interval width. The default is 0.95 (i.e., 95\% interval).}

\item{ci_type}{A string indicating whether to compute a \code{"confidence"} interval
for the predicted mean (\emph{i.e.}, uncertainty in the regression line) or a
\code{"prediction"} interval for individual outcomes. The default is
\code{"confidence"}. If \code{"prediction"}, the residual variance is added to the
variance of the fitted mean, resulting in a wider interval.}

\item{relative_h0}{Logical. If \code{TRUE} (default), hypothesis tests for the
predicted values (\code{predicted - h0}) assume \code{h0} is the model-estimated
mean of \code{y}. If \code{FALSE}, the null hypothesis is \code{h0 = 0}.}

\item{standardized}{Should coefficients be standardized beforehand?}

\item{xz}{The name of the interaction term (\code{x:z}). If \code{NULL}, it will
be created by combining \code{x} and \code{z} with a colon (e.g., \code{"x:z"}).
Some backends may remove or alter the colon symbol, so the function tries to
account for that internally.}

\item{...}{Additional arguments passed to lower-level functions or other internal
helpers.}
}
\value{
A \code{data.frame} (invisibly inheriting class \code{"simple_slopes"})
with columns:
\itemize{
  \item \code{vals_x}, \code{vals_z}: The requested grid values of \code{x} and \code{z}.
  \item \code{predicted}: The predicted value of \code{y} at that combination of
        \code{x} and \code{z}.
  \item \code{std.error}: The standard error of the predicted value.
  \item \code{z.value}, \code{p.value}: The z-statistic and corresponding p-value
        for testing the null hypothesis that \code{predicted == h0}.
  \item \code{ci.lower}, \code{ci.upper}: Lower and upper bounds of the confidence
        (or prediction) interval.
}
An attribute \code{"variable_names"} (list of \code{x}, \code{z}, \code{y})
is attached for convenience.
}
\description{
This function calculates simple slopes (predicted values of the outcome variable)
at user-specified values of the focal predictor (\code{x}) and moderator (\code{z})
in a structural equation modeling (SEM) framework. It supports interaction terms
(\code{xz}), computes standard errors (SE), and optionally returns confidence or
prediction intervals for these predicted values. It also provides p-values for
hypothesis testing. This is useful for visualizing and interpreting moderation
effects or to see how the slope changes at different values of the moderator.
}
\details{
\strong{Computation Steps}
\enumerate{
   \item The function extracts parameter estimates (and, if necessary, their covariance
         matrix) from the fitted SEM model (\code{model}).
   \item It identifies the coefficients for \code{x}, \code{z}, and \code{x:z} in the model's
         parameter table, as well as the variance of \code{x}, \code{z} and the residual variance
         of \code{y}.
   \item If \code{xz} is not provided, it will be constructed by combining \code{x} and
         \code{z} with a colon (\code{":"}). Depending on the approach used to estimate the model,
         the colon may be removed or replaced internally; the function attempts to reconcile that.

  \item A grid of \code{x} and \code{z} values is created from \code{vals_x} and
        \code{vals_z}. If \code{rescale = TRUE}, these values are transformed back into raw
        metric units for display in the output.
  \item For each point in the grid, a predicted value of \code{y} is computed via
        \code{(beta0 + beta_x * x + beta_z * z + beta_xz * x * z)} and, if included, a
        mean offset.
  \item The standard error (\code{std.error}) is derived from the covariance matrix of
        the relevant parameters, and if \code{ci_type = "prediction"}, adds the residual
        variance.
  \item Confidence (or prediction) intervals are formed using \code{ci_width} (defaulting
        to 95\%). The result is a table-like data frame with predicted values, CIs,
        standard errors, z-values, and p-values.
}
}
\examples{
\dontrun{
library(modsem)

m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)

# Simple slopes at X in [-3, 3] and Z in [-1, 1], rescaled to the raw metric.
simple_slopes(x = "X", z = "Z", y = "Y", model = est1)

# If the data or user wants unscaled values, set rescale = FALSE, etc.
simple_slopes(x = "X", z = "Z", y = "Y", model = est1, rescale = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{TPB_UK}
\alias{TPB_UK}
\title{TPB_UK}
\source{
Gathered from a replication study by Hagger et al. (2023).

Obtained from \doi{10.23668/psycharchives.12187}
}
\description{
A dataset based on the Theory of Planned Behaviour from a
UK sample. 4 variables with high communality were selected for each
latent variable (ATT, SN, PBC, INT, BEH), from two time points (t1 and t2).
}
\examples{

tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 # Causal Relationsships
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

est <- modsem(tpb_uk, data = TPB_UK)
summary(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reliablity_single_item.R
\name{relcorr_single_item}
\alias{relcorr_single_item}
\title{Reliability‑Corrected Single‑Item SEM}
\usage{
relcorr_single_item(
  syntax,
  data,
  choose = NULL,
  scale.corrected = TRUE,
  missing = "fiml",
  warn.lav = TRUE,
  group = NULL
)
}
\arguments{
\item{syntax}{A character string containing lavaan model syntax.  Must at
least include the measurement relations (\code{=~}).}

\item{data}{A \code{data.frame}, \code{tibble} or object coercible to a data frame that
holds the raw observed indicators.}

\item{choose}{\emph{Optional.} Character vector with the names of the latent
variables you wish to replace by single indicators.  Defaults to \strong{all}
first‑order latent variables in \code{syntax}.}

\item{scale.corrected}{Should reliability-corrected items be scale-corrected? If \code{TRUE}
reliability-corrected single items are corrected for differences in factor loadings between
the items. Default is \code{TRUE}.}

\item{missing}{How should missing values be handled in the CFA? Default is \code{"fiml"}.}

\item{warn.lav}{Should warnings from \code{lavaan::cfa} be displayed? If \code{FALSE}, they
are suppressed.}

\item{group}{Character. A variable name in the data frame defining the groups in a multiple
group analysis}
}
\value{
An object of S3 class \code{ModsemRelcorr} (a named list) with elements:
\describe{
  \item{\code{syntax}}{Modified lavaan syntax string.}
  \item{\code{data}}{Data frame with additional composite indicator columns.}
  \item{\code{parTable}}{Parameter table corresponding to `syntax`.}
  \item{\code{reliability}}{Named numeric vector of reliabilities (one per latent
    variable).}
  \item{\code{AVE}}{Named numeric vector with Average Variance Extracted values.}
  \item{\code{lVs}}{Character vector of latent variables that were corrected.}
  \item{\code{single.items}}{Character vector with names for the generated reliability corrected items}
}
}
\description{
Replace (some of) the first‑order latent variables in a lavaan measurement
model by single composite indicators whose error variances are fixed from
Cronbach's \eqn{\alpha}. The function returns a modified lavaan model
syntax together with an augmented data set that contains the newly created
composite variables, so that you can fit the full SEM in a single step.
}
\details{
The resulting object can be fed directly into \code{modsem} or \code{lavaan::sem}
by supplying \code{syntax = $syntax} and \code{data = $data}.
}
\examples{
\dontrun{
tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

corrected <- relcorr_single_item(syntax = tpb_uk, data = TPB_UK)
print(corrected)

syntax <- corrected$syntax
data   <- corrected$data

est_dca <- modsem(syntax, data = data, method = "dblcent")
est_lms <- modsem(syntax, data = data, method="lms", nodes=32)
summary(est_lms)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/generics_modsem_pi.R,
%   R/predict_da.R
\name{modsem_predict}
\alias{modsem_predict}
\alias{modsem_predict.modsem_pi}
\alias{modsem_predict.modsem_da}
\title{Predict From \code{modsem} Models}
\usage{
modsem_predict(object, ...)

\method{modsem_predict}{modsem_pi}(object, ...)

\method{modsem_predict}{modsem_da}(
  object,
  newdata = NULL,
  method = c("EBM", "ML", "Bartlett", "Regression"),
  type = c("lv", "ov", "all"),
  standardized = FALSE,
  se = FALSE,
  drop.list.single.group = TRUE,
  ...
)
}
\arguments{
\item{object}{An object of class \code{modsem_pi} or \code{modsem_da},
respectively.}

\item{...}{Further arguments passed to \code{lavaan::predict} for
\code{\link{modsem_pi}} objects; currently ignored by the
\code{\link{modsem_da}} method.}

\item{newdata}{Optional \code{data.frame} or matrix. If supplied, factor
scores are computed for these observations instead of the data used during
model estimation. Must contain all observed-variable columns required by
the model; missing values are handled via FIML. Currently only supported
by the \code{\link{modsem_da}} method.}

\item{method}{Character. Scoring method. One of \code{"EBM"} (Empirical
Bayes Modal; MAP estimate of the posterior), \code{"ML"} (maximum
likelihood), \code{"Regression"} (alias for \code{"EBM"}), or
\code{"Bartlett"} (alias for \code{"ML"}). For \code{\link{modsem_da}}
objects this selects the optimisation-based scoring algorithm. For
\code{\link{modsem_pi}} objects the argument is forwarded to
\code{lavaan::lavPredict}, which accepts the same \code{method} values.}

\item{type}{Character. Which scores to return: \code{"lv"} for latent
variable scores only (default), \code{"ov"} for model-implied observed
variable scores, or \code{"all"} for both combined. For
\code{\link{modsem_da}} objects this is handled directly; for
\code{\link{modsem_pi}} objects it is forwarded to
\code{lavaan::lavPredict}, which supports the same \code{type} values.}

\item{standardized}{Logical. If \code{TRUE}, each score column is
standardised (mean 0, SD 1) before returning. Currently only used by the
\code{\link{modsem_da}} method.}

\item{se}{Logical. If \code{TRUE}, attach standard-error and asymptotic
covariance attributes for \code{\link{modsem_da}} predictions. Standard
errors are returned in the \code{"se"} attribute and covariance matrices in
the \code{"acov"} attribute.}

\item{drop.list.single.group}{Logical. If \code{TRUE}, covariance attributes
for single-group \code{\link{modsem_da}} predictions drop the outer group
list, matching \code{lavaan::lavPredict}.}
}
\value{
* For \code{\link{modsem_pi}}: whatever \code{lavaan::predict()} returns,
  which is usually a matrix of factor scores.
* For \code{\link{modsem_da}}: a numeric matrix with one row per
  observation. Columns depend on \code{type}: latent variable scores
  (\code{"lv"}), model-implied observed-variable scores (\code{"ov"}), or
  both (\code{"all"}). Columns are optionally standardised if
  \code{standardized = TRUE}.
}
\description{
A generic function (and corresponding methods) that produces predicted
values or factor scores from \code{\link{modsem}} models.
}
\section{Methods (by class)}{
\itemize{
\item \code{modsem_predict(modsem_pi)}: Wrapper for \code{lavaan::predict}

\item \code{modsem_predict(modsem_da)}: Computes factor scores or model-implied observed-variable scores for a
\code{\link{modsem_da}} model via MAP optimisation.

}}
\examples{
m1 <- '
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
'

est_dca <- modsem(m1, oneInt, method = "dblcent")
head(modsem_predict(est_dca))

\dontrun{
est_lms <- modsem(m1, oneInt, method = "lms")
head(modsem_predict(est_lms))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twostep.R
\name{twostep}
\alias{twostep}
\title{Estimate interaction effects in structural equation models (SEMs) using a twostep procedure}
\usage{
twostep(model.syntax, data, method = "lms", ...)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{data}{dataframe}

\item{method}{method to use:
\describe{
  \item{\code{"dblcent"}}{double centering approach (passed to \code{lavaan}).}
  \item{\code{"ca"}}{constrained approach (passed to \code{lavaan}).}
  \item{\code{"rca"}}{residual centering approach (passed to \code{lavaan}).}
  \item{\code{"uca"}}{unconstrained approach (passed to \code{lavaan}).}
  \item{\code{"pind"}}{prod ind approach, with no constraints or centering (passed to \code{lavaan}).}
  \item{\code{"lms"}}{latent moderated structural equations (not passed to \code{lavaan}).}
  \item{\code{"qml"}}{quasi maximum likelihood estimation (not passed to \code{lavaan}).}
  \item{\code{"custom"}}{use parameters specified in the function call (passed to \code{lavaan}).}
}}

\item{...}{arguments passed to other functions depending on the method (see \code{\link{modsem_pi}} and \code{\link{modsem_da}})}
}
\value{
\code{modsem} object with class \code{\link{modsem_pi}} or \code{\link{modsem_da}}.
}
\description{
Estimate an interaction model using a twostep procedure. For the PI approaches, the \code{lavaan::sam} function
  is used to optimize the models, instead of \code{lavaan::sem}. Note that the product indicators are still used,
  and not the newly developed SAM approach to estimate latent interactions. For the DA approaches (LMS and QML)
  the measurement model is estimated using a CFA (\code{lavaan::cfa}). The structural model is estimated using
  \code{\link{modsem_da}}, where the estimates in the measurement model are fixed, based on the CFA estimates.
  Note that standard errors are uncorrected (i.e., naive), and do not account for the uncertainty in the CFA estimates.
  \strong{NOTE, this is an experimental feature!}
}
\examples{
library(modsem)
m1 <- '
  # Outer Model
  X =~ x1 + x2 +x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'

est_dblcent <- twostep(m1, oneInt, method = "dblcent")
summary(est_dblcent)

\dontrun{
est_lms <- twostep(m1, oneInt, method = "lms")
summary(est_lms)

est_qml <- twostep(m1, oneInt, method = "qml")
summary(est_qml)
}

tpb_uk <- "
# Outer Model (Based on Hagger et al., 2007)
 ATT =~ att3 + att2 + att1 + att4
 SN =~ sn4 + sn2 + sn3 + sn1
 PBC =~ pbc2 + pbc1 + pbc3 + pbc4
 INT =~ int2 + int1 + int3 + int4
 BEH =~ beh3 + beh2 + beh1 + beh4

# Inner Model (Based on Steinmetz et al., 2011)
 # Causal Relationsships
 INT ~ ATT + SN + PBC
 BEH ~ INT + PBC
 BEH ~ INT:PBC
"

uk_dblcent <- twostep(tpb_uk, TPB_UK, method = "dblcent")
summary(uk_dblcent)

\dontrun{
  uk_qml <- twostep(tpb_uk, TPB_UK, method = "qml")

  uk_lms <- twostep(tpb_uk, TPB_UK, method = "lms", nodes = 32, adaptive.quad = TRUE)
  summary(uk_lms)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_jn.R
\name{plot_jn}
\alias{plot_jn}
\title{Plot Interaction Effect Using the Johnson-Neyman Technique}
\usage{
plot_jn(
  x,
  z,
  y,
  model,
  min_z = -3,
  max_z = 3,
  sig.level = 0.05,
  alpha = 0.2,
  detail = 1000,
  sd.line = 2,
  standardized = FALSE,
  xz = NULL,
  greyscale = FALSE,
  plot.jn.points = TRUE,
  type = c("direct", "indirect", "total"),
  mc.quantiles = FALSE,
  mc.reps = 10000,
  ...
)
}
\arguments{
\item{x}{The name of the predictor variable (as a character string).}

\item{z}{The name of the moderator variable (as a character string).}

\item{y}{The name of the outcome variable (as a character string).}

\item{model}{A fitted model object of class \code{modsem_da}, \code{modsem_mplus}, \code{modsem_pi}, or \code{lavaan}.}

\item{min_z}{The minimum value of the moderator variable \code{z} to be used in the plot (default is -3). It is relative to the mean of z.}

\item{max_z}{The maximum value of the moderator variable \code{z} to be used in the plot (default is 3). It is relative to the mean of z.}

\item{sig.level}{The alpha-criterion for the confidence intervals (default is 0.05).}

\item{alpha}{alpha setting used in \code{ggplot} (i.e., the opposite of opacity)}

\item{detail}{The number of generated data points to use for the plot (default is 1000). You can increase this value for smoother plots.}

\item{sd.line}{A thick black line showing \code{+/- sd.line * sd(z)}. NOTE: This line will be truncated by \code{min_z} and \code{max_z} if
the sd.line falls outside of \code{[min_z, max_z]}.}

\item{standardized}{Should coefficients be standardized beforehand?}

\item{xz}{The name of the interaction term. If not specified, it will be created using \code{x} and \code{z}.}

\item{greyscale}{Logical. If \code{TRUE} the plot is plotted in greyscale.}

\item{plot.jn.points}{Logical. If \code{TRUE}, omit the numeric annotations for the JN-points from the plot.}

\item{type}{Which effect to display. One of \code{"direct"}, \code{"indirect"}, or \code{"total"}.}

\item{mc.quantiles}{Should JN quantiles be calculated using Monte-Carlo estimates?}

\item{mc.reps}{Number of Monte Carlo replicates used to approximate the confidence
bands when \code{mc.quantile} is \code{TRUE} and/or \code{type} is \code{"indirect"} or \code{"total"}.}

\item{...}{Additional arguments (currently not used).}
}
\value{
A \code{ggplot} object showing the interaction plot with regions of significance.
}
\description{
This function plots the simple slopes of an interaction effect across different values of a moderator variable using the Johnson-Neyman technique. It identifies regions where the effect of the predictor on the outcome is statistically significant.
}
\details{
The function calculates the simple slopes of the predictor variable \code{x} on the outcome variable \code{y} at different levels of the moderator variable \code{z}. It uses the Johnson-Neyman technique to identify the regions of \code{z} where the effect of \code{x} on \code{y} is statistically significant.

When plotting indirect or total effects, the function relies on Monte Carlo
draws from the estimated sampling distribution of the parameters to approximate
the conditional effect and its confidence interval across the moderator range.

It extracts the necessary coefficients and variance-covariance information from the fitted model object. The function then computes the critical t-value and solves the quadratic equation derived from the t-statistic equation to find the Johnson-Neyman points.

The plot displays:
\itemize{
  \item The estimated simple slopes across the range of \code{z}.
  \item Confidence intervals around the slopes.
  \item Regions where the effect is significant (shaded areas).
  \item Vertical dashed lines indicating the Johnson-Neyman points.
  \item Text annotations providing the exact values of the Johnson-Neyman points.
}
}
\examples{
\dontrun{
library(modsem)

m1 <-  '
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9

  visual ~ speed + textual + speed:textual
'

est <- modsem(m1, data = lavaan::HolzingerSwineford1939, method = "ca")
plot_jn(x = "speed", z = "textual", y = "visual", model = est, max_z = 6)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootstrap.R
\name{bootstrap_modsem}
\alias{bootstrap_modsem}
\alias{bootstrap_modsem.modsem_pi}
\alias{bootstrap_modsem.modsem_da}
\alias{bootstrap_modsem.function}
\title{Bootstrap a modsem Model}
\usage{
bootstrap_modsem(model = modsem, FUN, ...)

\method{bootstrap_modsem}{modsem_pi}(model, FUN = "coef", ...)

\method{bootstrap_modsem}{modsem_da}(
  model,
  FUN = "coef",
  R = 1000L,
  P.max = 1e+05,
  type = c("nonparametric", "parametric"),
  verbose = interactive(),
  calc.se = FALSE,
  optimize = FALSE,
  convergence.abs = 10 * model$args$convergence.abs,
  convergence.rel = 10 * model$args$convergence.rel,
  algorithm = "EM",
  ...
)

\method{bootstrap_modsem}{`function`}(
  model = modsem,
  FUN = "coef",
  data,
  R = 1000L,
  verbose = interactive(),
  FUN.args = list(),
  cluster.boot = NULL,
  ...
)
}
\arguments{
\item{model}{A fitted \code{modsem} object, or a function to be bootstrapped
(e.g., \code{\link{modsem}}, \code{\link{modsem_da}} and \code{\link{modsem_pi}})}

\item{FUN}{A function that returns the statistic of interest when applied to
a fitted model.  The function must accept a single argument, the model
object, and should ideally return a numeric vector; see Value.}

\item{...}{Additional arguments forwarded to \code{lavaan::bootstrapLavaan}
for \code{\link{modsem_pi}} objects, or \code{\link{modsem_da}} for
\code{\link{modsem_da}} objects.}

\item{R}{number of bootstrap replicates.}

\item{P.max}{ceiling for the simulated population size.}

\item{type}{Bootstrap flavour, see Details.}

\item{verbose}{Should progress information be printed to the console?}

\item{calc.se}{Should standard errors for each replicate. Defaults to \code{FALSE}.}

\item{optimize}{Should starting values be re-optimized for each replicate. Defaults to \code{FALSE}.}

\item{convergence.abs}{Absolute convergence criteria. Defaults to 10 times the original criteria.}

\item{convergence.rel}{Relative convergence criteria. Defaults to 10 times the original criteria.}

\item{algorithm}{Which algorithm should be used for the LMS approach? Defaults to \code{"EM"}, as
opposed to \code{"EMA"}.}

\item{data}{Dataset to be resampled.}

\item{FUN.args}{Arguments passed to \code{FUN}}

\item{cluster.boot}{Variable to cluster bootstrapping by}
}
\value{
Depending on the return type of \code{FUN} either
  \describe{
    \item{numeric}{A matrix with \code{R} rows (bootstrap replicates) and as
      many columns as \code{length(FUN(model))}.}
    \item{other}{A list of length \code{R}; each element is the raw output of
      \code{FUN}. \strong{NOTE}: Only applies for \code{\link{modsem_da}} objects}
  }
}
\description{
A generic interface for parametric and non-parametric bootstrap procedures
for structural equation models estimated with the \pkg{modsem} ecosystem.
The function dispatches on the class of \code{model}; currently dedicated
methods exist for \code{\link{modsem_pi}} (product‑indicator approach) and
\code{\link{modsem_da}} (distributional‑analytic approach).
}
\details{
A thin wrapper around \code{lavaan::bootstrapLavaan()} that performs the
  necessary book‑keeping so that \code{FUN} receives a fully‑featured
  \code{modsem_pi} object—rather than a bare \code{lavaan} fit—at every
  iteration.

The function internally resamples the observed data (non-parametric
  case) or simulates from the estimated parameter table (parametric case),
  feeds the sample to \code{\link{modsem_da}}, evaluates \code{FUN} on the
  refitted object and finally collates the results.

This is a more general version of \code{boostrap_modsem} for
  bootstrapping \code{modsem} functions, not modsem objects.
  \code{model} is now a function to be boostrapped, and \code{...}
  are now passed to the function (\code{model}), not \code{FUN}. To
  pass arguments to \code{FUN} use \code{FUN.args}.
}
\section{Methods (by class)}{
\itemize{
\item \code{bootstrap_modsem(modsem_pi)}: Bootstrap a \code{modsem_pi} model by delegating
to \code{\link[lavaan]{bootstrapLavaan}}.

\item \code{bootstrap_modsem(modsem_da)}: Parametric or non-parametric bootstrap for
\code{modsem_da} models.

\item \code{bootstrap_modsem(`function`)}: Non-parametric bootstrap of \code{modsem} functions

}}
\examples{

m1 <- '
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

  Y ~ X + Z + X:Z
'

fit_pi <- modsem(m1, oneInt)
bootstrap_modsem(fit_pi, FUN = coef, R = 10L)


m1 <- '
  X =~ x1 + x2
  Z =~ z1 + z2
  Y =~ y1 + y2

  Y ~ X + Z + X:Z
'

\dontrun{
fit_lms <- modsem(m1, oneInt, method = "lms")
bootstrap_modsem(fit_lms, FUN = coef, R = 10L)
}

tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

\dontrun{
boot <- bootstrap_modsem(model = modsem,
                         model.syntax = tpb, data = TPB,
                         method = "dblcent", rcs = TRUE,
                         rcs.scale.corrected = TRUE,
                         FUN = "coef", R = 50L)
coef <- apply(boot, MARGIN = 2, FUN = mean, na.rm = TRUE)
se   <- apply(boot, MARGIN = 2, FUN = sd, na.rm = TRUE)

cat("Parameter Estimates:\n")
print(coef)

cat("Standard Errors: \n")
print(se)

}
}
\seealso{
\code{\link[lavaan]{bootstrapLavaan}}, \code{\link{modsem_pi}},
  \code{\link{modsem_da}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_interaction.R
\name{plot_interaction}
\alias{plot_interaction}
\title{Plot Interaction Effects in a SEM Model}
\usage{
plot_interaction(
  x,
  z,
  y,
  model,
  vals_x = seq(-3, 3, 0.001),
  vals_z,
  alpha_se = 0.15,
  digits = 2,
  ci_width = 0.95,
  ci_type = "confidence",
  rescale = TRUE,
  standardized = FALSE,
  xz = NULL,
  greyscale = FALSE,
  ...
)
}
\arguments{
\item{x}{A character string specifying the focal predictor (x-axis variable).}

\item{z}{A character string specifying the moderator variable.}

\item{y}{A character string specifying the outcome (dependent) variable.}

\item{model}{An object of class \code{\link{modsem_pi}}, \code{\link{modsem_da}},
\code{\link{modsem_mplus}}, or possibly a \code{lavaan} object. Must be a fitted
SEM model containing paths for \code{y ~ x + z + x:z}.}

\item{vals_x}{A numeric vector of values at which to compute and plot the focal
predictor \code{x}. The default is \code{seq(-3, 3, .001)}, which provides a
relatively fine grid for smooth lines. If \code{rescale=TRUE}, these values
are in standardized (mean-centered and scaled) units, and will be converted back
to the original metric in the internal computation of predicted means.}

\item{vals_z}{A numeric vector of values of the moderator \code{z} at which to draw
separate regression lines. Each distinct value in \code{vals_z} defines a separate
group (plotted with a different color). If \code{rescale=TRUE}, these values
are also assumed to be in standardized units.}

\item{alpha_se}{A numeric value in \eqn{[0, 1]} specifying the transparency of
the confidence/prediction interval ribbon. Default is \code{0.15}.}

\item{digits}{An integer specifying the number of decimal places to which the
moderator values (\code{z}) are rounded for labeling/grouping in the plot.}

\item{ci_width}{A numeric value in \eqn{(0,1)} indicating the coverage of the
confidence (or prediction) interval. The default is \code{0.95} for a 95\%
interval.}

\item{ci_type}{A character string specifying whether to compute
\code{"confidence"} intervals (for the mean of the predicted values, default)
or \code{"prediction"} intervals (which include residual variance).}

\item{rescale}{Logical. If \code{TRUE} (default), \code{vals_x} and \code{vals_z}
are interpreted as standardized units, which are mapped back to the raw scale
before computing predictions. If \code{FALSE}, \code{vals_x} and \code{vals_z}
are taken as raw-scale values directly.}

\item{standardized}{Should coefficients be standardized beforehand?}

\item{xz}{A character string specifying the interaction term (\code{x:z}).
If \code{NULL}, the term is created automatically as \code{paste(x, z, sep = ":")}.
Some SEM backends may handle the interaction term differently (for instance, by
removing or modifying the colon), and this function attempts to reconcile that
internally.}

\item{greyscale}{Logical. If \code{TRUE} the plot is plotted in greyscale.}

\item{...}{Additional arguments passed on to \code{\link{simple_slopes}}.}
}
\value{
A \code{ggplot} object that can be further customized (e.g., with
  additional \code{+ theme(...)} layers). By default, it shows lines for each
  moderator value and a shaded region corresponding to the interval type
  (confidence or prediction).
}
\description{
This function creates an interaction plot of the outcome variable (\code{y}) as a function
of a focal predictor (\code{x}) at multiple values of a moderator (\code{z}). It is
designed for use with structural equation modeling (SEM) objects (e.g., those from
\code{\link{modsem}}). Predicted means (or predicted individual values) are calculated
via \code{\link{simple_slopes}}, and then plotted with \code{ggplot2} to display
multiple regression lines and confidence/prediction bands.
}
\details{
\strong{Computation Steps:}
\enumerate{
  \item Calls \code{\link{simple_slopes}} to compute the predicted values of \code{y}
        at the specified grid of \code{x} and \code{z} values.
  \item Groups the resulting predictions by unique \code{z}-values (rounded to
        \code{digits}) to create colored lines.
  \item Plots these lines using \code{ggplot2}, adding ribbons for confidence
        (or prediction) intervals, with transparency controlled by \code{alpha_se}.
}

\strong{Interpretation:}
Each line in the plot corresponds to the regression of \code{y} on \code{x} at
a given level of \code{z}. The shaded region around each line (ribbon) shows
either the confidence interval for the predicted mean (if \code{ci_type =
"confidence"}) or the prediction interval for individual observations (if
\code{ci_type = "prediction"}). Where the ribbons do not overlap, there is
evidence that the lines differ significantly over that range of \code{x}.
}
\examples{
\dontrun{
library(modsem)

# Example 1: Interaction of X and Z in a simple SEM
m1 <- "
# Outer Model
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3

# Inner Model
  Y ~ X + Z + X:Z
"
est1 <- modsem(m1, data = oneInt)

# Plot interaction using a moderate range of X and two values of Z
plot_interaction(x = "X", z = "Z", y = "Y", xz = "X:Z",
                 vals_x = -3:3, vals_z = c(-0.2, 0), model = est1)

# Example 2: Interaction in a theory-of-planned-behavior-style model
tpb <- "
# Outer Model
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN  =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC
  BEH ~ PBC:INT
"
est2 <- modsem(tpb, data = TPB, method = "lms", nodes = 32)

# Plot with custom Z values and a denser X grid
plot_interaction(x = "INT", z = "PBC", y = "BEH",
                 xz = "PBC:INT",
                 vals_x = seq(-3, 3, 0.01),
                 vals_z = c(-0.5, 0.5),
                 model = est2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{TPB_1SO}
\alias{TPB_1SO}
\title{TPB_1SO}
\description{
A simulated dataset based on the Theory of Planned Behaviour,
where INT is a higher order construct of ATT, SN, and PBC.
}
\examples{
tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PBC =~ pbc1 + pbc2 + pbc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + PBC + SN

  # Higher order interaction
  INTxPBC =~ ATT:PBC + SN:PBC + PBC:PBC

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

\dontrun{
est_ca <- modsem(tpb, data = TPB_1SO, method = "ca")
summary(est_ca)

est_dblcent  <- modsem(tpb, data = TPB_1SO, method = "dblcent")
summary(est_dblcent)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{TPB_2SO}
\alias{TPB_2SO}
\title{TPB_2SO}
\description{
A simulated dataset based on the Theory of Planned Behaviour,
where INT is a higher order construct of ATT and SN, and PBC is a higher order
construct of PC and PB.
}
\examples{
tpb <- '
  # First order constructs
  ATT =~ att1 + att2 + att3
  SN  =~ sn1 + sn2 + sn3
  PB =~ pb1 + pb2 + pb3
  PC =~ pc1 + pc2 + pc3
  BEH =~ b1 + b2

  # Higher order constructs
  INT =~ ATT + SN
  PBC =~ PC + PB

  # Higher order interaction
  INTxPBC =~ ATT:PC + ATT:PB + SN:PC + SN:PB

  # Structural model
  BEH ~ PBC + INT + INTxPBC
'

\dontrun{
est <- modsem(tpb, data = TPB_2SO, method = "ca")
summary(est)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{TPB}
\alias{TPB}
\title{TPB}
\description{
A simulated dataset based on the Theory of Planned Behaviour
}
\examples{

tpb <- "
# Outer Model (Based on Hagger et al., 2007)
  ATT =~ att1 + att2 + att3 + att4 + att5
  SN =~ sn1 + sn2
  PBC =~ pbc1 + pbc2 + pbc3
  INT =~ int1 + int2 + int3
  BEH =~ b1 + b2

# Inner Model (Based on Steinmetz et al., 2011)
  INT ~ ATT + SN + PBC
  BEH ~ INT + PBC + INT:PBC
"

est <- modsem(tpb, data = TPB)
summary(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modsem_pi.R
\name{get_pi_syntax}
\alias{get_pi_syntax}
\title{Get \code{lavaan} syntax for product indicator approaches}
\usage{
get_pi_syntax(
  model.syntax,
  method = "dblcent",
  match = FALSE,
  data = NULL,
  ...
)
}
\arguments{
\item{model.syntax}{\code{lavaan} syntax}

\item{method}{method to use:
\code{"rca"} = residual centering approach,
\code{"uca"} = unconstrained approach,
\code{"dblcent"} = double centering approach,
\code{"pind"} = prod ind approach, with no constraints or centering,
\code{"custom"} = use parameters specified in the function call}

\item{match}{should the product indicators be created by using the match-strategy}

\item{data}{Optional. Dataset to use, usually not relevant.}

\item{...}{arguments passed to other functions (e.g., \link{modsem_pi})}
}
\value{
\code{character} vector
}
\description{
\code{get_pi_syntax()} is a function for creating the \code{lavaan} syntax used for estimating
latent interaction models using one of the product indicator approaches.
}
\examples{
library(modsem)
library(lavaan)
m1 <- '
  # Outer Model
  X =~ x1 + x2 + x3
  Y =~ y1 + y2 + y3
  Z =~ z1 + z2 + z3

  # Inner model
  Y ~ X + Z + X:Z
'
syntax <- get_pi_syntax(m1)
data <- get_pi_data(m1, oneInt)
est <- sem(syntax, data)
summary(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_modsem_da.R
\name{fit_modsem_da}
\alias{fit_modsem_da}
\title{Fit measures for QML and LMS models}
\usage{
fit_modsem_da(
  model,
  chisq = TRUE,
  lav.fit = FALSE,
  drop.list.single.group = TRUE
)
}
\arguments{
\item{model}{fitted model. Thereafter, you can use 'compare_fit()'
to assess the comparative fit of the models. If the interaction effect makes
the model better, and e.g., the RMSEA is good for the baseline model,
the interaction model likely has a good RMSEA as well.}

\item{chisq}{should Chi-Square based fit-measures be calculated?}

\item{lav.fit}{Should fit indices from the \code{lavaan} model used to optimize
the starting parameters be included (if available)? This is usually only approprioate
for linear models (i.e., no interaction effects), where the parameter estimates
for LMS and QML are equivalent to ML estimates from lavaan.}

\item{drop.list.single.group}{Logical. If FALSE, (some) results are returned as
a list, where each element corresponds to a group (even if there is only a single group).
If TRUE, the list will be unlisted if there is only a single group}
}
\description{
Calculates chi-sq test and p-value, as well as RMSEA for
the LMS and QML models. Note that the Chi-Square based fit measures should be calculated
for the baseline model, i.e., the model without the interaction effect
}
